#!/bin/sh

#  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
#
#  This file is part of Lightweight Bartering Grid.
#  Lightweight Bartering Grid is free software;
#  you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  Lightweight Bartering Grid is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Lightweight Bartering Grid; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Authors:      Cyril Briquet, Xavier Dalem
# Last revised: October 29th, 2007

# ----------------------------------------------------------------

# Need to export it : we'll use other scripts which need it
export CLUSTER_DEPLOYMENT=true
export LBG_DEP_DIR=/tmp/.lbg/
. lbg_common.sh
SEP=':'
DATE_FORMAT='%x %X'
# Copy of config dir, we'll override the SRC_CONFIG_DIR later to load the files
# from the correct place
BASE_CONFIG_DIR=$SRC_CONFIG_DIR
FTP_CMD="cd /tmp/batch_user_storage; ~/usr/vsftpd-2.0.5/vsftpd ~/usr/vsftpd-2.0.5/vsftpd_grid.conf > /dev/null 2>&1 &"
LOCAL_LBG_JAR="../../jars/lightweight_bartering_grid.jar"

function edfValidation() {
	if [ -z "$experiment_id" ]; then
		echo "Error (EDF): empty experiment id"
		return 1
	elif [ -z "$experiment_dir" ]; then
		echo "Warning (EDF): empty experiment directory"
	elif [ -z "$jobs" ]; then
		echo "Error (EDF): "
		return 1
	elif [ -z "$tasks_per_job" ]; then
		echo "Error (EDF): "
		return 1
	elif [ -z "$ddr" ]; then
		echo "Error (EDF): "
		return 1
	elif [ -z "$data_size" ]; then
		echo "Error (EDF): "
		return 1
	elif [ -z "$data_dir" ]; then
		echo "Error (EDF): "
		return 1
	elif [ -z "$data_per_task" ]; then
		echo "Error (EDF): "
		return 1
	elif [ -z "$search_engine" ]; then
		echo "Error (EDF): "
		return 1
	elif [ -z "$consumer_peer" ]; then
		echo "Error (EDF): "
		return 1
	elif [ -z "$consumer_user" ]; then
		echo "Error (EDF): "
		return 1
	elif [ -z "$user_port" ]; then
		echo "Error (EDF): "
		return 1
	elif [ -z "$ftp_port" ]; then
		echo "Error (EDF): "
		return 1
	elif [ -z "$CLUSTER_LIST" ]; then
		echo "Error (EDF): "
		return 1
	fi
	return 0
}

function handleExperiment() {
	# Load cluster list, don't add consumer peer yet
	CLUSTER_LIST=`awk \
		'BEGIN {clusters=0} \
		{if (clusters) print $0}
		/^clusters[\t ]*'$SEP'/ {clusters=1}' < $1`

	# Load other variables, custom, user-friendly format didn't allow shell
	# syntax but this it's close enough to be translated
	eval `cat $1 | grep -E "^[a-z_]*.[[:space:]]*:" | grep -v clusters | sed 's/\s*:\s*/=/'`

	# Only start reporting when we have the experiment id
	echo "Starting experiment $experiment_id at "`date +"$DATE_FORMAT"`

	# Validate input
	if ! edfValidation; then
		echo "Error: Invalid EDF file $1"
		echo "Aborted experiment $experiment_id at "`date +"$DATE_FORMAT"`
		echo "---"
		return
	fi

	if (echo $CLUSTER_LIST | grep $search_engine > /dev/null); then
		echo "Warning: search engine ($search_engine) is in cluster list"
	fi

	if (echo $CLUSTER_LIST | grep $consumer_peer > /dev/null); then
		echo "Error: consumer peer ($consumer_peer) is in cluster list"
		echo "Aborted experiment $experiment_id at "`date +"$DATE_FORMAT"`
		echo "---"
		return
	fi

	# Set the config directory to current experiment
	# NOTE: this has the desired side-effect of uploading the config files from
	#       the right place when deploying. See uploadDeps function
	export SRC_CONFIG_DIR=$BASE_CONFIG_DIR/$experiment_dir

	# Load some variables from config files
	PEER_USER_PORT=`grep USER_PORT $SRC_CONFIG_DIR/peerconfig.properties | cut -d '=' -f 2-`
	PEER_CACHE_DIR=`grep CACHE_DIR $SRC_CONFIG_DIR/peerconfig.properties | cut -d '=' -f 2-`
	PEER_CACHE_DIR=`dirname $PEER_CACHE_DIR/foo` # Strip trailing slash
	RES_CACHE_DIR=`grep CACHE_DIR $SRC_CONFIG_DIR/resourceconfig.properties | cut -d '=' -f 2-`
	RES_CACHE_DIR=`dirname $RES_CACHE_DIR/foo` # Strip trailing slash

	# Clean
	echo -e "\n---\nShutting down grid..."
	echo -e "$CLUSTER_LIST\n$consumer_peer\n$consumer_user" | ./lbg_shutdown_grid.sh $search_engine
	# Kill FTP, send sigkill: vsftp isn't always reactive enough
	$SSH_BIN $consumer_peer "killall -9 vsftpd 2> /dev/null" > /dev/null 2>&1

	# Prepare consumer peer storage, /tmp is too small
	# Warning: highly dependent on consumer peer filesystem
	$SSH_BIN $consumer_peer "ln -nsf /usr/local/grid_consumer_storage $PEER_CACHE_DIR"

	# Deploy, -e needed to keep new lines in CLUSTER_LIST
	echo -e "\n---\nDeploying grid..."
	echo -e "$CLUSTER_LIST\n$consumer_peer" | ./lbg_deploy_grid.sh $search_engine

	# Launch user FTP server. Requires the test data to be already present
	echo "Launching consumer FTP server ($consumer_peer)"
	$SSH_BIN $consumer_peer $FTP_CMD

	# Get some useful hostnames
	extractHostname $consumer_user
	consumer_user_h=$EXTRACT_PEER_HOSTNAME_RETVAL
	extractHostname $consumer_peer
	consumer_peer_h=$EXTRACT_PEER_HOSTNAME_RETVAL

	# Prepare user stats dir
	${SSH_BIN} $consumer_user "rm -rf $LBG_DEP_DIR ; mkdir -p $LBG_DEP_DIR/stats $LBG_DEP_DIR/logs"
	
	# Start user
	uploadDeps $consumer_user
	echo "Running user"
	$SSH_BIN $consumer_user \
    	"$SHELL_BIN -c \"$JAVA_BIN -cp '$LBG_LIBS' -ea lbg.middleware.AutomaticJobSubmitter \
			-H $consumer_peer_h -P $PEER_USER_PORT -h $consumer_user_h -p $user_port \
			-i $consumer_user_h -n $jobs --sleep_lo 1 --sleep_hi 1 --block true --ftp_host $consumer_peer_h \
			--ftp_port $ftp_port --dir '$data_dir' --size $data_size -t $tasks_per_job \
			-d $data_per_task --ddr $ddr -s '$LBG_DEP_DIR/stats' -j '$LBG_DEP_DIR/gridhasher.jar' \
			-o '$LBG_DEP_DIR' > '$LBG_DEP_DIR/logs/submitter' 2>&1\""

	# Clean
	echo -e "\n---\nShutting down grid..."
	echo -e "$CLUSTER_LIST\n$consumer_peer\n$consumer_user" | ./lbg_shutdown_grid.sh $search_engine
	# Kill FTP, send sigkill: vsftp isn't always reactive enough
	$SSH_BIN $consumer_peer "killall -9 vsftpd 2> /dev/null" > /dev/null 2>&1

	# Gather logs from clusters and user
	mkdir -p $SRC_CONFIG_DIR/logs $SRC_CONFIG_DIR/stats
	for machine in $CLUSTER_LIST $consumer_peer $consumer_user $search_engine ; do
		echo -ne "Gathering logs from $machine\r"
		$SCP_BIN $machine:$LBG_DEP_DIR/logs/* $SRC_CONFIG_DIR/logs/ > /dev/null 2>&1
		$SCP_BIN $machine:$LBG_DEP_DIR/stats/* $SRC_CONFIG_DIR/stats/ > /dev/null 2>&1
		# Now delete logs, stats, jars and output .dat files would be done
		# automatically at next iteration but maybe there is none
		$SSH_BIN $machine "rm -rf $LBG_DEP_DIR"
	done
	echo

	# Clean caches (emptying them, then removing them also ensures
	# the symlinked directories will be emptied
	for machine in $CLUSTER_LIST $consumer_peer ; do
		echo -ne "Cleaning cache on $machine\r"
		$SSH_BIN $machine "rm -rf $PEER_CACHE_DIR/* $PEER_CACHE_DIR/.??* $RES_CACHE_DIR/* $RES_CACHE_DIR/.??*;
			rm -rf $PEER_CACHE_DIR $RES_CACHE_DIR"
	done
	echo

	CACHE_HIT_FILES=`mktemp`
	AGGREGATE=$SRC_CONFIG_DIR/aggregated_stats.txt
	echo "Aggregating statistics in $AGGREGATE"
	# Get cache hits
	echo -e "$experiment_id\n" > $AGGREGATE
	find $SRC_CONFIG_DIR/stats/ -name '*.cache_hit' > $CACHE_HIT_FILES
	java -cp $LOCAL_LBG_JAR lbg.common.math.SummarizedMeasurement $CACHE_HIT_FILES >> $AGGREGATE
	# Only one file, but arbitrary name
	cat $SRC_CONFIG_DIR/stats/*.job_runtime >> $AGGREGATE
	rm -f $CACHE_HIT_FILES

	echo "Finished experiment $experiment_id at "`date +"$DATE_FORMAT"`
	echo "---"
}

while [ $# -gt 0 ] ; do
	handleExperiment "$1"
	shift
done

# ----------------------------------------------------------------
