#!/bin/sh

#  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
#
#  This file is part of Lightweight Bartering Grid.
#  Lightweight Bartering Grid is free software;
#  you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  Lightweight Bartering Grid is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Lightweight Bartering Grid; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Lightweight Bartering Grid Background Load Generator
# Author:       Cyril Briquet
# Last revised: April 8th, 2008

# Usage example:
# ./lbg_blg.sh 3 2 5 120 12 12 true myapp ~/lbg/deployment/gridapp_examples/ < ~/gridsites.conf
# to submit to 3 Peers 2 Jobs of 5 Tasks after 2 minutes, spaced by 12 seconds,
# the Grid application being myapp.jar available in ~/lbg/deployment/gridapp_examples/

# ----------------------------------------------------------------

if ! test $1 || ! test $2 || ! test $3 || ! test $4 || ! test $5 || ! test $6 || ! test $7 ; then
  echo "Usage: $0 peer_count jobs_count tasks_per_job activation_delay inter_submission_lo inter_submission_hi block_before_resubmission [app_jar_base_name] [app_jar_path] < gridsites.conf" ;
  exit
fi

# ----------------

PEER_COUNT=$1
if [ "${PEER_COUNT}" -lt 1 ] ; then
  echo Illegal Peer count ${PEER_COUNT}
  exit
fi

JOBS_COUNT=$2
if [ "${JOBS_COUNT}" -lt 1 ] ; then
  echo Illegal Jobs count ${JOBS_COUNT}
  exit
fi

TASKS_PER_JOB=$3
if [ "${TASKS_PER_JOB}" -lt 1 ] ; then
  echo Illegal number of Tasks per Job ${TASKS_PER_JOB}
  exit
fi
#if [ "${TASKS_PER_JOB}" -gt 5 ] ; then
#  echo Illegal number of Resources per Peer ${TASKS_PER_JOB}
#  exit
#fi

ACTIVATION_DELAY=$4
if [ "${ACTIVATION_DELAY}" -lt 1 ] ; then
  echo Illegal activation delay ${ACTIVATION_DELAY}
  exit
fi

if [ "$5" = "trace" ] && [ "$6" = "trace" ] ; then
# (currently) experimental/undocumented feature:
# use a Job submissions trace file
# instead of a random inter-arrival times distribution
  TRACE_FN="/tmp/trace.txt"

  SUBMISSION_TIMES_FLAGS="--trace ${TRACE_FN}"
else
# set random inter-arrival times distribution parameters
  INTER_SUBMISSION_LO=$5
  if [ "${INTER_SUBMISSION_LO}" -lt 0 ] ; then
    echo Illegal Jobs inter-submission interval low boundary ${INTER_SUBMISSION_LO}
    exit
  fi
  
  INTER_SUBMISSION_HI=$6
  if [ "${INTER_SUBMISSION_HI}" -lt 0 ] ; then
    echo Illegal Jobs inter-submission interval high boundary ${INTER_SUBMISSION_HI}
    exit
  fi

  if [ "${INTER_SUBMISSION_HI}" -lt "${INTER_SUBMISSION_LO}" ] ; then
    echo Illegal Jobs inter-submission interval ${INTER_SUBMISSION_LO}-${INTER_SUBMISSION_HI}
    exit
  fi
  
  SUBMISSION_TIMES_FLAGS="--sleep_lo ${INTER_SUBMISSION_LO} --sleep_hi ${INTER_SUBMISSION_HI}"
fi


BBR=$7
if [ "${BBR}" != "true" ] ; then
  if [ "${BBR}" != "false" ] ; then
    echo Illegal block-before-resubmission flag ${BBR}
    exit
  fi
fi

# ----------------

# optional parameters

MY_APP=myapp
if test $8 ; then
  MY_APP=$8
fi

MY_APP_PATH=${HOME}/lbg/deployment/gridapp_examples/
if test $9 ; then
  MY_APP_PATH=$9
fi

# ----------------

JAR_DIR=/tmp/.grid_dep_dir
JAR_DEP="${JAR_DIR}/lightweight_bartering_grid.jar:${JAR_DIR}/Azureus2.5.0.4.jar:${JAR_DIR}/commons-logging-1.1.jar:${JAR_DIR}/commons-math-1.1.jar:${JAR_DIR}/edtftpj.jar:${JAR_DIR}/ftpserver-dev.jar:${JAR_DIR}/jfreechart-1.0.9.jar:${JAR_DIR}/jcommon-1.0.12.jar:${JAR_DIR}/iText-2.0.8.jar:${MY_APP_PATH}/${MY_APP}.jar"
AJS_JOB_FLAGS=" -i ats_user -d 0 -t ${TASKS_PER_JOB} -n ${JOBS_COUNT} ${SUBMISSION_TIMES_FLAGS} --block ${BBR} -j ${MY_APP_PATH}/${MY_APP}.jar -m ${MY_APP} -s /tmp -o /tmp"
#AJS_MONITORING_FLAG="-agentlib:yjpagent"

# ----------------------------------------------------------------

echo "Lightweight Bartering Grid Background Load Generator"
echo "  `date`"
echo "  will now submit ${JOBS_COUNT} BoTs of ${TASKS_PER_JOB} Tasks each to (up to) ${PEER_COUNT} Peers, "
echo "  every ${INTER_SUBMISSION_LO}-${INTER_SUBMISSION_HI} seconds, starting in ${ACTIVATION_DELAY} seconds..."

# init loop counter
peer_i=0
TARGET_PEERS=" "
while read SITE; do
  if [ -n "${SITE}" ]; then

    # isolate current Peer from stdin
    CURRENT_PEER_TARGET=`echo "${SITE}" | cut -s -f1 -d" "`
    CURRENT_PEER="`echo ${CURRENT_PEER_TARGET} | cut -f2 -d\"@\"`"
    AJS_NET_FLAGS="-H ${CURRENT_PEER} -P 50102 -h ${CURRENT_PEER} -p 50107"
    TARGET_PEERS="${TARGET_PEERS} ${CURRENT_PEER}"
    #--echo "debug xx ${CURRENT_PEER_TARGET} ${CURRENT_PEER} peer_i=${peer_i} xx"

    # actually submit 1 BoT of 'Resources-per-Peer' Tasks to current Peer
    SLEEP_CMD="sleep ${ACTIVATION_DELAY}"
    AJS_CMD="java -cp ${JAR_DEP} ${AJS_MONITORING_FLAG} -ea lbg.middleware.AutomaticJobSubmitter ${AJS_JOB_FLAGS} ${AJS_NET_FLAGS}"
    CMD="${SLEEP_CMD} ; ${AJS_CMD}"
    #--echo "debug xx ${CMD} xx"
    (ssh ${CURRENT_PEER_TARGET} /bin/bash -c "\"${CMD}\"" < /dev/null) &

    # update loop counter, complete script when Peer count reached
    peer_i=$((peer_i+1))
    if [ "${peer_i}" = "${PEER_COUNT}" ]; then
      echo "${TARGET_PEERS}"
      exit
    fi
  fi
done
# if specified Peer count is greater than actual cluster count, the loop
# completes correctly (i.e. all clusters have been processed once), and silently

echo "${TARGET_PEERS}"
exit

# ----------------------------------------------------------------
