/**
 * Copyright (c) 2008, Gerard Dethier
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of Liège nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY GERARD DETHIER ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL ﻿GERARD DETHIER BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.Properties;

/**
 * An UAClient can be used to submit jobs from an external software
 * to a Lightweight Bartering Grid User Agent.
 * <p>To do so, it communicates with a LiveUserAgentServer
 * (package lbg.user.server).
 * <p>The UAClient class class also illustrates
 * the text-based protocol used by the LiveUserAgentServer.
 * <p>The UAClient class can, and should, be sub-classed to be fitted to
 * a particular application.
 * @author Gerard Dethier
 */
public class UAClient {

	protected int userAgentServerPort;
	protected PrintStream toUser;
	protected BufferedReader fromUser;
	protected File tmpFile;

	/**
	 * Constructor of the UAClient.
	 * It connects to LiveUserAgentServer running on the local host.
	 * 
	 * @param userAgentServerPort The port on which the LiveUserAgent is listening.
	 * 
	 * @throws IOException When the connection to the LiveUserAgent could not be established.
	 */
	public UAClient(int userAgentServerPort) throws IOException {

		this.userAgentServerPort = userAgentServerPort;

		// Connection to the LiveUserAgentServer
		Socket sock = new Socket();
		sock.connect(new InetSocketAddress(InetAddress.getLocalHost(), userAgentServerPort), 1000);
		
		toUser = new PrintStream(sock.getOutputStream());
		fromUser = new BufferedReader(new InputStreamReader(sock.getInputStream()));
		
		// Creation of temporary file used for the jdf files on job submissions
		tmpFile = new File("UAClient.tmp");
		tmpFile.deleteOnExit();

	}

	/**
	 * A call to this method leads to the shutdown of the LiveUserAgentServer to which
	 * this interface is connected. It blocks until the LiveUserAgentServer's shutdown.
	 * 
	 * @throws IOException If a problem occurs while reading the LiveUserAgentServer's
	 * notification.
	 */
	public void shutdownUser() throws IOException {
		toUser.println("ShutdownUser");
		fromUser.readLine(); // Wait for the user to shutdown
	}

	/**
	 * submitJobToTheGrid submits a job to LBG given a job description.
	 * 
	 * @param jobDescription The description of the job to submit.
	 * This description should be generated using the <code>describeEmbeddedJob</code> methods
	 * .
	 * @return The job id if the job was correctly submitted.
	 * 
	 * @throws GridException If the job could not be submitted by the LiveUserAgentServer.
	 * @throws IOException If the temporary property file could not be generated or the
	 * LiveUserAgentServer's acknowledgement could not be read.
	 */
	public int submitJobToTheGrid(Properties jobDescription) throws IOException {

		// Creation of jdf file.
		FileOutputStream fos = new FileOutputStream(tmpFile);
		jobDescription.store(fos, "");
		fos.close();
		
		// Submit job request to LiveUserAgentServer
		toUser.println("JobSubmission");
		toUser.println(tmpFile.getAbsolutePath());
		
		// Reception of acknowledgement of the LiveUserAgentServer
		int jobId;
		String ack = fromUser.readLine();
		if("OK".equals(ack)) {
			jobId = Integer.parseInt(fromUser.readLine());
		} else {
			String message = fromUser.readLine();
			System.err.println(message);
			return -1;
		}

		return jobId;
	}
	
	/**
	 * describeEmbeddedJob generates a description of a Job given
	 * the jar files, the "main" classes and the parameters of each Task.
	 * 
	 * @param jarFiles An array of paths to jar files. jarFiles[i] is the jar file name of Task i.
	 * @param taskMains An array of class names. taskMains[i] is the "main" class name of Task i.
	 * @param parameters An array of parameters. parameters[i] are the parameters of Task i.
	 * The parameters of a Task must be separated by white spaces.
	 * 
	 * @return The properties describing a Job or null if the given arrays do not have the same size.
	 * 
	 */
	public Properties describeEmbeddedJob(String[] jarFiles, String[] taskMains, String[] parameters) {

		/* Expected Properties:
		task_list = embedded_task0, embedded_task1, ...
		...
		embedded_taski.jar = eval(jarFiles[i])
		embedded_taski.main_class = eval(taskMains[i])
		embedded_taski.parameters = eval(parameters[i])
		...
		 */

		if(jarFiles.length != taskMains.length || taskMains.length != parameters.length) {
			return null;
		}

		int task_count = jarFiles.length;

		Properties props = new Properties();
		StringBuilder task_list = new StringBuilder();

		String base_name = "embedded_task";
		for (int i = 0 ; i < task_count ; i++) {
			// Set properties of Task i
			String task_name = base_name+i;

			props.setProperty(task_name+".jar", jarFiles[i]);
			props.setProperty(task_name+".main_class", taskMains[i]);
			props.setProperty(task_name+".parameters", parameters[i]);

			// Add Task i to Task list
			task_list.append(task_name);
			if (i < task_count-1) task_list.append(", ");
		}

		// Set Task list property
		props.setProperty("task_list",task_list.toString());

		return props;

	}
	
	/**
	 * describeEmbeddedJob generates a description of a Job of identical tasks.
	 * 
	 * @param task_count The number of Tasks to put in the Job.
	 * @param jarFile The jar file name containing the Task classes.
	 * @param taskMain The "main" class name.
	 * @param parameters The parameters of the tasks.
	 * 
	 * @return The properties describing a Job.
	 */
	public Properties describeEmbeddedJob(int task_count, String jarFile, String taskMain, String parameters) {

		/* Expected Properties:
		task_list = embedded_task0, embedded_task1, ...
		...
		embedded_taski.jar = jarFile
		embedded_taski.main_class = taskMain
		embedded_taski.parameters = parameters
		...
		 */

		Properties props = new Properties();
		StringBuilder task_list = new StringBuilder();

		String base_name = "embedded_task";

		for (int i = 0 ; i < task_count ; i++) {
			// Set properties of Task i
			String task_name = base_name+i;

			props.setProperty(task_name+".jar", jarFile);
			props.setProperty(task_name+".main_class", taskMain);
			props.setProperty(task_name+".parameters", parameters);

			// Add Task i to Task list
			task_list.append(task_name);
			if (i < task_count-1) task_list.append(", ");
		}

		// Set Task list property
		props.setProperty("task_list",task_list.toString());

		return props;

	}
	
	/**
	 * describeEmbeddedJob generates a description of a Job of Tasks differing only by parameters.
	 * 
	 * @param task_count The number of Tasks to put in the Job.
	 * @param jarFile The jar file name of the jar containing the Task classes.
	 * @param taskMain The "main" class name.
	 * @param parameters The parameters of the tasks.
	 * 
	 * @return The properties describing a Job.
	 */
	public Properties describeEmbeddedJob(String jarFile, String taskMain, String[] parameters) {

		/* Expected Properties:
		task_list = embedded_task0, embedded_task1, ...
		...
		embedded_taski.jar = jarFile
		embedded_taski.main_class = taskMain
		embedded_taski.parameters = parameters[i]
		...
		 */

		Properties props = new Properties();
		StringBuilder task_list = new StringBuilder();
		
		int task_count = parameters.length;

		String base_name = "embedded_task";

		for (int i = 0 ; i < task_count ; i++) {
			// Set properties of Task i
			String task_name = base_name+i;

			props.setProperty(task_name+".jar", jarFile);
			props.setProperty(task_name+".main_class", taskMain);
			props.setProperty(task_name+".parameters", parameters[i]);

			// Add Task i to Task list
			task_list.append(task_name);
			if (i < task_count-1) task_list.append(", ");
		}

		// Set Task list property
		props.setProperty("task_list",task_list.toString());

		return props;

	}

}

