/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LightweightBarteringGrid.java

package lbg;

import static lbg.VirtualizationLevel.*;
import lbg.common.exception.*;
import lbg.simulator.*;

/** The <b>Lightweight Bartering Grid</b> middleware allows Peers to gridify
 * their computational and storage Resources,
 * and barter them globally following various negotiation/accounting strategies.
 * @author Cyril Briquet
 */
public abstract class LightweightBarteringGrid {

	public static String welcome() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("Lightweight Bartering Grid, ");
		sb.append("a P2P Grid middleware and discrete-event simulator\n");
		sb.append("  Version ");
		sb.append(LightweightBarteringGrid.VERSION);
		sb.append("\n");
		sb.append("  Copyright (c) 2005-2008, ");
		sb.append("Cyril Briquet, parts Xavier Dalem\n");
		sb.append("  Contact: C.Briquet@ulg.ac.be\n");

		return sb.toString();
		
	}

	// ****************************************************************

	/** Sets current operating mode (callable only once).
	 * @param mode target operating mode
	 * @exception GridException if target mode is illegal,
	 * or a mode has already been set
	 */
	public static void setMode(VirtualizationLevel mode) throws GridException {
		
		if (LightweightBarteringGrid.mode_is_set == true) {
			throw new GridException("setMode(): illegal operation");
		}
		
		switch (mode) {
		case LIVE_MODE:
		case SIM_MODE:
			LightweightBarteringGrid.current_mode = mode;
			break;
		default:
			throw new GridException("setMode(): illegal mode");
		}
		
		LightweightBarteringGrid.mode_is_set = true;
		
	}
	
	/** Returns current operating mode.
	 * @return current operating mode
	 */
	public static VirtualizationLevel getMode() {

		return LightweightBarteringGrid.current_mode;
		
	}

	/** Sets current Simulator and Scenario (legal in SIM mode only).
	 * @param sim target Simulator
	 * @param sc target Scenario
	 * @exception GridException if target Scenario is illegal,
	 * or current mode is illegal
	 */
	public static synchronized void setSimulator(Simulator sim,Scenario sc)
		throws GridException {
		
		switch (LightweightBarteringGrid.current_mode) {
		case LIVE_MODE:
			throw new GridException("setSimulator(): illegal operation");
		case SIM_MODE:
			if (sim == null) {
				throw new GridException("setSimulator(): illegal Simulator");
			}
			if (sc == null)  {
				throw new GridException("setSimulator(): illegal Scenario");
			}
			LightweightBarteringGrid.current_simulator = sim;
			LightweightBarteringGrid.current_scenario = sc;
			break;
		default:
			throw new GridException("setSimulator(): unsupported mode");
		}
		
	}

	/** Returns current Scenario (legal in SIM mode only).
	 * @exception GridException if current mode is illegal
	 */
	public static Scenario getScenario() throws GridException {

		switch (LightweightBarteringGrid.current_mode) {
		case LIVE_MODE:
			throw new GridException("getScenario(): illegal operation");
		case SIM_MODE:
			return LightweightBarteringGrid.current_scenario;
		default:
			throw new GridException("getScenario(): unsupported mode");
		}
		
	}

	/** Returns current time:
	 * <ul>
	 * <li>SIM mode - in seconds since the beginning of the simulation
	 * <li>LIVE mode - in seconds since the epoch
	 * </ul>
	 * @return current time (in seconds)
	 * @exception GridException if time status is corrupted
	 */
	public static synchronized long getTime() throws GridException {
		
		switch (LightweightBarteringGrid.current_mode) {
		case LIVE_MODE:
			return System.currentTimeMillis() / 1000;
		case SIM_MODE:
			return LightweightBarteringGrid.time;
		default:
			throw new GridException("getTime(): unsupported mode");
		}
		
	}

	/** Sets current time to target time (legal in SIM mode only).
	 * @param time target time
	 * @return current time (in seconds)
	 * @exception GridException if target time is illegal,
	 * current mode is illegal, or time status is corrupted
	 */
	public static synchronized long setTime(long time) throws GridException {
		
		switch (LightweightBarteringGrid.current_mode) {
		case LIVE_MODE:
			throw new GridException("setTime(): illegal operation");
		case SIM_MODE:
			if (LightweightBarteringGrid.time > time) {
				throw new GridException("setTime(): illegal time");
			}
			return LightweightBarteringGrid.time = time;
		default:
			throw new GridException("setTime(): unsupported mode");
		}
		
	}

	/** Shifts current time forward by target amount (legal in SIM mode only).
	 * @param timeshift target time shift
	 * @return current time (in seconds)
	 * @exception GridException if target time shift is illegal,
	 * current mode is illegal, or time status is corrupted
	 */
	public static synchronized long shiftTime(long timeshift)
		throws GridException {
		
		switch (LightweightBarteringGrid.current_mode) {
		case LIVE_MODE:
			throw new GridException("shiftTime(): illegal operation");
		case SIM_MODE:
			if (timeshift < 0) {
				throw new GridException("setTime(): illegal time shift");
			}
			return LightweightBarteringGrid.time += timeshift;
		default:
			throw new GridException("shiftTime(): unsupported mode");
		}
		
	}
	
	// ****************************************************************

	public static synchronized void handleGridException(GridException e) {
		
		e.printStackTrace();           // stderr output

		System.out.println();
		e.printStackTrace(System.out); // stdout output
		System.out.println();

		if (LightweightBarteringGrid.current_simulator != null) {
			System.out.println(LightweightBarteringGrid.current_simulator);
		}

		if (LightweightBarteringGrid.getMode() == SIM_MODE) {
			System.exit(0);
		}
		
	}

	// ****************************************************************

	/** Lightweight Bartering Grid current version.
	 */
	public static final String VERSION = "0.4.1";

	public static String[] getJarDependencies() {
		
		return LightweightBarteringGrid.JAR_DEPENDENCIES.clone();
		
	}
	
	private static final String[] JAR_DEPENDENCIES = new String[] {
		"Azureus2.5.0.4.jar",
		"commons-logging-1.1.jar",
		"commons-math-1.1.jar",
		"edtftpj.jar",
		"ftpserver-dev.jar",
		"jfreechart-1.0.9.jar",
		"jcommon-1.0.12.jar",
		"iText-2.0.8.jar"
	};
	
	// ****************************************************************

	static {

		LightweightBarteringGrid.mode_is_set = false;
		LightweightBarteringGrid.current_mode = LIVE_MODE;
		LightweightBarteringGrid.time = 0;
		LightweightBarteringGrid.current_scenario = null;
		LightweightBarteringGrid.current_simulator = null;

	}

	// ****************************************************************

	private static boolean mode_is_set;
	private static VirtualizationLevel current_mode;
	private static long time;
	private static Scenario current_scenario;
	private static Simulator current_simulator;

}

