/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ThreadPool.java

package lbg.common;

import lbg.common.exception.*;
import lbg.common.math.CommonMath;
import static lbg.common.net.CommonNetHandling.*;


/**
 * @author Cyril Briquet
 */
public class ThreadPool {

	public ThreadPool(Runnable runnable,int thread_pool_size,
					  boolean set_daemon,boolean hi_priority,String thread_name)
		throws GridException {
		
		if (runnable == null) {
			throw new GridException("ThreadPool(): illegal Runnable");
		}
		if (thread_pool_size < 1) {
			throw new GridException("ThreadPool(): illegal thread count");
		}
		if (thread_name == null) {
			throw new GridException("ThreadPool(): illegal Thread name");
		}

		this.thread_name         = thread_name;
		this.runnable            = runnable;
		this.thread_pool_size    = thread_pool_size;
		this.set_daemon          = set_daemon;
		this.hi_priority         = hi_priority;
		this.thread_pool_created = false;
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ThreadPool]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized void startService() throws GridException {

		if (this.thread_pool_created == false) {

			// Launch multiple threads
			for (int ti = 0 ; ti < this.thread_pool_size ; ti++) {
				Thread t = new Thread(this.runnable,thread_name+"-"+ti);
				t.setDaemon(this.set_daemon);
				if (this.hi_priority) {
					t.setPriority(Math.min(t.getPriority()+2,
										   Thread.MAX_PRIORITY));
				}
				t.start();
				//System.out.println("Launched thread #"+ti);
			}

			this.thread_pool_created = true;
		}
		
	}

	// ****************************************************************

	public static void sleep(long seconds) throws GridException {
		
		nap(seconds*1000L);
		
	}

	public static void nap(long millis) throws GridException {
		
		if (millis < 0) {
			throw new GridException("nap(): illegal *ZZZ* length");
		}
		try {
			Thread.sleep(millis);
		}
		catch (InterruptedException e) {
			throw new GridException(e);
		}
		
	}
	
	public static final int DEFAULT_THREAD_POOL_SIZE = getCoresCount()*2;

	// ****************************************************************

	private final String   thread_name;
	private final int      thread_pool_size;
	private final Runnable runnable;
	private final boolean  hi_priority;
	private final boolean  set_daemon;
	private       boolean  thread_pool_created;

}

