/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ArgsParser.java

package lbg.common.io;

import java.util.TreeMap;
import java.util.LinkedList;
import java.text.ParseException;

/** A parser for command-line arguments.
 * <br>Supports optional and required parameters, short and long options,
 * flags and arguments taking a list of values.
 * <br>Currently doesn't support arguments strings with a inderminate number of
 * parameters at the end (like a list of files grep-style). Can be easily
 * bypassed with a OptionType.LIST option.
 * <h4>Version history</h4>
 * <dl>
 * <dt>1.1.3 (2007-03-27)</dt><dd>
 * {@link #addIntOption(char, String, boolean, Integer, String)}
 * requires a default value if option isn't required.</dd>
 * <dt>1.1.2 (2006-12-01)</dt><dd>Check passed parameters.</dd>
 * <dt>1.1.1 (2006-10-17)</dt><dd>Usage auto-documents LISTs if needed.</dd>
 * <dt>1.1 (2006-10-15)</dt><dd>Added support for lists.</dd>
 * <dt>1.0.1 (2006-08-27)</dt><dd>Improved the looks of usage information.</dd>
 * <dt>1.0 (2006-08-26)</dt><dd>First working version, added support for custom
 * helper, see addHelper() ; descriptions ; improved usage information
 * ; still lacks lists.</dd>
 * <dt>0.1.1</dt><dd>Support for types.</dd>
 * <dt>0.1.0</dt><dd>Objectives, method signatures and main
 * structures.</dd>
 * </dl>
 * <h4>Todo:</h4>
 * <ul><li>Add support for lists</li></ul>
 * @version 1.1.3 (2007-03-27)
 * @author Xavier Dalem
 */
public class ArgsParser {
	/** The type of the option. */
	private enum OptionType { STRING, INT, FLAG, LIST }

	/** This can be thrown when the parser meets an inconsistency in the
	 * options.
	 * This is a developper exception, the end-user should never see this. */
	private static class OptionException extends RuntimeException {
		static final long serialVersionUID = 2109728264299034428L;
		public OptionException(String s) {
			super(s);
		}
	}

	/** Option groups both regular options taking an argument and flags */
	private static class Option {
		// Declared finals because we won't modify them and it's easier to
		// access
		final char sh;
		final String lg;
		final OptionType type;
		final boolean required;
		private final String desc;
		private Object value;

		Option(char sh, String lg, Object defVal, OptionType type,
				boolean required, String desc) {
			if (type == null)
				throw new NullPointerException("Option got null type");
			
			this.sh = sh;
			this.lg = lg;
			this.value = defVal;
			this.type = type;
			this.required = required;
			this.desc = (desc != null)?desc:"";
		}

		void setValue(String val) {
			switch (type) {
				case STRING:
					value = val;
					break;
				case INT:
					value = new Integer(val);
					break;
				case LIST:
					value = val.split(",");
					break;
				default:
					throw new Error("Cannot assign string value to boolean" +
						" flag");
			}
		}

		void setFlag() {
			value = Boolean.valueOf(true);
		}

		Object getValue() {
			return value;
		}

		public String toString() {
			StringBuilder sb = new StringBuilder("");
			
			if (sh != ' ' && lg != null)
				sb.append("  -" + sh + ", -" + lg);
			else if (sh != ' ')
				sb.append("  -" + sh);
			else
				sb.append("  -" + lg);
			
			switch (type) {
				case STRING:
					sb.append("  STRING");
					break;
				case INT:
					sb.append("  INT");
					break;
				case LIST:
					sb.append("  LIST");
					break;
				default:
			}
			
			sb.append("\n\t" + desc + (required?" (required)":" (optional)"));
			
			return sb.toString();
		}
	}

	/** The program name, as shown in usage information */
	private String program;
	/** A map of short (one characater) options */
	private TreeMap<Character, Option> shortOptions;
	/** A map of long options */
	private TreeMap<String, Option> longOptions;
	/** A list of all required options, consulted for parse() only and
	 * sequentially.
	 * Linked list was chosent to be fast */
	private LinkedList<Option> required;
	/** A list of all options, consulted for usage() only and sequentially.
	 * Linked list was chosen to be fast */
	private LinkedList<Option> whole;
	/** Was there an error preventing parsing ? */
	private boolean error;
	/** Is there at least one LIST ? Useful for usage() */
	private boolean hasList = false;
	/** The helper */
	private Option helper;

	/** Prepare an argument parser.
	 * @param program The program name, as it will appear in usage information
	 */
	public ArgsParser(String program) {
		if (program == null)
			throw new NullPointerException("ArgsParser needs non-null program");
		
		this.program = program;
		shortOptions = new TreeMap<Character, Option>();
		longOptions = new TreeMap<String, Option>();
		required = new LinkedList<Option>();
		whole = new LinkedList<Option>();
		error = false;
		helper = null;
	}

	/** Register a string option for the Arguments Parser.
	 * @param sh The short option character (for -f options), ' ' if none
	 * @param lg The long option string, without first leading '-'. For '--foo'
	 * style options, make only one leading '-', null if none
	 * @param required true if the option is required
	 * @param defValue The default value for that option, null for none. Not
	 * taken into account if option is required.
	 * @param desc A description in usage information
	 * @throws OptionException In case of option inconsistency
	 */
	public void addStringOption(char sh, String lg, boolean required,
		String defValue, String desc) throws OptionException { 
		addOption(new Option(sh, lg, required?null:defValue, OptionType.STRING,
			required, desc));
	}

	/** Register an list option for the Arguments Parser.
	 * Default values aren't supported. The list is a comma-separated list
	 * of string values.
	 * @param sh The short option character (for -f options), ' ' if none
	 * @param lg The long option string, without first leading '-'. For '--foo'
	 * style options, make only one leading '-', null if none
	 * @param required true if the option is required
	 * @param desc A description in usage information
	 * @throws OptionException In case of option inconsistency
	 */
	public void addListOption(char sh, String lg, boolean required,
		String desc) throws OptionException { 
		addOption(new Option(sh, lg, null, OptionType.LIST,	required, desc));
		hasList = true;
	}

	/** Register an int option for the Arguments Parser.
	 * @param sh The short option character (for -f options), ' ' if none
	 * @param lg The long option string, without first leading '-'. For '--foo'
	 * style options, make only one leading '-', null if none
	 * @param required true if the option is required
	 * @param defValue The default value for that option. Not
	 * taken into account if option is required.
	 * @param desc A description in usage information
	 * @throws OptionException In case of option inconsistency
	 */
	public void addIntOption(char sh, String lg, boolean required,
		Integer defValue, String desc) throws OptionException { 
		if (!required && defValue == null)
			throw new IllegalArgumentException(
				"Optionnal int Option requires default value");
		addOption(new Option(sh, lg, required?null:defValue, OptionType.INT,
			required, desc));
	}

	/** Register a flag for the Arguments Parser.
	 * Flags do not require an argument, and are always optionnal.
	 * @param sh The short option character (for -f options), ' ' if none
	 * @param lg The long option string, without first leading '-'. For '--foo'
	 * style options, make only one leading '-', null if none
	 * @param desc A description in usage information
	 * @throws OptionException In case of option inconsistency
	 */
	public void addFlag(char sh, String lg, String desc)throws OptionException {
		addOption(new Option(sh, lg, Boolean.valueOf(false), OptionType.FLAG,
			false, desc));
	}

	/** Adds a special kind of flag that will cause the parser to stop without
	 * paying attention to the rest of the command line. Nothing will be done
	 * but setting the appropriate flag.
	 * <br>Only one helper can be set
	 * @param sh The short option character (for -f options), ' ' if none
	 * @param lg The long option string, without first leading '-'. For '--foo'
	 * style options, make only one leading '-', null if none
	 * @param desc A description in usage information
	 * @throws OptionException In case of option inconsistency
	 */
	public void addHelper(char sh, String lg, String desc)
		throws OptionException {
		if (helper != null)
			throw new OptionException("You cannot set two helpers");
		
		helper = new Option(sh, lg, Boolean.valueOf(false), OptionType.FLAG,
			false, desc);
		addOption(helper);
	}

	/** Add option o to the parser
	 * @throws OptionException In case of option inconsistency
	 */
	private void addOption(Option o) throws OptionException {
		if (o.lg != null && o.lg.equals(""))
			throw new OptionException("Long option can be null but not empty");

		if (o.sh == ' ' && o.lg == null)
			throw new OptionException("Option is not well qualified");

		if (o.sh != ' ') {
			Character c = Character.valueOf(o.sh);
			if (shortOptions.put(c, o) != null) {
				error = true;
				throw new OptionException("Option '-" + o.sh
					+ "' is already registered");
			}
		}

		if (o.lg != null) {
			if (longOptions.put(o.lg, o) != null) {
				error = true;
				throw new OptionException("Option '-" + o.lg
					+ "' is already registered");
			}
		}

		if (o.required)
			this.required.add(o);

		whole.add(o);
	}

	/** Parse the arguments with currently registered options and flags
	 * @param argv The arguments to parse
	 * @throws ParseException if parsing fails. Offset from that exception
	 * is the number of the argument
	 */
	public void parse(String[] argv) throws ParseException {
		if (error)
			throw new ParseException("Cannot parse arguments: options are "
				+ "inconsistent", 0);
		
		if (argv == null)
			throw new IllegalArgumentException("Null argument array");
		
		for (int argc = 0; argc < argv.length; argc++) {
			if (argv[argc].charAt(0) != '-')
				throw new ParseException("'" + argv[argc]
					+ "': not a valid option", argc);
			
			// Handle options
			Option o = null;
			
			if (argv[argc].length() == 2) // short
				o = shortOptions.get(Character.valueOf(argv[argc].charAt(1)));
			else // long
				o = longOptions.get(argv[argc].substring(1));
			
			if (o == null)
				throw new ParseException("'" + argv[argc]
					+ "': not a valid option", argc);

			if (o.type == OptionType.FLAG) {
				o.setFlag();
				if (o == helper)
					return;
			}
			else {
				if (++argc == argv.length)
					throw new ParseException("'" + argv[argc-1]
						+ "': option requires an argument", argc);
				try {
					o.setValue(argv[argc]);
				} catch (NumberFormatException e) {
					throw new ParseException("'" + argv[argc-1]
						+ "': option requires an integer argument", argc);
				}
			}
		}
		
		// Check that all required options have been set
		for (Option o : required) {
			if (o.getValue() == null) {
				if (o.sh != ' ')
					throw new ParseException("Error: option '-" + o.sh
						+ "' is required", 0);
				else
					throw new ParseException("Error: option '-" + o.lg
						+ "' is required", 0);
			}
		}
	}

	/** Get value of option lg.
	 * If that option was also registered with a character, you may
	 * indiferently check both and get the same result.
	 * <br>This value is guaranteed to be non null if option was required and
	 * parse() didn't throw an exception.
	 * @param lg The long option
	 * @throws OptionException If the option doesn't exist
	 */
	public String strValueOf(String lg) throws OptionException {
		Option o = longOptions.get(lg);
		if (o == null)
			throw new OptionException("'-" + lg + "': No such option");
		
		if (o.type != OptionType.STRING)
			throw new OptionException("'-" + lg + "': Not a string option");
		
		return (String) o.getValue();
	}

	/** Get value of option sh.
	 * If that option was also registered with a string, you may
	 * indiferently check both and get the same result.
	 * <br>This value is guaranteed to be non null if option was required and
	 * parse() didn't throw an exception.
	 * <br>This method can be more efficient than its equivalent taking a
	 * String as first argument.
	 * @param sh The short option
	 * @throws OptionException If the option doesn't exist
	 */
	public String strValueOf(char sh) throws OptionException {
		Option o = shortOptions.get(Character.valueOf(sh));
		if (o == null)
			throw new OptionException("'-" + sh + "': No such option");
		
		if (o.type != OptionType.STRING)
			throw new OptionException("'-" + sh + "': Not a string option");
		
		return (String) o.getValue();
	}

	/** Get value of option lg as a string array.
	 * If that option was also registered with a character, you may
	 * indiferently check both and get the same result.
	 * <br>This value is guaranteed to be non null if option was required and
	 * parse() didn't throw an exception.
	 * @param lg The long option
	 * @throws OptionException If the option doesn't exist
	 */
	public String[] listValueOf(String lg) throws OptionException {
		Option o = longOptions.get(lg);
		if (o == null)
			throw new OptionException("'-" + lg + "': No such option");
		
		if (o.type != OptionType.LIST)
			throw new OptionException("'-" + lg + "': Not a string option");
		
		return (String[]) o.getValue();
	}

	/** Get value of option sh as a string array.
	 * If that option was also registered with a string, you may
	 * indiferently check both and get the same result.
	 * <br>This value is guaranteed to be non null if option was required and
	 * parse() didn't throw an exception.
	 * <br>This method can be more efficient than its equivalent taking a
	 * String as first argument.
	 * @param sh The short option
	 * @throws OptionException If the option doesn't exist
	 */
	public String[] listValueOf(char sh) throws OptionException {
		Option o = shortOptions.get(Character.valueOf(sh));
		if (o == null)
			throw new OptionException("'-" + sh + "': No such option");
		
		if (o.type != OptionType.LIST)
			throw new OptionException("'-" + sh + "': Not a string option");
		
		return (String[]) o.getValue();
	}

	/** Get value of option lg as an int.
	 * If that option was also registered with a character, you may
	 * indiferently check both and get the same result.
	 * <br>This method gets through the task of typecasting the argument.
	 * <br>This value is guaranteed to exist if option was required and
	 * parse() didn't throw an exception.
	 * @param lg The long option
	 * @throws OptionException If the option doesn't exist
	 */
	public int intValueOf(String lg) throws OptionException {
		Option o = longOptions.get(lg);
		if (o == null)
			throw new OptionException("'-" + lg + "': No such option");
		
		if (o.type != OptionType.INT)
			throw new OptionException("'-" + lg + "': Not an integer option");
		
		return ((Integer) o.getValue()).intValue();
	}

	/** Get value of option sh as an int.
	 * If that option was also registered with a character, you may
	 * indiferently check both and get the same result.
	 * <br>This method gets through the task of typecasting the argument.
	 * <br>This value is guaranteed to exist if option was required and
	 * parse() didn't throw an exception.
	 * <br>This method can be more efficient than its equivalent taking a
	 * String as first argument.
	 * @param sh The short option
	 * @throws OptionException If the option doesn't exist
	 */
	public int intValueOf(char sh) throws OptionException {
		Option o = shortOptions.get(Character.valueOf(sh));
		if (o == null)
			throw new OptionException("'-" + sh + "': No such option");
		
		if (o.type != OptionType.INT)
			throw new OptionException("'-" + sh + "': Not an integer option");
		
		return ((Integer) o.getValue()).intValue();
	}

	/** Get value of flag lg.
	 * If that option was also registered with a character, you may
	 * indiferently check both and get the same result.
	 * <br>This value is guaranteed to exist.
	 * @param lg The long option
	 * @return true if the flag was present
	 * @throws OptionException If the option doesn't exist
	 */
	 public boolean getFlag(String lg) throws OptionException {
		Option o = longOptions.get(lg);
		if (o == null)
			throw new OptionException("'-" + lg + "': No such option");
		
		if (o.type != OptionType.FLAG)
			throw new OptionException("'-" + lg + "': Not a flag");
		
		return ((Boolean) o.getValue()).booleanValue();
	}

	/** Get value of flag sh.
	 * If that option was also registered with a character, you may
	 * indiferently check both and get the same result.
	 * <br>This value is guaranteed to exist.
	 * <br>This method can be more efficient than its equivalent taking a
	 * String as first argument.
	 * @param sh The short option
	 * @return true if the flag was present
	 * @throws OptionException If the option doesn't exist
	 */
	public boolean getFlag(char sh) throws OptionException {
	Option o = shortOptions.get(Character.valueOf(sh));
		if (o == null)
			throw new OptionException("'-" + sh + "': No such option");
		
		if (o.type != OptionType.FLAG)
			throw new OptionException("'-" + sh + "': Not a flag");
		
		return ((Boolean) o.getValue()).booleanValue();
	}

	/** Returns usage infomation, to be printed in console or screen.
	 * This command also outputs the format of LISTs, so it shouldn't
	 * be mentionned in independant descriptions. */
	public String usage() {
		StringBuilder sb = new StringBuilder("Usage: "
			+ program + " OPTIONS\nwhere OPTIONS can be:\n");
		
		for (Option o : whole)
			sb.append(o.toString() + "\n");
		
		if (hasList)
			sb.append("\nLISTs are comma-delimited lists of values\n");
		
		return sb.toString();
	}

	/** Demo program for ArgsParser */
	public static void main(String[] argv) throws OptionException {
		ArgsParser p = new ArgsParser("xr.util.ArgsParser");
		
		p.addHelper('h', "-help", "Show this help and exit");
		p.addFlag('s', "stats", "Report stats");
		p.addStringOption(' ', "-file", false, null, "Use specified input file");
		p.addIntOption('n', null, true, null, "Iterate INT times");
		p.addListOption('l', null, true, "List of output files");
		
		try {
			p.parse(argv);
		} catch (ParseException pe) {
			System.err.println(pe.getMessage());
			System.err.println(p.usage());
			return;
		}
		
		if (p.getFlag('h')) {
			System.err.println("Demo program for argument parser (does"
				+ " nothing)\n\n" + p.usage());
			return;
		}
		
		String[] list = p.listValueOf('l');
		StringBuilder sbl = new StringBuilder("{ ");
		for (int i = 0; i < list.length; i++) {
			sbl.append(list[i]);
			if (i < list.length - 1)
				sbl.append(" , ");
		}
		sbl.append(" }");
		
		System.out.println("Successfully parsed arguments, values are:"
			+ "\n s | stats : " + p.getFlag("stats")
			+ "\n n         : " + p.intValueOf('n')
			+ "\n file      : " + p.strValueOf("-file")
			+ "\n l         : " + sbl.toString()
			);
	}
}

