/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// CommonChartHandling.java

package lbg.common.io;

import lbg.common.exception.*;

import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import static org.jfree.chart.plot.PlotOrientation.*;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.DatasetUtilities;

import com.lowagie.text.Document;
import com.lowagie.text.DocumentException;
import com.lowagie.text.Rectangle;
import com.lowagie.text.pdf.DefaultFontMapper;
import com.lowagie.text.pdf.PdfContentByte;
import com.lowagie.text.pdf.PdfTemplate;
import com.lowagie.text.pdf.PdfWriter;

/**
 * @author Cyril Briquet
 */
public class CommonChartHandling {

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[CommonChartHandling]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	public static JFreeChart generateChart(String title,
										   String x_label,String y_label,
										   double[] values)
		throws GridException {
		
		System.out.println("Generating chart...");

		/*null title permitted -- if (title == null) {
			throw new GridException("generateChart(): illegal title");
		}*/
		if (x_label == null) {
			throw new GridException("generateChart(): illegal X axis label");
		}
		if (y_label == null) {
			throw new GridException("generateChart(): illegal X axis label");
		}
		
		System.out.println("Generating chart dataset...");

		CategoryDataset ds = generateDataset(values);

		System.out.println("Completed generation of chart dataset");
		
		JFreeChart chart = ChartFactory.
			createBarChart(title,x_label,y_label,ds,VERTICAL,false,true,true);

		System.out.println("Completed generation of chart");

		return chart;

	}

    public static CategoryDataset generateDataset(double[] values)
		throws GridException {
		
		if (values == null) {
			throw new GridException("generateDataset(): illegal values");
		}

		double values_store[][] = new double[1][];
		values_store[0] = values; 

		return DatasetUtilities.
			createCategoryDataset("Series","Category",values_store);

    }

	public static BufferedImage generateChartImage(JFreeChart chart)
		throws GridException {
		
		if (chart == null) {
			throw new GridException("generateChartImage(): illegal chart");
		}

		System.out.println("Saving chart as an image...");

		BufferedImage image =
			chart.createBufferedImage(DEFAULT_IMG_CHART_WIDTH,
									  DEFAULT_IMG_CHART_HEIGHT);
	

		System.out.println("Completed saving of chart as an image");

		return image;
		
	}
	
	// ****************************************************************

    /** Saves target chart as a .pdf file.
	 * @param chart target chart
	 * @param fn target .pdf filename
	 * @return chart target chart, to enable method chaining
	 * @exception GridException if a parameter is illegal
	 */
	public static JFreeChart saveAsPDF(JFreeChart chart,String fn)
		throws GridException {
		
		return saveAsPDF(chart,fn,
					 	 DEFAULT_PDF_CHART_WIDTH,DEFAULT_PDF_CHART_HEIGHT);
		
	}

    /** Saves target chart as a .pdf file.
	 * @param chart target chart
	 * @param fn target .pdf filename
	 * @param w width of target .pdf file
	 * @param h height of target .pdf file
	 * @return chart target chart, to enable method chaining
	 * @exception GridException if a parameter is illegal
     */
    public static JFreeChart saveAsPDF(JFreeChart chart,String fn,int w,int h)
		throws GridException {

		System.out.println("Saving chart as .pdf...");

		// The code of this method is largely based on iText online tutorial:
		// http://itextdocs.lowagie.com/tutorial/

		if (chart == null) {
			throw new GridException("saveAsPDF(): illegal chart");
		}
		if (CommonFileHandling.isPDFFileName(fn) == false) {
			throw new GridException("saveAsPDF(): illegal filename "+fn);
		}
		if (w < 1) {
			throw new GridException("saveAsPDF(): illegal width "+w);
		}
		if (h < 1) {
			throw new GridException("saveAsPDF(): illegal height "+h);
		}
		
		// http://www.mail-archive.com/itext-questions@lists.sourceforge.net/
		//		msg18399.html
		// 		msg16646.html
		org.jfree.text.TextUtilities.setUseDrawRotatedStringWorkaround(false);

    	Document doc = new Document(new Rectangle(w,h));
    	try {
    		PdfWriter writer =
				PdfWriter.getInstance(doc,new FileOutputStream(fn));
    		doc.open();
			
    		PdfContentByte cb = writer.getDirectContent();
    		PdfTemplate tp = cb.createTemplate(w,h);
    		Graphics2D g2d = tp.createGraphics(w,h,new DefaultFontMapper());
    		Rectangle2D r2d = new Rectangle2D.Double(0,0,w,h);
    		chart.draw(g2d,r2d);
    		g2d.dispose();
    		cb.addTemplate(tp,0,0);
    	}
    	catch(DocumentException e) {
    		throw new GridException(e);
    	}
    	catch (FileNotFoundException e) {
    		throw new GridException(e);
    	}
		finally {
			doc.close();
		}

		System.out.println("Completed saving of chart as .pdf...");
		
		return chart;
		
    }

	// ****************************************************************

	public static final int DEFAULT_IMG_CHART_WIDTH   = 240; //288;
	
	public static final int DEFAULT_IMG_CHART_HEIGHT  = 180; //224;

	public static final int DEFAULT_PDF_CHART_WIDTH   = 400;
	
	public static final int DEFAULT_PDF_CHART_HEIGHT  = 600;

}

