/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// CommonFileHandling.java

package lbg.common.io;

import lbg.common.exception.*;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.channels.FileChannel;
import java.util.jar.Attributes;
import java.util.jar.JarInputStream;
import java.util.jar.Manifest;
import java.util.Properties;

/** A CommonFileHandling provides common File handling routines.
 * @author Cyril Briquet, Xavier Dalem
 */
public abstract class CommonFileHandling {

    /** Opens a BufferedReader encapsulating
     * target filename.
     * @param fn target filename
     * @return encapsulating BufferedReader
	 * @exception GridException if target file cannot be opened
     */
    public static BufferedReader openBR(String fn) throws GridException {

        FileReader fr = null;
        try {
            fr = new FileReader(fn);
            if (fr == null) {
                throw new FileNotFoundException();
            }
        }
        catch (FileNotFoundException e) {
			throw new GridException(e);
            //System.out.println("Error : cannot open (ro) "+fn);
        }
        return new BufferedReader(fr);

    }

    /** Closes target BufferedReader encapsulating
     * target filename.
     * @param br target encapsulating BufferedReader
     * @param fn target filename
     * @exception GridException if target file cannot be closed
     */
    public static void closeBR(BufferedReader br,String fn)
		throws GridException {

        try {
            if (br == null) {
                throw new IOException("null BufferedReader");
            }
            br.close();
        }
        catch (IOException e) {
			throw new GridException(e);
            //System.out.println("Error : cannot close "+fn);
        }

    }

    /** Reads one text line from target BufferedReader encapsulating
     * target filename.
     * @param br target encapsulating BufferedReader
     * @param fn target filename
     * @return read line, or null if there is no more line to read
     * @exception GridException if current line cannot be read
     */
    public static String readLine(BufferedReader br,String fn)
		throws GridException {

        String s = null;
        try {
            s = br.readLine();
        }
        catch (IOException e) {
			throw new GridException(e);
            //System.out.println("Error : unexpected I/O error "+
            //                   "while reading from "+fn);
        }
        return s;

    }

	/** Reads text file contents given target filename.
	 * @param fn target filename
	 * @return read file
	 * @exception GridException if a parameter is illegal,
	 * or if an error occurs while reding the file
	 */
	public static String readFile(String fn) throws GridException {
	
		BufferedReader br = CommonFileHandling.openBR(fn);
		StringBuilder sb = new StringBuilder();
		
		String s = null;
		while ((s = CommonFileHandling.readLine(br,fn)) != null) {
			sb.append(s);
			sb.append("\n");
		}

		CommonFileHandling.closeBR(br,fn);
	
		return sb.toString();
		
	}

	/** Reads text file contents given target filename.
	 * @param fn target filename
	 * @param dont_complain if true, no Exception will be thrown,
	 * and null will be returned
	 * @return read file, or null if dont_complain is true,
	 * and either a parameter is illegal,
	 * or an error occurs while reding the file
	 * @exception GridException if dont_complain is false,
	 * and either a parameter is illegal,
	 * or an error occurs while reding the file
	 */
	public static String readFile(String fn,boolean dont_complain)
		throws GridException {
		
		String retval = null;
		try {
			retval = readFile(fn);
		}
		catch (GridException e) {
			if (dont_complain == false) {
				throw e;
			}
			else {
				retval = null;
			}
		}
		return retval;

	}
	
	// ****************************************************************

    /** Writes target data to target file.
     * @param fn target filename
     * @param data target data
     * @return true if write operation was a success
     * @exception GridException if target data cannot be written
     */
    public static boolean writeFile(String fn,String data)
		throws GridException {

        BufferedWriter bw = null;
        PrintWriter    pw = null;
		FileWriter     fw = null;
        try {
            fw = new FileWriter(fn);
        }
        catch (IOException e) {
            System.out.println("Error : cannot open (wo) "+fn);
            return false;
        }

		bw = new BufferedWriter(fw);
		pw = new PrintWriter(bw);
		
        pw.print(data);
		pw.flush();
		pw.close();

		/* old code
        try {
			bw.flush();
			bw.close();
        }
        catch (IOException e) {
			throw new GridException(e);
            //System.out.println("Error : cannot close "+fn);
        }
		*/

		return true;

    }
    
    /** Writes target data to target file.
     * @param f target file
     * @param data target data
     * @throws GridException if target data cannot be written
     */
    public static void writeFile(File f, String data) throws GridException {

    	try {
    		FileWriter fw = new FileWriter(f);
    		fw.write(data);
    		fw.flush();
    		fw.close();
    	} catch (IOException e) {
    		throw new GridException(e);
    	}

    }
    
	/** Copies target source file to target destination file.
	 * @param fn1 target source filename
	 * @param fn2 target destination filename
	 * @exception GridException if an error occurs
	 * while copying the file
	 */
	public static void copyFile(String fn1,String fn2) throws GridException {
		
		try {
			File f_in = new File(fn1);
			File f_out = new File(fn2);
			FileChannel src_channel = new FileInputStream(f_in).getChannel();
			FileChannel dst_channel = new FileOutputStream(f_out).getChannel();
			src_channel.transferTo(0,src_channel.size(),dst_channel);
			// or dst_channel.transferFrom(src_channel,0,src_channel.size());
			src_channel.close();
			dst_channel.close();
		}
		catch (IOException e) {
			throw new GridException(e);
		}
		
	}

	// ****************************************************************

    /** Returns The first non-existing file matching the original one.
     * <p>Files are checked in a wget fashion. If the supplied files doesn't
     * exist, it is returned. If it exists, a new file of the same name, but
     * with a ".0" appended to it. As long as the files exist, the suffix will
     * be incremented, returning the first non-existing file.</p>
     * <p>For example, if the files "foo", "foo.0" and "foo.2" exist, this
     * method applied to <code>new File("foo")</code> will return a
     * <code>File("foo.1")</code></p>
     * @param f base file
     * @return file that matches the above contract
     */
    public static File getFirstMatchingFile(File f) {
    	String base_filename = f.getAbsolutePath();
    	int offset = -1;
    	
    	while (f.exists())
    		f = new File(base_filename + "." + ++offset);
    	
    	return f;
    }

	// ****************************************************************

	/** Creates a new directory.
	 * @param target_dir target directory to create
	 * @exception GridException if a parameter is illegal
	 */
	public static void createDirectory(String target_dir)
		throws GridException {
		
		if (target_dir == null) {
			throw new GridException("createDirectory(): "+
									"illegal target directory");
		}
		
		File f = new File(target_dir);
		if (f.exists() == true) {
			if (f.isDirectory() == false) {
				throw new GridException("createDirectory(): "+
										target_dir + "already exists, "+
										"but is not a directory");
			}
		}
		else /* if (f.exists() == false) */ {
			if (f.mkdirs() == false) {
				throw new GridException("createDirectory(): cannot create "+
										target_dir);
			}
		}
		
	}

	/** Recursively removes a directory and all its contents,
	 * including subdirectories.
	 * @param target_dir target directory to remove
	 * @exception GridException if a parameter is illegal
	 */
	public static void removeDirectory(String target_dir) throws GridException {

		removeDirectory(target_dir,false);
		
	}

	/** Recursively removes a directory and all its contents,
	 * including subdirectories.
	 * @param target_dir target directory to remove
	 * @param fail_silently if true, no GridException is thrown
	 * @exception GridException if a parameter is illegal
	 */
	public static void removeDirectory(String target_dir,boolean fail_silently)
		throws GridException {

		try {
			if ((target_dir == null) || (target_dir.length() < 1)) {
				throw new GridException("removeDirectory(): "+
										"illegal target directory");
			}
			
			removeDirectoryHelper(new File(target_dir));
		}
		catch (GridException e) {
			if (fail_silently == false) {
				throw e;
			}
		}
		
	}

	private static void removeDirectoryHelper(File f) throws GridException {
		
		if (f == null) {
			throw new GridException("removeDirectoryHelper(): illegal file");
		}

		if (f.exists() == false) {
			throw new GridException("removeDirectoryHelper(): "+
									"file "+f.getPath()+" not found");
		}

		String fn = f.getName();
		if ((fn.compareTo(".") == 0) || (fn.compareTo("..") == 0)) {
			return;
		}
		
		if (f.isDirectory()) {
			File[] dir_contents = f.listFiles();
			for (int i = 0 ; i < dir_contents.length ; i++) {
				removeDirectoryHelper(dir_contents[i]);
			}
		}

		if (f.delete() == false) {
			throw new GridException("removeDirectoryHelper(): "+
									"cannot delete "+f.getPath());
		}
		
		
	}

	// ****************************************************************

	/** Loads Properties from target text file.
	 * @param fn target text filename
	 * @exception GridException if an error occurs
	 * while loading the Properties text file
	 */
	public static Properties loadProperties(String fn) throws GridException {
		
		Properties props = null;
		
		try {
			FileInputStream fis = new FileInputStream(fn);
			props = new Properties();
			props.load(fis);
			fis.close();
		}
		catch (IOException e) {
			throw new GridException(e);
		}
		
		return props;
		
	}

	// ****************************************************************

	/** Dumps target Object to standard output.
	 * @param obj target Object to dump to standard output
	 * @return true if Object is legal
	 */
 	public static boolean dump(Object obj) {
		
		System.out.print(obj);
		
		return (obj != null);
		
	}

	/** Dumps separating line to System console standard output.
	 */
 	public static boolean dumpSeparator() {
		
		System.out.println("\n");
		for (int i = 0 ; i < 16 ; i++) System.out.print("====");				
		System.out.println("\n");
		
		return true;
		
	}

	// ****************************************************************

	/** Extracts file name from target file name with complete path.
	 * @param fn target file name with complete path
	 * @return extracted file name
	 * @exception GridException if target file name with complete path
	 * is illegal
	 */
	public static String extractFileNameFromPath(String fn)
		throws GridException {
		
		if (fn == null) {
			throw new GridException("extractFileNameFromPath(): "+
									"illegal target filename");
		}
		
		int idx = fn.lastIndexOf(FILE_SEP);
		if (idx == -1) {
			return fn;
		}
		
		return fn.substring(idx+1);
		
	}
	
	// ****************************************************************

	/** Extracts base file name (without extension)
	 * from target file name without complete path.
	 * @param fn target file name without complete path
	 * @param ext target extension
	 * @return extracted base file name
	 * @exception GridException if target file name without complete path
	 * or target extension is illegal
	 */
	public static String extractBaseFileNameFromFileName(String fn,String ext)
		throws GridException {
		
		if (fn == null) {
			throw new GridException("extractBaseFileNameFromFileName(): "+
									"illegal target filename");
		}
		if (ext == null) {
			throw new GridException("extractBaseFileNameFromFileName(): "+
									"illegal target extension");
		}

		int idx = fn.lastIndexOf(ext);
		if (idx == -1) {
			throw new GridException("extractBaseFileNameFromFileName(): "+
									"target extension not present");
		}
		
		return fn.substring(0,idx);
		
	}

	/** Extracts base file name (without extension)
	 * from target file name with complete path.
	 * @param fn target file name with complete path
	 * @param ext target extension
	 * @return extracted base file name
	 * @exception GridException if target file name with complete path
	 * or target extension is illegal
	 */
	public static String extractBaseFileNameFromPath(String fn,String ext)
		throws GridException {
		
		if (fn == null) {
			throw new GridException("extractBaseFileNameFromPath(): "+
									"illegal target filename");
		}
		if (ext == null) {
			throw new GridException("extractBaseFileNameFromPath(): "+
									"illegal target extension");
		}
		
		String fn_tmp = extractFileNameFromPath(fn);
		
		return extractBaseFileNameFromFileName(fn_tmp,ext);
		
	}
	
	// ****************************************************************

	/** Checks if target file name has a licit extension.
	 * @param fn target file name
	 * @param ext target extension
	 * @return true if target file name has a licit extension
	 */
	private static boolean hasLicitExtension(String fn,String ext) {
	
		if ((ext == null) || (fn == null)) {
			return false;
		}
		
		int fn_len = fn.length();
		int ext_len = ext.length();
		return ((fn_len >= ext_len+1) &&
				(fn.lastIndexOf(ext) == fn_len-ext_len));
		
	}

	/** Checks if target file name has a licit prefix.
	 * @param fn target file name
	 * @param pfx target prefix
	 * @return true if target file name has a licit extension
	 */
	private static boolean hasLicitPrefix(String fn,String pfx) {
	
		if ((pfx == null) || (fn == null)) {
			return false;
		}
		
		int fn_len = fn.length();
		int pfx_len = pfx.length();
		return ((fn_len >= pfx_len+1) && (fn.indexOf(pfx) == 0));
		
	}

	/** Checks if target path is absolute.
	 * @param path target path to check
	 * @return true if target path is absolute
	 * @see java.io.File#isAbsolute()
	 */
	public static boolean isAbsolutePath(String path) {
		
		return new File(path).isAbsolute();
		
	}

	// ****************************************************************

	/** Checks if target file name is a Java archive file name.
	 * @param fn target file name
	 * @return true if target file name is a Java archive file name
	 */
	public static boolean isJarFileName(String fn) {
	
		return hasLicitExtension(fn,CommonFileHandling.JAR_EXT);
		
	}
	
	/** Checks if target file name is a text file name.
	 * @param fn target file name
	 * @return true if target file name is a text file name
	 */
	public static boolean isTextFileName(String fn) {
	
		return hasLicitExtension(fn,CommonFileHandling.TXT_EXT);
		
	}

	/** Checks if target file name is a CSV spreadsheet file name.
	 * @param fn target file name
	 * @return true if target file name is a CSV spreadsheet file name
	 */
	public static boolean isCSVFileName(String fn) {
	
		return hasLicitExtension(fn,CommonFileHandling.CSV_EXT);
		
	}
	
	/** Checks if target file name is a PDF file name.
	 * @param fn target file name
	 * @return true if target file name is a PDF file name
	 */
	public static boolean isPDFFileName(String fn) {
	
		return hasLicitExtension(fn,CommonFileHandling.PDF_EXT);
		
	}

	/** Checks if target file name is a log file name.
	 * @param fn target file name
	 * @return true if target file name is a log file name
	 */
	public static boolean isLogFilename(String fn) {

		return hasLicitPrefix(fn,LOG_PREFIX) && hasLicitExtension(fn,LOG_EXT);

	}

	// ****************************************************************

	/** Extracts the main class of the target jar file's embedded manifest.
	 * @param jarfn target jar file name
	 * @return main class of the target jar,
	 * or null if the main class cannot be found
	 * @exception GridException if a parameter is illegal,
	 * or if the target jar cannot be processed
	 */
	public static String extractMainClass(String jarfn) throws GridException {
		
		if (isJarFileName(jarfn) == false) {
			throw new GridException("extractMainClass(): "+
									"illegal jar file name");
		}

		String main_class = null;
		JarInputStream jar = null;
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(jarfn);
			jar = new JarInputStream(fis);

			Manifest mf = jar.getManifest();
			if (mf == null) {
				main_class = null;
			}
			else {
				Attributes att = mf.getMainAttributes();
				main_class = att.getValue("Main-Class");
			}
		}
		catch (FileNotFoundException e) {
			throw new GridException("extractMainClass(): jar not found");
		}
		catch (IOException e) {
			throw new GridException("extractMainClass(): corrupted jar");
		}
		finally {
			try {
				if (jar != null) {
					jar.close();
				}
				if (fis != null) {
					fis.close();
				}
			}
			catch (IOException e) { /* perdu */ }
		}
		
		return main_class;

	}
	
	// ****************************************************************

	/** Filesystem file separator (i.e. / or \ depending on O.S. type).
     */
    public static final String FILE_SEP = File.separator;
		//System.getProperty("file.separator");
 
	/** Class file default extension.
	 */
	public static final String CLASS_EXT = ".class";

	/** CSV spreadsheet file default extension.
	 */
	public static final String CSV_EXT = ".csv";

	/** JAR file default extension.
	 */
	public static final String JAR_EXT = ".jar";

	/** Log file default extension.
	 */
	public static final String LOG_EXT = ".txt";

	/** Log file default prefix.
	 */
	public static final String LOG_PREFIX = "log-";

	/** Output data file default extension.
	 */
	public static final String OUTPUT_DATA_EXT = ".dat";
	
	/** PDF file default extension.
	 */
	public static final String PDF_EXT = ".pdf";

	/** Text file default extension.
	 */
	public static final String TXT_EXT = ".txt";


}

