/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// CommonStringHandling.java

package lbg.common.io;

import lbg.common.exception.*;
import lbg.protocol.data.MetaGridData;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.util.Calendar;
import java.util.Date;
import java.util.Properties;
import java.util.StringTokenizer;

/** A CommonStringHandling provides common String handling routines.
 * @author Cyril Briquet
 */
public abstract class CommonStringHandling {

	// ****************************************************************

	/** Loads property from target Properties given target key.
	 * @param key target key
	 * @param props target Properties
	 * @return property from target Properties given target key
	 * @exception GridException if target key or Properties are illegal
	 */
	public static String loadProperty(String key,Properties props)
		throws GridException {

		if (key == null) {
			throw new GridException("loadProperty(): illegal key");
		}
		if (props == null) {
			throw new GridException("loadProperty(): illegal Properties");
		}
		
		String p = props.getProperty(key);
		
		if (p == null) {
			throw new GridException("loadProperty(): key "+key+" undefined");
		}
		
		return p;
		
	}
	
	/** Sets target Properties with target (key, value) pair.
	 */
	public static void setProperty(Properties props,String key,String value)
		throws GridException {
		
		if (props == null) {
			throw new GridException("setProperty(): illegal Properties");
		}
		if (key == null) {
			throw new GridException("setProperty(): illegal key");
		}
		if (value == null) {
			throw new GridException("setProperty(): illegal value");
		}
		
		props.setProperty(key,value);
		
	}
	
	/** Sets target Properties with target (key, group value) pair.
	 */
	public static void setGroupProperty(Properties props,
										String key,String value,int s)
		throws GridException {
		
		if (props == null) {
			throw new GridException("setGroupProperty(): illegal Properties");
		}
		if (key == null) {
			throw new GridException("setGroupProperty(): illegal key");
		}
		if (value == null) {
			throw new GridException("setGroupProperty(): illegal value");
		}
		if (s < 1) {
			throw new GridException("setGroupProperty(): illegal group size");
		}

		StringBuilder sb_value = new StringBuilder();
		
		sb_value.append("{ ");
		for (int i = 0 ; i < s ; i++) {
			sb_value.append(value);
			if (i < s-1) sb_value.append(", ");
		}
		sb_value.append(" }");
		
		props.setProperty(key,sb_value.toString());
		
	}

	/** Sets target Properties with target (key, group value) pair.
	 */
	public static void setGroupProperty(Properties props,
										String key,String value_all_except_one,
										String value_last,int s)
		throws GridException {
		
		if (props == null) {
			throw new GridException("setGroupProperty(): illegal Properties");
		}
		if (key == null) {
			throw new GridException("setGroupProperty(): illegal key");
		}
		if (value_all_except_one == null) {
			throw new GridException("setGroupProperty(): illegal value "+
									"(all but last)");
		}
		if (value_last == null) {
			throw new GridException("setGroupProperty(): illegal value (last)");
		}
		if (s < 1) {
			throw new GridException("setGroupProperty(): illegal group size");
		}

		StringBuilder sb_value = new StringBuilder();
		
		sb_value.append("{ ");
		for (int i = 0 ; i < s-1 ; i++) {
			sb_value.append(value_all_except_one);
			sb_value.append(", ");
		}
		sb_value.append(value_last);
		sb_value.append(" }");
		
		props.setProperty(key,sb_value.toString());
		
	}

	/** Sets target Properties with target (key, group values) pair.
	 */
	public static void setGroupProperty(Properties props,
										String key,int[] values)
		throws GridException {
		
		if (props == null) {
			throw new GridException("setGroupProperty(): illegal Properties");
		}
		if (key == null) {
			throw new GridException("setGroupProperty(): illegal key");
		}
		if (values == null) {
			throw new GridException("setGroupProperty(): illegal values");
		}
		if (values.length < 1) {
			throw new GridException("setGroupProperty(): illegal group size");
		}

		StringBuilder sb_value = new StringBuilder();
		
		sb_value.append("{ ");
		for (int i = 0 ; i < values.length ; i++) {
			sb_value.append(values[i]);
			if (i < values.length-1) sb_value.append(", ");
		}
		sb_value.append(" }");
		
		props.setProperty(key,sb_value.toString());
		
	}

	/** Sets target Properties with target (key, group values) pair.
	 */
	public static void setGroupProperty(Properties props,
										String key,long[] values)
		throws GridException {
		
		if (props == null) {
			throw new GridException("setGroupProperty(): illegal Properties");
		}
		if (key == null) {
			throw new GridException("setGroupProperty(): illegal key");
		}
		if (values == null) {
			throw new GridException("setGroupProperty(): illegal values");
		}
		if (values.length < 1) {
			throw new GridException("setGroupProperty(): illegal group size");
		}

		StringBuilder sb_value = new StringBuilder();
		
		sb_value.append("{ ");
		for (int i = 0 ; i < values.length ; i++) {
			sb_value.append(values[i]);
			if (i < values.length-1) sb_value.append(", ");
		}
		sb_value.append(" }");
		
		props.setProperty(key,sb_value.toString());
		
	}

	/** Sets target Properties with target (key, group values) pair.
	 */
	public static void setGroupProperty(Properties props,
										String key,float[] values)
		throws GridException {
		
		if (props == null) {
			throw new GridException("setGroupProperty(): illegal Properties");
		}
		if (key == null) {
			throw new GridException("setGroupProperty(): illegal key");
		}
		if (values == null) {
			throw new GridException("setGroupProperty(): illegal values");
		}
		if (values.length < 1) {
			throw new GridException("setGroupProperty(): illegal group size");
		}

		StringBuilder sb_value = new StringBuilder();
		
		sb_value.append("{ ");
		for (int i = 0 ; i < values.length ; i++) {
			sb_value.append(values[i]);
			if (i < values.length-1) sb_value.append(", ");
		}
		sb_value.append(" }");
		
		props.setProperty(key,sb_value.toString());
		
	}
	
	/** Sets target Properties with target (key, group values) pair.
	 */
	public static void setGroupProperty(Properties props,
										String key,double[] values)
		throws GridException {
		
		if (props == null) {
			throw new GridException("setGroupProperty(): illegal Properties");
		}
		if (key == null) {
			throw new GridException("setGroupProperty(): illegal key");
		}
		if (values == null) {
			throw new GridException("setGroupProperty(): illegal values");
		}
		if (values.length < 1) {
			throw new GridException("setGroupProperty(): illegal group size");
		}

		StringBuilder sb_value = new StringBuilder();
		
		sb_value.append("{ ");
		for (int i = 0 ; i < values.length ; i++) {
			sb_value.append(values[i]);
			if (i < values.length-1) sb_value.append(", ");
		}
		sb_value.append(" }");
		
		props.setProperty(key,sb_value.toString());
		
	}

	/** Sets target Properties with target (key, group values) pair.
	 */
	public static void setGroupProperty(Properties props,
										String key,boolean[] values)
		throws GridException {
		
		if (props == null) {
			throw new GridException("setGroupProperty(): illegal Properties");
		}
		if (key == null) {
			throw new GridException("setGroupProperty(): illegal key");
		}
		if (values == null) {
			throw new GridException("setGroupProperty(): illegal values");
		}
		if (values.length < 1) {
			throw new GridException("setGroupProperty(): illegal group size");
		}

		StringBuilder sb_value = new StringBuilder();
		
		sb_value.append("{ ");
		for (int i = 0 ; i < values.length ; i++) {
			sb_value.append(values[i]);
			if (i < values.length-1) sb_value.append(", ");
		}
		sb_value.append(" }");
		
		props.setProperty(key,sb_value.toString());
		
	}

	// ****************************************************************

	/** Loads a boolean from target String.
	 * @param s target String
	 * @return boolean loaded from target String
	 * @exception GridException if target String is illegal
	 */
	public static boolean loadBoolean(String s) throws GridException {
	
		if (s == null) {
			throw new GridException("loadBoolean(): illegal String");
		}

		return Boolean.parseBoolean(s);

	}

	/** Loads an integer from target String.
	 * @param s target String
	 * @return integer loaded from target String
	 * @exception GridException if target String is illegal
	 */
	public static int loadInt(String s) throws GridException {
	
		if (s == null) {
			throw new GridException("loadInt(): illegal String");
		}

		int i = 0;
		try {
			i = Integer.parseInt(s);
			return i;
		}
		catch(NumberFormatException e) {
			throw new GridException("loadInt(): illegal number "+s);
		}
		
	}
	
	/** Loads a log from target String.
	 * @param s target String
	 * @return long loaded from target String
	 * @exception GridException if target String is illegal
	 */
	public static long loadLong(String s) throws GridException {
	
		if (s == null) {
			throw new GridException("loadLong(): illegal String");
		}

		long l = 0;
		try {
			l = Long.parseLong(s);
			return l;
		}
		catch(NumberFormatException e) {
			throw new GridException("loadLong(): illegal number "+s);
		}
		
	}

	/** Loads a float from target String.
	 * @param s target String
	 * @return float loaded from target String
	 * @exception GridException if target String is illegal
	 */
	public static float loadFloat(String s) throws GridException {
	
		if (s == null) {
			throw new GridException("loadFloat(): illegal String");
		}

		float f = 0.0f;
		try {
			f = Float.parseFloat(s);
			return f;
		}
		catch(NumberFormatException e) {
			throw new GridException("loadFloat(): illegal number "+s);
		}
		
	}

	/** Loads a double from target String.
	 * @param s target String
	 * @return double loaded from target String
	 * @exception GridException if target String is illegal
	 */
	public static double loadDouble(String s) throws GridException {
	
		if (s == null) {
			throw new GridException("loadDouble(): illegal String");
		}

		double d = 0.0;
		try {
			d = Double.parseDouble(s);
			return d;
		}
		catch(NumberFormatException e) {
			throw new GridException("loadDouble(): illegal number "+s);
		}
		
	}

	/** Loads an integer array from target String.
	 * <p>Accepted syntax is exactly the following :
	 * <tt>{x, y, z}</tt>, i.e. comma separated, exactly one space behind
	 * the comma, the whole thing being enclosed in braces with no spaces
	 * before or after them.</p>
	 * @param s target String
	 * @return integer array loaded from target String
	 * @exception GridException if target String is illegal
	 */
	public static int[] loadIntArray(String s) throws GridException {
	
		// BEGIN sanity check
		if (s == null) {
			throw new GridException("loadIntArray(): illegal String");
		}
		
		StringTokenizer st = new StringTokenizer(s,"{}",true);
		if (st.countTokens() != 3) {
			throw new GridException("loadIntArray(): missing { or } in "+s);
		}
		String dummy = st.nextToken();
		if (dummy.equals("{") == false) {
			throw new GridException("loadIntArray(): missing { in "+s);
		}
		String s_body = st.nextToken();
		dummy = st.nextToken();
		if (dummy.equals("}") == false) {
			throw new GridException("loadIntArray(): missing } in "+s);
		}
		// END sanity check
		
		st = new StringTokenizer(s_body,", ",false);
		int[] a = new int[st.countTokens()];
		int i = 0;
		while (st.hasMoreTokens()) {
			String tok = st.nextToken();
			try {
				a[i++] = Integer.parseInt(tok);
			}
			catch (NumberFormatException e) {
				throw new GridException("loadIntArray(): illegal "+
										"token number "+tok);
			}
		}
		
		return a;
		
	}

	/** Loads a long array from target String.
	 * @param s target String
	 * @return long array loaded from target String
	 * @exception GridException if target String is illegal
	 */
	public static long[] loadLongArray(String s) throws GridException {
	
		// BEGIN sanity check
		if (s == null) {
			throw new GridException("loadLongArray(): illegal String");
		}

		StringTokenizer st = new StringTokenizer(s,"{}",true);
		if (st.countTokens() != 3) {
			throw new GridException("loadLongArray(): missing { or } in "+s);
		}
		String dummy = st.nextToken();
		if (dummy.equals("{") == false) {
			throw new GridException("loadLongArray(): missing { in "+s);
		}
		String s_body = st.nextToken();
		dummy = st.nextToken();
		if (dummy.equals("}") == false) {
			throw new GridException("loadLongArray(): missing } in "+s);
		}
		// END sanity check
		
		st = new StringTokenizer(s_body,", ",false);
		long[] a = new long[st.countTokens()];
		int i = 0;
		while (st.hasMoreTokens()) {
			String tok = st.nextToken();
			try {
				a[i++] = Long.parseLong(tok);
			}
			catch (NumberFormatException e) {
				throw new GridException("loadLongArray(): illegal "+
										"token number "+tok);
			}
		}
		
		return a;
		
	}

	/** Loads a float array from target String.
	 * @param s target String
	 * @return float array loaded from target String
	 * @exception GridException if target String is illegal
	 */
	public static float[] loadFloatArray(String s) throws GridException {
	
		// BEGIN sanity check
		if (s == null) {
			throw new GridException("loadFloatArray(): illegal String");
		}

		StringTokenizer st = new StringTokenizer(s,"{}",true);
		if (st.countTokens() != 3) {
			throw new GridException("loadFloatArray(): missing { or } in "+s);
		}
		String dummy = st.nextToken();
		if (dummy.equals("{") == false) {
			throw new GridException("loadFloatArray(): missing { in "+s);
		}
		String s_body = st.nextToken();
		dummy = st.nextToken();
		if (dummy.equals("}") == false) {
			throw new GridException("loadFloatArray(): missing } in "+s);
		}
		// END sanity check
		
		st = new StringTokenizer(s_body,", ",false);
		float[] a = new float[st.countTokens()];
		int i = 0;
		while (st.hasMoreTokens()) {
			String tok = st.nextToken();
			try {
				a[i++] = Float.parseFloat(tok);
			}
			catch (NumberFormatException e) {
				throw new GridException("loadFloatArray(): illegal "+
										"token number "+tok);
			}
		}
		
		return a;
		
	}

	/** Loads a double array from target String.
	 * @param s target String
	 * @return double array loaded from target String
	 * @exception GridException if target String is illegal
	 */
	public static double[] loadDoubleArray(String s) throws GridException {
	
		// BEGIN sanity check
		if (s == null) {
			throw new GridException("loadDoubleArray(): illegal String");
		}

		StringTokenizer st = new StringTokenizer(s,"{}",true);
		if (st.countTokens() != 3) {
			throw new GridException("loadDoubleArray(): missing { or } in "+s);
		}
		String dummy = st.nextToken();
		if (dummy.equals("{") == false) {
			throw new GridException("loadDoubleArray(): missing { in "+s);
		}
		String s_body = st.nextToken();
		dummy = st.nextToken();
		if (dummy.equals("}") == false) {
			throw new GridException("loadDoubleArray(): missing } in "+s);
		}
		// END sanity check
		
		st = new StringTokenizer(s_body,", ",false);
		double[] a = new double[st.countTokens()];
		int i = 0;
		while (st.hasMoreTokens()) {
			String tok = st.nextToken();
			try {
				a[i++] = Double.parseDouble(tok);
			}
			catch (NumberFormatException e) {
				throw new GridException("loadDoubleArray(): illegal "+
										"token number "+tok);
			}
		}
		
		return a;
		
	}

	/** Loads a String array from target String.
	 * @param s target String
	 * @return String array loaded from target String
	 * @exception GridException if target String is illegal
	 */
	public static String[] loadStringArray(String s) throws GridException {
	
		if (s == null) {
			throw new GridException("loadStringArray(): illegal String");
		}
		
		StringTokenizer st = new StringTokenizer(s,"{}, ",false);
		String[] a = new String[st.countTokens()];
		int i = 0;
		while (st.hasMoreTokens()) {
			String tok = st.nextToken();
			a[i++] = tok;
		}
		
		return a;
		
	}

	/** Loads a boolean array from target String.
	 * @param s target String
	 * @return boolean array loaded from target String
	 * @exception GridException if target String is illegal
	 */
	public static boolean[] loadBooleanArray(String s) throws GridException {
	
		if (s == null) {
			throw new GridException("loadBooleanArray(): illegal String");
		}
		
		StringTokenizer st = new StringTokenizer(s,"{}, ",false);
		boolean[] a = new boolean[st.countTokens()];
		int i = 0;
		while (st.hasMoreTokens()) {
			String tok = st.nextToken();
			a[i++] = Boolean.parseBoolean(tok);
		}
		
		return a;
		
	}
	
	// ****************************************************************

	/** Generates a String representation of target int,
	 * filled with heading zeros if target int if smaller than target limit.
	 * <p>Example: fillWithHeadingZerosLimit(42,100) -> "042".
	 * @param i target int
	 * @param j target limit
	 * @return String representation of target int, filled with heading zeros
	 * if target int if smaller than pow(10,round(ceil(log10(j))))
	 * @exception GridException if a parameter is negative
	 */
	public static String fillWithHeadingZerosLimit(long i,long j)
		throws GridException {

		if (i < 0) {
			throw new GridException("fillWithHeadingZerosLimit(): "+
									"illegal target int");
		}
		if (j < 0) {
			throw new GridException("fillWithHeadingZerosLimit(): "+
									"illegal target limit");
		}
		
		int len = 1;
		for (int j0 = 1 ; j0 < (j+1) ; j0 *= 10,len++);
		
		return fillWithHeadingZeros(i,--len);
		
	}
	
	/** Generates a String representation of target int,
	 * filled with heading zeros, if target int is smaller
	 * than target power of 10. In other words, the target power of 10
	 * represents the number of digits in the returned String representation.
	 * <p>Example: fillWithHeadingZeros(42,3) -> "042".
	 * @param i target int
	 * @param len base 10 power of target limit
	 * @return String representation of target int, filled with heading zeros
	 * if target int if smaller than pow(10,len-1)
	 * @exception GridException if a parameter is negative
	 */
	public static String fillWithHeadingZeros(long i,int len)
		throws GridException {

		return String.format("%0"+len+"d", i);
		//return fillWithHeadingCharacter(i,len,'0');
	
	}
	
	/** Generates a String representation of target int,
	 * filled with heading spaces, if target int is smaller
	 * than target power of 10. In other words, the target power of 10
	 * represents the number of digits in the returned String representation.
	 * <p>Example: fillWithHeadingZeros(42,3) -> " 42".
	 * @param i target int
	 * @param len base 10 power of target limit
	 * @return String representation of target int, filled with heading zeros
	 * if target int if smaller than pow(10,len-1)
	 * @exception GridException if a parameter is negative
	 */
	public static String fillWithHeadingSpaces(long i,int len)
		throws GridException {

		return fillWithHeadingCharacter(i,len,' ');
	
	}

	/** Generates a String representation of target int,
	 * filled with target heading character, if target int is smaller
	 * than target power of 10. In other words, the target power of 10
	 * represents the number of digits in the returned String representation.
	 * <p>Example: fillWithHeadingZeros(42,3) -> "c42".
	 * @param i target int
	 * @param len base 10 power of target limit
	 * @param c target character
	 * @return String representation of target int, filled with heading zeros
	 * if target int if smaller than pow(10,len-1)
	 * @exception GridException if a parameter is negative
	 */
	public static String fillWithHeadingCharacter(long i,int len,char c)
		throws GridException {
		
		if (i < 0) {
			throw new GridException("fillWithHeadingCharacter(): "+
									"illegal target int");
		}
		if (len < 0) {
			throw new GridException("fillWithHeadingCharacter(): "+
									"illegal target limit");
		}

		StringBuilder sb = new StringBuilder();
		
		for (int base=10,limit=(int)Math.pow(10,len) ; base<limit ; base*=10) {
			if (i < base) sb.append(c);
		}
		sb.append(i);
		
		return sb.toString();
		
	}
	
	// ****************************************************************

	/** Replace a tag by another tag at the head of target String.
	 * Replacement takes place only when old tag is heading target String.
	 * @param s target String
	 * @param old_tag old tag
	 * @param new_tag new tag
	 */
	public static String replaceTag(String s,String old_tag,String new_tag) {
		
		if (s == null) {
			return s;
		}
		
		if (! s.startsWith(old_tag))
			return s;
		
		return new_tag+s.substring(old_tag.length());
		
	}

	// ****************************************************************

	/** Merges the two target String arrays into one new String array,
	 * so that the resulting array is a concatenation of the two target arrays.
	 * @param a1 1st target String array
	 * @param a2 2nd target String array
	 * @return merged String array
	 * @exception GridException if a parameter is illegal
	 * @deprecated Use {@link #mergeArrays} instead
	 */
	@Deprecated
	public static String[] mergeStringArrays(String[] a1,String[] a2)
		throws GridException {

		if ((a1 == null) || (a1.length < 0)) {
			throw new GridException("mergeStringArrays(): illegal array a1");
		}
		if ((a2 == null) || (a2.length < 0)) {
			throw new GridException("mergeStringArrays(): illegal array a2");
		}
		
		String[] am = new String[a1.length+a2.length];
		System.arraycopy(a1,0,am,0,a1.length);
		System.arraycopy(a2,0,am,a1.length,a2.length);
		
		return am;
		
	}
	
	/** Merges the target arrays into one new array,
	 * so that the resulting array is a concatenation of the target arrays.
	 * @param a1 1st target String array
	 * @param ax The other arrays
	 * @return the merged array
	 * @throws NullPointerException if any of the arrays is null
	 */
	public static <T> T[] mergeArrays(T[] a1,T[]... ax) {
		
		if (a1 == null) {
			throw new NullPointerException("mergeArrays(): "+
										   "null array as argument 0");
		}
		
		if (ax.length == 0) {
			return a1.clone();
		}
		
		// Compute total resulting length and create array
		int totalLen = a1.length;
		for (int i = 0; i < ax.length; i++) {
			if (ax[i] == null) {
				throw new NullPointerException("mergeArrays(): "+
											   "null array as argument "+(i+1));
			}
			totalLen += ax[i].length;
		}
		
		@SuppressWarnings("unchecked")
 		T[] ret = (T[])java.lang.reflect.Array.newInstance(
 			a1.getClass().getComponentType(),totalLen);
		
		// Copy base array to destination
		System.arraycopy(a1, 0, ret, 0, a1.length);
		
		// Copy following arrays
		int from = a1.length;
		for (int i = 0; i < ax.length; i++) {
			System.arraycopy(ax[i], 0, ret, from, ax[i].length);
			from += ax[i].length;
		}
		
		return ret;
	}

	// ****************************************************************

	/** Formats floating-point literal as a String,
	 * removing trailing .0 if present.
	 * @param d target floating-poing literal
	 * @return String representing target floating-poing literal
	 */
	public static String formatIntegerDoubleAsInteger(double d) {
		
		String s = Double.toString(d);
		
		return (s.endsWith(".0")) ? s.substring(0,s.length()-2) : s;
		
	}
	
	// ****************************************************************

	/** Generates a timestamp, i.e. 20080303141924.
	 * @return generated timestamp
	 */
	public static String generateTimeStamp() throws GridException {
		
		Calendar cal = Calendar.getInstance();
		StringBuilder sb = new StringBuilder();

		sb.append(cal.get(Calendar.YEAR)).
			append(fillWithHeadingZeros(1+cal.get(Calendar.MONTH),2)).
			append(fillWithHeadingZeros(cal.get(Calendar.DAY_OF_MONTH),2)).
			append(fillWithHeadingZeros(cal.get(Calendar.HOUR_OF_DAY),2)).
			append(fillWithHeadingZeros(cal.get(Calendar.MINUTE),2)).
			append(fillWithHeadingZeros(cal.get(Calendar.SECOND),2));
		
		return sb.toString();
		
	}
	
	// ****************************************************************

	/** Returns a DO NOT EDIT notice.
	 * @return DO NOT EDIT notice
	 */
	public static String doNotEditNotice() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("# DO NOT EDIT -- AUTOMATICALLY GENERATED FILE -- ").
			append(new Date()).
			append("\n");
		
		return sb.toString();
	
	}
	
	// ****************************************************************

	/** Formats as a String the stack trace of target Throwable.
	 * @param thr target Throwable
	 * @return stack trace of target Throwable, as a String
	 */
	 public static String getStackTrace(Throwable thr) {
		 
		 // BEGIN handling of null Throwable
		 if (thr == null) {
			 String retval = null;
			 try {
			 	retval = getStackTrace("getStackTrace(): illegal Throwable");
			 }
			 catch (GridException e) {
				 retval = "getStackTrace(): internal error";
			 }
			 return retval;
		 }
		 // END handling of null Throwable
		 
		 Writer sw = new StringWriter();
		 PrintWriter pw = new PrintWriter(sw);
		 thr.printStackTrace(pw);
		 
		 return sw.toString();
	 }

	/** Formats as a String the current stack trace using target message.
	 * @param msg target message
	 * @return stack trace of target Throwable, as a String
	 */
	 public static String getStackTrace(String msg) throws GridException {
		 
		 if (msg == null) {
			 throw new GridException("getStackTrace(): illegal message");
		 }
		 
		 try {
			 throw new GridException(msg);
		 }
		 catch (GridException e) {
			 return getStackTrace(e);
		 }

	 }
	 
	// ****************************************************************

	/** Zero-sized, dummy String, intended to be used as a dummy execution log.
	 */
	public static final byte[] DUMMY_LOG = new byte[0];

	/** Dummy String.
	 */
	public static final String DUMMY_STRING = "";

	/** Zero-sized, dummy String array.
	 */
	public static final String[] DUMMY_STRING_ARRAY = new String[0];

	/** Zero-sized, dummy MetaGridData array.
	 */
	public static final MetaGridData[] DUMMY_MGD_ARRAY = new MetaGridData[0];

	/** Zero-sized, dummy Object array.
	 */
	public static final Object[] DUMMY_OBJECT_ARRAY = new Object[0];

	/** Zero-sized, dummy byte array.
	 */
	public static final byte[] DUMMY_BYTE_ARRAY = new byte[0];

	/** Zero-sized, dummy byte matrix.
	 */
	public static final byte[][] DUMMY_BYTE_MATRIX = new byte[0][0];
	
}

