/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// EnhancedByteArrayOutputStream.java

package lbg.common.io;

import java.io.ByteArrayOutputStream;

/** <p>Additional methods for the standard ByteArrayOutputStream. 
 * This new class allows to copy its buffer in an already allocated
 * byte array rather than creating a new one and can support a limit of
 * characters.</p>
 * 
 * <p>If the stream is not large enough to fit the given information, it is
 * silently dropped. This is useful in cases where you need to start a
 * trustless child thread than should only output a limited amount of
 * data.</p> 
 * 
 * @author Xavier Dalem
 */
public class EnhancedByteArrayOutputStream extends ByteArrayOutputStream {
	private int limit;
	
	public EnhancedByteArrayOutputStream() {
		super();
		this.limit = 0;
	}
	
	public EnhancedByteArrayOutputStream(int limit) {
		super();
		this.limit = limit;
	}
	
	public EnhancedByteArrayOutputStream(int limit, int size) {
		super(size);
		this.limit = limit;
	}
	
	/** Sets the limit size for this stream to a new value.
	 * Set to 0 for no limit.
	 * 
	 * @param new_limit The new byte limit
	 * @return The old limit
	 */
	public int setLimit(int new_limit) {
		int old_limit = limit;
		limit = new_limit;
		return old_limit;
	}
	
	/** Copies the buffer of this stream to the given byte array. If the given
	 * array is too small, a new one is created as for {@link #toByteArray()}.
	 * 
	 * @param array The array where data should be put.
	 * @return A copy of the buffer
	 */
	public byte[] toByteArray(byte[] array) {
		if (array.length < count)
			return toByteArray();
		
		System.arraycopy(buf, 0, array, 0, count);
		
		return array;
	}

	/** {@inheritDoc}
	 * <p>If the contents don't fit in the stream because of the limit,
	 * it is truncated.</p>
	 */
	public void write(byte[] b, int off, int len) {
		if (limit != 0 && count + 1 > limit)
			return;
		else if (limit != 0 && count + len >= limit)
			super.write(b, off, (limit-count));
		super.write(b, off, len);
	}
	
	/** {@inheritDoc}
	 * <p>If the content don't fit in the stream because of limit,
	 * it is dropped.</p>
	 */
	public void write(int b) {
		if (count + 1 > limit)
			return;
		super.write(b);
	}
}
