/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// CommonMath.java

package lbg.common.math;

import lbg.common.exception.*;
import java.util.Random;

/**
 * @author Cyril Briquet
 */
public abstract class CommonMath {

	// ****************************************************************

	/** Samples a uniform distribution.
	 * The returned number belongs to the target range.
	 * @param lo low boundary
	 * @param hi high boundary
	 * @return random int number belonging to [lo..hi]
	 */
	public synchronized static int sampleUniformDist(int lo,int hi)
		throws GridException {
		
		if (lo > hi) {
			throw new GridException("CommonMath.sampleUniformDist(): "+
									"illegal range ["+lo+"-"+hi+"]");
		}
		
		return Math.round(CommonMath.rand.nextFloat()*(hi-lo))+lo;
		
	}
	
	/** Samples a uniform distribution.
	 * The returned number belongs to the target range.
	 * @param lo low boundary
	 * @param hi high boundary
	 * @return random int number belonging to [lo..hi]
	 */
	public synchronized static long sampleUniformDist(long lo,long hi)
		throws GridException {
		
		if (lo > hi) {
			throw new GridException("CommonMath.sampleUniformDist(): "+
									"illegal range ["+lo+"-"+hi+"]");
		}
		
		return Math.round(CommonMath.rand.nextDouble()*(hi-lo))+lo;
		
	}

	/** Samples a uniform distribution.
	 * The returned number may be any int value.
	 * @return random int number
	 */
	public synchronized static int sampleUniformDist() throws GridException {
		
		return CommonMath.rand.nextInt();
		
	}

	/** Samples a uniform distribution.
	 * The returned number belongs to the interval [0.0f..1.0f].
	 * @return random float number belonging to [0.0f..1.0f]
	 */
	public synchronized static float sampleUniformDistWithinUnity()
		throws GridException {
		
		return CommonMath.rand.nextFloat();
		
	}

	/** Samples a random file size sampled from a random distribution
	 * for target filename.
	 * <p>Returned values are guaranteed
	 * to be in the interval [2 << 10..2 << 22 + 2 << 10]
	 * <p>Different calls with the same target filename
	 * are guaranteed to result in the same file size.
	 * @param filename targer filename
	 * @return random file size
	 * @exception GridException if target filename is illegal
	 */
	public synchronized static int sampleRandomFileSize(String filename)
		throws GridException {

		CommonMath.fs_rand.setSeed(filename.hashCode());
		
		return 1024 + CommonMath.fs_rand.nextInt(2 << 22);
		
	}
	
	/** Returns a random boolean.
	 * @return random boolean
	 */
	public synchronized static boolean sampleRandomBoolean() {
		
		return CommonMath.rand.nextBoolean();
		
	}

	// ****************************************************************

	public static String div_label(long a) {
		
		return div_label(a,1);
		
	}
	
	public static String div_label(long a,long b) {
		
		return div_label((double)a,(double)b,1.0);
		
	}

	public static String div_label(double a) {
		
		return div_label(a,1.0);
		
	}

	public static String div_label(double a,double b) {
		
		return div_label(a,b,1.0);
		
	}

	public static String div_label(long a,long b,double factor) {

		return CommonMath.div_label((double)a,(double)b,factor);
		
	}
	
	public static String div_label(double a,double b,double factor) {

		if (b == 0) return "INFINITY";
		
		return Long.toString(Math.round(a/b*factor));
		
	}

	// ****************************************************************

	public static String div_label_2dd(long a) {

		return CommonMath.div_label_2dd((double)a,1.0);
	
	}
	
	public static String div_label_2dd(long a,long b) {
		
		return CommonMath.div_label_2dd((double)a,(double)b,1.0);
		
	}

	public static String div_label_2dd(long a,long b,double factor) {

		return CommonMath.div_label_2dd((double)a,(double)b,factor);
		
	}

	public static String div_label_2dd(double a) {

		return CommonMath.div_label_2dd(a,1.0);
		
	}

	public static String div_label_2dd(double a,double b) {

		return CommonMath.div_label_2dd(a,b,1.0);
		
	}
	
	public static String div_label_2dd(double a,double b,double factor) {

		if (b == 0) return "INFINITY";
		
		return new Double(Math.round(a/b*factor*100.0)/100.0).toString();
		
	}

	// ****************************************************************

	/** Computes the mean of the two taget values.
	 * @return mean of the two target values
	 */
	public static double mean(double x1,double x2) {
	
		return x1 + ((x2-x1) / 2.0);
		
	}
	
	/** Computes the standard deviation of N observations, given their
	 * sum (i.e. sum of the x_i) and squared sum (i.e. sum of the (x_i)^2).
	 * @param sum_x sum of the observations
	 * @param sum_squared_x sum of the squared observations
	 * @param N number of observations
	 * @return standard deviation of the observations
	 * @exception GridException if a parameter is illegal
	 * @deprecated
	 */
	@Deprecated
	public static double stddev(double sum_x,double sum_squared_x,long N)
		throws GridException {
		
		if (sum_squared_x < 0) {
			throw new GridException("stddev(): "+
									"negative sum of squared observations");
		}
		if (N <= 0) {
			throw new GridException("stddev(): "+
									((N == 0) ? "no observation" : "negative "+
									"number of observations"));
		}
		
		// std_dev = sqrt( E(X^2) - ((E(X))^2 )
		
		double mean_of_squared_obs = sum_squared_x / ((double)N);
			
		double mean_of_obs = sum_x / ((double)N);

		return Math.sqrt(mean_of_squared_obs - (mean_of_obs * mean_of_obs));
		
	}

	/** Computes the standard deviation of N observations, given their
	 * sum (i.e. sum of the x_i) and squared sum (i.e. sum of the (x_i)^2).
	 * <p>Note that this implementation does not throw an Exception,
	 * but instead returns a conventional, illegal value.
	 * @param sum_x sum of the observations
	 * @param sum_squared_x sum of the squared observations
	 * @param N number of observations
	 * @return standard deviation of the observations,
	 * -1.0 if a parameter is illegal
	 * @deprecated
	 */
	@Deprecated
	public static double uncheckedStddev(double sum_x,double sum_squared_x,
										 long N) {
			
		double retval = -1.0;
		try {
			retval = stddev(sum_x,sum_squared_x,N);
		}
		catch (GridException e) {
			e.printStackTrace();
			System.out.println(e);
		}
		
		return retval;

	}
	
	// ****************************************************************

	/** Filters target double-precision floating point number
	 * to zero if its value is not a number,
	 * else leaves it unchanged.
	 * @param d target double-precision floating point number
	 * @return target number itself, or zero if target number is not a number
	 */
	public static final double filterToZeroIfNaN(double d) {

		return (Double.isNaN(d)) ? 0 : d;
		
	}
	
	// ****************************************************************

	/** Normalizes target base value with respect to target maximum value.
	 * <p>Note: target base value greater than target maximum value
	 * are considered equal to target maximum value.
	 * @param base_val target base value
	 * @param max_val target maximum value
	 * @return normalized (i.e. included in [0.0..1.0]) target base value
	 * @exception GridException if a parameter is illegal
	 */
	public static double normalizeValue(double base_val,double max_val)
		throws GridException {
		
		if (base_val < 0.0) {
			throw new GridException("normalizeValue(): illegal base value");
		}
		if (max_val < 0.0) {
			throw new GridException("normalizeValue(): illegal maximum value");
		}
		
		return (max_val <= base_val) ?
			0.0 : ((max_val - base_val) / max_val);

	}

	/** Normalizes target base value with respect to target maximum value.
	 * @param base_val target base value
	 * @param max_val target maximum value
	 * @return normalized (i.e. included in [0.0..1.0]) target base value
	 * @exception GridException if target base value is greater
	 * than target maximum value, or if a parameter is illegal
	 */
	public static double normalizeCheckedValue(double base_val,double max_val)
		throws GridException {
		
		if (base_val > max_val) {
			throw new GridException("normalizeValue(): "+
									"base value greater than maximum value");
		}
		
		return normalizeValue(base_val,max_val);
		
	}

	// ****************************************************************

	/** Returns a hash code for given number.
	 * <p>The hash code is guaranteed to be different from Integer.MIN_VALUE,
	 * which can facilitate caching and lazy initilization of hash codes.
	 * @param target target number
	 * @return hash code
	 */
	public static int generateHashCode(int target) {
		
		int h = target + 0x1337;
		
		return (h == Integer.MIN_VALUE) ? ++h : h;
		
	}

	/** Returns a hash code for given number.
	 * <p>The hash code is guaranteed to be different from Integer.MIN_VALUE,
	 * which can facilitate caching and lazy initilization of hash codes.
	 * @param target target number
	 * @return hash code
	 */
	public static int generateHashCode(long target) {
		
		int h = (int)(target ^ (target >>> 32)) + 0x1337;
		
		return (h == Integer.MIN_VALUE) ? ++h : h;
		
	}

	/** Returns a hash code for given String.
	 * <p>The hash code is guaranteed to be different from Integer.MIN_VALUE,
	 * which can facilitate caching and lazy initilization of hash codes.
	 * @param target target String
	 * @return hash code
	 */
	public static int generateHashCode(String target) {
		
		int h = ((target != null) ? target.hashCode() : 0) + 0x1337;
		
		return (h == Integer.MIN_VALUE) ? ++h : h;
		
	}

	/** Returns a hash code for given pair of numbers.
	 * <p>The hash code is guaranteed to be different from Integer.MIN_VALUE,
	 * which can facilitate caching and lazy initilization of hash codes.
	 * @param target_principal target number
	 * @param target_secondary target number
	 * @return hash code
	 */
	public static int generateHashCode(int target_principal,
									   int target_secondary) {
		
		int dummy = target_secondary % 0x400;
		int h = ((target_principal << 10) + dummy) + 0x1337;
		
		return (h == Integer.MIN_VALUE) ? ++h : h;
		
	}

	// ****************************************************************

	public synchronized static void setSeed(long seed) {
		
		CommonMath.rand = new Random(seed);
		
	}

	// ****************************************************************

	static {
	
		CommonMath.rand    = new Random(System.currentTimeMillis());
		CommonMath.fs_rand = new Random(42);
	
	}
	
	private static Random rand;
	private static Random fs_rand;

}

