/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DescriptiveMeasurement.java

package lbg.common.math;

import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonFileHandling.*;
import java.util.Properties;
import java.util.StringTokenizer;
import org.apache.commons.math.stat.descriptive.DescriptiveStatistics;

/**
 * @author Cyril Briquet
 */
public class DescriptiveMeasurement extends Measurement {

	/** Creates a new DescriptiveMeasurement.
	 * @param label measurement  label
	 * @exception GridException if a parameter is illegal
	 */
	public DescriptiveMeasurement(String label) throws GridException {
		
		super(label);
		
		this.sum_x_all = 0.0;
		this.sum_squared_x_all = 0.0;
		this.x_i_min = Double.MAX_VALUE;
		this.x_i_max = Double.MIN_VALUE;
		this.x_i_min_tag = null;
		this.x_i_max_tag = null;
		this.n_obs = 0;
		this.stats = DescriptiveStatistics.newInstance();
		
	}

	/** Creates a new DescriptiveMeasurement.
	 * @param label measurement  label
	 * @param ws window size
	 * @exception GridException if a parameter is illegal
	 */
	public DescriptiveMeasurement(String label,int ws)
		throws GridException {
		
		this(label);
		
		if (ws < 1) {
			throw new GridException("DescriptiveMeasurement(): "+
									"illegal window size");
		}
		
		this.stats.setWindowSize(ws);
		
	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[Measurement","[DescriptiveMeasurement"));
		
		int nm5 = (int)this.stats.getN()-5;
		if (nm5 >= 1) {
			sb.append("mean (N-5)=");
			sb.append(this.getMeanOfFirstKTermsWithoutException(nm5));
			sb.append("\n");
		}

		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized void addObservation(double x_i)
		throws GridException {
		
		addObservation(x_i,"dummy-tag");

	}
		
	public synchronized void addObservation(double x_i,String tag)
		throws GridException {
		
		if (Double.isNaN(x_i)) {
			throw new GridException("addObservation(): illegal observation");
		}
		if (tag == null) {
			throw new GridException("addObservation(): illegal tag");
		}
		
		this.sum_x_all += x_i;
		this.sum_squared_x_all += (x_i * x_i);
		this.n_obs++;
		this.stats.addValue(x_i);
		
		if (x_i < this.x_i_min) {
			this.x_i_min = x_i;
			this.x_i_min_tag = tag;
		}
		if (x_i > this.x_i_max) {
			this.x_i_max = x_i;
			this.x_i_max_tag = tag;
		}
		
	}
	
	// ****************************************************************

	public synchronized long getObservations() {
		
		return this.n_obs;
		
	}

	public synchronized double getSum() {

		
		return this.sum_x_all;
	
	}
	
	public synchronized double getSquaredSum() {

		
		return this.sum_squared_x_all;
	
	}
	
	public synchronized double getMin() {
	
		return this.x_i_min;
		
	}

	public synchronized double getMax() {
	
		return this.x_i_max;
		
	}

	public synchronized String getMinTag() {
	
		return this.x_i_min_tag;
		
	}

	public synchronized String getMaxTag() {
	
		return this.x_i_max_tag;
		
	}

	public synchronized double getMean() {

		return this.sum_x_all / this.n_obs;
		
	}
	
	public synchronized double getMeanOfFirstKTerms(long k)
		throws GridException {
		
		if (k < 1) {
			throw new GridException("getMeanOfFirstKTerms(): "+
									"illegal terms count");
		}

		double s_k = 0.0;
		int lim = (int)Math.min(k,this.stats.getN());
		for (int i = 0 ; i < lim ; ++i) {
			s_k += this.stats.getElement(i);
		}

		return s_k / (double)lim;
		
	}
	
	private synchronized double getMeanOfFirstKTermsWithoutException(long k) {
	
		double retval = Double.NaN;
		try {
			retval = getMeanOfFirstKTerms(k);
		}
		catch (GridException e) {
			retval = Double.NaN;
		}
		return retval;
		
	}
	
	public synchronized double getStandardDeviation() {

		return this.stats.getStandardDeviation();
		
	}
	
	// ****************************************************************

	private double sum_x_all;
	private double sum_squared_x_all;
	private double x_i_min;
	private double x_i_max;
	private String x_i_min_tag;
	private String x_i_max_tag;
	private long n_obs;
	private final DescriptiveStatistics stats;

}

