/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Measurement.java

package lbg.common.math;

import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonFileHandling.*;
import java.util.Properties;
import java.util.StringTokenizer;

// See Jakarta Commons Math
//   http://jakarta.apache.org/commons/math/userguide/stat.html
//   http://jakarta.apache.org/commons/math/api-1.1/index.html

/**
 * @author Cyril Briquet
 */
public abstract class Measurement {

	protected Measurement(String label) throws GridException {

		if (label == null) {
			throw new GridException("SummarizedMeasurement(): illegal label");
		}
		
		this.label = label;
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[Measurement]");
		sb.append("\n");
		sb.append("label=");
		sb.append(this.label);
		sb.append("\n");
		sb.append("observations=");
		sb.append(this.getObservations());
		sb.append("\n");
		sb.append("mean=");
		sb.append(this.getMean());
		sb.append("\n");
		sb.append("std_dev=");
		sb.append(this.getStandardDeviation());
		sb.append("\n");
		sb.append("sum_x_all=");
		sb.append(this.getSum());
		sb.append("\n");
		sb.append("sum_squared_x_all=");
		sb.append(this.getSquaredSum());
		sb.append("\n");
		sb.append("x_i_min=");
		sb.append(this.getMin());
		sb.append("\n");
		sb.append("x_i_max=");
		sb.append(this.getMax());
		sb.append("\n");
	
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized void addObservation(double x_i)
		throws GridException {
		
		addObservation(x_i,"dummy-tag");

	}
		
	public abstract void addObservation(double x_i,String tag)
		throws GridException;
	
	// ****************************************************************

	public synchronized String getLabel() {
		
		return this.label;
		
	}
	
	public synchronized long getN() {
		
		return getObservations();
		
	}

	// ****************************************************************

	public abstract long getObservations();


	public abstract double getSum();
	
	public abstract double getSquaredSum();

	public abstract double getMin();

	public abstract double getMax();
	
	public abstract String getMinTag();

	public abstract String getMaxTag();

	public abstract double getMean();
	
	public abstract double getStandardDeviation();
	
	// ****************************************************************

	public static Measurement
		createMetaMeasurement(String list_of_measurements)
		throws GridException {
		
		if (list_of_measurements == null) {
			throw new GridException("createMetaMeasurement(): illegal "+
									"list of measurement filenames");
		}

		Measurement meta_measurement =
			MeasurementFactory.createSummarizedMeasurement("meta-measurement");
		
		String s = CommonFileHandling.readFile(list_of_measurements);
		if (s == null) {
			throw new GridException("createMetaMeasurement(): illegal "+
									"list of measurement filenames");
		}
		StringTokenizer st = new StringTokenizer(s);
		while (st.hasMoreTokens()) {
			String measurement_filename = st.nextToken();
			if ((measurement_filename == null) ||
				(measurement_filename.length() < 1)) {
			throw new GridException("createMetaMeasurement(): illegal "+
									"measurement filename");
			}
			
			try {
				Properties props =
					CommonFileHandling.loadProperties(measurement_filename);
				Measurement m = new SummarizedMeasurement(props);
				if ((m.getObservations() > 0) ||
					(Double.isNaN(m.getMean()) == false)) {
					meta_measurement.addObservation(m.getMean());
				}
			}
			catch (GridException e) {
				System.out.println("*WARNING* - failed to load "+
								   measurement_filename);
			}
		}

		return meta_measurement;
		
	}
	
	// ****************************************************************

	public static void test_main(String[] args) {
		
		try {
			Measurement m = MeasurementFactory.
				createSummarizedMeasurement("task");
			
			m.addObservation(10);
			m.addObservation(8);
			m.addObservation(12);
			m.addObservation(12);
			
			System.out.println(m);
			
			m = MeasurementFactory.
				createDescriptiveMeasurement("task");
			
			m.addObservation(10);
			m.addObservation(8);
			m.addObservation(12);
			m.addObservation(12);
			
			System.out.println(m);
		}
		catch (GridException e) {
			e.printStackTrace();
		}
		finally {
			System.exit(0);
		}
		
	}

	public static void main(String[] args) {
		
		try {
			checkArgs(args);
			
			System.out.println();
			
			Measurement m = Measurement.createMetaMeasurement(args[0]);
			
			System.out.println();
			System.out.println(m);
		}
		catch (GridException e) {
			e.printStackTrace();
		}
		finally {
			System.exit(0);
		}
		
	}

	// ****************************************************************

	private static void checkArgs(String[] args) throws GridException {
	
		if ((args.length != 1) ||
			(args[0] == null) ||
			(args[0].length() < 1+TXT_EXT.length())) {
			System.out.println("Usage: "+
							   "java -cp classpath "+
							   Measurement.class.getName()+
							   " list_of_measurements.txt\n");
			throw new GridException("checkArgs(): illegal arguments");
		}
		
	}

	// ****************************************************************

	private final String label;

}

