/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// BitTorrentLeechersSetDownloadTime.java

package lbg.common.net;

import lbg.common.exception.*;

/**
 * @author Cyril Briquet
 */
public class BitTorrentLeechersSetDownloadTime {

	public BitTorrentLeechersSetDownloadTime() throws GridException {
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		sb.append("[BitTorrentLeechersSetDownloadTime]");

		return sb.toString();
		
	}
	
	/** Predicts the average download time of target file
	 * by the target leechers set using the BitTorrent protocol.
	 * <p><b>WARNING: experimental parameters, use with great caution.
	 * @param s target file size (in megabytes)
	 * @param n number of leechers
	 * @return predicted download time using BitTorrent
	 */
	public double myPredict(double s,int n)
		throws GridException {
	
		return predict(s,n,20,BETA_SENDER/(2*n));
		
	}

	/** Predicts the average download time of target file
	 * by the target leechers set using the BitTorrent protocol.
	 * @param s target file size (in megabytes)
	 * @param n number of leechers
	 * @param alpha Latency of the BitTorrent protocol (in seconds).
	 * @param beta Available bandwidth (in megabytes / seconds)
	 * of the initial seeder.
	 * @return predicted download time using BitTorrent
	 */
	public double predict(double s,int n,double alpha,double beta)
		throws GridException {
		
		if ((s <= 0.0) || (s > 2000.0)) { // 2 Gigabytes is enough for now ;)
			throw new GridException("predict(): illegal file size "+
									s+" megabytes");
		}
		if ((n < 1) || (n > 10000)) {
			throw new GridException("predict(): illegal number of leechers "+n);
		}
		if ((alpha < 0.0) || (n > 1000.0)) {
			throw new GridException("predict(): illegal BT latency "+alpha);
		}
		if ((beta <= 0.0) || (n > 2000.0)) {
			throw new GridException("predict(): illegal bandwidth "+
									beta+" megabytes / second");
		}
		
		// T(S,N) = (alpha_bt + ((a * S) / beta_sender)) * (ln N / ln a)
		// Source: "Scheduling Independent Tasks Sharing Large Data
		// 			Distributed with BitTorrent", Wei et al., 2005
		
		double a = 5.0;
		
		double f1 = alpha + ((a * s) / beta);
		double f2 = Math.log((double)n) / Math.log(a);
		
		return f1 * f2;
		
	}
	
	/** Latency of the BitTorrent protocol (in seconds).
	 */
	public static final double ALPHA_BT = 5.0;
	
	/** Available bandwidth (in megabytes / seconds) of the initial seeder.
	 */
	public static final double BETA_SENDER = 100.0/8.0;

	// ****************************************************************

	public static void main(String[] args) {
	
		System.out.println("BT leechers set download time modelling\n");
		System.out.println("Usage: java "+
						   BitTorrentLeechersSetDownloadTime.class.getName()+
						   " file_size leechers_number\n");
		
		try {
			BitTorrentLeechersSetDownloadTime bt_model =
				new BitTorrentLeechersSetDownloadTime();
			
			int s = Integer.parseInt(args[0]);
			int n = Integer.parseInt(args[1]);
			double prediction = bt_model.myPredict(s,n);
			System.out.println("T("+s+","+n+") = "+prediction);
			
		}
		catch (GridException e) {
			System.out.println(e);
			System.exit(0);	
		}
		
	}
	
}

