/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LinkedListDeque.java

package lbg.common.queueing.deque;

import lbg.common.exception.*;
import java.util.Iterator;

/** Implements a Deque using a doubly linked list.
 * <p>Original, unlicensed version written by Geoff Knagge (9806135),
 * last modified 14/8/1999.
 * <p>This version synthesized and heavily patched by Cyril Briquet,
 * last modified 04/02/2007.
 * <p>Note: the sole purpose of this class is to provide
 * a J2SE 5.0 implementation of a Deque.
 * @author Cyril Briquet
 */
public class LinkedListDeque<T> implements Deque<T> {

	public LinkedListDeque() {

		this.head = new ListNode<T>(null,null,null);
		this.tail = new ListNode<T>(null,null,head); // head is prev node
		this.head.setNext(this.tail);
		this.dequeSize = 0;

	}

	// ****************************************************************

	public synchronized boolean isEmpty() {
		
		// returns true if there are no nodes between the head and tail
		return this.dequeSize == 0;
		
	}

	public synchronized int size() {
		
		// returns the number of nodes between the head and the tail
		return this.dequeSize;
		
	}

	// ****************************************************************

	/** Inserts target Object at the head of the Deque.
     */
	public synchronized void insertFirst(T o) throws GridException {

		if (o == (T)null) {
			throw new GridException("insertFirst(): illegal object");
		}
		
		/* Inserts a new ListNode containing target Object
		   between the head sentinel
		   and the node that the head was originally linked to.
		 */

		ListNode<T> newNode = new ListNode<T>(o,this.head.getNext(),this.head);
		this.head.getNext().setPrev(newNode);
		this.head.setNext(newNode);
		this.dequeSize++;

	}

	/** Inserts target Object at the tail of the Deque.
     */
	public synchronized void insertLast(T o) throws GridException {

		if (o == (T)null) {
			throw new GridException("insertLast(): illegal object");
		}
		
		/* Inserts a new ListNode containing target Object
		   between the tail sentinel
		   and the node that the tail was originally linked to.
		 */
	
		ListNode<T> newNode = new ListNode<T>(o,this.tail,this.tail.getPrev());
		this.tail.getPrev().setNext(newNode);
		this.tail.setPrev(newNode);
		this.dequeSize++;

	}

	/** Returns the object to the head of the Deque, and removes that Object.
	 */
	public synchronized T removeFirst() throws GridException {

		if (this.dequeSize == 0) {
			throw new GridException("removeFirst(): empty Deque");
		}
		
		ListNode<T> node = this.head.getNext();
		this.head.setNext(node.getNext());
		node.getNext().setPrev(this.head);
		this.dequeSize--;
		
		return node.getElement();
		
	}

	/** Returns the object to the tail of the Deque, and removes that Object.
	 */
	public synchronized T removeLast() throws GridException {

		if (this.dequeSize == 0) {
			throw new GridException("removeLast(): empty Deque");
		}
		
		ListNode<T> node = this.tail.getPrev();
		this.tail.setPrev(node.getPrev());
		node.getPrev().setNext(this.head);
		this.dequeSize--;
		return node.getElement();
		
	}

	// ****************************************************************
	
	/** Returns the object to the head of the Deque.
	 */
	public synchronized T firstElement() throws GridException {

		if (this.dequeSize == 0) {
			throw new GridException("firstElement(): empty Deque");
		}
		
		return head.getNext().getElement();
		
	}

	/** Returns the object to the tail of the Deque.
	 */
	public synchronized T lastElement() throws GridException {

		if (this.dequeSize == 0) {
			throw new GridException("lastElement(): empty Deque");
		}
		
		return tail.getPrev().getElement();
		
	}
	
	// ****************************************************************

	public synchronized Iterator<T> iterator() {

		return new ListNodeAscendingIterator<T>(this.head);		
		
	}

	public synchronized Iterator<T> descendingIterator() {

		return new ListNodeDescendingIterator<T>(this.tail);		
		
	}

	// ****************************************************************

	private int dequeSize;
	private ListNode<T> head;
	private ListNode<T> tail;
	
	// ****************************************************************

	private class ListNode<E> {
	
		public ListNode(E item,ListNode<E> newNext,ListNode<E> newPrev) {
			
			this.thisNode = item;    // the object that the node points to
			this.next = newNext;	 // the next node		
			this.previous = newPrev; // the preceding node
			
		}
	
		// ****************************************************************
	
		// returns the object pointed to by the node
		public synchronized E getElement() {
	
			return this.thisNode;
	
		}
	
		// returns the node which follows this one
		public synchronized ListNode<E> getNext() {
	
			return this.next;
	
		}
	
		// returns the node which preceeds this one
		public synchronized ListNode<E> getPrev() {
	
			return this.previous;
	
		}
	
		// ****************************************************************
	
		// specify the object to be pointed to by the node
		public synchronized void setElement(E item) {
	
			this.thisNode = item;
	
		}
	
		// specify which node follows the current one
		public synchronized void setNext(ListNode<E> newNext) {
		
			this.next = newNext;
			
		}
	
		// specify which node precedes the current one
		public synchronized void setPrev(ListNode<E> newPrev) {
	
			this.previous = newPrev;
	
		}
	
		// ****************************************************************
	
		private E thisNode;
		private ListNode<E> next;
		private ListNode<E> previous;
	
	}
	
	// ****************************************************************

	private class ListNodeAscendingIterator<E> implements Iterator<E> {
		
		public ListNodeAscendingIterator(ListNode<E> head) {

			this.cur = head.getNext();
			
		}

		public boolean hasNext() {
			
			return ((this.cur != null) && (this.cur.getElement() != null));
			
		}
		
		public E next() {

			if ((this.cur == null) || (this.cur.getElement() == null)) {
				return null;
			}
			
			ListNode<E> dummy = this.cur;
			this.cur = this.cur.getNext();
			
			return dummy.getElement();
			
		}
		
		public void remove() {
		
			
			throw new UnsupportedOperationException();
			
		}
		
		private ListNode<E> cur;
		
	}

	private class ListNodeDescendingIterator<E> implements Iterator<E> {
		
		public ListNodeDescendingIterator(ListNode<E> tail) {
		
			this.cur = tail.getPrev();
			
		}

		public boolean hasNext() {
			
			return ((this.cur != null) && (this.cur.getElement() != null));
			
		}
		
		public E next() {

			if ((this.cur == null) || (this.cur.getElement() == null)) {
				return null;
			}
			
			ListNode<E> dummy = this.cur;
			this.cur = this.cur.getPrev();
			
			return dummy.getElement();
			
		}
		
		
		public void remove() {
		
			
			throw new UnsupportedOperationException();
			
		}
		
		private ListNode<E> cur;
		
	}

	// ****************************************************************

}

