/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DoublyLinkedList.java

package lbg.common.queueing.priority;

import lbg.common.exception.*;

/** Helper class used to create double-linked lists.
 * @author Cyril Briquet
 */
public class DoublyLinkedList<E> {
		
	public DoublyLinkedList() {

		this.size = 0;
		this.head = null;
		this.tail = null;
		
	}
		
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[DoublyLinkedList]" );
		sb.append(this.size);
		sb.append(" elements");
		for (int i = 0 ; i < this.size ; i++) {
			DoublyLinkedNode<E> p = this.head;
			sb.append(p);
			p = p.getNext();			
		}
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized int size() {
		
		return this.size;
		
	}
	
	public synchronized DoublyLinkedNode<E> head() {
		
		return this.head;
		
	}
	
	public synchronized DoublyLinkedNode<E> tail() {
		
		return this.tail;
		
	}

	/** Appends target element to the tail of the doubly-linked list.
	 * @param e target element
	 * @return doubly-linked node that was appended
	 * @exception GridException if target element is illegal
	 */
	public synchronized DoublyLinkedNode<E> append(E e) throws GridException {
		
		assert(sanityCheck());

		if (e == (E)null) {
			throw new GridException("append(): illegal element");
		}
	
		this.size++;
		
		if (head == null) {
			this.head = this.tail = new DoublyLinkedNode<E>(e);
		}
		else {
			DoublyLinkedNode<E> p = new DoublyLinkedNode<E>(e,tail);
			this.tail.setNext(p);
			this.tail = p;
		}
		
		assert(sanityCheck());

		return this.tail;
		
	}

	/** Appends target element to the tail of the doubly-linked list,
	 * removing as many elements as required from the head of the list
	 * so that its size does not exceed the target limit,
	 * after appending the target element.
	 * @param e target element
	 * @param limit target size
	 * @return doubly-linked node that was appended
	 * @exception GridException if target element is illegal
	 */
	public synchronized DoublyLinkedNode<E> appendAndTrim(E e,int limit)
		throws GridException {

		assert(sanityCheck());

		if (limit < 1) {
			throw new GridException("append(): illegal limit");
		}
		
		append(e);
		
		while (limit < this.size) {
			removeHead();
		}
		
		assert(sanityCheck());

		return this.tail;
		
	}
	
	/** Removes the head element from the head of the doubly-linked list.
	 * @exception GridException if internal data is corrupted
	 */
	public synchronized void removeHead() throws GridException {
		
		assert(sanityCheck());
		
		this.size--;

		if (head != null) {
			DoublyLinkedNode<E> p = this.head;
			this.head = this.head.getNext();
			this.head.setPrev(null);
			p.setNext(null);
		}

		assert(sanityCheck());
		
	}

	// ****************************************************************

	private boolean sanityCheck() {
		
		if (this.size == 0) {
			return ((this.head == null) && (this.tail == null));
		}
		else {
			return ((this.head != null) && (this.tail != null));
			
		}
		
	}

	// ****************************************************************

	private int size;
	private DoublyLinkedNode<E> head;
	private DoublyLinkedNode<E> tail;
	
}

