/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PriorityQueue.java

package lbg.common.queueing.priority;

import lbg.common.exception.*;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.TreeMap;

/** Synchronized implementation of an unbounded PriorityQueue
 * backed by a balanced tree and a doubly-linked list.
 * <p>Nominal use is to keep a set of elements sorted
 * with respect to their order of insertion.
 * <p>This class has been developped to replace LinkedList.
 * They may both be used to implement Tasks queues in the QueueManager.
 * As inserting an element into such a queue is legal
 * only if the element is not already present, 
 * it is required that a search operation is also performed.
 * The cost of the search operation in LinkedList is linear,
 * while it is logarithmic in PriorityQueue, because the latter is backed
 * by a balanced tree.
 * <p>IMPORTANT NOTE: as int != Integer, it is preferred not to use int keys
 * when Integer keys are expected, to avoid that autoboxing
 * skews Object.equals() tests.
 * <p>NOTE: current implementation is not robust to non-Comparable keys.
 * @author Cyril Briquet
 */
/*
 * Execution times of different scenarios,
 * with increasing number of Jobs per Peer and a RandomNegotiator.
 * <pre>
Jobs/Peer |   Self-Ref bug |   LinkedList    | PriorityQueue
   5      | =   10 s       | =     3 s       | =  2-3 s
  20 (x4) | =   80 s ( x8) | =    12 s (x4)  | =    9 s (x4)
  80 (x4) | = 1840 s (x23) | =   103 s (x9)  | =   37 s (x4)
 320 (x4) | =    /         | =  1376 s (x13) | =  150 s (x4)
1280 (x4) | =    /         | = 35133 s (x26) | = 2269 s (x15, but thrashing)
</pre>
 * <p>If some numeric ID type is used as a key,
 * it is important to use a wrapper class instead of a primitive type,
 * because multiple type promotions may lead
 * to the creation of different objects
 * indeed wrapping the same integer number but which won't compare identically.
*/
public class PriorityQueue<E> implements Iterable<E> {

	/** Builds a new PriorityQueue.
	 */
	public PriorityQueue() {

		this.tm = new TreeMap<E,DoublyLinkedNode<E>>();
		this.ptr_first = null;
		this.ptr_last = null;
		
	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PriorityQueue]");
		
		return sb.toString();
		
	}
	
	// ****************************************************************

	/** Returns the number of elements in the PriorityQueue.
	 * @return the number of elements in the PriorityQueue
	 */
	public synchronized int size() {
		
		return this.tm.size();
		
	}
	
	/** Returns an iterator over the elements in the PriorityQueue.
	 * The elements are returned by decreasing priority.
	 * @return an iterator over the elements in the PriorityQueue
	 */
	public synchronized Iterator<E> iterator() {

		return new PriorityQueueIterator<E>(this.ptr_first);
		
	}

	/** Returns if target element is stored in the PriorityQueue.
	 * @return true if target element is stored in the PriorityQueue
	 */
	public synchronized boolean hasElement(E e ) {

		if (e == (E)null) {
			throw new NullPointerException("hasElement(): null element");
		}

		return (this.tm.get(e) != null);
		
	}

	/** Removes then returns
	 * the element with the higest priority in the PriorityQueue.
	 * @return the element with the higest priority in the PriorityQueue,
	 * null if there is no element in the PriorityQueue
	 */
	public synchronized E extractHighestPriorityElement() throws GridException {
		
		assert(sanityCheck());

		if (this.tm.size() == 0) {
			return (E)null;
		}
		
		DoublyLinkedNode<E> ptr_node =
			this.tm.remove(this.tm.firstKey());
			
		if (ptr_node == null) {
			throw new GridException("extractHighestPriorityElement(): "+
									"corrupted data");
		}
		
		assert(sanityCheck());

		return ptr_node.info();
		
	}

	/** Adds target element to the end of the PriorityQueue,
	 * if not already present.
	 * @param e target element
	 * @return true if target element was not already present
	 * in the PriorityQueue
	 * @exception GridException if data has been corrupted
	 * @exception NullPointerException if target element is null
	 */
	public synchronized boolean add(E e)
		throws GridException, NullPointerException {
		
		if (e == (E)null) {
			throw new NullPointerException("add(): null element");
		}

		assert(sanityCheck());
		
		if (this.tm.get(e) != null) {
			return false;
		}

		DoublyLinkedNode<E> ptr_node =
			new DoublyLinkedNode<E>(e,this.ptr_last);
		if (this.tm.put(e,ptr_node) != null) {
			throw new GridException("add(): corrupted data");
		}
		
		if (this.ptr_first == null) {
			assert(this.ptr_last == null);
			this.ptr_first = ptr_node;
		}
		if (this.ptr_last != null) {
			this.ptr_last.setNext(ptr_node);
		}
		this.ptr_last = ptr_node;

		assert(sanityCheck());
		
		return true;
		
	}

	/* -- * Adds target element to the PriorityQueue,
	 * at the index computed its priority.
	 * @param e target element
	 * @return true if target element was not already present
	 * in the PriorityQueue
	 * @exception NullPointerException if target element is null
	 * -- /
	 
	 THIS METHOD SHOULD BE EXPLICITLY TESTED BEFORE IT IS RELEASED
	 
	private synchronized boolean addAccordingToPriority(E e)
		throws NullPointerException {
		
		if (e == (E)null) {
			throw new NullPointerException("add(): null element");
		}

		assert(sanityCheck());

		// Empty PriorityQueue
		if (this.tm.size() == 0) {
			return add(e);
		}
		
		// Discover element with immediately lower priority
		DoublyLinkedNode<E> ptr_node_next =
			this.tm.get(this.tm.headMap(e).firstKey());

		// There is no element with lower priority
		if (ptr_node_next == null) {
			return add(e);
		}
		
		// There are elements with lower priority
		boolean retval = true;
		DoublyLinkedNode<E> ptr_node_prev = ptr_node_next.getPrev();
		DoublyLinkedNode<E> ptr_node =
			new DoublyLinkedNode<E>(e,ptr_node_prev,ptr_node_next);
		if (this.tm.put(e,ptr_node) != null) {
			retval = false;
		}

		// There is no element with higher prioity
		if (ptr_node_next == this.ptr_first) {
			assert(ptr_node_prev == null);
			this.ptr_first = ptr_node;
			ptr_node_next.setPrev(ptr_node);
		}
		// There are elements with higher prioity
		else // if (ptr_node_next != this.ptr_first) {
			assert(ptr_node_prev != null);
			ptr_node_prev.setNext(ptr_node);
			ptr_node_next.setPrev(ptr_node);
		}
		
		assert(sanityCheck());

		return retval;
		
	}
	*/

	/** Removes target element from the PriorityQueue if it is present.
	 * @param e target element
	 * @return true if target element was present
	 * in the PriorityQueue
	 * @exception NullPointerException if target element is null
	 */
	public synchronized boolean remove(E e) throws NullPointerException {
		
		if (e == (E)null) {
			throw new NullPointerException("remove(): null element");
		}
		
		assert(sanityCheck());

		DoublyLinkedNode<E> ptr_node = this.tm.remove(e);
		if (ptr_node == (E)null) {
			return false;
		}
		
		DoublyLinkedNode<E> ptr_prev = ptr_node.getPrev();
		DoublyLinkedNode<E> ptr_next = ptr_node.getNext();

		if (ptr_prev != null) {
			ptr_prev.setNext(ptr_next);
		}
		else {
			// The removed element was the first element
			this.ptr_first = ptr_next;
		}
		if (ptr_next != null) {
			ptr_next.setPrev(ptr_prev);
		}
		else {
			// The removed element was the last element
			this.ptr_last = ptr_prev;
		}

		assert(sanityCheck());

		return true;
		
	}
	
	// ****************************************************************

	private synchronized boolean sanityCheck() {
		
		if (this.tm.size() == 0) {
			return ((this.ptr_first == null) &&
					(this.ptr_last == null));
		}
		else {
			return ((this.ptr_first != null) &&
					(this.ptr_last != null));
			
		}
		
	}

	// ****************************************************************

	private TreeMap<E,DoublyLinkedNode<E>> tm;
	private DoublyLinkedNode<E> ptr_first;
	private DoublyLinkedNode<E> ptr_last;

	// ****************************************************************

	/** Helper class of PriorityQueue.
	 */
	private class PriorityQueueIterator<PQI_E> implements Iterator<PQI_E> {
	
		/** Builds a new PriorityQueueIterator.
		 * @param e first element of the PriorityQueue.
		 */
		public PriorityQueueIterator(DoublyLinkedNode<PQI_E> e) {
	
			this.ptr_current = e;
			
		}
		
		public synchronized String toString() {
			
			StringBuilder sb = new StringBuilder();
			
			sb.append("[PriorityQueueIterator]");
			if (this.ptr_current != null) {
				sb.append("\n  ->  prev=");
				sb.append(this.ptr_current.getPrev());
				sb.append("\n  ->  next=");
				sb.append(this.ptr_current.getNext());
			}
			
			return sb.toString();
			
		}
	
		// ****************************************************************
	
		/** Iterator implementation.
		 */
		public synchronized boolean hasNext() {
			
			return (this.ptr_current != null);
			
		}
			 
		/** Iterator implementation.
		 */
		public synchronized PQI_E next() {
	
			if (this.ptr_current == (null)) {
				throw new NoSuchElementException("next(): no more elements");
			}
			
			PQI_E retval = this.ptr_current.info();
			this.ptr_current = this.ptr_current.getNext();
			
			return retval;
			
		}
		
		/** Iterator implementation.
		 */
		public synchronized void remove() {
		
			throw new UnsupportedOperationException("remove(): "+
													"illegal operation");
			
		}
	
		// ****************************************************************
	
		private DoublyLinkedNode<PQI_E> ptr_current;
	
	}
	
}

