/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// AutomaticJobSubmitter.java

package lbg.middleware;

import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.common.math.*;
import static lbg.common.io.CommonFileHandling.FILE_SEP;
import lbg.middleware.util.*;
import lbg.peer.service.*;
import lbg.peer.workflow.*;
import lbg.protocol.data.*;
import lbg.protocol.rpc.*;
import lbg.task.Task;
import lbg.user.*;

import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.logging.Logger;

/** Utility software to automatically submit, multiple times, one synthetic Job.
 * @author Xavier Dalem, Cyril Briquet
 * @since 2007-02-01
 */
public class AutomaticJobSubmitter implements LiveUserAgentListener {

	/** Constructs a new automatic Job submitter.
	 */
	public AutomaticJobSubmitter(LiveUserAgent ua,JobConfigurator jc,
								 int jobs_to_submit,
								 boolean block_before_resubmission,
								 SubmissionTimesGenerator stg)
		throws GridException {

		if (ua == null) {
			throw new GridException("AutomaticJobSubmitter(): "+
									"illegal UserAgent");
		}
		if (jc == null) {
			throw new GridException("AutomaticJobSubmitter(): "+
									"illegal Job configurator");
		}
		if (jobs_to_submit < 1) {
			throw new GridException("AutomaticJobSubmitter(): "+
									"illegal Job count");
		}
		if (stg == null) {
			throw new GridException("AutomaticJobSubmitter(): "+
									"illegal submission times generator");
		}

		this.q = new LinkedBlockingQueue<Integer>();
		this.ua = ua;
		this.jc = jc;
		this.jobs_to_submit = jobs_to_submit;
		this.submitted_jobs = new TreeSet<Integer>();
		this.completed_jobs = 0;

		this.block_before_resubmission = block_before_resubmission;
		this.stg = stg;
		
		// Register AJS with UserAgent
		this.ua.registerListener(this);

	}

	// ****************************************************************

	/** LiveUserAgentListener implementation.
	 * {@inheritDoc}
	 */
	public void jobCompletion(JobDescriptor jd,int job_id)
		throws GridException {

		boolean removed_submitted_job = false;
		synchronized (this.submitted_jobs) {
			removed_submitted_job = this.submitted_jobs.remove(job_id);
		}
		
		if (removed_submitted_job == false) {
			USERLOGGER.info("AutomaticJobSubmitter @ "+this.ua.host()+
							": unexpectedly got completed Job "+job_id);
		}
		
		try {
			this.q.put(job_id);
		}
		catch (InterruptedException e) {
			throw new GridException(e);
		}
		++this.completed_jobs;

	}
	
	/** LiveUserAgentListener implementation.
	 * {@inheritDoc}
	 */
	public void taskCompletion(Task t) {}

	// ****************************************************************

	public synchronized void submitJobs() throws GridException {

		synchronized (this.submitted_jobs) {
			if ((this.submitted_jobs.size() > 0) || (this.completed_jobs > 0)) {
				System.out.println("AutomaticJobSubmitter @ "+this.ua.host()+
								   ": already submitted "+
								   (this.submitted_jobs.size()+
									this.completed_jobs)+
								   " Jobs, "+this.jobs_to_submit+" to go");
				return;
			}
		}

		// Submit Job
		System.out.println("AutomaticJobSubmitter @ "+this.ua.host()+
						   ": submitting "+
						   ((this.jobs_to_submit >= 0) ?
							   this.jobs_to_submit : "an infinite number of")+
						   " Jobs of "+this.jc.getTasksPerJob() + " Tasks "+
						   "with input data files each,\n"+stg.summary()+
						   ((this.block_before_resubmission) ?
							   ", waiting results of each Job "+
							   "before resubmitting": "")+",\n"+
						   this.jc.getDataCount()+" distinct input data files "+
						   "will be uploaded to the Peer");

		this.stg.skipWaitingBeforeFirstSubmission();
		while (this.ua.isAlive() && (this.jobs_to_submit != 0)) {
			// 2nd guard but not be "> 0",
			// as -1 is a legal value used to simulate infinite submissions

			// Synthesize Job
			JobDescriptor jd = this.jc.configureJobDescriptor();

			int job_id = submitJob(this.ua,jd);
			System.out.println("UserAgent @ "+this.ua.host()+
							   ": submitted job #"+job_id+" to Peer @ "+
							   this.ua.usedPeer().id());
			
			// Theoretically, we could reach an underflow here.
			// Practically, I'd be surprised to see the program run for 68 years
			--this.jobs_to_submit;

			if (this.block_before_resubmission) {
				this.waitForJobsCompletion();
			}
			if (this.jobs_to_submit > 0) {
				this.stg.waitBeforeNextSubmission();
			}
		}

		if (this.block_before_resubmission == false) {
			System.out.println("AutomaticJobSubmitter @ "+this.ua.host()+
							   ": submitted all Jobs, will "+
							   "exit when all Job results have been retrieved.");
			this.waitForJobsCompletion();
		}

		// Leave some time for the User Agent's Job completion operations
		// (e.g. logging might be buffered)
		ThreadPool.sleep(2);

		System.out.println("AutomaticJobSubmitter @ "+this.ua.host()+
						   ": shutdown");
		
	}
	
	// ****************************************************************

	/** Periodically submits BoTs to a given Peer.
	 */
	public static void main(String[] argv) { /* {{{ */

		try {
			ArgsParser ap =
				new ArgsParser("java lbg.middleware.AutomaticJobSubmitter");
			
			// Help
			ap.addHelper('?', "-help", "Show this help and exit");
			// Connection options
			ap.addStringOption('H', "-peerhost", true, null, "Peer host");
			ap.addIntOption('P', "-peerport", true, null, "Peer port");
			ap.addStringOption('h', "-userhost", true, null, "User host, as seen from the rest of the grid");
			ap.addIntOption('p', "-userport", true, null, "User port");
			ap.addStringOption('i', "-id", true, null, "User ID");
			ap.addStringOption('o', null, false, System.getProperty("user.dir"), "Output directory for task result, defaults to current directory");
			ap.addStringOption('s', null, false, System.getProperty("user.dir"), "Stats directory for job runtime result, defaults to current directory");
			// Synthetic workload options
			ap.addStringOption('j', "-jar", false, "jars/gridhasher.jar", "Path to Grid application jar, defaults to jars/gridhasher.jar");
			ap.addStringOption('m', "-mainclass", false, "lbg.examples.hash.GridHasher", "Main class of Grid application, defaults to lbg.examples.hash.GridHasher");
			ap.addIntOption('n', "-num_jobs", false, -1, "Number of job submissions, defaults to infinite (sends always the same job)");
			ap.addIntOption(' ', "-sleep_lo", false, 60, "Low boundary of the time interval (in seconds) to sleep between submissions, defaults to 60");
			ap.addIntOption(' ', "-sleep_hi", false, 60, "High boundary of the time interval (in seconds) to sleep between submissions, defaults to 60");
			ap.addStringOption(' ', "-trace", false, null, "filename of submission times trace (mutually exclusive with the --sleep_xx options)");
			ap.addStringOption(' ',"-block", false, "true", "Block before resubmission flag, defaults to true");
			ap.addIntOption('t', null, false, 10, "Tasks per job, defaults to 10");
			ap.addIntOption('d', null, false, 3, "Data's per Task, defaults to 3");
			ap.addIntOption(' ', "-ddr", false, 30, "Share data ration (in percents), defaults to 30");
			ap.addStringOption(' ', "-dir", false, System.getProperty("user.dir"), "Where to find random data files, default to working directory");
			ap.addIntOption(' ', "-size", false, 256, "Random data size in MB, default to 256");
			ap.addFlag(' ', "-regen_datas", "Regenerate different data's for each Job submission (only relevant for more than one Job). Else, the same data would be used for each Job.");
			ap.addStringOption(' ', "-ftp_host", false, null, "FTP server hosting data files for user");
			ap.addIntOption(' ', "-ftp_port", false, 2424, "FTP server hosting data files for user");
			
			try {
				ap.parse(argv);
			}
			catch (java.text.ParseException pe) {
				System.err.println(pe.getMessage());
				System.err.println(ap.usage());
				return;
			}
			
			if (ap.getFlag('?')) {
				System.err.println("Lightweight Bartering Grid -- "+
								   "AutomaticJobSubmitter\n\n"+ap.usage());
				return;
			}
			
			// User parameters
			String peer_host  = ap.strValueOf('H');
			int    peer_port  = ap.intValueOf('P');
			int    local_port = ap.intValueOf('p');
	
			// Jobs submission parameters
			int num_jobs = ap.intValueOf('n');
			SubmissionTimesGenerator stg = null;
			if (ap.strValueOf("-trace") != null) {
				stg =
					new SubmissionTimesTrace(ap.strValueOf("-trace"),num_jobs);
			}
			else {
				int inter_submission_lo  = ap.intValueOf("-sleep_lo");
				int inter_submission_hi  = ap.intValueOf("-sleep_hi");
				stg = new SubmissionTimesDistribution(inter_submission_lo,
													  inter_submission_hi);
			}
			
			boolean bbr = Boolean.valueOf(ap.strValueOf("-block"));
			
			/* **************************************************************** */
	
			lbg.LightweightBarteringGrid.setMode(lbg.VirtualizationLevel.LIVE_MODE);
			
			// ATS instantiation
			JobConfigurator jc = new JobConfigurator(
				ap.intValueOf('t'),ap.intValueOf('d'),ap.intValueOf("-ddr"),
				ap.intValueOf("-size"), !ap.getFlag("-regen_datas"),
				ap.strValueOf("-ftp_host"),
				ap.intValueOf("-ftp_port"), ap.strValueOf("-dir"),
				SubmissionUtils.loadJar(ap.strValueOf('j')),
				ap.strValueOf('m'));
	
			// Generate User
			LiveUserAgent user_agent =
				new LiveUserAgent(ap.strValueOf('i'),ap.strValueOf('h'),local_port,
								  HandleDownloader.
									downloadPeerHandle(peer_host,peer_port),
								  ap.strValueOf('o'),ap.strValueOf('s'));
			user_agent.startService(true);
			
			AutomaticJobSubmitter ajs =
				new AutomaticJobSubmitter(user_agent,jc,num_jobs,bbr,stg);
			
			// Welcome message
			System.out.println(lbg.LightweightBarteringGrid.welcome());
			
			ajs.submitJobs();
		}
		catch (Throwable t) {
			t.printStackTrace();
		}
		
	} /* }}} */

	// ****************************************************************

	/** Submits the given job and register it for later waiting.
	 * @param user The user that should submit the job
	 * @param jd The job to submit
	 * @return The job ID
	 * @see #waitForJobsCompletion()
	 */
	private synchronized int submitJob(UserAgent user,JobDescriptor jd)
		throws GridException {
		
		int job_id = user.submitJob(jd);
		synchronized(this.submitted_jobs) {
			this.submitted_jobs.add(job_id);
		}
		return job_id;
		
	}
	
	/** Waits for the completion of all jobs from {@link #submitJob}
	 */
	private synchronized void waitForJobsCompletion() throws GridException {
		
		for ( ; ; ) {
			synchronized(this.submitted_jobs) {
				if (this.submitted_jobs.size() == 0) break;
			}
			try {
				Integer job_id_as_key = this.q.take();
			}
			catch (InterruptedException e) {
				throw new GridException(e);
			}
		}
		
	}
	
	// ****************************************************************

	private final LinkedBlockingQueue<Integer> q;
	private final LiveUserAgent ua;
	private final JobConfigurator jc;
	private final boolean block_before_resubmission;
	private final SubmissionTimesGenerator stg;
	private       int jobs_to_submit;
	private final Set<Integer> submitted_jobs;
	private       int completed_jobs;

	private static final Logger USERLOGGER = Logger.getLogger("lbg.user");

}

