/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LaunchPeer.java

package lbg.middleware;

import lbg.*;
import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.common.net.*;
import lbg.peer.*;
import lbg.peer.core.*;
import lbg.peer.service.*;
import lbg.protocol.rpc.*;
import lbg.peer.config.*;

import java.util.Properties;

/** Peer Launcher
 * @author Xavier Dalem, Cyril Briquet
 */
public class LaunchPeer {

	public static void main(String[] argv) throws GridException {

		final ArgsParser ap = new ArgsParser("lbg.LaunchPeer");
		
		// Help
		ap.addHelper('?', "-help", "Show this help and exit");
		// Options
		ap.addStringOption('c', "-config", true, null, "Peer configuration");

		//ap.addStringOption('i', "-id", false, null, "Peer id");
		ap.addStringOption('i', "-id", false,
						   CommonNetHandling.getLocalHostName(),"Peer id");
		//ap.addStringOption('h', "-localhost", false, null, "Peer hostname");
		ap.addStringOption('h', "-localhost",false,
						   CommonNetHandling.getLocalHostName(),
						   "Peer hostname");
		ap.addStringOption('s',"-searchhost",false,
						   null, // null, not getLocalHostName()
						   "search engine hostname");
		// Flags
		ap.addFlag('q', "-quiet", "Do not print statistics while running");
		
		try {
			ap.parse(argv);
		}
		catch (java.text.ParseException pe) {
			System.err.println(pe.getMessage());
			System.err.println(ap.usage());
			return;
		}
		
		if (ap.getFlag('?')) {
			System.err.println("Lightweight Bartering Grid -- "
				+ "Peer launcher\n\n" + ap.usage());
			return;
		}
		
		final Properties conf =
			CommonFileHandling.loadProperties(ap.strValueOf('c'));
		String id_option = ap.strValueOf('i');
		if (id_option != null) {
			conf.setProperty("PEER_ID",id_option);
		}
		String hostname_option = ap.strValueOf('h');
		if (hostname_option != null) {
			conf.setProperty("PEER_HOSTNAME",hostname_option);
		}
		String search_host_option = ap.strValueOf('s');
		if (search_host_option != null) {
			conf.setProperty("SEARCH_HOSTNAME",search_host_option);
		}
		
		/* **************************************************************** */

		LightweightBarteringGrid.setMode(VirtualizationLevel.LIVE_MODE);
		
		PeerNetworkConfiguration pnc = new PeerNetworkConfiguration(conf);
		LivePeer p = new LivePeer(new PeerBasicConfiguration(conf),
								  pnc,
								  new PeerDataConfiguration(conf),
								  new PeerStrategyConfiguration(conf),
								  new PeerServiceFactory(),
								  new ControllerFactory(),
								  HandleDownloader.
								  	downloadSearchEngineHandle(pnc.SEARCH_HOSTNAME,
															   pnc.SEARCH_PORT));
		p.startService();
		
		if (!(ap.getFlag('q'))) {
			System.out.println( "\n"+LightweightBarteringGrid.welcome()+
							   p.toStringSummary());
		}

		boolean stats = !(ap.getFlag('q'));
		while (p.isAlive()) {
			// Print stats
			if (stats) {
				int busy = p.rms().countBusyResources();
				System.out.print("Queued (loc/sup): "
					+ p.qms().countWaitingLocalTasks() + "/"
					+ p.qms().countWaitingSupplyingTasks()
					+ " | Res: " + p.rms().countTotalResources()
					+ " (" + busy + " busy)"
					+ "\r");
				System.out.flush();
			}
			ThreadPool.sleep(1);
		}
		
		p.checkLiveness();
		
		System.out.println("Peer daemon terminated");
		
	}
	

}

