/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LaunchResource.java

package lbg.middleware;

import lbg.*;
import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.common.net.*;
import lbg.protocol.rpc.*;
import lbg.resource.*;
import lbg.resource.config.*;

import java.util.Date;
import java.util.Properties;

/** Resource Launcher
 * @author Xavier Dalem, Cyril Briquet
 */
public class LaunchResource {

	public static void main(String[] argv) throws GridException {

		final ArgsParser ap = new ArgsParser("lbg.LaunchResource");
		
		// Help
		ap.addHelper('?', "-help", "Show this help and exit");
		// Options
		ap.addStringOption('c', "-config", true, null, "Resource configuration");
		//ap.addStringOption('i', "-id", false, null, "Resource id");
		ap.addStringOption('i', "-id",false,
						   CommonNetHandling.getLocalHostName(),
						   "Resource id");
		//ap.addStringOption('h', "-localhost", false, null, "Resource hostname");
		ap.addStringOption('h', "-localhost",false,
							 CommonNetHandling.getLocalHostName(),
							 "Resource hostname");
		ap.addStringOption('H', "-peerhost", false, null, "Peer hostname");
		// Flags
		ap.addFlag('q', "-quiet", "Do not print statistics while running");
		
		try {
			ap.parse(argv);
		}
		catch (java.text.ParseException pe) {
			System.err.println(pe.getMessage());
			System.err.println(ap.usage());
			return;
		}
		
		if (ap.getFlag('?')) {
			System.err.println("Lightweight Bartering Grid -- "
				+ "Resource launcher\n\n" + ap.usage());
			return;
		}
		
		final Properties conf =
			CommonFileHandling.loadProperties(ap.strValueOf('c'));

		String id_option = ap.strValueOf('i');
		if (id_option != null) {
			conf.setProperty("RESOURCE_ID",id_option);
		}
		String hostname_option = ap.strValueOf('h');
		if (hostname_option != null) {
			conf.setProperty("RESOURCE_HOSTNAME",hostname_option);
		}
		String peer_host_option = ap.strValueOf('H');
		if (peer_host_option != null) {
			conf.setProperty("PEER_HOST",peer_host_option);
		}

		/* **************************************************************** */
		
		LightweightBarteringGrid.setMode(VirtualizationLevel.LIVE_MODE);
		
		ResourceNetworkConfiguration netconf =
			new ResourceNetworkConfiguration(conf);
		
		LiveResource res =
			new LiveResource(new ResourceBasicConfiguration(conf),
							 netconf,
							 new ResourceDataConfiguration(conf),
							 HandleDownloader.
							 	downloadPeerHandle(netconf.HOST(),
												   netconf.peerPort()));
		res.startService();

		if (!(ap.getFlag('q'))) {
			System.out.println("\n"+LightweightBarteringGrid.welcome()+
							   res.toString());
		}
		
		boolean stats = !(ap.getFlag('q'));
		while (res.isAlive()) {
			// Print stats
			if (stats) {
				StringBuilder sb = new StringBuilder();
				if (res.service().idle())
					sb.append("Idle");
				else {
					sb.append("Busy on task for ");
					sb.append(res.runningTask().consumerID());
					sb.append(", started ");
					sb.append(new Date(res.taskStartTime()*1000).toString());
				}
				while (sb.length() < 80) {
					sb.append(" "); // Let's fill line remainder with blanks
				}
				sb.append("\r");
				System.out.print(sb.toString());
			}
			
			// With such a sleeping time, instant Tasks will not be shown
			// but these should be quite rare in a grid.
			// Moreover, not metioningsuch a tiny power consumption seems OK.
			ThreadPool.sleep(1);
		}
		
		System.out.println("Resource daemon terminated");
	}

}

