/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LaunchUserAgent.java

package lbg.middleware;

import lbg.*;
import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.common.net.*;
import lbg.middleware.util.*;
import lbg.peer.workflow.*;
import lbg.protocol.rpc.*;
import lbg.user.*;
import java.io.File;
import java.util.Properties;

/** User Launcher (to submit jobs)
 * @author Xavier Dalem, Cyril Briquet
 */
public class LaunchUserAgent {
	
	public static void main(String[] argv) throws GridException {

		final ArgsParser ap = new ArgsParser("java lbg.LaunchUserAgent");
		
		// Help
		ap.addHelper('?', "-help", "Show this help and exit");
		// Options
		ap.addStringOption('H', "-peerhost", true, null, "Peer host");
		ap.addIntOption('P', "-peerport", true, null, "Peer port");
		//ap.addStringOption('h', "-userhost", true, null, "User host, as seen from the rest of the grid"); -- CB
		ap.addStringOption('h', "-userhost", false,
						   CommonNetHandling.getLocalHostName(),
						   "User hostname");
		ap.addIntOption('p', "-userport", true, null, "User port");
		ap.addStringOption('i', "-id", true, null, "User ID");
		ap.addStringOption('o', null, false, System.getProperty("user.dir"), "Output directory for task result, defaults to current directory");
		ap.addStringOption('s', null, false, System.getProperty("user.dir"), "Stats directory for job runtime, defaults to current directory");
		ap.addListOption('f', "-file", true, "Job to submit (*.jdf files)");
		
		try {
			ap.parse(argv);
		} catch (java.text.ParseException pe) {
			System.err.println(pe.getMessage());
			System.err.println(ap.usage());
			return;
		}
		
		if (ap.getFlag('?')) {
			System.err.println("Lightweight Bartering Grid -- "
				+ "User launcher\n\n" + ap.usage());
			return;
		}
		
		/* **************************************************************** */

		LightweightBarteringGrid.setMode(lbg.VirtualizationLevel.LIVE_MODE);
		
		String peer_host = ap.strValueOf('H');
		int peer_port = ap.intValueOf('P');
		int local_port = ap.intValueOf('p');
		
		LiveUserAgent user_agent = new LiveUserAgent(ap.strValueOf('i'),
							  ap.strValueOf('h'),
							  local_port,
							  HandleDownloader.
							 	downloadPeerHandle(peer_host,peer_port),
							  ap.strValueOf('o'),
							  ap.strValueOf('s'));
		user_agent.startService();
		
		System.out.println("\n"+LightweightBarteringGrid.welcome()+
						   user_agent.toString());

		try {
			for (String job : ap.listValueOf('f')) {
				Properties props = CommonFileHandling.loadProperties(job);
				String basedir =
					new File(job).getAbsoluteFile().getParent()+
					CommonFileHandling.FILE_SEP;

				PropertiesJobLoader pjl = new PropertiesJobLoader();

				JobDescriptor jd =
					pjl.loadJobDescription(user_agent,props,basedir);
				int job_id = user_agent.submitJob(jd);
				System.out.println("Submitted job #"+job_id);
			}
		}
		catch (GridException e) {
			e.printStackTrace();
		}
		
		System.out.println("Waiting for results from Peer " + peer_host + ":"+
						   peer_port);
		while (user_agent.isAlive()) {
			// Nothing to report, we're waiting
			ThreadPool.sleep(1);
		}
		
		System.out.println("User Agent daemon terminated");
		
	}

}

