/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// GridClassLoader.java

package lbg.middleware.execution;

import lbg.common.exception.*;
import lbg.peer.workflow.*;
import lbg.protocol.api.GridApplication;
import lbg.task.*;

import java.lang.reflect.Constructor;
import java.util.TreeSet;

/** Job/Task classes loader.
 * This class can be used on the Resource to load everything need for an
 * execution, but also on the User to make sure class loading will be OK on
 * the Resource.
 * @author Xavier Dalem
 * @since 2007-09-09
 */
public class GridClassLoader extends ClassLoader {

	/** Safety guard for infinite recursion.
	 * While circular inheritance is illegal and should be forbidden by the
	 * compilator, I suppose code could be crafted to show such a property and
	 * make our loader crash if it didn't check that heritage is linear.
	 */
	/* Example :
	 * Trying to load A extends B.
	 * B is not loaded, try to load it before defining A.
	 * Trying to load B extends A
	 * A is not loaded (we needed B), try to load it before defining B
	 * Loop, and let the universe collapse in an infinite recursion.
	 */
	private final TreeSet<String> loadingStack;

	public GridClassLoader() {
		loadingStack = new TreeSet<String>();
	}

	/** Tries to load all Tasks from the Job description
	 * @param jd The Job description
	 * @see #load(Task)
	 */
	public synchronized void load(JobDescriptor jd) throws GridException {
		int i = 0;

		for (Task t : jd.toTasks("dummy", 0, 0)) {
			try {
				load(t);
			} catch (GridException e) {
				throw new GridException("load(): Failed to load Task " + i, e);
			}

			i++;
		}
	}

	/** Loads all classes required to run the Task, and creates the Application.
	 * @todo check dependency, and don't forget circular ones with a stack
	 */
	public synchronized GridApplication load(Task t) throws GridException {
		// Get main class name
		String main_s = t.mainClass();

		// Load all classes
		String[] names   = t.names();
		for (int i = 0; i < names.length; i++) {
			try {
				load(t, i);
			} catch (Error e) {
				throw new GridException("Failed to load class " + names[i], e);
			}
		}

		GridApplication gapp;

		// Loads and sets the GridApplication,
		// and checks that it implements the GridApplication interface
		try {
			Class<?> main_c = Class.forName(main_s, true, this);
			Constructor cons = main_c.getConstructor(new Class[0]);
			if (cons == (Constructor)null) {
				throw new GridException("load(): "+
										"target Task is not "+
										"a legal GridApplication "+
										"(argument-less constructor not found)");
			}

			Object o = cons.newInstance(new Object[0]);
			if (! (o instanceof GridApplication)) {
				throw new GridException("load(): "+
										"target Task is not a GridApplication");
			}
			gapp = (GridApplication)o;
		}
		catch (ClassNotFoundException e1) {
			throw new GridException("load(): "+
									"could not load main class " + main_s,e1);
		}
		catch (Exception e2) {
			throw new GridException("load(): "+
									"could not load GridApplication",e2);
		}

		return gapp;
	}

	/** Loads a given class from a Task, trying to resolve the hierarchy.
	 * @param t The Task containing the classes
	 * @param class_index The number of the class to load in the Task
	 * @todo Maybe sorting the names array in JobDescriptor.toTasks would be
	 * a good idea, but this only becomes interesting if we make more than
	 * log(N) searches, N being the total number of classes
	 */
	private synchronized void load(Task t, int class_index) throws
			NoClassDefFoundError, ClassCircularityError {
		byte[][] classes = t.jars();
		String[] names = t.names();

		// Inheritance circularity check
		if (loadingStack.contains(names[class_index]))
			throw new ClassCircularityError(names[class_index]
				+ " (in)directly depends on itself");

		// Avoid linkage errors when trying to load twice the same class
		try {
			super.loadClass(names[class_index]);
			return; // Class was already loaded
		} catch (ClassNotFoundException e) {
			// Class is absent and should be loaded
		}

		while (true) {
			try {
				super.defineClass(names[class_index], classes[class_index], 0,
					classes[class_index].length);
				break;
			} catch (NoClassDefFoundError err) {
				String missing = err.getMessage().replace('/', '.');
				//debugging code -- System.err.println("Searching for missing class " + missing);

				int i;
				for (i = names.length-1; i >= 0; i--)
					if (names[i].equals(missing))
						break;

				if (i < 0) // Didn't find the missing one
					throw err;
				else {
					loadingStack.add(names[class_index]);
					load(t, i);
					loadingStack.remove(names[class_index]);
				}
			}
		}
	}

}

