/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// JobConfigurator.java

package lbg.middleware.util;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.common.net.*;
import static lbg.common.io.CommonFileHandling.FILE_SEP;
import lbg.peer.service.*;
import lbg.peer.workflow.*;
import lbg.protocol.data.*;
import lbg.protocol.rpc.*;
import lbg.task.Task;
import lbg.user.*;

import java.io.File;
import java.util.Set;
import java.util.TreeSet;

/** Configurator of Jobs composed of multiple Tasks.
 * <p>One input data file can be assigned to each Task.
  * <p>The level of redundancy between files of one Job can be defined.
 * @author Xavier Dalem, Cyril Briquet
 * @since 2007-02-01
 */
public class JobConfigurator {

	/** Constructs an automatic submitter.
	 * @param tpj Tasks per job
	 * @param dpt Data per taks
	 * @param ddr Share data ratio
	 * @param data_size Data size (for ddr mode)
	 * @param keep_datas <tt>true</tt> if the same data should be used in
	 * all calls to {@link #configureJobDescriptor()}
	 * @param ftp_host Host of the FTP server giving data for user
	 * @param ftp_port Port of the FTP server giving data for user
	 * @param data_source_directory data source directory
	 * @param mj Grid application jar
	 * @param main_class Grid application main class
	 */
	public JobConfigurator(int tpj,int dpt,int ddr,int data_size,
						   boolean keep_datas, String ftp_host, int ftp_port,
						   String data_source_directory,
						   SubmissionUtils.MemoryJar mj,String main_class)
		throws GridException {
		
		if (tpj <= 0) {
			throw new GridException("tpj must be strictly positive");
		}
		this.tpj = tpj;
		
		if (dpt < 0) {
			throw new GridException("dpt must be positive");
		}
		this.dpt = dpt;
		
		if ((ddr < 0) || (ddr > 100)) {
			throw new GridException("DDR must be an integer percentage");
		}
		this.ddr = ddr;

		if (data_size < 1) {
			throw new GridException("illegal data size");
		}
		this.data_size = data_size;
		
		// keep_datas may be null
		this.keep_datas = keep_datas;
		
		if ((dpt > 0) && (ftp_host == null)) {
			throw new GridException("illegal FTP host");
		}
		this.ftp_host = ftp_host;
		
		if ((dpt > 0) && (CommonNetHandling.isValidPort(ftp_port) == false)) {
			throw new GridException("illegal FTP port");
		}
		this.ftp_port = ftp_port;
		
		if (data_source_directory == null) {
			throw new GridException("illegal data source directory");
		}
		this.data_source_directory = data_source_directory;

		if (mj == null) {
			throw new GridException("illegal Grid application jar");
		}
		this.mj = mj;
		
		if (main_class == null) {
			throw new GridException("illegal Grid application main class");
		}
		this.main_class = main_class;
		
		// The following line will be updated in the future,
		// when AutomaticJobSubmitter supports the description of
		// parameters to pass to the submitted Tasks
		this.parameters = (this.main_class.endsWith("GridHasher")) ?
			new Object[] { "SHA-1" } : new Object[] { };
			
		this.data_count = computeDataCount();
		this.last_datas = null;

	}

	// ****************************************************************

	public synchronized int getTasksPerJob() {
	
		return this.tpj;
		
	}

	public synchronized int getDataCount() {
	
		return this.data_count;
		
	}
	
	// ****************************************************************

	public synchronized JobDescriptor configureJobDescriptor()
		throws GridException {
		
		JobDescriptor jd = new JobDescriptor();
		jd.add(this.mj.classnames,this.mj.bytecode);

		addDataDescription(jd,this.data_count);
		
		addTasks(jd,this.last_datas);
		
		return jd;
	}

	// ****************************************************************

	private synchronized int computeDataCount() throws GridException {

		// We should create tpj*dpt*ddr unique data's, then try to add
		// them to Tasks
		int data_count = (this.tpj*this.dpt*this.ddr)/100;
		// However, as we need unique data's in one task, we should have at
		// least dpt datas
		if (data_count < this.dpt) {
			data_count = this.dpt;
			System.err.println("DDR value is too low (" +
							   CommonMath.div_label_2dd(this.ddr,100)+
							   "), using "+((float)1/this.tpj)+" instead");
		}
		data_count = (data_count > this.dpt) ? data_count : this.dpt;

		return data_count;
	
	}
	
	/** Fills data using a 'Diversity Data Ratio' schema.
	 * Tasks are filled with random data (all of the same size) according to
	 * the following formula:
	 * <tt>total_unique_data_size = ddr * total_data_size</tt>
	 * @param jd The JobDescriptor to fill
	 */
	private synchronized void
		addDataDescription(JobDescriptor jd,int data_count)
		throws GridException {
		
		// Check if we can use cached results:
		// generate data's either on first call
		// or if we want to regenerate them for each submitted BoT
		if (this.last_datas != null && this.keep_datas) {
			jd.add(this.last_dataIDs, this.last_datas);
			return;
		}
		
		// size is fixed, nb of data per Task is fixed
		
		GridData[][] datas =
			new GridData[data_count][GridDataManager.NUM_SHARE_OPTIONS];
			
		String[] dataIDs = new String[datas.length];

		// Pre-hash file. It will always be the same so computing it
		// datas.length
		byte[] hash = null;
		
		if (data_count > 0) {
			hash = GridDataManager.hash(
				new File(this.data_source_directory+FILE_SEP+
						 "random_data_"+this.data_size+"m"));
		}
		
		// Generate random data
		for (int k = 0; k < datas.length; k++) {
			// Random ids...
			dataIDs[k] = "AutomaticData_"+CommonMath.sampleUniformDist();
			
			for (int j = 0; j < datas[k].length; j++) {
				datas[k][j] = null;
			}
			
			// ... but all data have the same physical source
			datas[k][GridDataManager.SHARED_FTP] =
				new FTPGridData(dataIDs[k],((long)this.data_size)*1024L*1024L,
								this.ftp_host, this.ftp_port,
								"random_data_"+this.data_size+"m",
								"anonymous","",hash);
		}
		
		// Add data ids to job descriptor
		jd.add(dataIDs, datas);
		
		// Cache results
		this.last_dataIDs = dataIDs;
		this.last_datas = datas;
		
		return;
		
	}

	private synchronized void addTasks(JobDescriptor jd,GridData[][] datas)
		throws GridException {
		
		// Add tasks to JobDescriptor
		int current_data = 0;
		
		// TaskDescriptors creation.
		/* Consider all data's off all tasks as one array A (practically, it's
		 * a bidimensional one, hence the two loops). We make copies of the
		 * already generated array of data's B in A so that A={B, B, part of B}
		 * That way, data's are distributed so that tasks never have the same
		 * data twice, and we try to avoid bunches of data-equal tasks (not
		 * possible if B.length == dpt)
		 */
		for (int cur_task = 0; cur_task < this.tpj; cur_task++) {
			String[] related_data = new String[this.dpt];
			for (int k = 0; k < related_data.length; k++) {
				related_data[k] = 
					datas[current_data][GridDataManager.SHARED_FTP].id;
				current_data = (current_data + 1) % datas.length;
			}
			
			TaskDescriptor td =
				new TaskDescriptor(this.mj.classnames,this.main_class,
								   this.parameters,related_data);
			jd.add(td);
		}
		
	}

	// ****************************************************************

	private final int tpj;
	private final int dpt;
	private final int ddr;
	private final int data_size;
	private final boolean keep_datas;
	private final int ftp_port;
	private       int data_count;
	private final SubmissionUtils.MemoryJar mj;
	private final String main_class;
	private final Object[] parameters;
	private final String ftp_host;
	private final String data_source_directory;

	private GridData[][] last_datas;
	private String[] last_dataIDs;
}

