/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PropertiesJobLoader.java

package lbg.middleware.util;

import lbg.common.exception.*;
import lbg.common.io.CommonFileHandling;
import static lbg.common.io.CommonFileHandling.FILE_SEP;
import static lbg.middleware.util.SubmissionUtils.MemoryJar;
import lbg.peer.workflow.JobDescriptor;
import lbg.peer.workflow.TaskDescriptor;
import lbg.protocol.data.*;
import lbg.user.UserAgent;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.util.Map;
import java.util.Properties;
import java.util.TreeMap;

/** Utility class to load a JobDescriptor from a file in Properties format.
 * @since 2006-12-13
 * @author Xavier Dalem, Cyril Briquet
 * @see java.util.Properties
 */
public class PropertiesJobLoader {
	
	public PropertiesJobLoader() {
	
		this.jar_cache = new TreeMap<String,MemoryJar>();
		
	}
	
	/** Loads a JobDescriptor form a set of properties coming fro a JDF file.
	 * Pathanmes are resolved from <tt>user.dir</tt>
	 * 
	 * @deprecated Relying on the current working directory is dangerous and
	 * can yield confusing results for the user. The best is usually to specify
	 * the directory containing the JDF file using
	 * {@link #loadJobDescription(UserAgent, Properties, String)}
	 */
	@Deprecated
	public synchronized JobDescriptor
		loadJobDescription(UserAgent user,Properties props)
		throws GridException{
		return loadJobDescription(user, props, System.getProperty("user.dir"));
	}

	/** Loads a JobDescriptor form a set of properties coming from a JDF file.
	 * 
	 * @param user The user that will manage the JobDescriptor
	 * @param props The properties defining the JobDescriptor
	 * @param basedir The base directory from where the pathnames (e.g. to data,
	 * archive files) will be resolved.
	 * @return A fully formed JobDescriptor
	 * @see JobDescriptor
	 * @throws GridException
	 */
	public synchronized JobDescriptor
		loadJobDescription(UserAgent user,Properties props, String basedir)
		throws GridException {
		
		if (basedir == null)
			throw new NullPointerException("Null base directory");
		
		if (! basedir.endsWith(FILE_SEP))
            basedir += FILE_SEP;

		String cur_prop = props.getProperty("task_list");
		if (cur_prop == null || cur_prop.equals(""))
			throw new GridException("PropertiesJobLoader: task_list must"
				+ " contain at least a task");
		
		JobDescriptor jd = new JobDescriptor();
		
		// For each task (strip heading and trailing spaces in their names)
		for (String task : cur_prop.split("\\s?,\\s?")) {
			// Load bytecode from given jar
			cur_prop = props.getProperty(task + ".jar");
			if (cur_prop == null || cur_prop.equals(""))
				throw new GridException("PropertiesJobLoader: missing jar file"
					+ " for task '" + task + "'");
			if (! CommonFileHandling.isAbsolutePath(cur_prop))
				cur_prop = basedir + cur_prop;
			
			
			MemoryJar mj = this.jar_cache.get(cur_prop);
			if (mj == null) {
				mj = SubmissionUtils.loadJar(cur_prop);
				this.jar_cache.put(cur_prop,mj);
			}

			// Load all data files {{{
			GridData[][] datas = null;
			String[] ids = null;
			
			cur_prop = props.getProperty(task + ".datas");
			if (cur_prop == null || cur_prop.equals("")) {
				datas = new GridData[0][0];
				ids = new String[0];
			}
			
			else {
			
				String[] data_paths = cur_prop.split("\\s?,\\s?");
				datas = new GridData
					[data_paths.length][GridDataManager.NUM_SHARE_OPTIONS];
				ids = new String[data_paths.length];
				
				for (int i = 0; i < data_paths.length; i++) {
					if (! CommonFileHandling.isAbsolutePath(data_paths[i]))
						data_paths[i] = basedir + data_paths[i];
					File in = new File(data_paths[i]);
					byte[] data = new byte[(int) in.length()];
					try {
						DataInputStream dis =
							new DataInputStream(new FileInputStream(in));
						dis.readFully(data);
						dis.close();
					} catch (java.io.IOException ioe) {
						throw new GridException("Failed to read data file "
							+ data_paths[i], ioe);
					}
					
					ids[i] = user.usedPeer().id() + "."	+ user.id()	+ "."
						+ in.getName();
					datas[i][GridDataManager.SHARED_INLINE] =
						new InlineGridData(ids[i], data);
				}
			}
			// }}}
			
			// Get parameters {{{
			Object[] params = null;
			cur_prop = props.getProperty(task + ".parameters");
			if (cur_prop == null || cur_prop.equals(""))
				params = new Object[0];
			else
				params = cur_prop.split("\\s?,\\s?");
			// }}}
			
			// Get main class, last property for task description creation
			cur_prop = props.getProperty(task + ".main_class");
			if (cur_prop == null || cur_prop.equals(""))
				throw new GridException("PropertiesJobLoader: missing main"
				+ " class for task '" + task + "'");
			
			jd.add(new TaskDescriptor(mj.classnames, cur_prop, params, ids),
				   mj.classnames, mj.bytecode, ids, datas);
		}
		
		return jd;
	}

	// *very* useful for parameter sweeps...
	private final Map<String,MemoryJar> jar_cache;
	
}

