/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SubmissionTimesTrace.java

package lbg.middleware.util;

import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.DUMMY_STRING;

import java.io.BufferedReader;
import java.util.Arrays;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.Date;
import java.util.StringTokenizer;

/**
 * @author Cyril Briquet
 */
public class SubmissionTimesTrace implements SubmissionTimesGenerator {

	/** Constructs a new submission times trace-based generator.
	 */
	public SubmissionTimesTrace(String trace_fn,int expected_jobs)
		throws GridException {

		if (trace_fn == null) {
			throw new GridException("SubmissionTimesTrace(): "+
									"illegal trace filename");
		}
		if (expected_jobs < 0) {
			throw new GridException("SubmissionTimesTrace(): "+
									"illegal expected Job count");
		}

		this.trace_fn = trace_fn;
		this.expected_jobs = expected_jobs;
		this.submission_times = new int[this.expected_jobs];
		this.wait_defined_period_calls = new AtomicInteger(0);
		
		loadTrace();

	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[SubmissionTimesTrace]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	public void waitBeforeNextSubmission() throws GridException {

		int inter_submission_t = -1;
		
		synchronized (this) {
			int w = this.wait_defined_period_calls.get();
			if (w == this.expected_jobs) {
				throw new GridException("waitBeforeNextSubmission(): "+
										"too many calls");
			}
			
			if (w == 0) {
				inter_submission_t = this.submission_times[0];
			}
			else{
				inter_submission_t =
					this.submission_times[w]-this.submission_times[w-1];
			}
			
			this.wait_defined_period_calls.incrementAndGet();
		}
		
		ThreadPool.sleep(inter_submission_t);
		
	}
	
	public void skipWaitingBeforeFirstSubmission() throws GridException {
		
		if (this.wait_defined_period_calls.get() != 0) {
			throw new GridException("skipWaitingBeforeFirstSubmission(): "+
									"already waited for a few submissions");
		}
		this.wait_defined_period_calls.incrementAndGet();
		
	}


	// ****************************************************************

	public String summary() {

		StringBuilder sb = new StringBuilder();
		
		sb.append("from trace file ").append(this.trace_fn);

		return sb.toString();
		
	}

	// ****************************************************************

	// For testing purposes
	public static void main(String[] args) throws GridException {
		
		if (args.length != 2) {
			throw new GridException("Illegal command line parameters");
		}
		
		String tfn = args[0];
		int jc = Integer.parseInt(args[1]);
		SubmissionTimesGenerator stg = new SubmissionTimesTrace(tfn,jc);
		
		System.out.println("[INIT] "+new Date());
		for (int i = 0 ; i < jc ; ++i) {
			stg.waitBeforeNextSubmission();
			System.out.println("["+i+"]    "+new Date());
		}
		
	}
	
	// ****************************************************************

	private void loadTrace() throws GridException {
		
		BufferedReader br = CommonFileHandling.openBR(this.trace_fn);
		
		String line = DUMMY_STRING;
		int l = 0;
		while ((line = CommonFileHandling.readLine(br,this.trace_fn)) != null) {
			line = line.trim();
			if (line.startsWith("#") || line.length() == 0) continue;

			if (l >= expected_jobs) {
				throw new GridException("load(): too many lines");
			}
			
			StringTokenizer st = new StringTokenizer(line);
			if (st.countTokens() != 2) {
				throw new GridException("loadTrace(): illegal line "+
										" ("+line+")");
			}
			int submission_time = -1;
			try {
				submission_time = Integer.parseInt(st.nextToken());
			}
			catch (NumberFormatException e) {
				throw new GridException("loadTrace(): illegal line  ("+line+")",
										e);
			}

			if (submission_time < 0) {
				throw new GridException("loadTrace(): illegal submission time "+
										submission_time);
			}

			this.submission_times[l] = submission_time;
			
			++l;
		}
		
		CommonFileHandling.closeBR(br,this.trace_fn);
		
		if (this.expected_jobs != l) {
			throw new GridException("loadTrace(): Job count mismatch (got "+l+
									", expected "+this.expected_jobs+")");
		}
		
		if (this.expected_jobs > 0) {
			Arrays.sort(this.submission_times);
		}
		
		System.out.println("Loaded "+l+
						   " Job submission/completion times pairs");
		
	}
	
	// ****************************************************************

	private final String trace_fn;
	private final int    expected_jobs;
	private final int[]  submission_times;
	private final AtomicInteger wait_defined_period_calls;

}

