/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SubmissionUtils.java

package lbg.middleware.util;

import lbg.common.exception.*;
import static lbg.common.io.CommonFileHandling.CLASS_EXT;
import java.io.DataInputStream;
import java.util.Enumeration;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.jar.JarEntry;

/** A set of utils to facilitate BoT submission.
 * Capabilities are:<ul>
 * <li>bytecode loading from a jar</li>
 * </ul>
 * @author Xavier Dalem
 * @since 2007-02-01
 */
public class SubmissionUtils {
	
	/** Initial size of Vectors */
	private static final int V_INIT = 1;
	/** Increment size of Vectors */
	private static final int V_INCR = 2;
	
	/** Loads the bytecode of a jar in memory.
	 * @param path Path to the jar file
	 * @return The content of the jar file
	 */
	public static MemoryJar loadJar(String path) throws GridException {
		JarFile jar = null;
		Vector<byte[]> bytecodes_v = null;
		Vector<String> names_v = null;
		
		try {
			jar = new JarFile(path);
			bytecodes_v = new Vector<byte[]>(V_INIT,V_INCR);
			names_v = new Vector<String>(V_INIT,V_INCR);

			for (Enumeration e = jar.entries(); e.hasMoreElements();) {
				JarEntry je = (JarEntry)e.nextElement();
				// skip everything but classes
				if (je.getName().endsWith(CLASS_EXT)) {
					int i = je.getName().lastIndexOf(CLASS_EXT);
					DataInputStream dis = new DataInputStream(jar.getInputStream(je));
					byte[] bytes = new byte[(int) je.getSize()];

					// Tricky : InputStream does not fill the byte array. Long
					// class files ended up being truncated (i.e. padded with
					// zeroes)
					dis.readFully(bytes);
					dis.close();
					bytecodes_v.add(bytes);

					// Define the name based on location, without the trailing .class
					String name = je.getName().replace('/','.').substring(0,i);
					names_v.add(name);
				}
			}
		} catch (java.io.IOException ioe) {
			throw new GridException("Failed to open jar " + path, ioe);
		}

		MemoryJar mj = new MemoryJar();
		mj.classnames = new String[names_v.size()];
		mj.classnames = names_v.toArray(mj.classnames);
		names_v = null;
		mj.bytecode = new byte[bytecodes_v.size()][];
		mj.bytecode = bytecodes_v.toArray(mj.bytecode);
		bytecodes_v = null;
		
		return mj;
	}
	
	// Nested classes {{{
	/** Holds the content of a given jar in memory.
	 * Only classes are kept: resources are discarded.
	 */
	public static class MemoryJar {
		public String[] classnames = null;
		public byte[][] bytecode = null;
		
		MemoryJar() {}
	}
	// }}}
}

