/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LivePeer.java

package lbg.peer;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.peer.config.*;
import lbg.peer.core.*;
import lbg.peer.core.negotiator.*;
import lbg.peer.core.qms.*;
import lbg.peer.core.qms.prequeue.*;
import lbg.peer.core.scheduler.*;
import lbg.peer.service.*;
import lbg.peer.web.*;
import lbg.protocol.data.*;
import lbg.protocol.rpc.*;
import lbg.searchengine.*;
import lbg.searchengine.service.*;
import java.util.concurrent.*;
import java.util.logging.Logger;
import java.util.logging.Level;

/** Peer (LIVE mode).
 * @author Cyril Briquet
 */
public class LivePeer extends Peer {
	
	/** Creates a new Peer (LIVE mode).
	 * @param peer_basic_config Peer basic configuration
	 * @param peer_network_config Peer network configuration
	 * @param peer_data_config Peer data configuration
	 * @param peer_strategy_config Peer strategy configuration
	 * @param se_handle search engine handle
	 * @param peer_service_factory Peer service factory
	 * @param ctrlr_factory controller factory
	 */
	public LivePeer(PeerBasicConfiguration peer_basic_config,
					PeerNetworkConfiguration peer_network_config,
					PeerDataConfiguration peer_data_config,
					PeerStrategyConfiguration peer_strategy_config,
					PeerServiceFactory peer_service_factory,
					ControllerFactory ctrlr_factory,
					SearchEngineHandle se_handle)
		throws GridException {
		
		super(peer_basic_config,peer_network_config,peer_data_config,
			  peer_strategy_config,peer_service_factory,ctrlr_factory,
			  se_handle);
		
		// ****************

		this.logger = Logger.getLogger("lbg.peer");

		this.gdm = new GridDataManager(peer_network_config.hostname(),
           peer_data_config.CACHE_DIR(), peer_network_config.btDataPort(),
           peer_network_config.ftpDataPort(),
           peer_network_config.hostname(),
           peer_network_config.btTrackerPort(),
           peer_data_config.SUPPLYING_CACHE_CAPACITY(),
           peer_data_config.DATA_HOST_POLICY(),
           null)
           .setFTPConfig(peer_network_config.ftpRequestPort());
		
		this.websvr = new PeerWebServer(this);
		
		// ****************

		this.res_svc_thread  = null;
		this.ua_svc_thread   = null;
		this.peer_svc_thread = null;
		this.sched_ex_svc    = null;
		this.online = false;

	}

	// ****************************************************************
	
	public String toString() {
		
		return CommonStringHandling.
			replaceTag(super.toString(),"[Peer","[LivePeer");

	}

	// ****************************************************************

	public synchronized void startService() throws GridException {

		if (this.online == true) {
			return;
		}
		else {
			this.online = true;
		}

		// ****************

		RemoteCallDaemon res_rc_daemon =
			new RemoteCallDaemon(RemoteCall.CallType.PEER,
								 this.netConfig().resourcePort(),
								 this.service(),null);
		this.res_svc_thread =
			new Thread(res_rc_daemon,"Resource-to-Peer-daemon");
		//this.res_svc_thread.start(); -- no, all Threads started together

		// ****************

		RemoteCallDaemon ua_rc_daemon =
			new RemoteCallDaemon(RemoteCall.CallType.PEER,
								 this.netConfig().userPort(),
								 this.service(),null);

		this.ua_svc_thread = new Thread(ua_rc_daemon,"User-to-Peer-daemon");
		//this.ua_svc_thread.start(); -- no, all Threads started together

		// ****************

		final LivePeer self = this;
		final ExternalPeerHandle self_external_peer_handle =
			(ExternalPeerHandle)this.service().createHandle();
		RemoteCallStatusListener rsl =
			new RemoteCallStatusListener() {
				public void isOnline() {
					try {
						self.searchEngine().
							setOwnerPeer(self_external_peer_handle);
					}
					catch (GridException e) {
						//Logger.getLogger("lbg.peer").
						logger.log(Level.WARNING,
							"Could not register Peer with Search engine",e);
					}
				}
			};
		RemoteCallDaemon peer_rc_daemon =
			new RemoteCallDaemon(RemoteCall.CallType.PEER,
								 this.netConfig().peerPort(),
								 this.service(),rsl);

		this.peer_svc_thread = new Thread(peer_rc_daemon,"Peer-to-Peer-daemon");
		//this.peer_svc_thread.start(); -- no, all Threads started together

		// ****************

		LiveSearchEngineClient se_client =
			(LiveSearchEngineClient)super.searchEngine();
		se_client.startService();
		
		// ****************

		SelfEventProcessorThread sep_thread =
			new SelfEventProcessorThread(this);
		// no, all Threads started together
		//this.sched_ex_svc = Executors.newScheduledThreadPool(1);
		//this.sched_fut = this.sched_ex_svc.
		//	scheduleAtFixedRate(sep_thread,1,1,TimeUnit.SECONDS);

		// ****************

		LiveJobPreQueue prequeue =
			(LiveJobPreQueue)this.qms().prequeue();
		prequeue.startService();

		// ****************

		LiveSchedulerController sched_ctrlr =
			(LiveSchedulerController)this.schedulerController();
		sched_ctrlr.startService();
		
		LiveNegotiatorController neg_ctrlr =
			(LiveNegotiatorController)this.negotiatorController();
		neg_ctrlr.startService();

		// ****************

		this.res_svc_thread.start();
		
		this.ua_svc_thread.start();
		
		this.peer_svc_thread.start();
		
		this.sched_ex_svc = Executors.newScheduledThreadPool(1);
		this.sched_ex_svc.
			scheduleAtFixedRate(sep_thread,1,1,TimeUnit.SECONDS);

		// ****************

		this.websvr.startService();
		
	}

	public boolean isAlive() {
		
		return
			((this.res_svc_thread != null) ?
				this.res_svc_thread.isAlive() : false) &&
			((this.ua_svc_thread != null) ?
				this.ua_svc_thread.isAlive() : false) &&
			((this.peer_svc_thread != null) ?
				this.peer_svc_thread.isAlive() : false);
	}

	public void checkLiveness() {
		
		if (res_svc_thread.isAlive() == false) {
			System.err.println("Resource-to-Peer daemon died unexpectedly");
		}
		if (ua_svc_thread.isAlive() == false) {
			System.err.println("User-to-Peer daemon died unexpectedly");
		}
		if (peer_svc_thread.isAlive() == false) {
			System.err.println("Peer-to-Peer daemon died unexpectedly");
		}

	}
	
	// ****************************************************************

	public GridDataManager gdm() {
		
		return this.gdm;
		
	}
	
	public Logger logger() {
		
		return this.logger;
		
	}

	// ****************************************************************

	private final GridDataManager gdm;
	private final Logger          logger;
	private final PeerWebServer   websvr;
	
	private       ScheduledExecutorService sched_ex_svc;
	//private ScheduledFuture sched_fut;

	private       Thread res_svc_thread;
	private       Thread ua_svc_thread;
	private       Thread peer_svc_thread;
	private       boolean online;

}

