/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Peer.java

package lbg.peer;

import static lbg.LightweightBarteringGrid.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.exception.*;
import lbg.peer.config.*;
import lbg.peer.core.*;
import lbg.peer.core.qms.*;
import lbg.peer.core.rms.*;
import lbg.peer.core.xms.*;
import lbg.peer.core.negotiator.*;
import lbg.peer.core.scheduler.*;
import lbg.peer.register.*;
import lbg.peer.service.*;
import lbg.peer.strategy.*;
import lbg.protocol.data.*;
import lbg.searchengine.*;
import lbg.searchengine.service.*;
import java.util.logging.Logger;

/** Peer.
 * @author Cyril Briquet, Xavier Dalem
 */
public abstract class Peer {
	
	/** Creates a new Peer.
	 * @param peer_basic_config Peer basic configuration
	 * @param peer_network_config Peer network configuration
	 * @param peer_data_config Peer data configuration
	 * @param peer_strategy_config Peer strategy configuration
	 * @param peer_service_factory Peer service factory
	 * @param ctrlr_factory controller factory
	 * @param se_handle search engine handle
	 */
	Peer(PeerBasicConfiguration peer_basic_config,
	     PeerNetworkConfiguration peer_network_config,
		 PeerDataConfiguration peer_data_config,
		 PeerStrategyConfiguration peer_strategy_config,
		 PeerServiceFactory peer_service_factory,
		 ControllerFactory ctrlr_factory,
		 SearchEngineHandle se_handle)
		throws GridException {

		if (peer_basic_config == null) {
			throw new GridException("Peer(): illegal "+
									"Peer Basic Configuration");
		}
		if (peer_network_config == null) {
			throw new GridException("Peer(): illegal "+
									"Peer Network configuration");
		}
		if (peer_data_config == null) {
			throw new GridException("Peer(): illegal "+
									"Peer Data configuration");
		}
		if (peer_strategy_config == null) {
			throw new GridException("Peer(): illegal "+
									"Peer Strategy configuration");
		}
		if (peer_service_factory == null) {
			throw new GridException("Peer(): illegal "+
									"Peer service factory");
		}
		if (ctrlr_factory == null) {
			throw new GridException("Peer(): illegal Controller Factory");
		}
		if (se_handle == null) {
			throw new GridException("Peer(): illegal Search Engine handle");
		}
		
		// ****************
		
		this.id = peer_basic_config.PEER_ID();
		this.peer_basic_config = peer_basic_config;
		this.peer_network_config = peer_network_config;
		this.peer_data_config = peer_data_config;
		this.peer_strategy_config = peer_strategy_config;
		this.se_client = SearchEngineClientFactory.
			createSearchEnglineClient(se_handle,null,
									  peer_basic_config.SEARCH_TIMEOUT);

		this.strategy =
			new StrategyDescriptor(this,peer_strategy_config);
		this.register =
			new PeerRegister(this,this.strategy,peer_strategy_config);
		this.negotiator = strategy.createNegotiator();
		this.scheduler  = strategy.createScheduler(this.negotiator);

		this.rms = new ResourceManager(this,
			peer_data_config.PEER_STORAGE_AFFINITY(),
			peer_data_config.PEER_DATA_REPLICATION(),
			peer_data_config.PEER_IDLE_REPLICATION_RATIO(),
			peer_strategy_config.PEER_PSUFFERAGE());
		this.qms = new QueueManager(this,this.scheduler,
									peer_data_config.PEER_TTG_POLICY());
		this.xms = new ExecutionManager(this,
			peer_strategy_config.CONSUMPTION_TIMEOUT());
		//this.gdm = ...;
		
		this.negotiator_controller =
			ctrlr_factory.createNegotiatorController(this,this.negotiator);
		this.scheduler_controller =
			ctrlr_factory.createSchedulerController(this,this.scheduler);

		this.peer_service =
			peer_service_factory.createPeerService(this);
		
	}
	
	public String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[Peer ");
		sb.append(this.id);
		sb.append("]");
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.peer_basic_config);
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.peer_network_config);
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.peer_data_config);
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.peer_strategy_config);
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.se_client);
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.rms);
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.qms);
		if (this.gdm() != null) {
			sb.append("\n  ->  \n  ->  ");
			sb.append(this.gdm());
		}
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.xms);
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.strategy);
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.negotiator);
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.scheduler);
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.peer_service);
		sb.append("\n  ->  \n  ->  ");
		sb.append(this.register);

		return sb.toString();

	}

	public String toStringSummary() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[Peer ").append(this.id).append("] @ ").
			append(peer_network_config.hostname()).
			append(':').append(this.peer_network_config.peerPort()).
			append(',').append(this.peer_network_config.resourcePort()).
			append(',').append(this.peer_network_config.userPort());
		
		return sb.toString();

	}

	// ****************************************************************

	public PeerService service() {
		
		return this.peer_service;
		
	}

	// ****************************************************************

	public String id() {

		return this.id;
		
	}

	public PeerBasicConfiguration basicConfig() {
		
		return this.peer_basic_config;
		
	}

	public PeerNetworkConfiguration netConfig() {
		
		return this.peer_network_config;
		
	}
	
	public PeerDataConfiguration dataConfig() {
		
		return this.peer_data_config;
		
	}
	
	public PeerStrategyConfiguration strategyConfig() {
		
		return this.peer_strategy_config;
		
	}
	
	public SearchEngineClient searchEngine() {
		
		return this.se_client;
		
	}

	// ****************************************************************

	public ResourceManager rms() {
		
		return this.rms;
		
	}
	
	public QueueManager qms() {
	
		return this.qms;
		
	}
	
	public ExecutionManager xms() {
	
		return this.xms;
		
	}

	public abstract GridDataManager gdm();
	
	public abstract Logger logger();
	
	public PeerRegister register() {
		
		return this.register;
		
	}

	public ISchedulerStats scheduler() {
	
		return this.scheduler;
		
	}
	
	public SchedulerController schedulerController() {
		
		return this.scheduler_controller;
		
	}
	
	public NegotiatorController negotiatorController() {
		
		return this.negotiator_controller;
		
	}

	// ****************************************************************

	public void processSelfEvents() throws GridException {

		synchronized(this.rms) {

			// Update Request Manager statistics
			this.qms.updateStatistics();
			
			// Preempt/Cancel timed-out Tasks
			boolean timed_out_tasks = false;

			//synchronized (this.rms) {
			timed_out_tasks = this.xms.processTimedOutConsumptionTasks();
			//}
			if (timed_out_tasks) {
				this.scheduler_controller.
					activateLocalAndConsumptionScheduling();
			}

			//synchronized (this.rms) {
			timed_out_tasks = this.xms.processTimedOutLocalTasks();
			//}
			if (timed_out_tasks) {
				this.scheduler_controller.activateLocalAndSupplyingScheduling();
			}
			
			//synchronized (this.rms) {
			timed_out_tasks = this.xms.processTimedOutSupplyingTasks();
			//}
			if (timed_out_tasks) {
				this.scheduler_controller.activateLocalAndSupplyingScheduling();
			}
			
			// Evaluate pending Negotiation signals
			NeighborhoodNegotiationProfile grid_negotiation_profile =
				this.register.getNeighborhoodNegotiationProfile();
			if (grid_negotiation_profile.couldEvaluateSupplyingRequests()) {
				this.negotiator_controller.
					activateSupplyingRequestsEvaluation();
			}
			if (grid_negotiation_profile.couldEvaluateConsumptionGrants()) {
				this.scheduler_controller.
					activateConsumptionScheduling();
			}
			
			// Generate new Supplying Requests (AFTER Negotiation signals eval)
			if (grid_negotiation_profile.couldGenerateSupplyingRequests() &&
				(this.qms.countWaitingLocalTasks() > 0)) {
				this.negotiator_controller.
					activateSupplyingRequestsGeneration();
			}
			
			if ((getMode() == LIVE_MODE) && (getTime() % 5 == 0)) {
				this.rms().updateUtilizationValue();
			}
			
		}
		
	}

	// ****************************************************************

	private final String id;
	private PeerBasicConfiguration    peer_basic_config;
	private PeerNetworkConfiguration  peer_network_config;
	private PeerDataConfiguration     peer_data_config;
	private PeerStrategyConfiguration peer_strategy_config;
	private SearchEngineClient        se_client;

	private ResourceManager    rms;
	private QueueManager       qms;
	private ExecutionManager   xms;
	private PeerRegister       register;
	private StrategyDescriptor strategy;
	        Negotiator         negotiator;
	        Scheduler          scheduler;
	private PeerService        peer_service;
	
	private SchedulerController  scheduler_controller;
	private NegotiatorController negotiator_controller;

	// ****************************************************************

}

