/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PeerDataConfiguration.java

package lbg.peer.config;

import lbg.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.protocol.data.*;
import lbg.simulator.config.*;
import java.util.Properties;
import java.util.StringTokenizer;

/**
 * @author Cyril Briquet, Xavier Dalem
 */
public class PeerDataConfiguration {

	public PeerDataConfiguration(Properties props) throws GridException {
		
		if (props == null) {
			throw new GridException("PeerDataConfiguration(): "+
									"illegal Properties");
		}

		this.CACHE_DIR = loadProperty("CACHE_DIR",props);
		
		this.DATA_HOST_POLICY = buildHostPolicy(
			props.getProperty("DATA_HOST_POLICY")); // Host policy can be null
		
		String candidate_capacity =
			loadProperty("SUPPLYING_CACHE_CAPACITY",props);
		this.SUPPLYING_CACHE_CAPACITY =
			(candidate_capacity == null ||
			 candidate_capacity.length() == 0) ? 
			 GridDataManager.INFINITE_CAPACITY :
			 loadInt(candidate_capacity);
		this.PEER_TTG_POLICY =
			loadBoolean(loadProperty("PEER_TTG_POLICY",props));
		this.PEER_STORAGE_AFFINITY =
			loadBoolean(loadProperty("PEER_STORAGE_AFFINITY",props));
		this.PEER_DATA_REPLICATION =
			loadBoolean(loadProperty("PEER_DATA_REPLICATION",props));
		this.PEER_IDLE_REPLICATION_RATIO =
			loadFloat(loadProperty("PEER_IDLE_REPLICATION_RATIO",props));

		// ****************

		checkProperties();

	}
	
	public PeerDataConfiguration(GridConfiguration grid_config,int g)
		throws GridException {
		
		if (LightweightBarteringGrid.getMode() != SIM_MODE) {
			throw new GridException("PeerDataConfiguration(): "+
									"wrong constructor");
		}
		
		if (grid_config == null) {
			throw new GridException("PeerDataConfiguration(): "+
									"illegal Grid Strategy configuration");
		}
		
		if ((g < 0) || (g >= grid_config.size())) {
			throw new GridException("PeerDataConfiguration(): "+
									"illegal group index");
		}

		this.CACHE_DIR = "/tmp/peer_cache_dir/";
		this.DATA_HOST_POLICY = 0;
		this.SUPPLYING_CACHE_CAPACITY =
			GridDataManager.INFINITE_CAPACITY;
		this.PEER_TTG_POLICY = grid_config.PEER_TTG_POLICY()[g];
		this.PEER_STORAGE_AFFINITY =
			grid_config.PEER_STORAGE_AFFINITY()[g];
		this.PEER_DATA_REPLICATION =
			grid_config.PEER_DATA_REPLICATION()[g];
		this.PEER_IDLE_REPLICATION_RATIO =
			grid_config.PEER_IDLE_REPLICATION_RATIO()[g];

		// ****************

		checkProperties();

	}
	
	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PeerDataConfiguration]");
		sb.append(format(true));
		
		return sb.toString();
		
	}
	
	/** Outputs the PeerDataConfiguration,
	 * formatted either as a property file or as a log file.
	 * @param log_format true to format PeerDataConfiguration as a log file
	 * @return formatted PeerDataConfiguration
	 */
	public String format(boolean log_format) {
		
		StringBuilder sb = new StringBuilder();

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("CACHE_DIR                    = ");
		sb.append(this.CACHE_DIR);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("DATA_HOST_POLICY             = ");
		sb.append(this.DATA_HOST_POLICY);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("SUPPLYING_CACHE_CAPACITY     = ");
		sb.append(this.SUPPLYING_CACHE_CAPACITY);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_TTG_POLICY              = ");
		sb.append(this.PEER_TTG_POLICY);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_STORAGE_AFFINITY        = ");
		sb.append(this.PEER_STORAGE_AFFINITY);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_DATA_REPLICATION        = ");
		sb.append(this.PEER_DATA_REPLICATION);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_IDLE_REPLICATION_RATIO  = ");
		sb.append(this.PEER_IDLE_REPLICATION_RATIO);

		return sb.toString();

	}
	
	// ****************************************************************

	public String CACHE_DIR() {
		
		return this.CACHE_DIR;
		
	}

	public int DATA_HOST_POLICY() {
		
		return this.DATA_HOST_POLICY;
		
	}

	public int SUPPLYING_CACHE_CAPACITY() {
		
		return this.SUPPLYING_CACHE_CAPACITY;
		
	}
	
	public boolean PEER_TTG_POLICY() {
		
		return this.PEER_TTG_POLICY;
		
	}

	public boolean PEER_STORAGE_AFFINITY() {
		
		return this.PEER_STORAGE_AFFINITY;
		
	}

	public boolean PEER_DATA_REPLICATION() {
		
		return this.PEER_DATA_REPLICATION;
		
	}

	public float PEER_IDLE_REPLICATION_RATIO() {
		
		return this.PEER_IDLE_REPLICATION_RATIO;
		
	}
	
	// ****************************************************************

	/** Gets a downloading policy for data.
	 * Download options currently are <i>INLINE</i> (implicitly set as the
	 * best one), <i>TORRENT</i> and <i>FTP</i>.
	 * @param policy An ordered list of download options, highest priority first
	 * @return An array of download options, in decreasing order of preference.
	 * null if no policy was given. The returned
	 * array will always all valid options.
	 * @throws GridException if the policy is is badly formed or incomplete
	 * @deprecated You shoudln't need to build any of these, the receiving
	 * policy is now enforced by the GridDataManager.
	 */
	@Deprecated public static int[] buildRcvPolicy(String policy) throws GridException {
		if (policy == null)
			return null;
		
		String[] options = policy.split("\\s*,\\s*");
		
		int[] prefs = new int[options.length+1];
		prefs[0] = GridDataManager.SHARED_INLINE; // Forced as highest
		
		boolean has_torrent = false, has_ftp = false;
		
		for (int i = 0; i < options.length; i++) {
			if (options[i].equals("INLINE"))
				prefs[i+1] = GridDataManager.SHARED_INLINE;
			else if (options[i].equals("TORRENT")) {
				prefs[i+1] = GridDataManager.SHARED_TORRENT;
				has_torrent = true;
			}
			else if (options[i].equals("FTP")) {
				prefs[i+1] = GridDataManager.SHARED_FTP;
				has_ftp = true;
			}
			else
				throw new GridException("Unknown reception policy: " + policy);
		}
		
		if (!has_ftp || !has_torrent)
			throw new GridException("Reception policy must include both FTP"
				+ " and TORRENT: you are forced to support both");
		
		return prefs;
	}

	/** Get a host policy for data.
	 * Host options currently are <i>TORRENT</i> and <i>FTP</i>.
	 * @param policy A list of host options.
	 * @return A valid policy (can be empty)
	 */
	private static int buildHostPolicy(String policy) throws GridException {
		int flags = 0;
		
		if (policy == null || policy.equals(""))
			return flags;
		
		StringTokenizer st = new StringTokenizer(policy, ",");
		while (st.hasMoreTokens()) {
			String tok = st.nextToken();
			if (tok.equals("TORRENT"))
				flags |= GridDataManager.HOST_TORRENT_TRACK;
			else if (tok.equals("FTP"))
				flags |= GridDataManager.HOST_FTP;
			else
				throw new GridException("Unknown sharing policy: " + policy);
		}
		
		return flags;
	}

	// ****************************************************************

	private void checkProperties() throws GridException {
		
		if ((this.CACHE_DIR == null) ||
			(this.CACHE_DIR.length() < 1)) {
			throw new GridException("PeerDataConfiguration(): "+
									"illegal Peer cache directory");
		}
		if (this.SUPPLYING_CACHE_CAPACITY < 1
			&& this.SUPPLYING_CACHE_CAPACITY != GridDataManager
			.INFINITE_CAPACITY) {
			throw new GridException("PeerDataConfiguration(): "+
									"illegal Resource "+
									"Supplying cache capacity");
		}
		
		// Since implementation of the hosting heuristics, host policy can
		// be empty
	}

	// ****************************************************************

	private String CACHE_DIR;
	private int DATA_HOST_POLICY;
	private int SUPPLYING_CACHE_CAPACITY;
	private boolean PEER_TTG_POLICY;
	private boolean PEER_STORAGE_AFFINITY;
	private boolean PEER_DATA_REPLICATION;
	private float PEER_IDLE_REPLICATION_RATIO;

	// ****************************************************************

}

