/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PeerNetworkConfiguration.java

package lbg.peer.config;

import lbg.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.common.net.*;
import java.util.Properties;

/**
 * @author Cyril Briquet
 */
public class PeerNetworkConfiguration {

	public PeerNetworkConfiguration(Properties props) throws GridException {
		
		if (props == null) {
			throw new GridException("PeerNetworkConfiguration(): "+
									"illegal Properties");
		}

		this.PEER_HOSTNAME = loadProperty("PEER_HOSTNAME",props);
		this.PEER_PORT = loadInt(loadProperty("PEER_PORT",props));
		this.RESOURCE_PORT = loadInt(loadProperty("RESOURCE_PORT",props));
		this.USER_PORT = loadInt(loadProperty("USER_PORT",props));
		this.BITTORRENT_TRACKER_PORT =
			loadInt(loadProperty("BITTORRENT_TRACKER_PORT",props));
		this.BITTORRENT_DATA_PORT =
			loadInt(loadProperty("BITTORRENT_DATA_PORT",props));
		this.FTP_DATA_PORT =
			loadInt(loadProperty("FTP_DATA_PORT",props));
		this.FTP_REQUEST_PORT =
			loadInt(loadProperty("FTP_REQUEST_PORT",props));
		this.SEARCH_HOSTNAME =
			loadProperty("SEARCH_HOSTNAME",props);
		this.SEARCH_PORT =
			loadInt(loadProperty("SEARCH_PORT",props));

		// ****************
		
		checkProperties();

	}

	public PeerNetworkConfiguration() throws GridException {
		
		if (LightweightBarteringGrid.getMode() != SIM_MODE) {
			throw new GridException("PeerNetworkConfiguration(): "+
									"wrong constructor");
		}

		this.PEER_HOSTNAME = CommonNetHandling.getLocalHostName();
		this.PEER_PORT               = 24;
		this.RESOURCE_PORT           = 24;
		this.USER_PORT               = 24;
		this.BITTORRENT_TRACKER_PORT = 24;
		this.BITTORRENT_DATA_PORT    = 24;
		this.FTP_DATA_PORT           = 24;
		this.FTP_REQUEST_PORT        = 24;
		this.SEARCH_HOSTNAME = CommonNetHandling.getLocalHostName();
		this.SEARCH_PORT             = 24;

		// ****************
		
		checkProperties();

	}
			
	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PeerNetworkConfiguration]");
		sb.append(format(true));
		
		return sb.toString();
		
	}
	
	/** Outputs the PeerNetworkConfiguration,
	 * formatted either as a property file or as a log file.
	 * @param log_format true to format PeerNetworkConfiguration as a log file
	 * @return formatted PeerNetworkConfiguration
	 */
	public String format(boolean log_format) {
		
		StringBuilder sb = new StringBuilder();

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_HOSTNAME            = ");
		sb.append(this.PEER_HOSTNAME);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_PORT                = ");
		sb.append(this.PEER_PORT);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("RESOURCE_PORT            = ");
		sb.append(this.RESOURCE_PORT);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("USER_PORT                = ");
		sb.append(this.USER_PORT);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("BITTORRENT_TRACKER_PORT  = ");
		sb.append(this.BITTORRENT_TRACKER_PORT);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("BITTORRENT_DATA_PORT     = ");
		sb.append(this.BITTORRENT_DATA_PORT);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("FTP_DATA_PORT            = ");
		sb.append(this.FTP_DATA_PORT);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("FTP_REQUEST_PORT         = ");
		sb.append(this.FTP_REQUEST_PORT);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("SEARCH_HOSTNAME          = ");
		sb.append(this.SEARCH_HOSTNAME);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("SEARCH_PORT              = ");
		sb.append(this.SEARCH_PORT);

		return sb.toString();

	}
	
	// ****************************************************************

	/**
	 * @return Peer host name, as reachable by other Peers
	 */
	public String hostname() {
		
		return this.PEER_HOSTNAME;
		
	}
	
	/**
	 * @return valid port for Peer-to-Peer communication
	 */
	public int peerPort() {
		
		return this.PEER_PORT;
		
	}
	
	/**
	 * @return valid port for Resource-to-Peer communication
	 */
	public int resourcePort() {
		
		return this.RESOURCE_PORT;
		
	}
	
	/**
	 * @return valid port for user-to-peer communication
	 */
	public int userPort() {
		
		return this.USER_PORT;
		
	}
	
	/**
	 * @return BitTorrent tracker port
	 */
	public int btTrackerPort() {
		
		return this.BITTORRENT_TRACKER_PORT;
		
	}
	
	/**
	 * @return BitTorrent data port
	 */
	public int btDataPort() {
		
		return this.BITTORRENT_DATA_PORT;
		
	}
	
	public int ftpDataPort() {
		
		return this.FTP_DATA_PORT;
		
	}
	
	public int ftpRequestPort() {
		
		return this.FTP_REQUEST_PORT;
		
	}

	public int webPort() {
		
		return this.WEB_PORT;
		
	}

	// ****************************************************************

	private void checkProperties() throws GridException {

		if ((this.PEER_HOSTNAME == null) ||
			(this.PEER_HOSTNAME.length() < 1)) {
			throw new GridException("checkProperties(): "+
									"illegal Peer hostname");
		}
		if (CommonNetHandling.isValidPort(this.PEER_PORT) == false) {
			throw new GridException("checkProperties(): "+
									"illegal Peer port");
		}
		if (CommonNetHandling.isValidPort(this.RESOURCE_PORT) == false) {
			throw new GridException("checkProperties(): "+
									"illegal Resource port");
		}
		if (CommonNetHandling.isValidPort(this.USER_PORT) == false) {
			throw new GridException("checkProperties(): "+
									"illegal User port");
		}
		if (CommonNetHandling.isValidPort(this.BITTORRENT_TRACKER_PORT) ==
			false) {
			throw new GridException("checkProperties(): "+
									"illegal BitTorrent tracker port");
		}
		if (CommonNetHandling.isValidPort(this.BITTORRENT_DATA_PORT) ==
			false) {
			throw new GridException("checkProperties(): "+
									"illegal BitTorrent data port");
		}
		if (CommonNetHandling.isValidPort(this.FTP_DATA_PORT) == false) {
			throw new GridException("checkProperties(): "+
									"illegal FTP data port");
		}
		if (CommonNetHandling.isValidPort(this.FTP_REQUEST_PORT) ==
			false) {
			throw new GridException("checkProperties(): "+
									"illegal FTP request port");
		}
		if ((this.SEARCH_HOSTNAME == null) ||
				(this.SEARCH_HOSTNAME.length() < 1)) {
				throw new GridException("checkProperties(): "+
										"illegal Peer hostname");
			}
		if (CommonNetHandling.isValidPort(this.SEARCH_PORT) ==
			false) {
			throw new GridException("checkProperties(): "+
									"illegal search port");
		}

	}

	// ****************************************************************

	private final String PEER_HOSTNAME;
	private final int PEER_PORT;
	private final int RESOURCE_PORT;
	private final int USER_PORT;
	private final int BITTORRENT_TRACKER_PORT;
	private final int BITTORRENT_DATA_PORT;
	private final int FTP_DATA_PORT;
	private final int FTP_REQUEST_PORT;
	public final String SEARCH_HOSTNAME;
	public final int SEARCH_PORT;
	public final int WEB_PORT = 50119;

}

