/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PeerStrategyConfiguration.java

package lbg.peer.config;

import lbg.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.peer.strategy.accounting.*;
import lbg.peer.strategy.negotiation.*;
import lbg.peer.strategy.scheduling.*;
import lbg.simulator.config.*;

import java.nio.charset.CoderMalfunctionError;
import java.util.Map;
import java.util.Properties;
import java.util.TreeMap;

/**
 * @author Cyril Briquet
 */
public class PeerStrategyConfiguration {

	public PeerStrategyConfiguration(Properties props) throws GridException {
		
		if (props == null) {
			throw new GridException("PeerStrategyConfiguration(): "+
									"illegal Properties");
		}
		
		this.PEER_DEFAULT_ACCOUNTANT =
			loadProperty(BarteringAccountantType.getPDPLabel(),props);
		this.PEER_NEGOTIATION_SUPPLYING_POLICY =
			loadProperty(SupplyingNegotiationType.getPDPLabel(),props);
		this.PEER_NEGOTIATION_CONSUMPTION_POLICY =
			loadProperty(ConsumptionNegotiationType.getPDPLabel(),props);

		this.PEER_SCHEDULING_LOCAL_POLICY =
			loadProperty(LocalSchedulingType.getPDPLabel(),props);
		this.PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY =
			loadProperty(SupplyingRunningPreemptionType.getPDPLabel(),props);
		this.PEER_PREEMPTION_SUPPLYING_WAITING_POLICY =
			loadProperty(SupplyingWaitingPreemptionType.getPDPLabel(),props);
		this.PEER_FILTERING_SUPPLYING_POLICY =
			loadProperty(SupplyingFilteringType.getPDPLabel(),props);
		this.PEER_SCHEDULING_SUPPLYING_POLICY =
			loadProperty(SupplyingSchedulingType.getPDPLabel(),props);
		this.PEER_SCHEDULING_CONSUMPTION_POLICY =
			loadProperty(ConsumptionSchedulingType.getPDPLabel(),props);

		this.REQUESTORS_COUNT_THRESHOLD =
			loadInt(loadProperty("REQUESTORS_COUNT_THRESHOLD",props));
		this.REQUESTORS_TIME_THRESHOLD =
			loadInt(loadProperty("REQUESTORS_TIME_THRESHOLD",props));
		this.GRANTORS_COUNT_THRESHOLD =
			loadInt(loadProperty("GRANTORS_COUNT_THRESHOLD",props));
		this.GRANTORS_TIME_THRESHOLD =
			loadInt(loadProperty("GRANTORS_TIME_THRESHOLD",props));
		this.REQUEST_SUPPLYING_TIME_THRESHOLD =
			loadInt(loadProperty("REQUEST_SUPPLYING_TIME_THRESHOLD",props));
		
		this.LOCAL_Q_LEN_PREEMPTION_THRESHOLD =
			loadInt(loadProperty("LOCAL_Q_LEN_PREEMPTION_THRESHOLD",props));
		
		this.PEER_PSUFFERAGE =
			loadBoolean(loadProperty("PEER_PSUFFERAGE",props));

		this.PEER_Q_FILTERING_THRESHOLD =
			loadInt(loadProperty("PEER_Q_FILTERING_THRESHOLD",props));
		this.CONSUMPTION_TIMEOUT =
			loadLong(loadProperty("CONSUMPTION_TIMEOUT",props));

		this.DO_CONSUMPTION_BLACKLIST =
			loadBoolean(loadProperty("DO_CONSUMPTION_BLACKLIST",props));
		this.CONSUMPTION_BLACKLIST_PROBA =
			loadFloat(loadProperty("CONSUMPTION_BLACKLIST_PROBA",props));

		// ****************
		
		checkProperties();

	}
	
	public PeerStrategyConfiguration(GridStrategyConfiguration
									 grid_strategy_config,int g)
		throws GridException {
		
		if (LightweightBarteringGrid.getMode() != SIM_MODE) {
			throw new GridException("PeerStrategyConfiguration(): "+
									"wrong constructor");
		}

		if (grid_strategy_config == null) {
			throw new GridException("PeerStrategyConfiguration(): "+
									"illegal Grid Strategy configuration");
		}
		
		if ((g < 0) || (g >= grid_strategy_config.size())) {
			throw new GridException("PeerStrategyConfiguration(): "+
									"illegal group index");
		}
		
		this.PEER_DEFAULT_ACCOUNTANT =
			grid_strategy_config.
				PEER_DEFAULT_ACCOUNTANT()[g];
		this.PEER_NEGOTIATION_SUPPLYING_POLICY =
			grid_strategy_config.
				PEER_NEGOTIATION_SUPPLYING_POLICY()[g];
		this.PEER_NEGOTIATION_CONSUMPTION_POLICY =
			grid_strategy_config.
				PEER_NEGOTIATION_CONSUMPTION_POLICY()[g];
		this.PEER_SCHEDULING_LOCAL_POLICY =
			grid_strategy_config.
				PEER_SCHEDULING_LOCAL_POLICY()[g];
		this.PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY =
			grid_strategy_config.
				PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY()[g];
		this.PEER_PREEMPTION_SUPPLYING_WAITING_POLICY =
			grid_strategy_config.
				PEER_PREEMPTION_SUPPLYING_WAITING_POLICY()[g];
		this.PEER_FILTERING_SUPPLYING_POLICY =
			grid_strategy_config.
				PEER_FILTERING_SUPPLYING_POLICY()[g];
		this.PEER_SCHEDULING_SUPPLYING_POLICY =
			grid_strategy_config.
				PEER_SCHEDULING_SUPPLYING_POLICY()[g];
		this.PEER_SCHEDULING_CONSUMPTION_POLICY =
			grid_strategy_config.
				PEER_SCHEDULING_CONSUMPTION_POLICY()[g];
		
		this.REQUESTORS_COUNT_THRESHOLD =
			grid_strategy_config.REQUESTORS_COUNT_THRESHOLD()[g];
		this.REQUESTORS_TIME_THRESHOLD =
			grid_strategy_config.REQUESTORS_TIME_THRESHOLD()[g];
		this.GRANTORS_COUNT_THRESHOLD =
			grid_strategy_config.GRANTORS_COUNT_THRESHOLD()[g];
		this.GRANTORS_TIME_THRESHOLD =
			grid_strategy_config.GRANTORS_TIME_THRESHOLD()[g];
		this.REQUEST_SUPPLYING_TIME_THRESHOLD =
			grid_strategy_config.REQUEST_SUPPLYING_TIME_THRESHOLD()[g];
		
		this.LOCAL_Q_LEN_PREEMPTION_THRESHOLD =
			grid_strategy_config.LOCAL_Q_LEN_PREEMPTION_THRESHOLD()[g];
			
		this.PEER_PSUFFERAGE =
			grid_strategy_config.PEER_PSUFFERAGE()[g];

		this.PEER_Q_FILTERING_THRESHOLD =
			grid_strategy_config.PEER_Q_FILTERING_THRESHOLD()[g];
		this.CONSUMPTION_TIMEOUT =
			grid_strategy_config.CONSUMPTION_TIMEOUT()[g];

		this.DO_CONSUMPTION_BLACKLIST =
			grid_strategy_config.DO_CONSUMPTION_BLACKLIST()[g];
		this.CONSUMPTION_BLACKLIST_PROBA =
			grid_strategy_config.CONSUMPTION_BLACKLIST_PROBA()[g];

		// ****************

		checkProperties();
		
	}
	
	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PeerStrategyConfiguration]");
		sb.append(format(true));
		
		return sb.toString();
		
	}
	
	/** Outputs the PeerStrategyConfiguration,
	 * formatted either as a property file
	 * or as a log file.
	 * @param log_format true to format PeerStrategyConfiguration as a log file
	 * @return formatted PeerStrategyConfiguration
	 */
	public String format(boolean log_format) {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_SCHEDULING_LOCAL_POLICY         = ");
		sb.append(this.PEER_SCHEDULING_LOCAL_POLICY);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY    = ");
		sb.append(this.PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_PREEMPTION_SUPPLYING_WAITING_POLICY    = ");
		sb.append(this.PEER_PREEMPTION_SUPPLYING_WAITING_POLICY);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_FILTERING_SUPPLYING_POLICY     = ");
		sb.append(this.PEER_FILTERING_SUPPLYING_POLICY);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_SCHEDULING_SUPPLYING_POLICY     = ");
		sb.append(this.PEER_SCHEDULING_SUPPLYING_POLICY);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_SCHEDULING_CONSUMPTION_POLICY   = ");
		sb.append(this.PEER_SCHEDULING_CONSUMPTION_POLICY);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_NEGOTIATION_SUPPLYING_POLICY    = ");
		sb.append(this.PEER_NEGOTIATION_SUPPLYING_POLICY);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_NEGOTIATION_CONSUMPTION_POLICY  = ");
		sb.append(this.PEER_NEGOTIATION_CONSUMPTION_POLICY);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_DEFAULT_ACCOUNTANT                       = ");
		sb.append(this.PEER_DEFAULT_ACCOUNTANT);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("REQUESTORS_COUNT_THRESHOLD                    = ");
		sb.append(this.REQUESTORS_COUNT_THRESHOLD);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("REQUESTORS_TIME_THRESHOLD                     = ");
		sb.append(this.REQUESTORS_TIME_THRESHOLD);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("GRANTORS_COUNT_THRESHOLD                      = ");
		sb.append(this.GRANTORS_COUNT_THRESHOLD);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("GRANTORS_TIME_THRESHOLD                       = ");
		sb.append(this.GRANTORS_TIME_THRESHOLD);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("REQUEST_SUPPLYING_TIME_THRESHOLD              = ");
		sb.append(this.REQUEST_SUPPLYING_TIME_THRESHOLD);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("LOCAL_Q_LEN_PREEMPTION_THRESHOLD              = ");
		sb.append(this.LOCAL_Q_LEN_PREEMPTION_THRESHOLD);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_PSUFFERAGE                               = ");
		sb.append(this.PEER_PSUFFERAGE);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_Q_FILTERING_THRESHOLD                    = ");
		sb.append(this.PEER_Q_FILTERING_THRESHOLD);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("CONSUMPTION_TIMEOUT                           = ");
		sb.append(this.CONSUMPTION_TIMEOUT);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("DO_CONSUMPTION_BLACKLIST                      = ");
		sb.append(this.DO_CONSUMPTION_BLACKLIST);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("CONSUMPTION_BLACKLIST_PROBA                   = ");
		sb.append(this.CONSUMPTION_BLACKLIST_PROBA);

		return sb.toString();

	}

	// ****************************************************************

	public String PEER_DEFAULT_ACCOUNTANT() {
	
		return this.PEER_DEFAULT_ACCOUNTANT;
		
	}

	public String PEER_NEGOTIATION_SUPPLYING_POLICY() {
	
		return this.PEER_NEGOTIATION_SUPPLYING_POLICY;
		
	}

	public String PEER_NEGOTIATION_CONSUMPTION_POLICY() {
	
		return this.PEER_NEGOTIATION_CONSUMPTION_POLICY;
		
	}

	public String PEER_SCHEDULING_LOCAL_POLICY() {
	
		return this.PEER_SCHEDULING_LOCAL_POLICY;
		
	}
	
	public String PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY() {
	
		return this.PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY;
		
	}

	public String PEER_PREEMPTION_SUPPLYING_WAITING_POLICY() {
	
		return this.PEER_PREEMPTION_SUPPLYING_WAITING_POLICY;
		
	}

	public String PEER_FILTERING_SUPPLYING_POLICY() {
	
		return this.PEER_FILTERING_SUPPLYING_POLICY;
		
	}

	public String PEER_SCHEDULING_SUPPLYING_POLICY() {
	
		return this.PEER_SCHEDULING_SUPPLYING_POLICY;
		
	}

	public String PEER_SCHEDULING_CONSUMPTION_POLICY() {
	
		return this.PEER_SCHEDULING_CONSUMPTION_POLICY;
		
	}
	
	public int REQUESTORS_COUNT_THRESHOLD() {
		
		return this.REQUESTORS_COUNT_THRESHOLD;
		
	}
	
	public int REQUESTORS_TIME_THRESHOLD() {
		
		return this.REQUESTORS_TIME_THRESHOLD;
		
	}

	public int GRANTORS_COUNT_THRESHOLD() {
		
		return this.GRANTORS_COUNT_THRESHOLD;
		
	}
	
	public int GRANTORS_TIME_THRESHOLD() {
		
		return this.GRANTORS_TIME_THRESHOLD;
		
	}
	
	public int REQUEST_SUPPLYING_TIME_THRESHOLD() {
	
		return this.REQUEST_SUPPLYING_TIME_THRESHOLD;
		
	}
	
	public int LOCAL_Q_LEN_PREEMPTION_THRESHOLD() {
		
		return this.LOCAL_Q_LEN_PREEMPTION_THRESHOLD;
		
	}
	
	public boolean PEER_PSUFFERAGE() {
		
		return this.PEER_PSUFFERAGE;
		
	}

	public int PEER_Q_FILTERING_THRESHOLD() {
		
		return this.PEER_Q_FILTERING_THRESHOLD;
		
	}
	
	public long CONSUMPTION_TIMEOUT() {
		
		return this.CONSUMPTION_TIMEOUT;
		
	}
	
	public boolean DO_CONSUMPTION_BLACKLIST() {
	
		return this.DO_CONSUMPTION_BLACKLIST;
		
	}

	public float CONSUMPTION_BLACKLIST_PROBA() {
		
		return this.CONSUMPTION_BLACKLIST_PROBA;
		
	}
	
	// ****************************************************************

	private void checkProperties() throws GridException {
		
		if (this.PEER_DEFAULT_ACCOUNTANT == null) {
			throw new GridException("checkProperties(): "+
									"illegal Peer default Accountant");
		}

		if (this.PEER_NEGOTIATION_SUPPLYING_POLICY == null) {
			throw new GridException("checkProperties(): "+
									"illegal Peer target "+
									"Negotiation Supplying policy");
		}
		if (this.PEER_NEGOTIATION_CONSUMPTION_POLICY == null) {
			throw new GridException("checkProperties(): "+
									"illegal Peer target "+
									"Negotiation Consumption policy");
		}

		if (this.PEER_SCHEDULING_LOCAL_POLICY == null) {
			throw new GridException("checkProperties(): "+
									"illegal Peer target "+
									"Scheduling Local policy");
		}
		if (this.PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY == null) {
			throw new GridException("checkProperties(): "+
									"illegal Peer target "+
									"Preemption Supplying running policy");
		}
		if (this.PEER_PREEMPTION_SUPPLYING_WAITING_POLICY == null) {
			throw new GridException("checkProperties(): "+
									"illegal Peer target "+
									"Preemption Supplying waiting policy");
		}
		if (this.PEER_FILTERING_SUPPLYING_POLICY == null) {
			throw new GridException("checkProperties(): "+
									"illegal Peer target "+
									"filtering Supplying policy");
		}
		if (this.PEER_SCHEDULING_SUPPLYING_POLICY == null) {
			throw new GridException("checkProperties(): "+
									"illegal Peer target "+
									"Scheduling Supplying policy");
		}
		if (this.PEER_SCHEDULING_CONSUMPTION_POLICY == null) {
			throw new GridException("checkProperties(): "+
									"illegal Peer target "+
									"Scheduling Consumption policy");
		}
		if (this.REQUESTORS_COUNT_THRESHOLD < 1) {
			throw new GridException("checkProperties(): "+
									"illegal Peer requestors count threshold");
		}
		if (this.REQUESTORS_TIME_THRESHOLD < 0) {
			throw new GridException("checkProperties(): "+
									"illegal Peer requestors time threshold");
		}
		if (this.GRANTORS_COUNT_THRESHOLD < 1) {
			throw new GridException("checkProperties(): "+
									"illegal Peer grantors count threshold");
		}
		if (this.GRANTORS_TIME_THRESHOLD < 0) {
			throw new GridException("checkProperties(): "+
									"illegal Peer grantors time threshold");
		}
		if (this.REQUEST_SUPPLYING_TIME_THRESHOLD < 0) {
			throw new GridException("checkProperties(): "+
									"illegal Peer Request Supplying "+
									"time threshold");
		}
		if (this.LOCAL_Q_LEN_PREEMPTION_THRESHOLD < 1) {
			throw new GridException("checkProperties(): "+
									"illegal Peer local queue length "+
									"preemption threshold");
		}
		if (this.PEER_Q_FILTERING_THRESHOLD < 1) {
			throw new GridException("checkProperties(): "+
									"illegal Peer queue filtering "+
									"threshold");
		}
		if (this.PEER_Q_FILTERING_THRESHOLD >
			this.LOCAL_Q_LEN_PREEMPTION_THRESHOLD) {
			throw new GridException("checkProperties(): "+
									"illegal Peer local queue length "+
									"preemption threshold (smaller than "+
									"Peer queue filtering threshold)");
		}
		if (this.CONSUMPTION_TIMEOUT < 1) {
			throw new GridException("checkProperties(): "+
									"illegal Consumption Tasks timeout");
		}

		if ((this.CONSUMPTION_BLACKLIST_PROBA < 0.0f) ||
			(this.CONSUMPTION_BLACKLIST_PROBA > 1.0f)) {
			throw new GridException("checkProperties(): illegal "+
									"consumption blacklist probabililty");
		}

	}

	// ****************************************************************

	public static final int getPolicyCount() {
		
		return POLICY_COUNT;
		
	}

	public static final int getPDPCount() {
		
		return PDP_PROPS.size();
		
	}

	public static final String[] getPDPList() {
	
		String[] pdp_list = new String[PDP_PROPS.size()];
		int i = -1;
		for (String pdp : PDP_PROPS.keySet()) {
			pdp_list[++i] = pdp;
		}

		return pdp_list;
		
	}

	public static final String[] getPDPLabels(String pdp) throws GridException {
		
		if (pdp == null) {
			throw new GridException("getPDPLabels(): illegal PDP");
		}
		String[] labels = PDP_PROPS.get(pdp);
		if (labels == null) {
			throw new GridException("getPDPLabels(): no such PDP "+pdp);
		}
		
		return labels;
		
	}
	
	// ****************************************************************

	private static final int POLICY_COUNT;
	
	private static final Map<String,String[]> PDP_PROPS =
		new TreeMap<String,String[]>();

	private static final void addToPDPProps(String key,String[] values,
											GridException e) {
		
		if (e == null) {
			e = new GridException("addToPDPProps(): "+
								  "thorough code urgently review required");
		}
		if ((key == null) || (values == null) ||
			(PDP_PROPS.put(key,values)  != null)) {
			throw new CoderMalfunctionError(e);
		}
		
	}
	
	static {
	
		int count = 1;
		count *= BarteringAccountantType.size();
		count *= ConsumptionNegotiationType.size();
		count *= SupplyingNegotiationType.size();
		count *= LocalSchedulingType.size();
		count *= SupplyingRunningPreemptionType.size();
		count *= SupplyingWaitingPreemptionType.size();
		count *= SupplyingFilteringType.size();
		count *= ConsumptionSchedulingType.size();
		count *= SupplyingSchedulingType.size();
		POLICY_COUNT = count;

		GridException e =
			new GridException("PeerStrategyConfiguration static initializer: "+
							  "review code to remove illegal or duplicate PDP");
		
		addToPDPProps(LocalSchedulingType.getPDPLabel(),
					  LocalSchedulingType.getPolicyLabels(),e);
		
		addToPDPProps(SupplyingRunningPreemptionType.getPDPLabel(),
					  SupplyingRunningPreemptionType.getPolicyLabels(),e);
		
		addToPDPProps(SupplyingWaitingPreemptionType.getPDPLabel(),
					  SupplyingWaitingPreemptionType.getPolicyLabels(),e);
		
		addToPDPProps(SupplyingFilteringType.getPDPLabel(),
					  SupplyingFilteringType.getPolicyLabels(),e);
		
		addToPDPProps(SupplyingSchedulingType.getPDPLabel(),
					  SupplyingSchedulingType.getPolicyLabels(),e);
		
		addToPDPProps(ConsumptionSchedulingType.getPDPLabel(),
					  ConsumptionSchedulingType.getPolicyLabels(),e);
		
		addToPDPProps(SupplyingNegotiationType.getPDPLabel(),
					  SupplyingNegotiationType.getPolicyLabels(),e);
		
		addToPDPProps(ConsumptionNegotiationType.getPDPLabel(),
					  ConsumptionNegotiationType.getPolicyLabels(),e);
		
		addToPDPProps(BarteringAccountantType.getPDPLabel(),
					  BarteringAccountantType.getPolicyLabels(),e);
		
	}

	// ****************************************************************

	private final String PEER_DEFAULT_ACCOUNTANT;
	private final String PEER_NEGOTIATION_SUPPLYING_POLICY;
	private final String PEER_NEGOTIATION_CONSUMPTION_POLICY;
	private final String PEER_SCHEDULING_LOCAL_POLICY;
	private final String PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY;
	private final String PEER_PREEMPTION_SUPPLYING_WAITING_POLICY;
	private final String PEER_FILTERING_SUPPLYING_POLICY;
	private final String PEER_SCHEDULING_SUPPLYING_POLICY;
	private final String PEER_SCHEDULING_CONSUMPTION_POLICY;

	private final int REQUESTORS_COUNT_THRESHOLD;
	private final int REQUESTORS_TIME_THRESHOLD;
	private final int GRANTORS_COUNT_THRESHOLD;
	private final int GRANTORS_TIME_THRESHOLD;
	private final int REQUEST_SUPPLYING_TIME_THRESHOLD;
	private final int LOCAL_Q_LEN_PREEMPTION_THRESHOLD;
	private final boolean PEER_PSUFFERAGE;
	private final int PEER_Q_FILTERING_THRESHOLD;
	private final long CONSUMPTION_TIMEOUT;

	private final boolean DO_CONSUMPTION_BLACKLIST;
	private final float CONSUMPTION_BLACKLIST_PROBA;

}

