/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ControllerFactory.java

package lbg.peer.core;

import lbg.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.core.negotiator.*;
import lbg.peer.core.scheduler.*;
import lbg.simulator.*;

/** Factory for scheduling and negotiation controllers.
 * <p>Note: methods are not static because a ControllerFactory
 * may be instanciated multiple times within the same VM,
 * e.g. inside the LBG Simulator.
 * @author Cyril Briquet
 */
public class ControllerFactory {

	/** Creates a ControllerFactory (legal in LIVE mode only).
	 */
	public ControllerFactory() throws GridException {

		if (LightweightBarteringGrid.getMode() == SIM_MODE) {
			throw new GridException("ControllerFactory(): "+
									"unsupported mode");
		}

		this.env_ctrlr = null;
		this.PEER_BASE_NAME = null;

	}

	/** Creates a ControllerFactory (legal in SIM mode only).
	 * @param env_ctrlr Environment Controller
	 * @exception GridException if a parameter is illegal
	 */
	public ControllerFactory(EnvironmentController env_ctrlr,
							 String PEER_BASE_NAME)
		throws GridException {
	
		if (LightweightBarteringGrid.getMode() == LIVE_MODE) {
			throw new GridException("ControllerFactory(): "+
									"unsupported mode");
		}

		if (env_ctrlr == null)  {
			throw new GridException("ControllerFactory(): "+
									"illegal Environment Controller");
		}
		if (PEER_BASE_NAME == null) {
			throw new GridException("ControllerFactory(): "+
									"illegal Peer base name");
		}

		this.env_ctrlr = env_ctrlr;
		this.PEER_BASE_NAME = PEER_BASE_NAME;
		
	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ControllerFactory]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized SchedulerController
		createSchedulerController(Peer peer,Scheduler scheduler)
		throws GridException {
		
		if (peer == null) {
			throw new GridException("createSchedulerController(): "+
									"illegal Peer");
		}
		if (scheduler == null) {
			throw new GridException("createSchedulerController(): "+
									"illegal Scheduler");
		}
		
		switch (LightweightBarteringGrid.getMode()) {
		case LIVE_MODE:
			return new LiveSchedulerController(peer,scheduler);
		case SIM_MODE:
			return new SimulatedSchedulerController(peer,this.env_ctrlr,
													this.PEER_BASE_NAME);
		default:
			throw new GridException("createSchedulerController(): "+
									"unsupported mode");
		}
		
	}
	
	public synchronized NegotiatorController
		createNegotiatorController(Peer peer,Negotiator negotiator)
		throws GridException {
		
		switch (LightweightBarteringGrid.getMode()) {
		case LIVE_MODE:
			return new LiveNegotiatorController(peer,negotiator);
		case SIM_MODE:
			return new SimulatedNegotiatorController(peer,this.env_ctrlr,
													 this.PEER_BASE_NAME);
		default:
			throw new GridException("createSchedulerController(): "+
									"unsupported mode");
		}
		
	}

	// ****************************************************************
	
	private EnvironmentController env_ctrlr;
	private String PEER_BASE_NAME;

}

