/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LiveNegotiatorController.java

package lbg.peer.core.negotiator;

import lbg.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.getStackTrace;
import lbg.peer.*;

import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * @author Cyril Briquet
 */
public class LiveNegotiatorController extends NegotiatorController
	implements Runnable {

	public LiveNegotiatorController(Peer managed_peer,Negotiator negotiator)
		throws GridException {
		
		super(managed_peer);
		
		if (LightweightBarteringGrid.getMode() == SIM_MODE) {
			throw new GridException("LiveNegotiatorController(): "+
									"unsupported mode");
		}

		if (managed_peer == null) {
			throw new GridException("LiveNegotiatorController(): illegap Peer");
		}
		if (negotiator == null) {
			throw new GridException("LiveNegotiatorController(): "+
									"illegal Negotiator");
		}

		this.managed_peer = managed_peer;
		this.negotiator = negotiator;
		this.ctrl_flags = new NegotiatorControlFlags();
		this.q = new LinkedBlockingQueue<Boolean>();
		this.thread_pool =
			new ThreadPool(this,1,true,false,"Negotiator-daemon");
		
		this.max_size = new AtomicInteger(0);

	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[NegotiatorController]",
							   "[LiveNegotiatorController]"));
		
		return sb.toString();
		
	}

	// ****************************************************************


	/** Runnable implementation.
	 */
	public void run() {

		for ( ; ; ) {
			try {
				this.q.take();
				
				if (this.couldNegotiate() == true) {
					this.getAndResetControlFlags(this.ctrl_flags);
					synchronized(this.managed_peer.rms()) {
						this.negotiator.negotiate(this.ctrl_flags);
					}
					this.ctrl_flags.clear();
				}
			}
			catch (Throwable thr) {
				this.managed_peer.logger().
					info("Negotiator.negotiate(): "+getStackTrace(thr));
				if (this.ctrl_flags != null) this.ctrl_flags.clear();
			}
		}

	}

	// ****************************************************************

	public synchronized void startService() throws GridException {
		
		this.thread_pool.startService();
		
	}

	public synchronized void awakeNegotiator() throws GridException {
	
		try {
			int current_q_size = this.q.size();
			if (current_q_size < threshold) {
				this.q.put(Boolean.TRUE);
			}
			
			if (++current_q_size > this.max_size.get()) {
				this.max_size.set(current_q_size);
			}
		}
		catch (InterruptedException e) {
			throw new GridException(e);
		}
		
	}
	
	public int maxSize() {
		
		return this.max_size.get();
		
	}
	
	// ****************************************************************

	private final Peer managed_peer;
	private final Negotiator negotiator;
	private final NegotiatorControlFlags ctrl_flags;
	private final LinkedBlockingQueue<Boolean> q;;
	private final ThreadPool thread_pool;

	private static final int threshold = 2;
	private        AtomicInteger max_size;

}

