/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Negotiator.java

package lbg.peer.core.negotiator;

import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.strategy.negotiation.*;

/**
 * @author Cyril Briquet
 */
public class Negotiator {

	public Negotiator(Peer managed_peer,
					  String target_negotiation_supplying_policy,
					  String target_negotiation_consumption_policy)
		throws GridException {
		
		if (managed_peer == null) {
			throw new GridException("Negotiator(): illegal Peer");
		}
		if (target_negotiation_supplying_policy == null) {
			throw new GridException("Negotiator(): illegal target "+
									"Negotiation Supplying policy");
		}
		if (target_negotiation_consumption_policy == null) {
			throw new GridException("Negotiator(): illegal target "+
									"Negotiation Consumption policy");
		}

		this.managed_peer = managed_peer;
		this.target_negotiation_supplying_policy =
			target_negotiation_supplying_policy;
		this.target_negotiation_consumption_policy =
			target_negotiation_consumption_policy;

		try {
			SupplyingNegotiationType snt =
				SupplyingNegotiationType.
					valueOf(this.target_negotiation_supplying_policy);
			this.supplying_policy =
				snt.createSupplyingNegotiationPolicy(this.managed_peer);
		}
		catch (IllegalArgumentException e) {
			throw new GridException("Negotiator(): illegal "+
									"Negotiation Supplying policy "+
									this.target_negotiation_supplying_policy,
									e);
		}
		try {
			ConsumptionNegotiationType cnt =
				ConsumptionNegotiationType.
					valueOf(this.target_negotiation_consumption_policy);
			this.consumption_policy =
				cnt.createConsumptionNegotiationPolicy(this.managed_peer);
		}
		catch (IllegalArgumentException e) {
			throw new GridException("Negotiator(): illegal "+
									"Negotiation Consumption policy "+
									this.target_negotiation_consumption_policy,
									e);
		}
		
	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[Negotiator ");
		sb.append(this.managed_peer.id());
		sb.append("]");
		
		sb.append("\n  ->  ");
		sb.append("Negotiation Supplying policy=");
		sb.append(this.target_negotiation_supplying_policy);
		sb.append("\n  ->  ");
		sb.append("Negotiation Consumption policy=");
		sb.append(this.target_negotiation_consumption_policy);
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized void negotiate(NegotiatorControlFlags ctrl_flags)
		throws GridException {

		boolean gen_supplying_requests_flag =
			ctrl_flags.getGenSupplyingRequestsFlag();
		boolean eval_supplying_requests_flag =
			ctrl_flags.getEvalSupplyingRequestsFlag();
			
		if ((gen_supplying_requests_flag == false) &&
			(eval_supplying_requests_flag == false)) {
			throw new GridException("negotiate(): Negotiator activated "+
									"but there is nothing to negotiate");
		}
		
		// ****************

		if (gen_supplying_requests_flag) {
			synchronized(this.managed_peer.rms()) {
				this.generateSupplyingRequests();
			}
		}

		if (eval_supplying_requests_flag) {
			synchronized(this.managed_peer.rms()) {
				this.evaluateSupplyingRequests();
			}
		}

	}
	
	// ****************************************************************

	private synchronized void generateSupplyingRequests()
		throws GridException {
			
		// Synchronize on RMS in order to avoid race conditions due to long,
		// multi-calls transactions btwn TaskManager, ResourceManager
		synchronized(this.managed_peer.rms()) {
			this.consumption_policy.generateSupplyingRequests();
		}
		
	}

	private synchronized void evaluateSupplyingRequests() throws GridException {
		
		// Synchronize on RMS in order to avoid race conditions due to long,
		// multi-calls transactions btwn TaskManager, ResourceManager
		synchronized(this.managed_peer.rms()) {
			this.supplying_policy.evaluateSupplyingRequests();
		}
		
	}

	// ****************************************************************

	private Peer managed_peer;
	private String target_negotiation_supplying_policy;
	private String target_negotiation_consumption_policy;

	private SupplyingNegotiationPolicy supplying_policy;
	private ConsumptionNegotiationPolicy consumption_policy;

}

