/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// QueueManager.java

package lbg.peer.core.qms;

import lbg.*;
import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.math.*;
import static lbg.common.io.CommonStringHandling.DUMMY_LOG;
import lbg.common.queueing.priority.*;
import lbg.peer.*;
import static lbg.peer.core.qms.RequestStatus.*;
import static lbg.peer.core.qms.RequestType.*;
import lbg.peer.core.qms.prequeue.*;
import lbg.peer.core.qms.queue.*;
import lbg.peer.core.scheduler.*;
import lbg.peer.register.*;
import lbg.peer.register.learning.consumption.*;
import lbg.peer.strategy.ranking.*;
import lbg.peer.workflow.*;
import lbg.task.*;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
	  
/** Queueing Management System, or Request Manager.
 * <p>Local Jobs and Supplying Jobs are maintained by order of submission time
 * in their own queue respectively.
 * <p>Each Local Job is currently composed of 1 Bag of Tasks.
 * <p>Each Supplying Job is currently composed of 1 Bag of 1 Task.
 * @author Cyril Briquet
 */
public class QueueManager {

	public QueueManager(Peer managed_peer,Scheduler scheduler,
						boolean enable_ttg)
		throws GridException {

		if (managed_peer == null) {
			throw new GridException("QueueManager(): illegal Peer");
		}
		if (scheduler == null) {
			throw new GridException("QueueManager(): illegal Scheduler");
		}
		
		this.managed_peer = managed_peer;
		this.prequeue = JobPreQueueFactory.
			createJobPreQueue(this.managed_peer,scheduler,enable_ttg);
		this.scheduler = scheduler;
		
		this.local_jobs_heap = new TreeMap<Integer,Job>();
		this.waiting_local_jobs_queue = new PriorityQueue<Integer>();
		this.running_local_jobs_queue = new PriorityQueue<Integer>();
		this.waiting_local_tasks = new AtomicInteger(0);
		this.running_local_tasks = 0;

		this.total_waiting_local_tasks = 0;
		this.total_waiting_local_tasks_count = 0;
		this.total_running_local_tasks = 0;
		this.total_running_local_tasks_count = 0;
		this.total_local_jobs = 0;
		this.total_local_jobs_count = 0;
		
		this.supplying_jobs_heap = new TreeMap<Integer,Job>();
		this.waiting_supplying_jobs_queue = new PriorityQueue<Integer>();
		this.running_supplying_jobs_queue = new PriorityQueue<Integer>();
		this.waiting_supplying_tasks = new AtomicInteger(0);
		this.running_supplying_tasks = 0;
		this.waiting_consumer_peers =
			new ConsumerQueue(this.managed_peer,true);
		this.running_consumer_peers =
			new ConsumerQueue(this.managed_peer,false);

		this.total_waiting_supplying_tasks = 0;
		this.total_waiting_supplying_tasks_count = 0;
		this.total_running_supplying_tasks = 0;
		this.total_running_supplying_tasks_count = 0;
		this.total_supplying_jobs = 0;
		this.total_supplying_jobs_count = 0;

		this.completed_tasks = 0;
		this.completed_tasks_total_runtime = 0;
		this.completed_tasks_total_nominal_runtime = 0;
		this.completed_local_tasks = 0;
		this.completed_local_tasks_total_runtime = 0;
		this.completed_supplying_tasks = 0;
		this.completed_supplying_tasks_total_runtime = 0;
		this.completed_local_jobs = new AtomicLong(0);
		this.completed_local_jobs_total_runtime = 0;
		this.completed_local_jobs_total_runtime_measurement =
			MeasurementFactory.
			  createDescriptiveMeasurement("completed-local-jobs-total-runtime");
		
		this.most_recentely_completed_local_job_time = 0;
		final boolean store_data = false;// true;
		this.local_q_history =
			new LocalQueueHistory(this.managed_peer.id(),store_data);
		
		this.currently_preempting_supplying_tasks_flag = false;

	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[QueueManager ");
		sb.append(this.managed_peer.id());
		sb.append("]");
		
		sb.append("\n  ->  ");
		sb.append("waiting requests { Local Jobs=");
		sb.append(this.local_jobs_heap.size());
		sb.append(", ");
		sb.append("Local Tasks=");
		int local_tasks = 0;
		for (Integer job_id_as_key : this.waiting_local_jobs_queue) {
			local_tasks +=
				this.local_jobs_heap.get(job_id_as_key).
					root().countUnscheduledTasks();
		}
		sb.append(local_tasks);
		int wlt = this.waiting_local_tasks.get();
		if (local_tasks != wlt) {
			sb.append(" (CORRUPTED DATA, EXPECTED ");
			sb.append(wlt);
			sb.append(")");
		}
		sb.append(", ");
		sb.append("Supplying Tasks=");
		int supplying_tasks = 0;
		for (Integer job_id_as_key : this.waiting_supplying_jobs_queue) {
			int count =
				this.supplying_jobs_heap.get(job_id_as_key).
					root().countUnscheduledTasks();
			supplying_tasks += count;
			assert(count == 1);
		}
		sb.append(supplying_tasks);
		int wst = this.waiting_supplying_tasks.get();
		if (supplying_tasks != wst) {
			sb.append(" (CORRUPTED DATA, EXPECTED ");
			sb.append(wst);
			sb.append(")");
		}

		sb.append(" }");
		
		sb.append("\n  ->  ");
		sb.append("running requests { Local Tasks=");
		local_tasks = 0;
		for (Integer job_id_as_key : this.running_local_jobs_queue) {
			local_tasks +=
				this.local_jobs_heap.get(job_id_as_key).
					root().countScheduledTasks();
		}
		sb.append(local_tasks);
		if (local_tasks != this.running_local_tasks) {
			sb.append(" (CORRUPTED DATA, EXPECTED ");
			sb.append(this.running_local_tasks);
			sb.append(")");
		}
		sb.append(", ");
		sb.append("Supplying Tasks=");
		supplying_tasks = 0;
		for (Integer bot_key : this.running_supplying_jobs_queue) {
			int count =
				this.supplying_jobs_heap.get(bot_key).
					root().countScheduledTasks();
			supplying_tasks += count;
			assert(count == 1);
		}
		sb.append(supplying_tasks);
		if (supplying_tasks != this.running_supplying_tasks) {
			sb.append(" (CORRUPTED DATA, EXPECTED ");
			sb.append(this.running_supplying_tasks);
			sb.append(")");
		}
		sb.append(" }");
		
		sb.append("\n  ->  ");
		sb.append("completed requests { Local Jobs=");
		sb.append(this.completed_local_jobs.get());
		sb.append(", ");
		sb.append("Local Tasks=");
		sb.append(this.completed_local_tasks);
		sb.append(", ");
		sb.append("Supplying Tasks=");
		sb.append(this.completed_supplying_tasks);
		sb.append(" }");
		
		sb.append("\n  ->  ");
		sb.append("mean waiting requests { Local Jobs=");
		sb.append(CommonMath.div_label(this.total_local_jobs,
									   this.total_local_jobs_count));
		sb.append(", ");
		sb.append("Local Tasks=");
		sb.append(CommonMath.div_label(this.total_waiting_local_tasks,
									   this.total_waiting_local_tasks_count));
		sb.append(", ");
		sb.append("Supplying Tasks=");
		sb.append(CommonMath.div_label(this.total_waiting_supplying_tasks,
									   this.total_waiting_supplying_tasks_count));
		sb.append(" }");
		
		sb.append("\n  ->  ");
		sb.append("mean running requests { Local Tasks=");
		sb.append(CommonMath.div_label(this.total_running_local_tasks,
									   this.total_running_local_tasks_count));
		sb.append(", ");
		sb.append("Supplying Tasks=");
		sb.append(CommonMath.div_label(this.total_running_supplying_tasks,
									   this.total_running_supplying_tasks_count));
		sb.append(" }");

		sb.append("\n  ->  ");
		sb.append("mean completion time { Local Job=");
		sb.append(CommonMath.div_label(this.completed_local_jobs_total_runtime,
									   this.completed_local_jobs.get()));

		sb.append(", Local Tasks=");
		sb.append(CommonMath.div_label(this.completed_local_tasks_total_runtime,
									   this.completed_local_tasks));
		sb.append(", Supplying Tasks=");
		sb.append(CommonMath.div_label(this.completed_supplying_tasks_total_runtime,
									   this.completed_supplying_tasks));
		sb.append(", All Tasks=");
		sb.append(CommonMath.div_label(this.completed_tasks_total_runtime,
									   this.completed_tasks));

		sb.append(" }");
		
		sb.append("\n  ->  ");
		sb.append("{SIM} nominal Task completion time=");
		sb.append(CommonMath.div_label((double)this.
										completed_tasks_total_nominal_runtime,
									   ((double)this.completed_tasks) *
										this.managed_peer.rms().
									   	meanNominalResourcePower()));
		
		sb.append("\n  ->  ");
		sb.append("most recent Local Job completion=");
		sb.append(this.most_recentely_completed_local_job_time);
		sb.append(", current time=");
		try {
			sb.append(LightweightBarteringGrid.getTime());
		}
		catch (GridException doomed) { }

		sb.append("\n  ->  ");
		sb.append(this.prequeue);
		
		return sb.toString();

	}

	// ****************************************************************

	public JobPreQueue prequeue() {
		
		return this.prequeue;
		
	}
	
	public boolean hasWaitingLocalTasks() {
		
		return countWaitingLocalTasks() > 0;
		
	}
	
	public boolean hasWaitingSupplyingTasks() {
		
		return countWaitingSupplyingTasks() > 0;
		
	}

	public int countWaitingLocalTasks() {
		
		//this.local_jobs_heap.size() > 0; -- BUG: NO, too "optimistic" ;)
		return this.waiting_local_tasks.get();
		
	}
	
	public int countWaitingSupplyingTasks() {
		
		return this.waiting_supplying_tasks.get();
		
	}

	// ****************************************************************

	public synchronized double getMeanJobCompletionTime() {
		
		return
			((double)this.completed_local_jobs_total_runtime)/
			((double)this.completed_local_jobs.get());
				
	}

	public synchronized double getMeanJobCompletionTimeOfFirstKJobs(int k)
		throws GridException {
		
		if (k < 1) {
			throw new GridException("getMeanJobCompletionTimeOfFirstKJobs(): "+
									"illegal Job count");
		}
		
		return ((DescriptiveMeasurement)this.
				completed_local_jobs_total_runtime_measurement).
				getMeanOfFirstKTerms(k);
		
	}
	
	public synchronized double getMeanJobCompletionTimeOfNm5FirstKJobs()
		throws GridException {
	
			return getMeanJobCompletionTimeOfFirstKJobs(
						Math.max(1,(int)this.completed_local_jobs.get()-5));
		
		}
	
	public synchronized double getStdDevJobCompletionTime() {

		return (this.completed_local_jobs.get() <= 0) ? Double.NaN :
			this.completed_local_jobs_total_runtime_measurement.
				getStandardDeviation();

	}
	
	public              long getCompletedLocalJobsCount() {
		
		return this.completed_local_jobs.get();
		
	}
	
	// ****************************************************************

	/** Queues a Local Job.
	 * @param job Local Job to queue
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized void queueLocalJob(Job job) throws GridException {
		
		if (job == null) {
			throw new GridException("queueLocalJob(): illegal Job");
		}

		// Add Job to local_jobs_heap and waiting_local_jobs_queue
		if (this.local_jobs_heap.put(job.idAsKey(),job) != null) {
			throw new GridException("queueLocalJob(): Job id conflict");
		}
		if (this.waiting_local_jobs_queue.add(job.idAsKey()) == false) {
			throw new GridException("queueLocalJob(): "+
									"corrupted Local Jobs queue");
		}

		this.waiting_local_tasks.addAndGet(job.root().countUnscheduledTasks());

	}

	/** Queues a Supplying Job.
	 * @param job Supplying Job to queue
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized void queueSupplyingJob(Job job,String consumer_id)
		throws GridException {
		
		if (job == null) {
			throw new GridException("queueSupplyingJob(): illegal Job");
		}
		if (consumer_id == null) {
			throw new GridException("queueSupplyingJob(): illegal Consumer id");
		}
		
		// Add Job to supplying_job_heap and waiting_supplying_jobs_queue
		if (this.supplying_jobs_heap.put(job.idAsKey(),job) != null) {
			throw new GridException("queueSupplyingJob(): Job id conflict");
		}
		if (this.waiting_supplying_jobs_queue.add(job.idAsKey()) == false) {
			throw new GridException("queueSupplyingJob(): "+
									"corrupted Supplying Jobs queue");
		}
		// Add Consumer Peer to waiting_consumer_peers
		this.waiting_consumer_peers.addConsumer(consumer_id);

		this.waiting_supplying_tasks.incrementAndGet(); //++

	}

	// ****************************************************************

	/** Returns the number of completed Tasks of target Local Job.
	 * @return number of completed Tasks of target Local Job
	 * @exception GridException if an unexpected error occurs
	 */
	public synchronized int queryLocalJobStatus(int job_id)
		throws GridException {
		
		Job job = getLocalJob(job_id);
		BoT bot = job.root();
		
		return bot.countCompletedTasks();
		
	}
	
	public synchronized boolean isSupplyingTaskRunning(Task t)
		throws GridException {
		
		if (t == null) {
			throw new GridException("isSupplyingTaskRunning(): illegal Task");
		}
		
		Integer tmp_job_id_as_key = Tank.newInteger(t.getSupplyingSpaceJobID());

		boolean is_waiting =
			this.waiting_supplying_jobs_queue.hasElement(tmp_job_id_as_key);
		boolean is_running =
			this.running_supplying_jobs_queue.hasElement(tmp_job_id_as_key);
		
		if (is_waiting == is_running) {
			if (is_waiting == true) {
				throw new GridException("isSupplyingTaskRunning(): "+
										"corrupted data");
			}
			else {
				throw new GridException("isSupplyingTaskRunning(): "+
										"Task not found");
			}
		}

		return is_running;
		
	}
	
	// ****************************************************************

	/** Selects for Local execution
	 * the first queued, yet unscheduled Local Task.
	 * @param head_job_only if true, selects Tasks from head Local Job only
	 * @return Local Task selected for Local execution 
	 * as well as its scheduling context,
	 * or null if there is no Local Task to be scheduled
	 * @exception GridException if queue is corrupted
	 */
	public synchronized RequestContainer
		scheduleFirstLocalTask(boolean head_job_only) throws GridException {

		// Are there waiting Local Tasks ?
		if (this.waiting_local_jobs_queue.size() > 0) {

			// Scan all waiting Local BoT ...
			for (Integer job_id_as_key : this.waiting_local_jobs_queue) {

				Job job = getLocalJob(job_id_as_key);
				BoT bot = job.root();
				// ... until one that has some unscheduled Tasks
				if (bot.hasUnscheduledTasks()) {
					RequestContainer rc = bot.selectFirstLocalTask();

					this.waiting_local_tasks.decrementAndGet(); // --
					this.running_local_tasks++;

					this.running_local_jobs_queue.add(job_id_as_key);
						// add(): no check of retval
					 if (bot.hasUnscheduledTasks() == false) {
						 this.waiting_local_jobs_queue.
						 	remove(job_id_as_key);
					 }
					 return rc;
				}
				
				if (head_job_only) {
					return null;
				}
				
			}
			
			throw new GridException("scheduleFirstLocalTask(): "+
									"corrupted QueueManager data");

		}
		
		return null;
		
	}

	/** Selects for Consumption execution
	 * the first queued, yet unscheduled Local Task.
	 * @param head_job_only if true, selects Tasks from head Local Job only
	 * @return Local Task selected for Consumption execution,
	 * as well as its scheduling context,
	 * or null if there is no Local Task to be scheduled
	 * @exception GridException if queue is corrupted
	 */
	public synchronized RequestContainer
		scheduleFirstConsumptionTask(boolean head_job_only)
		throws GridException {
			
		// Are there waiting Local Tasks ?
		if (this.waiting_local_jobs_queue.size() > 0) {

			// Scan all waiting Local BoT ...
			for (Integer job_id_as_key : this.waiting_local_jobs_queue) {

				Job job = getLocalJob(job_id_as_key);

				// ... until one is not restricted to Local execution and ...
				if (job.consumptionOutcomeTracker().
					shouldRestrictJobToLocalExecution()) {
					continue;
				}
				
				// ... has some unscheduled Tasks
				BoT bot = job.root();
				if (bot.hasUnscheduledTasks()) {
					RequestContainer rc = bot.selectFirstConsumptionTask();
					JobConsumptionOutcomeTracker jcot =
						job.consumptionOutcomeTracker();
					rc.setJobConsumptionOutcomeTracker(jcot);
					JobConsumptionDataTracker jcdt =
						job.consumptionDataTracker();
					rc.setJobConsumptionDataTracker(jcdt);

					this.waiting_local_tasks.decrementAndGet(); //--
					this.running_local_tasks++;

					this.running_local_jobs_queue.add(job_id_as_key);
					// add(): must not check of retval
					if (bot.hasUnscheduledTasks() == false) {
						this.waiting_local_jobs_queue.remove(job_id_as_key);
					}
					return rc;
				}
				
				if (head_job_only) {
					return null;
				}

			}

			// Due to restrictedToLocalExecution(), it's possible to be here

		}
		
		return null;
		
	}
	
	/** Selects for Consumption execution
	 * the next queued, yet unscheduled Local Task
	 * @return Local Task selected for Consumption execution,
	 * as well as its scheduling context,
	 * or null if there remains no Local Task to be scheduled
	 * that is less prioritary than the Task memorized
	 * by the Local queue cursor
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized RequestContainer
		scheduleNextConsumptionTask(LocalQueueCursor lqc) throws GridException {
		
		// Are there waiting Local Tasks ?
		if (this.waiting_local_jobs_queue.size() > 0) {

			if (lqc == null) {
				throw new GridException("scheduleNextConsumptionTask(): "+
										"illegal Local queue cursor");
			}
			Integer previous_job_id_as_key = lqc.jobIDAsKey();

			// Scan all waiting Local BoT ...
			for (Integer job_id_as_key : this.waiting_local_jobs_queue) {
				
				// (while skipping previously enumerated Jobs)
				if (job_id_as_key < previous_job_id_as_key) {
					continue;
				}

				Job job = getLocalJob(job_id_as_key);

				// ... until one is not restricted to Local execution and ...
				if (job.consumptionOutcomeTracker().
					shouldRestrictJobToLocalExecution()) {
					continue;
				}
				
				// ... has some unscheduled Tasks
				BoT bot = job.root();
				if (bot.hasUnscheduledTasks()) {
					RequestContainer rc = bot.selectNextConsumptionTask(lqc);
					if (rc == null) {
						continue; // skip
					}
					JobConsumptionOutcomeTracker jcot =
						job.consumptionOutcomeTracker();
					rc.setJobConsumptionOutcomeTracker(jcot);
					JobConsumptionDataTracker jcdt =
						job.consumptionDataTracker();
					rc.setJobConsumptionDataTracker(jcdt);
					lqc.update(rc); // update Local queue cursor

					this.waiting_local_tasks.decrementAndGet(); //--
					this.running_local_tasks++;

					this.running_local_jobs_queue.add(job_id_as_key);
					// add(): must not check of retval
					if (bot.hasUnscheduledTasks() == false) {
						this.waiting_local_jobs_queue.remove(job_id_as_key);
					}
					return rc;
				}
				
			}
			
		}
		
		return null;
		
	}

	/** Selects for Local execution
	 * the first queued, yet unscheduled Supplying Task
	 * @return Supplying Task selected for Local execution,
	 * as well as its scheduling context,
	 * or null if there is no Supplying Task to be scheduled
	 * @exception GridException if queue is corrupted
	 */
	public synchronized RequestContainer
		scheduleFirstSupplyingTask() throws GridException {
		
		// Are there waiting Supplying Tasks ?
		if (this.waiting_supplying_jobs_queue.size() > 0) {

			// Scan all waiting Supplying BoT ...
			for (Integer job_id_as_key : this.waiting_supplying_jobs_queue) {
				Job job = getSupplyingJob(job_id_as_key);
				BoT bot = job.root();
				// ... until one that has some unscheduled Tasks
				if (bot.hasUnscheduledTasks()) {
					RequestContainer rc = bot.selectFirstSupplyingTask();

					this.waiting_supplying_tasks.decrementAndGet(); //--
					this.running_supplying_tasks++;

					this.running_supplying_jobs_queue.add(job_id_as_key);
						// add(): no check of retval
					if (bot.hasUnscheduledTasks() == false) {
						this.waiting_supplying_jobs_queue.
							remove(job_id_as_key);
						this.waiting_consumer_peers.
							removeConsumer(rc.getTask().consumerID());
						this.running_consumer_peers.
							addConsumer(rc.getTask().consumerID());
					}
					return rc;
				}
			}
			
			throw new GridException("scheduleFirstSupplyingTask(): "+
									"corrupted QueueManager data");

		}
		
		return null;
		
	}

	/** Selects for Local execution
	 * the best Favors-ranked, queued, yet unscheduled Supplying Task
	 * @return Supplying Task selected for Local execution,
	 * as well as its scheduling context,
	 * or null if there is no Supplying Task to be scheduled
	 * @exception GridException if queue is corrupted
	 */
	public synchronized RequestContainer
		scheduleBestRankedSupplyingTask() throws GridException {
		
		// Are there waiting Supplying Tasks ?
		if (this.waiting_supplying_jobs_queue.size() > 0) {

			PeerProfile[] waiting_consumers =
				this.waiting_consumer_peers.getConsumers();
			// Rank waiting Consumer Peers
			Ranking r = new FavorsSupplyingRanking(waiting_consumers);
			if (r.getRankedPeerCount() < 1) {
				throw new GridException("scheduleBestRankedSupplyingTask(): "+
										"inconsistent FavorsSupplyingRanking");
			}
			int[] sorted_waiting_consumer_index = r.peersRanking();
			assert(sorted_waiting_consumer_index != null);
			int i = sorted_waiting_consumer_index[0];
			// Extract best ranked Peer from waiting_supplying_jobs_queue
			String consumer_id = waiting_consumers[i].id();
			
			// Scan all waiting Supplying BoT ...
			for (Integer job_id_as_key : this.waiting_supplying_jobs_queue) {
				Job job = getSupplyingJob(job_id_as_key);
				if (job.ownerID().equals(consumer_id) == false) {
					continue; // skip to next Job until target Consumer found
				}
				BoT bot = job.root();
				// ... until one that has some unscheduled Tasks
				if (bot.hasUnscheduledTasks() == false) {
					throw new GridException("scheduleBestRanked"+
											"SupplyingTask(): "+
											"corrupted QueueManager data");
				}
				RequestContainer rc = bot.selectFirstSupplyingTask();

				this.waiting_supplying_tasks.decrementAndGet(); //--
				this.running_supplying_tasks++;

				this.running_supplying_jobs_queue.add(job_id_as_key);
					// add(): no check of retval
				if (bot.hasUnscheduledTasks() == false) {
					this.waiting_supplying_jobs_queue.
						remove(job_id_as_key);
					this.waiting_consumer_peers.
						removeConsumer(rc.getTask().consumerID());
					this.running_consumer_peers.
						addConsumer(rc.getTask().consumerID());
				}
				return rc;
			}
			
			throw new GridException("scheduleBestRankedSupplyingTask(): "+
									"corrupted QueueManager data");

		}
		
		if (this.waiting_consumer_peers.size() != 0) {
			throw new GridException("scheduleBestRankedSupplyingTask(): "+
									"corrupted QueueManager data ");
		}
		
		return null;
		
	}

	// ****************************************************************

	/** Dequeues a Local Task after its execution has been completed
	 * @param t target Task
	 * @param xstart_time Task execution start time
	 * @param runtime Task run time
	 * @return completion level status (useful for Workflow Jobs only)
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized
		RequestStatus dequeueCompletedLocalTask(Task t,long xstart_time,
												long runtime)
		throws GridException {
		
		return dequeueCompletedLocalOrConsumptionTask(t,xstart_time,
												      runtime,false);
		
	}
	
	/** Dequeues a Consumption Task after its execution has been completed
	 * @param t target Task
	 * @param xstart_time Task execution start time
	 * @param runtime Task run time
	 * @return completion level status (useful for Workflow Jobs only)
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized
		RequestStatus dequeueCompletedConsumptionTask(Task t,long xstart_time,
													  long runtime)
		throws GridException {

		return dequeueCompletedLocalOrConsumptionTask(t,xstart_time,
													  runtime,true);

	}
	
	/** Dequeues a Local or Consumption Task after its execution
	 * has been completed
	 * @param t target Task
	 * @param xstart_time Task execution start time
	 * @param runtime Task run time
	 * @param is_consumption_task true if Task is a Consumption Task
	 * @return completion level status (useful for Workflow Jobs only)
	 * @exception GridException if a parameter is illegal
	 */
	private synchronized RequestStatus
		dequeueCompletedLocalOrConsumptionTask(Task t,long xstart_time,
											   long runtime,
											   boolean is_consumption_task)
		throws GridException {
		
		if (t == null) {
			throw new GridException("dequeueCompletedLocalOr"+
									"ConsumptionTask(): illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("dequeueCompletedLocalOr"+
									"ConsumptionTask(): "+
									"illegal execution start time");
		}
		if (runtime < 0) {
			throw new GridException("dequeueCompletedLocalOr"+
									"ConsumptionTask(): illegal run time");
		}

		// Locate Job in Job heap
		Job job = getLocalJob(t);

		// Locate BoT
		BoT bot = job.root();
		if (bot.id() != t.botID()) {
			throw new GridException("dequeueCompletedLocalOr"+
									"OrConsumptionTask(): BoT id conflict");
		}

		// Flag the Task as completed, update Job Consumption history
		RequestContainer rc = bot.flagTaskAsCompleted(t);
		if (is_consumption_task) {
			job.consumptionOutcomeTracker().
				trackCompletedTask(t.output().supplierID(),runtime);
			job.consumptionDataTracker().trackData(t.output().supplierID(),t);
			this.managed_peer.rms().
				dataTracker().trackData(t.output().supplierID(),t);
			
			if (rc.hasSchedulingContext()) {
				rc.getSchedulingContext().
					setOutcome(t,xstart_time,CONSUMPTION_TASK); // [ML]
			}
		}
		else {
			if (rc.hasSchedulingContext()) {
				/* rc.getSchedulingContext().
					setOutcome(t,xstart_time,LOCAL_TASK); // [ML]
				*/
			}
		}

		// Update stats
		this.running_local_tasks--;
		long now = LightweightBarteringGrid.getTime();
		long rt = now-xstart_time;
		this.completed_tasks++;
		this.completed_tasks_total_runtime += rt;
		this.completed_tasks_total_nominal_runtime += t.nominalRuntime();
		this.completed_local_tasks++;
		this.completed_local_tasks_total_runtime += rt;
		
		// If BoT is completed, remove it from local Task queue
		// If Job is completed, remove it from local Job heap
		//                      and notify the User Agent
		if (bot.isCompleted()) {
			// Notify UserAgent
			this.notifyCompletedJobToUser(t);

			// Remove BoT from running_local_jobs_queue
			if (this.running_local_jobs_queue.remove(t.jobIDAsKey()) == false) {
				throw new GridException("dequeueCompletedLocalOr"+
										"ConsumptionTask(): "+
										"corrupted running Local Jobs queue");
			}

			// Remove Job from Local Job heap
			if (this.local_jobs_heap.remove(t.jobIDAsKey()) == null) {
				throw new GridException("dequeueCompletedLocalOr"+
										"ConsumptionTask(): "+
										"corrupted Local Jobs heap");
			}
			// TODO [WORKFLOW] throw Exception,
			// if all Tasks of the Job are not completed
			this.completed_local_jobs.incrementAndGet();
			rt = now-job.submissionTime();
			this.completed_local_jobs_total_runtime += rt;
			this.completed_local_jobs_total_runtime_measurement.
				addObservation(rt);
			
			this.local_q_history.addJob(job.submissionTime(),now);
			
			this.most_recentely_completed_local_job_time = now;
			
			// TODO [WORKFLOW] replace retval => return BOT_COMPLETED
			return LOCAL_JOB_COMPLETED;
		}

		// BoT is not completed yet
		else {
			// Remove BoT from running_local_jobs_queue
			if (bot.hasScheduledTasks() == false) {
				if (this.running_local_jobs_queue.remove(t.jobIDAsKey()) ==
					false) {
				throw new GridException("dequeueCompletedLocalOr"+
										"ConsumptionTask(): "+
										"corrupted running Local Jobs queue");
				}
			}

			// Notify UserAgent
			this.notifyCompletedTaskToUser(t);

			return LOCAL_TASK_COMPLETED;
		}

	}
	 
	/** Dequeues a Supplying Task after its execution has been completed
	 * @param t target Task
	 * @param xstart_time Task execution start time	 
	 * @param runtime Task run time
	 * @return completion level status
	 * @exception GridException if Task or queue is corrupted
	 */
	public synchronized RequestStatus
		dequeueCompletedSupplyingTask(Task t,long xstart_time,long runtime)
		throws GridException {
		
		if (t == null) {
			throw new GridException("dequeueCompletedSupplyingTask(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("dequeueCompletedSupplyingTask(): "+
									"illegal execution start time");
		}
		if (runtime < 0) {
			throw new GridException("dequeueCompletedSupplyingTask(): "+
									"illegal run time");
		}

		// Locate Job in Job heap
		Job job = getSupplyingJob(t);

		// Locate BoT
		BoT bot = job.root();
		if (bot.id() != t.getSupplyingSpaceBoTID()) {
			throw new GridException("dequeueCompletedSupplyingTask(): "+
									"BoT id conflict");
		}

		// Flag the Task as completed
		RequestContainer rc = bot.flagTaskAsCompleted(t);
		if (rc.hasSchedulingContext()) {
			/*
			rc.getSchedulingContext().
				setOutcome(t,xstart_time,SUPPLYING_TASK); // [ML]
			*/
		}

		// Update stats
		this.running_supplying_tasks--;
		long rt = LightweightBarteringGrid.getTime()-xstart_time;
		this.completed_tasks++;
		this.completed_tasks_total_runtime += rt;
		this.completed_tasks_total_nominal_runtime += t.nominalRuntime();
		this.completed_supplying_tasks++;
		this.completed_supplying_tasks_total_runtime += rt;

		if (!bot.isCompleted()) {
			throw new GridException("dequeueCompletedSupplyingTask(): "+
									"corrupted BoT holding Supplying Task");
		}
		
		// Remove BoT from running_supplying_jobs_queue
		if (this.running_supplying_jobs_queue.remove(job.idAsKey()) ==
			false) {
			throw new GridException("dequeueCompletedSupplyingTask(): "+
									"corrupted running Supplying Jobs Queue");
		}
			
		// Remove Job from Supplying Job heap
		if (this.supplying_jobs_heap.remove(job.idAsKey()) == null) {
			throw new GridException("dequeueCompletedSupplyingTask(): "+
									"corrupted Supplying Jobs heap");
		}
		// TODO [WORKFLOW] throw Exception,
		// if all Tasks of the Job are not completed,
		// i.e. if another alien Task has inserted itself into the Job ...

		return SUPPLYING_TASK_COMPLETED;

	}

	// ****************************************************************

	/** Requeues a Local Task
	 * that has been unqueued, but not executed,
	 * i.e. scheduled by the owner Peer.
	 * <p>Unqueued Tasks are in a bizarre state,
	 * as they have been removed from the waiting queue,
	 * but haven't yet been added to the running queue.
	 * It is therefore very important to put them back into the waiting queue.
	 * @param t target Task
	 * @exception GridException if Task or queue is corrupted
	 */
	public synchronized void requeueUnqueuedLocalTask(Task t) 
		throws GridException {
	
		requeueUnqueuedLocalOrConsumptionTask(t);
		
	}

	/** Requeues a Consumption Task
	 * that has been unqueued, but not executed,
	 * i.e. accepted by a Supplier Peer.
	 * <p>Unqueued Tasks are in a bizarre state,
	 * as they have been removed from the waiting queue,
	 * but haven't yet been added to the running queue.
	 * It is therefore very important to put them back into the waiting queue.
	 * @param t target Task
	 * @exception GridException if Task or queue is corrupted
	 */
	public synchronized void requeueUnqueuedConsumptionTask(Task t) 
		throws GridException {

		requeueUnqueuedLocalOrConsumptionTask(t);
		
	}

	private synchronized void requeueUnqueuedLocalOrConsumptionTask(Task t) 
		throws GridException {

		if (t == null) {
			throw new GridException("requeueUnqueuedLocalOrConsumptionTask(): "+
									"illegal Task");
		}
		
		// Locate Job in Job heap
		Job job = getLocalJob(t);
		
		// Locate BoT
		BoT bot = job.root();
		if (bot.id() != t.botID()) {
			throw new GridException("requeueUnqueuedLocalOrConsumptionTask(): "+
									"BoT id conflict");
		}

		// Flag the Task as unscheduled
		/*RequestContainer rc = */ bot.flagTaskAsUnscheduled(t,false);
		
		// Update stats
		this.waiting_local_tasks.incrementAndGet(); //++
		this.running_local_tasks--;
		
		// Add BoT to waiting_local_jobs_queue
		if (bot.countUnscheduledTasks() == 1) {
			if (this.waiting_local_jobs_queue.add(t.jobIDAsKey()) == false) {
				throw new GridException("requeueUnqueuedLocal"+
										"OrConsumptionTask(): "+
										"corrupted Local Jobs heap");
			}
		}
		// Remove Job from running_local_jobs_queue
		if (bot.hasScheduledTasks() == false) {
			if (this.running_local_jobs_queue.remove(t.jobIDAsKey()) == false) {
				throw new GridException("requeueUnqueuedLocal"+
										"OrConsumptionTask(): "+
										"corrupted Local Jobs heap");
			}
		}
		
	}
	
	/** Requeues a Local Task after its execution has been cancelled.
	 * @param t target Local Task
	 * @param xstart_time Task execution start time	 
	 * @param runtime Task run time
	 * @param supplier_id Supplier Peer id,
	 * in this case equals to the Task owner Peer id
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized
		void requeuePreemptedLocalTask(Task t,long xstart_time,long runtime,
									   String supplier_id)
		throws GridException {
		
		requeuePreemptedLocalOrConsumptionTask(t,xstart_time,runtime,
											   false,supplier_id);

		// This Task is obviously too-hard to schedule right, so cancel it
		if (t.getPreemptionCount() > ALLOWED_PREEMPTIONS_BEFORE_CANCELLATION) {
			this.dequeueWaitingLocalTask(t);
		}
		
	}
		
	/** Requeues a Consumption Task after its execution has been cancelled.
	 * @param t target Consumption Task
	 * @param xstart_time Task execution start time	 
	 * @param runtime Task run time
	 * @param supplier_id Supplier Peer id
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized
		void requeuePreemptedConsumptionTask(Task t,long xstart_time,
											 long runtime,String supplier_id)
		throws GridException {
		
		requeuePreemptedLocalOrConsumptionTask(t,xstart_time,runtime,
											   true,supplier_id);
		
	}

	/** Requeues a Local/Consumption Task after its execution
	 * has been cancelled.
	 * @param t target Task
	 * @param xstart_time Task execution start time
	 * @param runtime Task run time
	 * @param is_consumption_task true if Task is a Consumption Task
	 * @param supplier_id Supplier Peer id
	 * @exception GridException if a parameter is illegal
	 */
	private synchronized void
		requeuePreemptedLocalOrConsumptionTask(Task t,
											   long xstart_time,long runtime,
											   boolean is_consumption_task,
											   String supplier_id)
		throws GridException {

		if (t == null) {
			throw new GridException("requeuePreempted"+
									"LocalOrConsumptionTask(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("requeuePreempted"+
									"LocalOrConsumptionTask(): "+
									"illegal execution start time");
		}
		if (runtime < 0) {
			throw new GridException("requeuePreempted"+
									"LocalOrConsumptionTask(): "+
									"illegal run time");
		}
		if (supplier_id == null) {
			throw new GridException("requeuePreempted"+
									"LocalOrConsumptionTask(): "+
									"illegal Supplier ID");
		}
		if (is_consumption_task == false) {
			if (supplier_id.equals(this.managed_peer.id()) == false) {
			throw new GridException("requeuePreempted"+
									"LocalOrConsumptionTask(): "+
									"illegal access");
			}
		}
		
		// Locate Job in Job heap
		Job job = getLocalJob(t);
		
		// Locate BoT
		BoT bot = job.root();
		if (bot.id() != t.botID()) {
			throw new GridException("requeuePreempted"+
									"LocalOrConsumptionTask(): "+
									"BoT id conflict");
		}

		// Flag the Task as preempted
		t.flagAsPreempted();
		
		// Flag the Task as unscheduled, update Job Consumption history
		RequestContainer rc = bot.flagTaskAsUnscheduled(t,is_consumption_task);
		if (is_consumption_task) {
			job.consumptionOutcomeTracker().
				trackCancelledTask(supplier_id,runtime);

			if (rc.hasSchedulingContext()) {
				rc.getSchedulingContext().
					setOutcome(t,xstart_time,CONSUMPTION_TASK); // [ML]
			}
		}
		else {
			if (rc.hasSchedulingContext()) {
				/*
				rc.getSchedulingContext().
					setOutcome(t,xstart_time,LOCAL_TASK); // [ML]
				*/
			}
		}
		
		// Update stats
		this.waiting_local_tasks.incrementAndGet(); //++
		this.running_local_tasks--;
		
		// Add BoT to waiting_local_jobs_queue
		if (bot.countUnscheduledTasks() == 1) {
			if (this.waiting_local_jobs_queue.add(t.jobIDAsKey()) == false) {
				throw new GridException("requeuePreempted"+
										"LocalOrConsumptionTask(): "+
										"corrupted Local Jobs heap");
			}
		}
		// Remove Job from Local Job heap
		if (bot.hasScheduledTasks() == false) {
			if (this.running_local_jobs_queue.remove(t.jobIDAsKey()) == false) {
				throw new GridException("requeuePreempted"+
										"LocalOrConsumptionTask(): "+
										"corrupted Local Jobs heap");
			}
		}

	}
	
	/** Requeues a Supplying Task
	 * that has been unqueued, but not executed,
	 * i.e. scheduled by the owner Peer.
	 * <p>Unqueued Tasks are in a bizarre state,
	 * as they have been removed from the waiting queue,
	 * but haven't yet been added to the running queue.
	 * It is therefore very important to put them back into the waiting queue.
	 * @param t target Task
	 * @exception GridException if Task or queue is corrupted
	 */
	public synchronized void requeueUnqueuedSupplyingTask(Task t)
		throws GridException {

		if (t == null) {
			throw new GridException("requeueUnqueuedSupplyingTask(): "+
									"illegal Task");
		}
		
		// Locate Job in Job heap
		Job job = getSupplyingJob(t);
		
		// Locate BoT
		BoT bot = job.root();
		if (bot.id() != t.getSupplyingSpaceBoTID()) {
			throw new GridException("requeueUnqueuedSupplyingTask(): "+
									"BoT id conflict");
		}

		// Flag the Task as unscheduled
		/*RequestContainer rc = */ bot.flagTaskAsUnscheduled(t,false);
		
		// Update stats
		this.waiting_supplying_tasks.incrementAndGet(); //++
		this.running_supplying_tasks--;
		
		// Add BoT to waiting_supplying_jobs_queue
		if (bot.countUnscheduledTasks() == 1) {
			if (this.waiting_supplying_jobs_queue.add(job.idAsKey()) == false) {
				throw new GridException("requeueUnqueuedSupplyingTask()"+
										"corrupted Supplying Jobs heap");
			}
		}
		// Add Consumer Peer to waiting_consumer_peers
		this.waiting_consumer_peers.addConsumer(t.consumerID());
		this.running_consumer_peers.removeConsumer(t.consumerID());
		
		// Remove Job from Supplying Job heap
		if (bot.hasScheduledTasks() == false) {
			if (this.running_supplying_jobs_queue.remove(job.idAsKey()) ==
				false) {
				throw new GridException("requeueUnqueuedSupplyingTask()"+
										"corrupted Supplying Jobs heap");
			}
		}
		
	}

	/** Requeues a Supplying Task after its execution has been cancelled.
	 * @param t target Task
	 * @param xstart_time Task execution start time	 
	 * @exception GridException if Task or queue is corrupted
	 */
	public synchronized
		void requeuePreemptedSupplyingTask(Task t,long xstart_time)
		throws GridException {

		if (t == null) {
			throw new GridException("requeuePreemptedSupplyingTask(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("requeuePreemptedSupplyingTask(): "+
									"illegal execution start time "+ xstart_time);
		}

		// Locate Job in Job heap
		Job job = getSupplyingJob(t);
		
		// Locate BoT
		BoT bot = job.root();
		if (bot.id() != t.getSupplyingSpaceBoTID()) {
			throw new GridException("requeuePreemptedSupplyingTask(): "+
									"BoT id conflict");
		}

		// Flag the Task as preempted
		t.flagAsPreempted();

		// Flag the Task as unscheduled
		RequestContainer rc = bot.flagTaskAsUnscheduled(t,true);
		if (rc.hasSchedulingContext()) {
			/*
			rc.getSchedulingContext().
				setOutcome(t,xstart_time,SUPPLYING_TASK); // [ML]
			*/
		}
		
		// Update stats
		this.waiting_supplying_tasks.incrementAndGet(); //++
		this.running_supplying_tasks--;

		// Add BoT to waiting_supplying_jobs_queue
		if (bot.countUnscheduledTasks() == 1) { //Always true with 1 BoT per Job
			if (this.waiting_supplying_jobs_queue.add(job.idAsKey()) == false) {
				throw new GridException("requeuePreemptedSupplyingTask(): "+
										"corrupted Local Jobs heap");
			}
		}
		// Add Consumer Peer to waiting_consumer_peers
		this.waiting_consumer_peers.addConsumer(t.consumerID());
		this.running_consumer_peers.removeConsumer(t.consumerID());
		
		// Remove Job from Supplying Job heap
		if (bot.hasScheduledTasks() == false) {
			if (this.running_supplying_jobs_queue.remove(job.idAsKey()) ==
				false) {
				throw new GridException("requeuePreemptedSupplyingTask(): "+
										"corrupted Local Jobs heap");
			}
		}

		// This Task is obviously too-hard to schedule right, so cancel it
		if (t.getPreemptionCount() > ALLOWED_PREEMPTIONS_BEFORE_CANCELLATION) {
			this.dequeueWaitingSupplyingTask(true,t);
		}
		
	}
	
	/** Dequeues a Local Job.
	 *
	 * <p>Note: as User-generated Job cancellation is not currently supported,
	 * this method is currently not implemented
	 * and will throw a GridException.
	 *
	 * @param job_id target job_id
	 * @exception GridException if Task or queue is corrupted
	 */
	public synchronized void dequeueLocalJob(int job_id)
		throws GridException {

		if (job_id < 0) {
			throw new GridException("dequeueLocalJob(): illegal Job id");
		}
		
		// Locate Job in Job heap
		Job job = getLocalJob(job_id);
		if (job == null) {
			throw new GridException("dequeueLocalJob(): no such Job");
		}
		
		// Locate BoT
		BoT bot = job.root();

		// Cancel waiting Tasks
		Iterator<Task> unscheduled_tasks = bot.unscheduledTasks();
		while (unscheduled_tasks.hasNext()) {
			Task t = unscheduled_tasks.next();
			this.dequeueWaitingLocalTask(t);
		}

		// Cancel running Tasks
		Iterator<Task> scheduled_tasks = bot.scheduledTasks();
		while (scheduled_tasks.hasNext()) {
			Task t = scheduled_tasks.next();
			// Identify Task type
			boolean is_local = true;
			RequestType rt = this.managed_peer.xms().identifyRunningTaskType(t);
			switch (rt) {
			case LOCAL_TASK:
				is_local = true;
				break;
			case CONSUMPTION_TASK:
				is_local = false;
				break;
			case SUPPLYING_TASK:
			default:
				throw new GridException("dequeueLocalJob(): illegal Task type");
			}
			
			// Actually cancel the Task
			if (is_local) {
				this.managed_peer.xms().
					cancelLocalTaskOnLocalResource(t,DUMMY_LOG);
			}
			else /* if (is_local == false) */ {
				this.managed_peer.xms().
					cancelLocalTaskOnSupplyingResource(t);
				this.dequeueWaitingLocalTask(t);
			}
		}

	}

	/** Dequeues a Local Task after its execution has been cancelled.
	 * @param t target Local Task
	 * @param xstart_time Task execution start time
	 * @param runtime Task run time
	 * @param supplier_id Supplier Peer id,
	 * in this case equals to the Task owner Peer id
	 * @exception GridException if Task or queue is corrupted
	 */
	public synchronized void dequeueCancelledLocalTask(Task t,long xstart_time,
													   long runtime,
													   String supplier_id)
		throws GridException {

		// implementation follows the pattern of requeuePreemptedLocalTask()

		requeuePreemptedLocalOrConsumptionTask(t,xstart_time,runtime,
											   false,supplier_id);

		this.dequeueWaitingLocalTask(t);

	}

	/** Dequeues a waiting Local Task.
	 * @param t target Task
	 * @exception GridException if Task or queue is corrupted
	 */
	public synchronized void dequeueWaitingLocalTask(Task t)
		throws GridException {

		// implementation follows the pattern of
		// dequeueCompletedLocalOrConsumptionTask(): if last Task, dequeue Job

		if (t == null) {
			throw new GridException("dequeueWaitingLocalTask(): illegal Task");
		}

		// Locate Job in Job heap
		Job job = getLocalJob(t);

		// Locate BoT
		BoT bot = job.root();
		if (bot.id() != t.botID()) {
			throw new GridException("dequeueWaitingLocalTask(): "+
									"BoT id conflict");
		}
		
		// Flag the Task as "completed"
		bot.flagTaskAsCompleted(t);

		// Update stats
		this.waiting_local_tasks.decrementAndGet(); //--
		/* Task was not completed, so the following stats are meaningless:
		long now = LightweightBarteringGrid.getTime();
		long rt = now-xstart_time;
		this.completed_tasks++;
		this.completed_tasks_total_runtime += rt;
		this.completed_tasks_total_nominal_runtime += t.nominalRuntime();
		this.completed_local_tasks++;
		this.completed_local_tasks_total_runtime += rt;
		*/
		
		// If BoT is completed, remove it from local Task queue
		// If Job is completed, remove it from local Job heap
		//                      and notify the User Agent
		if (bot.isCompleted()) {
			// Notify UserAgent
			this.notifyCompletedJobToUser(t);

			// Remove BoT from running_local_jobs_queue
			if (this.waiting_local_jobs_queue.remove(t.jobIDAsKey()) == false) {
				throw new GridException("dequeueWaitingLocalTask(): "+
										"corrupted waiting Local Jobs queue");
			}

			// Remove Job from Local Job heap
			if (this.local_jobs_heap.remove(t.jobIDAsKey()) == null) {
				throw new GridException("dequeueWaitingLocalTask(): "+
										"corrupted Local Jobs heap");
			}

			// TODO [WORKFLOW] throw Exception,
			// if all Tasks of the Job are not completed
			this.completed_local_jobs.incrementAndGet();
			long now = LightweightBarteringGrid.getTime();
			long rt = now-job.submissionTime();
			this.completed_local_jobs_total_runtime += rt;
			this.completed_local_jobs_total_runtime_measurement.
				addObservation(rt);
			
			this.local_q_history.addJob(job.submissionTime(),now);
			
			this.most_recentely_completed_local_job_time = now;
			
			// TODO [WORKFLOW] replace retval => return BOT_COMPLETED
			//return LOCAL_JOB_COMPLETED;
		}

		// BoT is not completed yet
		else {
			// Notify UserAgent
			this.notifyCompletedTaskToUser(t);

			//return LOCAL_TASK_COMPLETED;
		}

		
	}

	/** Dequeues a Supplying Task after its execution has been cancelled.
	 * @param t target Task
	 * @param xstart_time Task execution start time	 
	 * @exception GridException if Task or queue is corrupted
	 */
	public synchronized
		void dequeueCancelledSupplyingTask(Task t,long xstart_time)
		throws GridException {

		if (t == null) {
			throw new GridException("dequeueCancelledSupplyingTask(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("dequeueCancelledSupplyingTask(): "+
									"illegal execution start time");
		}

		// Locate Job in Job heap
		Job job = getSupplyingJob(t);

		// Locate BoT
		BoT bot = job.root();
		if (bot.id() != t.getSupplyingSpaceBoTID()) {
			throw new GridException("dequeueCancelledSupplyingTask(): "+
									"BoT id conflict");
		}

		// Flag the Task as cancelled (see debug-cancel-waiting-supplying-tasks)
		t.flagAsCancelled(this.managed_peer.id());

		// Flag the Task as unscheduled
		RequestContainer rc = bot.flagTaskAsUnscheduled(t,true);
		if (rc.hasSchedulingContext()) {
			/*
			rc.getSchedulingContext().
				setOutcome(t,xstart_time,SUPPLYING_TASK); // [ML]
			*/
		}

		// Update stats
		this.running_supplying_tasks--;

		// Remove BoT from waiting_supplying_jobs_queue
		if (this.running_supplying_jobs_queue.remove(job.idAsKey()) ==
			false) {
			throw new GridException("dequeueCancelledSupplyingTask(): "+
									"corrupted running Supplying Jobs Queue");
		}
		/* As it was not waiting, *DO NOT* remove Consumer Peer
		   from waiting_consumer_peers,
		   else it would be ejected from QueueManager
		// Remove Consumer Peer from waiting_consumer_peers
		this.waiting_consumer_peers.removeConsumer(t.consumerID());*/
		
		// Remove Job from Supplying Job heap
		if (this.supplying_jobs_heap.remove(job.idAsKey()) == null) {
			throw new GridException("dequeueCancelledSupplyingTask(): "+
									"corrupted Supplying Jobs heap");
		}

	}

	// ****************************************************************

	/** Dequeues target waiting Supplying Task.
	 * @param notify_consumer communicates cancellation to consumer
	 * @param t target Task
	 * @exception GridException if Task or queue is corrupted
	 */
	public synchronized void
		dequeueWaitingSupplyingTask(boolean notify_consumer,Task t)
		throws GridException {

		if (t == null) {
			throw new GridException("dequeueWaitingSupplyingTask(): "+
									"illegal Task");
		}

		// Locate Job in Job heap
		Job job = getSupplyingJob(t);

		// Actually dequeue the waiting Supplying Task
		dequeueTargetWaitingSupplyingTask(notify_consumer,t,job);

	}
	
	/** Dequeues the oldest waiting Supplying Task of target consumer.
	 * @param notify_consumer communicates cancellation to consumer
	 * @param consumer_id target consumer
	 * @exception GridException if Task or queue is corrupted
	 */
	public synchronized void
		dequeueOldestWaitingSupplyingTaskOfConsumer(boolean notify_consumer,
													String consumer_id)
		throws GridException {

		if (consumer_id == null) {
			throw new GridException("dequeueWaitingSupplyingTask(): "+
									"illegal consumer");
		}
		
		if (this.waiting_supplying_jobs_queue.size() < 1) {
			throw new GridException("dequeueWaitingSupplyingTask(): "+
									"no waiting consumer");
		}
		
		Iterator<Integer> it = this.waiting_supplying_jobs_queue.iterator();
		Job job = null;
		while (it.hasNext()) {
			int job_id = it.next();
			Job job_dummy = getSupplyingJob(job_id);
			if (job_dummy.ownerID().equals(consumer_id)) {
				job = job_dummy;
				break;
			}
		}
		if (job == null) {
			throw new GridException("dequeueWaitingSupplyingTask(): "+
									"target consumer currently not waiting");
		}
		BoT bot = job.root();
		
		if (bot.hasUnscheduledTasks() == false) {
			throw new GridException("dequeueOldestWaitingSupplyingTaskOf"+
									"Consumer(): corrupted data");
		}
		Task t = bot.peekFirstSupplyingTask();
		
		dequeueTargetWaitingSupplyingTask(notify_consumer,t,job);
		
	}
	
	/** Dequeues all waiting Supplying Tasks.
	 * @param notify_consumers communicates cancellations to consumers
	 * @exception GridException if queue is corrupted
	 */
	public synchronized
		void dequeueAllWaitingSupplyingTasks(boolean notify_consumers)
		throws GridException {
		
		final int nb_supplying_tasks = this.waiting_supplying_jobs_queue.size();
		if (nb_supplying_tasks > 0) {
			dequeueMostRecentWaitingSupplyingTasks(notify_consumers,
											   	   nb_supplying_tasks);
		}
		
	}

	/** Dequeues the k most recently queued waiting Supplying Tasks.
	 * @param notify_consumers communicates cancellations to consumers
	 * @param k number of Supplying Tasks to dequeue
	 * @exception GridException if queue is corrupted
	 */
	public synchronized
		void dequeueMostRecentWaitingSupplyingTasks(boolean notify_consumers,
													int k)
		throws GridException {

		final int nb_supplying_tasks = this.waiting_supplying_jobs_queue.size();

		if ((k < 1) || (k > nb_supplying_tasks)) {
			throw new GridException("dequeueMostRecent"+
									"WaitingSupplyingTasks(): illegal "+
									"number of Supplying Tasks ("+k+")");
		}
		
		Iterator<Integer> it = this.waiting_supplying_jobs_queue.iterator();
		ArrayList<Integer> al =
			new ArrayList<Integer>(this.waiting_supplying_jobs_queue.size());
		final int lim = nb_supplying_tasks-k;
		for (int i = 0 ; i < lim ; ++i) {
			assert(it.hasNext());
			it.next();
		}
		while (it.hasNext()) {
			al.add(it.next());
		}
		
		dequeueTargetWaitingSupplyingTasks(notify_consumers,al);
		
	}
	
	/** Dequeues the k worst NoF-ranked  queued waiting Supplying Tasks.
	 * @param notify_consumers communicates cancellations to consumers
	 * @param k number of Supplying Tasks to dequeue
	 * @exception GridException if queue is corrupted
	 */
	public synchronized
		void dequeueWorstRankedWaitingSupplyingTasks(boolean notify_consumers,
													 int k)
		throws GridException {
		
		final int nb_supplying_tasks = this.waiting_supplying_jobs_queue.size();

		if ((k < 1) || (k > nb_supplying_tasks)) {
			throw new GridException("dequeueWorstRanked"+
									"WaitingSupplyingTasks(): illegal "+
									"number of Supplying Tasks ("+k+")");
		}
		
		PeerProfile[] waiting_consumers =
			this.waiting_consumer_peers.getConsumers();
		// Rank waiting Consumer Peers
		FavorsSupplyingRanking r =
			new FavorsSupplyingRanking(waiting_consumers);
		PeerProfile[] ranked_peers = r.getRankedPeers();
		int dequeued_tasks = 0;
		final ArrayList<Integer> al = new ArrayList<Integer>(1);
		final ArrayList<Job> supplying_job_list =
			new ArrayList<Job>(this.supplying_jobs_heap.size());
		// Enumerate all Consumers by decreasing NoF balance
		for (int p = ranked_peers.length-1 ; p >= 0 ; --p) {
			// Seek all Supplying Tasks of target consumer
			// -- yes, an auxiliary data structure would accelerate the lookup
			//    to be done when refacotring the QueueManager
			PeerProfile peer_profile = ranked_peers[p];
			String consumer_id = peer_profile.id();
			Iterator<Job> it = this.supplying_jobs_heap.values().iterator();
			supplying_job_list.clear();
			while (it.hasNext()) {
				supplying_job_list.add(it.next());
			}
			
			for (int j = 0 ; j < supplying_job_list.size() ; ++j) {
				Job job = supplying_job_list.get(j);
				if (job.ownerID().equals(consumer_id)) {
					al.clear();
					al.add(job.id());
					int tmp_dequeued_tasks =
						dequeueTargetWaitingSupplyingTasks(notify_consumers,al);
					assert(tmp_dequeued_tasks < 2);
					if (tmp_dequeued_tasks > 0) {
						dequeued_tasks += tmp_dequeued_tasks;
						if (dequeued_tasks >= k) {
							return;
						}
					}
				}
			}
			
		}
		
	}
	
	/** Dequeues target waiting Supplying Tasks.
	 * @param notify_consumers communicates cancellations to consumers
	 * @param al list of target Supplying Tasks to dequeue
	 * @return number of Supplying Tasks actually dequeued
	 * @exception GridException if queue is corrupted
	 */
	private synchronized
		int dequeueTargetWaitingSupplyingTasks(boolean notify_consumers,
											   ArrayList<Integer> al)
		throws GridException {
	
		if ((al == null) || (al.size() <= 0)) {
			throw new GridException("dequeueTargetWaitingSupplyingTasks(): "+
									"illegal target Supplying Tasks list");
		}
		
		int dequeued_tasks = 0;
		for (Integer job_id : al) {
			
			Job job = getSupplyingJob(job_id);
			BoT bot = job.root();
			if (bot.hasUnscheduledTasks() == false) {
				continue; // skip
			}
			
			Task t = job.root().peekFirstSupplyingTask();
			
			// Actually dequeue the waiting Supplying Task
			dequeueTargetWaitingSupplyingTask(notify_consumers,t,job);
			
			++dequeued_tasks;
		}
		
		return dequeued_tasks;
		
	}
	
	/** Dequeues target waiting Supplying Task.
	 * @param notify_consumer communicates cancellation to consumer
	 * @param t target waiting Supplying Task
	 * @param job Job of target waiting Supplying Task
	 * @exception GridException if queue is corrupted
	 */
	private synchronized void
		dequeueTargetWaitingSupplyingTask(boolean notify_consumer,
										  Task t,Job job)
		throws GridException {
	
		if (t == null) {
			throw new GridException("dequeueTargetWaitingSupplyingTask(): "+
									"illegal target Supplying Task");
		}
		if (job == null) {
			throw new GridException("dequeueTargetWaitingSupplyingTask(): "+
									"illegal Job of target Supplying Task");
		}

		// Flag the Task as cancelled
		t.flagAsCancelled(this.managed_peer.id());
		
		/*if (rc.hasSchedulingContext()) {
			/ *
			rc.getSchedulingContext().
				setOutcome(t,xstart_time,SUPPLYING_TASK); // [ML]
			* /
		}*/

		// Update stats
		this.waiting_supplying_tasks.decrementAndGet(); //--
	
		// Remove BoT from waiting_supplying_jobs_queue
		if (this.waiting_supplying_jobs_queue.remove(job.idAsKey()) == false) {
			throw new GridException("dequeueTargetWaitingSupplyingTask(): "+
									"corrupted waiting Supplying Jobs Queue");
		}

		// Remove Consumer Peer from waiting_consumer_peers
		this.waiting_consumer_peers.removeConsumer(t.consumerID());

		// Remove Job from Supplying Job heap
		if (this.supplying_jobs_heap.remove(job.idAsKey()) == null) {
			throw new GridException("dequeueTargetWaitingSupplyingTask(): "+
									"corrupted Supplying Jobs heap");
		}
		
		// Notify consumer of the cancellation
		if (notify_consumer) {
			lbg.peer.service.ExternalPeerHandle peer_handle =
				this.managed_peer.register().getHandle(t.consumerID());
			String supplier_id = this.managed_peer.id();
			peer_handle.
				cancelLocalTaskOnSupplyingResource(supplier_id,t);
		}
	
	}
	
	// ****************************************************************

	/** Indicates whether waiting Supplying Tasks should be dequeued,
	 * i.e. cancelled, that is if the number of waiting Local Tasks exceeds
	 * LOCAL_Q_LEN_PREEMPTION_THRESHOLD
	 * @return true, if waiting Supplying Tasks should be dequeued,
	 * i.e. cancelled
	 */
	public synchronized boolean shouldPreemptWaitingSupplyingTasks() {

		int outstanding_local_tasks = // ... without considering consumption
			this.waiting_local_tasks.get()
			-this.managed_peer.rms().countIdleResources();
		
		if (outstanding_local_tasks < 0) { // enough Resources... don't dequeue
			return false;
		}
		
		return (outstanding_local_tasks >=
				this.managed_peer.strategyConfig().
				LOCAL_Q_LEN_PREEMPTION_THRESHOLD());
		
	}
	
	public synchronized void preemptWaitingSupplyingTasks()
		throws GridException {
		
		if (this.shouldPreemptWaitingSupplyingTasks()) {
			if (this.currently_preempting_supplying_tasks_flag == false) {
				this.currently_preempting_supplying_tasks_flag = true;
				this.scheduler.preemptWaitingSupplyingTasks();
				this.currently_preempting_supplying_tasks_flag = false;
				/* flag, semaphore: see also
				   http://java.sun.com/j2se/1.5.0/docs/api/java/util/concurrent/Semaphore.html
				*/
			}
		}
		
	}
	
	/** Profiles of all consumer Peers with running Supplying Tasks.
	 * @return profiles of all consumer Peers with running Supplying Tasks.
	 */
	public PeerProfile[] getRunningConsumers() {
		
		return this.running_consumer_peers.getConsumers();
		
	}

	/** Profiles of all consumer Peers with waiting Supplying Tasks.
	 * @return profiles of all consumer Peers with waiting Supplying Tasks.
	 */
	public PeerProfile[] getWaitingConsumers() {
		
		return this.waiting_consumer_peers.getConsumers();
		
	}

	/** Profiles of all consumer Peers with queued Supplying Tasks.
	 * @return profiles of all consumer Peers with queued Supplying Tasks.
	 */
	public PeerProfile[] getAllConsumers() {
		
		PeerProfile[] r_c = this.running_consumer_peers.getConsumers();
		PeerProfile[] w_c = this.waiting_consumer_peers.getConsumers();
		
		if (r_c.length == 0) {
			return w_c;
		}
		else if (w_c.length == 0) {
			return r_c;
		}
		else {
			// Merge the running and waiting arrays
			Set<PeerProfile> a_c = new TreeSet<PeerProfile>();
			for (int i = 0 ; i < r_c.length ; ++i) {
				a_c.add(r_c[i]);
			}
			for (int i = 0 ; i < w_c.length ; ++i) {
				a_c.add(w_c[i]);
			}
			
			PeerProfile[] peer_profiles = new PeerProfile[a_c.size()];
			return a_c.toArray(peer_profiles);
		}
		
	}
	
	// ****************************************************************

	public synchronized void updateStatistics() {
		
		this.total_waiting_local_tasks += this.waiting_local_tasks.get();
		this.total_waiting_local_tasks_count++;
		this.total_waiting_supplying_tasks +=
			this.waiting_supplying_tasks.get();
		this.total_waiting_supplying_tasks_count++;
		this.total_local_jobs += this.local_jobs_heap.size();
		this.total_local_jobs_count++;
		
		this.total_running_local_tasks += this.running_local_tasks;
		this.total_running_local_tasks_count++;
		this.total_running_supplying_tasks += this.running_supplying_tasks;
		this.total_running_supplying_tasks_count++;
		this.total_supplying_jobs += this.supplying_jobs_heap.size();
		this.total_supplying_jobs_count++;

	}

	// ****************************************************************

	private synchronized void notifyCompletedTaskToUser(Task t)
		throws GridException {

		if (t == null) {
			throw new GridException("notifyCompletedTaskToUser(): "+
									"illegal Task");
		}
		
		Job job = getLocalJob(t);
		try {
			job.user().notifyCompletedTask(t);
		}
		catch (GridNetworkException e) {
			e.printStackTrace();
			// The User is unreachable ... results return to star dust ...
			// ... as long as there is no result persistence mechanism on Peer
		}

	}
	
	private synchronized void notifyCompletedJobToUser(Task t)
		throws GridException {
		
		if (t == null) {
			throw new GridException("notifyCompletedJobToUser(): "+
									"illegal Task");
		}

		Job job = getLocalJob(t);
		try {
			job.user().notifyCompletedJob(t);
		}
		catch (GridNetworkException e) {
			// The User is unreachable ... results return to star dust ...
			// ... as long as there is no result persistence mechanism on Peer
		}

	}

	// ****************************************************************

	private synchronized Job getLocalJob(Task t) throws GridException {

		if (t == null) {
			throw new GridException("getLocalJob(): illegal Task");
		}
		
		Integer job_id_as_key = t.jobIDAsKey();
		Job job = this.local_jobs_heap.get(job_id_as_key);
		if (job == null) {
			throw new GridException("getLocalJob(): "+
									"corrupted Local Jobs heap, "+
									"cannot find Job "+job_id_as_key);
		}

		// Check that Supplying Task belongs to the Peer that wants to cancel it
		if (job.ownerID().equals(t.consumerID()) == false) {
			throw new GridException("getLocalJob(): "+
									"illegal access "+
									"(incoming Task consumer id="+
									t.consumerID()+
									", Job owner id="+job.ownerID()+
									", Job id="+t.jobID()+
									", BoT id="+t.botID()+
									", managed Peer id="+
									this.managed_peer.id()+")");
		}
		
		return job;
		
	}

	private synchronized Job getLocalJob(Integer job_id_as_key)
		throws GridException {

		if (job_id_as_key == null) {
			throw new GridException("getLocalJob(): illegal Job key");
		}
		
		Job job = this.local_jobs_heap.get(job_id_as_key);
		if (job == null) {
			throw new GridException("getLocalJob(): "+
									"corrupted Local Jobs heap, "+
									"cannot find Job "+job_id_as_key);
		}

		return job;
		
	}

	private synchronized Job getSupplyingJob(Task t) throws GridException {

		if (t == null) {
			throw new GridException("getSupplyingJob(): illegal Task");
		}
		
		Integer tmp_job_id_as_key = Tank.newInteger(t.getSupplyingSpaceJobID());
		Job job = this.supplying_jobs_heap.get(tmp_job_id_as_key);

		if (job == null) {
			throw new GridException("getSupplyingJob(): "+
									"corrupted Supplying Jobs heap, "+
									"cannot find Job "+
									t.getSupplyingSpaceBoTID());
		}
		// Check that Supplying Task belongs to the Peer that wants to cancel it
		if (job.ownerID().equals(t.consumerID()) == false) {
			throw new GridException("getSupplyingJob(): "+
									"illegal access "+
									"(incoming Task consumer id="+
									t.consumerID()+
									", Job owner id="+job.ownerID()+
									", Supplying space Job id="+
									t.getSupplyingSpaceJobID()+
									", Supplying space BoT id="+
									t.getSupplyingSpaceBoTID()+
									", managed Peer id="+
									this.managed_peer.id()+")");
		}
		
		return job;
		
	}

	private synchronized Job getSupplyingJob(Integer job_id_as_key)
		throws GridException {

		if (job_id_as_key == null) {
			throw new GridException("getSupplyingJob(): illegal Job key");
		}
		
		Job job = this.supplying_jobs_heap.get(job_id_as_key);
		if (job == null) {
			throw new GridException("getSupplyingJob(): "+
									"corrupted Local Jobs heap, "+
									"cannot find Job "+job_id_as_key);
		}

		return job;
		
	}

	// ****************************************************************

	private synchronized void debugging_msg_1(Job job) throws GridException {
		
		lbg.protocol.data.MetaGridData[] dummy =
			job.root().getUniqueMetaGridDatas();
		System.out.println("BoT "+job.id());
		for (int i=0 ; i<dummy.length ; ++i) System.out.print(dummy[i].id+" ");
		System.out.println();
		
	}

	// ****************************************************************

	/** Number of allowed preemptions (expectedly a high value, e.g. 256)
	 * before a Task is automatically cancelled.
	 * <p>This Task control mechanism prevents Tasks-that-can-never-complete
	 * (i.e.~because they are bugged) to become zombies.
	 */
	public static final int ALLOWED_PREEMPTIONS_BEFORE_CANCELLATION = 256;

	// ****************************************************************

	private final Peer        managed_peer;
	private final JobPreQueue prequeue;
	private final Scheduler   scheduler;
	
	// Local Jobs data

	private final TreeMap<Integer,Job> local_jobs_heap;
	private final PriorityQueue<Integer> waiting_local_jobs_queue;
	private final PriorityQueue<Integer> running_local_jobs_queue;
	private final AtomicInteger waiting_local_tasks;
	private int running_local_tasks;

	private long total_waiting_local_tasks;
	private long total_waiting_local_tasks_count;
	private long total_running_local_tasks;
	private long total_running_local_tasks_count;
	private long total_local_jobs;
	private long total_local_jobs_count;
	
	private long most_recentely_completed_local_job_time;

	// Supplying Jobs data

	private final TreeMap<Integer,Job> supplying_jobs_heap;
	private final PriorityQueue<Integer> waiting_supplying_jobs_queue;
	private final PriorityQueue<Integer> running_supplying_jobs_queue;
	private final AtomicInteger waiting_supplying_tasks;
	private int running_supplying_tasks;

	private final ConsumerQueue waiting_consumer_peers;
	private final ConsumerQueue running_consumer_peers;
	
	private long total_waiting_supplying_tasks;
	private long total_waiting_supplying_tasks_count;
	private long total_running_supplying_tasks;
	private long total_running_supplying_tasks_count;
	private long total_supplying_jobs;
	private long total_supplying_jobs_count;

	private boolean currently_preempting_supplying_tasks_flag;
	
	// Stats data

	private long completed_tasks;
	private long completed_tasks_total_runtime;
	private long completed_tasks_total_nominal_runtime;

	private final LocalQueueHistory local_q_history;

	private long completed_local_tasks;
	private long completed_local_tasks_total_runtime;
	
	private long completed_supplying_tasks;
	private long completed_supplying_tasks_total_runtime;
	
	private final AtomicLong completed_local_jobs;
	private long completed_local_jobs_total_runtime;
	private final Measurement completed_local_jobs_total_runtime_measurement;
	
	// ****************************************************************

	/** Helper class of QueueManager
	 */
	private static class ConsumerQueue {

		public ConsumerQueue(Peer managed_peer,boolean waiting)
			throws GridException {
			
			if (managed_peer == null) {
				throw new GridException("ConsumerQueue(): illegal Peer");
			}
			this.managed_peer = managed_peer;
			
			this.q = new TreeMap<String,PeerProfile>();
			this.waiting = waiting;
			
		}
		
		public synchronized int size() {
			
			return this.q.size();
			
		}
		
		public synchronized void addConsumer(String consumer_id)
			throws GridException {
			
			if (consumer_id == null) {
				throw new GridException("addConsumer(): illegal Consumer id");
			}

			PeerProfile peer_profile = this.q.get(consumer_id);
			
			if (peer_profile != null) {
				if (this.waiting) {
					peer_profile.increaseWaitingJobsCount();
				}
				else {
					peer_profile.increaseRunningJobsCount();
				}
			}
			else {
				peer_profile =
					this.managed_peer.register().getProfile(consumer_id);
				if (this.waiting) {
					if (peer_profile.increaseWaitingJobsCount() != 1) {
						throw new GridException("addConsumer(): "+
												"corrupted waiting count");
					}
				}
				else {
					if (peer_profile.increaseRunningJobsCount() != 1) {
						throw new GridException("addConsumer(): "+
												"corrupted running count");
					}
				}
				this.q.put(consumer_id,peer_profile);
			}
			
		}
		
		public synchronized void removeConsumer(String consumer_id)
			throws GridException {
			
			if (consumer_id == null) {
				throw new GridException("removeConsumer(): "+
										"illegal Consumer id");
			}

			PeerProfile peer_profile = this.q.get(consumer_id);
			
			if (peer_profile == null) {
				throw new GridException("removeConsumer(): corrupted data");
			}

			if (waiting) {
				if (peer_profile.decreaseWaitingJobsCount() == 0) {
					this.q.remove(consumer_id);
				}
			}
			else {
				if (peer_profile.decreaseRunningJobsCount() == 0) {
					this.q.remove(consumer_id);
				}
			}
			
		}
		
		public synchronized PeerProfile[] getConsumers() {
		
			PeerProfile[] peer_profiles = new PeerProfile[this.q.size()];
			
			return this.q.values().toArray(peer_profiles);
			
		}
		
		private final Peer managed_peer;
		private final TreeMap<String,PeerProfile> q;
		private final boolean waiting;
		
	}
	
	// ****************************************************************

}

