/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// JobPreQueue.java

package lbg.peer.core.qms.prequeue;

import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.core.scheduler.*;
import lbg.peer.workflow.*;
import lbg.task.*;
import lbg.user.service.*;
	  
/** Pre-queue of incoming requests, that will be transferred to
 * the Queueing Management System as soon as the Peer global lock is released.
 * <p>Each Local Job is currently composed of 1 Bag of Tasks.
 * <p>Each Supplying Job is currently composed of 1 Bag of 1 Task.
 * @author Cyril Briquet
 */
public abstract class JobPreQueue {

	JobPreQueue(Peer managed_peer,Scheduler scheduler,boolean enable_ttg)
		throws GridException {

		if (managed_peer == null) {
			throw new GridException("JobPreQueue(): illegal Peer");
		}
		if (scheduler == null) {
			throw new GridException("JobPreQueue(): illegal Scheduler");
		}
		
		this.managed_peer = managed_peer;
		this.scheduler    = scheduler;
		this.jf = new JobFactory(this.managed_peer.id(),enable_ttg);
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[JobPreQueue ");
		sb.append(this.managed_peer.id());
		sb.append("]");

		sb.append("\n  ->  ");
		sb.append("queue threshold = ").
			append(this.managed_peer.strategyConfig().
				   PEER_Q_FILTERING_THRESHOLD());

		return sb.toString();

	}

	// ****************************************************************

	/**
	 * @return JobPreQueue size
	 * @exception GridException implementations may throw a GridException
	 * if operation is not supported
	 */
	public abstract int size() throws GridException;
	
	public int transferLocalJobToQMS(UserHandle uh,JobDescriptor jd)
		throws GridException {

		if (uh == null) {
			throw new GridException("transferLocalJobToQMS(): illegal User");
		}
		if (jd == null) {
			throw new GridException("transferLocalJobToQMS(): "+
									"illegal Job Descriptor");
		}

		Job job = this.jf.createLocalJob(uh,jd);
		
		/* done in subclass
		this.managed_peer.qms().queueLocalJob(job);*/
		this.transferJob(new RequestHolder(job,null));
		
		return job.id();

	}
	
	/**
	 * @return a Consumption Task id if the submitted Supplying Task
	 * has been accepted, or null if it has not been accepted
	 */
	public ConsumptionTaskID transferSupplyingTaskToQMS(Task t)
		throws GridException {
		
		if (t == null) {
			throw new GridException("transferSupplyingTaskToQMS(): "+
									"illegal Task");
		}

		// Check if Supplying Task should be filtered out
		if (this.scheduler.filterOutSupplyingTask(t)) {
			return new ConsumptionTaskID(); // no-Task
		}

		// Create new Supplying Job containing 1 BoT of 1 (Supplying) Task
		Job job = this.jf.createSupplyingJob(t);
		ConsumptionTaskID accepted =
			new ConsumptionTaskID(t.getSupplyingSpaceJobID(),
						  		  t.getSupplyingSpaceBoTID());
		
		/* done in subclass
		this.managed_peer.qms().queueSupplyingJob(job,t.consumerID());*/
		this.transferJob(new RequestHolder(job,t.consumerID()));
		
		return accepted;

	}
	
	// ****************************************************************

	protected abstract void transferJob(RequestHolder req) throws GridException;
	
	protected synchronized void actuallyTransferJob(RequestHolder req)
		throws GridException {
		
		if (req == null) {
			throw new GridException("actuallyTransferJob(): illegal request");
		}

		synchronized(this.managed_peer.rms()) {
			if (req.consumer_id == null) {
				this.managed_peer.qms().queueLocalJob(req.job);
				this.managed_peer.schedulerController().
					activateLocalScheduling();
			}
			else {
				this.managed_peer.qms().
					queueSupplyingJob(req.job,req.consumer_id);
				this.managed_peer.schedulerController().
					activateSupplyingScheduling();
			}
		}
		
	}
	
	// ****************************************************************

	private final Peer managed_peer;
	private final Scheduler scheduler;
	private final JobFactory jf;

	// ****************************************************************

	protected static class RequestHolder {
	
		public RequestHolder(Job job,String consumer_id) {
			this.job         = job;
			this.consumer_id = consumer_id;
		}
		
		public final Job job;
		public final String consumer_id;
		
	}
	
}

