/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LocalQueueHistory.java

package lbg.peer.core.qms.queue;

import lbg.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonFileHandling.*;
import lbg.peer.core.qms.*;
import lbg.task.*;

import java.util.ArrayList;

/** A Local queue history tracks the submission/completion times
 * local Jobs.
 * <p>The entries of the tracked times are sorted on the Job completion times,
 * not on the Job submission times.</p>
 * <p>The LocalQueueHistory can be stored to disk by a shutdown hook
 * added to the VM where the managed Peer is running.</p>
 * @author Cyril Briquet
 */
public class LocalQueueHistory {

	/** Constructs a new LocalQueueHistory.
	 * @param peer_id managed Peer id
	 * @param store_data if true, the LocalQueueHistory is stored to disk
	 * by a shutdown hook added to the VM where the managed Peer is running
	 */
	public LocalQueueHistory(String peer_id,boolean store_data)
		throws GridException {
		
		if (peer_id == null) {
			throw new GridException("LocalQueueHistory(): illegal Peer id");
		}
		this.peer_id = peer_id;
		this.store_data = store_data;
		
		if (this.store_data) {
			this.submission_t_list = new ArrayList<Long>();
			this.completion_t_list = new ArrayList<Long>();
			
			Runnable runnable = new Runnable() {
				public void run() {
					storeJobs();
				}
			};
			Runtime.getRuntime().addShutdownHook(new Thread(runnable));
			
		}
		else {
			this.submission_t_list = null;
			this.completion_t_list = null;
		}

	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LocalQueueHistory]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	/**
	 * @return true if data was indeed stored,
	 * or false if the Local queue history was configured not to store data
	 */
	public synchronized boolean addJob(long submission_t,long completion_t)
		throws GridException {
		
		if (submission_t < 0) {
			throw new GridException("addJob(): illegal Job submission time");
		}
		if (completion_t < 0) {
			throw new GridException("addJob(): illegal Job completion time");
		}
		if ((submission_t > completion_t) ||
			(completion_t > LightweightBarteringGrid.getTime())) {
			throw new GridException("addJob(): illegal Job times");
		}
		
		boolean retval = false;
		if (this.store_data) {
			assert(areListsConsistent());

			this.submission_t_list.add(submission_t);
			this.completion_t_list.add(completion_t);
			
			retval = true;
			assert(areListsConsistent());
		}

		return retval;

	}
	
	/**
	 * @return true if data was indeed stored,
	 * or false if the Local queue history was configured not to store data
	 */
	public synchronized boolean addJobs(String jobs_list) throws GridException {
		
		if (jobs_list == null) {
			throw new GridException("addJobs(): illegal Jobs list");
		}
		
		boolean retval = false;
		if (this.store_data) {
			assert(areListsConsistent());
			
			if (1 < 2) throw new GridException("yet to be implemented");
			
			retval = true;
			assert(areListsConsistent());
		}
		
		return retval;
		
	}
	
	/**
	 * @return true if data was indeed stored,
	 * or false if the Local queue history was configured not to store/save data
	 */
	public synchronized boolean storeJobs() {
		
		boolean retval = false;
		if (this.store_data) {
			assert(areListsConsistent());

			StringBuilder sb_fn = new StringBuilder();
			sb_fn.append("__").append(this.peer_id).
				append("_local_jobs_history").append(LOG_EXT);			
			
			StringBuilder sb_data = new StringBuilder();
			for (int l = 0 ; l < this.submission_t_list.size() ; ++l) {
				sb_data.append(this.submission_t_list.get(l)).append(" ").
					append(this.completion_t_list.get(l)).append("\n");
			}
			
			try {
				CommonFileHandling.
					writeFile(sb_fn.toString(),sb_data.toString());
			}
			catch (GridException e) {
				System.out.println("storeJobs(): could not store "+
								   "Job submission times to "+sb_fn.toString()+
								   "\n"+sb_data.toString()+"\n");
			}
			
			retval = true;
		}
		
		return retval;
		
	}
	
	// ****************************************************************

	private synchronized boolean areListsConsistent() {
	
		return this.submission_t_list.size() == this.completion_t_list.size();
		
	}
	
	
	// ****************************************************************

	private final String peer_id;
	private final boolean store_data;
	private final ArrayList<Long> submission_t_list;
	private final ArrayList<Long> completion_t_list;

}

