/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ResourceDataTracker.java

package lbg.peer.core.rms;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.*;
import static lbg.peer.workflow.TemporalTasksGrouping.*;
import lbg.protocol.data.*;
import lbg.resource.service.*;
import lbg.task.*;
import java.util.*;

/** A ResourceDataTracker tracks the (Local and Supplying) data stored
 * on all Resources of the Peer.
 * <p>The tracked data are those sent in a WorkingSet to a Resource.
 * To each data id is associated the set of Resources
 * where the data has been sent.
 * <p>The tracked data can be searched to locate Resources
 * following various criteria using the Storage Affinity metric.
 * Of particular interest is the
 * locateIdleResourceWithBestStorageAffinity() method.
 * @author Cyril Briquet
 */
public class ResourceDataTracker {

	/** Constructs a new ResourceDataTracker.
	 */
	public ResourceDataTracker(Peer managed_peer) throws GridException {

		if (managed_peer == null) {
			throw new GridException("ResourceDataTracker(): illegal Peer");
		}
		
		this.managed_peer = managed_peer;
		this.db = new TreeMap<String,Set<ResourceHandle>>();
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ResourceDataTracker] ");

		sb.append("\n  ->  Tracked data DB size = ");
		sb.append(this.db.size());
		sb.append("\n  ->  Tracked data = { ");
		for (String data_id : this.db.keySet()) {
			sb.append(data_id);
			sb.append(" ");
		}
		sb.append("}");

		return sb.toString();

	}
	
	// ****************************************************************

	/**
	 * @return true if target data was not already tracked on target Resource
	 */
	public synchronized boolean trackData(String data_id,ResourceHandle rh)
		throws GridException {
		
		if (data_id == null) {
			throw new GridException("trackData(): illegal data id");
		}
		if (rh == null) {
			throw new GridException("trackData(): illegal ResourceHandle");
		}
		
		Set<ResourceHandle> data_item = this.db.get(data_id);
		if (data_item == null) {
			data_item = new TreeSet<ResourceHandle>();
			if (this.db.put(data_id,data_item)!=null) {
				throw new GridException("trackData(): corrupted data");
			}
		}

		return data_item.add(rh);
		
	}
	
	public synchronized void untrackData(String data_id,ResourceHandle rh)
		throws GridException {

		if (data_id == null) {
			throw new GridException("untrackData(): illegal data id");
		}
		if (rh == null) {
			throw new GridException("untrackData(): illegal ResourceHandle");
		}
		
		Set<ResourceHandle> data_item = this.db.get(data_id);
		if (data_item == null) {
			throw new GridException("untrackData(): data "+data_id+
									" not tracked");
		}
		
		if (data_item.remove(rh) == false) {
			throw new GridException("untrackData(): data not tracked "+
									"on target Resource");
		}
		
		if (data_item.size() == 0) {
			if (this.db.remove(data_id) == null) {
				throw new GridException("untrackData(): corrupted data");
			}
		}
		
	}

	public synchronized void untrackData(String[] data_ids,ResourceHandle rh)
		throws GridException {
		
		if (data_ids == null) {
			throw new GridException("untrackData(): illegal data ids");
		}
		if (rh == null) {
			throw new GridException("untrackData(): illegal ResourceHandle");
		}
		
		for (String data_id : data_ids) {
			untrackData(data_id,rh);
		}

	}

	/** Untrack all data for target Resource.
	 * Checks if target data includes all data tracked for target Resource.
	 */
	public synchronized void untrackAllData(String[] data_ids,ResourceHandle rh)
		throws GridException {
		
		untrackData(data_ids,rh);
		
		for (Set<ResourceHandle> set : db.values()) {
			if (set.contains(rh)) {
				throw new GridException("untrackAllData(): corrupted data");
			}
		}

	}

	// ****************************************************************

	/** Locate Resources which have target data in their WorkingSet.
	 * @return null if no Resource has target data in her WorkingSet
	 */
	public synchronized Set<ResourceHandle>
		locateResourcesWithReplicatedData(String data_id)
		throws GridException {
	
		if (data_id == null) {
			throw new GridException("locateResourcesWithReplicatedData(): "+
									"illegal data id");
		}
		
		Set<ResourceHandle> s = this.db.get(data_id);

		return  s;

	}

	/** Locate Resources which have all target datas in their WorkingSet.
	 * @return null if no Resource has all target datas in her WorkingSet
	 */
	public synchronized Set<ResourceHandle>
		locateResourcesWithReplicatedData(String[] data_ids)
		throws GridException {
	
		if (data_ids == null) {
			throw new GridException("locateResourcesWithReplicatedData(): "+
									"illegal data ids");
		}
		
		Set<ResourceHandle> inter_s = null;
		for (String data_id : data_ids) {
			if (data_id == null) {
				throw new GridException("locateResourcesWithReplicatedData(): "+
										"illegal data id");
			}
			
			Set<ResourceHandle> s = this.db.get(data_id);
			
			if (s == null) {
				continue;
			}
			
			if (inter_s == null) {
				inter_s = new TreeSet<ResourceHandle>(s);
			}
			else {
				inter_s.retainAll(s);
			}
		}

		return ((inter_s == null) ||
				(inter_s.size() == 0)) ?
			null : inter_s;

	}

	// ****************************************************************

	/** Locate Resources which have some data of target Task
	 * in their WorkingSet.
	 * @deprecated
	 * @return null if no idle Resource has any target data in her WorkingSet
	 */
	@Deprecated
	public synchronized Set<TransferCostData>
		locateIdleResourcesWithStorageAffinity(Task t)
		throws GridException {
	
		if (t == null) {
			throw new GridException("locateResourcesWithStorageAffinity(): "+
									"illegal Task");
		}
		
		Set<ResourceHandle> union_s = new TreeSet<ResourceHandle>();
		for (MetaGridData data : t.getInputMetaGridDataSet()) {
			if (data.id == null) {
				throw new GridException("locateResources"+
										"WithStorageAffinity(): "+
										"illegal data id");
			}
			
			Set<ResourceHandle> s = this.db.get(data.id);
			
			if (s == null) {
				continue;
			}
			
			union_s.addAll(s);
		}
		
		if (union_s.size() == 0) {
			return null;
		}

		// union_s contains all Resources
		// which have at least one data in common with target Task
		
		TreeSet<TransferCostData> sorted_s =
			new TreeSet<TransferCostData>();
		long HIGHEST_TRANSFER_COST = t.getInputDataSetSize();
		
		for (ResourceHandle rh : union_s) {
			WorkingSet ws =
				this.managed_peer.rms().getWorkingSetOfIdleResource(rh);
			if (ws == null) {
				// Resource is busy !
				if (rh.idle() == true) {
					throw new GridException("locateResources"+
											"WithStorageAffinity(): "+
											"corrupted RMS data");
				}
				continue; // insert nothing
			}
			// What's the transfer cost of sending t on rh ?
			long d = distance(ws.getMetaGridData(),t.getInputMetaGridDataSet());
			assert(d <= HIGHEST_TRANSFER_COST);
			TransferCostData tch = new TransferCostData(rh,d);
			sorted_s.add(tch);			
		}
		
		// sorted_s is the sorted (by increasing transfer costs) set
		// of Resources sharing some data with target Task

		return (sorted_s.size() > 0) ?
			sorted_s : null;

	}

	/** Locate best (Storage Affinity-wise) Resource
	 * which has a mximum of data required by target Task in her WorkingSet.
	 * <p>If there are several such Resources, the one with the largest
	 * number of unused cache slots (i.e. free space) is selected.
	 * @return null if no idle Resource has any target data in her WorkingSet
	 */
	public synchronized ResourceHandle
		locateIdleResourceWithBestStorageAffinity(Task t)
		throws GridException {
	
		if (t == null) {
			throw new GridException("locateIdleResource"+
									"WithBestStorageAffinity(): "+
									"illegal Task");
		}
		
		// Seek Resources with at least one data in common with target Task
		Set<ResourceHandle> union_s = new TreeSet<ResourceHandle>();
		for (MetaGridData data : t.getInputMetaGridDataSet()) {
			if (data.id == null) {
				throw new GridException("locateIdleResource"+
										"WithBestStorageAffinity(): "+
										"illegal data id");
			}
			
			Set<ResourceHandle> s = this.db.get(data.id);
			
			if (s == null) {
				continue;
			}
			
			union_s.addAll(s);
		}
		
		if (union_s.size() == 0) {
			return null;
		}

		// union_s now contains all Resources
		// which have at least one data in common with target Task
		
		ResourceHandle best_rh = null;
		long HIGHEST_TRANSFER_COST = t.getInputDataSetSize();
		long lowest_transfer_cost = HIGHEST_TRANSFER_COST+1;
		int  lowest_ws_size = Integer.MAX_VALUE;

		// Maximize StorageAffinity by minimizing remaining transfer cost metric
		boolean flag = false;
		for (ResourceHandle rh : union_s) {
			WorkingSet ws =
				this.managed_peer.rms().getWorkingSetOfIdleResource(rh);
			if (ws == null) {
				// Resource is busy
				// (DO NOT check status with idle() => risk of race condition)
				flag = true;
				continue; // don't process
			}

			// What's the transfer cost of sending t on rh ?
			long d = distance(ws.getMetaGridData(),t.getInputMetaGridDataSet());
			assert(d <= HIGHEST_TRANSFER_COST);
			// New minimum transfer cost
			if (d < lowest_transfer_cost) {
				best_rh = rh;
				lowest_transfer_cost = d;
				lowest_ws_size = ws.size();
			}
			// Another Resource with (current) minimum transfer cost ...
			else if (d == lowest_transfer_cost) {
				int ws_size = ws.size();
				// ... with largest number of unused cache slots
				if (ws_size < lowest_ws_size) {
					best_rh = rh;
					lowest_transfer_cost = d;
					lowest_ws_size = ws_size;
				}
			}

		}
		
		// Consistency check: if all Resources in union_s were idle,
		// *_transfer_*cost must have been modified
		if ((flag == false) && (lowest_transfer_cost > HIGHEST_TRANSFER_COST)) {
				throw new GridException("locateIdleResource"+
										"WithBestStorageAffinity(): "+
										"corrupted data");
		}
		
		return best_rh;

	}

	// ****************************************************************

	/** Locate data in target data set that is the most replicated
	 * on this Peer Resources.
	 */
	public synchronized String
		locateMostLocallyReplicatedData(String[] data_ids)
		throws GridException {
			
		if ((data_ids == null) || (data_ids.length == 0)) {
			throw new GridException("locateMostLocallyReplicatedData(): "+
									"illegal data ids");
		}

		String max_data_id = null;
		int max_freq = -1;
		for (String data_id : data_ids) {
			if (data_id == null) {
				throw new GridException("locateMostLocallyReplicatedData(): "+
										"illegal data id");
			}
			
			Set<ResourceHandle> s = this.db.get(data_id);
			if (s == null) {
				continue;
			}

			int freq = s.size();
			if (freq > max_freq) {
				max_freq = freq;
				max_data_id = data_id;
			}
		}
		
		return max_data_id;
	
	}
	
	/** Locate data in target data set that is the less replicated
	 * on this Peer Resources.
	 */
	public synchronized String
		locateLessLocallyReplicatedData(String[] data_ids)
		throws GridException {
			
		if ((data_ids == null) || (data_ids.length == 0)) {
			throw new GridException("locateLessLocallyReplicatedData(): "+
									"illegal data ids");
		}

		String min_data_id = null;
		int min_freq = this.db.size()+1;
		for (String data_id : data_ids) {
			if (data_id == null) {
				throw new GridException("locateLessLocallyReplicatedData(): "+
										"illegal data id");
			}
			
			Set<ResourceHandle> s = this.db.get(data_id);
			if (s == null) {
				continue;
			}

			int freq = s.size();
			if (freq < min_freq) {
				min_freq = freq;
				min_data_id = data_id;				
			}
		}
		
		return min_data_id;
	
	}

	// ****************************************************************

	private Peer managed_peer;
	private Map<String,Set<ResourceHandle>> db;

	// ****************************************************************

	/** Helper class of Data Tracker.
	 * Stores data about a Resource and the associated transfer cost
	 * to that Resource of an implicit target Task.
	 */
	public static class TransferCostData
		implements Comparable <TransferCostData> {
	
		/** Constructs a new TransferCostData.
		 */
		public TransferCostData(ResourceHandle rh,long distance)
			throws GridException {
				
			if (rh == null) {
				throw new GridException("TransferCostData(): "+
										"illegal ResourceHandle");
			}
			if (distance < 0) {
				throw new GridException("TransferCostData(): "+
										"illegal distance");
			}
	
			this.rh = rh;
			this.distance = distance;
			this.hashcode = Integer.MIN_VALUE;
			
		}
		
		public String toString() {
		
			StringBuilder sb = new StringBuilder();
			
			sb.append("[TransferCostData] ");
	
			return sb.toString();
	
		}
		
		/**
		 * <p>Sort by increasing size.
		 */
		public int compareTo(TransferCostData tch) {
		
			long my_size    = this.distance;
			long other_size = tch.distance();
			
			if (my_size < other_size) {
				return -1;
			}
			else if (my_size > other_size) {
				return +1;
			}
			else /* if (my_size == other_size) */ {
				return 0;
			}
			
		}
		
		@Override
		public boolean equals(Object obj) {
			
			if (obj instanceof TransferCostData == false) {
				return false;
			}
			
			return (this.compareTo((TransferCostData)obj) == 0);
			
		}

		@Override
		public int hashCode() {
		
			if (this.hashcode == Integer.MIN_VALUE) {
				this.hashcode = CommonMath.generateHashCode(this.distance);
			}
			return this.hashcode;
		
		}
		
		public ResourceHandle resource() {
		
			return this.rh;
			
		}
		
		public long distance() {
		
			return this.distance;
			
		}
	
		private final ResourceHandle rh;
		private final long distance;
		private int hashcode;
		
	}

	// ****************************************************************

}

