/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ResourceLocator.java

package lbg.peer.core.rms;

import lbg.common.exception.*;
import static lbg.common.math.CommonMath.generateHashCode;
import lbg.peer.*;
import lbg.resource.service.*;
import lbg.task.*;
import java.util.*;

/** A ResourceLocator tracks Task -> Resource mappings
 * so as to locate the Resource running a given Task.
 * @author Cyril Briquet
 */
public class ResourceLocator {

	public ResourceLocator(Peer managed_peer) throws GridException {
		
		if (managed_peer == null) {
			throw new GridException("ResourceLocator(): illegal Peer");
		}

		this.managed_peer = managed_peer;
		this.rs_locator_local =
			new TreeMap<ResourceLocatorKey,ResourceHandle>();
		this.rs_locator_supplying =
			new TreeMap<ResourceLocatorKey,ResourceHandle>();

	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ResourceLocator ");
		sb.append(this.managed_peer.id());
		sb.append("]");
		
		sb.append("\n  ->  ");
		sb.append("Tracked Resource DB size = ");
		sb.append(this.rs_locator_local.size()+
				  this.rs_locator_supplying.size());
		
		return sb.toString();

	}
	
	// ****************************************************************

	public synchronized
		void trackBusyLocalResource(Task t,ResourceHandle rh)
		throws GridException {
		
		if (t == null) {
			throw new GridException("trackBusyLocalResource(): "+
									"illegal Task");
		}
		if (rh == null) {
			throw new GridException("trackBusyLocalResource(): "+
									"illegal Resource");
		}
		
		ResourceLocatorKey rlk = new ResourceLocatorKey(t,true);

		if (this.rs_locator_local.put(rlk,rh) != null) {
			throw new GridException("trackBusyLocalResource(): "+
									"corrupted Busy Local Resource "+
									"locator data");
		}
		
	}
	
	public synchronized
		void trackBusySupplyingResource(Task t,ResourceHandle rh)
		throws GridException {
		
		if (t == null) {
			throw new GridException("trackBusySupplygingResource(): "+
									"illegal Task");
		}
		if (rh == null) {
			throw new GridException("trackBusySupplyingResource(): "+
									"illegal Resource");
		}
		
		ResourceLocatorKey rlk = new ResourceLocatorKey(t,false);

		if (this.rs_locator_supplying.put(rlk,rh) != null) {
			throw new GridException("trackBusySupplyingResource(): "+
									"corrupted Busy Supplying Resource "+
									"locator data");
		}
	
	}

	public synchronized
		void untrackBusyLocalResource(Task t) throws GridException {

		if (t == null) {
			throw new GridException("untrackBusyLocalResource(): "+
									"illegal Task");
		}

		ResourceLocatorKey rlk = new ResourceLocatorKey(t,true);

		if (this.rs_locator_local.remove(rlk) == null) {
			throw new GridException("untrackBusyLocalResource(): "+
									"corrupted Busy Local Resource "+
									"locator data");
		}
		
	}
	
	public synchronized
		void untrackBusySupplyingResource(Task t) throws GridException {
		
		if (t == null) {
			throw new GridException("untrackBusySupplyingResource(): "+
									"illegal Task");
		}

		ResourceLocatorKey rlk = new ResourceLocatorKey(t,false);

		if (this.rs_locator_supplying.remove(rlk) == null) {
			throw new GridException("untrackBusySupplyingResource(): "+
									"corrupted Busy Supplying Resource "+
									"locator data");
		}
		
	}
	
	// ****************************************************************

	/** Locates a Busy Local Resource running target Task (may be null).
	 * @param t target Task
	 * @return Busy Local Resource running target Task,
	 * or null if there no such Resource can be found
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized ResourceHandle locateBusyLocalResource(Task t)
		throws GridException {
		
		if (t == null) {
			throw new GridException("locateBusyLocalResource(): "+
									"illegal Task");
		}

		ResourceLocatorKey rlk = new ResourceLocatorKey(t,true);

		ResourceHandle rh = this.rs_locator_local.get(rlk);
		
		return rh;
		
	}

	/** Locates a Busy Supplying Resource running target Task (may be null).
	 * @param t target Task
	 * @return Busy Supplying Resource running target Task,
	 * or null if there no such Resource can be found
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized ResourceHandle locateBusySupplyingResource(Task t)
		throws GridException {
		
		if (t == null) {
			throw new GridException("locateBusySupplyingResource(): "+
									"illegal Task");
		}

		ResourceLocatorKey rlk = new ResourceLocatorKey(t,false);

		ResourceHandle rh = this.rs_locator_supplying.get(rlk);
		
		return rh;
		
	}

	/** Locates a Busy Local Resource running target Task
	 *  (guaranteed to be non-null).
	 * @param t target Task
	 * @return Busy Local Resource running target Task
	 * (guaranteed to be non-null)
	 * @exception GridException if a parameter is illegal
	 * or if no such Resource may be found
	 */
	public synchronized ResourceHandle
		locateAndCheckBusyLocalResource(Task t) throws GridException {

		ResourceHandle rh = this.locateBusyLocalResource(t);
		if (rh == null) {
			throw new GridException("locateAndCheckBusyLocalResource(): "+
									"Resource not found");
		}
		
		return rh;
		
	}

	/** Locates a Busy Supplying Resource running target Task
	 *  (guaranteed to be non-null).
	 * @param t target Task
	 * @return Busy Supplying Resource running target Task
	 * (guaranteed to be non-null)
	 * @exception GridException if a parameter is illegal
	 * or if no such Resource may be found
	 */
	public synchronized ResourceHandle
		locateAndCheckBusySupplyingResource(Task t) throws GridException {
		
		ResourceHandle rh = this.locateBusySupplyingResource(t);
		if (rh == null) {
			throw new GridException("locateAndCheckBusySupplyingResource(): "+
									"Resource not found");
		}
		
		return rh;
		
	}

	// ****************************************************************

	private final Peer managed_peer;
	private final Map<ResourceLocatorKey,ResourceHandle> rs_locator_local;
	private final Map<ResourceLocatorKey,ResourceHandle> rs_locator_supplying;

	// ****************************************************************

	/** Helper class of ResourceLocator.
	 * Allows to identify the Resource running the target Task.
	 */
	public static class ResourceLocatorKey
		implements Comparable<ResourceLocatorKey> {
		
		/**
		 * @param t target Task
		 * @param is_local_task true if target Task is Local,
		 * false if target Task is Supplying
		 * @exception GridException if a parameter is illegal
		 */
		public ResourceLocatorKey(Task t,boolean is_local_task)
			throws GridException {
	
			if (t == null) {
				throw new GridException("ResourceLocatorKey(): "+
										"illegal Task");
			}
	
			this.is_local_task = is_local_task;
	
			if (this.is_local_task) {
				this.job_id = t.jobID();
				this.bot_id = t.botID();
			}
			else {
				this.job_id = t.getSupplyingSpaceJobID();
				this.bot_id = t.getSupplyingSpaceBoTID();
			}
			this.task_id = t.id();
			
			this.hashcode = Integer.MIN_VALUE;
			
		}
			
		public String toString() {
				
			StringBuilder sb = new StringBuilder();
				
			sb.append("[ResourceLocatorKey]");
			sb.append(this.is_local_task ? "Local" : "Supplying");
			sb.append(" Job id=");
			sb.append(this.job_id);
			sb.append(", ");
			sb.append(is_local_task ? "Local" : "Supplying");
			sb.append(" BoT id=");
			sb.append(this.bot_id);
			sb.append(", Task id=");
			sb.append(this.task_id);
				
			return sb.toString();
				
		}
		
		public int compareTo(ResourceLocatorKey other) {
	
			int other_job_id = other.job_id;
			int other_bot_id = other.bot_id;
			int other_task_id = other.task_id;
				
			if (this.job_id < other_job_id ) {
				return -1;				
			}
			else if (this.job_id > other_job_id) {
				return +1;
			}
			else /* if (this.job_id == other_job_id) */ {
				if (this.bot_id < other_bot_id) {
					return -1;
				}
				else if (this.bot_id > other_bot_id) {
					return +1;
				}
				else /* if (this.bot_id == other_bot_id) */ {
					if (this.task_id < other_task_id) {
						return -1;
					}
					else if (this.task_id > other_task_id) {
						return +1;
					}
					else /* if (this.task_id == other_task_id) */ {
						return 0;
					}
				}
			}
			
		}

		@Override
		public boolean equals(Object obj) {
			
			if (obj instanceof ResourceLocatorKey == false) {
				return false;
			}
			
			return (this.compareTo((ResourceLocatorKey)obj) == 0);
			
		}
		
		@Override
		public int hashCode() {
		
			if (this.hashcode == Integer.MIN_VALUE) {
				this.hashcode =
					generateHashCode(this.job_id,
									 generateHashCode(
									   generateHashCode(this.bot_id),
									   generateHashCode(this.task_id)));
			}
			return this.hashcode;
		
		}

		public final boolean is_local_task;
		public final int job_id;
		public final int bot_id;
		public final int task_id;
		private int hashcode;
		
	}

	// ****************************************************************

}

