/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ResourceManager.java

package lbg.peer.core.rms;

import lbg.common.exception.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.peer.*;
import static lbg.peer.core.rms.ResourceState.*;
import lbg.peer.register.learning.consumption.*;
import lbg.peer.register.learning.supplying.*;
import lbg.protocol.data.*;
import lbg.resource.service.*;
import lbg.task.*;
import java.util.*;

/**
 * @author Cyril Briquet
 */
public class ResourceManager {

	public ResourceManager(Peer managed_peer,boolean data_present_enabled,
						   boolean data_replication_enabled,
						   float idle_replication_ratio,
						   boolean psufferage_enabled) throws GridException {

		if (managed_peer == null) {
			throw new GridException("ResourceManager(): illegal Peer");
		}
		if ((idle_replication_ratio < 0.0) || (idle_replication_ratio > 1.0)) {
			throw new GridException("ResourceManager(): "+
									"illegal idle replication share");
		}
		
		this.managed_peer = managed_peer;
		this.resource_locator = new ResourceLocator(this.managed_peer);

		this.resource_data_tracker =
			new ResourceDataTracker(this.managed_peer);
		this.consumption_data_tracker =
			new PeerConsumptionDataTracker(128);
		this.supplying_data_tracker =
			new PeerSupplyingDataTracker(this.managed_peer,128);
		this.data_present_enabled     = data_present_enabled;
		this.data_replication_enabled = data_replication_enabled;
		this.idle_replication_ratio   = idle_replication_ratio;
		this.psufferage_enabled       = psufferage_enabled;

		this.resource_tracker =
			new ResourceTracker(this.managed_peer,this.resource_data_tracker);
		this.cached_utilization_value = 0.0;
		
		this.power_tracker = new TreeMap<ResourceHandle,Integer>();
		this.total_power = 0;
		this.available_power = 0;
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ResourceManager ");
		sb.append(this.managed_peer.id());
		sb.append("]");

		sb.append("\n  ->  ");
		sb.append("data present Resource location heuristic ");
		sb.append(this.data_present_enabled ? "enabled" : "disabled");
		sb.append("\n  ->  ");
		sb.append("proactive data replication heuristic ");
		sb.append(this.data_replication_enabled ? "enabled" : "disabled");
		sb.append(", ");
		sb.append("idle replication ratio = ");
		sb.append(this.idle_replication_ratio);
		sb.append("\n  ->  ");
		sb.append("PSufferage Resource selection heuristic "+
				  "(for Supplying Task preemption) ");
		sb.append(this.psufferage_enabled ? "enabled" : "disabled");
		
		sb.append("\n  ->  nominal power=");
		sb.append(this.total_power);
		sb.append(", available power=");
		sb.append(this.available_power);
		/*for (ResourceHandle rh : this.power_tracker.keySet()) {
			sb.append("\n  ->  ");
			sb.append(rh.id());
			sb.append(" power=");
			sb.append(this.power_tracker.get(rh));
		}*/
		
		sb.append("\n  ->  ");
		sb.append(this.resource_tracker);
		sb.append("\n  ->  ");
		sb.append(this.resource_locator);
		sb.append("\n  ->  ");
		sb.append(this.resource_data_tracker);
		sb.append("\n  ->  ");
		sb.append(this.consumption_data_tracker);
		sb.append("\n  ->  ");
		sb.append(this.supplying_data_tracker);
		
		return sb.toString();

	}

	// ****************************************************************

	public synchronized double meanNominalResourcePower() {
	
		return ((double)this.total_power) /
			((double)this.resource_tracker.countResources());
		
	}

	public synchronized void addResource(ResourceHandle rh,int power)
		throws GridException {
		
		assert(threeSetsInvariant());
		
		if (rh == null) {
			throw new GridException("addResource(): illegal ResourceHandle");
		}
		if (power < 1) {
			throw new GridException("addResource(): illegal Resource power");
		}
		
		this.resource_tracker.trackResource(rh);
		
		this.total_power += power;
		this.available_power += power;
		this.power_tracker.put(rh,power); // retval != null => re-registration
		
		assert(threeSetsInvariant());
		
	}

	/** Removes a Resource from the Peer.
	 * If a Task was running on this Resource:
	 * (1) the ResourceManager informs the other managers,
	 * and (2) the Resource is supposed to actually cancel the Task.
	 */
	public synchronized void removeResource(ResourceHandle rh)
		throws GridException {

		assert(threeSetsInvariant());

		if (rh == null) {
			throw new GridException("removeResource(): illegal ResourceHandle");
		}

		Integer res_power_dummy = this.power_tracker.remove(rh);
		if (res_power_dummy == null) {
			throw new GridException("removeResource(): Resource not found "+
									"(this may or may not be expected)");
		}
		int res_power = res_power_dummy.intValue();
		this.total_power -= res_power;

		ResourceUsageData rud = this.resource_tracker.getResourceUsageData(rh);

		if (rud == null) {
			throw new GridException("removeResource(): corrupted data");
		}
		
		switch (rud.state()) {
		case IDLE:
			this.available_power -= res_power;
			break;
		case BUSY_LOCAL:
			Task t = rud.runningTask();
			// Untrack Task/Resource mapping (*BEFORE* calling clearResource())
			this.resource_locator.untrackBusyLocalResource(t);
			// Put Managers in consistent state
			this.managed_peer.xms().
				preemptLocalTaskOnLocalResource(t,DUMMY_LOG);
			//this.available_power += rh_power; -- not needed: Resource's busy
			break;
		case BUSY_SUPPLYING:
			t = rud.runningTask();
			// Untrack Task/Resource mapping (*BEFORE* calling clearResource())
			this.resource_locator.untrackBusySupplyingResource(t);
			// Put Managers in consistent state
			this.managed_peer.xms().
				preemptSupplyingTaskOnLocalResource(t,DUMMY_LOG);
			//this.available_power += rh_power; -- not needed: Resource's busy
			break;
		default:
			throw new GridException("removeResource(): unsupported state");
		}

		// Untrack all data
		this.resource_data_tracker.
			untrackAllData(rud.working_set.getDataId(),rh);

		// Actually remove the Resource
		this.resource_tracker.untrackResource(rh);

		assert(threeSetsInvariant());

	}

	// ****************************************************************

	public int countTotalResources() {
		
		return countIdleResources()+countBusyResources();
		
	}
	
	public int countIdleResources() {

		return this.resource_tracker.countIdleResources();

	}
	
	public int countBusyResources() {
		
		return countBusyLocalResources()+countBusySupplyingResources();
		
	}

	public int countBusyLocalResources() {

		return this.resource_tracker.countBusyLocalResources();
		
	}

	public int countBusySupplyingResources() {

		return this.resource_tracker.countBusySupplyingResources();
		
	}

	public boolean hasIdleResource() {

		return this.resource_tracker.hasIdleResource();
		
	}

	public boolean hasBusyLocalResource() throws GridException {

		return this.resource_tracker.hasBusyLocalResource();
			
	}

	public boolean hasBusySupplyingResource() throws GridException {

		return this.resource_tracker.hasBusySupplyingResource();
		
	}
	
	/** Returns the Peer utilization, i.e. total busy time of
	 * all currently registered Resources,
	 * divided by their total registration time.
	 * @return Peer utilization, in [0.0..1.0]
	 * @exception GridException if Resource usage data is corrupted
	 */
	public synchronized double utilization() throws GridException {

		return this.resource_tracker.utilization();
	
	}

	/** Returns the most cached value of Peer utilization,
	 * i.e. total busy time of
	 * all currently registered Resources,
	 * divided by their total registration time.
	 * This value may at most 5 seconds stale.
	 * @return Peer utilization, in [0.0..1.0]
	 */
	public double getCachedUtilizationValue() {

		return this.cached_utilization_value;
	
	}
	
	public synchronized void updateUtilizationValue() throws GridException {
		
		this.cached_utilization_value = this.utilization();
		
	}
	
	// ****************************************************************

	public synchronized ConsumptionDataTracker dataTracker() {
		
		return this.consumption_data_tracker;
		
	}

	/** Locates a Resource currentlty in target state
	 * with maximum Storage Affinity with target Task.
	 * @param state target state
	 * @param t target Task
	 * @return Resource in target state with maximum Storage Affinity
	 * with target Task, or null if no such Task can be found
	 */
	public synchronized ResourceHandle
		locateResource(ResourceState state,Task t) throws GridException {
		
		if (state == null) {
			throw new GridException("locateResource(): illegal state");
		}
		if (t == null) {
			throw new GridException("locateResource(): illegal Task");
		}

		// First check if it is worthwhile to apply search algorithm
		if (isSomeResourceAvailable(state) == false) {
			return null;
		}
		
		ResourceHandle rh = null;

		// Storage Affinity search
		if ((this.data_present_enabled) && (state == IDLE)) {
			rh = locateIdleResourceWithBestSA(state,t);
		}

		// PSufferage search
		if ((this.psufferage_enabled) && (state == BUSY_SUPPLYING)) {
			rh = locateBusySupplyingResourceWithNewestTask(state);
		}
		
		// Random search
		if (rh == null) {
			rh = locateResourceRandomly(state);
		}
		
		// Sanity check
		checkLocatedResource(state,rh);
		
		return rh;
	
	}
	
	private synchronized boolean isSomeResourceAvailable(ResourceState state)
		throws GridException {
		
		if (state == null) {
			throw new GridException("isSomeResourceAvailable(): illegal state");
		}
		
		switch (state) {
		case IDLE:
			if (this.resource_tracker.hasIdleResource() == false) {
				return false;
			}
			break;
		case BUSY_LOCAL:
			if (this.resource_tracker.hasBusyLocalResource() == false) {
				return false;
			}
			break;
		case BUSY_SUPPLYING:
			if (this.resource_tracker.hasBusySupplyingResource() == false) {
				return false;
			}
			break;
		default:
			throw new GridException("isSomeResourceAvailable(): "+
									"unsupported state");
		}
		
		return true;
		
	}
	
	private synchronized ResourceHandle
		locateResourceRandomly(ResourceState state) throws GridException {
		
		if (state == null) {
			throw new GridException("locateResourceRandomly(): illegal state");
		}

		ResourceHandle rh = null;
		switch (state) {
		case IDLE:
			rh = this.resource_tracker.getRandomIdleResource();
			break;
		case BUSY_LOCAL:
			rh = this.resource_tracker.getRandomBusyLocalResource();
			break;
		case BUSY_SUPPLYING:
			rh = this.resource_tracker.getRandomBusySupplyingResource();
			break;
		default:
			throw new GridException("locateResourceRandomly(): "+
									"unsupported state");
		}
		//debugging_msg_2(state,t,rh);

		return rh;
		
	}

	/**
	 * @return idle Resource with the data cache
	 * having the highest Storage Affinity with the target prioritary Task,
	 * or null if no idle Resource has Storage Affinity with that Task
	 */
	private synchronized ResourceHandle
		locateIdleResourceWithBestSA(ResourceState state,Task t)
		throws GridException {
		
		if (state == null) {
			throw new GridException("locateIdleResourceWithBestSA(): "+
									"illegal state");
		}
		if (state != IDLE) {
			throw new GridException("locateIdleResourceWithBestSA(): "+
									"illegal state (busy "+
									((state == BUSY_LOCAL)?"local":"supplying")+
									")");
		}
		if (t == null) {
			throw new GridException("locateIdleResourceWithBestSA(): "+
									"illegal Task");
		}
		
		MetaGridData[] data = t.getInputMetaGridDataSet();
		if (data == null) {
			throw new GridException("locateIdleResourceWithBestSA(): "+
									"corrupted Task");
		}

		ResourceHandle rh = null;

		if (data.length > 0) {
			// Data present (Storage Affinity) search algorithm,
			// applicable only for Tasks with some input data
			rh = this.resource_data_tracker.
				locateIdleResourceWithBestStorageAffinity(t);
			//if (rh != null) debugging_msg_1(state,t,rh);
		}

		return rh;

	}

	/**
	 * @return Resource running the ungraced Supplying Task that has been
	 * running for the less time, or null if no such Resource can be found
	 */
	private synchronized ResourceHandle
		locateBusySupplyingResourceWithNewestTask(ResourceState state)
		throws GridException {
		
		if (state == null) {
			throw new GridException("locateBusySupplyingResource"+
									"WithNewestTask(): illegal state");
		}
		if (state != BUSY_SUPPLYING) {
			throw new GridException("locateBusySupplyingResource"+
									"WithNewestTask(): illegal state ( "+
									((state == BUSY_LOCAL)?"busy local":"idle")+
									")");
		}
		
		Task t = this.managed_peer.xms().
			getMostRecentlyRunningAndUngracedSupplyingTask();
		if (t == null) {
			return null;
		}
		
		return this.resource_locator.locateAndCheckBusySupplyingResource(t);
	
	}
	
	private synchronized void
		checkLocatedResource(ResourceState state,ResourceHandle rh)
		throws GridException {
		
		if (state == null) {
			throw new GridException("checkLocatedResource(): illegal state");
		}
		if (rh == null) {
			throw new GridException("checkLocatedResource(): illegal Resource");
		}
		
		switch (state) {
		case IDLE:
			if (rh.idle() == false) {
				this.managed_peer.logger().
					info(getStackTrace("checkLocatedResource(): "+
									   "Resource status conflict (busy)"));
			}
			break;
		case BUSY_LOCAL:
		case BUSY_SUPPLYING:
			if (rh.idle() == true) {
				this.managed_peer.logger().
					info(getStackTrace("checkLocatedResource(): "+
									   "Resource status conflict (idle)"));
			}
			break;
		default:
			throw new GridException("checkLocatedResource(): "+
									"unsupported state");
		}
		
	}

	// ****************************************************************

	/** Runs target Task on target Resource.
	 * @param rh target Resource
	 * @param t target Task
	 * @return true if the target Resource can run the target Task,
	 * or false if the target Resource cannot run the target Task
	 * (e.g. typically because the WorkingSet is too large for the Resource,
	 * which would be caused by uncareful scheduling)
	 */
	public synchronized boolean runLocalTask(ResourceHandle rh,Task t)
		throws GridException {

		assert(threeSetsInvariant());

		if (rh == null) {
			throw new GridException("runLocalTask(): illegal ResourceHandle");
		}
		
		if (rh.idle() == false) {
			throw new GridException("runLocalTask(): resource busy");
		}

		WorkingSet ws =
			this.resource_tracker.
				updateResourceStatusToBusy(rh,BUSY_LOCAL,t);
		ws.clearAddAndLockLocalInputData(t);
		
		this.resource_locator.trackBusyLocalResource(t,rh);
		this.available_power -= this.power_tracker.get(rh);

		boolean accepted = rh.runTask(t,ws.getData());
		if (accepted == false) {
			this.managed_peer.logger().
				info("runLocalTask(): working set rejected by Resource "+
					 ws.managedResource().id());
			this.setIdleResourceState(t,rh,BUSY_LOCAL);
		}

		assert(threeSetsInvariant());

		return accepted;

	}

	/** Runs target Task on target Resource.
	 * @param rh target Resource
	 * @param t target Task
	 * @return true if the target Resource can run the target Task,
	 * or false if the target Resource cannot run the target Task
	 * (e.g. typically because the WorkingSet is too large for the Resource,
	 * which would be caused by uncareful scheduling)
	 */
	public synchronized boolean runSupplyingTask(ResourceHandle rh,Task t)
		throws GridException {

		assert(threeSetsInvariant());

		if (rh == null) {
			throw new GridException("runSupplyingTask(): "+
									"illegal ResourceHandle");
		}

		if (rh.idle() == false) {
			throw new GridException("runSupplyingTask(): resource busy");
		}

		WorkingSet ws =
			this.resource_tracker.
				updateResourceStatusToBusy(rh,BUSY_SUPPLYING,t);
		ws.clearAddAndLockSupplyingInputData(t);
		if (this.data_replication_enabled) {
			this.supplying_data_tracker.add(t.getInputMetaGridDataSet());
		}

		this.resource_locator.trackBusySupplyingResource(t,rh);
		this.available_power -= this.power_tracker.get(rh);

		boolean accepted = rh.runTask(t,ws.getData());
		if (accepted == false) {
			this.managed_peer.logger().
				info("runSupplyingTask(): working set rejected by Resource "+
					 ws.managedResource().id());
			this.setIdleResourceState(t,rh,BUSY_SUPPLYING);
		}

		assert(threeSetsInvariant());
		
		return accepted;

	}
	
	// ****************************************************************

	/** Clears the state of the target Resource.
	 * <p>It is expected that the Resource is idle
	 * and has already completed the target Task,
	 * because freeResource() <b>will not perform Task completion
	 * on the Resource</b>.
	 * @param t target Local Task
	 * @return target Resource state change time
	 * (i.e. the time at which the Resource became busy),
	 * or -1 if is_resource_alive is false
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized long freeBusyLocalResource(Task t)
		throws GridException {
		
		ResourceHandle rh =
			this.resource_locator.locateAndCheckBusyLocalResource(t);

		return this.clearResource(t,rh,true,BUSY_LOCAL,true);
		
	}

	/** Clears the state of the target Resource.
	 * <p>It is expected that the Resource is idle
	 * and has already completed the target Task,
	 * because freeResource() <b>will not perform Task completion
	 * on the Resource</b>.
	 * @param t target Supplying Task
	 * @return target Resource state change time
	 * (i.e. the time at which the Resource became busy),
	 * or -1 if is_resource_alive is false
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized long freeBusySupplyingResource(Task t)
		throws GridException {
		
		ResourceHandle rh =
			this.resource_locator.locateAndCheckBusySupplyingResource(t);

		return this.clearResource(t,rh,true,BUSY_SUPPLYING,true);
		
	}

	/** Clears the state of the target Resource.
	 * <p>It is expected that the Resource is busy
	 * and the target Task has not been cancelled yet,
	 * because preemptResource() <b>will perform Task cancellation
	 * on the Resource</b>.
	 * @param t target Task
	 * @return target Resource state change time
	 * (i.e. the time at which the Resource became busy),
	 * or -1 if is_resource_alive is false
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized long preemptBusyLocalResource(Task t)
		throws GridException {
		
		ResourceHandle rh =
			this.resource_locator.locateBusyLocalResource(t);

		return this.clearResource(t,rh,false,BUSY_LOCAL,(rh != null));
		
	}

	/** Clears the state of the target Resource.
	 * <p>It is expected that the Resource is busy
	 * and the target Task has not been cancelled yet,
	 * because clearResource() <b>will perform Task cancellation
	 * on the Resource</b>.
	 * @param t target Supplying Task
	 * @return target Resource state change time
	 * (i.e. the time at which the Resource became busy),
	 * or -1 if is_resource_alive is false
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized long preemptBusySupplyingResource(Task t)
		throws GridException {
		
		ResourceHandle rh =
			this.resource_locator.locateBusySupplyingResource(t);

		return this.clearResource(t,rh,false,BUSY_SUPPLYING,(rh != null));
		
	}

	/** Clears the state of the target Resource.
	 * <p>It is expected that the Resource is busy
	 * and the target Task has not been cancelled yet,
	 * because clearResource() <b>will perform Task cancellation
	 * on the Resource</b>.
	 * @param t target Supplying Task
	 * @param rh target ResourceHandle
	 * @return target Resource state change time
	 * (i.e. the time at which the Resource became busy),
	 * or -1 if is_resource_alive is false
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized long preemptBusySupplyingResource(Task t,
														  ResourceHandle rh)
		throws GridException {
		
		return this.clearResource(t,rh,false,BUSY_SUPPLYING,true);
		
	}

	/** Clears the state of the target Resource.
	 * <p>It is expected that the Resource is idle
	 * if it has already completed the target Task,
	 * because clearResource() <b>will not perform Task completion
	 * on the Resource</b>.
	 * It is also expected that the Resource is busy
	 * if the target Task has to be cancelled,
	 * because clearResource() <b>will perform Task cancellation
	 * on the Resource</b>.
	 * @param t target Task
	 * @param rh target (Busy) Resource
	 * @param complete_task true if the goal is to complete the target Task,
	 * false if the goal is to cancel the Target Task
	 * @param state target state (legal values: BUSY_LOCAL, BUSY_SUPPLYING)
	 * @param is_busy_resource_alive true if the target (Busy) Resource
	 * may be safely contacted through the given ResourceHandle
	 * @return target Resource state change time
	 * (i.e. the time at which the Resource became busy),
	 * or -1 if is_resource_alive is false
	 * @exception GridException if a parameter is illegal
	 */
	private synchronized long clearResource(Task t,ResourceHandle rh,
											boolean complete_task,
										    ResourceState state,
											boolean is_busy_resource_alive)
		throws GridException {

		assert(threeSetsInvariant());

		if (t == null) {
			throw new GridException("clearResource(): illegal Task");
		}
		
		if (is_busy_resource_alive == false) {
			// goal is to removeResource()
			return -1;
		}
		
		// goal is to complete or cancel
		if (rh == null) {
			throw new GridException("clearResource(): "+
									"illegal ResourceHandle");
		}

		if (rh.idle() == true) {
			if (!complete_task) {
				// goal is to cancel, but Resource is idle
				this.managed_peer.logger().
					info(getStackTrace("clearResource(): Resource idle, "+
						"rh.cancelTask() must not have been called before"));
			}
			else {
				// goal is to complete, and Resource is idle as expected
				// skip
			}
		}
		else /* if (rh.idle() == false) */ {
			if (complete_task) {
				// goal is to complete, but Resource is busy
				this.managed_peer.logger().
					info(getStackTrace("clearResource(): Resource busy, "+
						 "rh.completeTask() must have been called before"));
			}
			else {
				// goal is to cancel, and Resource is busy as expected
				rh.cancelTask();
			}
		}

		long task_start_time = // get value BEFORE updating ResourceTracker
			this.resource_tracker.
				getResourceUsageData(rh).getStateChangedTime();
		
		setIdleResourceState(t,rh,state);
		
		assert(threeSetsInvariant());

		return task_start_time;
		
	}
	
	private synchronized void setIdleResourceState(Task t,ResourceHandle rh,
												   ResourceState state)
		throws GridException {
		
		assert(threeSetsInvariant());

		if (t == null) {
			throw new GridException("setIdleResourceState(): illegal Task");
		}
		if (rh == null) {
			throw new GridException("setIdleResourceState(): "+
									"illegal ResourceHandle");
		}

		WorkingSet ws =
			this.resource_tracker.updateResourceStatusToIdle(rh,state);

		this.available_power += this.power_tracker.get(rh);

		switch (state) {
		case BUSY_LOCAL:
			ws.unlockLocalInputData(t);
			this.resource_locator.untrackBusyLocalResource(t);
			break;
		case BUSY_SUPPLYING:
			ws.unlockSupplyingInputData(t);
			this.resource_locator.untrackBusySupplyingResource(t);
			break;
		case IDLE:
		default:
			throw new GridException("setIdleResourceState(): "+
									"unsupported state");
		}
		
		assert(threeSetsInvariant());

	}

	// ****************************************************************

	/** Returns the Task running on target Busy Resource.
	 * @param rh target Busy Resource
	 * @return Task running on target Busy Resource
	 */
	public synchronized Task getRunningTask(ResourceHandle rh)
		throws GridException {
		
		if (rh == null) {
			throw new GridException("getRunningTask(): illegal Resource");
		}
		
		return this.resource_tracker.getResourceUsageData(rh).runningTask();
		
	}
	
	// ****************************************************************
	
	/** Returns the WorkingSet of the target Resource,
	 * if target Resource is idle.
	 * <p>Note: does not assume that Resource is idle,
	 * but rather returns a WorkingSet only when it is idle.
	 */
	public synchronized WorkingSet
		getWorkingSetOfIdleResource(ResourceHandle rh) throws GridException {

		return this.resource_tracker.getWorkingSetOfIdleResource(rh);
		
	}

	// ****************************************************************

	/** Replicates Supplying data to Resources.
	 * <p>Note: only a subset of idle Resources is targeted
	 * by the replication process.
	 * @exception GridException
	 */
	public synchronized void replicateSupplyingData() throws GridException {

		/* Implementation notes:
		
		   When to replicate Supplying Tasks data ?
		   (i.e. when this method should be called ?)
		   After Supplying scheduling loop in Scheduler.scheduleAll()
		   
		   Replicate Supplying data, following BT+P2P Grid paper proposal,
		   by sending working sets on a fraction of idle/busy Resources
		 */
		
		if 	((this.data_replication_enabled == false) ||
			 (this.idle_replication_ratio == 0.0)) {
			return;
		}
		
		// Compute replication count
		int select_idle_r_count =
			(int)Math.floor(this.idle_replication_ratio *
							this.resource_tracker.countIdleResources());
		
		if (select_idle_r_count < 1) {
			return;
		}
		
		if (1 <2) {
				throw new GridException("replicateSupplyingData(): "+
										"currently uncertified operation");
		}
		
		// Skip Resources that will be the first to be selected
		// for Task execution (supposing they already store some input data),
		// so as to not penalize them by lengthy replication
		int skip_idle_r_count =
			this.resource_tracker.countIdleResources() - select_idle_r_count;
		Iterator<ResourceUsageData> it =
			this.resource_tracker.getIdleResourcesIterator();
		while (it.hasNext() && (skip_idle_r_count-- > 0)) {
			it.next();
		}
		// The Iterator now references the Resources subset
		// that will probably be last selected for scheduling
		
		// Replicate data
		this.supplying_data_tracker.replicateData(it,select_idle_r_count);
		
	}

	// ****************************************************************

	private synchronized boolean threeSetsInvariant() {
		
		return
			(this.power_tracker != null) &&
			(this.total_power >= 0) && (this.available_power >= 0) &&
			(this.total_power >= this.available_power);
		
	}
	
	// ****************************************************************

	private void debugging_msg_1(ResourceState state,Task t,
							 ResourceHandle rh) throws GridException {
		
		int count = -1;
		StringBuilder sb_list = new StringBuilder("unavailable information");
		switch (state) {
		case IDLE:
			count = this.resource_tracker.countIdleResources();
			sb_list = new StringBuilder();
			Iterator<ResourceUsageData> it =
				this.resource_tracker.getIdleResourcesIterator();
			while (it.hasNext()) {
				ResourceUsageData rud = it.next();
				WorkingSet ws =
					this.resource_tracker.getWorkingSetOfIdleResource(rud.rh);
				sb_list.append("\n\t");
				sb_list.append(rud.id);
				sb_list.append(" { ");
				MetaGridData[] dummy = ws.getMetaGridData();
				for (int i = 0 ; i < dummy.length ; i++) {
					sb_list.append(dummy[i].id);
					sb_list.append(" ");
				}
				sb_list.append(" } ");
			}
		break;
		case BUSY_LOCAL:
			count = this.resource_tracker.countBusyLocalResources();
			break;
		case BUSY_SUPPLYING: count =
			this.resource_tracker.countBusySupplyingResources();
			break;
		}

		System.out.println("Supplier "+this.managed_peer.id()+".rms { "+
							t.getInputDataSetID()+"}:\n\t"+
							"found Storage-optimal "+state+" Resource "+
							rh.id()+" among "+count+" Resources "+
							sb_list.toString()+"\n");
		
	}

	private void debugging_msg_2(ResourceState state,Task t,
							 	 ResourceHandle rh) throws GridException {
		
		StringBuilder sb_list = new StringBuilder();
		Iterator<ResourceUsageData> it =
					this.resource_tracker.getIdleResourcesIterator();
		while (it.hasNext()) {
			ResourceUsageData rud = it.next();
			WorkingSet ws =
				this.resource_tracker.getWorkingSetOfIdleResource(rud.rh);
			sb_list.append("\n\t");
			sb_list.append(rud.id);
			sb_list.append(" { ");
			MetaGridData[] dummy = ws.getMetaGridData();
			for (int i = 0 ; i < dummy.length ; i++) {
				sb_list.append(dummy[i].id);
				sb_list.append(" ");
			}
			sb_list.append(" } ");
		}

		System.out.println("Supplier "+this.managed_peer.id()+".rms { "+
						   t.getInputDataSetID()+" }:\n\t"+
						   "found arbitrary "+state+" Resource "+
						   rh.id()+sb_list.toString()+"\n");

	}

	// ****************************************************************

	private final Peer managed_peer;
	private final ResourceTracker resource_tracker;
	private final ResourceLocator resource_locator;
	private 	  double cached_utilization_value;

	private final ResourceDataTracker        resource_data_tracker;
	private final PeerConsumptionDataTracker consumption_data_tracker;
	private final PeerSupplyingDataTracker   supplying_data_tracker;
	private final boolean data_present_enabled;
	private final boolean data_replication_enabled;
	private final float idle_replication_ratio;
	private final boolean psufferage_enabled;

	private final Map<ResourceHandle,Integer> power_tracker;
	private       int total_power;
	private       int available_power;

	// ****************************************************************

}

