/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ResourceTracker.java

package lbg.peer.core.rms;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.*;
import static lbg.peer.core.rms.ResourceState.*;
import lbg.resource.service.*;
import lbg.task.*;

import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.atomic.AtomicInteger;

/** A ResourceTracker tracks Resources for the ResourceManager that uses it.
 * @author Cyril Briquet
 */
public class ResourceTracker {

	public ResourceTracker(Peer managed_peer,
						   ResourceDataTracker data_tracker)
		throws GridException {
		
		if (managed_peer == null) {
			throw new GridException("ResourceTracker(): illegal Peer");
		}
		if (data_tracker == null) {
			throw new GridException("ResourceTracker(): illegal "+
									"Resource-level DataTracker");
		}

		this.managed_peer = managed_peer;
		this.data_tracker = data_tracker;
		
		this.rs_count = 0;
		this.rs_idle =
			new TreeMap<ResourceHandle,ResourceUsageData>();
		this.rs_idle_s = new AtomicInteger(0);
		this.rs_busy_local =
			new TreeMap<ResourceHandle,ResourceUsageData>();
		this.rs_busy_local_s = new AtomicInteger(0);
		this.rs_busy_supplying =
			new TreeMap<ResourceHandle,ResourceUsageData>();
		this.rs_busy_supplying_s = new AtomicInteger(0);
		this.rs_all =
			new TreeMap<ResourceHandle,ResourceState>();

	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ResourceTracker ");
		sb.append(this.managed_peer.id());
		sb.append("]");
		
		sb.append("\n  ->  ");
		sb.append(this.rs_count);
		sb.append(" Resources {");
		sb.append(" idle=");
		sb.append(this.rs_idle.size());
		sb.append(", busy Local=");
		sb.append(this.rs_busy_local.size());
		sb.append(", busy Supplying=");
		sb.append(this.rs_busy_supplying.size());
		sb.append(" }");

		sb.append("\n  ->  utilization { ");
		long total_idle_time = 0;
		long total_busy_local_time = 0;
		long total_busy_supplying_time = 0;
		for (ResourceUsageData rud : this.rs_idle.values()) {
			total_idle_time += rud.getTotalIdleTime();
			total_busy_local_time += rud.getTotalBusyLocalTime();
			total_busy_supplying_time += rud.getTotalBusySupplyingTime();
		}
		for (ResourceUsageData rud : this.rs_busy_local.values()) {
			total_idle_time += rud.getTotalIdleTime();
			total_busy_local_time += rud.getTotalBusyLocalTime();
			total_busy_supplying_time += rud.getTotalBusySupplyingTime();
		}
		for (ResourceUsageData rud : this.rs_busy_supplying.values()) {
			total_idle_time += rud.getTotalIdleTime();
			total_busy_local_time += rud.getTotalBusyLocalTime();
			total_busy_supplying_time += rud.getTotalBusySupplyingTime();
		}
		long total_time = total_idle_time +
			total_busy_local_time + total_busy_supplying_time;
		sb.append(CommonMath.div_label(total_idle_time,total_time,100.0));
		sb.append("% idle, ");
		sb.append(CommonMath.div_label(total_busy_local_time,
									   total_time,100.0));
		sb.append("% busy Local, ");
		sb.append(CommonMath.div_label(total_busy_supplying_time,
									   total_time,100.0));
		sb.append("% busy Supplying");
		sb.append(" }");
		
		/* displaying each Resource is not really useful, but code's ready
		sb.append("\n  ->  ");
		sb.append(this.rs_idle.size());
		sb.append(" idle Resources");
		for (ResourceHandle rh : this.rs_idle.keySet()) {
			sb.append("\n  ->  ");
			sb.append(rh);
		}
		sb.append("\n  ->  ");
		sb.append(this.rs_busy_local.size());
		sb.append(" busy Local Resources");
		for (ResourceHandle rh : this.rs_busy_local.keySet()) {
			sb.append("\n  ->  ");
			sb.append(rh);
		}
		sb.append("\n  ->  ");
		sb.append(this.rs_busy_supplying.size());
		sb.append(" busy Supplying Resources");
		for (ResourceHandle rh : this.rs_busy_supplying.keySet()) {
			sb.append("\n  ->  ");
			sb.append(rh);
		}*/

		return sb.toString();

	}
	
	// ****************************************************************

	public int countResources() {
		
		return countIdleResources()+
			countBusyLocalResources()+countBusySupplyingResources();

	}
	
	public int countIdleResources() {

		return this.rs_idle_s.get();
		
		/* BAD BIG BUG -- rs_busy_local.size()

		What's the net effect of this bug ?
		Longer Supplying Tasks queues, leading to abnormal memory consumption
		and, ultimately, to thrashing ...
		
		800 jobs scenario
		LocalScheduler
			208 s, memory peak ~   20 Mo
		SupplyingNonpreemptiveScheduler + NoBarteringNegotiator + NoEvalAccountant:
			211 s, memory peak ~   20 Mo
		SupplyingNonpreemptiveScheduler + ClassicFavorsNegotiator +
			OGTimeAccountant:
			419 s, memory peak ~  410+ Mo
			312 s, memory peak ~  410  Mo -- null rather than clear
			412 s, memory peak ~  348  Mo -- System.gc()
		SupplyingPreemptiveScheduler + ClassicFavorsNegotiator + OGTimeAccountant:
			597 s, memory peak ~ 1352  Mo
		*/

	}

	public int countBusyLocalResources() {

		return this.rs_busy_local_s.get();
		
	}

	public int countBusySupplyingResources() {

		return this.rs_busy_supplying_s.get();
		
	}

	public boolean hasIdleResource() {

		return countIdleResources() > 0;
		
	}

	public boolean hasBusyLocalResource() {

		return countBusyLocalResources() > 0;                                   
			
	}

	public boolean hasBusySupplyingResource() {

		return countBusySupplyingResources() > 0;
		
	}

	/** Returns the Peer utilization, i.e. total busy time of
	 * all currently registered Resources,
	 * divided by their total registration time.
	 * @return Peer utilization, in [0.0..1.0]
	 */
	public synchronized double utilization() throws GridException {
	
		long total_idle_time = 0;
		long total_busy_local_time = 0;
		long total_busy_supplying_time = 0;
		
		for (ResourceUsageData rud : this.rs_idle.values()) {
			total_idle_time += rud.getTotalIdleTime();
			total_busy_local_time += rud.getTotalBusyLocalTime();
			total_busy_supplying_time += rud.getTotalBusySupplyingTime();
		}
		for (ResourceUsageData rud : this.rs_busy_local.values()) {
			total_idle_time += rud.getTotalIdleTime();
			total_busy_local_time += rud.getTotalBusyLocalTime();
			total_busy_supplying_time += rud.getTotalBusySupplyingTime();
		}
		for (ResourceUsageData rud : this.rs_busy_supplying.values()) {
			total_idle_time += rud.getTotalIdleTime();
			total_busy_local_time += rud.getTotalBusyLocalTime();
			total_busy_supplying_time += rud.getTotalBusySupplyingTime();
		}
		
		final double total_busy_time =
			total_busy_local_time + total_busy_supplying_time;
		final double total_time = total_idle_time + total_busy_time;

		final double utilization = total_busy_time / total_time;
		
		if ((utilization < 0.0) || (utilization > 1.0)) {
			throw new GridException("utilization(): corrupted Resource usage "+
									"data (utilization="+utilization+")");
		}
		
		return utilization;
		
	}
	
	// ****************************************************************

	public synchronized ResourceState
		getResourceState(ResourceHandle rh) throws GridException {

		if (rh == null) {
			throw new GridException("getResourceState(): "+
									"illegal Resource Handle");
		}
		
		ResourceState state = this.rs_all.get(rh);
		
		if (state == null) {
			throw new GridException("getResourceState(): "+
									"corrupted state data");
		}
		
		return state;
		
	}
	
	public synchronized ResourceUsageData
		getResourceUsageData(ResourceHandle rh) throws GridException {
		
		if (rh == null) {
			throw new GridException("getResourceUsageData(): "+
									"illegal Resource Handle");
		}
		
		ResourceState state = this.getResourceState(rh);
		TreeMap<ResourceHandle,ResourceUsageData> target_rs =
			null;
		switch (state) {
		case IDLE:
			target_rs = this.rs_idle;
			break;
		case BUSY_LOCAL:
			target_rs = this.rs_busy_local;
			break;
		case BUSY_SUPPLYING:
			target_rs = this.rs_busy_supplying;
			break;
		default:
			throw new GridException("getResourceUsageData(): "+
									"unsupported state");
		}
		
		ResourceUsageData rud = target_rs.get(rh);
		
		if (rud == null) {
			throw new GridException("getResourceUsageData(): "+
									"corrupted "+state+" Resource set");
		}
		
		return rud;
		
	}
	
	// ****************************************************************

	/** Starts the tracking of the target Resource.
	 * @param rh target Resource
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized void trackResource(ResourceHandle rh)
		throws GridException {
		
		assert(areResourceSetsConsistentInvariant());
		
		if (rh == null) {
			throw new GridException("trackResource(): illegal Resource Handle");
		}
		
		// Add Resource to idle Resource set
		ResourceUsageData rud =
			new ResourceUsageData(this.managed_peer,rh,this.data_tracker);
		if (this.rs_idle.put(rh,rud) != null) {
			throw new GridException("trackResource(): "+
									"Resource already tracked");
		}
		this.rs_idle_s.incrementAndGet();
		
		// Add Resource to global Resource set
		if (this.rs_all.put(rh,IDLE) != null) {
			throw new GridException("trackResource(): "+
									"Resource state already tracked");
		}
		this.rs_count++;

		assert(areResourceSetsConsistentInvariant());
		
	}
	
	/** Stops the tracking of the target Resource.
	 * @param rh target Resource
	 * @return Resource usage data)
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized ResourceUsageData untrackResource(ResourceHandle rh)
		throws GridException {
		
		assert(areResourceSetsConsistentInvariant());
		
		if (rh == null) {
			throw new GridException("untrackResource(): "+
									"illegal Resource Handle");
		}

		// Identify appropriate Resource set
		ResourceState state = this.getResourceState(rh);
		TreeMap<ResourceHandle,ResourceUsageData> target_rs = null;
		AtomicInteger target_rs_s;
		switch (state) {
		case IDLE:
			target_rs = this.rs_idle;
			target_rs_s = this.rs_idle_s;
			break;
		case BUSY_LOCAL:
			target_rs = this.rs_busy_local;
			target_rs_s = this.rs_busy_local_s;
			break;
		case BUSY_SUPPLYING:
			target_rs = this.rs_busy_supplying;
			target_rs_s = this.rs_busy_supplying_s;
			break;
		default:
			throw new GridException("untrackResource(): unsupported state");
		}
		
		// Remove Resource from appropriate Resource set
		ResourceUsageData rud = target_rs.remove(rh);
		if (rud == null) {
			throw new GridException("untrackResource(): "+
									"corrupted "+state+" Resource set");
		}
		target_rs_s.decrementAndGet();
		
		// Remove Resource from global Resource set
		if (this.rs_all.remove(rh) != state) {
			throw new GridException("untrackResource(): "+
									"corrupted Resource state set");
		}
		this.rs_count--;
		
		assert(areResourceSetsConsistentInvariant());
		
		return rud;
		
	}
	
	// ****************************************************************

	/** Returns an Iterator over idle Resources usage data.
	 * @return Iterator over idle Resources usage data
	 */
	public synchronized Iterator<ResourceUsageData> getIdleResourcesIterator() {

		return this.rs_idle.values().iterator();
		
	}
	
	/** Returns the WorkingSet of the target Resource,
	 * if target Resource is idle.
	 * <p>Note: does not assume that Resource is idle,
	 * but rather returns a WorkingSet only when it is idle.
	 * @param rh target Resource
	 * @return WorkingSet of target Resource if target Resource is idle,
	 * or null if target Resource is not idle
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized WorkingSet
		getWorkingSetOfIdleResource(ResourceHandle rh)
		throws GridException {

		if (rh == null) {
			throw new GridException("getWorkingSetOfIdleResoucre(): "+
									"illegal ResourceHandle");
		}

		if (this.getResourceState(rh) != IDLE) {
			return null;
		}

		ResourceUsageData rud = this.rs_idle.get(rh);
		if (rud == null) {
			throw new GridException("getWorkingSetOfIdleResource(): "+
									"corrupted idle Resource set");
		}
		
		return rud.working_set;
		
	}
	
	// ****************************************************************

	/** Returns an arbitrary (i.e. the first) idle Resource.
	 * @return arbitrary idle Resource
	 * @deprecated
	 */
	@Deprecated
	public synchronized ResourceHandle getFirstIdleResource() {
		
		return this.rs_idle.firstKey();
		
	}
	
	/** Returns an arbitrary (i.e. the first) busy local Resource.
	 * @return arbitrary busy local Resource
	 * @deprecated
	 */
	@Deprecated
	public synchronized ResourceHandle getFirstBusyLocalResource() {
		
		return this.rs_busy_local.firstKey();
		
	}

	/** Returns an arbitrary (i.e. the first) busy supplying Resource.
	 * @return arbitrary busy supplying Resource
	 * @deprecated
	 */
	@Deprecated
	public synchronized ResourceHandle getFirstBusySupplyingResource() {
		
		return this.rs_busy_supplying.firstKey();
		
	}

	/** Returns a random idle Resource.
	 * @return random idle Resource
	 */
	public synchronized ResourceHandle getRandomIdleResource()
		throws GridException {
		
		return locateRandomResource(this.rs_idle);
		
	}
	
	/** Returns a random busy local Resource.
	 * @return random busy local Resource
	 */
	public synchronized ResourceHandle getRandomBusyLocalResource()
		throws GridException {
		
		return locateRandomResource(this.rs_busy_local);
		
	}

	/** Returns a random busy supplying Resource.
	 * @return random busy supplying Resource
	 */
	public synchronized ResourceHandle getRandomBusySupplyingResource()
		throws GridException {
		
		return locateRandomResource(this.rs_busy_supplying);
		
	}
	
	// ****************************************************************

	/** Updates target Resource status from idle
	 * to busy local or busy supplying, depending on the value of target state.
	 * <p> rs_idle = rs_idle \ { rh } ; rs_busy_x = rs_busy_x U { rh } ;
	 * @param rh target Resource
	 * @param state target state (legal values: BUSY_LOCAL, BUSY_SUPPLYING)
	 * @param t Task that is going to be run on target Resource
	 * @return Resource WorkingSet
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized
		WorkingSet updateResourceStatusToBusy(ResourceHandle rh,
											  ResourceState state,
											  Task t)
		throws GridException {

		assert(areResourceSetsConsistentInvariant());
		
		if (rh == null) {
			throw new GridException("updateResourceStatusToBusy(): "+
									"illegal Resource Handle");
		}

		TreeMap<ResourceHandle,ResourceUsageData> target_rs = null;
		AtomicInteger target_rs_s;
		switch (state) {
		case BUSY_LOCAL:
			target_rs = this.rs_busy_local;
			target_rs_s = this.rs_busy_local_s;
			break;
		case BUSY_SUPPLYING:
			target_rs = this.rs_busy_supplying;
			target_rs_s = this.rs_busy_supplying_s;
			break;
		case IDLE:
		default:
			throw new GridException("updateResourceStatusToBusy(): "+
									"unsupported state");
		}
		
		ResourceUsageData rud = this.rs_idle.remove(rh);
		if (rud == null) {
			throw new GridException("updateResourceStatusToBusy(): "+
									"corrupted idle Resource set");
		}
		rs_idle_s.decrementAndGet();
		
		if (target_rs.put(rh,rud) != null) {
			throw new GridException("updateResourceStatusToBusy(): "+
									"corrupted "+state+" Resource set");
		}
		target_rs_s.incrementAndGet();
		
		if (this.rs_all.put(rh,state) != IDLE) {
			throw new GridException("updateResourceStatusToBusy(): "+
									"corrupted Resource state set");
		}

		// Update Resource usage data
		rud.updateTotalIdleTime(state,t);
		
		assert(areResourceSetsConsistentInvariant());
		
		return rud.working_set;
	
	}

	/** Updates target Resource status from busy local or busy supplying,
	 * depending on the value of target state, to idle.
	 * <p> rs_busy_x = rs_busy_x \ { rh } ; rs_idle = rs_idle U { rh } ;
	 * @param rh target Resource
	 * @param state target state (legal values: BUSY_LOCAL, BUSY_SUPPLYING)
	 * @return Resource WorkingSet
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized
		WorkingSet updateResourceStatusToIdle(ResourceHandle rh,
											  ResourceState state)
		throws GridException {

		assert(areResourceSetsConsistentInvariant());
		
		if (rh == null) {
			throw new GridException("updateResourceStatusToIdle(): "+
									"illegal Resource Handle");
		}

		TreeMap<ResourceHandle,ResourceUsageData> target_rs = null;
		AtomicInteger target_rs_s;
		switch (state) {
		case BUSY_LOCAL:
			target_rs = this.rs_busy_local;
			target_rs_s = this.rs_busy_local_s;
			break;
		case BUSY_SUPPLYING:
			target_rs = this.rs_busy_supplying;
			target_rs_s = this.rs_busy_supplying_s;
			break;
		case IDLE:
		default:
			throw new GridException("updateResourceStatusToIdle(): "+
									"unsupported state");
		}
		
		ResourceUsageData rud = target_rs.remove(rh);
		if (rud == null) {
			throw new GridException("updateResourceStatusToIdle(): "+
									"corrupted idle Resource set");
		}
		target_rs_s.decrementAndGet();
		
		if (this.rs_idle.put(rh,rud) != null) {
			throw new GridException("updateResourceStatusToIdle(): "+
									"corrupted "+state+" Resource set");
		}
		this.rs_idle_s.incrementAndGet();
		
		if (this.rs_all.put(rh,IDLE) != state) {
			throw new GridException("updateResourceStatusToIdle(): "+
									"corrupted Resource state set");
		}
		
		// Update Resource usage data
		switch (state) {
		case BUSY_LOCAL:
			rud.updateTotalBusyLocalTime();
			break;
		case BUSY_SUPPLYING:
			rud.updateTotalBusySupplyingTime();
			break;
		case IDLE:
		default:
			throw new GridException("updateResourceStatusToIdle(): "+
									"unsupported state");
		}

		assert(areResourceSetsConsistentInvariant());
		
		return rud.working_set;
		
	}

	
	// ****************************************************************

	private synchronized ResourceHandle
		locateRandomResource(Map<ResourceHandle,ResourceUsageData> ts)
		throws GridException {
		
		if (ts == null) {
			throw new GridException("locateRandomResource(): "+
									"illegal Resource set");
		}
		
		int key_idx = CommonMath.sampleUniformDist(0,ts.size()-1);
		assert((key_idx >= 0) && (key_idx < ts.size()));

		Iterator<ResourceHandle> it = ts.keySet().iterator();
		for (int i = 0 ; i < key_idx ; ++i) {
			it.next();
		}
		
		return it.next();
		
	}
	
	private synchronized boolean areResourceSetsConsistentInvariant() {
		
		boolean is_rs_count_consistent = (this.rs_count >= 0);

		boolean are_rs_sets_consistent = 
			this.rs_count == (this.rs_idle.size() +
			 this.rs_busy_local.size() + this.rs_busy_supplying.size());
		
		boolean are_rs_states_set_consistent = 
			this.rs_count == this.rs_all.size();
			
		boolean are_rs_sizes_consistent =
			(this.rs_idle.size() == this.rs_idle_s.get()) &&
			(this.rs_busy_local.size() == this.rs_busy_local_s.get()) &&
			(this.rs_busy_supplying.size() == this.rs_busy_supplying_s.get());
		
		return is_rs_count_consistent && are_rs_sets_consistent &&
			are_rs_states_set_consistent && are_rs_sizes_consistent;
		
	}

	// ****************************************************************

	private final Peer managed_peer;
	private final ResourceDataTracker data_tracker;

	private final TreeMap<ResourceHandle,ResourceUsageData> rs_idle;
	private final AtomicInteger rs_idle_s;
	private final TreeMap<ResourceHandle,ResourceUsageData> rs_busy_local;
	private final AtomicInteger rs_busy_local_s;
	private final TreeMap<ResourceHandle,ResourceUsageData> rs_busy_supplying;
	private final AtomicInteger rs_busy_supplying_s;
	private final Map<ResourceHandle,ResourceState> rs_all;
	private       int rs_count;

	// ****************************************************************

}

