/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ResourceUsageData.java

package lbg.peer.core.rms;

import lbg.*;
import lbg.common.exception.*;
import static lbg.peer.core.rms.ResourceState.*;
import lbg.peer.*;
import lbg.resource.service.*;
import lbg.task.*;

/** Helper class of ResourceManager.
 * Stores usage statistics and WorkingSet of target Resource.
 * @author Cyril Briquet
 */
public class ResourceUsageData {
	
	public ResourceUsageData(Peer managed_peer,ResourceHandle rh,
							 ResourceDataTracker data_tracker)
		throws GridException {

		if (managed_peer == null) {
			throw new GridException("ResourceUsageData(): illegal Peer");
		}
		if (rh == null) {
			throw new GridException("ResourceUsageData(): "+
									"illegal ResourceHandle");
		}
		if (data_tracker == null) {
			throw new GridException("ResourceUsageData(): illegal "+
									"Resource-level DataTracker");
		}
		
		this.id = rh.id();
		this.rh = rh;
		this.state_changed_time = LightweightBarteringGrid.getTime();
		this.idle_time = 0;
		this.busy_local_time = 0;
		this.busy_supplying_time = 0;
		
		this.working_set = new WorkingSet(managed_peer,rh,data_tracker);
		this.running_task = null;
		this.state = IDLE;

	}
		
	public synchronized String toString() {
			
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ResourceUsageData]");
		sb.append(" status=");
		sb.append(this.state);
		sb.append(" state change time=");
		sb.append(this.state_changed_time);
		sb.append(", total idle time=");
		sb.append(this.idle_time);
		sb.append(", total busy Local time=");
		sb.append(this.busy_local_time);
		sb.append(", total busy Supplying time=");
		sb.append(this.busy_supplying_time);
		
		sb.append("\n  ->  ");
		sb.append(this.working_set);
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized ResourceState state() {
		
		return this.state;
		
	}
	
	public synchronized Task runningTask() throws GridException {
	
		if (this.state == IDLE) {
			throw new GridException("runningTask(): Resource is idle");
		}
		if (this.running_task == null) {
			throw new GridException("runningTask(): no running Task");
		}
		
		return this.running_task;
		
	}
	
	// ****************************************************************

	/** Returns state change time of the monitored Resource.
	 * @return state change time of the monitored Resource
	 */
	public synchronized long getStateChangedTime() {
		
		return this.state_changed_time;
		
	}

	/** Returns total idle time of the monitored Resource.
	 * @return total idle time of the monitored Resource
	 */
	public synchronized long getTotalIdleTime() {
		
		return this.idle_time;
		
	}

	/** Returns total busy Local time of the monitored Resource.
	 * @return total busy Local time of the monitored Resource
	 */
	public synchronized long getTotalBusyLocalTime() {
		
		return this.busy_local_time;
		
	}

	/** Returns total busy Supplying time of the monitored Resource.
	 * @return total busy Supplying time of the monitored Resource
	 */
	public synchronized long getTotalBusySupplyingTime() {
		
		return this.busy_supplying_time;
		
	}

	// ****************************************************************

	/** Updates total idle time of the monitored Resource.
	 * @param state new state monitored Resource
	 * @param t running Task
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized void updateTotalIdleTime(ResourceState state,
												 Task t)
		throws GridException {
		
		if (t == null) {
			throw new GridException("updateTotalIdleTime(): illegal Task");
		}
		switch (state) {
		case BUSY_LOCAL:
		case BUSY_SUPPLYING:
			// skip
			break;
		case IDLE:
		default:
			throw new GridException("updateTotalIdleTime(): "+
									"unsupported state");
		}
		
		long previous_time = this.state_changed_time;
		this.state_changed_time = LightweightBarteringGrid.getTime();
		
		this.idle_time += (this.state_changed_time-previous_time);
		
		this.running_task = t;
		this.state = state;

	}
		
	/** Updates total busy Local time of the monitored Resource.
	 */
	public synchronized void updateTotalBusyLocalTime()
		throws GridException {
			
		long previous_time = this.state_changed_time;
		this.state_changed_time = LightweightBarteringGrid.getTime();
		
		this.busy_local_time += (this.state_changed_time-previous_time);
		
		this.running_task = null;
		this.state = IDLE;
		
	}

	/** Updates total busy Supplying time of the monitored Resource.
	 */
	public synchronized void updateTotalBusySupplyingTime()
		throws GridException {
			
		long previous_time = this.state_changed_time;
		this.state_changed_time = LightweightBarteringGrid.getTime();
			
		this.busy_supplying_time += (this.state_changed_time-previous_time);

		this.running_task = null;
		this.state = IDLE;
		
	}

	// ****************************************************************

	public final WorkingSet working_set;

	// ****************************************************************

	public final String id;
	public final ResourceHandle rh;
	private long state_changed_time;
	private long idle_time;
	private long busy_local_time;
	private long busy_supplying_time;
	private Task running_task;
	private ResourceState state;
	
}

