/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// WorkingSet.java

package lbg.peer.core.rms;

import lbg.common.exception.*;
import lbg.peer.*;
import lbg.protocol.data.*;
import lbg.resource.service.*;
import lbg.task.*;
import java.util.*;

/** Resource Working Set.
 * <p>It is guarantueed that the Working Set always contains and locks
 * the input data of running Tasks.
 * <p>It is guarantueed that the Working Set size is never larger
 * than the cache size of the Resource that is managed.
 * <p>When a data id is added to, or cleared from, the WorkingSet,
 * the ResourceManager ResourceDataTracker is updated.
 * @author Cyril Briquet
 */
public class WorkingSet {

	/* NOTE: in ResourceManager, the WorkingSet of a given Resource
	 *       MUST be updated after each call
	 *            to ResourceTracker.updateResourceStatus*()
	 */

	/** Constructs a new WorkingSet.
	 * @param managed_peer managed Peer
	 * @param rh managed Resource Handle
	 * @param data_tracker managed Peer Data Tracker
	 * @exception GridException if a parameter is illegal
	 */
	public WorkingSet(Peer managed_peer,
					  ResourceHandle rh,ResourceDataTracker data_tracker)
		throws GridException {

		if (managed_peer == null) {
			throw new GridException("WorkingSet(): illegal Peer");
		}
		if (rh == null) {
			throw new GridException("WorkingSet(): "+
									"illegal ResourceHandle");
		}
		if (data_tracker == null) {
			throw new GridException("WorkingSet(): illegal Resource-level "+
									"DataTracker");
		}

		this.peer_prefix = managed_peer.id()+".";
		this.capacity = rh.cacheCapacity();
		if (this.capacity < 1) {
			throw new GridException("WorkingSet(): illegal capacity");
		}
		this.data_tracker = data_tracker;
		this.rh = rh;
		
		this.local_db            = new TreeSet<MetaGridData>();
		this.supplying_db        = new TreeSet<MetaGridData>();
		this.local_locked_db     = new TreeSet<MetaGridData>();
		this.supplying_locked_db = new TreeSet<MetaGridData>();
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[WorkingSet] ");
		sb.append((this.local_db.size()+this.supplying_db.size()));
		sb.append("/");
		sb.append(this.capacity);
		sb.append(" stored data id");

		sb.append("\n  ->  Local data id DB = { ");
		int s = local_db.size();
		for (MetaGridData data : local_db) {
			String data_id = data.id;
			sb.append(data_id);
			if (this.local_locked_db.contains(data)) {
				sb.append(" (locked)");
			}
			if (--s > 0) sb.append(", ");
		}
		sb.append(" } ");
		
		sb.append("\n  ->  Supplying data id DB = { ");
		s = supplying_db.size();
		for (MetaGridData data : supplying_db) {
			String data_id = data.id;
			sb.append(data_id);
			if (this.supplying_locked_db.contains(data)) {
				sb.append(" (locked)");
			}
			if (--s > 0) sb.append(", ");
		}
		sb.append(" } ");

		return sb.toString();

	}
	
	// ****************************************************************

	/** Managed Resource Handle.
	 */
	public synchronized ResourceHandle managedResource() {

		return this.rh;
	
	}
	
	/** Returns the nominal data storage nominal capacity.
	 */
	public synchronized int nominalCapacity() {
		
		return this.capacity;
		
	}
	
	/** Returns the number of all stored data.
	 */
	public synchronized int size() {
		
		return this.local_db.size() + this.supplying_db.size();
		
	}

	/** Returns the number of stored Local data.
	 */
	public synchronized int localSize() {
		
		return this.local_db.size();
		
	}

	/** Returns the number of stored Supplying data.
	 */
	public synchronized int supplyingSize() {
		
		return this.supplying_db.size();
		
	}
	
	
	/** Returns the number of locked data (whether Local or Supplying).
	 */
	public synchronized int lockedSize() {

		return this.local_locked_db.size() + this.supplying_locked_db.size();
		
	}

	/** Returns the nominal data storage  available capacity
	 * (i.e. ws.capacity() - ws.lockedSize()).
	 */
	public synchronized int availableCapacity() {
		
		return this.capacity -
			(this.local_locked_db.size() + this.supplying_locked_db.size());
		
	}

	// ****************************************************************

	public synchronized boolean containsData(MetaGridData data)
		throws GridException {
		
		if (data == null) {
			throw new GridException("containsData(): illegal MetaGridData");
		}
		
		return this.local_db.contains(data) || this.supplying_db.contains(data);

	}

	public synchronized boolean containsLocalData(MetaGridData data)
		throws GridException {
		
		if (data == null) {
			throw new GridException("containsLocalData(): "+
									"illegal MetaGridData");
		}
		
		return this.local_db.contains(data);

	}

	public synchronized boolean containsSupplyingData(MetaGridData data)
		throws GridException {
		
		if (data == null) {
			throw new GridException("containsSupplyingData(): "+
									"illegal MetaGridData");
		}
		
		return this.supplying_db.contains(data);

	}

	// ****************************************************************

	public synchronized String[] getDataId() throws GridException {

		String[] dummy =
			new String[this.local_db.size()+this.supplying_db.size()];
		int i = 0;
		for (MetaGridData data : this.local_db) {
			dummy[i++] = data.id;
		}
		for (MetaGridData data : this.supplying_db) {
			dummy[i++] = data.id;
		}
		if (i != dummy.length) {
			throw new GridException("getData(): corrupted data");
		}

		return dummy;

	}
	
	public synchronized String[] getLocalDataId() throws GridException {

		String[] dummy = new String[this.local_db.size()];
		int i = 0;
		for (MetaGridData data : this.local_db) {
			dummy[i++] = data.id;
		}

		return dummy;

	}

	public synchronized String[] getSupplyingDataId() throws GridException {

		String[] dummy = new String[this.supplying_db.size()];
		int i = 0;
		for (MetaGridData data : this.supplying_db) {
			dummy[i++] = data.id;
		}

		return dummy;

	}

	// ****************************************************************

	/** Returns the GridData stored in the WorkingSet,
	 * so as to communicate them to the managed Resource.
	 */
	public synchronized GridData[][] getData() throws GridException {

		GridData[][] dummy =
			new GridData[this.local_db.size()+this.supplying_db.size()][];

		int i = 0;
		for (MetaGridData data : this.local_db) {
			dummy[i++] = data.grid_data;
		}
		for (MetaGridData data : this.supplying_db) {
			dummy[i++] = data.grid_data;
		}
		if (i != dummy.length) {
			throw new GridException("getData(): corrupted data");
		}

		return dummy;

	}

	public synchronized GridData[][] getLocalData() throws GridException {

		GridData[][] dummy = new GridData[this.local_db.size()][];
		int i = 0;
		for (MetaGridData data : this.local_db) {
			dummy[i++] = data.grid_data;
		}

		return dummy;

	}

	public synchronized GridData[][] getSupplyingData() throws GridException {

		GridData[][] dummy = new GridData[this.supplying_db.size()][];
		int i = 0;
		for (MetaGridData data : this.supplying_db) {
			dummy[i++] = data.grid_data;
		}

		return dummy;

	}

	public synchronized MetaGridData[] getMetaGridData() throws GridException {

		MetaGridData[] dummy =
			new MetaGridData[this.local_db.size()+this.supplying_db.size()];
		int i = 0;
		for (MetaGridData data : this.local_db) {
			dummy[i++] = data;
		}
		for (MetaGridData data : this.supplying_db) {
			dummy[i++] = data;
		}
		if (i != dummy.length) {
			throw new GridException("getMetaGridData(): corrupted data");
		}

		return dummy;

	}

	public synchronized MetaGridData[] getLocalGridMetaData()
		throws GridException {

		MetaGridData[] dummy = new MetaGridData[this.local_db.size()];
		int i = 0;
		for (MetaGridData data : this.local_db) {
			dummy[i++] = data;
		}

		return dummy;

	}

	public synchronized MetaGridData[] getSupplyingMetaGridData()
		throws GridException{

		MetaGridData[] dummy = new MetaGridData[this.supplying_db.size()];
		int i = 0;
		for (MetaGridData data : this.supplying_db) {
			dummy[i++] = data;
		}

		return dummy;

	}

	// ****************************************************************

	/**
	 * <p>Note: the Resource-level DataTracker is updated by this operation
	 */
	public synchronized void clearAllUnlockedData() throws GridException {

		clearLocalUnlockedData();
		clearSupplyingUnlockedData();
	
	}

	/**
	 * <p>Note: the Resource-level DataTracker is updated by this operation
	 */
	public synchronized void clearLocalUnlockedData() throws GridException {

		if (this.local_db.size() > 0) {
			for (MetaGridData data : this.local_db) {
				this.data_tracker.untrackData(data.id,this.rh);
			}
			this.local_db.clear();
			this.local_db.addAll(this.local_locked_db);
		}
		
		assert(checkSizeInvariant());
		
	}

	/**
	 * <p>Note: the Resource-level DataTracker is updated by this operation
	 */
	public synchronized void clearSupplyingUnlockedData() throws GridException {

		if (this.supplying_db.size() > 0) {
			for (MetaGridData data : this.supplying_db) {
				this.data_tracker.untrackData(data.id,this.rh);
			}
			this.supplying_db.clear();
			this.supplying_db.addAll(this.supplying_locked_db);
		}
		
		assert(checkSizeInvariant());
		
	}

	// ****************************************************************

	/** Clears WorkingSet, then add-and-lock target Local data.
	 */
	public synchronized void clearAddAndLockLocalInputData(Task t)
		throws GridException {

		if (local_locked_db.size() != 0) {
			throw new GridException("clearAddAndLockLocalInputData(): "+
									"some Local data is already locked");
		}
		if (supplying_locked_db.size() != 0) {
			throw new GridException("clearAddAndLockLocalInputData(): "+
									"some Supplying data is already locked");
		}
		if (t == null) {
			throw new GridException("clearAddAndLockLocalInputData(): "+
									"illegal Task");
		}

		MetaGridData[] data_to_lock = t.getInputMetaGridDataSet();

		if (data_to_lock.length > this.capacity) {
			throw new GridException("clearAddAndLockLocalInputData(): "+
									"WorkingSet overflow");
		}
		
		clearAllUnlockedData();
		
		for (MetaGridData data : data_to_lock) {
			this.local_db.add(data);
			this.local_locked_db.add(data);
			this.data_tracker.trackData(data.id,this.rh);
		}

		assert(checkSizeInvariant());
		
	}

	/** Clears WorkingSet, then add-and-lock target Supplying data.
	 */
	public synchronized void clearAddAndLockSupplyingInputData(Task t)
		throws GridException {
		
		if (local_locked_db.size() != 0) {
			throw new GridException("clearAddAndLockSupplyingInputData(): "+
									"some Local data is already locked");
		}
		if (supplying_locked_db.size() != 0) {
			throw new GridException("clearAddAndLockSupplyingInputData(): "+
									"some Supplying data is already locked");
		}
		if (t == null) {
			throw new GridException("clearAddAndLockSupplyingInputData(): "+
									"illegal Task");
		}

		MetaGridData[] data_to_lock = t.getInputMetaGridDataSet();

		if (data_to_lock.length > this.capacity) {
			throw new GridException("clearAddAndLockSupplyingInputData(): "+
									"WorkingSet overflow");
		}

		clearAllUnlockedData();

		for (MetaGridData data : data_to_lock) {
			this.supplying_db.add(data);
			this.supplying_locked_db.add(data);
			this.data_tracker.trackData(data.id,this.rh);
		}

		assert(checkSizeInvariant());
		
	}

	public synchronized void unlockLocalInputData(Task t)
		throws GridException {

		if (supplying_locked_db.size() != 0) {
			throw new GridException("unlockLocalInputData(): "+
									"some Supplying data was locked");
		}
		if (t == null) {
			throw new GridException("unlockLocalInputData(): "+
									"illegal Task");
		}
		if (t.getInputMetaGridDataSet().length !=
			this.local_locked_db.size()) {
			throw new GridException("unlockLocalInputData(): "+
									"corrupted Local locked data "+
									"(input data set = "+
									t.getInputMetaGridDataSet().length+
									", Local locked data size ="+
									this.local_locked_db.size()+")");
		}

		this.local_locked_db.clear();
		
		assert(checkSizeInvariant());
		
	}

	public synchronized void unlockSupplyingInputData(Task t)
		throws GridException {
		
		if (local_locked_db.size() != 0) {
			throw new GridException("unlockLocalInputData(): "+
									"some Supplying data was locked");
		}
		if (t == null) {
			throw new GridException("unlockLocalInputData(): "+
									"illegal Task");
		}
		if (t.getInputMetaGridDataSet().length !=
			this.supplying_locked_db.size()) {
			throw new GridException("unlockSupplyingInputData(): "+
									"corrupted Supplying locked data"+
									"(input data set = "+
									t.getInputMetaGridDataSet().length+
									", Supplying locked data size ="+
									this.supplying_locked_db.size()+")");
		}

		this.supplying_locked_db.clear();
		
		assert(checkSizeInvariant());
		
	}
	
	// ****************************************************************
	
	/**
	 * <p>Note: the Resource-level DataTracker is updated by this operation
	 * @return true if target data_id was not already in the working set
	 */
	public synchronized boolean addLocalData(MetaGridData data)
		throws GridException {
		
		if (data == null) {
			throw new GridException("addLocalData(): illegal MetaGridData");
		}
		String data_id = data.id;
		if (isLocalData(data_id) == false) {
			throw new GridException("addLocalData(): not Local data "+
									data_id);
		}
		if (this.size() == this.capacity) {
			throw new GridException("addLocalData(): WorkingSet overflow");
		}
		
		boolean retval = true;
		if (this.local_db.add(data) == false) {
			retval = false;
		}
		
		this.data_tracker.trackData(data_id,this.rh);

		assert(checkSizeInvariant());

		return retval;
		
	}
	
	/**
	 * <p>Note: the Resource-level DataTracker is updated by this operation
	 * @return true if target data_id was not already in the working set
	 */
	public synchronized boolean addSupplyingData(MetaGridData data)
		throws GridException {
		
		if (data == null) {
			throw new GridException("addSupplyingData(): illegal MetaGridData");
		}
		String data_id = data.id;
		if (isSupplyingData(data_id) == false) {
			throw new GridException("addSupplyingData(): not Supplying data "+
									data_id);
		}
		if (this.size() == this.capacity) {
			throw new GridException("addSupplyingData(): WorkingSet overflow");
		}
		
		boolean retval = true;
		if (this.supplying_db.add(data) == false) {
			retval = false;
		}
		
		this.data_tracker.trackData(data_id,this.rh);

		assert(checkSizeInvariant());

		return retval;
		
	}

	/**
	 * <p>Note: the Resource-level DataTracker is updated by this operation
	 * @return true if no target data_id was already in the working set
	 */
	public synchronized boolean addLocalData(MetaGridData[] ds)
		throws GridException {
		
		if ((ds == null) || (ds.length < 1)) {
			throw new GridException("addLocalData(): illegal MetaGridData");
		}

		boolean retval = true;
		for (MetaGridData data : ds) {
			retval = retval && addLocalData(data);
		}
		
		return retval;
		
	}
		
	/**
	 * <p>Note: the Resource-level DataTracker is updated by this operation
	 * @return true if no target data_id was already in the working set
	 */
	public synchronized boolean addSupplyingData(MetaGridData[] ds)
		throws GridException {
		
		if ((ds == null) || (ds.length < 1)) {
			throw new GridException("addSupplyingData(): "+
									"illegal MetaGridData");
		}

		boolean retval = true;
		for (MetaGridData data : ds) {
			retval = retval && addSupplyingData(data);
		}
		
		return retval;
		
	}

	// ****************************************************************

	private boolean isLocalData(String data_id) throws GridException {
		
		if (data_id == null) {
			throw new GridException("isLocalData(): illegal data id");
		}
		
		return (data_id.indexOf(this.peer_prefix) == 0);
		
	}
	
	private boolean isSupplyingData(String data_id) throws GridException {
		
		if (data_id == null) {
			throw new GridException("isSupplyingData(): illegal data id");
		}
		
		return (data_id.indexOf(this.peer_prefix) != 0);
		
	}

	private synchronized boolean checkSizeInvariant() {
	
		int local_db_s = this.local_db.size();
		int supplying_db_s = this.supplying_db.size();
		int local_locked_db_s = this.local_locked_db.size();
		int supplying_locked_db_s = this.supplying_locked_db.size();
		
		return
			(local_db_s + supplying_db_s <= this.capacity) &&
			/*(local_locked_db_s + supplying_locked_db_s <= this.capacity) &&*/
			(local_locked_db_s <= local_db_s) &&
			(supplying_locked_db_s <= supplying_db_s) &&
			((local_locked_db_s == 0) || (supplying_locked_db_s == 0)) ;
		
	}
	
	// ****************************************************************

	private final String peer_prefix;
	private final int capacity;
	private final ResourceDataTracker data_tracker;
	private final ResourceHandle rh;

	private final Set<MetaGridData> local_db;
	private final Set<MetaGridData> supplying_db;
	private final Set<MetaGridData> local_locked_db;
	private final Set<MetaGridData> supplying_locked_db;
	
}

