/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LiveSchedulerController.java

package lbg.peer.core.scheduler;

import lbg.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.getStackTrace;
import lbg.peer.*;

import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * @author Cyril Briquet
 */
public class LiveSchedulerController extends SchedulerController
	implements Runnable {

	public LiveSchedulerController(Peer managed_peer,Scheduler scheduler)
		throws GridException {

		super(managed_peer);

		if (LightweightBarteringGrid.getMode() == SIM_MODE) {
			throw new GridException("LiveSchedulerController(): "+
									"unsupported mode");
		}

		if (managed_peer == null) {
			throw new GridException("LiveSchedulerController(): illegap Peer");
		}
		if (scheduler == null) {
			throw new GridException("LiveSchedulerController(): "+
									"illegal Scheduler");
		}
		
		this.managed_peer = managed_peer;
		this.scheduler = scheduler;
		this.ctrl_flags = new SchedulerControlFlags();
		this.q = new LinkedBlockingQueue<Boolean>();
		this.thread_pool =
			new ThreadPool(this,1,true,false,"Scheduler-daemon");

		this.max_size = new AtomicInteger(0);
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();

		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[SchedulerController]",
							   "[LiveSchedulerController]"));
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** Runnable implementation.
	 */
	public void run() {

		for ( ; ; ) {
			try {
				this.q.take();
				
				if (this.couldSchedule() == true) {
					this.getAndResetControlFlags(this.ctrl_flags);
					synchronized(this.managed_peer.rms()) {
						this.scheduler.schedule(this.ctrl_flags);
					}
					this.ctrl_flags.clear();
				}
			}
			catch (Throwable thr) {
				this.managed_peer.logger().
					info("Scheduler.schedule(): "+getStackTrace(thr));
				if (this.ctrl_flags != null) this.ctrl_flags.clear();
			}
		}

	}

	// ****************************************************************

	public synchronized void startService() throws GridException {
		
		this.thread_pool.startService();
		
	}
	
	public synchronized void awakeScheduler() throws GridException {
	
		try {
			int current_q_size = this.q.size();
			if (current_q_size < threshold) {
				this.q.put(Boolean.TRUE);
			}
			
			if (++current_q_size > this.max_size.get()) {
				this.max_size.set(current_q_size);
			}
		}
		catch (InterruptedException e) {
			throw new GridException(e);
		}
		
	}
	
	public int maxSize() {
		
		return this.max_size.get();
		
	}

	// ****************************************************************

	private final Peer managed_peer;
	private final Scheduler scheduler;
	private final SchedulerControlFlags ctrl_flags;
	private final LinkedBlockingQueue<Boolean> q;;
	private final ThreadPool thread_pool;

	private static final int threshold = 2;
	private       AtomicInteger max_size;
	
}

