/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Scheduler.java

package lbg.peer.core.scheduler;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.*;
import lbg.peer.core.negotiator.*;
import lbg.peer.strategy.scheduling.*;
import lbg.task.*;

/** Schedules Local Tasks, both locally and as Consumption Tasks,
 * and Supplying Tasks. Also Preempts Supplying Tasks as needed.
 * @author Cyril Briquet
 */
public class Scheduler implements ISchedulerStats {

	/** Creates a new Scheduler.
	 * @param managed_peer managed Peer
	 * @param negotiator Negotiator of the managed Peer
	 * @param target_scheduling_local_policy selected Local scheduling policy
	 * @param target_scheduling_running_preemption_policy
	 * selected running Preemption policy
	 * @param target_scheduling_waiting_preemption_policy
	 * selected waiting Preemption policy
	 * @param target_supplying_filtering_policy
	 * selected Supplying filtering policy
	 * @param target_scheduling_supplying_policy
	 * selected Supplying scheduling policy
	 * @param target_scheduling_consumption_policy
	 * selected Consumption scheduling policy
	 */
	public Scheduler(Peer managed_peer,Negotiator negotiator,
					 String target_scheduling_local_policy,
					 String target_scheduling_running_preemption_policy,
					 String target_scheduling_waiting_preemption_policy,
					 String target_supplying_filtering_policy,
					 String target_scheduling_supplying_policy,
					 String target_scheduling_consumption_policy)
		throws GridException {
	
		if (managed_peer == null) {
			throw new GridException("Scheduler(): illegal Peer");
		}
		if (negotiator == null) {
			throw new GridException("Scheduler(): illegal Peer");
		}
		if (target_scheduling_local_policy == null) {
			throw new GridException("Scheduler(): "+
									"illegal target "+
									"Scheduling Local policy");
		}
		if (target_scheduling_running_preemption_policy == null) {
			throw new GridException("Scheduler(): "+
									"illegal target "+
									"waiting Scheduling Preemption policy");
		}
		if (target_scheduling_waiting_preemption_policy == null) {
			throw new GridException("Scheduler(): "+
									"illegal target "+
									"running Scheduling Preemption policy");
		}
		if (target_supplying_filtering_policy == null) {
			throw new GridException("Scheduler(): "+
									"illegal target "+
									"Supplying filtering policy");
		}
		if (target_scheduling_supplying_policy == null) {
			throw new GridException("Scheduler(): "+
									"illegal target "+
									"Scheduling Supplying policy");
		}
		if (target_scheduling_consumption_policy == null) {
			throw new GridException("Scheduler(): "+
									"illegal target "+
									"Scheduling Consumption policy");
		}

		this.managed_peer = managed_peer;
		this.target_scheduling_local_policy =
			target_scheduling_local_policy;
		this.target_scheduling_running_preemption_policy =
			target_scheduling_running_preemption_policy;
		this.target_scheduling_waiting_preemption_policy =
			target_scheduling_waiting_preemption_policy;
		this.target_supplying_filtering_policy =
			target_supplying_filtering_policy;
		this.target_scheduling_supplying_policy =
			target_scheduling_supplying_policy;
		this.target_scheduling_consumption_policy =
			target_scheduling_consumption_policy;
		
		boolean tmp_is_bartering_strategy = true;
		

		try {
			LocalSchedulingType lst =
				LocalSchedulingType.
					valueOf(this.target_scheduling_local_policy);
			this.local_policy =
				lst.createLocalSchedulingPolicy(this.managed_peer,this);
		}
		catch (IllegalArgumentException e) {
			throw new GridException("Scheduler(): illegal "+
									"Local Scheduling policy"+
									this.target_scheduling_local_policy,
									e);
		}
		try {
			SupplyingRunningPreemptionType rpt =
				SupplyingRunningPreemptionType.
					valueOf(this.target_scheduling_running_preemption_policy);
			this.running_preemption_policy =
				rpt.createSupplyingRunningPreemptionPolicy(this.managed_peer,this);
		}
		catch (IllegalArgumentException e) {
			throw new GridException("Scheduler(): illegal "+
									"running Preemption Scheduling policy"+
									this.target_scheduling_running_preemption_policy,
									e);
		}
		try {
			SupplyingWaitingPreemptionType wpt =
				SupplyingWaitingPreemptionType.
					valueOf(this.target_scheduling_waiting_preemption_policy);
			this.waiting_preemption_policy =
				wpt.createSupplyingWaitingPreemptionPolicy(this.managed_peer);
		}
		catch (IllegalArgumentException e) {
			throw new GridException("Scheduler(): illegal "+
									"waiting Preemption Scheduling policy"+
									this.target_scheduling_waiting_preemption_policy,
									e);
		}
		try {
			SupplyingFilteringType sft =
				SupplyingFilteringType.
					valueOf(this.target_supplying_filtering_policy);
			this.supplying_filtering_policy =
				sft.createSupplyingFilteringPolicy(this.managed_peer);
		}
		catch (IllegalArgumentException e) {
			throw new GridException("Scheduler(): illegal "+
									"Supplying filtering policy "+
									this.target_supplying_filtering_policy,
									e);
		}
		try {
			SupplyingSchedulingType sst =
				SupplyingSchedulingType.
					valueOf(this.target_scheduling_supplying_policy);
			if (sst.isBarteringStrategy() == false) {
				tmp_is_bartering_strategy = false;
			}
			this.supplying_policy =
				sst.createSupplyingSchedulingPolicy(this.managed_peer,this);
		}
		catch (IllegalArgumentException e) {
			throw new GridException("Scheduler(): illegal "+
									"Supplying Scheduling policy"+
									this.target_scheduling_supplying_policy,
									e);
		}
		try {
			ConsumptionSchedulingType cst = 
				ConsumptionSchedulingType.
					valueOf(this.target_scheduling_consumption_policy);
			if (cst.isBarteringStrategy() == false) {
				tmp_is_bartering_strategy = false;
			}
			this.consumption_policy =
				cst.createConsumptionSchedulingPolicy(this.managed_peer,this);
		}
		catch (IllegalArgumentException e) {
			throw new GridException("Scheduler(): illegal "+
									"Consumption Scheduling policy"+
									this.target_scheduling_consumption_policy,
									e);
		}

		this.is_bartering_strategy = tmp_is_bartering_strategy;
		
		this.scheduler_decision_local       = 0;
		this.scheduler_decision_preemption  = 0;
		this.scheduler_decision_supplying   = 0;
		this.scheduler_decision_consumption = 0;
		this.m_blacklisted_supplier_ratio = MeasurementFactory.
			createSummarizedMeasurement("blacklisted supplier ratio");

	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[Scheduler ");
		sb.append(this.managed_peer.id());
		sb.append("]");
		
		sb.append("\n  ->  Local       Scheduling = ");
		sb.append(this.target_scheduling_local_policy);
		sb.append("\n  ->  Preemption (running)   = ");
		sb.append(this.target_scheduling_running_preemption_policy);
		sb.append("\n  ->  Preemption (waiting)   = ");
		sb.append(this.target_scheduling_waiting_preemption_policy);
		sb.append("\n  ->  Supplying   filtering  = ");
		sb.append(this.target_supplying_filtering_policy);
		sb.append("\n  ->  Supplying   Scheduling = ");
		sb.append(this.target_scheduling_supplying_policy);
		sb.append("\n  ->  Consumption Scheduling = ");
		sb.append(this.target_scheduling_consumption_policy);
		sb.append("\n  ->  Bartering ");
		sb.append(this.is_bartering_strategy ? "enabled" : "disabled");

		long scheduler_decisions =
			this.scheduler_decision_local+
			this.scheduler_decision_preemption+
			this.scheduler_decision_consumption+
			this.scheduler_decision_supplying;
		sb.append("\n  ->  scheduling decisions = { ");
		sb.append(CommonMath.div_label(this.scheduler_decision_local,
							 		   scheduler_decisions,100.0));
		sb.append("% Local, ");
		sb.append(CommonMath.div_label(this.scheduler_decision_preemption,
							 		   scheduler_decisions,100.0));
		sb.append("% Local preemption, ");
		sb.append(CommonMath.div_label(this.scheduler_decision_consumption,
							 		   scheduler_decisions,100.0));
		sb.append("% Consumption, ");
		sb.append(CommonMath.div_label(this.scheduler_decision_supplying,
							 		   scheduler_decisions,100.0));
		sb.append("% Supplying");
		sb.append(" }");
		
		sb.append("\n  ->  blacklisting decisions = { ").
			append(CommonMath.
					div_label(this.m_blacklisted_supplier_ratio.getMean(),
							  scheduler_decisions,100.0)).
			append(" % blacklisted suppliers").
			append(" }");
		
		return sb.toString();
		
	}
	
	// ****************************************************************
	
	/** Activates scheduling of Local, Supplying and Consumption Tasks
	 * following target scheduling control flags.
	 * <p>Also asks the Negotiator to generate Supplying Requests as needed.
	 * @param ctrl_flags target scheduling control flags
	 * @exception GridException if a parameter is illegal,
	 * or an error happens while scheduling
	 */
	public synchronized void schedule(SchedulerControlFlags ctrl_flags)
		throws GridException {

		if (ctrl_flags == null) {
			throw new GridException("schedule(): illegal control flags");
		}
		
		boolean local_scheduling_flag =
			ctrl_flags.getLocalSchedulingFlag();
		boolean supplying_scheduling_flag =
			ctrl_flags.getSupplyingSchedulingFlag();
		boolean consumption_scheduling_flag =
			ctrl_flags.getConsumptionSchedulingFlag();
		
		if ((local_scheduling_flag == false) &&
			(supplying_scheduling_flag == false) &&
			(consumption_scheduling_flag == false)) {
			throw new GridException("schedule(): illegal control flags");
		}

		// ****************
		
		boolean gen_supplying_requests = false;
		
		if (local_scheduling_flag || consumption_scheduling_flag) {
			this.scheduleLocalTasks();
			
			gen_supplying_requests =
				this.managed_peer.qms().hasWaitingLocalTasks();
		}

		if (supplying_scheduling_flag) {
			this.scheduleSupplyingTasks();
		}

		if (consumption_scheduling_flag) {
			this.scheduleConsumptionTasks();

			gen_supplying_requests =
				this.managed_peer.qms().hasWaitingLocalTasks();
		}
	
		if (gen_supplying_requests) {
			synchronized(this.managed_peer.rms()) {
				this.managed_peer.negotiatorController().
					activateSupplyingRequestsGeneration();
			}
		}

	}
	
	// ****************************************************************

	private synchronized void scheduleLocalTasks() throws GridException {

		// Synchronize on RMS in order to avoid race conditions due to long,
		// multi-calls transactions btwn TaskManager, ResourceManager
		synchronized(this.managed_peer.rms()) {
			this.local_policy.scheduleLocalTasks();
		}
		
	}

	/** Preempts some Supplying Tasks on Local Resources
	 * so as to enable the Local scheduling of target prioritary Local Task.
	 * @param prioritary_task target prioritary Local Task
	 * @exception GridException if a parameter is illegal,
	 * or an error happens while preempting
	 */
	public synchronized void preemptLocalResources(Task prioritary_task)
		throws GridException {

		// Synchronize on RMS in order to avoid race conditions due to long,
		// multi-calls transactions btwn TaskManager, ResourceManager
		synchronized(this.managed_peer.rms()) {
			this.running_preemption_policy.preemptLocalResources(prioritary_task);
		}
		
	}
	
	/** Preempts some waiting Supplying Tasks.
	 */
	public synchronized void preemptWaitingSupplyingTasks()
		throws GridException {
		
		// Synchronize on RMS in order to avoid race conditions due to long,
		// multi-calls transactions btwn TaskManager, ResourceManager
		synchronized(this.managed_peer.rms()) {
			this.waiting_preemption_policy.preemptWaitingSupplyingTasks();
		}
		
	}

	/** Indicates whether target Supplying Task should be filtered out,
	 * or may be safely queued.
	 * @return true if target Supplying Task should be filtered out,
	 * false it it can be accepted
	 */
	public synchronized boolean filterOutSupplyingTask(Task t)
		throws GridException {

		// Synchronize on RMS in order to avoid race conditions due to long,
		// multi-calls transactions btwn TaskManager, ResourceManager
		synchronized(this.managed_peer.rms()) {
			return this.supplying_filtering_policy.filterOutSupplyingTask(t);
		}
		
	}

	private synchronized boolean scheduleConsumptionTasks()
		throws GridException {

		// Synchronize on RMS in order to avoid race conditions due to long,
		// multi-calls transactions btwn TaskManager, ResourceManager
		synchronized(this.managed_peer.rms()) {
			return this.consumption_policy.scheduleConsumptionTasks() > 0;
		}

	}

	private synchronized void scheduleSupplyingTasks()
		throws GridException {

		// Synchronize on RMS in order to avoid race conditions due to long,
		// multi-calls transactions btwn TaskManager, ResourceManager
		synchronized(this.managed_peer.rms()) {
			this.supplying_policy.scheduleSupplyingTasks();
			this.managed_peer.rms().replicateSupplyingData();
		}

	}

	// ****************************************************************

	public synchronized void accountSchedulerLocalDecision() {
		
		this.scheduler_decision_local++;
		
	}

	public synchronized void accountSchedulerPreemptionDecision() {
		
		this.scheduler_decision_preemption++;
		
	}

	public synchronized void
		accountSchedulerConsumptionDecision(long scheduled_tasks,
											double blacklisted_supplier_ratio)
		throws GridException {
		
		if (scheduled_tasks <= 0) {
			throw new GridException("accountSchedulerConsumptionDecision(): "+
									"illegal Consumption Tasks scheduling "+
									"decisions count");
		}
		if ((blacklisted_supplier_ratio < 0.0) ||
			(blacklisted_supplier_ratio > 1.0)) {
			throw new GridException("accountSchedulerConsumptionDecision(): "+
									"illegal Consumption Tasks scheduling "+
									"blacklisted supplier ratio");
		}

		this.scheduler_decision_consumption += scheduled_tasks;
		this.m_blacklisted_supplier_ratio.
			addObservation(blacklisted_supplier_ratio);

	}

	public synchronized void accountSchedulerSupplyingDecision() {
		
		this.scheduler_decision_supplying++;

	}
	
	public synchronized double getBlacklistedSupplierRatio() {
		
		return m_blacklisted_supplier_ratio.getMean();
		
	}
	
	// ****************************************************************

	private final Peer managed_peer;
	private final String target_scheduling_local_policy;
	private final String target_scheduling_running_preemption_policy;
	private final String target_scheduling_waiting_preemption_policy;
	private final String target_supplying_filtering_policy;
	private final String target_scheduling_supplying_policy;
	private final String target_scheduling_consumption_policy;

	private final boolean is_bartering_strategy;
	private final LocalSchedulingPolicy local_policy;
	private final SupplyingRunningPreemptionPolicy running_preemption_policy;
	private final SupplyingWaitingPreemptionPolicy waiting_preemption_policy;
	private final SupplyingFilteringPolicy supplying_filtering_policy;
	private final ConsumptionSchedulingPolicy consumption_policy;
	private final SupplyingSchedulingPolicy supplying_policy;
	
	private       long scheduler_decision_local;
	private       long scheduler_decision_preemption;
	private       long scheduler_decision_supplying;
	private       long scheduler_decision_consumption;
	private final Measurement m_blacklisted_supplier_ratio;

}

