/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ExecutionManager.java

package lbg.peer.core.xms;

import lbg.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.*;
import lbg.peer.core.qms.*;
import static lbg.peer.core.qms.RequestStatus.*;
import static lbg.peer.core.qms.RequestType.*;
import lbg.peer.core.qms.prequeue.*;
import lbg.peer.service.*;
import lbg.resource.service.*;
import lbg.task.*;

/**
 * @author Cyril Briquet
 */
public class ExecutionManager {

	public ExecutionManager(Peer managed_peer,long consumption_timeout)
		throws GridException {
	
		if (managed_peer == null) {
			throw new GridException("ExecutionManager(): illegal Peer");
		}
		if (consumption_timeout <= 0) {
			throw new GridException("ExecutionManager(): "+
									"illegal Consumption timeout");
		}

		this.managed_peer = managed_peer;
		this.consumption_timeout = consumption_timeout;
		this.local_timeout       = this.LOCAL_TIMEOUT;
		this.supplying_timeout   = this.LOCAL_TIMEOUT;

		this.run_local_tasks_count = 0;
		this.run_supplying_tasks_count = 0;
		this.run_consumption_tasks_count = 0;

		this.cancelled_local_tasks_total_runtime = 0;
		this.cancelled_local_tasks_count = 0;
		this.cancelled_supplying_tasks_total_runtime = 0;
		this.cancelled_supplying_tasks_count = 0;
		this.cancelled_consumption_tasks_total_runtime = 0;
		this.cancelled_consumption_tasks_count = 0;

		this.completed_local_tasks_total_runtime = 0;
		this.completed_local_tasks_count = 0;
		this.completed_supplying_tasks_total_runtime = 0;
		this.completed_supplying_tasks_count = 0;
		this.completed_consumption_tasks_total_runtime = 0;
		this.completed_consumption_tasks_count = 0;

		// [TIMEOUT] keep track of running Tasks,
		//           for timeout enforcement and monitoring purposes
		Peer p = this.managed_peer;
		// Consumption Tasks
		TaskComparator consumption_tasks_cmp =
			LocalTaskComparator.getDefaultComparator();
		this.running_consumption_tasks =
			new RunningConsumptionTaskManager(p,consumption_tasks_cmp,
											  "ConsumptionTaskRuntimeData");
		// Local Tasks
		TaskComparator local_tasks_cmp =
			LocalTaskComparator.getDefaultComparator();
		this.running_local_tasks =
			new RunningLocalTaskManager(p,local_tasks_cmp,
										"LocalTaskRuntimeData");
		// Supplying Tasks
		TaskComparator supplying_tasks_cmp =
			SupplyingTaskComparator.getDefaultComparator();
		this.running_supplying_tasks =
			new RunningSupplyingTaskManager(p,supplying_tasks_cmp,
											"SupplyingTaskRuntimeData");

	}
	
	public String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ExecutionManager ");
		sb.append(this.managed_peer.id());
		sb.append("]\n");
		
		sb.append("  ->  Local Tasks       { run=");
		sb.append(this.run_local_tasks_count);

		sb.append(", cancelled=");
		sb.append(this.cancelled_local_tasks_count);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.cancelled_local_tasks_count,
									   this.run_local_tasks_count,100));
		sb.append("%)");

		sb.append(", completed=");
		sb.append(this.completed_local_tasks_count);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.completed_local_tasks_count,
									   this.run_local_tasks_count,100));
		sb.append("%)");
		sb.append(", total runtime=");
		sb.append(this.completed_local_tasks_total_runtime);
		sb.append(", mean runtime=");
		sb.append(CommonMath.div_label(this.completed_local_tasks_total_runtime,
									   this.completed_local_tasks_count));
		sb.append(" }\n");

		sb.append("  ->  Supplying Tasks   { run=");
		sb.append(this.run_supplying_tasks_count);

		sb.append(", cancelled=");
		sb.append(this.cancelled_supplying_tasks_count);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.cancelled_supplying_tasks_count,
									   this.run_supplying_tasks_count,100));
		sb.append("%)");

		sb.append(", completed=");
		sb.append(this.completed_supplying_tasks_count);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.completed_supplying_tasks_count,
									   this.run_supplying_tasks_count,100));
		sb.append("%)");
		sb.append(", total runtime=");
		sb.append(this.completed_supplying_tasks_total_runtime);
		sb.append(", mean runtime=");
		sb.append(CommonMath.
					div_label(this.completed_supplying_tasks_total_runtime,
							  this.completed_supplying_tasks_count));
		sb.append(" }\n");

		sb.append("  ->  Consumption Tasks { run=");
		sb.append(this.run_consumption_tasks_count);

		sb.append(", cancelled=");
		sb.append(this.cancelled_consumption_tasks_count);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.cancelled_consumption_tasks_count,
									   this.run_consumption_tasks_count,100));
		sb.append("%)");

		sb.append(", completed=");
		sb.append(this.completed_consumption_tasks_count);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.completed_consumption_tasks_count,
									   this.run_consumption_tasks_count,100));
		sb.append("%)");
		sb.append(", total runtime=");
		sb.append(this.completed_consumption_tasks_total_runtime);
		sb.append(", mean runtime=");
		sb.append(CommonMath.
					div_label(this.completed_consumption_tasks_total_runtime,
							  this.completed_consumption_tasks_count));
		sb.append(" }\n");

		long run_tasks_count = this.run_local_tasks_count +
			this.run_supplying_tasks_count + this.run_consumption_tasks_count;
		long completed_tasks_count = this.completed_local_tasks_count +
			this.completed_supplying_tasks_count +
			this.completed_consumption_tasks_count;
		long cancelled_tasks_count = this.cancelled_local_tasks_count +
			this.cancelled_supplying_tasks_count +
			this.cancelled_consumption_tasks_count;
		long completed_tasks_total_runtime =
			this.completed_local_tasks_total_runtime +
			this.completed_supplying_tasks_total_runtime +
			this.completed_consumption_tasks_total_runtime;
		sb.append("  ->  All Tasks         { run=");
		sb.append(run_tasks_count);

		sb.append(", cancelled=");
		sb.append(cancelled_tasks_count);
		sb.append(" (");
		sb.append(CommonMath.div_label(cancelled_tasks_count,
									   run_tasks_count,100));
		sb.append("%)");

		sb.append(", completed=");
		sb.append(completed_tasks_count);
		sb.append(" (");
		sb.append(CommonMath.div_label(completed_tasks_count,
									   run_tasks_count,100));
		sb.append("%)");
		sb.append(", total runtime=");
		sb.append(completed_tasks_total_runtime);
		sb.append(", mean runtime=");
		sb.append(CommonMath.div_label(completed_tasks_total_runtime,
									   completed_tasks_count));
		sb.append(" }");
		
		sb.append("\n  ->  ");
		sb.append(this.running_consumption_tasks);
		sb.append("\n  ->  ");
		sb.append(this.running_local_tasks);
		sb.append("\n  ->  ");
		sb.append(this.running_supplying_tasks);
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized long getRunLocalTasksCount() {
		
		return this.run_local_tasks_count;
		
	}
	
	public synchronized long getRunSupplyingTasksCount() {
		
		return this.run_supplying_tasks_count;
		
	}
	
	public synchronized long getRunConsumptionTasksCount() {
		
		return this.run_consumption_tasks_count;
		
	}

	public synchronized long getCancelledLocalTasksTotalRuntime() {
		
		return this.cancelled_local_tasks_total_runtime;
		
	}

	public synchronized long getCancelledLocalTasksCount() {
		
		return this.cancelled_local_tasks_count;
		
	}

	public synchronized long getCancelledSupplyingTasksTotalRuntime() {
		
		return this.cancelled_supplying_tasks_total_runtime;
		
	}

	public synchronized long getCancelledSupplyingTasksCount() {
		
		return this.cancelled_supplying_tasks_count;
		
	}

	public synchronized long getCancelledConsumptionTasksTotalRuntime() {
		
		return this.cancelled_consumption_tasks_total_runtime;
		
	}

	public synchronized long getCancelledConsumptionTasksCount() {
		
		return this.cancelled_consumption_tasks_count;
		
	}

	public synchronized long getCompletedLocalTasksTotalRuntime() {
		
		return this.completed_local_tasks_total_runtime;
		
	}

	public synchronized long getCompletedLocalTasksCount() {
		
		return this.completed_local_tasks_count;
		
	}

	public synchronized long getCompletedSupplyingTasksTotalRuntime() {
		
		return this.completed_supplying_tasks_total_runtime;
		
	}

	public synchronized long getCompletedSupplyingTasksCount() {
		
		return this.completed_supplying_tasks_count;
		
	}

	public synchronized long getCompletedConsumptionTasksTotalRuntime() {
		
		return this.completed_consumption_tasks_total_runtime;
		
	}

	public synchronized long getCompletedConsumptionTasksCount() {
		
		return this.completed_consumption_tasks_count;
		
	}
	
	public synchronized double estimateMeanLocalTaskRuntime() {
	
		return (this.completed_local_tasks_count > 0) ?
	 			(((double)this.completed_local_tasks_total_runtime)/
				 ((double)this.completed_local_tasks_count)) : 0.0;
	
	}

	public synchronized double estimateMeanSupplyingTaskRuntime() {
	
		return (this.completed_supplying_tasks_count > 0) ?
	 			(((double)this.completed_supplying_tasks_total_runtime)/
				 ((double)this.completed_supplying_tasks_count)) : 0.0;
	
	}
	
	public synchronized double getLocalToTotalTasksRatio() {
		
		long completed_local_and_supplying_tasks_count =
			this.completed_local_tasks_count+
			this.completed_supplying_tasks_count;

		return (completed_local_and_supplying_tasks_count > 0) ?
				(((double)this.completed_local_tasks_count) /
				 ((double)completed_local_and_supplying_tasks_count)) : 0.0;

	}

	public synchronized double getSupplyingToTotalTasksRatio() {
		
		long completed_local_and_supplying_tasks_count =
			this.completed_local_tasks_count+
			this.completed_supplying_tasks_count;

		return (completed_local_and_supplying_tasks_count > 0) ?
				(((double)this.completed_supplying_tasks_count) /
				 ((double)completed_local_and_supplying_tasks_count)) : 0.0;

	}

	public synchronized double estimateMeanSelfRuntime() {

		return (this.estimateMeanLocalTaskRuntime() *
				this.getLocalToTotalTasksRatio()) +
			   (this.estimateMeanSupplyingTaskRuntime() *
				this.getSupplyingToTotalTasksRatio());
		
	}

	public synchronized double estimateMeanSupplierRuntime() {

		return (this.completed_consumption_tasks_count > 0) ?
			(((double)this.completed_consumption_tasks_total_runtime) /
			 ((double)this.completed_consumption_tasks_count)) : 0.0;

	}

	// ****************************************************************

	/** Runs target Task on target Resource.
	 * @param rh target Resource
	 * @param t target Task
	 * @return true if the target Resource can run the target Task,
	 * or false if the target Resource cannot run the target Task
	 * (e.g. typically because the WorkingSet is too large for the Resource,
	 * which would be caused by uncareful scheduling)
	 */
	public synchronized boolean
		runLocalTaskOnLocalResource(Task t,ResourceHandle rh)
		throws GridException {
		
		if (t == null) {
			throw new GridException("runLocalTaskOnLocalResource(): "+
									"illegal Task");
		}
		if (rh == null) {
			throw new GridException("runLocalTaskOnLocalResource(): "+
									"illegal Resource handle");
		}
		
		// ****************

		if (this.managed_peer.rms().runLocalTask(rh,t) == false) {
			return false;
		}

		// ****************

		// Update stats
		this.run_local_tasks_count++;

		// ****************

		// [TIMEOUT]
		this.running_local_tasks.
			addRunningTask(t,this.local_timeout,null,
						   this.managed_peer.id(),false);// null 3rd param 8)
	   
		return true;

	}
	
	/** Runs target Task on target Resource.
	 * @param rh target Resource
	 * @param t target Task
	 * @return true if the target Resource can run the target Task,
	 * or false if the target Resource cannot run the target Task
	 * (e.g. typically because the WorkingSet is too large for the Resource,
	 * which would be caused by uncareful scheduling)
	 */
	public synchronized boolean
		runSupplyingTaskOnLocalResource(Task t,ResourceHandle rh)
		throws GridException {
		
		if (t == null) {
			throw new GridException("runSupplyingTaskOnLocalResource(): "+
									"illegal Task");
		}
		if (rh == null) {
			throw new GridException("runSupplyingTaskOnLocalResource(): "+
									"illegal Resource handle");
		}
		
		// ****************

		if (this.managed_peer.rms().runSupplyingTask(rh,t) == false) {
			return false;
		}
		
		// ****************

		// Update stats
		this.run_supplying_tasks_count++;

		// ****************
		
		// [TIMEOUT]
		ExternalPeerHandle peer_handle =
			this.managed_peer.register().getHandle(t.consumerID());
		this.running_supplying_tasks.
			addRunningTask(t,this.supplying_timeout,peer_handle,
						   this.managed_peer.id(),true);
			
		return true;

	}

	/** Returns the target Task on target Supplier Peer.
	 * @param t target Task
	 * @param peer_handle target Supplier Peer
	 * @return true if target Supplier Peer has accepted to run target Task
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized boolean
		runLocalTaskOnSupplyingResource(Task t,
										ExternalPeerHandle peer_handle)
		throws GridException {

		if (t == null) {
			throw new GridException("runLocalTaskOnSupplyingResource(): "+
									"illegal Task");
		}
		if (peer_handle == null) {
			throw new GridException("runLocalTaskOnSupplyingResource(): "+
									"illegal Peer handle");
		}

		// ****************

		ConsumptionTaskID is_accepted = null;
		
		try {
			is_accepted =
				peer_handle.computeSupplyingTask(this.managed_peer.id(),t);
		}
		catch (GridException e) {
			this.managed_peer.logger().
				info("runLocalTaskOnSupplyingResource(): could not contact "+
					  peer_handle.id());
			is_accepted = null;
		}
		finally {
			if ((is_accepted == null) || (is_accepted.job_id < 0)) {
				return false;
			}
		}

		// Update the Consumption Task with the Supplying space id
		// returned by the Consumer upon Task acceptation
		t.setSupplyingSpaceJobID(is_accepted.job_id); // ULTRA IMPORTANT
		t.setSupplyingSpaceBoTID(is_accepted.bot_id); // ULTRA IMPORTANT
		
		// ****************

		// Update stats
		this.run_consumption_tasks_count++;

		// ****************

		// [TIMEOUT]
		this.running_consumption_tasks.
			addRunningTask(t,this.consumption_timeout,peer_handle,
						   peer_handle.id(),true);

		return true;
		
	}

	// ****************************************************************

	/** Returns the start time of target Task's execution.
	 * @param t target Task
	 * @return start time
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized long getStartTime(Task t,RequestType rt)
		throws GridException {
		
		if (t == null) {
			throw new GridException("getStartTime(): illegal Task");
		}
		
		RunningTaskManager rtm = null;
		switch (rt) {
		case CONSUMPTION_TASK:
			rtm = this.running_consumption_tasks;
			break;
		case SUPPLYING_TASK:
			rtm = this.running_supplying_tasks;
			break;
		case LOCAL_TASK:
			rtm = this.running_local_tasks;
			break;
		default:
			throw new GridException("getStartTime(): illegal Task type");
		}

		return rtm.getStartTime(t);
		
	}
	
	/** Updates the time-out and action tag associated with the target
	 * Supplying Task.
	 * @param t target Task
	 * @param do_cancel time-out action tag
	 * @param time_out time-out
	 * @return true if the time-out and action tag have been updated,
	 * or false if they were not updated
	 * (i.e. the action tag has already been updated)
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized boolean
		updateSupplyingTaskTimeOutAction(Task t,boolean do_cancel,long time_out)
		throws GridException {
		
		return this.running_supplying_tasks.
			updateTimeOutActionTagAndValue(t,do_cancel,time_out);
		
	}
	
	/** Identifies if target Supplying Task has been graced
	 * @param t target Task
	 * @return true if target Supplying Task has been graced
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized boolean isSupplyingTaskGraced(Task t)
		throws GridException {
		
		return this.running_supplying_tasks.isTaskGraced(t);
		
	}

	/** Returns the most recently running Supplying Task that has not
	 * been graced
	 * (useful for PSufferage Resource selection to preempt
	 * a running Supplying Task).
	 * @return most recently running Supplying Task that has not been graced,
	 * or null if there is no running Supplying Task ungraced
	 */
	public synchronized Task getMostRecentlyRunningAndUngracedSupplyingTask()
		throws GridException {
		
		return this.running_supplying_tasks.
			getMostRecentlyRunningAndUngracedTask();

	}
	
	/** Identity running Task type.
	 * @param t target Task
	 * @return running Task type
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized RequestType identifyRunningTaskType(Task t)
		throws GridException {
		
		if (t == null) {
			throw new GridException("identifyRunningTaskType(): illegal Task");
		}
		
		boolean is_running_local =
			(this.running_local_tasks.identifySupplier(t) != null);
		boolean is_running_supplying =
			(this.running_supplying_tasks.identifySupplier(t) != null);
		boolean is_running_consumption =
			(this.running_consumption_tasks.identifySupplier(t) != null);

		if (is_running_local) {
			if (is_running_supplying || is_running_consumption) {
				throw new GridException("identifyRunningTaskType(): "+
											"corrupted data (1)");
			}
			return LOCAL_TASK;
		}

		if (is_running_supplying) {
			if (is_running_local || is_running_consumption) {
				throw new GridException("identifyRunningTaskType(): "+
											"corrupted data (2)");
			}
			return SUPPLYING_TASK;
		}

		if (is_running_consumption) {
			if (is_running_local || is_running_supplying) {
				throw new GridException("identifyRunningTaskType(): "+
											"corrupted data (3)");
			}
			return CONSUMPTION_TASK;
		}

		throw new GridException("identifyRunningTaskType(): "+
								"Task not running");
		
	}
	
	// ****************************************************************

	public synchronized RequestStatus
		completeLocalTaskOnLocalResource(Task t) throws GridException {

		if (t == null) {
			throw new GridException("completeLocalTaskOnLocalResource(): "+
									"illegal Task");
		}

		long now = LightweightBarteringGrid.getTime();

		// ****************

		// [TIMEOUT]
		long xstart_time = -1;
		//	incorrect code -- long xstart_time = now-t.output().runtime();
		try {
			xstart_time = this.running_local_tasks.removeRunningTask(t);
		}
		catch (GridException e) {
			// Local Task has already been processed
			return TASK_DISCARDED;
		}
		

		// Update stats
		long runtime = now - xstart_time;
		assert(runtime >= 0);
		this.completed_local_tasks_total_runtime += runtime;
		this.completed_local_tasks_count++;

		// ****************

		// Account local activity
		this.managed_peer.register().getNeighborhoodBarteringProfile().
			accountLocalActivityCompletion(t);

		// Free Request (Peer QueueManager viewpoint)
		RequestStatus status = this.managed_peer.qms().
			dequeueCompletedLocalTask(t,xstart_time,runtime);
		
		// Free Resource (Peer ResourceManager viewpoint)
		this.managed_peer.rms().freeBusyLocalResource(t);

		// ****************
		
		return status;

	}

	public synchronized RequestStatus
		completeSupplyingTaskOnLocalResource(Task t) throws GridException {
		
		if (t == null) {
			throw new GridException("completeSupplyingTaskOnLocalResource(): "+
									"illegal Task");
		}

		long now = LightweightBarteringGrid.getTime();

		// ****************

		// [TIMEOUT]
		long xstart_time = -1;
		//	incorrect code -- long xstart_time = now-t.output().runtime();
		try {
			xstart_time = this.running_supplying_tasks.removeRunningTask(t);
		}
		catch (GridException e) {
			// Supplying Task has already been processed
			return TASK_DISCARDED;
		}

		// Update stats
		long runtime = now - xstart_time;
		assert(runtime >= 0);
		this.completed_supplying_tasks_total_runtime += runtime;
		this.completed_supplying_tasks_count++;

		// ****************

		// Account supplying
		this.managed_peer.register().getNeighborhoodBarteringProfile().
			accountSupplyingCompletion(t,xstart_time);

		// Free Request (Peer QueueManager viewpoint)
		RequestStatus status = this.managed_peer.qms().
			dequeueCompletedSupplyingTask(t,xstart_time,runtime);

		// Free Resource (Peer ResourceManager viewpoint)
		this.managed_peer.rms().freeBusySupplyingResource(t);

		// ****************

		return status;

	}

	public synchronized RequestStatus
		completeLocalTaskOnSupplyingResource(Task t) throws GridException {
		
		if (t == null) {
			throw new GridException("completeLocalTaskOnSupplyingResource: "+
									"illegal Task");
		}

		long now = LightweightBarteringGrid.getTime();
		
		// ****************

	
		// [TIMEOUT]
		long xstart_time = -1;
		try {
			xstart_time = this.running_consumption_tasks.removeRunningTask(t);
		}
		catch (GridException e) {
			// Consumption Task has already been processed
			return TASK_DISCARDED;
		}

		// Update stats
		long runtime = now - xstart_time;
		assert(runtime >= 0);
		this.completed_consumption_tasks_total_runtime += runtime;
		this.completed_consumption_tasks_count++;

		// ****************

		// Account consumption
		this.managed_peer.register().getNeighborhoodBarteringProfile().
			accountConsumptionCompletion(t,xstart_time);

		// Free Request (Peer QueueManager viewpoint)
		RequestStatus status =  this.managed_peer.qms().
			dequeueCompletedConsumptionTask(t,xstart_time,runtime);

		// ****************

		return status;

	}

	// ****************************************************************

	/** The managed Peer instructs itself to cancel a Local Task
	 * running locally.
	 */
	public synchronized void
		cancelLocalTaskOnLocalResource(Task t,byte[] logs)
		throws GridException {

		if (t == null) {
			throw new GridException("cancelLocalTaskOnLocalResource(): "+
									"illegal Task");
		}
		if (logs == null) {
			throw new GridException("cancelLocalTaskOnLocalResource(): "+
									"illegal logs array");
		}
		
		// ****************

		// [TIMEOUT]
		long xstart_time = this.running_local_tasks.removeRunningTask(t);
		//	incorrect code -- long xstart_time = now-t.output().runtime();

		// ****************

		// Account Local non-activity
		this.managed_peer.register().getNeighborhoodBarteringProfile().
			accountLocalActivityCancellation(t);
		
		// Free Resource (Peer ResourceManager viewpoint)
		long rms_xstart_time =
			this.managed_peer.rms().preemptBusyLocalResource(t);
		
		if ((xstart_time != rms_xstart_time) &&
			(LightweightBarteringGrid.getMode() == SIM_MODE)) {
			throw new GridException("cancelLocalTaskOnLocalResource(): "+
									"XMS X start time="+xstart_time+
									", RMS X start_time="+rms_xstart_time);
		}

		// Free Request (Peer QueueManager viewpoint)
		long runtime = LightweightBarteringGrid.getTime() - xstart_time;
		assert(runtime >= 0);
		this.managed_peer.qms().
			dequeueCancelledLocalTask(t,xstart_time,runtime,t.consumerID());

		// ****************

		// Update stats
		this.cancelled_local_tasks_total_runtime +=
			LightweightBarteringGrid.getTime()-xstart_time;
		this.cancelled_local_tasks_count++;

		// ****************

	}

	/** The managed Peer instructs itself to preempt a Local Task
	 * running locally.
	 * <p>Note: Task preemption is non-terminating cancellation of a Task.
	 */
	public synchronized void
		preemptLocalTaskOnLocalResource(Task t,byte[] logs)
		throws GridException {

		if (t == null) {
			throw new GridException("preemptLocalTaskOnLocalResource(): "+
									"illegal Task");
		}
		if (logs == null) {
			throw new GridException("preemptLocalTaskOnLocalResource(): "+
									"illegal logs array");
		}
		
		// ****************

		// [TIMEOUT]
		long xstart_time = this.running_local_tasks.removeRunningTask(t);
		//	incorrect code -- long xstart_time = now-t.output().runtime();

		// ****************

		// Account Local non-activity
		this.managed_peer.register().getNeighborhoodBarteringProfile().
			accountLocalActivityCancellation(t);
		
		// Free Resource (Peer ResourceManager viewpoint)
		long rms_xstart_time =
			this.managed_peer.rms().preemptBusyLocalResource(t);
		
		if ((xstart_time != rms_xstart_time) &&
			(LightweightBarteringGrid.getMode() == SIM_MODE)) {
			throw new GridException("preemptLocalTaskOnLocalResource(): "+
									"XMS X start time="+xstart_time+
									", RMS X start_time="+rms_xstart_time);
		}

		// Free Request (Peer QueueManager viewpoint)
		long runtime = LightweightBarteringGrid.getTime() - xstart_time;
		assert(runtime >= 0);
		this.managed_peer.qms().
			requeuePreemptedLocalTask(t,xstart_time,runtime,t.consumerID());

		// ****************

		// Update stats
		this.cancelled_local_tasks_total_runtime +=
			LightweightBarteringGrid.getTime()-xstart_time;
		this.cancelled_local_tasks_count++;

		// ****************

	}
	
	/** The managed Peer instructs itself to preempt a Supplying Task
	 * running locally.
	 * <p>Note: Task preemption is non-terminating cancellation of a Task.
	 * <p>Note: as opposed to the pair of cancelXTaskOnXResource() methods,
	 * there is no symmetry for preemptSupplyingTaskOnLocalResource(),
	 * because there is no interest to announce delays in Supplying Task
	 * execution through preemptLocalTaskOnSupplyingResource().
	 */
	public synchronized void
		preemptSupplyingTaskOnLocalResource(Task t,byte[] logs)
			throws GridException {

		if (t == null) {
			throw new GridException("preemptSupplyingTaskOnLocalResource(): "+
									"illegal Task");
		}
		if (logs == null) {
			throw new GridException("preemptSupplyingTaskOnLocalResource(): "+
									"illegal logs array");
		}

		// ****************

		boolean is_supplying_task_running;
		try {
			is_supplying_task_running =
				this.managed_peer.qms().isSupplyingTaskRunning(t);
		}
		catch (GridException e) {
			// Task is not queued, incoming preemption request is ignored
			return;
		}
		if (is_supplying_task_running == false) {
			// Task is not actually running: it's pretty easy to preempt it :)
			return;
		}

		// Task is running, go on
		
		// ****************
		
		// [TIMEOUT]
		long xstart_time = this.running_supplying_tasks.removeRunningTask(t);
		//	incorrect code -- long xstart_time = now-t.output().runtime();

		// ****************

		// Free Resource (Peer ResourceManager viewpoint)
		long rms_xstart_time =
			this.managed_peer.rms().preemptBusySupplyingResource(t);

		if ((xstart_time != rms_xstart_time) &&
			(LightweightBarteringGrid.getMode() == SIM_MODE)) {
			throw new GridException("preemptSupplyingTaskOnLocalResource(): "+
									"XMS X start time="+xstart_time+
									", RMS X start_time="+rms_xstart_time);
		}
		
		// Free Request (Peer QueueManager viewpoint)
		this.managed_peer.qms().requeuePreemptedSupplyingTask(t,xstart_time);

		// Account non-Supplying
		this.managed_peer.register().getNeighborhoodBarteringProfile().
			accountSupplyingCancellation(t,xstart_time);

		// ****************
		
		// Update stats
		this.cancelled_supplying_tasks_total_runtime +=
			LightweightBarteringGrid.getTime()-xstart_time;
		this.cancelled_supplying_tasks_count++;

		// ****************

	}
	
	/** Asks this (Supplier) Peer to cancel target Supplying Task.
	 * <p>Note: cancelSupplyingTaskOnLocalResource() would be called
	 * by this Peer, or by another Peer (through the ExternalPeerHandle).
	 * <p><ul>
	 * <li>Cancellation initiated by this Peer.
	 * In case of Resource preemption.
	 * <li>Cancellation initiated by the Consumer Peer.
	 * If another (Consumer) Peer does not need this (Supplier) Peer
	 * to complete the Task, it can politely tell to this (Supplier) Peer
	 * through its ExternalPeerHandle so that it does not waste Resources.
	 * An example when this would be needed is when the Consumer Peer
	 * becomes impatient when the time spent after submission time
	 * has reached some (internal to Consumer Peer) timeout.
	 * </ul>
	 * <p>Note: Task cancellation is terminating cancellation of a Task.
	 */
	public synchronized void cancelSupplyingTaskOnLocalResource(Task t)
		throws GridException {

		if (t == null) {
			throw new GridException("cancelSupplyingTaskOnLocalResource(): "+
									"illegal Task");
		}

		// ****************

		boolean is_supplying_task_running;
		try {
			is_supplying_task_running =
				this.managed_peer.qms().isSupplyingTaskRunning(t);
		}
		catch (GridException e) {
			// Task is not queued, incoming cancellation request is ignored
			return;
		}
		if (is_supplying_task_running == false) {
			// Task is not actually running: it's pretty easy to cancel it :)

			// Free Request (Peer QueueManager viewpoint)
			this.managed_peer.qms().dequeueWaitingSupplyingTask(false,t);
			return;
		}

		// Task is running, go on
		
		// ****************
		
		// [TIMEOUT]
		long xstart_time = this.running_supplying_tasks.removeRunningTask(t);
		//	incorrect code -- long xstart_time = now-t.output().runtime();

		// ****************

		// Free Resource (Peer ResourceManager viewpoint)
		long rms_xstart_time = -1;
		try {
			rms_xstart_time =
				this.managed_peer.rms().preemptBusySupplyingResource(t);
		}
		catch (GridException e) {
			// Resource was already idle, so no Task could have been preempted
			return;
		}

		if ((xstart_time != rms_xstart_time) &&
			(LightweightBarteringGrid.getMode() == SIM_MODE)) {
			throw new GridException("cancelSupplyingTaskOnLocalResource(): "+
									"XMS X start time="+xstart_time+
									", RMS X start_time="+rms_xstart_time);
		}
		
		// Free Request (Peer QueueManager viewpoint)
		this.managed_peer.qms().dequeueCancelledSupplyingTask(t,xstart_time);

		// Account non-Supplying
		this.managed_peer.register().getNeighborhoodBarteringProfile().
			accountSupplyingCancellation(t,xstart_time);

		// ****************

		// Update stats
		this.cancelled_supplying_tasks_total_runtime +=
			LightweightBarteringGrid.getTime()-xstart_time;
		this.cancelled_supplying_tasks_count++;
		
		// ****************

	}

	/** Notifies Consumer Peer of the cancellation of target Task
	 * on the Supplier Peer.
	 * <p>Note: cancelLocalTaskOnSupplyingResource() would be called
	 * by this Peer, or by another Peer (through the ExternalPeerHandle).
	 * <p><ul>
	 * <li>Cancellation initiated by this Peer.
	 * In case of timeout.
	 * <li>Cancellation initiated by the Supplier Peer.
	 * In case of Resource preemption.
	 * </ul>
	 * <p>Note: Task cancellation is terminating cancellation of a Task.
	 */
	public synchronized void
		cancelLocalTaskOnSupplyingResource(Task t) throws GridException {
		
		if (t == null) {
			throw new GridException("cancelLocalTaskOnSupplyingResource(): "+
									"illegal Task");
		}
		
		String supplier_id = this.running_consumption_tasks.identifySupplier(t);
		
		if (supplier_id == null) {
			throw new GridException("cancelLocalTaskOnSupplyingResource(): "+
									"corrupted internal data, target "+
									"Consumption Task not currently running");
		}
		
		cancelLocalTaskOnSupplyingResource(supplier_id,t);

	}
	
	/** Notifies Consumer Peer of the cancellation of target Task
	 * on the Supplier Peer.
	 * <p>Note: cancelLocalTaskOnSupplyingResource() would be called
	 * by this Peer, or by another Peer (through the ExternalPeerHandle).
	 * <p><ul>
	 * <li>Cancellation initiated by this Peer.
	 * In case of timeout.
	 * <li>Cancellation initiated by the Supplier Peer.
	 * In case of Resource preemption.
	 * </ul>
	 * <p>Note: Task cancellation is terminating cancellation of a Task.
	 */
	public synchronized void
		cancelLocalTaskOnSupplyingResource(String supplier_id,Task t)
		throws GridException {
		
		if (t == null) {
			throw new GridException("cancelLocalTaskOnSupplyingResource(): "+
									"illegal Task");
		}
		if (supplier_id == null) {
			throw new GridException("cancelLocalTaskOnSupplyingResource(): "+
									"illegal supplier id");
		}
		
		// ****************

		// [TIMEOUT]
		long xstart_time = -1;
		
		try {
			xstart_time = this.running_consumption_tasks.removeRunningTask(t);
		}
		catch (GridException e) {
			return;
		}
		
		// ****************

		// Account non-Consumption
		this.managed_peer.register().getNeighborhoodBarteringProfile().
			accountConsumptionCancellation(supplier_id,t,xstart_time);
		
		// Free Request (Peer QueueManager viewpoint)
		long runtime = LightweightBarteringGrid.getTime() - xstart_time;
		assert(runtime >= 0);
		this.managed_peer.qms().
			requeuePreemptedConsumptionTask(t,xstart_time,runtime,supplier_id);
		
		// ****************

		// Update stats
		this.cancelled_consumption_tasks_total_runtime +=
			LightweightBarteringGrid.getTime()-xstart_time;
		this.cancelled_consumption_tasks_count++;

		// ****************

		this.managed_peer.qms().preemptWaitingSupplyingTasks();

	}
	
	// ****************************************************************

	/** Cancels timed-out Consumption Tasks.
	 * @return true if at least one Consumption Task has been cancelled
	 */
	public synchronized boolean processTimedOutConsumptionTasks()
		throws GridException {

		return this.running_consumption_tasks.processTimedOutTasks();
	
	}
	
	/** Cancels timed-out Local Tasks.
	 * @return true if at least one Local Task has been cancelled
	 */
	public synchronized boolean processTimedOutLocalTasks()
		throws GridException {

		return this.running_local_tasks.processTimedOutTasks();
	
	}

	/** Cancels timed-out Supplying Tasks.
	 * @return true if at least one Supplying Task has been cancelled
	 */
	public synchronized boolean processTimedOutSupplyingTasks()
		throws GridException {

		return this.running_supplying_tasks.processTimedOutTasks();
	
	}

	// ****************************************************************

	public synchronized int countRunningLocalTasks()
		throws GridException {
		
		return this.running_local_tasks.countRunningTasks();
		
	}

	public synchronized int countRunningSupplyingTasks()
		throws GridException {
		
		return this.running_supplying_tasks.countRunningTasks();
		
	}
	
	public synchronized int countRunningConsumptionTasks()
		throws GridException {
		
		return this.running_consumption_tasks.countRunningTasks();
		
	}
	
	/** Lists running Local Tasks, in decreasing order of execution time
	 * (longest-running first).
	 * @return running Local Tasks, in decreasing order of execution time
	 */
	public synchronized Task[] listRunningLocalTasks() {
		
		return this.running_local_tasks.listRunningTasks();
		
	}

	/** Lists running Supplying Tasks, in decreasing order of execution time
	 * (longest-running first).
	 * @return running Supplying Tasks, in decreasing order of execution time
	 */
	public synchronized Task[] listRunningSupplyingTasks() {
		
		return this.running_supplying_tasks.listRunningTasks();
		
	}

	/** Lists running Consumption Tasks, in decreasing order of execution time
	 * (longest-running first).
	 * @return running Consumption Tasks, in decreasing order of execution time
	 */
	public synchronized Task[] listRunningConsumptionTasks() {
		
		return this.running_consumption_tasks.listRunningTasks();
		
	}
	
	// ****************************************************************

	private final Peer managed_peer;

	private long run_local_tasks_count;
	private long run_supplying_tasks_count;
	private long run_consumption_tasks_count;

	private long cancelled_local_tasks_total_runtime;
	private long cancelled_local_tasks_count;
	private long cancelled_supplying_tasks_total_runtime;
	private long cancelled_supplying_tasks_count;
	private long cancelled_consumption_tasks_total_runtime;
	private long cancelled_consumption_tasks_count;

	private long completed_local_tasks_total_runtime;
	private long completed_local_tasks_count;
	private long completed_supplying_tasks_total_runtime;
	private long completed_supplying_tasks_count;
	private long completed_consumption_tasks_total_runtime;
	private long completed_consumption_tasks_count;

	// [TIMEOUT]
	private final long consumption_timeout;
	private final RunningTaskManager running_consumption_tasks;

	private final long local_timeout;
	private final RunningTaskManager running_local_tasks;

	private final long supplying_timeout;
	private final RunningTaskManager running_supplying_tasks;

	/* Impact of timeout management on ExecutionManager methods:
	   
	   [L,S]  runLocalTaskOnLocalResource
	   [L,S]  runSupplyingTaskOnLocalResource
	   [C]   *runLocalTaskOnSupplyingResource
	   [L,S]  completeLocalTaskOnLocalResource
	   [L,S]  completeSupplyingTaskOnLocalResource
	   [C]   *completeLocalTaskOnSupplyingResource
	   [L,S]  preemptLocalTaskOnLocalResource
	   [L,S]  preemptSupplyingTaskOnLocalResource
	   [L,S]  cancelSupplyingTaskOnLocalResource
	   [C]   *cancelLocalTaskOnSupplyingResource
	   
	   Implementation note:
	   only Consumption Tasks (* methods) are currently timeout-managed
	   
	 */
	
	// ****************************************************************

	// 1-month limit == 2592000s, 1-week limit == 604800s
	private static final long LOCAL_TIMEOUT = 604800;

}

