/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LocalTaskRuntimeData.java

package lbg.peer.core.xms;

import lbg.*;
import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.service.*;
import static lbg.common.math.CommonMath.generateHashCode;
import lbg.task.*;

/**
 * @author Cyril Briquet
 */
public class LocalTaskRuntimeData extends TaskRuntimeData {
	
	public LocalTaskRuntimeData(Task t,long time_out,
								ExternalPeerHandle peer_handle,
								String supplier_id,
								boolean do_cancel)
		throws GridException {
		
		super();

		if (t == null) {
			throw new GridException("LocalTaskRuntimeData(): "+
									"illegal Task");
		}
		if (time_out <= 0) {
			throw new GridException("LocalTaskRuntimeData(): "+
									"illegal timeout");
		}
		if (peer_handle != null) {
			throw new GridException("LocalTaskRuntimeData(): "+
									"not a dummy Peer handle 8)");
		}
		if (supplier_id == null) {
			throw new GridException("LocalTaskRuntimeData(): "+
									"illegal supplier id");
		}
		
		long now = LightweightBarteringGrid.getTime();
		this.task = t;
		this.start_time = now;
		this.time_out = now + time_out;
		this.supplier_id = supplier_id;
		this.do_cancel = do_cancel;
		this.hashcode = Integer.MIN_VALUE;
		this.tag_modified = false;
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LocalTaskRuntimeData]\n");
		
		sb.append("  ->  ");
		sb.append("Task ");
		sb.append(this.task.id());
		sb.append(", supplier=");
		sb.append(this.supplier_id);
		sb.append(", timeout=");
		sb.append(this.time_out);
		sb.append(", action=");
		sb.append(this.do_cancel ? "cancel" : "preempt");

		return sb.toString();
		
	}

	@Override
	public synchronized int compareTo(TaskRuntimeData other) {
	
		long dummy = this.time_out - other.getTimeOut();
		return (dummy == 0) ?
			/* same timeout */ task_cmp.compare(this.task,other.getTask())
			:
			/* different timeout */ ((dummy < 0) ? -1 : +1) ;
		
	}

	@Override
	public synchronized boolean equals(Object obj) {
		
		if (obj instanceof LocalTaskRuntimeData == false) {
			return false;
		}
		
		return (this.compareTo((TaskRuntimeData)obj) == 0);
		
	}

	@Override
	public synchronized int hashCode() {
	
		if (this.hashcode == Integer.MIN_VALUE) {
			this.hashcode = generateHashCode(this.task.jobID(),this.task.id());
		}
		
		return generateHashCode(generateHashCode(this.time_out),this.hashcode);
	
	}

	// ****************************************************************

	public Task getTask() {
		
		return this.task;
		
	}
	
	public long getStartTime() {
		
		return this.start_time;
		
	}

	public synchronized long getTimeOut() {
		
		return this.time_out;
		
	}
	
	public synchronized void setTimeOut(long time_out) throws GridException {

		if (time_out < 0) {
			throw new GridException("setTimeOut(): illegal timeout");
		}

		long now = LightweightBarteringGrid.getTime();
		this.time_out = now+time_out;
		
	}

	/**
	 * @return null
	 */
	public ExternalPeerHandle getPeerHandle() {
		
		return null;
		
	}
	
	public String getSupplierID() {
		
		return this.supplier_id;
		
	}
	
	public synchronized boolean getTimeOutActionTag() {
		
		return this.do_cancel;
		
	}
	
	public synchronized void setTimeOutActionTag(boolean do_cancel) {
		
		this.do_cancel = do_cancel;
		this.tag_modified = true;
		
	}
	
	public synchronized boolean isTimeOutActionTagModified() {
		
		return this.tag_modified;
		
	}

	// ****************************************************************

	private final Task task;
	private final long start_time;
	private long time_out;
	private boolean do_cancel;
	private final String supplier_id;
	private int hashcode;
	private boolean tag_modified;
	
	// ****************************************************************

	private final LocalTaskComparator task_cmp =
		LocalTaskComparator.getDefaultComparator();

}

