/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// RunningTaskManager.java

package lbg.peer.core.xms;

import lbg.*;
import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.service.*;
import lbg.task.*;
import java.util.*;

/**
 * @author Cyril Briquet
 */
public abstract class RunningTaskManager {

	RunningTaskManager(Peer managed_peer,TaskComparator task_cmp,
					   String task_runtime_data_type)
		throws GridException {
		
		if (managed_peer == null) {
			throw new GridException("RunningTaskManager(): "+
									"illegal Peer");
		}
		if (task_cmp == null) {
			throw new GridException("RunningTaskManager(): "+
									"illegal Task comparator");
		}
		if (task_runtime_data_type == null) {
			throw new GridException("CommonConsumptionScheduling(): "+
									"illegal TaskRuntimeData factory type");
		}
		
		this.managed_peer = managed_peer;
		this.running_tasks_set = new TreeMap<Task,TaskRuntimeData>(task_cmp);
		this.running_tasks_timeouts = new TreeSet<TaskRuntimeData>();
		this.trdf = TaskRuntimeDataFactory.valueOf(task_runtime_data_type);

	}


	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[RunningTaskManager ");
		sb.append(this.managed_peer.id());
		sb.append("]\n");

		sb.append("  ->  ");
		sb.append(this.running_tasks_set.size());
		sb.append(" managed running Consumption Tasks");
		
		return sb.toString();
		
	}
	
	// ****************************************************************

	/** Adds target Task to this running Tasks Manager.
	 * @param t target Task
	 * @param timeout timeout value
	 * @param peer_handle handle of Peer to contact
	 * @param supplier_id supplier id
	 * @param do_cancel true to cancel the Task upon time-out,
	 * false to only preempt the Task upon time-out
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized void addRunningTask(Task t,long timeout,
								 			ExternalPeerHandle peer_handle,
											String supplier_id,
											boolean do_cancel)
		throws GridException {
	
		assert(twoSetsInvariant());
		
		TaskRuntimeData trd = this.trdf.
			createTaskRuntimeData(t,timeout,peer_handle,supplier_id,do_cancel);
		
		if (this.running_tasks_set.put(t,trd) != null) {
			throw new GridException("addRunningTask(): "+
									"Task already managed (1)");
		}
		if (this.running_tasks_timeouts.add(trd) == false) {
			throw new GridException("addRunningTask(): "+
									"Task already managed (2)");
		}
		
		assert(twoSetsInvariant());

	}
	
	/** Removes target Task from this running Tasks Manager.
	 * @param t target Task
	 * @return Task start time
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized long removeRunningTask(Task t) throws GridException {
	
		assert(twoSetsInvariant());

		if (t == null) {
			throw new GridException("removeRunningTask(): illegal Task");
		}
		
		TaskRuntimeData trd = this.running_tasks_set.remove(t);
		
		if (trd == null) {
			throw new GridException("removeRunningTask(): corrupted data (1)");
		}
		if (this.running_tasks_timeouts.remove(trd) == false) {
			throw new GridException("removeRunningTask(): corrupted data (2) "+
									trd);
		}
		
		assert(twoSetsInvariant());

		return trd.getStartTime();
		
	}

	// ****************************************************************

	/** Identifies which Peer is running the target Task.
	 * @param t target Task
	 * @return Peer running the Task, or null no such Task is running
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized String identifySupplier(Task t) throws GridException {
		
		if (t == null) {
			throw new GridException("identifySupplier(): illegal Task");
		}
		
		TaskRuntimeData trd = this.running_tasks_set.get(t);
		
		return (trd != null) ? trd.getSupplierID() : null;

	}

	/** Returns the start time of target Task's execution.
	 * @param t target Task
	 * @return start time
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized long getStartTime(Task t) throws GridException {
		
		if (t == null) {
			throw new GridException("getStartTime(): illegal Task");
		}
		
		TaskRuntimeData trd = this.running_tasks_set.get(t);
		
		if (trd == null) {
			throw new GridException("getStartTime(): target Task "+
									"currently not running");
		}
		
		return trd.getStartTime();

	}

	// ****************************************************************

	/** Updates the time-out action tag associated with the target Task.
	 * @param t target Task
	 * @param do_cancel time-out action tag
	 * @return true if the action tag has been updated,
	 * or false if it were not updated
	 * (i.e. the action tag has already been updated)
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized boolean updateTimeOutActionTag(Task t,boolean do_cancel)
		throws GridException {
		
		if (t == null) {
			throw new GridException("updateTimeOutActionTag(): illegal Task");
		}
		
		TaskRuntimeData trd = this.running_tasks_set.get(t);
		
		if (trd == null) {
			throw new GridException("updateTimeOutActionTag(): target Task "+
									"currently not running");
		}
		
		boolean retval = (trd.isTimeOutActionTagModified() == false);
		if (retval == true) {
			trd.setTimeOutActionTag(do_cancel);
		}
		
		return retval;

	}

	/** Updates the time-out associated with the target Task.
	 * @param t target Task
	 * @param time_out time-out
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized void updateTimeOutValue(Task t,long time_out)
		throws GridException {
		
		if (t == null) {
			throw new GridException("updateTimeOutValue(): illegal Task");
		}
		if (time_out < 0) {
			throw new GridException("updateTimeOutValue(): "+
									"illegal time-out");
		}
		
		TaskRuntimeData trd = this.running_tasks_set.get(t);
		
		if (trd == null) {
			throw new GridException("updateTimeOutValue(): target Task "+
									"currently not running");
		}
		
		if (this.running_tasks_timeouts.remove(trd) == false) {
			throw new GridException("updateTimeOutValue(): "+
									"corrupted data (1)");
		}

		trd.setTimeOut(time_out);
		
		if (this.running_tasks_timeouts.add(trd) == false) {
			throw new GridException("updateTimeOutValue(): "+
									"corrupted data (2)");
		}

	}
	
	/** Updates the time-out and action tag associated with the target Task.
	 * @param t target Task
	 * @param do_cancel time-out action tag
	 * @param time_out time-out
	 * @return true if the time-out and action tag have been updated,
	 * or false if they were not updated
	 * (i.e. the action tag has already been updated)
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized boolean
		updateTimeOutActionTagAndValue(Task t,boolean do_cancel,long time_out)
		throws GridException {
		
		if (t == null) {
			throw new GridException("updateTimeOutActionTagAndValue(): "+
									"illegal Task");
		}
		if (time_out < 0) {
			throw new GridException("updateTimeOutActionTagAndValue(): "+
									"illegal time-out");
		}
		
		TaskRuntimeData trd = this.running_tasks_set.get(t);
		
		if (trd == null) {
			throw new GridException("updateTimeOutActionTagAndValue(): "+
									"target Task currently not running");
		}
		
		if (this.running_tasks_timeouts.remove(trd) == false) {
			throw new GridException("updateTimeOutActionTagAndValue(): "+
									"corrupted data (1)");
		}

		boolean retval = (trd.isTimeOutActionTagModified() == false);
		if (retval == true) {
			trd.setTimeOutActionTag(do_cancel);
			trd.setTimeOut(time_out);
		}

		if (this.running_tasks_timeouts.add(trd) == false) {
			throw new GridException("updateTimeOutActionTagAndValue(): "+
									"corrupted data (2)");
		}
		
		return retval;

	}
	
	/** Identifies if target Task has been graced
	 * @param t target Task
	 * @return true if target Task has been graced
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized boolean isTaskGraced(Task t)
		throws GridException {
		
		TaskRuntimeData trd = this.running_tasks_set.get(t);
		
		if (trd == null) {
			throw new GridException("isTaskGraced(): "+
									"target Task currently not running");
		}
		
		return trd.isTimeOutActionTagModified();

	}

	// ****************************************************************

	public synchronized boolean areSomeTasksTimedOut() throws GridException {
		
		if (this.running_tasks_timeouts.size() == 0) {
			return false;
		}
		
		TaskRuntimeData trd = this.running_tasks_timeouts.first();
		long time_out = trd.getTimeOut();
		
		return (time_out < LightweightBarteringGrid.getTime());
		
	}

	public synchronized boolean areSomeTasksRunning() throws GridException {
		
		return (this.running_tasks_timeouts.size() > 0);

	}
	
	public synchronized int countRunningTasks() throws GridException {
	
		return this.running_tasks_timeouts.size();
		
	}

	/** Lists running Tasks, in decreasing order of execution time
	 * (longest-running first).
	 * @return running Tasks, in decreasing order of execution time
	 */
	public synchronized Task[] listRunningTasks() {
		
		Task[] running_tasks_array = new Task[this.running_tasks_set.size()];
		Iterator<Task> it = this.running_tasks_set.keySet().iterator();
		
		int i = 0;
		while (it.hasNext()) {
			running_tasks_array[i++] = it.next();
		}
		
		return running_tasks_array;
		
	}

	/** Returns the most recently running Task that has not been graced
	 * (useful for PSufferage Resource selection to preempt
	 * a running Supplying Task).
	 * @return most recently running Task that has not been graced,
	 * or null if there is no running Task ungraced
	 */
	public synchronized Task getMostRecentlyRunningAndUngracedTask()
		throws GridException {

		// I'm feeling lucky: the Task with the highest time-out value
		//                    is probably ungraced, so check it first
		TaskRuntimeData trd = this.running_tasks_timeouts.last();
		Task ungraced_t = trd.getTask();
		if (isTaskGraced(ungraced_t) == false) {
			return ungraced_t;
		}
		
		// Iterate over running Tasks by increasing time-out value,
		// and return the ungraced Task with the highest time-out value
		Iterator<TaskRuntimeData> it = this.running_tasks_timeouts.iterator();
		ungraced_t = null; // in case all running Task are graced
		while (it.hasNext()) {
			trd = it.next();
			Task cur_t = trd.getTask();
			if (isTaskGraced(cur_t) == false) {
				ungraced_t = cur_t;
			}
		}

		/* implementation notes -- please read before patching to optimize
		   
		   in the corner case of periods of time
		   where lots of Supplying Tasks are graced
		   and there are a lot of preemption attempts
		   (i.e. lots of calls to this algorithm, with few ungraced Tasks),
		   then this search algorithm is clearly not the most efficient;
		   however, such periods are necessarily short
		   (except if the Peer is unduly configured with a long grace peiod), so
		   the benefits of maintaining further data structures is probably low;
		   moreover, there is indeed a penalty to maintain such data structures
		   and they yield no benefit in the general case;
		   therefore, we do not optimize for this corner case
		   
		   in the corner case where
		     ((ResourceManager.supplying_timeout <
			   AdaptivePreemption.GRACE_PERIOD) &&
		      (RunningTaskManager.running_tasks_timeouts.size() is high))
		   then this search algorithm is not the most efficient either;
		   however, one can remark there is no benefit in configuring a Peer
		   with a supplying time-out shorter than the preemption grace period,
		   so we do not optimize for this irrelevant corner case
		*/
	
		return ungraced_t;
		
	}

	// ****************************************************************
	
	/** Processes timed-out Tasks.
	 * @return true if at least one Task has been preempted or cancelled
	 */
	public abstract boolean processTimedOutTasks() throws GridException;

	// ****************************************************************

	synchronized ArrayList<TaskRuntimeData> listTimedOutTasks()
		throws GridException {
	
		ArrayList<TaskRuntimeData> al = new ArrayList<TaskRuntimeData>();
		
		// Iterate over running Tasks until 1 non timed-out Task is found
		// Precondition: running Tasks sorted by increasing timeout
		for (TaskRuntimeData trd : this.running_tasks_timeouts) {
			long time_out = trd.getTimeOut();
			
			if (time_out < LightweightBarteringGrid.getTime()) {
				// Timeout !
				al.add(trd);
			}
			else {
				// No timeout, break
				break;
			}
		}
		
		return al;
		
	}
	
	// ****************************************************************
	
	private synchronized boolean twoSetsInvariant() {
	
		return this.running_tasks_set.size() ==
			   this.running_tasks_timeouts.size();
		
	}

	// ****************************************************************

	private final Peer managed_peer;
	private final TreeMap<Task,TaskRuntimeData> running_tasks_set;
	private final TreeSet<TaskRuntimeData> running_tasks_timeouts;
	private final TaskRuntimeDataFactory trdf;

}

