/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// NeighborhoodBarteringProfile.java

package lbg.peer.register;

import lbg.*;
import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.*;
import lbg.peer.register.bartering.*;
import lbg.task.*;

/**
 * @author Cyril Briquet
 */
public class NeighborhoodBarteringProfile {

	public NeighborhoodBarteringProfile(Peer managed_peer,PeerRegister grid_register)
		throws GridException {
	
		if (managed_peer == null) {
			throw new GridException("NeighborhoodBarteringProfile(): illegal Peer");
		}
		if (grid_register == null) {
			throw new GridException("NeighborhoodBarteringProfile(): "+
									"illegal PeerRegister");
		}
		
		this.managed_peer = managed_peer;
		this.grid_register = grid_register;

		this.total_consumed_local_power = 0;
		this.total_supplied_local_power = 0;
		this.total_consumed_external_power = 0;
		this.m_supplying_completion_runtime =
			MeasurementFactory.createDescriptiveMeasurement(
				"Grid-level Supplying Tasks completion times",100);
		this.total_completed_local_tasks = 0;
		this.total_completed_supplied_tasks = 0;
		this.total_completed_consumed_tasks = 0;
		this.total_cancelled_local_tasks = 0;
		this.total_cancelled_supplied_tasks = 0;
		this.total_cancelled_consumed_tasks = 0;

	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[NeighborhoodBarteringProfile ");
		sb.append(this.managed_peer.id());
		sb.append("]");
		
		// FR = consumed external power / supplied local power
		sb.append("\n  ->  ");
		sb.append("{SIM} Favor Ratio = ");
		sb.append(this.total_consumed_external_power);
		sb.append(" / ");
		sb.append(this.total_supplied_local_power);
		sb.append(" = ");
		sb.append(CommonMath.div_label_2dd(this.total_consumed_external_power,
										   this.total_supplied_local_power));
		sb.append("\n  ~>  ");
		sb.append(this.managed_peer.id());
		sb.append(".favors_ratio = ");
		sb.append(((double)this.total_consumed_external_power)/
				  this.total_supplied_local_power);
		// RG = consumed local power + external power / consumed local power
		sb.append("\n  ->  ");
		sb.append("{SIM} Resource Gain = ");
		sb.append(CommonMath.div_label_2dd((this.total_consumed_local_power+
					 						this.total_consumed_external_power),
										   (this.total_consumed_local_power)));

		// Power repartition
		sb.append("\n  ->  power repartition { ");
		long total_used_power = this.total_consumed_local_power +
			this.total_supplied_local_power+this.total_consumed_external_power;
		sb.append("Local power=");
		sb.append(this.total_consumed_local_power);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.total_consumed_local_power,
									   total_used_power,100));
		sb.append("%), Supplied power=");
		sb.append(this.total_supplied_local_power);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.total_supplied_local_power,
									   total_used_power,100));
		sb.append("%), Consumed power=");
		sb.append(this.total_consumed_external_power);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.total_consumed_external_power,
									   total_used_power,100));
		sb.append("%) }");

		return sb.toString();
		
	}
	
	// ****************************************************************

	public synchronized void accountLocalActivityCompletion(Task t)
		throws GridException {

		if (t == null) {
			throw new GridException("accountLocalActivityCompletion(): "+
									"illegal Task");
		}
		if (t.consumerID().equals(this.managed_peer.id()) == false) {
			throw new GridException("accountLocalActivityCompletion(): "+
									"external referencing");
		}

		// Store local power consumption/self-supplying
		// so as to estimate own performance
		TaskOutput task_output = t.output();
		long runtime = task_output.runtime();
		long consumed_local_power =
			((long)task_output.resourcePower()) * runtime;
		this.total_consumed_local_power += consumed_local_power;

		this.total_completed_local_tasks++;

	}

	public synchronized void accountSupplyingCompletion(Task t,long xstart_time)
		throws GridException {
		
		if (t == null) {
			throw new GridException("accountSupplyingCompletion(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("accountSupplyingCompletion(): "+
									"illegal execution start time");
		}

		TaskOutput task_output = t.output();
		long runtime = task_output.runtime();
		String consumer_id = t.consumerID();
		if (consumer_id.equals(this.managed_peer.id())) {
			throw new GridException("accountSupplyingCompletion(): "+
									"self-referencing");
		}
		PeerProfile peer_profile = this.grid_register.getProfile(consumer_id);
		
		// Store local power Supplying (computed with perfect accounting)
		// so as to estimate own performance
		long supplied_local_power =
			((long)task_output.resourcePower()) * runtime;
		this.total_supplied_local_power += supplied_local_power;

		// Account Supplying completion
		PeerBarteringProfile bartering_profile =
			peer_profile.getBarteringProfile();
		bartering_profile.
			accountSupplyingCompletion(t,xstart_time,supplied_local_power);

		long now = LightweightBarteringGrid.getTime();
		long xruntime = now-xstart_time;
		this.m_supplying_completion_runtime.addObservation(xruntime);
		
		this.total_completed_supplied_tasks++;

	}

	public synchronized void
		accountConsumptionCompletion(Task t,long xstart_time)
		throws GridException {
		
		if (t == null) {
			throw new GridException("accountConsumptionCompletion(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("accountConsumptionCompletion(): "+
									"illegal execution start time");
		}

		TaskOutput task_output = t.output();
		long runtime = task_output.runtime();
		String supplier_id = task_output.supplierID();
		if (supplier_id.equals(this.managed_peer.id())) {
			throw new GridException("accountConsumptionCompletion(): "+
									"self-referencing");
		}
		PeerProfile peer_profile = this.grid_register.getProfile(supplier_id);
		 
		// Store supplied power Consumption (computed with perfect accounting)
		// so as to estimate own performance
		long consumed_external_power =
			((long)task_output.resourcePower()) * runtime;
		this.total_consumed_external_power += consumed_external_power;
		
		// Account Consumption completion
		PeerBarteringProfile bartering_profile =
			peer_profile.getBarteringProfile();
		bartering_profile.
			accountConsumptionCompletion(t,xstart_time,consumed_external_power);
		
		this.total_completed_consumed_tasks++;

	}
	
	// ****************************************************************

	public synchronized void accountLocalActivityCancellation(Task t)
		throws GridException {

		if (t == null) {
			throw new GridException("accountLocalActivityCancellation(): "+
									"illegal Task");
		}
		
		this.total_cancelled_local_tasks++;

	}

	public synchronized void
		accountSupplyingCancellation(Task t,long start_time)
		throws GridException {

		if (t == null) {
			throw new GridException("accountSupplyingCancellation(): "+
									"illegal Task");
		}
		if (start_time < 0) {
			throw new GridException("accountSupplyingCancellation(): "+
									"illegal start time");
		}

		String consumer_id = t.consumerID();
		if (consumer_id.equals(this.managed_peer.id())) {
			throw new GridException("accountSupplyingCancellation(): "+
									"self-referencing");
		}
		PeerProfile peer_profile = this.grid_register.getProfile(consumer_id);

		// Account Supplying completion
		PeerBarteringProfile bartering_profile =
			peer_profile.getBarteringProfile();
		bartering_profile.accountSupplyingCancellation(t,start_time);
		
		this.total_cancelled_supplied_tasks++;
		
	}
	
	public synchronized void
		accountConsumptionCancellation(String supplier_id,Task t,long start_time)
		throws GridException {
		
		if (supplier_id == null) {
			throw new GridException("accountConsumptionCancellation(): "+
									"illegal Supplier ID");
		}
		if (start_time < 0) {
			throw new GridException("accountConsumptionCancellation(): "+
									"illegal start time");
		}
		
		if (t == null) {
			throw new GridException("accountConsumptionCancellation(): "+
									"illegal Task");
		}

		if (supplier_id.equals(this.managed_peer.id())) {
			throw new GridException("accountConsumptionCancellation(): "+
									"self-referencing");
		}
		PeerProfile peer_profile = this.grid_register.getProfile(supplier_id);

		// Account Consumption completion
		PeerBarteringProfile bartering_profile =
			peer_profile.getBarteringProfile();
		bartering_profile.accountConsumptionCancellation(t,start_time);

		this.total_cancelled_consumed_tasks++;

	}
	
	// ****************************************************************

	public synchronized long getTotalConsumedLocalPower() {

		return this.total_consumed_local_power;
		
	}
	
	/** Returns Favor Ratio of the managed Peer.
	 * <p>Favor Ratio = consumed external power / supplied local power
	 */
	public synchronized double getFavorRatio() {
		
		return 
			((double)this.total_consumed_external_power) /
			((double)this.total_supplied_local_power);
		
	}

	/** Returns Resource Gain of the managed Peer.
	 * <p>Resource Gain =
	 * consumed local power + external power / consumed local power
	 */
	public synchronized double getResourceGain() {
		
		long dummy = 
			this.total_consumed_local_power+this.total_consumed_external_power;
		
		return ((double)dummy) / ((double)this.total_consumed_local_power);
		
	}
	
	// ****************************************************************

	/**
	 * @return an estimate of the runtime of a Supplying Task
	 * over all profiled Peers (>= 0.0), or -1.0 if no estimate is available
	 */
	public synchronized double estimateSupplyingTaskRuntime() {
	
		if (this.m_supplying_completion_runtime.getObservations() > 1) {
			return this.m_supplying_completion_runtime.getMean()+
				   this.m_supplying_completion_runtime.getStandardDeviation();
		}
		else {
			return -1.0;
		}
		
	}
	
	// ****************************************************************

	private Peer managed_peer;
	private PeerRegister grid_register;

	// Bartering outcome data
	private long total_consumed_local_power;
	private long total_supplied_local_power;
	private long total_consumed_external_power;

	private final Measurement m_supplying_completion_runtime;

	private long total_completed_local_tasks;
	private long total_completed_supplied_tasks;
	private long total_completed_consumed_tasks;

	private long total_cancelled_local_tasks;
	private long total_cancelled_supplied_tasks;
	private long total_cancelled_consumed_tasks;

}

