/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// NeighborhoodNegotiationProfile.java

package lbg.peer.register;

import lbg.*;
import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.*;
import lbg.peer.config.*;
import lbg.peer.register.negotiation.*;
import java.util.Iterator;
import java.util.TreeMap;

/**
 * @author Cyril Briquet
 */
public class NeighborhoodNegotiationProfile {

	public NeighborhoodNegotiationProfile(Peer managed_peer,PeerRegister grid_register,
								  PeerStrategyConfiguration
								  peer_strategy_config)
		throws GridException {
	
		if (managed_peer == null) {
			throw new GridException("NeighborhoodNegotiationProfile(): illegal Peer");
		}
		if (grid_register == null) {
			throw new GridException("NeighborhoodNegotiationProfile(): "+
									"illegal PeerRegister");
		}
		if (peer_strategy_config == null) {
			throw new GridException("NeighborhoodNegotiationProfile(): illegal "+
									"Peer Strategy configuration");
		}
		
		this.REQUESTORS_COUNT_THRESHOLD =
			peer_strategy_config.REQUESTORS_COUNT_THRESHOLD();
		this.REQUESTORS_TIME_THRESHOLD =
			peer_strategy_config.REQUESTORS_TIME_THRESHOLD();
		this.GRANTORS_COUNT_THRESHOLD =
			peer_strategy_config.GRANTORS_COUNT_THRESHOLD();
		this.GRANTORS_TIME_THRESHOLD =
			peer_strategy_config.GRANTORS_TIME_THRESHOLD();
		this.REQUEST_SUPPLYING_TIME_THRESHOLD =
			peer_strategy_config.REQUEST_SUPPLYING_TIME_THRESHOLD();

		this.managed_peer = managed_peer;
		this.grid_register = grid_register;

		this.supplying_requestors = new TreeMap<String,PeerProfile>();
		this.requestors_set_timestamp = -1;
		this.total_supplying_requestors = 0;
		this.total_received_supplying_requests = 0;

		this.total_consumption_grants_receivers = 0;
		this.total_sent_consumption_grants = 0;

		this.request_supplying_timestamp = -1;
		this.total_potential_suppliers = 0;
		this.total_sent_supplying_requests = 0;

		this.consumption_grantors = new TreeMap<String,PeerProfile>();
		this.grantors_set_timestamp = -1;
		this.total_consumption_grantors = 0;
		this.total_received_consumption_grants = 0;

		this.total_used_consumption_grants = 0;

	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[NeighborhoodNegotiationProfile ");
		sb.append(this.managed_peer.id());
		sb.append("]");

		sb.append("\n  ->  Requestors count threshold ");
		sb.append("(to evaluate Supplying Requests) = ");
		sb.append(REQUESTORS_COUNT_THRESHOLD);

		sb.append("\n  ->  Requestors time  threshold ");
		sb.append("(to evaluate Supplying Requests) = ");
		sb.append(REQUESTORS_TIME_THRESHOLD);
		sb.append(" s");

		sb.append("\n  ->  Grantors   count threshold ");
		sb.append("(to evaluate Consumption Grants) = ");
		sb.append(GRANTORS_COUNT_THRESHOLD);

		sb.append("\n  ->  Grantors   time  threshold ");
		sb.append("(to evaluate Consumption Grants) = ");
		sb.append(GRANTORS_TIME_THRESHOLD);
		sb.append(" s");

		sb.append("\n  ->  previous reception of Supplying Requests at t = ");
		sb.append(this.requestors_set_timestamp);
		sb.append(" s");
		sb.append(" (outstanding Peers: ");
		sb.append(this.supplying_requestors.size());
		sb.append(")");
		sb.append("\n  ->  previous reception of Consumption Grants at t = ");
		sb.append(this.grantors_set_timestamp);
		sb.append(" s");
		sb.append(" (outstanding Peers: ");
		sb.append(this.consumption_grantors.size());
		sb.append(")");
		sb.append("\n  ->  previous sending   of Supplying Requests at t = ");
		sb.append(this.request_supplying_timestamp);
		sb.append(" s");

		sb.append("\n  ->  ");
		sb.append("(Supplier role) granted Supplying Requests=");
		sb.append(this.total_sent_consumption_grants);
		sb.append("/");
		sb.append(this.total_received_supplying_requests);
		sb.append(" (");
		sb.append(CommonMath.
					div_label(this.total_sent_consumption_grants,
							  this.total_received_supplying_requests,100));
		sb.append("%)");
		
		sb.append("\n  ->  ");
		sb.append("(Consumer role) used Consumption Grants=");
		sb.append(this.total_used_consumption_grants);
		sb.append("/");
		sb.append(this.total_received_consumption_grants);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.total_used_consumption_grants,
									   this.total_received_consumption_grants,
									   100));
		sb.append("%)");
		sb.append("\n  ->  ");
		sb.append("(Consumer role) received Consumption Grants=");
		sb.append(this.total_received_consumption_grants);
		sb.append("/");
		sb.append(this.total_sent_supplying_requests);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.total_received_consumption_grants,
									   this.total_sent_supplying_requests,
									   100));
		sb.append("%)");

		// Sanity check
		Iterator<PeerProfile> peer_profiles = this.grid_register.getProfiles();
		int peers_count = 0;
		long tmp_total_sent_consumption_grants     = 0;
		long tmp_total_received_supplying_requests = 0;
		long tmp_total_used_consumption_grants     = 0;
		long tmp_total_sent_supplying_requests     = 0;
		long tmp_total_received_consumption_grants = 0;
		while (peer_profiles.hasNext()) {
			PeerNegotiationProfile negotiation_profile =
				peer_profiles.next().getNegotiationProfile();
			tmp_total_sent_consumption_grants     +=
				negotiation_profile.getTotalSentConsumptionGrants();
			tmp_total_received_supplying_requests +=
				negotiation_profile.getTotalReceivedSupplyingRequests();
			tmp_total_used_consumption_grants     +=
				negotiation_profile.getTotalUsedConsumptionGrants();
			tmp_total_sent_supplying_requests     +=
				negotiation_profile.getTotalSentSupplyingRequests();
			tmp_total_received_consumption_grants +=
				negotiation_profile.getTotalReceivedConsumptionGrants();
			peers_count++;
		}
		assert(peers_count == this.grid_register.size());
		assert(tmp_total_sent_consumption_grants ==
			   total_sent_consumption_grants);
		assert(tmp_total_received_supplying_requests ==
			   total_received_supplying_requests);
		assert(tmp_total_used_consumption_grants ==
			   total_used_consumption_grants);
		assert(tmp_total_sent_supplying_requests ==
			   total_sent_supplying_requests);
		assert(tmp_total_received_consumption_grants ==
			   total_received_consumption_grants);

		return sb.toString();
		
	}
	
	// ****************************************************************

	public synchronized void addReceivedSupplyingRequest(String peer_id,
														 int tasks_count)
		throws GridException {
		
		PeerProfile peer_profile = this.grid_register.getProfile(peer_id);
		if (peer_profile == null) {
			throw new GridException("addReceivedSupplyingRequest(): "+
									"illegal Peer "+peer_id);
		}
		if (tasks_count <= 0) {
			throw new GridException("addReceivedSupplyingRequest(): "+
									"illegal Tasks count");
		}

		PeerNegotiationProfile negotiation_profile =
			peer_profile.getNegotiationProfile();
		negotiation_profile.setReceivedSupplyingRequests(tasks_count);

		if (this.supplying_requestors.size() == 0) {
			this.requestors_set_timestamp = LightweightBarteringGrid.getTime();
		}
		this.supplying_requestors.put(peer_id,peer_profile);
		this.total_supplying_requestors++;
		this.total_received_supplying_requests += tasks_count;
		
	}

	public synchronized int countReceivedSupplyingRequests()
		throws GridException {
		
		return this.supplying_requestors.size();
		
	}
	
	/** Returns whether Supplying Requests can be evaluated,
	 * which is allowed when either the number of registered
	 * Supplying Requestors has reached the specified threshold,
	 * or when the time since a Supplying Requestor has been resgistered
	 * has reached the specified threshold, or both.
	 * @return true if Supplying Requests can be evaluated
	 * @exception GridException
	 */
	public synchronized boolean couldEvaluateSupplyingRequests()
		throws GridException {

		int s = this.supplying_requestors.size();
		
		return ((s >= REQUESTORS_COUNT_THRESHOLD) ||
				((s > 0) &&
				 (LightweightBarteringGrid.getTime() >=
				  this.requestors_set_timestamp+REQUESTORS_TIME_THRESHOLD)));
		
	}

	/** Returns received Supplying Requests.
	 * @return SupplyingRequestsContainer containing
	 * received Supplying Requests, or null if none has been received
	 * since they were last processed
	 */
	public synchronized SupplyingRequestsContainer
		getReceivedSupplyingRequests() throws GridException {

		if (this.supplying_requestors.size() == 0) {
			return null;
		}
		
		SupplyingRequestsContainer src =
			new SupplyingRequestsContainer(this.supplying_requestors.values(),
										   this.supplying_requestors.size());
		this.requestors_set_timestamp = -1;
			
		// (src.size() > 0) always hold given previous early return
		
		this.supplying_requestors.clear();

		//debugging code -- System.out.println("debug -- "+this.managed_peer.id()+": "+src.size()+" SupplyingRequests");

		return src;
		
	}

	public synchronized void accountSentConsumptionGrants(String peer_id,
														  int tasks_count)
		throws GridException {
		
		PeerProfile peer_profile = this.grid_register.getProfile(peer_id);
		if (peer_profile == null) {
			throw new GridException("accountSentConsumptionGrants(): "+
									"illegal Peer "+peer_id);
		}
		if (tasks_count <= 0) {
			throw new GridException("accountSentConsumptionGrants(): "+
									"illegal Tasks count");
		}
		
		PeerNegotiationProfile negotiation_profile =
			peer_profile.getNegotiationProfile();
		negotiation_profile.accountSentConsumptionGrants(tasks_count);
		
		this.total_consumption_grants_receivers++;
		this.total_sent_consumption_grants += tasks_count;

	}

	// ****************************************************************

	/** Returns whether Supplying can be Requested,
	 * which is allowed when the time since some Supplying has been Requested
	 * to other Peers has reached the specified threshold.
	 * @return true if Supplying can be Requestd
	 * @exception GridException
	 */
	public synchronized boolean couldGenerateSupplyingRequests()
		throws GridException {
		
		return (LightweightBarteringGrid.getTime() >=
				this.request_supplying_timestamp+
				REQUEST_SUPPLYING_TIME_THRESHOLD);
		
	}
	
	public synchronized void accountSentSupplyingRequests(String peer_id,
														  int tasks_count)
		throws GridException {
		
		PeerProfile peer_profile = this.grid_register.getProfile(peer_id);
		if (peer_profile == null) {
			throw new GridException("accountSentSupplyingRequests(): "+
									"illegal Peer "+peer_id);
		}
		if (tasks_count <= 0) {
			throw new GridException("accountSentSupplyingRequests(): "+
									"illegal Tasks count");
		}
		
		PeerNegotiationProfile negotiation_profile =
			peer_profile.getNegotiationProfile();
		negotiation_profile.accountSentSupplyingRequests(tasks_count);
		
		this.total_potential_suppliers++;
		this.total_sent_supplying_requests += tasks_count;
		
		this.request_supplying_timestamp = LightweightBarteringGrid.getTime();

	}

	public synchronized void addReceivedConsumptionGrant(String peer_id,
														 int tasks_count)
		throws GridException {
		
		PeerProfile peer_profile = this.grid_register.getProfile(peer_id);
		if (peer_profile == null) {
			throw new GridException("addReceivedConsumptionGrant(): "+
									"illegal Peer "+peer_id);
		}
		if (tasks_count <= 0) {
			throw new GridException("addReceivedConsumptionGrant(): "+
									"illegal Tasks count");
		}
		
		PeerNegotiationProfile negotiation_profile =
			peer_profile.getNegotiationProfile();
		negotiation_profile.setConsumptionGrantsCount(tasks_count);

		if (this.consumption_grantors.size() == 0) {
			this.grantors_set_timestamp = LightweightBarteringGrid.getTime();
		}
		this.consumption_grantors.put(peer_id,peer_profile);
		this.total_received_consumption_grants += tasks_count;
		this.total_consumption_grantors++;

	}

	public synchronized int countReceivedConsumptionGrants()
		throws GridException {
		
		return this.consumption_grantors.size();
		
	}

	/** Returns whether Consumption Grants can be evaluated,
	 * which is allowed when either the number of registered
	 * Consumption Grantors has reached the specified threshold,
	 * or when the time since a Consumption Grantor has been resgistered
	 * has reached the specified threshold, or both.
	 * @return true if Consumption Grants can be evaluated
	 * @exception GridException
	 */
	public synchronized boolean couldEvaluateConsumptionGrants()
		throws GridException {

		int s = this.consumption_grantors.size();
		
		return ((s >= GRANTORS_COUNT_THRESHOLD) ||
				((s > 0) &&
				 (LightweightBarteringGrid.getTime() >=
				  this.grantors_set_timestamp+GRANTORS_TIME_THRESHOLD)));
		
	}

	/** Returns received Consumption grants.
	 * @return ConsumptionGrantsContainer containing
	 * received Consumption Grants, or null if none has been received
	 * since they were last processed
	 */
	public synchronized ConsumptionGrantsContainer
		getReceivedConsumptionGrants() throws GridException {

		if (this.consumption_grantors.size() == 0) {
			return null;
		}
		
		ConsumptionGrantsContainer cgc =
			new ConsumptionGrantsContainer(this.consumption_grantors.values(),
										   this.consumption_grantors.size());
		this.grantors_set_timestamp = -1;

		// (cgc.size() > 0) always hold given previous early return

		this.consumption_grantors.clear();

		//debugging code -- System.out.println("debug -- "+this.managed_peer.id()+": "+cgc.size()+" Consunmption Grants");
		
		return cgc;
		
	}
	
	public synchronized void accountUsedConsumptionGrant(String peer_id)
		throws GridException {
		
		int tasks_count = 1;
			
		PeerProfile peer_profile = this.grid_register.getProfile(peer_id);
		if (peer_profile == null) {
			throw new GridException("accountUsedConsumptionGrant(): "+
									"illegal Peer "+peer_id);
		}
		if (tasks_count <= 0) {
			throw new GridException("accountUsedConsumptionGrant(): "+
									"illegal Tasks count");
		}
		
		PeerNegotiationProfile negotiation_profile =
			peer_profile.getNegotiationProfile();
		negotiation_profile.accountUsedConsumptionGrants(tasks_count);
		
		this.total_used_consumption_grants++;

	}

	// ****************************************************************

	/** Requestors count threshold.
	 * When the number of registered Supplying Requestors
	 * reaches the specified threshold, Supplying Requests can be evaluated.
	 */
	private final int REQUESTORS_COUNT_THRESHOLD; // = Peers_count-1 in SIM mode
	
	/** Requestors time threshold.
	 * When the time since a Supplying Requestor has been resgistered
	 * reaches the specified threshold, Supplying Requests can be evaluated.
	 */
	private final int REQUESTORS_TIME_THRESHOLD;

	/** Grantors count threshold.
	 * When the number of registered Consumption Grantors
	 * reaches the specified threshold, Consumption Grants can be evaluated.
	 */
	private final int GRANTORS_COUNT_THRESHOLD; // = Peers_count-1 in SIM mode

	/** Grantors time threshold.
	 * When the time since a Consumption Grantor has been resgistered
	 * reaches the specified threshold, Consumption Grants can be evaluated.
	 */
	private final int GRANTORS_TIME_THRESHOLD;
	
	/** Request Supplying tim threshold.
	 * When the time since some Supplying has been Requested to other Peers
	 * reaches the specified threshold, Supplying may again be requested.
	 */
	private final int REQUEST_SUPPLYING_TIME_THRESHOLD;

	// ****************************************************************

	private Peer managed_peer;
	private PeerRegister grid_register;
	
	// [SUPPLIER ROLE, REQUEST] Negotiation data (Supplying Requests to Peer)
/**/private TreeMap<String,PeerProfile> supplying_requestors;
/**/private long requestors_set_timestamp;
/**/private long total_supplying_requestors;
	private long total_received_supplying_requests;

	// [SUPPLIER ROLE, RESPONSE] Negotiation data (Consumption Grants from Peer)
/**/private long total_consumption_grants_receivers;
	private long total_sent_consumption_grants;

	// [CONSUMER ROLE, REQUEST] Negotiation data (Supplying Requests from Peer)
/**/private long request_supplying_timestamp;
/**/private long total_potential_suppliers;
	private long total_sent_supplying_requests;

	// [CONSUMER ROLE, RESPONSE] Negotiation data (Consumption Grants to Peer)
/**/private TreeMap<String,PeerProfile> consumption_grantors;
/**/private long grantors_set_timestamp;
	private long total_received_consumption_grants;
/**/private long total_consumption_grantors;

	// [CONSUMER ROLE, USE] Negotiation data (Consumption Grants used by Peer)
	private long total_used_consumption_grants;
	
}

