/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PeerProfile.java

package lbg.peer.register;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.*;
import lbg.peer.register.bartering.*;
import lbg.peer.register.learning.db.*;
import lbg.peer.register.negotiation.*;
import lbg.peer.service.*;
import lbg.peer.strategy.accounting.*;
import lbg.task.*;

/**
 * @author Cyril Briquet
 */
public class PeerProfile implements Comparable<PeerProfile> {

	public PeerProfile(Peer managed_peer,ExternalPeerHandle profiled_peer,
					   BarteringAccountantPolicy accountant)
		throws GridException {

		if (managed_peer == null) {
			throw new GridException("PeerProfile(): illegal managed Peer");
		}
		if (profiled_peer == null) {
			throw new GridException("PeerProfile(): illegal profiled Peer");
		}
		if (accountant == null) {
			throw new GridException("PeerProfile(): illegal Accountant");
		}

		this.managed_peer = managed_peer;
		this.profiled_peer = profiled_peer;
		this.id = profiled_peer.id();
		this.waiting_jobs_count = 0;
		this.running_jobs_count = 0;
		
		this.negotiation_profile =
			new PeerNegotiationProfile(this.managed_peer,this.profiled_peer);
		this.bartering_profile =
			new PeerBarteringProfile(this.managed_peer,
									 this.profiled_peer,
									 accountant);
		PeerInteractionsDataBaseFactory peer_interactions_db_factory =
			new PeerInteractionsDataBaseFactory();
		this.peer_interactions_db =
			peer_interactions_db_factory.
				createPeerInteractionsDataBase(this.managed_peer,
											   this.profiled_peer);
		this.tscf =
			this.peer_interactions_db.createTaskSchedulingContextFactory(this);

		this.hashcode = Integer.MIN_VALUE;
	
	}

	// don't call me...
	PeerProfile() throws GridException {
		
		this.managed_peer = null;
		this.profiled_peer = null;
		this.id = null;
		this.waiting_jobs_count = 0;
		this.running_jobs_count = 0;
		this.negotiation_profile = null;
		this.bartering_profile = null;
		this.peer_interactions_db = null;
		this.tscf = null;
		
	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PeerProfile of ");
		sb.append(this.profiled_peer.id());
		sb.append(",  by ");
		sb.append(this.managed_peer.id());
		sb.append("]");
		
		sb.append("\n  ->  ");
		sb.append(this.bartering_profile);
		sb.append("\n  ->  ");
		sb.append(this.negotiation_profile);
		sb.append("\n  ->  ");
		sb.append(this.peer_interactions_db);

		return sb.toString();
		
	}
	
	public int compareTo(PeerProfile other) {
	
		String self_id = this.id;
		String other_id = other.id();
		
		return self_id.compareToIgnoreCase(other_id);
		
	}

	@Override
	public boolean equals(Object obj) {
		
		if (obj instanceof PeerProfile == false) {
			return false;
		}
		
		return (this.compareTo((PeerProfile)obj) == 0);
		
	}
	
	@Override
	public int hashCode() {
	
		if (this.hashcode == Integer.MIN_VALUE) {
			this.hashcode = CommonMath.generateHashCode(this.id);
		}
		return this.hashcode;
	
	}
	
	// ****************************************************************

	public String id() {
		
		return this.id;
		
	}

	public ExternalPeerHandle getHandle() {
		
		return this.profiled_peer;
			
	}
			
	public PeerNegotiationProfile getNegotiationProfile() {
		
		return this.negotiation_profile;
		
	}
	
	public PeerBarteringProfile getBarteringProfile() {
		
		return this.bartering_profile;
		
	}

	public PeerInteractionsDataBase getInteractionsDataBase()
		throws GridException {
		
		return this.peer_interactions_db;
		
	}

	// ****************************************************************

	public synchronized int getWaitingJobsCount() throws GridException {
		
		return this.waiting_jobs_count;
		
	}
	
	public synchronized int increaseWaitingJobsCount() throws GridException {
		
		return (++this.waiting_jobs_count);
		
	}
	
	public synchronized int decreaseWaitingJobsCount() throws GridException {
		
		if (this.waiting_jobs_count == 0) {
			throw new GridException("decreaseWaitingJobsCount(): "+
									"illegal operation");
		}

		
		return (--this.waiting_jobs_count);
		
	}


	public synchronized int getRunningJobsCount() throws GridException {
		
		return this.running_jobs_count;
		
	}
	
	public synchronized int increaseRunningJobsCount() throws GridException {
		
		return (++this.running_jobs_count);
		
	}
	
	public synchronized int decreaseRunningJobsCount() throws GridException {
		
		if (this.running_jobs_count == 0) {
			throw new GridException("decreaseRunningJobsCount(): "+
									"illegal operation");
		}

		
		return (--this.running_jobs_count);
		
	}

	// ****************************************************************

	public synchronized TaskSchedulingContext
		createConsumptionTaskSchedulingContext(Task t) throws GridException {
		
		return
			this.tscf.
				createConsumptionTaskSchedulingContext(t);
		
	}

	public synchronized TaskSchedulingContext
		createSupplyingTaskSchedulingContext(Task t) throws GridException {

		return
			this.tscf.
				createSupplyingTaskSchedulingContext(t);
		
	}

	// ****************************************************************

	private final Peer managed_peer;
	private final ExternalPeerHandle profiled_peer;
	private final String id;
	private       int waiting_jobs_count;
	private       int running_jobs_count;
	
	private final PeerNegotiationProfile       negotiation_profile;
	private final PeerBarteringProfile         bartering_profile;
	private final PeerInteractionsDataBase     peer_interactions_db;
	private final TaskSchedulingContextFactory tscf;

	private       int hashcode;

}

