/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PeerRegister.java

package lbg.peer.register;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.*;
import lbg.peer.config.*;
import lbg.peer.service.*;
import lbg.peer.strategy.StrategyDescriptor;
import lbg.peer.strategy.accounting.*;
import java.util.Iterator;
import java.util.TreeMap;

/**
 * @author Cyril Briquet
 */
public class PeerRegister {

	public PeerRegister(Peer managed_peer,StrategyDescriptor strategy,
						PeerStrategyConfiguration peer_strategy_config)
		throws GridException {
		
		if (managed_peer == null) {
			throw new GridException("PeerRegister(): illegal Peer");
		}
		if (strategy == null) {
			throw new GridException("PeerRegister(): illegal Strategy");
		}
		if (peer_strategy_config == null) {
			throw new GridException("PeerRegister(): illegal "+
									"Peer Strategy configuration");
		}
		
		this.managed_peer = managed_peer;
		this.strategy = strategy;
		
		this.register = new TreeMap<String,PeerProfile>();
		this.negotiation_profile =
			new NeighborhoodNegotiationProfile(managed_peer,this,peer_strategy_config);
		this.bartering_profile =
			new NeighborhoodBarteringProfile(managed_peer,this);

	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PeerRegister ");
		sb.append(this.managed_peer.id());
		sb.append("]");
		
		sb.append("\n  ->  ");
		sb.append(this.bartering_profile);
		sb.append("\n  ->  ");
		sb.append(this.negotiation_profile);
		
		Iterator<PeerProfile> it = getProfiles();
		double total_favors_balance = 0.0;
		while (it.hasNext()) {
			PeerProfile peer_profile = it.next();
			total_favors_balance += 
				peer_profile.getBarteringProfile().getFavorsBalance();
			sb.append("\n  ->  \n  ->  ");
			sb.append(peer_profile);
		}
		
		sb.append("\n  ->  \n  ->  [All PeerProfiles]");
		sb.append("\n  ->  mean Favors balance=");
		sb.append(CommonMath.
					div_label(total_favors_balance,this.register.size()));
		sb.append("\n  ~>  ");
		sb.append(this.managed_peer.id());
		sb.append(".favors_balance.mean = ");
		sb.append(total_favors_balance/this.register.size());

		return sb.toString();
		
	}
	
	// ****************************************************************

	/** Returns the size of the PeerRegister.
	 * @return number of Peers registered into the PeerRegister
	 */
	public synchronized int size() {
		
		return this.register.size();
		
	}

	/** Creates a Profile in the PeerRegister for the target Peer.
	 * <p>DANGER: Peer authentication currently relies only on PeerHandle.id()
	 * @param peer_handle targer Peer
	 */
	public synchronized void createProfile(ExternalPeerHandle peer_handle)
		throws GridException {
		
		if (peer_handle == null) {
			throw new GridException("createProfile(): illegal Peer");
		}
		
		BarteringAccountantPolicy accountant =
			this.strategy.createDefaultAccountant();
		
		PeerProfile peer_profile =
			new PeerProfile(this.managed_peer,peer_handle,accountant);
			
		if (this.register.put(peer_handle.id(),peer_profile) !=
			null) {
			throw new GridException("createProfile(): Profile already stored");
		}
		
	}
	
	/** Removes a Profile from the PeerRegister for the target Peer.
	 * <p>DANGER: Peer authentication currently relies only on PeerHandle.id()
	 * @param peer_id targer Peer id
	 */
	public synchronized void removeProfile(String peer_id)
		throws GridException {

		if (peer_id == null) {
			throw new GridException("removeProfile(): illegal Peer id");
		}

		PeerProfile peer_profile = this.register.remove(peer_id);
		if (peer_profile == null) {
			throw new GridException("removeProfile(): Peer not stored");
		}
		if (peer_id.equals(peer_profile.getHandle().id()) == false) {
			throw new GridException("removeProfile(): id mismatch");
		}
		
	}

	/** Checks the Profile of the target Peer, and creates one if required.
	 * @param peer_id target Peer
	 * @return true if the managed Peer is authorized to communicate
	 * with target Peer
	 */
	public synchronized boolean checkProfile(String peer_id)
		throws GridException {

		// Avoid self-communications ...
		if (peer_id.equals(this.managed_peer.id())) {
			return false;
		}
		
		// If profile does not exist yet, create it
		if (this.register.get(peer_id) == null) {
			ExternalPeerHandle peer_handle =
				this.managed_peer.searchEngine().search(peer_id);
			this.createProfile(peer_handle);
		}
		
		return true;
		
	}

	/** Returns the Profile of the target Peer.
	 * <p>DANGER: Peer authentication currently relies only on PeerHandle.id()
	 * @param peer_id targer Peer id
	 * @return Profile of the target Peer
	 */
	public synchronized PeerProfile getProfile(String peer_id)
		throws GridException {
		
		if (peer_id == null) {
			throw new GridException("getProfile(): illegal Peer id");
		}

		PeerProfile peer_profile = this.register.get(peer_id);
		if (peer_profile == null) {
			throw new GridException("getProfile(): unknown Peer "+peer_id);
		}
		
		return peer_profile;
		
	}
	
	/** Returns the Profiles of the stored Peers.
	 * <p>DANGER: Peer authentication currently relies only on PeerHandle.id()
	 * @return Profiles of stored Peer
	 */
	public synchronized Iterator<PeerProfile> getProfiles() {
		
		return this.register.values().iterator();
		
	}
	
	/** Returns the Handle of stored Peer.
	 * @param peer_id targer Peer id
	 * @return Profile of the target Peer
	 */
	public synchronized ExternalPeerHandle getHandle(String peer_id)
		throws GridException {
		 
		return getProfile(peer_id).getHandle();

	}
	 
	// ****************************************************************

	public NeighborhoodNegotiationProfile getNeighborhoodNegotiationProfile() {
		
		return this.negotiation_profile;
		
	}
	
	public NeighborhoodBarteringProfile getNeighborhoodBarteringProfile() {
		
		return this.bartering_profile;
		
	}

	// ****************************************************************

	private Peer managed_peer;
	private StrategyDescriptor strategy;
	
	private TreeMap<String,PeerProfile> register;
	private NeighborhoodNegotiationProfile negotiation_profile;
	private NeighborhoodBarteringProfile bartering_profile;

}

