/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// IntervalStore.java

package lbg.peer.register.bartering;

import lbg.common.exception.*;
import lbg.common.math.*;
import static lbg.common.math.CommonMath.filterToZeroIfNaN;
import lbg.common.queueing.deque.*;

import java.util.ArrayList;

/**
 * @author Cyril Briquet
 */
public class IntervalStore {
	
	/**
	 * @param ivn number of intervals to store (> 0)
	 * @param ilen length of each interval, expressed in time units (> 0)
	 * @exception GridException if a parameter is illegal
	 */
	public IntervalStore(int ivn,int ilen) throws GridException {
	
		if (ivn < 1) {
			throw new GridException("IntervalStore(): illegal nb of intervals");
		}
		if (ilen < 1) {
			throw new GridException("IntervalStore(): illegal interval length");
		}
		
		this.ivn = ivn;
		this.ilen = ilen;
		this.m_interval_sizes = MeasurementFactory.
			createDescriptiveMeasurement("IntervalStore",this.ivn);
		this.recent_timestamps = new ArrayList<Long>();
		
	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[IntervalStore]");

		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized void storeTimestamp(long t) throws GridException {

		if (t < 1) {
			throw new GridException("storeTimestamp(): illegal timestamp");
		}
		
		final int s = recent_timestamps.size();
		if (s > 0) {
			Long most_recent_t = this.recent_timestamps.get(s-1);
			if (most_recent_t + ilen < t) {
				addInterval();
			}
			else {
				addTimestamp(t);
			}
		}
		else /* if (s == 0) */ {
			addTimestamp(t);
		}
		
	}
	
	public synchronized double getMeanIntervalSize() {
		
		int s = this.recent_timestamps.size();
		if (s > 0) {
			//TODO -- this.m_interval_sizes.addObservation(s+1);
		}
		double retval = filterToZeroIfNaN(this.m_interval_sizes.getMean());
		if (s > 0) {
			//TODO -- this.m_interval_sizes.removeLastObservation();
		}
		
		return retval;
		
	}
	
	// ****************************************************************

	private synchronized void addTimestamp(long t) throws GridException {
		
		if (t < 1) {
			throw new GridException("addTimestamp(): illegal timestamp");
		}

		this.recent_timestamps.add(t);
		
	}
	
	private synchronized void addInterval() throws GridException {

		final int s = recent_timestamps.size();
		this.m_interval_sizes.addObservation(s+1);
		this.recent_timestamps.clear();
		
	}
	
	// ****************************************************************

	private final int ivn;
	private final int ilen;
	private final Measurement m_interval_sizes;
	private final ArrayList<Long> recent_timestamps;
	
}

