/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PeerBarteringProfile.java

package lbg.peer.register.bartering;

import lbg.*;
import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.*;
import lbg.peer.service.*;
import lbg.peer.strategy.accounting.*;
import lbg.task.*;

/**
 * @author Cyril Briquet
 */
public class PeerBarteringProfile {
	
	public PeerBarteringProfile(Peer managed_peer,
								ExternalPeerHandle profiled_peer,
								BarteringAccountantPolicy accountant)
		throws GridException {
	
		if (managed_peer == null) {
			throw new GridException("PeerProfile(): illegal managed Peer");
		}
		if (profiled_peer == null) {
			throw new GridException("PeerProfile(): illegal profiled Peer");
		}
		if (accountant == null) {
			throw new GridException("PeerProfile(): illegal Accountant");
		}

		this.managed_peer = managed_peer;
		this.profiled_peer = profiled_peer;
		this.accountant = accountant;
		
		this.current_balance = 0.0;
		this.total_supplied_favors = 0;
		this.total_consumed_favors = 0;

		this.total_supplied_local_power = 0;
		this.total_consumed_external_power = 0;

		this.total_completed_consumption_tasks_runtime = 0;
		this.total_completed_consumption_tasks = 0;
		this.m_consumption_completion_runtime =
			MeasurementFactory.createDescriptiveMeasurement(
				"Peer-level completed Consumption Tasks runtimes",
				MAX_STORED_INTERACTIONS);
		this.total_completed_supplying_tasks_runtime = 0;
		this.total_completed_supplying_tasks = 0;
		this.m_supplying_completion_runtime =
			MeasurementFactory.createDescriptiveMeasurement(
				"Peer-level completed Supplying Tasks runtimes",
				MAX_STORED_INTERACTIONS);

		this.consumption_mc4_db = new IntervalStore(MAX_STORED_INTERACTIONS,30);
		this.total_cancelled_consumption_tasks = 0;
		this.m_cancelled_consumption_tasks_runtime =
			MeasurementFactory.createDescriptiveMeasurement(
				"Peer-level cancelled Consumption Tasks runtimes",
				MAX_STORED_INTERACTIONS);
		this.m_cancelled_supplying_tasks_runtime =
			MeasurementFactory.createDescriptiveMeasurement(
				"Peer-level cancelled Supplying Tasks runtimes",
				MAX_STORED_INTERACTIONS);
		this.total_cancelled_supplying_tasks = 0;

		this.completed_consumption_tasks_since_last_cancellation = 0;
		this.completed_supplying_tasks_since_last_cancellation = 0;
		this.m_completed_consumption_tasks_btwn_cancellation =
			MeasurementFactory.createDescriptiveMeasurement(
				"Peer-level completed Consumption Tasks between cancellations",
				MAX_STORED_INTERACTIONS);
		this.m_completed_supplying_tasks_btwn_cancellation =
			MeasurementFactory.createDescriptiveMeasurement(
				"Peer-level completed Supplying Tasks between cancellations",
				MAX_STORED_INTERACTIONS);
		
		this.cancelled_consumption_tasks_since_last_completion = 0;
		this.cancelled_supplying_tasks_since_last_completion = 0;

		this.most_recent_consumption_task_cancellation_time = -1;
		this.most_recent_supplying_task_cancellation_time = -1;
		this.m_time_btwn_consumption_task_cancellations =
			MeasurementFactory.createDescriptiveMeasurement(
				"Peer-level time between cancellations of Consumption Tasks",
				MAX_STORED_INTERACTIONS);
		this.m_time_btwn_supplying_task_cancellations =
			MeasurementFactory.createDescriptiveMeasurement(
				"Peer-level time between cancellations of Supplying Tasks",
				MAX_STORED_INTERACTIONS);

	}
	
	// don't call me...
	PeerBarteringProfile() throws GridException {
		
		this.managed_peer = null;
		this.profiled_peer = null;
		this.consumption_mc4_db = null;
		this.m_consumption_completion_runtime = null;
		this.m_supplying_completion_runtime = null;
		this.m_cancelled_consumption_tasks_runtime = null;
		this.m_cancelled_supplying_tasks_runtime = null;
		this.m_completed_consumption_tasks_btwn_cancellation = null;
		this.m_completed_supplying_tasks_btwn_cancellation = null;
		this.m_time_btwn_consumption_task_cancellations = null;
		this.m_time_btwn_supplying_task_cancellations = null;
		
	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PeerBarteringProfile of ");
		sb.append(this.profiled_peer.id());
		sb.append(",  by ");
		sb.append(this.managed_peer.id());
		sb.append("]");
		
		sb.append("\n  ->  Favors balance=");
		sb.append((long)this.current_balance);
		sb.append(" { Consumption favors=");
		sb.append((long)this.total_consumed_favors);
		sb.append(", Supplying favors=");
		sb.append((long)this.total_supplied_favors);
		sb.append(" }");
		
		// FR = consumed external power / supplied local power
		sb.append("\n  ->  ");
		sb.append("{SIM} Favor Ratio with Peer = ");
		sb.append(this.total_consumed_external_power);
		sb.append(" / ");
		sb.append(this.total_supplied_local_power);
		sb.append(" = ");
		sb.append(CommonMath.div_label_2dd(this.total_consumed_external_power,
										   this.total_supplied_local_power));
		// RG = consumed local power + external power / consumed local power
		sb.append("\n  ->  ");
		sb.append("{SIM} Resource Gain = ");
		long total_consumed_local_power =
			this.managed_peer.register().
				getNeighborhoodBarteringProfile().getTotalConsumedLocalPower();
		sb.append(CommonMath.div_label_2dd((total_consumed_local_power+
					 						this.total_consumed_external_power),
										   (total_consumed_local_power)));


										   sb.append("\n  ->  ");
		sb.append("Supplying Tasks   { mean runtime=");
		sb.append(CommonMath.
					div_label(this.total_completed_supplying_tasks_runtime,
							  this.total_completed_supplying_tasks));
		sb.append(", completed=");
		sb.append(this.total_completed_supplying_tasks);
		sb.append(", cancelled=");
		sb.append(this.total_cancelled_supplying_tasks);
		sb.append(" }");
		sb.append("\n  ->  Supplying reliability=");
		sb.append(CommonMath.
					div_label(getSupplyingTasksCompletionRatio(),1,100));
		sb.append("%");
		sb.append("\n  ->  ");
		sb.append("Supplying MTTC=");
		sb.append(CommonMath.div_label(getSupplyingMTTC(),1));
		sb.append(" s");
		sb.append("\n  ->  ");
		sb.append("Supplying MTBC=");
		sb.append(CommonMath.div_label(getSupplyingMTBC(),1));
		sb.append(" s");
		sb.append("\n  ->  ");
		sb.append("mean Supplying Completion stride=");
		sb.append(CommonMath.div_label(getMeanSupplyingCompletionStride(),1));
		
		sb.append("\n  ->  ");
		sb.append("Consumption Tasks { mean runtime=");
		sb.append(CommonMath.
					div_label(this.total_completed_consumption_tasks_runtime,
							  this.total_completed_consumption_tasks));
		sb.append(", completed=");
		sb.append(this.total_completed_consumption_tasks);
		sb.append(", cancelled=");
		sb.append(this.total_cancelled_consumption_tasks);
		sb.append(" }");
		sb.append("\n  ->  ");
		sb.append("Consumption reliability=");
		sb.append(CommonMath.
					div_label(getConsumptionTasksCompletionRatio(),1,100));
		sb.append("%");
		sb.append("\n  ->  ");
		sb.append("Consumption MTTC=");
		sb.append(CommonMath.div_label(getConsumptionMTTC(),1));
		sb.append(" s");
		sb.append("\n  ->  ");
		sb.append("Consumption MTBC=");
		sb.append(CommonMath.div_label(getConsumptionMTBC(),1));
		sb.append(" s");
		sb.append("\n  ->  ");
		sb.append("mean Consumption Completion stride=");
		sb.append(CommonMath.div_label(getMeanConsumptionCompletionStride(),1));

		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized BarteringAccountantPolicy getAccountant()
		throws GridException {
			
		return this.accountant;
		
	}

	public synchronized void setAccountant(BarteringAccountantPolicy accountant)
		throws GridException {
	
		if (accountant == null) {
			throw new GridException("setAccountant(): illegal Accountant");
		}
			
		this.accountant = accountant;
		
	}
	
	// ****************************************************************

	public synchronized double
		accountSupplyingCompletion(Task t,long xstart_time,
								   long supplied_local_power)
		throws GridException {

		if (t == null) {
			throw new GridException("accountSupplyingCompletion(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("accountSupplyingCompletion(): "+
									"illegal execution start time");
		}
		if (supplied_local_power < 0) {
			throw new GridException("accountSupplyingCompletion(): "+
									"illegal Supplied Local power ("+
									supplied_local_power+" s)");
		}
		
		this.total_supplied_local_power += supplied_local_power;
		
		long now = LightweightBarteringGrid.getTime();
		long runtime = now-xstart_time;
		this.total_completed_supplying_tasks_runtime += runtime;
		this.m_supplying_completion_runtime.addObservation(runtime);
		this.total_completed_supplying_tasks++;
		this.completed_supplying_tasks_since_last_cancellation++;
		this.cancelled_supplying_tasks_since_last_completion = 0;
		
		double supplied_favor =
			this.accountant.evaluateSupplying(t,xstart_time);
		
		if (supplied_favor < 0) {
			throw new GridException("accountSupplyingCompletion(): "+
									"illegal Supplied Favor");
		} /* see debug-supplying-accounting-fast-tasks if Exception is thrown */

		// Remember Supplied favors
		this.total_supplied_favors += supplied_favor;

		// Decrease favors balance
		if (this.current_balance > supplied_favor) {
			this.current_balance -= supplied_favor;
		}
		else /* if (this.current_balance <= supplied_favor) */ {
			this.current_balance = 0; // Favors balance always nonnegative
		}
		
		return supplied_favor;
		
	}

	public synchronized double
		accountConsumptionCompletion(Task t,long xstart_time,
									 long consumed_external_power)
		throws GridException {

		if (t == null) {
			throw new GridException("accountConsumptionCompletion(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("accountConsumptionCompletion(): "+
									"illegal execution start time");
		}
		if (consumed_external_power < 0) {
			throw new GridException("accountConsumptionCompletion(): "+
									"illegal Consumed External Power ("+
									consumed_external_power+" s)");
		}
		
		this.total_consumed_external_power += consumed_external_power;

		long now = LightweightBarteringGrid.getTime();
		long runtime = now-xstart_time;
		this.total_completed_consumption_tasks_runtime += runtime;
		this.m_consumption_completion_runtime.addObservation(runtime);
		this.total_completed_consumption_tasks++;
		this.completed_consumption_tasks_since_last_cancellation++;
		this.cancelled_consumption_tasks_since_last_completion = 0;

		double consumed_favor =
			this.accountant.evaluateConsumption(t,xstart_time);
		
		if (consumed_favor < 0) {
			throw new GridException("accountConsumptionCompletion(): "+
									"illegal Consumed Favor "+consumed_favor+
									" from Peer "+t.output().supplierID()+
									", started at t="+xstart_time);
		} /* see debug-supplying-accounting-fast-tasks if Exception is thrown */
		
		// Remember Consumed favors
		this.total_consumed_favors += consumed_favor;

		// Increase favors balance
		this.current_balance += consumed_favor;
		
		return consumed_favor;
		
	}
	
	// ****************************************************************
	
	public synchronized void
		accountSupplyingCancellation(Task t,long xstart_time)
		throws GridException {

		if (t == null) {
			throw new GridException("accountSupplyingCancellation(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("accountSupplyingCancellation(): "+
									"illegal execution start time");
		}

		long now = LightweightBarteringGrid.getTime();

		long runtime = now-xstart_time;
		this.m_cancelled_supplying_tasks_runtime.addObservation(runtime);
		this.total_cancelled_supplying_tasks++;
		this.m_completed_supplying_tasks_btwn_cancellation.addObservation(
			this.completed_supplying_tasks_since_last_cancellation);
		this.completed_supplying_tasks_since_last_cancellation = 0;
		this.cancelled_supplying_tasks_since_last_completion++;

		if (this.most_recent_supplying_task_cancellation_time > -1) {
			this.m_time_btwn_supplying_task_cancellations.addObservation(
				now-this.most_recent_supplying_task_cancellation_time);
		}
		this.most_recent_supplying_task_cancellation_time = now;

	}

	public synchronized void
		accountConsumptionCancellation(Task t,long xstart_time)
		throws GridException {
		
		if (t == null) {
			throw new GridException("accountConsumptionCancellation(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("accountConsumptionCancellation(): "+
									"illegal execution start time");
		}

		long now = LightweightBarteringGrid.getTime();
		
		this.consumption_mc4_db.storeTimestamp(now);
		long runtime = now-xstart_time;
		this.m_cancelled_consumption_tasks_runtime.addObservation(runtime);
		this.total_cancelled_consumption_tasks++;
		this.m_completed_consumption_tasks_btwn_cancellation.addObservation(
			this.completed_consumption_tasks_since_last_cancellation);
		this.completed_consumption_tasks_since_last_cancellation = 0;
		this.cancelled_consumption_tasks_since_last_completion++;

		if (this.most_recent_consumption_task_cancellation_time > -1) {
			this.m_time_btwn_consumption_task_cancellations.addObservation(
				now-this.most_recent_consumption_task_cancellation_time);
		}
		this.most_recent_consumption_task_cancellation_time = now;
		
	}

	// ****************************************************************

/*
	/**
	 * @deprecated
	 *
	 * /
	@Deprecated
	public synchronized long getTotalSuppliedLocalPower() throws GridException {
		
		return this.total_supplied_local_power;
		
	}
		
	/**
	 * @deprecated
	 *
	 * /
	@Deprecated
	public synchronized long getTotalConsumedExternalPower()
		throws GridException {
		
		return this.total_consumed_external_power;
		
	}
*/
	
	// ****************************************************************

	public synchronized double getTotalSuppliedFavors() {

		return this.total_supplied_favors;
		
	}

	public synchronized double getTotalConsumedFavors() {

		return this.total_consumed_favors;
		
	}

	/** Favors balance of managed Peer with target Peer
	 * (guaranteed to be nonnegative).
	 */
	public synchronized double getFavorsBalance() {

		return this.current_balance;
		
	}

	public synchronized long getTotalCompletedConsumptionTasks() {
	
		return this.total_completed_consumption_tasks;
		
	}

	public synchronized long getTotalCompletedSupplyingTasks() {
	
		return this.total_completed_supplying_tasks;
		
	}

	public synchronized long getTotalCancelledConsumptionTasks() {
	
		return this.total_cancelled_consumption_tasks;
		
	}

	public synchronized long getTotalCancelledSupplyingTasks() {
	
		return this.total_cancelled_supplying_tasks;
		
	}

	/** Current Consumption completion stride length.
	 * Number of completed Consumption Tasks since
	 * the most recent cancellation of such a Task
	 * (or since the instanciation of PeerRegister if no such Task
	 * was cancelled).
	 */
	public synchronized long getRecentlyCompletedConsumptionTasksCount() {
	
		return this.completed_consumption_tasks_since_last_cancellation;
		
	}

	/** Current Supplying completion stride length.
	 * Number of completed Supplying Tasks since
	 * the most recent cancellation of such a Task
	 * (or since the instanciation of PeerRegister if no such Task
	 * was cancelled).
	 */
	public synchronized long getRecentlyCompletedSupplyingTasksCount() {
	
		return this.completed_supplying_tasks_since_last_cancellation;
		
	}
	
	/** Current Consumption cancellation stride length.
	 * Number of cancelled Consumption Tasks since
	 * the most recent completion of such a Task
	 * (or since the instanciation of PeerRegister if no such Task
	 * hasn't already been completed).
	 */
	public synchronized long getRecentlyCancelledConsumptionTasksCount() {
	
		return this.cancelled_consumption_tasks_since_last_completion;
		
	}

	/** Current Supplying cancellation stride length.
	 * Number of cancelled Supplying Tasks since
	 * the most recent completion of such a Task
	 * (or since the instanciation of PeerRegister if no such Task
	 * hasn't already been completed).
	 */
	public synchronized long getRecentlyCancelledSupplyingTasksCount() {
	
		return this.cancelled_supplying_tasks_since_last_completion;
		
	}

	public long getMostRecentConsumptionTaskCancellationTime() {
		
		return this.most_recent_consumption_task_cancellation_time;
		
	}
		
	public long getMostRecentSupplyingTaskCancellationTime() {
		
		return this.most_recent_supplying_task_cancellation_time;
		
	}
	
	// ****************************************************************

	public synchronized double getMeanConsumptionCompletionStride() {
	
		if (this.total_cancelled_consumption_tasks < 1) {
			return this.completed_consumption_tasks_since_last_cancellation;
		}
		
		return m_completed_consumption_tasks_btwn_cancellation.getMean(); 
		
	}
	
	public synchronized double getMeanSupplyingCompletionStride() {
	
		if (this.total_cancelled_supplying_tasks < 1) {
			return this.completed_supplying_tasks_since_last_cancellation;
		}
		return m_completed_supplying_tasks_btwn_cancellation.getMean();
		
	}
	
	public synchronized double getConsumptionTasksCompletionRatio() {
		
		return ((double)this.total_completed_consumption_tasks) /
			   ((double)(this.total_completed_consumption_tasks +
						 this.total_cancelled_consumption_tasks));
		
	}

	public synchronized double getConsumptionTasksCancellationRatio() {
		
		return ((double)this.total_cancelled_consumption_tasks) /
			   ((double)(this.total_completed_consumption_tasks +
						 this.total_cancelled_consumption_tasks));
		
	}

	public synchronized double getSupplyingTasksCompletionRatio() {
		
		return ((double)this.total_completed_supplying_tasks) /
			   ((double)(this.total_completed_supplying_tasks +
						 this.total_cancelled_supplying_tasks));
		
	}

	public synchronized double getSupplyingTasksCancellationRatio() {
		
		return ((double)this.total_cancelled_supplying_tasks) /
			   ((double)(this.total_completed_supplying_tasks +
						 this.total_cancelled_supplying_tasks));
		
	}

	/** Consumption Tasks Mean Cancellation Notification Time
	 * (MTTC, pronounced <em>m cant</em>).
	 * @return Consumption Tasks MTTC,
	 * or Double.MAX_VALUE if no Task has been cancelled yet
	 */
	public synchronized double getConsumptionMTTC() {
		
		if (this.total_cancelled_consumption_tasks < 1) {
			return Double.MAX_VALUE;
		}
		
		return m_cancelled_consumption_tasks_runtime.getMean();
		
	}

	/** Supplying Tasks Mean Cancellation Notification Time.
	 * @return Supplying Tasks MTTC,
	 * or Double.MAX_VALUE if no Task has been cancelled yet
	 */
	public synchronized double getSupplyingMTTC() {
		
		if (this.total_cancelled_supplying_tasks < 1) {
			return Double.MAX_VALUE;
		}
		
		return m_cancelled_supplying_tasks_runtime.getMean();
		
	}
	
	/** Consumption Tasks Mean Time Between Cancellations.
	 * @return Consumption Task MTBC,
	 * or Double.MAX_VALUE if less than two Tasks have been cancelled yet
	 */
	public synchronized double getConsumptionMTBC() {
	
		if (this.total_cancelled_consumption_tasks < 2) {
			return Double.MAX_VALUE;
		}
		
		return this.m_time_btwn_consumption_task_cancellations.getMean();
		
	}

	/** Supplying Tasks Mean Time Between Cancellations.
	 * @return Supplying Task MTBC,
	 * or Double.MAX_VALUE if less than two Tasks have been cancelled yet
	 */
	public synchronized double getSupplyingMTBC() {
	
		if (this.total_cancelled_supplying_tasks < 2) {
			return Double.MAX_VALUE;
		}
		
		return this.m_time_btwn_supplying_task_cancellations.getMean();
		
	}
	
	public synchronized double getMeanConsumptionTaskRuntime() {

		if (this.total_completed_consumption_tasks < 1) {
			return Double.MAX_VALUE;
		}
		
		return this.m_consumption_completion_runtime.getMean();
		
	}

	public synchronized double getMeanSupplyingTaskRuntime() {
		
		if (this.total_completed_supplying_tasks < 1) {
			return Double.MAX_VALUE;
		}
		
		return this.m_supplying_completion_runtime.getMean();
		
	}

	/**
	 * @return an estimate of the runtime of a Supplying Task
	 * of the profiled Peer (>= 0.0), or -1.0 if no estimate is available
	 */
	public synchronized double estimateSupplyingTaskRuntime() {
	
		if (this.m_supplying_completion_runtime.getObservations() > 1) {
			return this.m_supplying_completion_runtime.getMean()+
				   this.m_supplying_completion_runtime.getStandardDeviation();
		}
		else {
			return this.managed_peer.register().
				getNeighborhoodBarteringProfile().
				estimateSupplyingTaskRuntime();
		}
		
	}
	
	/*public synchronized double getMeanRecentConsumptionTaskRuntime() {

		return (((double)this.completed_consumption_tasks_since_last_cancellation_runtime) /
				((double)this.completed_consumption_tasks_since_last_cancellation));
		
	}

	public synchronized double getMeanRecentSupplyingTaskRuntime() {
		
		return (((double)this.completed_supplying_tasks_since_last_cancellation_runtime) /
				((double)this.completed_supplying_tasks_since_last_cancellation));
		
	}*/
	
	public synchronized double getConsumptionMC4() {
	
		return this.consumption_mc4_db.getMeanIntervalSize();
		
	}
	
	public static final int MAX_STORED_INTERACTIONS = 100;

	// ****************************************************************

	private final Peer managed_peer;
	private final ExternalPeerHandle profiled_peer;
	private       BarteringAccountantPolicy accountant;

	// Bartering outcome data
	private       double current_balance; // Favors balance always >= 0.0
	private       double total_supplied_favors;
	private       double total_consumed_favors;
	
	private       long total_supplied_local_power;
	private       long total_consumed_external_power;

	private       long total_completed_consumption_tasks_runtime;
	private       long total_completed_consumption_tasks;
	private final Measurement m_consumption_completion_runtime;
	private       long total_completed_supplying_tasks_runtime;
	private       long total_completed_supplying_tasks;
	private final Measurement m_supplying_completion_runtime;

	private final IntervalStore consumption_mc4_db;
	private final Measurement m_cancelled_consumption_tasks_runtime;
    private       long total_cancelled_consumption_tasks;
	private final Measurement m_cancelled_supplying_tasks_runtime;
    private       long total_cancelled_supplying_tasks;
	
	private       long completed_consumption_tasks_since_last_cancellation;
	private       long completed_supplying_tasks_since_last_cancellation;
	private final Measurement m_completed_consumption_tasks_btwn_cancellation;
	private final Measurement m_completed_supplying_tasks_btwn_cancellation;

	private       long cancelled_consumption_tasks_since_last_completion;
	private       long cancelled_supplying_tasks_since_last_completion;

	private       long most_recent_consumption_task_cancellation_time;
	private       long most_recent_supplying_task_cancellation_time;
	private final Measurement m_time_btwn_consumption_task_cancellations;
	private final Measurement m_time_btwn_supplying_task_cancellations;
	
}

