/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// TestPeerBarteringProfile.java

package lbg.peer.register.bartering;

import lbg.*;
import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.service.*;
import lbg.peer.strategy.accounting.*;
import lbg.task.*;

/**
 * @author Cyril Briquet
 */
public class TestPeerBarteringProfile extends PeerBarteringProfile {
	
	public TestPeerBarteringProfile(String profiled_peer_id,
									String managed_peer_id)
		throws GridException {

		if (profiled_peer_id == null) {
			throw new GridException("TestPeerBarteringProfile(): "+
									"illegal profiled Peer id");
		}
		if (managed_peer_id == null) {
			throw new GridException("TestPeerBarteringProfile(): "+
									"illegal managed Peer id");
		}

		this.managed_peer_id = managed_peer_id;
		this.profiled_peer_id = profiled_peer_id;
		
		this.total_completed_consumption_tasks_runtime = 0;
		this.total_completed_consumption_tasks = 0;
		this.total_completed_supplying_tasks_runtime = 0;
		this.total_completed_supplying_tasks = 0;

		this.total_cancelled_consumption_tasks_runtime = 0;
		this.total_cancelled_consumption_tasks = 0;
		this.total_cancelled_supplying_tasks_runtime = 0;
		this.total_cancelled_supplying_tasks = 0;

		this.recently_completed_consumption_tasks = 0;
		this.recently_completed_supplying_tasks = 0;
		this.total_recently_completed_consumption_tasks = 0;
		this.total_recently_completed_supplying_tasks = 0;
		
		this.recently_cancelled_consumption_tasks = 0;
		this.recently_cancelled_supplying_tasks = 0;

		this.most_recent_consumption_task_cancellation_time = -1;
		this.most_recent_supplying_task_cancellation_time = -1;
		this.total_time_btwn_consumption_task_cancellations = 0;
		this.total_time_btwn_supplying_task_cancellations = 0;

	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[TestPeerBarteringProfile of ");
		sb.append(this.profiled_peer_id);
		sb.append(",  by ");
		sb.append(this.managed_peer_id);
		sb.append("]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized BarteringAccountantPolicy getAccountant()
		throws GridException {
			
		throw new GridException("not implemented");
		
	}

	public synchronized void setAccountant(BarteringAccountantPolicy accountant)
		throws GridException {
	
		throw new GridException("not implemented");
		
	}
	
	// ****************************************************************

	public synchronized double
		accountSupplyingCompletion(Task t,long xstart_time,
								   long supplied_local_power)
		throws GridException {

		if (t != null) {
			throw new GridException("accountSupplyingCompletion(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("accountSupplyingCompletion(): "+
									"illegal execution start time");
		}
		if (supplied_local_power >= 0) {
			throw new GridException("accountSupplyingCompletion(): "+
									"illegal Supplied Local power ("+
									supplied_local_power+" s)");
		}
		
		long now = LightweightBarteringGrid.getTime();
		this.total_completed_supplying_tasks_runtime += (now-xstart_time);
		this.total_completed_supplying_tasks++;
		this.recently_completed_supplying_tasks++;
		this.recently_cancelled_supplying_tasks = 0;
		
		return Double.NaN;
		
	}

	public synchronized double
		accountConsumptionCompletion(Task t,long xstart_time,
									 long consumed_external_power)
		throws GridException {

		if (t != null) {
			throw new GridException("accountConsumptionCompletion(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("accountConsumptionCompletion(): "+
									"illegal execution start time");
		}
		if (consumed_external_power >= 0) {
			throw new GridException("accountConsumptionCompletion(): "+
									"illegal Consumed External Power ("+
									consumed_external_power+" s)");
		}
		
		long now = LightweightBarteringGrid.getTime();
		this.total_completed_consumption_tasks_runtime += (now-xstart_time);
		this.total_completed_consumption_tasks++;
		this.recently_completed_consumption_tasks++;
		this.recently_cancelled_consumption_tasks = 0;

		return Double.NaN;
		
	}
	
	public synchronized void
		accountSupplyingCancellation(Task t,long xstart_time)
		throws GridException {

		if (t != null) {
			throw new GridException("accountSupplyingCancellation(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("accountSupplyingCancellation(): "+
									"illegal execution start time");
		}

		long now = LightweightBarteringGrid.getTime();

		this.total_cancelled_supplying_tasks_runtime += now-xstart_time;
		this.total_cancelled_supplying_tasks++;
		this.total_recently_completed_supplying_tasks +=
			this.recently_completed_supplying_tasks;
		this.recently_completed_supplying_tasks = 0;
		this.recently_cancelled_supplying_tasks++;

		if (this.most_recent_supplying_task_cancellation_time > -1) {
			this.total_time_btwn_supplying_task_cancellations +=
				(now-this.most_recent_supplying_task_cancellation_time);
		}
		this.most_recent_supplying_task_cancellation_time = now;

	}

	public synchronized void
		accountConsumptionCancellation(Task t,long xstart_time)
		throws GridException {
		
		if (t != null) {
			throw new GridException("accountConsumptionCancellation(): "+
									"illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("accountConsumptionCancellation(): "+
									"illegal execution start time");
		}

		long now = LightweightBarteringGrid.getTime();
		
		this.total_cancelled_consumption_tasks_runtime = now-xstart_time;
		this.total_cancelled_consumption_tasks++;
		this.total_recently_completed_consumption_tasks +=
			this.recently_completed_consumption_tasks;
		this.recently_completed_consumption_tasks = 0;
		this.recently_cancelled_consumption_tasks++;

		if (this.most_recent_consumption_task_cancellation_time > -1) {
			this.total_time_btwn_consumption_task_cancellations +=
				(now-this.most_recent_consumption_task_cancellation_time);
		}
		this.most_recent_consumption_task_cancellation_time = now;
		
	}

	// ****************************************************************

	/**
	 * @deprecated
	 */
	@Deprecated
	public synchronized long getTotalSuppliedLocalPower() {
		
		return Long.MIN_VALUE;
		
	}
		
	/**
	 * @deprecated
	 */
	@Deprecated
	public synchronized long getTotalConsumedExternalPower() {
		
		return Long.MIN_VALUE;
		
	}
	
	// ****************************************************************

	public synchronized double getTotalSuppliedFavors() {

		return Double.NaN;
		
	}

	public synchronized double getTotalConsumedFavors() {

		return Double.NaN;
		
	}

	/** Favors balance of managed Peer with target Peer
	 * (guaranteed to be nonnegative).
	 */
	public synchronized double getFavorsBalance() {

		return Double.NaN;
		
	}

	public synchronized long getTotalCompletedConsumptionTasks() {
	
		return this.total_completed_consumption_tasks;
		
	}

	public synchronized long getTotalCompletedSupplyingTasks() {
	
		return this.total_completed_supplying_tasks;
		
	}

	public synchronized long getTotalCancelledConsumptionTasks() {
	
		return this.total_cancelled_consumption_tasks;
		
	}

	public synchronized long getTotalCancelledSupplyingTasks() {
	
		return this.total_cancelled_supplying_tasks;
		
	}

	/** Current Consumption completion stride length.
	 * Number of completed Consumption Tasks since
	 * the most recent cancellation of such a Task
	 * (or since the instanciation of PeerRegister if no such Task
	 * was cancelled).
	 */
	public synchronized long getRecentlyCompletedConsumptionTasksCount() {
	
		return this.recently_completed_consumption_tasks;
		
	}

	/** Current Supplying completion stride length.
	 * Number of completed Supplying Tasks since
	 * the most recent cancellation of such a Task
	 * (or since the instanciation of PeerRegister if no such Task
	 * was cancelled).
	 */
	public synchronized long getRecentlyCompletedSupplyingTasksCount() {
	
		return this.recently_completed_supplying_tasks;
		
	}
	
	/** Current Consumption cancellation stride length.
	 * Number of cancelled Consumption Tasks since
	 * the most recent completion of such a Task
	 * (or since the instanciation of PeerRegister if no such Task
	 * hasn't already been completed).
	 */
	public synchronized long getRecentlyCancelledConsumptionTasksCount() {
	
		return this.recently_cancelled_consumption_tasks;
		
	}

	/** Current Supplying cancellation stride length.
	 * Number of cancelled Supplying Tasks since
	 * the most recent completion of such a Task
	 * (or since the instanciation of PeerRegister if no such Task
	 * hasn't already been completed).
	 */
	public synchronized long getRecentlyCancelledSupplyingTasksCount() {
	
		return this.recently_cancelled_supplying_tasks;
		
	}

	public long getMostRecentConsumptionTaskCancellationTime() {
		
		return this.most_recent_consumption_task_cancellation_time;
		
	}
		
	public long getMostRecentSupplyingTaskCancellationTime() {
		
		return this.most_recent_supplying_task_cancellation_time;
		
	}
	
	// ****************************************************************

	public synchronized double getMeanConsumptionCompletionStride() {
	
		if (this.total_cancelled_consumption_tasks < 1) {
			return this.recently_completed_consumption_tasks;
		}
		return ((double)this.total_recently_completed_consumption_tasks) /
			   ((double)this.total_cancelled_consumption_tasks);
		
	}
	
	public synchronized double getMeanSupplyingCompletionStride() {
	
		if (this.total_cancelled_supplying_tasks < 1) {
			return this.recently_completed_supplying_tasks;
		}
		return ((double)this.total_recently_completed_supplying_tasks) /
			   ((double)this.total_cancelled_supplying_tasks);
		
	}
	
	public synchronized double getConsumptionTasksCompletionRatio() {
		
		return ((double)this.total_completed_consumption_tasks) /
			   ((double)(this.total_completed_consumption_tasks +
						 this.total_cancelled_consumption_tasks));
		
	}

	public synchronized double getConsumptionTasksCancellationRatio() {
		
		return ((double)this.total_cancelled_consumption_tasks) /
			   ((double)(this.total_completed_consumption_tasks +
						 this.total_cancelled_consumption_tasks));
		
	}

	public synchronized double getSupplyingTasksCompletionRatio() {
		
		return ((double)this.total_completed_supplying_tasks) /
			   ((double)(this.total_completed_supplying_tasks +
						 this.total_cancelled_supplying_tasks));
		
	}

	public synchronized double getSupplyingTasksCancellationRatio() {
		
		return ((double)this.total_cancelled_supplying_tasks) /
			   ((double)(this.total_completed_supplying_tasks +
						 this.total_cancelled_supplying_tasks));
		
	}

	/** Consumption Tasks Mean Cancellation Notification Time
	 * (MTTC, pronounced <em>m cant</em>).
	 * @return Consumption Tasks MTTC,
	 * or Double.MAX_VALUE if no Task has been cancelled yet
	 */
	public synchronized double getConsumptionMTTC() {
		
		if (this.total_cancelled_consumption_tasks < 1) {
			return Double.MAX_VALUE;
		}
		return ((double)this.total_cancelled_consumption_tasks_runtime) /
			   ((double)this.total_cancelled_consumption_tasks);
		
	}

	/** Supplying Tasks Mean Cancellation Notification Time.
	 * @return Supplying Tasks MTTC,
	 * or Double.MAX_VALUE if no Task has been cancelled yet
	 */
	public synchronized double getSupplyingMTTC() {
		
		if (this.total_cancelled_supplying_tasks < 1) {
			return Double.MAX_VALUE;
		}
		return ((double)this.total_cancelled_supplying_tasks_runtime) /
			   ((double)this.total_cancelled_supplying_tasks);
		
	}
	
	/** Consumption Tasks Mean Time Between Cancellations.
	 * @return Consumption Task MTBC,
	 * or Double.MAX_VALUE if less than two Tasks have been cancelled yet
	 */
	public synchronized double getConsumptionMTBC() {
	
		if (this.total_cancelled_consumption_tasks < 2) {
			return Double.MAX_VALUE;
		}
		return ((double)this.total_time_btwn_consumption_task_cancellations) /
			   ((double)this.total_cancelled_consumption_tasks-1);
		
	}

	/** Supplying Tasks Mean Time Between Cancellations.
	 * @return Supplying Task MTBC,
	 * or Double.MAX_VALUE if less than two Tasks have been cancelled yet
	 */
	public synchronized double getSupplyingMTBC() {
	
		if (this.total_cancelled_supplying_tasks < 2) {
			return Double.MAX_VALUE;
		}
		return ((double)this.total_time_btwn_supplying_task_cancellations) /
			   ((double)this.total_cancelled_supplying_tasks-1);
		
	}
	
	public synchronized double getMeanConsumptionTaskRuntime() {

		return (((double)this.total_completed_consumption_tasks_runtime) /
				((double)this.total_completed_consumption_tasks));
		
	}

	public synchronized double getMeanSupplyingTaskRuntime() {
		
		return (((double)this.total_completed_supplying_tasks_runtime) /
				((double)this.total_completed_supplying_tasks));
		
	}

	// ****************************************************************

	private final String profiled_peer_id;
	private final String managed_peer_id;

	// Bartering outcome data

	private long total_completed_consumption_tasks_runtime;
	private long total_completed_consumption_tasks;
	private long total_completed_supplying_tasks_runtime;
	private long total_completed_supplying_tasks;

	private long total_cancelled_consumption_tasks_runtime;
	private long total_cancelled_consumption_tasks;
	private long total_cancelled_supplying_tasks_runtime;
	private long total_cancelled_supplying_tasks;
	
	private long recently_completed_consumption_tasks;
	private long recently_completed_supplying_tasks;
	private long total_recently_completed_consumption_tasks;
	private long total_recently_completed_supplying_tasks;

	private long recently_cancelled_consumption_tasks;
	private long recently_cancelled_supplying_tasks;

	private long most_recent_consumption_task_cancellation_time;
	private long most_recent_supplying_task_cancellation_time;
	private long total_time_btwn_consumption_task_cancellations;
	private long total_time_btwn_supplying_task_cancellations;
	
}

