/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// JobConsumptionOutcomeTracker.java

package lbg.peer.register.learning.consumption;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.service.*;
import java.util.*;

/** A JobConsumptionOutcomeTracker tracks completion/cancellation of
 * the Job's Tasks by the involved Supplier Peers,
 * providing a useful reliability metric
 * that may be used to black-list unreliable Suppliers.
 * @author Cyril Briquet
 */
public class JobConsumptionOutcomeTracker {

	public JobConsumptionOutcomeTracker(String owner_peer_id,int job_id,
										int total_job_tasks)
		throws GridException {
		
		if (owner_peer_id == null) {
			throw new GridException("JobConsumptionOutcomeTracker(): "+
									"illegal Peer id");
		}
		if (job_id < 0) {
			throw new GridException("JobConsumptionOutcomeTracker(): "+
									"illegal id");
		}
		if (total_job_tasks < 1) {
			throw new GridException("JobConsumptionOutcomeTracker(): "+
									"illegal Job Tasks count");
		}


		this.owner_peer_id = owner_peer_id;
		this.job_id = job_id;
		this.total_job_tasks = total_job_tasks;

		this.suppliers = new TreeMap<String,SupplierReliabilityHistory>();
		this.suppliers_ranking =
			new TreeMap<SupplierReliabilityHistory,Set<String>>();
		this.m_consumption_completion_runtime =
			MeasurementFactory.createSummarizedMeasurement(
				"Job-level Consumption Tasks completion times");
		this.total_completed_consumption_tasks = 0;
		this.total_cancelled_consumption_tasks = 0;

	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[JobConsumptionOutcomeTracker]");
		sb.append(" Job id=");
		sb.append(this.job_id);
		sb.append(", submitter=");
		sb.append(this.owner_peer_id);
		sb.append(", total Job Tasks=");
		sb.append(this.total_job_tasks);
		sb.append(", completed Consumption Tasks=");
		sb.append(this.total_completed_consumption_tasks);
		sb.append(", cancelled Consumption Tasks=");
		sb.append(this.total_cancelled_consumption_tasks);
		sb.append(", estimated completed Consumption Task runtime=");
		sb.append(this.estimateConsumptionTaskRuntime());
		
		if (trackerInvariant() == false) {
			sb.append(" (CORRUPTED JOB CONSUMPTION TRACKER)");
		}

		for (SupplierReliabilityHistory b : this.suppliers_ranking.keySet()) {
			sb.append("\n  ->  ");
			sb.append(b);
			sb.append("\n  ->                         { ");
			Set<String> peers_with_same_b = this.suppliers_ranking.get(b);
			int i = 0;
			for (String supplier_peer : peers_with_same_b) {
				sb.append(supplier_peer);
				if (++i < peers_with_same_b.size()) sb.append(", ");
			}
			sb.append(" }");
		}
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized void trackCompletedTask(String supplier_peer,
												long runtime)
		throws GridException {
		
		updatePeerPolicy(supplier_peer,true,runtime);
		this.m_consumption_completion_runtime.addObservation(runtime);
		
		assert(this.total_completed_consumption_tasks ==
			   this.m_consumption_completion_runtime.getObservations());
		
	}
	
	public synchronized void trackCancelledTask(String supplier_peer,
												long runtime)
		throws GridException {
		
		updatePeerPolicy(supplier_peer,false,runtime);

	}

	// ****************************************************************

	public synchronized double estimateConsumptionTaskRuntime() {
		
		if (this.total_completed_consumption_tasks > 0) {
			return
				this.m_consumption_completion_runtime.getMean()+
				this.m_consumption_completion_runtime.getStandardDeviation();
		}
		else {
			return Double.NaN;
		}
	
	}
	
	public synchronized int getCompletedConsumptionTasksCount() {
	
		return this.total_completed_consumption_tasks;
		
	}
	
	// ****************************************************************

	/** Indicates whether the Tasks of the tracked Job can be executed
	 * as Consumption Tasks, or if it is preferable not to,
	 * because it comes that these are extremely difficult
	 * to execute by Supplier Peers (either intrinsically,
	 * or because the environment is totally unreliable).
	 */
	public synchronized boolean shouldRestrictJobToLocalExecution()
		throws GridException {
		
		return (((float)this.total_cancelled_consumption_tasks) >
				MAX_CONSUMPTION_FAILURES_FACTOR*((float)this.total_job_tasks));
		
	}
	
	// ****************************************************************

	/** Indicates whether a supplier with target reliability behavior
	 * should be blacklisted, i.e. if it has been very unreliable
	 * when running Consumption Tasks of the tracked BoT.
	 */
	public synchronized boolean
		shouldBlackListPeer(SupplierReliabilityHistory b)
		throws GridException {
		
		if (b == null) {
			throw new GridException("shouldBlackListPeer(): illegal behavior");
		}
		
		return (b.getCancelledTasksCount() > MAX_SUPPLIER_FAILURES_FACTOR);
		
	}

	/** Indicates whether the target supplier
	 * should be blacklisted, i.e. if it has been very unreliable
	 * when running Consumption Tasks of the tracked BoT.
	 */
	public synchronized boolean shouldBlackListPeer(String supplier_id)
		throws GridException {
		
		if (supplier_id == null) {
			throw new GridException("shouldBlackListPeer(): "+
									"illegal supplier id");
		}
		
		SupplierReliabilityHistory b = this.suppliers.get(supplier_id);
		
		if (b == null) {
			return false;
		}
		
		return shouldBlackListPeer(b);
		
	}
	
	// ****************************************************************
	
	/** Return reliability ranking of Supplier Peers
	 * (i.e. Peers that have accepted the submission
	 * of at least 1 Task from the target Job),
	 * sorted by decreasing reliability,
	 * as defined by their SupplierReliabilityHistory.
	 * <p>The most unreliable Peers are excluded from the ranking.
	 * @return reliability ranking of Suppliers,
	 * ordered by decreasing reliability
	 * @exception GridException if an error occurs while ranking Suppliers
	 */
	public synchronized ArrayList<String> rankPeers() throws GridException {

		// Unreliable Suppliers (which exceed a given unreliability threshold)
		// are black-listed (for this Job), and thus excluded from the ranking
		
		ArrayList<String> dummy =
			new ArrayList<String>(this.suppliers_ranking.size());
		for (SupplierReliabilityHistory b : this.suppliers_ranking.keySet()) {
			if (shouldBlackListPeer(b)) {
				continue; // performance optimization: replace with "break;"
			}
			Set<String> peers_with_same_b = this.suppliers_ranking.get(b);
			for (String supplier_peer : peers_with_same_b) {
				dummy.add(supplier_peer);
			}
		}
		
		return dummy;
		
	}

	/** Return the set of Peers that have supplied Resources
	 * for Consumption Tasks of the Job tracked by this tracker.
	 * @return Suppliers set
	 */
	public synchronized Iterator<String> suppliers() {
		
		return this.suppliers.keySet().iterator();
		
	}
	
	// ****************************************************************

	/** For testing purposes only.
	 */
	public static void main(String[] args) {
	
		try {
			// Create Job Consumption tracker
			JobConsumptionOutcomeTracker consumption_tracker =
				new JobConsumptionOutcomeTracker("Consumer",24,10);
			
			// Create Suppliers
			int N = 5;
			ExternalPeerHandle[] suppliers = new ExternalPeerHandle[N];
			for (int i = 0 ; i < N ; i++) {
				suppliers[i] = new TestPeerHandle("Peer"+i);
			}
			
			// Fill tracker
			long runtime = 24;
			consumption_tracker.trackCompletedTask(suppliers[0].id(),runtime);
			consumption_tracker.trackCompletedTask(suppliers[1].id(),runtime);
			consumption_tracker.trackCompletedTask(suppliers[2].id(),runtime);
			consumption_tracker.trackCancelledTask(suppliers[2].id(),3);
			for (int i = 0 ; i < 24 ; i++) {
				System.out.println("*beat*");
				System.out.println(consumption_tracker);

				int p = CommonMath.sampleUniformDist(0,N-1);
				boolean has_completed_task = CommonMath.sampleRandomBoolean();
				System.out.println("\nPeer"+p+" "+(has_completed_task ?
								   "completed" : "cancelled")+ " one Task");
				if (has_completed_task) {
					consumption_tracker.
						trackCompletedTask(suppliers[p].id(),
										   runtime-10+(i%3*10));
				}
				else {
					consumption_tracker.
						trackCancelledTask(suppliers[p].id(),3);
				}
				System.out.println();
			}

			// Output tracker state
			System.out.println(consumption_tracker);
		}
		catch (GridException e) {
			e.printStackTrace();
			System.exit(0);
		}
		
	}
	
	// ****************************************************************

	private synchronized void
		updatePeerPolicy(String supplier_peer,boolean has_completed_task,
						   long runtime)
		throws GridException {
		
		if (supplier_peer == null) {
			throw new GridException("updatePeerPolicy(): illegal Peer");
		}
		if (runtime < 0) {
			throw new GridException("updatePeerPolicy(): illegal runtime");
		}
		
		// Update Suppliers map
		SupplierReliabilityHistory b = this.suppliers.get(supplier_peer);
		if (b == null) {
			b = new SupplierReliabilityHistory();
			if (this.suppliers.put(supplier_peer,b) !=
				null) {
				throw new GridException("updatePeerPolicy(): "+
										"corrupted data");
			}
		}
		else {
			this.unrankPeer(b,supplier_peer);
		}
		
		// Update Supplier behavior (AFTER unranking)
		if (has_completed_task) {
			b.accountCompletedTask();
			this.total_completed_consumption_tasks++;
		}
		else {
			b.accountCancelledTask();
			this.total_cancelled_consumption_tasks++;
		}

		// Update Suppliers ranking map
		this.rankPeer(b,supplier_peer);
		
	}
	
	// ****************************************************************

	private synchronized void 
		rankPeer(SupplierReliabilityHistory b,String supplier_peer)
		throws GridException {

		if (b == null) {
			throw new GridException("rankPeer(): illegal behavior");
		}
		if (supplier_peer == null) {
			throw new GridException("rankPeer(): illegal Peer");
		}
		
		Set<String> peers_with_same_b =	this.suppliers_ranking.get(b);
		if (peers_with_same_b == null) {
			peers_with_same_b = new TreeSet<String>();
			if (this.suppliers_ranking.put(new SupplierReliabilityHistory(b),
										   peers_with_same_b) !=
				null) {
					throw new GridException("rankPeer(): "+
											"corrupted Peers ranking");
			}
		}
		
		if (peers_with_same_b.add(supplier_peer) == false) {
			throw new GridException("rankPeer(): corrupted Peers ranking");
		}
		
	}
	
	private synchronized void
		unrankPeer(SupplierReliabilityHistory b,String supplier_peer)
		throws GridException {

		if (b == null) {
			throw new GridException("unrankPeer(): illegal behavior");
		}
		if (supplier_peer == null) {
			throw new GridException("unrankPeer(): illegal Peer");
		}
		
		Set<String> peers_with_same_b =	this.suppliers_ranking.get(b);
		if (peers_with_same_b == null) {
			throw new GridException("unrankPeer(): corrupted Peers ranking");
		}
		
		if (peers_with_same_b.remove(supplier_peer) == false) {
			throw new GridException("unrankPeer(): corrupted Peers ranking");
		}
		if (peers_with_same_b.size() == 0) {
			if (this.suppliers_ranking.remove(b) ==	null) {
				throw new GridException("unrankPeer(): "+
										"corrupted Peers ranking");
			}
		}
		
	}
	
	// ****************************************************************

	// JobConsumptionOutcomeTracker invariant
	private synchronized boolean trackerInvariant() {
	
		for (String supplier_peer : this.suppliers.keySet()) {
			SupplierReliabilityHistory b = this.suppliers.get(supplier_peer);
			if (b == null) {
				return false;
			}
			Set<String> peers_with_same_b =	this.suppliers_ranking.get(b);
			if (peers_with_same_b == null) {
				return false;
			}
			if (peers_with_same_b.contains(supplier_peer) == false) {
				return false;
			}
		}

		for (SupplierReliabilityHistory b : this.suppliers_ranking.keySet()) {
			Set<String> peers_with_same_b =	this.suppliers_ranking.get(b);
			if (peers_with_same_b == null) {
				return false;
			}
			for (String supplier_peer : peers_with_same_b) {
				if (this.suppliers.get(supplier_peer).equals(b) == false) {
					return false;
				}
			}
		}
		
		return true;
		
	}
	
	// ****************************************************************

	/** Number of allowed Task execution failures for the Tasks of the Job
	 * before its Tasks are restricted to execution as Local Tasks only,
	 * expressed as a factor of the number of Tasks of the Job.
	 */
	public static final float MAX_CONSUMPTION_FAILURES_FACTOR  = 5;

	/** Number of allowed Task execution failures for the Tasks of the Job
	 * before supplier Peer is black-listed (for the current Job),
	 * expressed as a factor of the number of Tasks of the Job.
	 */
	public static final float MAX_SUPPLIER_FAILURES_FACTOR     = 5;

	static {
		assert(MAX_CONSUMPTION_FAILURES_FACTOR >= MAX_SUPPLIER_FAILURES_FACTOR);
		assert(MAX_SUPPLIER_FAILURES_FACTOR > 0);
	}
	
	// ****************************************************************

	private final String owner_peer_id;
	private final int job_id;
	private final int total_job_tasks;

	private final Map<String,SupplierReliabilityHistory> suppliers;
	private final Map<SupplierReliabilityHistory,Set<String>> suppliers_ranking;
	private final Measurement m_consumption_completion_runtime;
	private       int total_completed_consumption_tasks;
	private       int total_cancelled_consumption_tasks;
	
}

