/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PeerConsumptionDataTracker.java

package lbg.peer.register.learning.consumption;

import lbg.common.exception.*;
import lbg.protocol.data.*;
import lbg.task.*;
import java.util.*;

/** A PeerConsumptionDataTracker tracks the data that are known
 * to have been donwloaded by Supplier Peers for recent Consumption Tasks.
 * <p>The tracked data were required
 * by successfully completed Consumption Tasks.
 * <p>To each data id is associated the set of Supplier Peers
 * which have downloaded the data.
 * <p>The tracked data can be searched to locate, and rank Peers
 * that may still have some Resources holding the data in their data cache.
 * @author Cyril Briquet
 */
public class PeerConsumptionDataTracker implements ConsumptionDataTracker {

	public PeerConsumptionDataTracker(int capacity) throws GridException {
		
		if (capacity < 1) {
			throw new GridException("PeerConsumptionDataTracker(): "+
									"illegal capacity");
		}
		
		this.capacity = capacity;
		this.data_downloads = new TreeMap<String,Set<String>>();

	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PeerConsumptionDataTracker]");

		for (String data_id : this.data_downloads.keySet()) {
			sb.append("\n  ->  ");
			sb.append(data_id);
			sb.append("  { ");
			//sb.append("\n  ->                         { ");
			Set<String> downloaders = this.data_downloads.get(data_id);
			int i = 0;
			for (String supplier_peer : downloaders) {
				sb.append(supplier_peer);
				if (++i < downloaders.size()) sb.append(", ");
			}
			sb.append(" }");
		}
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** Tracks target data, i.e. stores the fact that target Task data
	 * have been completyly downloaded by target Supplier Peer.
	 * @param supplier_peer target Supplier Peer
	 * @param t target Task
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized void trackData(String supplier_peer,Task t)
		throws GridException {

		if (supplier_peer == null) {
			throw new GridException("trackData(): illegal Peer id");
		}
		if (t == null) {
			throw new GridException("trackData(): illegal Task");
		}

		//debugging code -- System.out.println("BEGIN PeerConsumptionDataTracker");
		for (MetaGridData mgd : t.getInputMetaGridDataSet()) {
			//debugging_msg_1(supplier_peer,t,mgd);
			String data_id = mgd.id;
			Set<String> downloaders = this.data_downloads.get(data_id);
			if (downloaders == null) {
				downloaders = new TreeSet<String>();
				if (this.data_downloads.put(data_id,downloaders) !=
					null)  {
					throw new GridException("trackData(): "+
											"corrupted data downloads map");
				}
			}
			downloaders.add(supplier_peer);
		}
		//debugging code -- System.out.println("END PeerConsumptionDataTracker\n");

	}

	// ****************************************************************

	/** Return data-aware ranking of Supplier Peers,
	 * sorted by decreasing Storage Affinity,
	 * where input data of target Task may be present.
	 * <p>Note that Peers with no Storage Affinity are not ranked.
	 * <p>Note that input data of target Task may have been ejected
	 * from the data caches of the Suppliers Resources that once held it.
	 * @param t target Task
	 * @return data-aware ranking of Suppliers,
	 * which is a mapping of bytes count to be downloaded to Supplier Peers,
	 * sorted by decreasing Storage Affinity
	 * (i.e. best first, by increasing bytes count to be downloaded)
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized Map<Long,Set<String>> rankPeers(Task t)
		throws GridException {
		
		if (t == null) {
			throw new GridException("rankPeers(): illegal Task");
		}
		
		// Map Suppliers with some data to the bytes count supposed to be stored
		Map<String,Long> suppliers_with_data = new TreeMap<String,Long>();
		
		for (MetaGridData mgd : t.getInputMetaGridDataSet()) {
			String data_id = mgd.id;
			Set<String> downloaders = this.data_downloads.get(data_id);
			if (downloaders == null) {
				continue;
			}
			for (String peer_id : downloaders) {
				Long previous_sa = suppliers_with_data.get(peer_id);
				Long new_sa = null;
				if (previous_sa == null) {
					new_sa = Long.valueOf(mgd.size);
				}
				else {
					new_sa = Long.valueOf(previous_sa.longValue() + mgd.size);
				}
				suppliers_with_data.put(peer_id,new_sa);
			}
		}
		
		long max_sa = t.getInputDataSetSize();
		
		// Sort Suppliers with some data
		// by increasing bytes count to be downloaded
		Map<Long,Set<String>> ranked_suppliers_with_data =
			new TreeMap<Long,Set<String>>();
		Iterator<Map.Entry<String,Long>> suppliers_with_data_it =
			suppliers_with_data.entrySet().iterator();
		while (suppliers_with_data_it.hasNext()) {
			Map.Entry<String,Long> supplier_with_data =
				suppliers_with_data_it.next();
			String supplier_id = supplier_with_data.getKey();
			Long sa = supplier_with_data.getValue();
			// Compute bytes to download
			if (sa == null) {
				throw new GridException("rankPeers(): corrupted mapping");
			}
			long bytes_to_download = max_sa - sa.longValue();
			if ((bytes_to_download < 0) || (bytes_to_download >= max_sa)) {
				throw new GridException("rankPeers(): corrupted "+
										"Storage Affinity value ("+
										bytes_to_download+" bytes to download "+
										"by Peer "+supplier_id+" for Task "+t);
			}
			Long bytes_to_download_key = Long.valueOf(bytes_to_download);

			// Actually rank Supplier Peer
			Set<String> suppliers_with_same_sa =
				ranked_suppliers_with_data.get(bytes_to_download_key);
			if (suppliers_with_same_sa == null) {
				suppliers_with_same_sa = new TreeSet<String>();
				ranked_suppliers_with_data.
					put(bytes_to_download_key,suppliers_with_same_sa);
			}
			suppliers_with_same_sa.add(supplier_id);
		}
		
		return ranked_suppliers_with_data;
		
	}

	// ****************************************************************

	/** Returns the number of tracked data.
	 * @return tracked data count
	 */
	public synchronized int countTrackedData() {
	
		return this.data_downloads.size();
		
	}

	/** Returns an Iterator of the suppliers that presumably
	 * store target tracked data, i.e. tracked suppliers for target track data.
	 * @param data_id tracked data id
	 * @return tracked Suppliers for target tracked data,
	 * or null if data is not tracked
	 */
	public synchronized Iterator<String> trackedSuppliers(String data_id)
		throws GridException {

		if (data_id == null) {
			throw new GridException("trackedSuppliers(): illegal data id");
		}
		
		Set<String> tracked_suppliers = this.data_downloads.get(data_id);
		
		return (tracked_suppliers == null) ?
			null : tracked_suppliers.iterator();
		
	}

	// ****************************************************************

	private void debugging_msg_1(String supplier_peer,Task t,MetaGridData mgd)
		throws GridException {
		
		System.out.println("storing that PeerConsumptionDataTracker: "+
							supplier_peer+" has "+mgd.id+
							" for Job "+t.jobID()+" Task "+t.id());
		
	}
	
	// ****************************************************************

	private int capacity;
	
	// (data_id, (set of suppliers))
	private final Map<String,Set<String>> data_downloads;
	
	// ****************************************************************

}

