/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SupplierReliabilityHistory.java

package lbg.peer.register.learning.consumption;

import lbg.common.exception.*;
import static lbg.common.math.CommonMath.generateHashCode;

/** A SupplierReliabilityHistory describes the reliability of a Supplier Peer
 * for a given Job.
 * <p>Stores Task completion and cancellation counts
 * for target Supplier and Job.
 * @author Cyril Briquet
 */
public class SupplierReliabilityHistory
	implements Comparable<SupplierReliabilityHistory> {
	
	/** Constructs a new SupplierReliabilityHistory.
	 */
	public SupplierReliabilityHistory() {
		
		this.completed_tasks = 0;
		this.cancelled_tasks = 0;
		
	}
	
	/** Constructs a new SupplierReliabilityHistory,
	 * using target Supplier Reliability behavior to set initial
	 * Task completion/cancellation counts,
	 * i.e. this is a copy constructor.
	 * @exception GridException if a parameter is illegal
	 */
	public SupplierReliabilityHistory(SupplierReliabilityHistory b)
		throws GridException {
		
		if (b == null) {
			throw new GridException("SupplierReliabilityHistory(): "+
									"illegal behavior ");
		}
		
		this.completed_tasks = b.getCompletedTasksCount();
		this.cancelled_tasks = b.getCancelledTasksCount();
		
		if (this.completed_tasks < 0) {
			throw new GridException("SupplierReliabilityHistory(): "+
									"illegal completed Tasks count");
		}
		if (this.cancelled_tasks < 0) {
			throw new GridException("SupplierReliabilityHistory(): "+
									"illegal cancelled Tasks count");
		}
		
	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[SupplierReliabilityHistory]");
		sb.append(" completed ");
		sb.append(this.completed_tasks);
		sb.append(" Tasks, cancelled ");
		sb.append(this.cancelled_tasks);
		sb.append(" Tasks");
		
		return sb.toString();
		
	}

	/** Comparable implementation.
	 * <p>Compares 2 Supplier Reliability behaviors.
	 * <p>Smaller value is assigned to the more reliable Peer.
	 * Precisely, Peers with less cancelled Tasks is ranked first (i.e. lower).
	 * For couple of Peers with the same number of cancelled Tasks,
	 * the Peer with the most completed Tasks is ranked first (i.e. lower).
	 */
	public synchronized int compareTo(SupplierReliabilityHistory other) {
	
		int other_completed_tasks = other.getCompletedTasksCount();
		int other_cancelled_tasks = other.getCancelledTasksCount();

		if (this.cancelled_tasks > other_cancelled_tasks) {
			return +1;
		}
		else if (this.cancelled_tasks < other_cancelled_tasks) {
			return -1;
		}
		else /* if (this.cancelled_tasks == other_cancelled_tasks) */ {
			if (this.completed_tasks > other_completed_tasks) {
				return -1;
			}
			else if (this.completed_tasks < other_completed_tasks) {
				return +1;
			}
			else /* if (this.completed_tasks == other_completed_tasks) */ {
				return 0;
			}
		}
		
	}
	
	@Override
	public synchronized boolean equals(Object obj) {
		
		if (obj instanceof SupplierReliabilityHistory == false) {
			return false;
		}
		
		SupplierReliabilityHistory other = (SupplierReliabilityHistory)obj;
		
		return (this.completed_tasks == other.getCompletedTasksCount()) &&
			(this.cancelled_tasks == other.getCancelledTasksCount());
		
	}

	@Override
	public synchronized int hashCode() {
	
		return
			generateHashCode(generateHashCode(this.completed_tasks),
							 generateHashCode(this.cancelled_tasks));
	
	}
	
	// ****************************************************************

	public synchronized int getCompletedTasksCount() {
		
		return this.completed_tasks;
		
	}
	
	public synchronized int getCancelledTasksCount() {
		
		return this.cancelled_tasks;
		
	}

	public synchronized void accountCompletedTask() {
		
		this.completed_tasks++;
		
	}
	
	public synchronized void accountCancelledTask() {
		
		this.cancelled_tasks++;
		
	}

	// ****************************************************************

	private int completed_tasks;
	private int cancelled_tasks;
	
}

