/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PeerSupplyingDataTracker.java

package lbg.peer.register.learning.supplying;

import lbg.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.common.queueing.deque.*;
import lbg.peer.*;
import lbg.peer.core.rms.*;
import lbg.protocol.data.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

/** A PeerSupplyingDataTracker analyzes trends in stored data
 * for Supplying Tasks,
 * using Ranganthan's popularity metric and Briquet's metrics,
 * with a sliding window policy bounding the number of analyzed data.
 * <p>The main purpose of a PeerSupplyingDataTracker is to predict
 * which observed Supplying data will be needed again in the near future,
 * so as to replicate them on the Peer Resources.
 * @author Cyril Briquet
 */
public class PeerSupplyingDataTracker {

	/** Constructs a new PeerSupplyingDataTracker.
	 */
	public PeerSupplyingDataTracker(Peer managed_peer,int capacity)
		throws GridException {

		if (managed_peer == null) {
			throw new GridException("PeerSupplyingDataTracker(): "+
									"illegal Peer");
		}
		if (capacity < 1) {
			throw new GridException("PeerSupplyingDataTracker(): "+
									"illegal capacity");
		}
		
		this.managed_peer = managed_peer;
		this.capacity = capacity;

		this.tmp_most_pop_data   = new ArrayList<MetaGridData>(this.capacity);
		//this.tmp_largest_data    = new ArrayList<MetaGridData>(this.capacity);
		this.tmp_some_rand_data  = new ArrayList<MetaGridData>(this.capacity);
		//this.tmp_most_recent_dat = new ArrayList<MetaGridData>(this.capacity);
		this.ws_contents_weights = new WorkingSetContentsWeights();
		
		this.freq_db = new TreeMap<MetaGridData,PopularityKey>();
		this.q       = new LinkedListDeque<MetaGridData>();
		// J2SE6:      new ArrayDeque<MetaGridData>()
		this.pop_db  = new TreeMap<PopularityKey,Set<MetaGridData>>();
		DecreasingSizeMetaGridDataComparator cmp =
			new DecreasingSizeMetaGridDataComparator();
		this.size_db = new TreeSet<MetaGridData>(cmp);
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PeerSupplyingDataTracker] ");

		if (areDBConsistent() == false) {
			sb.append("*corrupted DB*");
		}
		
		sb.append("\n  ->  Tracked data trends DB size = ");
		sb.append(this.q.size());
		sb.append("\n  ->  Data (by increasing recency) = { ");
		Iterator<MetaGridData> it = this.q.iterator();
		while (it.hasNext()) {
			MetaGridData mgd = it.next();
			sb.append(mgd.id);
			sb.append(" (size=");
			sb.append(mgd.size);
			sb.append(" bytes, popularity=");
			sb.append(this.freq_db.get(mgd).i);
			sb.append(")");
			if (it.hasNext()) {
				sb.append(", ");
			}
		}
		sb.append(" }");
		
		sb.append("\n  ->  Data (by decreasing size)    = { ");
		it = this.size_db.iterator();
		while (it.hasNext()) {
			MetaGridData mgd = it.next();
			sb.append(mgd.id);
			sb.append(" (size=");
			sb.append(mgd.size);
			sb.append(" bytes, popularity=");
			sb.append(this.freq_db.get(mgd).i);
			sb.append(")");
			if (it.hasNext()) {
				sb.append(", ");
			}
		}
		sb.append(" }");

		Iterator<PopularityKey> it2 = this.pop_db.keySet().iterator();
		if (it2.hasNext()) sb.append("\n");
		// J2SE6: ((TreeMap<Integer,Set<MetaGridData>>)this.pop_db).
		//        descendingKeySet().iterator();
		while (it2.hasNext()) {
			sb.append("  ->  ");
			PopularityKey popularity = it2.next();
			sb.append(popularity.i);
			sb.append("* = { ");
			Set<MetaGridData> s = this.pop_db.get(popularity);
			Iterator<MetaGridData> dummy = s.iterator();
			while (dummy.hasNext()) {
				MetaGridData mgd = dummy.next();
				sb.append(mgd.id);
				if (dummy.hasNext()) {
					sb.append(", ");
				}
			}
			sb.append(" }");
			if (it2.hasNext()) sb.append("\n");
		}
		
		return sb.toString();

	}
	
	// ****************************************************************

	/** Add data (from a presumably Supplying Task) to be stored
	 * and analyzed.
	 */
	public synchronized void add(MetaGridData[] ds) throws GridException {

		if (ds == null) {
			throw new GridException("add(): illegal data set");
		}
		
		assert(qInvariant());
		
		for (MetaGridData newest : ds) {
			int q_s = q.size();
			if ((q_s < this.capacity) || (q_s == 0)) {
				// add new data
				this.addData(newest);
				// add new size
				this.addSize(newest);
			}
			else if (q_s > 0) /* && (q_s == this.capacity) */ {
				// Manage q overflow
				MetaGridData oldest = this.q.removeFirst();
				if (newest.equals(oldest)) {
					// simply reinsert oldest data into q
					this.q.insertLast(newest);
				}
				else {
					// remove old size
					this.removeSize(oldest);
					// decrease popularity of oldest in freq_db
					// update olddest in pop_db, to change its key
					this.decreasePopularityOfData(oldest);
					// add new data
					this.addData(newest);
					// add new size
					this.addSize(newest);
				}
			}
		}
		
		assert(qInvariant());

		/* Implementation notes:
		   
		   . this insertion algorithm is not optimal in the special case
		   	 where ds.length > this.capacity, but well ...
			 
		   . when there's an overflow, and newest != oldest:
		   	 * for each data in EJECTED data set:
			   remove from q ;
			   decrease popularity in freq Map (and remove if popularity == 0) ;
			   update popularity in pop Map
			 * for each data in ENTERING data set:
			   add to q ;
			   increase popularity in freq Map ;
			   update popularity in pop Map
		 */

	}

	// ****************************************************************

	/** Replicates data to target Resources, using prediction
	 * based on analysed trends.
	 * @param it target Resources (as an Iterator over their ResourceUsageData)
	 * @param res_count target Resources count
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized void replicateData(Iterator<ResourceUsageData> it,
										   int res_count)
		throws GridException {
		
		if (it == null) {
			throw new GridException("replicateData(): "+
									"illegal Resource iterator");
		}
		if (res_count < 1) {
			throw new GridException("replicateData(): illegal Resource count");
		}
	
		// First determine maximum WorkingSet availability
		// (i.e. number of free data slots in the WorkingSet)
		// for all *considered* Resources where replication will take place
		// (i.e. Resources preselected by the ResourceManager)
		ArrayList<WorkingSet> al = new ArrayList<WorkingSet>(res_count);
		int max_available_capacity = 0;
		while (it.hasNext()) {
			ResourceUsageData rud = it.next();
			WorkingSet ws = rud.working_set;
			int capacity = ws.availableCapacity();
			if (capacity > max_available_capacity) {
				max_available_capacity = capacity;
			}
			al.add(ws);
		}
		if (al.size() != res_count) {
			throw new GridException("replicateData(): corrupted Resource set");
		}

		// Extract data sets, using max_available_capacity
		int extracted_data = 0;
		extracted_data += extractMostPopularData(max_available_capacity);
		//extracted_data += extractLargestData(max_available_capacity);
		extracted_data += extractRandomData(max_available_capacity);
		
		// Actually replicate data:
		// (1) fill WorkingSet, (2) send WorkingSet to the managed Resource
		if (extracted_data > 0) {
			for (WorkingSet ws : al) {
				this.fillWorkingSet(ws);
				if (ws.managedResource().setWorkingSet(ws.getData()) == false) {
					this.managed_peer.logger().
						info("replicateData(): working set rejected by "+
							 "Resource "+ws.managedResource().id());
				}
			}
		}
	
		this.tmp_most_pop_data.clear();
		//this.tmp_largest_data.clear();
		this.tmp_some_rand_data.clear();
		//this.tmp_most_recent_data.clear();
		
	}
	
	// ****************************************************************

	/** Test method.
	 */
	public static void main(String[] args) {
		
		try {
			PeerSupplyingDataTracker dt =
				new PeerSupplyingDataTracker(null,2); //10);
				// a dummy Peer should be given to constructor instead of null
			
			GridData[] dummy_array = new GridData[1];
			MetaGridData[] ds = new MetaGridData[1];
			
			ds[0] = new MetaGridData("abc",1024,dummy_array);
			dt.add(ds);
			System.out.println("beat");
			System.out.println(dt);
			
			ds[0] = new MetaGridData("def",2048,dummy_array);
			dt.add(ds);
			System.out.println("beat");
			System.out.println(dt);

			ds[0] = new MetaGridData("def",2048,dummy_array);
			dt.add(ds);
			System.out.println("beat");
			System.out.println(dt);

			ds[0] = new MetaGridData("abc",1024,dummy_array);
			dt.add(ds);
			System.out.println("beat");
			System.out.println(dt);
			
			ds[0] = new MetaGridData("def",2048,dummy_array);
			dt.add(ds);
			System.out.println("beat");
			System.out.println(dt);

			ds[0] = new MetaGridData("abc",1024,dummy_array);
			dt.add(ds);
			System.out.println("beat");
			System.out.println(dt);

			ds[0] = new MetaGridData("abc",1024,dummy_array);
			dt.add(ds);
			System.out.println("beat");
			System.out.println(dt);

			ds[0] = new MetaGridData("abc",1024,dummy_array);
			dt.add(ds);
			System.out.println("beat");
			
			System.out.println(dt);

		}
		catch (GridException e) {
			e.printStackTrace();
		}
		finally {
			System.exit(0);
		}
		
	}
	
	// ****************************************************************

	/** Fills target WorkingSet with data that is predicted
	 * to be needed by the next Supplying Tasks.
	 * Prediction is currently based on Briquet's metrics
	 * (most popular data, plus some random data).
	 * @param ws target WorkingSet
	 * @exception GridException if a parameter is illegal
	 */
	private synchronized void fillWorkingSet(WorkingSet ws)
		throws GridException {

		/* Implementation notes:
		
		   Precondition: ArrayLists this.tmp_*_data MUST be filled !
		*/
		
		if (ws == null) {
			throw new GridException("fillWorkingSet(): illegal "+
									"target working set");
		}

		// Get working set available capacity
		// (== nominal capacity - locked data count)
		
		int ws_available_capacity = ws.availableCapacity();
		
		if (ws_available_capacity > this.q.size()) {
			ws_available_capacity = this.q.size();
		}

		// Determine the maximum number of data to extract
		// from each of the given data sets, and given the target weights

		this.ws_contents_weights.
			assignWeightsWithStrongPopularityBias(ws_available_capacity);

		// Clears the WorkingSet from unlocked data
		ws.clearAllUnlockedData();
		
		// Actually fill the WorkingSet
		// reminder: add*Data() retval indicates whether data was already stored

		for (int i = 0, d = this.ws_contents_weights.mostPopularDataWeight();
			 (i < d) && (i < this.tmp_most_pop_data.size()) ; i++) {
			ws.addSupplyingData(this.tmp_most_pop_data.get(i));
		}
		/*for (int i = 0, d = this.ws_contents_weights.largestDataWeight();
			 (i < d) && (i < this.tmp_largest_data.size()) ; i++) {
			ws.addSupplyingData(this.tmp_largest_data.get(i));
		}*/
		for (int i = 0, d = this.ws_contents_weights.randomDataWeight();
			 (i < d) && (i < this.tmp_some_rand_data.size()) ; i++) {
			ws.addSupplyingData(this.tmp_some_rand_data.get(i));
		}
		
	}

	/** Extracts, and copies, k most popular data to target ArrayList,
	 * by decreasing importance.
	 * @param k number of data to copy
	 * @return number of data actually copied
	 * @exception GridException if a parameter is illegal
	 */
	private synchronized int extractMostPopularData(int k)
		throws GridException {

		ArrayList<MetaGridData> al = this.tmp_most_pop_data;
		
		if (al == null) {
			throw new GridException("getMostPopular(): illegal ArrayList");
		}
		if (k < 1) {
			throw new GridException("getMostPopular(): illegal ArrayList");
		}

		int count = 0;
		Iterator<Set<MetaGridData>> it = this.pop_db.values().iterator();
		while (it.hasNext() && (count < k)) {
			Set<MetaGridData> s = it.next();
			for (MetaGridData mgd : s) {
				al.add(mgd);
				if (++count >= k) {
					break;
				}
			}
		}

		return count;
		
	}
	
	/** Extracts, and copies, k largest data to target ArrayList,
	 * by decreasing importance.
	 * @param k number of data to copy
	 * @return number of data actually copied
	 * @exception GridException if a parameter is illegal
	 */
	private synchronized int extractLargestData(int k) throws GridException {
		
		throw new UnsupportedOperationException("extractLargestData(): "+
												"currently unsupported");
		/*
		ArrayList<MetaGridData> al = this.tmp_largest_data;
		
		if (al == null) {
			throw new GridException("getLargestData(): illegal ArrayList");
		}
		if (k < 1) {
			throw new GridException("getLargestData(): illegal ArrayList");
		}

		int count = 0;
		Iterator<MetaGridData> it = this.size_db.iterator();
		while (it.hasNext()) {
			MetaGridData mgd = it.next();
			al.add(mgd);
			if (++count >= k) {
				break;
			}
		}

		return count;
		*/
		
	}
	
	/** Extracts, and copies k random data to target ArrayList,
	 * by decreasing importance.
	 * @param k number of data to copy
	 * @return number of data actually copied
	 * @exception GridException if a parameter is illegal
	 */
	private synchronized int extractRandomData(int k) throws GridException {
		
		ArrayList<MetaGridData> al = this.tmp_some_rand_data;
		
		if (al == null) {
			throw new GridException("getRandomData(): illegal ArrayList");
		}
		if (k < 1) {
			throw new GridException("getRandomData(): illegal ArrayList");
		}

		int len = this.freq_db.size();
		int count = 0;
		int i = 0;
		Iterator<MetaGridData> it = this.freq_db.keySet().iterator();
		while (it.hasNext()) {
			MetaGridData mgd = it.next();
			// If we're here, we're sure that (len >= 1)
			// float sample = CommonMath.sampleRandomDistWithinUnity();
			// float proba  = 1.0f / (float)len;
			// if (sample <= proba) -> success, the data is selected
			// len-i == remaining data to evaluate
			// k-count == number of data yet to get
			if (((len-i) <= (k-count)) ||
				(CommonMath.sampleUniformDistWithinUnity() <= 1.0f/(float)len)) {
				al.add(mgd);
				if (++count >= k) {
					break;
				}
			}
			++i;
		}
		
		return count;
		
	}

	/** Extracts, and copies, k most popular data to target ArrayList,
	 * by decreasing importance.
	 * @param k number of data to copy
	 * @return number of data actually copied
	 * @exception GridException if a parameter is illegal
	 */
	private synchronized int extractMostRecentData(int k) throws GridException {

		throw new UnsupportedOperationException("extractMostRecentData(): "+
												"currently unsupported");
		/*
		ArrayList<MetaGridData> al = this.tmp_most_recent_data;

		if (al == null) {
			throw new GridException("getMostRecentData(): illegal ArrayList");
		}
		if (k < 1) {
			throw new GridException("getMostRecentData(): illegal ArrayList");
		}

		int count = 0;
		Iterator<MetaGridData> it = this.q.descendingIterator();
		while (it.hasNext()) {
			MetaGridData mgd = it.next();
			al.add(mgd);
			if (++count >= k) {
				break;
			}
		}

		return count;
		*/
		
	}

	// ****************************************************************

	private synchronized void addData(MetaGridData newest)
		throws GridException {

		if (newest == null) {
			throw new GridException("addData(): illegal data");
		}
		
		// Update freq db
		PopularityKey freq = this.freq_db.get(newest);
		
		if (freq == null) {
			freq = new PopularityKey(1);
			this.freq_db.put(newest,freq);
		}
		else {
			// Remove old entry from popularity db
			removeFromPopularityDB(freq,newest);
			freq.increase();
		}

		// Update q
		this.q.insertLast(newest);
		
		// Update popularity db, with increased key
		addToPopularityDB(freq,newest);
		
	}
	
	private synchronized void decreasePopularityOfData(MetaGridData oldest)
		throws GridException {

		if (oldest == null) {
			throw new GridException("decreasePopularityOfData(): illegal data");
		}

		// Decrease popularity of data in freq_db
		PopularityKey freq = this.freq_db.get(oldest);
		if (freq == null) {
			throw new GridException("decreasePopularityOfData(): "+
									"corrupted freq db");
		}
		if (freq.i == 1) {
			if (this.freq_db.remove(oldest) == null) {
				throw new GridException("decreasePopularityOfData(): "+
										"corrupted freq db");
			}
		}
		
		// Remove old entry from popularity db
		removeFromPopularityDB(freq,oldest);

		// Update popularity db, with decreased key
		freq.decrease();
		if (freq.i > 0) {
			addToPopularityDB(freq,oldest);
		}
		
	}
	
	private synchronized void addToPopularityDB(PopularityKey popularity,
												MetaGridData mgd)
		throws GridException {
		
		if ((popularity == null) || (popularity.i <= 0)) {
			throw new GridException("addToPopularityDB(): illegal popularity");
		}
		if (mgd == null) {
			throw new GridException("addToPopularityDB(): illegal data");
		}
		
		Set<MetaGridData> s = this.pop_db.get(popularity); 
		if (s == null) {
			s = new TreeSet<MetaGridData>();
			if (this.pop_db.put(new PopularityKey(popularity.i),s) !=
				null) {
				throw new GridException("addToPopularityDB(): "+
										"corrupted pop db");
			}
		}
		s.add(mgd);
		
	}
	
	private synchronized void removeFromPopularityDB(PopularityKey popularity,
													 MetaGridData mgd)
		throws GridException {
		
		if ((popularity == null) || (popularity.i <= 0)) {
			throw new GridException("removeFromPopularityDB(): "+
									"illegal popularity");
		}
		if (mgd == null) {
			throw new GridException("removeFromPopularityDB(): illegal data");
		}

		Set<MetaGridData> s = this.pop_db.get(popularity);
		if (s == null) {
			throw new GridException("removeFromPopularityDB(): "+
									"corrupted popularity db");
		}
		if (s.remove(mgd) == false) {
			throw new GridException("removeFromPopularityDB(): "+
									"corrupted popularity db");
		}
		if (s.size() == 0) {
			if (this.pop_db.remove(popularity) == null) {
				throw new GridException("removeFromPopularityDB(): "+
										"corrupted popularity db");
			}
		}

	}
	
	private synchronized void addSize(MetaGridData mgd) throws GridException {
	
		if (mgd == null) {
			throw new GridException("addSize(): illegal data");
		}

		this.size_db.add(mgd);
		
	}
	
	// must be called BEFORE decreasePopularityOfData()
	private synchronized void removeSize(MetaGridData mgd)
		throws GridException {
	
		if (mgd == null) {
			throw new GridException("removeSize(): illegal data");
		}
		
		PopularityKey popularity = this.freq_db.get(mgd);
		if (popularity == null) {
			throw new GridException("removeSize(): corrupted freq db");
		}
		
		if (popularity.i > 1) {
			return;
		}

		if (this.size_db.remove(mgd) != true) {
			throw new GridException("removeSize(): corrupted size db");
		}
		
	}

	// ****************************************************************

	private synchronized boolean qInvariant() {
		
		int q_s       = this.q.size();
		int freq_db_s = this.freq_db.size();
		int pop_db_s  = this.pop_db.size();
		int size_db_s = this.size_db.size();
		
		boolean retval = (q_s >= 0) && (q_s <= this.capacity) &&
			((q_s > 0) ? (freq_db_s > 0) : (freq_db_s == 0)) &&
			(pop_db_s <= freq_db_s) && (freq_db_s <= q_s) &&
			(size_db_s == freq_db_s);
			
		if ((LightweightBarteringGrid.getMode() == LIVE_MODE) || (false)) {
			if (areDBConsistent() == false) {
				return false;
			}
		}
		
		return retval;
		
	}
	
	private synchronized boolean areDBConsistent() {

		// Check that frequencies in freq_db match q's contents
		boolean consistent_db = true;
		Iterator<MetaGridData> it = this.freq_db.keySet().iterator();
		while (it.hasNext()) {
			MetaGridData mgd = it.next();
			Iterator<MetaGridData> dummy = this.q.iterator();
			int count = 0;
			while (dummy.hasNext()) {
				MetaGridData dummy_data = dummy.next();
				if ((mgd.id.compareToIgnoreCase(dummy_data.id) == 0) &&
					(mgd.equals(dummy_data))) {
					count++;
				}
			}
			PopularityKey freq = this.freq_db.get(mgd);
			if ((freq == null) || (freq.i != count)) {
				consistent_db = false;
				break;
			}
		}

		if (consistent_db == false) return false;

		// Check that frequencies in pop_db match freq_db's contents
		Iterator<PopularityKey> it2 = this.pop_db.keySet().iterator();
		while (it2.hasNext()) {
			PopularityKey key = it2.next();
			Set<MetaGridData> s = this.pop_db.get(key);
			if (s == null) {
				consistent_db = false;
				break;
			}
			Iterator<MetaGridData> dummy = s.iterator();
			while (dummy.hasNext()) {
				MetaGridData mgd = dummy.next();
				PopularityKey freq = this.freq_db.get(mgd);
				if ((freq == null) || (key.i != freq.i)) {
					consistent_db = false;
					break;
				}
			}

		}
		
		if (consistent_db == false) return false;

		// Check that frequencies in pop_db match freq_db's contents
		it = this.size_db.iterator();
		while (it.hasNext()) {
			MetaGridData mgd = it.next();
			if (this.freq_db.get(mgd) == null) {
				consistent_db = false;
				break;
			}
		}
		
		return consistent_db;
		
	}

	// ****************************************************************

	private final Peer managed_peer;
	private final int capacity;
	
	private final Map<MetaGridData,PopularityKey> freq_db;     // frequency db
	private final Deque<MetaGridData> q;                       // recency queue

	private final Map<PopularityKey,Set<MetaGridData>> pop_db; // popularity db
	private final Set<MetaGridData> size_db;                   // size db
	
	private final ArrayList<MetaGridData> tmp_most_pop_data;
	//private final ArrayList<MetaGridData> tmp_largest_data;
	private final ArrayList<MetaGridData> tmp_some_rand_data;
	//private final ArrayList<MetaGridData> tmp_most_recent_data;
	private final WorkingSetContentsWeights ws_contents_weights;

	// ****************************************************************

	private static class PopularityKey implements Comparable<PopularityKey> {
		
		public PopularityKey(int i) throws GridException {

			if (i < 0) {
				throw new GridException("PopularityKey(): illegal popularity");
			}
			
			this.i = i;
			
		}
		
		// Sort by decreasing popularity
		public synchronized int compareTo(PopularityKey other) {
		
			if (this.i < other.i) {
				return +1;
			}
			else if (this.i > other.i) {
				return -1;
			}
			else /* if (this.i == other.i) */ {
				return 0;
			}
			
		}
		
		@Override
		public boolean equals(Object obj) {
			
			if (obj instanceof PopularityKey == false) {
				return false;
			}
			
			return (this.compareTo((PopularityKey)obj) == 0);
			
		}

		@Override
		public int hashCode() {
		
			return CommonMath.generateHashCode(this.i);
		
		}

		public synchronized void increase() {
		
			this.i++;
			
		}

		public synchronized void decrease() throws GridException {

			this.i--;
			
			if (i < 0) {
				throw new GridException("decrease(): illegal popularity");
			}

		}

		public int i;
		
	}
	
	private static class WorkingSetContentsWeights {
		
		public WorkingSetContentsWeights() throws GridException {
			
			this.most_pop_data_w = 0;
			this.largest_data_w = 0;
			this.rand_data_w = 0;
			
		}

		/** Assign weights uniformly.
		 * <p>{ Most Popular = 0.33, Largest = 0.33, Random = 0.33 }
		 */
		public synchronized
			void assignWeightsUniformly(int ws_available_capacity)
			throws GridException {

			if (ws_available_capacity < 1) {
				throw new GridException("assignWeightsUniformly(): "+
										"illegal Working Set capacity");
			}
			
			int d3 = ws_available_capacity / 3;
			int m3 = ws_available_capacity % 3;

			// Assign uniform weights
			this.most_pop_data_w  = d3 + ((m3  > 0) ? 1 : 0);
			this.largest_data_w   = d3 + ((m3 == 2) ? 1 : 0);
			this.rand_data_w      = d3;
			
			assert(areWeightsSummingToOne(ws_available_capacity));
			
		}
		
		/** Assign weights with a bias towards popularity.
		 * <p>{ Most Popular = 0.50, Largest = 0.25, Random = 0.25 }
		 */
		public synchronized void
			assignWeightsWithPopularityBias(int ws_available_capacity)
			throws GridException {

			if (ws_available_capacity < 1) {
				throw new GridException("assignWeightsWithPopularityBias(): "+
										"illegal Working Set capacity");
			}
			
			int d2 = ws_available_capacity / 2;
			int m2 = ws_available_capacity % 2;

			this.most_pop_data_w  = d2 + m2;
			
			int dd2 = d2 / 2;
			int mm2 = d2 % 2;
			
			this.largest_data_w   = dd2 + mm2;
			this.rand_data_w      = dd2;
			
			assert(areWeightsSummingToOne(ws_available_capacity));
			
		}

		/** Assign weights with a bias towards popularity.
		 * and against size.
		 * <p>{ Most Popular = 0.80, Largest = 0.00, Random = 0.20 }
		 */
		public synchronized void
			assignWeightsWithStrongPopularityBias(int ws_available_capacity)
			throws GridException {

			if (ws_available_capacity < 1) {
				throw new GridException("assignWeightsWith"+
										"StrongPopularityBias(): "+
										"illegal Working Set capacity");
			}
			
			this.most_pop_data_w  =
				(int)Math.ceil(((float)ws_available_capacity) * 0.8f);
			this.largest_data_w   = 0;
			this.rand_data_w      =
				ws_available_capacity - this.most_pop_data_w;
			
			assert(areWeightsSummingToOne(ws_available_capacity));
			
		}
		
		
		/** Assign weights with a complete bias towards popularity.
		 * <p>{ Most Popular = 1.0, Largest = 0.0, Random = 0.0 }
		 */
		public synchronized void
			assignWeightsWithCompletePopularityBias(int ws_available_capacity)
			throws GridException {

			if (ws_available_capacity < 1) {
				throw new GridException("assignWeightsWithPopularityBias(): "+
										"illegal Working Set capacity");
			}
			
			this.most_pop_data_w  = ws_available_capacity;
			this.largest_data_w   = 0;
			this.rand_data_w      = 0;
			
			assert(areWeightsSummingToOne(ws_available_capacity));
			
		}

		public synchronized int mostPopularDataWeight() {
		
			return this.most_pop_data_w;
			
		}
		
		public synchronized int largestDataWeight() {
			
			return this.largest_data_w;

		}
		
		public synchronized int randomDataWeight() {
		
			return this.rand_data_w;
			
		}
		
		private boolean areWeightsSummingToOne(int ws_available_capacity) {
			
			return ws_available_capacity ==
				this.most_pop_data_w + this.largest_data_w + this.rand_data_w;

		}
		
		private int most_pop_data_w;
		private int largest_data_w;
		private int rand_data_w;

	}

	// ****************************************************************

}

