/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ConsumptionGrantsContainer.java

package lbg.peer.register.negotiation;

import lbg.common.*;
import lbg.common.exception.*;
import lbg.peer.register.*;
import lbg.peer.service.*;
import java.util.*;

/** A ConsumptionGrantsContainer contains received Consumption grants
 * that have been received since the received Consumption grants
 * were last processed.
 * @author Cyril Briquet
 */
public class ConsumptionGrantsContainer {

	public ConsumptionGrantsContainer(Collection<PeerProfile> cgs,int s)
		throws GridException {
		
		if (cgs == null) {
			throw new GridException("ConsumptionGrantsContainer(): "+
									"illegal Grantors");
		}
		if (s <= 0) {
			throw new GridException("ConsumptionGrantsContainer(): "+
									"illegal Grantors count");
		}

		this.grantors = new PeerProfile[s];
		this.pending_grantors = new TreeMap<String,Integer>();
		this.pending_consumption_grants = new int[s];
		this.initial_consumption_grants = new int[s];
		this.used_consumption_grants    = new int[s];
		int dummy = 0;
		
		int i = 0;
		for (PeerProfile peer_profile : cgs) {
			this.grantors[i] = peer_profile;
			if (this.pending_grantors.
					put(peer_profile.id(),Tank.newInteger(i)) != null) {
				throw new GridException("ConsumptionGrantsContainer(): "+
										"corrupted data, more than one set of "+
										"Consumption Grants from "+
										peer_profile.id());
			}

			this.pending_consumption_grants[i] =
				peer_profile.getNegotiationProfile().
					getAndResetReceivedConsumptionGrants();
			this.initial_consumption_grants[i] =
				this.pending_consumption_grants[i];
			this.used_consumption_grants[i] = 0;

			dummy += this.pending_consumption_grants[i];
			++i;
		}
		if (i != s) { // consistency check
			throw new GridException("ConsumptionGrantsContainer(): "+
									"corrupted data (i="+i+", s="+s+")");
		}
		
		this.total_consumption_grants = dummy;
		this.pending_consumption_grants_count = this.total_consumption_grants;

		if (this.total_consumption_grants < 1) {
			throw new GridException("ConsumptionGrantsContainer(): "+
									"illegal number of Consumption Grants");
		}
		
	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ConsumptionGrantsContainer]");
		
		return sb.toString();
		
	}
	
	// ****************************************************************

	public synchronized int size() {
		
		return this.grantors.length;
		
	}
	
	public synchronized PeerProfile[] grantors() {
	
		return this.grantors.clone();
		
	}
	
	public synchronized PeerProfile grantorProfile(int i)
		throws GridException {
	
		if ((i < 0) || (i >= this.grantors.length)) {
			throw new GridException("grantorProfile(): illegal index");
		}
	
		return this.grantors[i];
		
	}

	public synchronized ExternalPeerHandle grantor(int i)
		throws GridException {
	
		if ((i < 0) || (i >= this.grantors.length)) {
			throw new GridException("grantor(): illegal index");
		}
	
		return this.grantors[i].getHandle();
		
	}

	// ****************************************************************

	/** Returns the id set of pending grantors.
	 * @return pending grantors id set
	 */
	public synchronized Collection<String> pendingGrantors() {
	
		return this.pending_grantors.keySet();
		
	}
	
	/** Returns the number of pending grantors.
	 * @return pending grantors count
	 */
	public synchronized int pendingGrantorsCount() {
	
		return this.pending_grantors.size();
		
	}
	
	/** Returns an Iterator of the index of pending grantors.
	 * @return pending grantors index Iterator
	 */
	public synchronized Iterator<Integer> pendingGrantorsIndexIterator() {
		
		return this.pending_grantors.values().iterator();
		
	}
	
	/** Returns the index of pending grantors.
	 * @return pending grantors index
	 */
	public synchronized Collection<Integer> pendingGrantorsIndex() {
		
		return this.pending_grantors.values();
		
	}
	
	/** Returns the index of target pending grantor.
	 * @param peer_profile target pending grantor profile
	 * @return pending grantor index
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized Integer pendingGrantorIndex(PeerProfile peer_profile)
		throws GridException {
		
		if (peer_profile == null) {
			throw new GridException("pendingGrantorIndex(): "+
									"illegal Peer profile");
		}
		
		return pendingGrantorIndex(peer_profile.id());
		
	}

	/** Returns the index of target pending grantor.
	 * @param peer_id target pending grantor id
	 * @return pending grantor index
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized Integer pendingGrantorIndex(String peer_id)
		throws GridException {
		
		if (peer_id == null) {
			throw new GridException("pendingGrantorIndex(): "+
									"illegal Peer id");
		}
		
		Integer idx = this.pending_grantors.get(peer_id);
		if (idx == null) {
			throw new GridException("pendingGrantoIndex(): "+
									"corrupted pending grantors mapping");
		}
		
		return idx;
		
	}

	/**
	 * @return random pending grantor index
	 */
	public synchronized int randomPendingGrantorIndex()
		throws GridException {
		
		if (this.pending_grantors.size() < 1) {
			throw new GridException("randomPendingGrantorIndex(): "+
									"no pending grantors");
		}
		
		return this.pending_grantors.get(this.pending_grantors.firstKey());
		
	}

	// ****************************************************************

	public synchronized int initialConsumptionGrants(int i)
		throws GridException {
		
		if ((i < 0) || (i >= this.grantors.length)) {
			throw new GridException("initialConsumptionGrants(): "+
									"illegal index");
		}

		return this.initial_consumption_grants[i];
		
	}
	
	public synchronized int pendingConsumptionGrants(int i)
		throws GridException {
		
		if ((i < 0) || (i >= this.grantors.length)) {
			throw new GridException("pendingConsumptionGrants(): "+
									"illegal index");
		}

		return this.pending_consumption_grants[i];
		
	}

	public synchronized int usedConsumptionGrants(int i)
		throws GridException {
		
		if ((i < 0) || (i >= this.grantors.length)) {
			throw new GridException("usedConsumptionGrants(): "+
									"illegal index");
		}

		return this.used_consumption_grants[i];
		
	}

	/** Uses all Consumption Grants from target grantor.
	 * <p>Target grantor is removed from the index of remaining grantors.
	 * @param i index of target grantor
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized void useAllConsumptionGrants(int i)
		throws GridException {

		useConsumptionGrants(i,this.pending_consumption_grants[i]);
		
	}
	
	/** Spills, i.e. fakes to use, all Consumption Grants from target grantor.
	 * <p>Target grantor is removed from the index of remaining grantors.
	 * @param i index of target grantor
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized void spillAllConsumptionGrants(int i)
		throws GridException {

		spillConsumptionGrants(i,this.pending_consumption_grants[i]);
		
	}

	/** Uses some Consumption Grants from target grantor.
	 * <p>If target grantor has no more remaining Consumption Grants,
	 * it is removed from the index of remaining grantors.
	 * @param i index of target grantor
	 * @param count number of used Consumption Grants
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized void useConsumptionGrants(int i,int count)
		throws GridException {
		
		useConsumptionGrants(i,count,false);

	}

	/** Spills, i.e. fakes to use, some Consumption Grants from target grantor.
	 * <p>If target grantor has no more remaining Consumption Grants,
	 * it is removed from the index of remaining grantors.
	 * @param i index of target grantor
	 * @param count number of used Consumption Grants
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized void spillConsumptionGrants(int i,int count)
		throws GridException {
		
		useConsumptionGrants(i,count,true);

	}
	
	/** Uses some Consumption Grants from target grantor.
	 * <p>If target grantor has no more remaining Consumption Grants,
	 * it is removed from the index of remaining grantors.
	 * @param i index of target grantor
	 * @param count number of used Consumption Grants
	 * @param spill true if grants are spilled rather than actually used
	 * @exception GridException if a parameter is illegal
	 */
	private synchronized void useConsumptionGrants(int i,int count,
												   boolean spill)
		throws GridException {
		
		if ((i < 0) || (i >= this.grantors.length)) {
			throw new GridException("useConsumptionGrants(): "+
									"illegal index");
		}
		if (count <= 0) {
			throw new GridException("useConsumptionGrants(): "+
									"illegal grants count");
		}
		if (count > this.pending_consumption_grants[i]) {
			throw new GridException("useConsumptionGrants(): "+
									"not enough available Consumption Grants");
		}

		this.pending_consumption_grants_count -= count;
		this.pending_consumption_grants[i] -= count;
		if (spill == false) {
			this.used_consumption_grants[i] += count;
		}
		
		if (this.pending_consumption_grants[i] == 0) {
			if (this.pending_grantors.remove(this.grantors[i].id()) == null) {
				throw new GridException("useConsumptionGrants(): "+
										"corrupted internal data");
			}
		}
		
	}

	public synchronized int totalConsumptionGrants() {
		
		return this.total_consumption_grants;
		
	}

	public synchronized int pendingConsumptionGrants() {
		
		return this.pending_consumption_grants_count;
		
	}

	public synchronized int usedConsumptionGrants() {
		
		return
			this.total_consumption_grants-this.pending_consumption_grants_count;
		
	}

	// ****************************************************************

	private final PeerProfile[] grantors;
	private final TreeMap<String,Integer> pending_grantors;
	private final int[] initial_consumption_grants;
	private final int[] pending_consumption_grants;
	private final int[] used_consumption_grants;
	private final int total_consumption_grants;
	private       int pending_consumption_grants_count;

}

