/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PeerNegotiationProfile.java

package lbg.peer.register.negotiation;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.*;
import lbg.peer.service.*;

/**
 * @author Cyril Briquet
 */
public class PeerNegotiationProfile {

	public PeerNegotiationProfile(Peer managed_peer,
								  ExternalPeerHandle profiled_peer)
		throws GridException {
			
		if (managed_peer == null) {
			throw new GridException("PeerProfile(): illegal managed Peer");
		}
		if (profiled_peer == null) {
			throw new GridException("PeerProfile(): illegal profiled Peer");
		}
		
		this.managed_peer = managed_peer;
		this.profiled_peer = profiled_peer;
		
		this.received_supplying_requests = 0;
		this.total_received_supplying_requests = 0;
		
		this.total_sent_consumption_grants = 0;

		this.total_sent_supplying_requests = 0;

		this.received_consumption_grants = 0;
		this.total_received_consumption_grants = 0;
		
		this.total_used_consumption_grants = 0;

	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();

		sb.append("[PeerNegotiationProfile of ");
		sb.append(this.profiled_peer.id());
		sb.append(",  by ");
		sb.append(this.managed_peer.id());
		sb.append("]");

		sb.append("\n  ->  ");
		sb.append("(Supplier role) granted Supplying Requests=");
		sb.append(this.total_sent_consumption_grants);
		sb.append("/");
		sb.append(this.total_received_supplying_requests);
		sb.append(" (");
		sb.append(CommonMath.
					div_label(this.total_sent_consumption_grants,
							  this.total_received_supplying_requests,100));
		sb.append("%)");
		
		sb.append("\n  ->  ");
		sb.append("(Consumer role) used Consumption Grants=");
		sb.append(this.total_used_consumption_grants);
		sb.append("/");
		sb.append(this.total_received_consumption_grants);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.total_used_consumption_grants,
									   this.total_received_consumption_grants,
									   100));
		sb.append("%)");
		sb.append("\n  ->  ");
		sb.append("(Consumer role) received Consumption Grants=");
		sb.append(this.total_received_consumption_grants);
		sb.append("/");
		sb.append(this.total_sent_supplying_requests);
		sb.append(" (");
		sb.append(CommonMath.div_label(this.total_received_consumption_grants,
									   this.total_sent_supplying_requests,
									   100));
		sb.append("%)");
		
		return sb.toString();
		
	}
	
	// ****************************************************************

	/** Increases Supplying Requests count by specified amount.
	 * @param tasks amount of Supplying Requests 
	 * by which to increase current Supplying Requests count
	 * @exception GridException if target parameter is illegal
	 */
	public synchronized void setReceivedSupplyingRequests(int tasks)
		throws GridException {
			
		if (tasks <= 0) {
			throw new GridException("setSupplyingRequests(): "+
									"illegal Tasks count ("+tasks+")");
		}
		
		this.received_supplying_requests = tasks;
		this.total_received_supplying_requests += tasks;
		
	}

	public synchronized int getAndResetReceivedSupplyingRequests()
		throws GridException {
		
		int retval = this.received_supplying_requests;
		this.received_supplying_requests = 0;
		return retval;
		
	}

	public synchronized int getReceivedSupplyingRequests() {
		
		return this.received_supplying_requests;
		
	}
	
	public synchronized long getTotalReceivedSupplyingRequests() {
		
		return this.total_received_supplying_requests;
		
	}

	public synchronized void
		accountSentConsumptionGrants(int tasks) throws GridException {
			
		if (tasks <= 0) {
			throw new GridException("accountSentConsumptionGrants(): "+
									"illegal Tasks count");
		}
		
		this.total_sent_consumption_grants += tasks;
			
	}

	public synchronized long getTotalSentConsumptionGrants() {
		
		return this.total_sent_consumption_grants;
		
	}

	// ****************************************************************
	
	public synchronized void
		accountSentSupplyingRequests(int tasks) throws GridException {
			
		if (tasks <= 0) {
			throw new GridException("accountSentSupplyingRequests(): "+
									"illegal Tasks count");
		}
		
		this.total_sent_supplying_requests += tasks;
			
	}
	
	public synchronized long getTotalSentSupplyingRequests() {
		
		return this.total_sent_supplying_requests;
		
	}

	/** Increases Consumption Grants count by specified amount.
	 * @param tasks amount of Consumption Grants 
	 * by which to increase current Consumption Grants count
	 * @exception GridException if target parameter is illegal
	 */
	public synchronized void setConsumptionGrantsCount(int tasks)
		throws GridException {
			
		if (tasks <= 0) {
			throw new GridException("setConsumptionGrants(): "+
									"illegal Tasks count");
		}
		
		this.received_consumption_grants = tasks;
		this.total_received_consumption_grants += tasks;
		
	}

	public synchronized int getAndResetReceivedConsumptionGrants()
		throws GridException {
		
		int retval = this.received_consumption_grants;
		this.received_consumption_grants = 0;
		return retval;
		
	}

	public synchronized int getReceviedConsumptionGrants() {
		
		return this.received_consumption_grants;
		
	}

	public synchronized long getTotalReceivedConsumptionGrants() {
		
		return this.total_received_consumption_grants;
		
	}

	public synchronized void accountUsedConsumptionGrants(int tasks)
		throws GridException {
			
		if (tasks <= 0) {
			throw new GridException("accountUsedConsumptionGrantsCount(): "+
									"illegal Tasks count");
		}
		
		this.total_used_consumption_grants += tasks;
		
	}

	public synchronized long getTotalUsedConsumptionGrants() {
		
		return this.total_used_consumption_grants;
	
	}

	// ****************************************************************

	private Peer managed_peer;
	private ExternalPeerHandle profiled_peer;

	// [SUPPLIER ROLE, REQUEST] Negotiation data (Supplying Requests to Peer)
	private int received_supplying_requests;
	private long total_received_supplying_requests;
	// [SUPPLIER ROLE, RESPONSE] Negotiation data (Consumption Grants from Peer)
	private long total_sent_consumption_grants;

	// [CONSUMER ROLE, REQUEST] Negotiation data (Supplying Requests from Peer)
	private long total_sent_supplying_requests;

	// [CONSUMER ROLE, RESPONSE] Negotiation data (Consumption Grants to Peer)
	private int received_consumption_grants;
	private long total_received_consumption_grants;

	// [CONSUMER ROLE, USE] Negotiation data (Consumption Grants used by Peer)
	private long total_used_consumption_grants;

}

