/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SupplyingRequestsContainer.java

package lbg.peer.register.negotiation;

import lbg.common.*;
import lbg.common.exception.*;
import lbg.peer.register.*;
import lbg.peer.service.*;
import java.util.*;

/**
 * @author Cyril Briquet
 */
public class SupplyingRequestsContainer {

	public SupplyingRequestsContainer(Collection<PeerProfile> srs,int s)
		throws GridException {
		
		if (srs == null) {
			throw new GridException("SupplyingRequestsContainer(): "+
									"illegal Requestors");
		}
		if (s <= 0) {
			throw new GridException("SupplyingRequestsContainer(): "+
									"illegal Requestors count");
		}

		this.requestors = new PeerProfile[s];
		this.pending_requestors = new TreeMap<PeerProfile,Integer>();
		this.pending_supplying_requests = new int[s];
		this.granted_supplying_requests = new int[s];
		this.total_supplying_requests = 0;
		
		int i = 0;
		for (PeerProfile peer_profile : srs) {
			this.requestors[i] = peer_profile;
			this.pending_requestors.put(peer_profile,Tank.newInteger(i));
			
			this.pending_supplying_requests[i] =
				peer_profile.getNegotiationProfile().
					getAndResetReceivedSupplyingRequests();
			this.granted_supplying_requests[i] = 0;
			
			this.total_supplying_requests += this.pending_supplying_requests[i];
			i++;
		}
		if (i != s) { // consistency check
			throw new GridException("SupplyingRequestsContainer(): "+
									"corrupted data (i="+i+", s="+s+")");
		}

		this.pending_supplying_requests_count = this.total_supplying_requests;
		
		if (this.total_supplying_requests < 1) {
			throw new GridException("SupplyingRequestsContainer(): "+
									"illegal number of Supplying Requests");
		}
		
	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[SupplyingRequestsContainer]");
		
		return sb.toString();
		
	}
	
	public synchronized int size() {
		
		return this.requestors.length;
		
	}
	
	public synchronized PeerProfile[] requestors() {
	
		return this.requestors;
		
	}
	
	/** Returns an Iterator of the index of each pending requestor.
	 * @return pending requestors indexes
	 */
	public synchronized Iterator<Integer> pendingRequestorsIndex() {
		
		return this.pending_requestors.values().iterator();
		
	}

	public synchronized ExternalPeerHandle requestor(int i)
		throws GridException {
		
		if ((i < 0) || (i >= this.requestors.length)) {
			throw new GridException("requestor(): illegal index");
		}
	
		return this.requestors[i].getHandle();
		
	}

	public synchronized int pendingSupplyingRequests(int i)
		throws GridException {
		
		if ((i < 0) || (i >= this.requestors.length)) {
			throw new GridException("pendingSupplyingRequests(): "+
									"illegal index");
		}
	
		return this.pending_supplying_requests[i];
		
	}
	
	public synchronized int grantedSupplyingRequests(int i)
		throws GridException {
		
		if ((i < 0) || (i >= this.requestors.length)) {
			throw new GridException("grantedSupplyingRequests(): "+
									"illegal index");
		}
	
		return this.granted_supplying_requests[i];
		
	}
	
	/**
	 * If target requestor has no more remaining Supplying Requests,
	 * it is removed from the index of remaining requestors.
	 * @param i index of target requestor
	 */
	public synchronized void grantSupplyingRequests(int i,int count)
		throws GridException {
		
		if ((i < 0) || (i >= this.requestors.length)) {
			throw new GridException("grantSupplyingRequests(): "+
									"illegal index");
		}
		if (count <= 0) {
			throw new GridException("grantSupplyingRequests(): "+
									"illegal requests count");
		}
		if (count > this.pending_supplying_requests[i]) {
			throw new GridException("grantSupplyingRequests(): "+
									"not enough available Supplying Requests");
		}
		
		this.pending_supplying_requests_count -= count;
		this.pending_supplying_requests[i] -= count;
		this.granted_supplying_requests[i] += count;
		
		if (this.pending_supplying_requests[i] == 0) {
			if (this.pending_requestors.remove(this.requestors[i]) ==
				null) {
					throw new GridException("grantSupplyingRequests(): "+
											"corrupted internal data");
			}
		}

	}
	
	public synchronized int totalSupplyingRequests() {
		
		return this.total_supplying_requests;
		
	}

	public synchronized int pendingSupplyingRequests() {
		
		return this.pending_supplying_requests_count;
		
	}
	
	public synchronized int grantedSuppplyingRequests() {
		
		return
			this.total_supplying_requests-this.pending_supplying_requests_count;
		
	}

	private PeerProfile[] requestors;
	private TreeMap<PeerProfile,Integer> pending_requestors;
	private int[] pending_supplying_requests;
	private int[] granted_supplying_requests;
	private int total_supplying_requests;
	private int pending_supplying_requests_count;

}

