/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LivePeerHandle.java

package lbg.peer.service;

import lbg.common.exception.*;
import lbg.common.math.*;
import static lbg.common.net.CommonNetHandling.getLocalHostName;
import lbg.peer.core.qms.*;
import lbg.peer.core.qms.prequeue.*;
import lbg.peer.workflow.*;
import lbg.protocol.rpc.*;
import lbg.resource.service.*;
import lbg.task.*;
import lbg.user.service.*;

import java.util.logging.Logger;

/** A peer handle grouping everything needed by other peers, resources and
 * user
 * @author Xavier Dalem
 */
public class LivePeerHandle implements
	ExternalPeerHandle, InternalPeerHandle, UserPeerHandle {
	// Code will be structured following that order

	static final long serialVersionUID = 2109728264299034428L;

	private final String id;
	private final String hostname;
	private final int peer_port;
	private final int resource_port;
	private final int user_port;
	private int hashcode;
	
	/** Create a handle to a peer
	 * @param id Peer ID
	 * @param hostname Peer hostname
	 * @param peer_port Peer-to-Peer communication port
	 * @param resource_port Resource-to-Peer communication port
	 * @param user_port User-to-Peer communication port
	 */
	public LivePeerHandle(String id, String hostname, int peer_port,
		int resource_port, int user_port) throws GridException {

		if (id == null || hostname == null || peer_port < 1
			|| resource_port < 1 || user_port < 1)
			throw new GridException("LivePeerHandle(): illegal parameters");
		
		this.id = id;
		this.hostname = hostname;
		this.peer_port = peer_port;
		this.resource_port = resource_port;
		this.user_port = user_port;
		this.hashcode = Integer.MIN_VALUE;
		
	}
	
	// *** Common methods ***
	
	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LivePeerHandle ");
		sb.append(id);
		sb.append("] at ");
		sb.append(hostname);
		sb.append(" on ports ");
		sb.append(peer_port);
		sb.append(", ");
		sb.append(resource_port);
		sb.append(", ");
		sb.append(user_port);
		
		return sb.toString();
		
	}

	public int compareTo(PeerHandle other) {
		return this.id.compareToIgnoreCase(other.id());
	}
	
	@Override
	public boolean equals(Object obj) {
		
		if (obj instanceof PeerHandle == false) {
			return false;
		}
		
		return (this.compareTo((PeerHandle)obj) == 0);
		
	}

	@Override
	public int hashCode() {
	
		if (this.hashcode == Integer.MIN_VALUE) {
			this.hashcode = CommonMath.generateHashCode(this.id);
		}
		return this.hashcode;
	
	}

	public String id() {
		return id;
	}
	
	public String hostname() {
		return hostname;
	}
	
	// *** Peer-to-peer methods ***
	
	public int peerPort() {
		return peer_port;
	}
	
	/** Request this peer to supply power for a number of tasks
	 * @param consumer_id The ID of the Peer that asks
	 * @param tasks_count The number of tasks consumers wished the Peer to
	 * handle
	 */
	public void requestSupplying(String consumer_id,int tasks_count)
			throws GridException {
		if (consumer_id == null) {
			throw new GridException("requestSupplying: null consumer");
		}
		/*// get Logger explicitly because this code executes in
		// the Peer using the handle, not in the Peer that created the handle
		Logger PEER_LOGGER = Logger.getLogger("lbg.peer");
		PEER_LOGGER.info("requestSupplying(): sending "+tasks_count+
						 " supplying requests to Peer "+this.hostname);*/
		
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.PEER,
				"requestSupplying", new Class[]
				{ Class.forName("java.lang.String"),
				  Integer.TYPE },
				new Object[] { consumer_id, tasks_count } );
			rc.send(hostname, peer_port, false);
		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}
	
	/** Grant a consumption for a number of tasks
	 * @param supplier_id The ID of the Peer that provides
	 * @param granted_count The number of tasks the provider agreed to run
	 */
	public void grantConsumption(String supplier_id,int granted_count)
			throws GridException {
		if (supplier_id == null) {
			throw new GridException("grantConsumption: null supplier");
		}
		/*// get Logger explicitly because this code executes in
		// the Peer using the handle, not in the Peer that created the handle
		Logger PEER_LOGGER = Logger.getLogger("lbg.peer");
		PEER_LOGGER.info("grantConsumption(): sending "+granted_count+
						 " consumption grants to Peer "+this.hostname);*/
		
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.PEER,
				"grantConsumption", new Class[]
				{ Class.forName("java.lang.String"),
				  Integer.TYPE },
				new Object[] { supplier_id, granted_count } );
			rc.send(hostname, peer_port, false);
		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}
	
	/** Compute a task.
	 * @param consumer_id The task owner
	 * @param t The task to run
	 */
	public ConsumptionTaskID computeSupplyingTask(String consumer_id,Task t)
			throws GridException {
		
		if (consumer_id == null) {
			throw new GridException("computeSupplyingTask: null consumer");
		}
		// get Logger explicitly because this code executes in
		// the Peer using the handle, not in the Peer that created the handle
		Logger PEER_LOGGER = Logger.getLogger("lbg.peer");
		PEER_LOGGER.info("computeSupplyingTask(): submitting 1 Supplying Task "+
						 "to Peer "+this.hostname);
		
		Object ret = null;
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.PEER,
				"computeSupplyingTask", new Class[]
				{ Class.forName("java.lang.String"),
				  Class.forName("lbg.task.Task") },
				new Object[] { consumer_id, t } );
			ret = rc.send(hostname, peer_port, true);
		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
		
		if ((ret != null) && (!(ret instanceof ConsumptionTaskID))) // TODO
			throw new GridException("Invalid return value");
		
		return ((ConsumptionTaskID)ret);
	}
	
	/** Preempts a task from another Peer running on this Peer */
	public void preemptSupplyingTaskOnLocalResource(String consumer_id,
													Task t,byte[] logs)
			throws GridException {
		
		if (consumer_id == null) {
			throw new GridException("preemptSupplyingTaskOnLocalResource: null"
				+ " consumer");
		}
		if (t == null) {
			throw new GridException("preemptSupplyingTaskOnLocalResource: null"
				+ " task");
		}
		// get Logger explicitly because this code executes in
		// the Peer using the handle, not in the Peer that created the handle
		Logger PEER_LOGGER = Logger.getLogger("lbg.peer");
		PEER_LOGGER.info("preemptSupplyingTaskOnLocalResource(): "+
						 "preempt 1 Supplying Task on Resouce ?");
		
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.PEER,
				"preemptSupplyingTaskOnLocalResource", new Class[]
				{ Class.forName("java.lang.String"),
				  Class.forName("lbg.task.Task"),
				  Class.forName("[B") },
				new Object[] { consumer_id, t, logs } );
			rc.send(hostname, peer_port, false);
		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}

	/** Cancel a task from another Peer running on this Peer */
	public void cancelSupplyingTaskOnLocalResource(String consumer_id,Task t)
			throws GridException {
		
		if (consumer_id == null) {
			throw new GridException("cancelSupplyingTaskOnLocalResource: null"
				+ " consumer");
		}
		if (t == null) {
			throw new GridException("cancelSupplyingTaskOnLocalResource: null"
				+ " task");
		}
		// get Logger explicitly because this code executes in
		// the Peer using the handle, not in the Peer that created the handle
		Logger PEER_LOGGER = Logger.getLogger("lbg.peer");
		PEER_LOGGER.info("cancelSupplyingTaskOnLocalResource(): cancelling "+
						 "1 Supplying Task from Peer "+this.hostname);
		
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.PEER,
				"cancelSupplyingTaskOnLocalResource", new Class[]
				{ Class.forName("java.lang.String"),
				  Class.forName("lbg.task.Task") },
				new Object[] { consumer_id, t } );
			rc.send(hostname, peer_port, false);
		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}

	/** Cancel a task from this Peer running on another Peer */
	public void cancelLocalTaskOnSupplyingResource(String supplier_id,Task t)
		throws GridException {
		
		if (supplier_id == null) {
			throw new GridException("cancelLocalTaskOnSupplyingResource: null"
				+ " consumer");
		}
		if (t == null) {
			throw new GridException("cancelLocalTaskOnSupplyingResource: null"
				+ " task");
		}
		// get Logger explicitly because this code executes in
		// the Peer using the handle, not in the Peer that created the handle
		Logger PEER_LOGGER = Logger.getLogger("lbg.peer");
		PEER_LOGGER.info("cancelLocalTaskOnSupplyingResource(): cancelling "+
						 "1 Consumption Task on Peer "+this.hostname);

		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.PEER,
				"cancelLocalTaskOnSupplyingResource", new Class[]
				{ Class.forName("java.lang.String"),
				  Class.forName("lbg.task.Task") },
				new Object[] { supplier_id, t } );
			rc.send(hostname, peer_port, false);
		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}

	public void uploadBySupplyingCompletedTask(String supplier_id,Task t)
			throws GridException {
		
		if (supplier_id == null) {
			throw new GridException("uploadBySupplyingCompletedTask: null"
				+ " consumer");
		}
		if (t == null) {
			throw new GridException("uploadBySupplyingCompletedTask: null"
				+ " task");
		}
		// get Logger explicitly because this code executes in
		// the Peer using the handle, not in the Peer that created the handle
		Logger PEER_LOGGER = Logger.getLogger("lbg.peer");
		PEER_LOGGER.info("uploadBySupplyingCompletedTask(): uploading results "+
						 " of 1 completed Supplying Task to Peer "+this.hostname);
		
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.PEER,
				"uploadBySupplyingCompletedTask", new Class[]
				{ Class.forName("java.lang.String"),
				  Class.forName("lbg.task.Task") },
				new Object[] { supplier_id, t } );
			rc.send(hostname, peer_port, false);
		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}
	
	// *** Resource-to-Peer methods ***
	
	public int resourcePort() {
		return resource_port;
	}
	
	/** Register a resource as part of this Peer
	 * @param rh A handle to the concerned resource
	 */
	public void addResource(ResourceHandle rh,int power) throws GridException {
		if (rh == null)
			throw new GridException("addResource: null resource");
		try {
			RemoteCall rc =
				new RemoteCall(RemoteCall.CallType.PEER, "addResource",
				new Class[] {Class.forName("lbg.resource.service.ResourceHandle"),
				 Integer.TYPE},
				new Object[] {rh,power});
			rc.send(hostname, resource_port, false);
		}
		catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}
	
	/** Remove a resource from this Peer
	 * @param rh A handle to the concerned resource
	 */
	public void removeResource(ResourceHandle rh) throws GridException {
		if (rh == null)
			throw new GridException("removeResource: illegal Resource");
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.PEER,
				"removeResource", new Class[]
				{Class.forName("lbg.resource.service.ResourceHandle")},
				new Object[] {rh} );
			rc.send(hostname, resource_port, false, false, true);
		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}
	
	/** Cancels a task from this Peer running on this Resource 
	 * @param t The task to cancel
	 */
	public void preemptLocalTaskOnLocalResource(Task t,byte[] logs)
		throws GridException {
		
		if (t == null) {
			throw new GridException("preemptLocalTaskOnLocalResource: "+
									"illegal Task");
		}

		// get Logger explicitly because this code executes in
		// the Peer using the handle, not in the Peer that created the handle
		Logger PEER_LOGGER = Logger.getLogger("lbg.peer");
		PEER_LOGGER.info("preemptLocalTaskOnLocalResource(): preempt "+
						 "1 Local Task on Resource ?");

		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.PEER,
				"preemptLocalTaskOnLocalResource", new Class[]
				{Class.forName("lbg.task.Task"), Class.forName("[B")},
				new Object[] {t,logs} );
			rc.send(hostname, resource_port, false);
		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}
	
	
	/** Upload a task completed by a resource to this Peer
	 * @param t The completed task
	 */
	public void uploadLocallyCompletedTask(Task t) throws GridException {

		if (t == null) {
			throw new GridException("uploadLocallyCompletedTask: null task");
		}
		// get Logger explicitly because this code executes in
		// the Peer using the handle, not in the Peer that created the handle
		Logger PEER_LOGGER = Logger.getLogger("lbg.peer");
		PEER_LOGGER.info("uploadLocallyCompletedTask(): uploading results "+
						 "of 1 completed Task to Peer "+this.hostname);
		
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.PEER,
				"uploadLocallyCompletedTask", new Class[]
				{Class.forName("lbg.task.Task")}, new Object[] {t} );
			rc.send(hostname, resource_port, false);
		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}
	
	// *** User-to-peer methods ***
	
	public int userPort() {
		return user_port;
	}
	
	/** Submit a job to peer
	 * @param uh A handle to contact the user
	 * @param jd A description of the job to submit
	 * @return the job ID
	 * @throws GridException if the submitting fails
	 */
	public int submitJob(UserHandle uh, JobDescriptor jd) throws GridException {
		if (uh == null)
			throw new GridException("submitJob: null user");
		if (jd == null)
			throw new GridException("submitJob: null job");
		Object ret = null;
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.PEER, "submitJob",
				new Class[] {Class.forName("lbg.user.service.UserHandle"),
				Class.forName("lbg.peer.workflow.JobDescriptor")}, new Object[]
				{uh, jd} );

			ret = rc.send(hostname, user_port, true,false);

		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
		
		if (ret == null) {
			throw new GridException("Peer "+getLocalHostName()+
									" illegal returned Job id");
		}
		if (ret instanceof Integer == false) {
			throw new GridException("Peer "+getLocalHostName()+
									" illegal non-integer returned Job id");
		}
		
		return ((Integer)ret).intValue();

	}
	
	/** Cancels a Job from the Peer
	 * @param job_id job id
	 * @throws GridException if the cancellation fails
	 */
	public void cancelJob(int job_id) throws GridException {
		
		if (job_id < 0) {
			throw new GridException("cancelJob(): illegal job id");
		}
		
		/*try {*/
			RemoteCall rc =
				new RemoteCall(RemoteCall.CallType.PEER, "cancelJob",
							   new Class[] { Integer.TYPE },
							   new Object[] { job_id });
			rc.send(hostname, user_port, false);
		/*}
		catch (ClassNotFoundException e) {
			throw new GridException("cancelJob(): critical, could not find "+
									"one of the base classes",e);
		}*/
		
	}

	/** Queries Job status from the Peer
	 * @param job_id job id
	 * @throws GridException if the status query
	 */
	public int queryJobStatus(int job_id) throws GridException {
		
		if (job_id < 0) {
			throw new GridException("queryJobStatus(): illegal job id");
		}
		
		Object ret = null;
		/*try {*/
			RemoteCall rc =
				new RemoteCall(RemoteCall.CallType.PEER, "queryJobStatus",
							   new Class[] { Integer.TYPE },
							   new Object[] { job_id });
			ret = rc.send(hostname, user_port, false);
		/*}
		catch (ClassNotFoundException e) {
			throw new GridException("cancelJob(): critical, could not find "+
									"one of the base classes",e);
		}*/
		
		if (ret == null) {
			throw new GridException("Peer "+getLocalHostName()+
									" illegal returned Job status");
		}
		if (ret instanceof Integer == false) {
			throw new GridException("Peer "+getLocalHostName()+
									" illegal non-integer returned Job status");
		}
		
		return ((Integer)ret).intValue();

	}
	
	/** Checks if this peer is alive (i.e. can be reached).
	 * 
	 * @return <tt>true</tt> if the peer is alive.
	 */
	public synchronized boolean isAlive() {
		Object ret = null;
		try {
			RemoteCall rc =
				new RemoteCall(RemoteCall.CallType.PEER,
							   "isAlive",new Class[0], new Object[0]);
				ret = rc.send(hostname, peer_port, true);
		}
		catch (GridException e) {
			if (! (e instanceof GridNetworkException))
				e.printStackTrace();
			return false;
		}
		
		if (ret == null || !(ret instanceof Boolean)
						|| ((Boolean) ret).booleanValue() != true)
			return false;
		
		return true;
	}
	
	// ****************************************************************

}

