/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LivePeerService.java

package lbg.peer.service;

import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.config.*;
import lbg.peer.core.qms.*;
import lbg.peer.core.qms.prequeue.*;
import lbg.peer.register.*;
import lbg.peer.workflow.*;
import lbg.resource.service.*;
import lbg.task.*;
import lbg.user.service.*;

/** PeerService, Live implementation.
 * <br>Most of the code is a direct adaptation of SimulatedPeerService
 * @author Cyril Briquet, Xavier Dalem
 */
public class LivePeerService implements PeerService {

	protected LivePeerService(Peer owner_peer) throws GridException {
		
		if (owner_peer == null) {
			throw new GridException("LivePeerService(): "+
									"illegal owner Peer");
		}
		
		this.owner_peer = owner_peer;
		PeerNetworkConfiguration pnc = this.owner_peer.netConfig();
		this.self_peer_handle =
			new LivePeerHandle(this.owner_peer.id(),pnc.hostname(),
							   pnc.peerPort(),pnc.resourcePort(),pnc.userPort());
		
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LivePeerService ");
		sb.append(this.owner_peer.id());
		sb.append("]");
		
		return sb.toString();
		
	}

	/** Creates a LivePeerHandle */
	public PeerHandle createHandle() throws GridException {
		
		return this.self_peer_handle;
		
	}
	
	// ****************************************************************

	public String id() {
		return owner_peer.id();
	}

	public String hostname() {
		return owner_peer.netConfig().hostname();
	}
	
	// ****************************************************************

	public int peerPort() {
		return owner_peer.netConfig().peerPort();
	}

	public void requestSupplying(String consumer_id,int tasks_count)
			throws GridException {
		if (consumer_id == null)
			throw new GridException("requestSupplying: null consumer");

		this.owner_peer.logger().
			entering("LivePeerService","requestSupplying",
					  new Object[] {consumer_id, tasks_count});
		/*this.owner_peer.logger().info("requestSupplying(): received "+tasks_count+
						 " supplying requests from Peer "+consumer_id);*/

		if (owner_peer.register().checkProfile(consumer_id) == false)
			return;

		NeighborhoodNegotiationProfile grid_negotiation_profile =
			this.owner_peer.register().getNeighborhoodNegotiationProfile();
		grid_negotiation_profile.
			addReceivedSupplyingRequest(consumer_id,tasks_count);

		if (grid_negotiation_profile.couldEvaluateSupplyingRequests()) {
			this.owner_peer.negotiatorController().
				activateSupplyingRequestsEvaluation();
		}

	}
	
	public void grantConsumption(String supplier_id,int granted_count)
			throws GridException {
		if (supplier_id == null)
			throw new GridException("grantConsumption: null supplier");

		this.owner_peer.logger().
			entering("LivePeerService","grantConsumption",
					 new Object[] {supplier_id, granted_count});
		/*this.owner_peer.logger().info("grantConsumption(): received "+granted_count+
						 " consumption grants from Peer "+supplier_id);*/
		
		if (owner_peer.register().checkProfile(supplier_id) == false)
			return;

		NeighborhoodNegotiationProfile grid_negotiation_profile =
			this.owner_peer.register().getNeighborhoodNegotiationProfile();
		grid_negotiation_profile.
			addReceivedConsumptionGrant(supplier_id,granted_count);
		
		if (grid_negotiation_profile.couldEvaluateConsumptionGrants()) {
			this.owner_peer.schedulerController().
				activateConsumptionScheduling();
		}

	}

	public ConsumptionTaskID computeSupplyingTask(String consumer_id,Task t)
		throws GridException {
		
		if (consumer_id == null) {
			throw new GridException("computeSupplyingTask: null consumer");
		}
		if (t == null) {
			throw new GridException("computeSupplyingTask: null task");
		}

		this.owner_peer.logger().
			entering("LivePeerService","computeSupplyingTask",
					 new Object[] {consumer_id, t});
		this.owner_peer.logger().
			info("computeSupplyingTask(): 1 Supplying Task submitted "+
				 "by Peer "+consumer_id);

		if (owner_peer.register().checkProfile(consumer_id) == false) {
			return null;
		}

		ConsumptionTaskID consumption_task_id =
			this.owner_peer.qms().prequeue().transferSupplyingTaskToQMS(t);
		
		if (consumption_task_id != null) {
			this.owner_peer.schedulerController().
				activateSupplyingScheduling();
		}
		
		return consumption_task_id;
		
	}
	
	public void cancelSupplyingTaskOnLocalResource(String consumer_id,Task t)
			throws GridException {
		if (consumer_id == null)
			throw new GridException("cancelSupplyingTaskOnLocalResource: null"
				+ " consumer");
		if (t == null)
			throw new GridException("cancelSupplyingTaskOnLocalResource: null"
				+ " task");

		this.owner_peer.logger().
			entering("LivePeerService","cancelSupplyingTaskOnLocalResource",
					 new Object[] {consumer_id, t});
		this.owner_peer.logger().
			info("cancelSupplyingTaskOnLocalResource(): Peer "+
				 consumer_id+" asks to cancel 1 Supplying Task");

		if (owner_peer.register().checkProfile(consumer_id) == false)
			return;

		synchronized(this.owner_peer.rms()) {
			owner_peer.xms().cancelSupplyingTaskOnLocalResource(t);
		}

		this.owner_peer.schedulerController().
			activateLocalAndSupplyingScheduling();

	}

	public void preemptSupplyingTaskOnLocalResource(String consumer_id,
													Task t,byte[] logs)
			throws GridException {
		if (consumer_id == null)
			throw new GridException("preemptSupplyingTaskOnLocalResource: null"
				+ " consumer");
		if (t == null)
			throw new GridException("preemptSupplyingTaskOnLocalResource: null"
				+ " task");

		this.owner_peer.logger().
			entering("LivePeerService","preemptSupplyingTaskOnLocalResource",
					 new Object[] {consumer_id, t, logs});
		this.owner_peer.logger().
			info("preemptSupplyingTaskOnLocalResource(): Resource ?"+
				 " has preempted 1 Supplying Task");
		
		
		if (owner_peer.register().checkProfile(consumer_id) == false)
			return;

		synchronized(this.owner_peer.rms()) {
			owner_peer.xms().preemptSupplyingTaskOnLocalResource(t,logs);
		}

		this.owner_peer.schedulerController().
			activateLocalAndSupplyingScheduling();

	}

	public void cancelLocalTaskOnSupplyingResource(String supplier_id,Task t)
			throws GridException {
		if (supplier_id == null)
			throw new GridException("cancelLocalTaskOnSupplyingResource: null"
				+ " supplier");
		if (t == null)
			throw new GridException("cancelLocalTaskOnSupplyingResource: null"
				+ " task");

		this.owner_peer.logger().
			entering("LivePeerService","cancelLocalTaskOnSupplyingResource",
					 new Object[] {supplier_id, t});
		this.owner_peer.logger().
			info("cancelLocalTaskOnSupplyingResource(): Peer "+
				  supplier_id+" has cancelled 1 Consumption Task");

		if (owner_peer.register().checkProfile(supplier_id) == false)
			return;

		synchronized(this.owner_peer.rms()) {
			owner_peer.xms().cancelLocalTaskOnSupplyingResource(supplier_id,t);
		}

		this.owner_peer.schedulerController().
			activateLocalAndConsumptionScheduling();

	}

	public void uploadBySupplyingCompletedTask(String supplier_id,Task t)
			throws GridException {
		if (supplier_id == null)
			throw new GridException("uploadBySupplyingCompletedTask: null"
				+ " supplier");
		if (t == null)
			throw new GridException("uploadBySupplyingCompletedTask: null"
				+ " task");

		this.owner_peer.logger().
			entering("LivePeerService","uploadBySupplyingCompletedTask",
					 new Object[] {supplier_id, t});
		this.owner_peer.logger().
			info("uploadBySupplyingCompletedTask(): "+
				 "1 Consumption Task completed by Peer "+supplier_id);
		
		if (owner_peer.register().checkProfile(supplier_id) == false)
			return;

		synchronized(this.owner_peer.rms()) {
			owner_peer.xms().completeLocalTaskOnSupplyingResource(t);
		}

		t.deleteMeatada();

		this.owner_peer.schedulerController().
			activateConsumptionScheduling();

	}

	public boolean isAlive() {
		return true;
	}

	// ****************************************************************

	public int resourcePort() {
		return owner_peer.netConfig().resourcePort();
	}

	public void addResource(ResourceHandle rh,int power) throws GridException {

		if (rh == null) {
			throw new GridException("addResource: null resource");
		}
		if (rh.isAlive() == false) {
			return;
		}
		
		owner_peer.rms().addResource(rh,power);
		
		this.owner_peer.schedulerController().
			activateLocalAndSupplyingScheduling();

	}
	
	public void removeResource(ResourceHandle rh) throws GridException {
		
		if (rh == null) {
			throw new GridException("removeResource: null resource");
		}
		
		owner_peer.rms().removeResource(rh);

		this.owner_peer.schedulerController().activateAllScheduling();
		
	}
	
	public void preemptLocalTaskOnLocalResource(Task t,byte[] logs)
		throws GridException {

		if (t == null) {
			throw new GridException("preemptLocalTaskOnLocalResource: "+
									"illegal Task");
		}
		
		this.owner_peer.logger().
			entering("LivePeerService","preemptLocalTaskOnLocalResource",
					 new Object[] { t, logs});
		this.owner_peer.logger().
			info("preemptLocalTaskOnLocalResource(): Resource ?"+
				 " has preempted 1 Local Task");
		
		synchronized(this.owner_peer.rms()) {
			owner_peer.xms().preemptLocalTaskOnLocalResource(t,logs);
		}

		this.owner_peer.schedulerController().
			activateLocalAndSupplyingScheduling();

	}

	public void uploadLocallyCompletedTask(Task t) throws GridException {

		if (t == null)
			throw new GridException("uploadLocallyCompletedTask: "+
				"null Task");
		
		this.owner_peer.logger().
			entering("LivePeerService","uploadLocallyCompletedTask",
					 new Object[] { t });
		this.owner_peer.logger().
			info("uploadLocallyCompletedTask(): "+
				 "1 Task completed by Resource ?");
		
		TaskOutput t_out = t.output();
		if (t_out == null) {
			throw new GridException("uploadLocallyCompletedTask: "+
				"uncompleted Task");
		}
		
		// Complete Task (Peer viewpoint)
		if (t_out.isLocalTask()) {
			// Fully Local Execution
			synchronized(this.owner_peer.rms()) {
				owner_peer.xms().completeLocalTaskOnLocalResource(t);
			}
		}
		else /* if (t_out.isSupplyingTask()) */ {
			// Execution by Resource Bartering
			// STEP 1: communicate results to self
			synchronized(this.owner_peer.rms()) {
				owner_peer.xms().completeSupplyingTaskOnLocalResource(t);
			}

			// STEP 2: communicate results to consumer Peer
			ExternalPeerHandle consumer_peer =
				owner_peer.register().getHandle(t.consumerID());
			consumer_peer.
				uploadBySupplyingCompletedTask(owner_peer.id(),t);
			
		}
		
		t.deleteMeatada();

		this.owner_peer.schedulerController().
			activateLocalAndSupplyingScheduling();

	}

	// ****************************************************************

	public int userPort() {
		return owner_peer.netConfig().userPort();
	}

	public int submitJob(UserHandle uh, JobDescriptor jd) throws GridException {

		jd.convert(owner_peer.gdm());

		int job_id = owner_peer.qms().prequeue().transferLocalJobToQMS(uh,jd);

		this.owner_peer.schedulerController().activateLocalScheduling();

		return job_id;

	}
	
	public void cancelJob(int job_id) throws GridException {
		
		synchronized(this.owner_peer.rms()) {
			this.owner_peer.qms().dequeueLocalJob(job_id);
		}
		
		this.owner_peer.schedulerController().activateAllScheduling();
		
	}

	public int queryJobStatus(int job_id) throws GridException {
		
		synchronized(this.owner_peer.rms()) {
			return this.owner_peer.qms().queryLocalJobStatus(job_id);
		}
		
	}

	// ****************************************************************

	private final Peer owner_peer;
	private final LivePeerHandle self_peer_handle;

}

