/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SimulatedPeerHandle.java

package lbg.peer.service;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.core.qms.*;
import lbg.peer.core.qms.prequeue.*;
import lbg.peer.workflow.*;
import lbg.resource.service.*;
import lbg.task.*;
import lbg.user.service.*;

/**
 * @author Cyril Briquet
 */
public class SimulatedPeerHandle implements
	ExternalPeerHandle, InternalPeerHandle, UserPeerHandle {

	public SimulatedPeerHandle(String id,SimulatedPeerService peer_service)
		throws GridException {

		if (id == null) {
			throw new GridException("SimulatedPeerHandle(): illegal id");
		}
		
		if (peer_service == null) {
			throw new GridException("SimulatedPeerHandle(): "+
									"illegal PeerService");
		}
		
		this.id = id;
		this.peer_service = peer_service;
		this.hashcode = Integer.MIN_VALUE;
		
	}
	
	public String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[SimulatedPeerHandle ");
		sb.append(this.id);
		sb.append("]");
		
		return sb.toString();

	}
	
	public int compareTo(PeerHandle other) {
	
		return this.id.compareToIgnoreCase(other.id());
		
	}
	
	@Override
	public boolean equals(Object obj) {
		
		if (obj instanceof PeerHandle == false) {
			return false;
		}
		
		return (this.compareTo((PeerHandle)obj) == 0);
		
	}

	@Override
	public int hashCode() {
	
		if (this.hashcode == Integer.MIN_VALUE) {
			this.hashcode = CommonMath.generateHashCode(this.id);
		}
		return this.hashcode;
	
	}

	// ****************************************************************

	public String id() {
		
		return this.id;
		
	}

	public String hostname() {
		
		return this.peer_service.hostname();
		
	}
	
	// ****************************************************************

	public int peerPort() {
		
		return this.peer_service.peerPort();
		
	}

	public void requestSupplying(String consumer_id,int tasks_count)
		throws GridException {
		
		this.peer_service.requestSupplying(consumer_id,tasks_count);
		
	}
	
	public void grantConsumption(String supplier_id,int granted_count)
		throws GridException {
		
		this.peer_service.grantConsumption(supplier_id,granted_count);
		
	}

	public ConsumptionTaskID computeSupplyingTask(String consumer_id,Task t)
		throws GridException {
		
		return this.peer_service.computeSupplyingTask(consumer_id,t);
		
	}
	
	public void cancelSupplyingTaskOnLocalResource(String consumer_id,Task t)
		throws GridException {
		
		this.peer_service.cancelSupplyingTaskOnLocalResource(consumer_id,t);

	}

	public void cancelLocalTaskOnSupplyingResource(String supplier_id,Task t)
		throws GridException {
			
		this.peer_service.cancelLocalTaskOnSupplyingResource(supplier_id,t);
			
	}
			
	public void uploadBySupplyingCompletedTask(String supplier_id,Task t)
		throws GridException {
		
		this.peer_service.uploadBySupplyingCompletedTask(supplier_id,t);
		
	}

	// ****************************************************************

	public int resourcePort() {
		
		return this.peer_service.resourcePort();
		
	}

	public void addResource(ResourceHandle rh,int power) throws GridException {
		
		this.peer_service.addResource(rh,power);
		
	}
	
	public void removeResource(ResourceHandle rh) throws GridException {
		
		this.peer_service.removeResource(rh);
		
	}

	public void preemptLocalTaskOnLocalResource(Task t,byte[] logs)
		throws GridException {
		
		this.peer_service.preemptLocalTaskOnLocalResource(t,logs);

	}
	
	public void preemptSupplyingTaskOnLocalResource(String consumer_id,
													Task t,byte[] logs)
		throws GridException {
		
		this.peer_service.
			preemptSupplyingTaskOnLocalResource(consumer_id,t,logs);

	}

	public void uploadLocallyCompletedTask(Task t)
		throws GridException {
		
		this.peer_service.uploadLocallyCompletedTask(t);
		
	}
	
	public boolean isAlive() {
	
		return this.peer_service.isAlive();
		
	}

	// ****************************************************************

	public int userPort() {
		
		return this.peer_service.userPort();
		
	}

	public int submitJob(UserHandle uh,JobDescriptor jd) throws GridException {
		
		return this.peer_service.submitJob(uh,jd);
		
	}

	public void cancelJob(int job_id) throws GridException {
		
		this.peer_service.cancelJob(job_id);
		
	}

	public int queryJobStatus(int job_id) throws GridException {
		
		return this.peer_service.queryJobStatus(job_id);
		
	}

	// ****************************************************************
	
	static final long serialVersionUID = 3892512113398044387L;

	// ****************************************************************

	private final String id;
	private transient final SimulatedPeerService peer_service;
	private int hashcode;

}
