/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SimulatedPeerService.java

package lbg.peer.service;

import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.core.qms.*;
import lbg.peer.core.qms.prequeue.*;
import lbg.peer.register.*;
import lbg.peer.workflow.*;
import lbg.resource.service.*;
import lbg.task.*;
import lbg.user.service.*;

/**
 * @author Cyril Briquet
 */
public class SimulatedPeerService implements PeerService {

	protected SimulatedPeerService(Peer owner_peer) throws GridException {
	
		if (owner_peer == null) {
			throw new GridException("SimulatedPeerService(): "+
									"illegal owner Peer");
		}
		
		this.owner_peer = owner_peer;
		
	}

	public String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[SimulatedPeerService ");
		sb.append(this.owner_peer.id());
		sb.append("]");
		
		return sb.toString();

	}

	public PeerHandle createHandle() throws GridException {
		
		return new SimulatedPeerHandle(this.owner_peer.id(),this);
		
	}
	
	// ****************************************************************

	public String id() {

		return this.owner_peer.id();
		
	}
	
	public String hostname() {
		
		return this.owner_peer.netConfig().hostname();
		
	}
	
	// ****************************************************************

	public int peerPort() {
		
		return this.owner_peer.netConfig().peerPort();

	}

	public void requestSupplying(String consumer_id,int tasks_count)
		throws GridException {

		if (this.owner_peer.register().checkProfile(consumer_id) == false) {
			return;
		}

		NeighborhoodNegotiationProfile grid_negotiation_profile =
			this.owner_peer.register().getNeighborhoodNegotiationProfile();
		grid_negotiation_profile.
			addReceivedSupplyingRequest(consumer_id,tasks_count);
		
		if (grid_negotiation_profile.couldEvaluateSupplyingRequests()) {
			this.owner_peer.negotiatorController().
				activateSupplyingRequestsEvaluation();
		}

	}
	
	public void grantConsumption(String supplier_id,int granted_count)
		throws GridException {
		
		if (this.owner_peer.register().checkProfile(supplier_id) == false) {
			return;
		}

		NeighborhoodNegotiationProfile grid_negotiation_profile =
			this.owner_peer.register().getNeighborhoodNegotiationProfile();
		grid_negotiation_profile.
			addReceivedConsumptionGrant(supplier_id,granted_count);
		
		if (grid_negotiation_profile.couldEvaluateConsumptionGrants()) {
			this.owner_peer.schedulerController().
				activateConsumptionScheduling();
		}
		
	}

	public ConsumptionTaskID computeSupplyingTask(String consumer_id,Task t)
		throws GridException {
		
		if (this.owner_peer.register().checkProfile(consumer_id) == false) {
			return null;
		}

		ConsumptionTaskID accepted = null;
		synchronized(this.owner_peer.rms()) {
			accepted = this.owner_peer.qms().prequeue().
				transferSupplyingTaskToQMS(t);
		}
		
		if (accepted != null) {
			this.owner_peer.schedulerController().activateSupplyingScheduling();
		}
		
		return accepted;
		
	}
	
	public void cancelSupplyingTaskOnLocalResource(String consumer_id,Task t)
		throws GridException {
		
		if (this.owner_peer.register().checkProfile(consumer_id) == false) {
			return;
		}
		
		synchronized(this.owner_peer.rms()) {
			this.owner_peer.xms().cancelSupplyingTaskOnLocalResource(t);
		}

		this.owner_peer.schedulerController().
			activateLocalAndSupplyingScheduling();

	}

	public void cancelLocalTaskOnSupplyingResource(String supplier_id,Task t)
		throws GridException {
		
		if (this.owner_peer.register().checkProfile(supplier_id) == false) {
			return;
		}
		
		synchronized(this.owner_peer.rms()) {
			this.owner_peer.xms().cancelLocalTaskOnSupplyingResource(supplier_id,t);
		}

		this.owner_peer.schedulerController().
			activateLocalAndConsumptionScheduling();
		
	}

	public void uploadBySupplyingCompletedTask(String supplier_id,Task t)
		throws GridException {
		
		if (this.owner_peer.register().checkProfile(supplier_id) == false) {
			return;
		}

		synchronized(this.owner_peer.rms()) {
			this.owner_peer.xms().completeLocalTaskOnSupplyingResource(t);
		}

		this.owner_peer.schedulerController().activateConsumptionScheduling();

	}

	// ****************************************************************

	public int resourcePort() {
		
		return this.owner_peer.netConfig().resourcePort();

	}

	public void addResource(ResourceHandle rh,int power) throws GridException {

		if (rh == null) {
			throw new GridException("addResource(): illegal Resource");
		}
		if (rh.isAlive() == false) {
			return;
		}

		this.owner_peer.rms().addResource(rh,power);

		this.owner_peer.schedulerController().
			activateLocalAndSupplyingScheduling();
		
	}
	
	public void removeResource(ResourceHandle rh) throws GridException {
		
		this.owner_peer.rms().removeResource(rh);

		this.owner_peer.schedulerController().activateAllScheduling();
		
	}

	public void preemptLocalTaskOnLocalResource(Task t,byte[] logs)
		throws GridException {
		
		synchronized(this.owner_peer.rms()) {
			this.owner_peer.xms().preemptLocalTaskOnLocalResource(t,logs);
		}

		this.owner_peer.schedulerController().
			activateLocalAndSupplyingScheduling();
	
	}

	public void preemptSupplyingTaskOnLocalResource(String consumer_id,
													Task t,byte[] logs)
		throws GridException {
		
		if (this.owner_peer.register().checkProfile(consumer_id) == false) {
			return;
		}
		
		synchronized(this.owner_peer.rms()) {
			this.owner_peer.xms().preemptSupplyingTaskOnLocalResource(t,logs);
		}
		
		this.owner_peer.schedulerController().
			activateLocalAndSupplyingScheduling();

	}

	public void uploadLocallyCompletedTask(Task t) throws GridException {
		
		if (t == null) {
			throw new GridException("uploadLocallyCompletedTask(): "+
									"illegal Task");
		}
		TaskOutput t_out = t.output();
		if (t_out == null) {
			throw new GridException("uploadLocallyCompletedTask(): "+
									"uncompleted Task");
		}
		
		// Complete Task (Peer viewpoint)
		if (t_out.isLocalTask()) {
			// Fully Local Execution
			synchronized(this.owner_peer.rms()) {
				this.owner_peer.xms().completeLocalTaskOnLocalResource(t);
			}
		}
		else /* if (t_out.isSupplyingTask()) */ {
			// Execution by Resource Bartering
			// STEP 1: communicate results to self
			synchronized(this.owner_peer.rms()) {
				this.owner_peer.xms().completeSupplyingTaskOnLocalResource(t);
			}

			// STEP 2: communicate results to consumer Peer
			ExternalPeerHandle consumer_peer =
				this.owner_peer.register().getHandle(t.consumerID());
			consumer_peer.
				uploadBySupplyingCompletedTask(this.owner_peer.id(),t);
			
		}
		
		this.owner_peer.schedulerController().
			activateLocalAndSupplyingScheduling();

	}
	
	public boolean isAlive() {
	
		return true;
		
	}

	// ****************************************************************

	public int userPort() {
		
		return this.owner_peer.netConfig().userPort();

	}

	public int submitJob(UserHandle uh,JobDescriptor jd) throws GridException {
	
		int job_id = -1;
		synchronized(this.owner_peer.rms()) {
			job_id = this.owner_peer.qms().prequeue().
				transferLocalJobToQMS(uh,jd);
		}

		this.owner_peer.schedulerController().activateLocalScheduling();
		
		return job_id;
		
	}

	public void cancelJob(int job_id) throws GridException {
		
		synchronized(this.owner_peer.rms()) {
			this.owner_peer.qms().dequeueLocalJob(job_id);
		}
		
		this.owner_peer.schedulerController().activateAllScheduling();
		
	}

	public int queryJobStatus(int job_id) throws GridException {
		
		synchronized(this.owner_peer.rms()) {
			return this.owner_peer.qms().queryLocalJobStatus(job_id);
		}
		
	}

	// ****************************************************************

	private final Peer owner_peer;

}

