/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// StrategyDescriptor.java

package lbg.peer.strategy;

import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.config.*;
import lbg.peer.core.negotiator.*;
import lbg.peer.core.scheduler.*;
import lbg.peer.strategy.accounting.*;

/**
 * @author Cyril Briquet
 */
public class StrategyDescriptor {

	public StrategyDescriptor(Peer managed_peer,
							  PeerStrategyConfiguration peer_strategy_config)
		throws GridException {
		
		if (managed_peer == null) {
			throw new GridException("StrategyDescriptor(): illegal Peer");
		}
		if (peer_strategy_config == null) {
			throw new GridException("StrategyDescriptor(): illegal "+
									"Peer Strategy configuration");
		}
		
		this.managed_peer = managed_peer;
		this.default_accountant =
			peer_strategy_config.PEER_DEFAULT_ACCOUNTANT();
		this.target_negotiation_supplying_policy =
			peer_strategy_config.PEER_NEGOTIATION_SUPPLYING_POLICY();
		this.target_negotiation_consumption_policy =
			peer_strategy_config.PEER_NEGOTIATION_CONSUMPTION_POLICY();
		this.target_scheduling_local_policy =
			peer_strategy_config.PEER_SCHEDULING_LOCAL_POLICY();
		this.target_scheduling_running_preemption_policy =
			peer_strategy_config.PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY();
		this.target_scheduling_waiting_preemption_policy =
			peer_strategy_config.PEER_PREEMPTION_SUPPLYING_WAITING_POLICY();
		this.target_supplying_filtering_policy =
			peer_strategy_config.PEER_FILTERING_SUPPLYING_POLICY();
		this.target_scheduling_supplying_policy =
			peer_strategy_config.PEER_SCHEDULING_SUPPLYING_POLICY();
		this.target_scheduling_consumption_policy =
			peer_strategy_config.PEER_SCHEDULING_CONSUMPTION_POLICY();
	
	}
	
	public String toString() {

		StringBuilder sb = new StringBuilder();
		
		sb.append("[StrategyDescriptor ");
		sb.append(this.managed_peer.id());
		sb.append("]");
		sb.append("\n  ->  ");
		sb.append(this.default_accountant);
		sb.append("\n  ->  ");
		sb.append(this.target_negotiation_supplying_policy);
		sb.append("\n  ->  ");
		sb.append(this.target_negotiation_consumption_policy);
		sb.append("\n  ->  ");
		sb.append(this.target_scheduling_local_policy);
		sb.append("\n  ->  ");
		sb.append(this.target_scheduling_running_preemption_policy);
		sb.append("\n  ->  ");
		sb.append(this.target_scheduling_waiting_preemption_policy);
		sb.append("\n  ->  ");
		sb.append(this.target_supplying_filtering_policy);
		sb.append("\n  ->  ");
		sb.append(this.target_scheduling_supplying_policy);
		sb.append("\n  ->  ");
		sb.append(this.target_scheduling_consumption_policy);
		
		return sb.toString();

	}

	// ****************************************************************

	public Negotiator createNegotiator() throws GridException {
		
		return new Negotiator(this.managed_peer,
							  this.target_negotiation_supplying_policy,
							  this.target_negotiation_consumption_policy);

	}

	public Scheduler createScheduler(Negotiator negotiator)
		throws GridException {

		return new Scheduler(this.managed_peer,negotiator,
							 this.target_scheduling_local_policy,
							 this.target_scheduling_running_preemption_policy,
							 this.target_scheduling_waiting_preemption_policy,
							 this.target_supplying_filtering_policy,
							 this.target_scheduling_supplying_policy,
							 this.target_scheduling_consumption_policy);
		
	}

	public BarteringAccountantPolicy createDefaultAccountant()
			throws GridException {
		
		BarteringAccountantPolicy accountant = null;
		
		try {
			accountant =
				BarteringAccountantType.valueOf(this.default_accountant).
					createAccountant(this.managed_peer);
		}
		catch (IllegalArgumentException e) {
			throw new GridException("createDefaultAccountant(): "+
									"illegal Default Accountant "+
									this.default_accountant,e);
		}

		return accountant;

	}

	// ****************************************************************

	private Peer managed_peer;
	private String default_accountant;
	private String target_negotiation_supplying_policy;
	private String target_negotiation_consumption_policy;
	private String target_scheduling_local_policy;
	private String target_scheduling_running_preemption_policy;
	private String target_scheduling_waiting_preemption_policy;
	private String target_supplying_filtering_policy;
	private String target_scheduling_supplying_policy;
	private String target_scheduling_consumption_policy;

}

