/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// OGPerfectAccountant.java

package lbg.peer.strategy.accounting;

import lbg.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.exception.*;
import lbg.peer.*;
import lbg.task.*;

/**
 * <p>Note: this perfect accounting implementation (in the sense of OurGrid)
 * is correct only in stable environments,
 * i.e. supplied power is constant over time (when supplied, of course).
 * <p>Implementation note:
 * this implementation relies on Task output information.
 * @author Cyril Briquet
 */
public class OGPerfectAccountant implements BarteringAccountantPolicy {

	/** Constructs a new OGPerfectAccountant (SIM_MODE only).
	 * @throws GridException if a parameter is illegal
	 */
	public OGPerfectAccountant(Peer managed_peer)
		throws GridException {

		if (managed_peer == null) {
			throw new GridException("OGPerfectAccountant(): illegal Peer");
		}

		if (LightweightBarteringGrid.getMode() != SIM_MODE) {
			throw new GridException("OGPerfectAccountant(): unsupported mode");
		}

		this.managed_peer = managed_peer;
		
	}

	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[OGPerfectAccountant ");
		sb.append(this.managed_peer.id());
		sb.append("]");
		
		return sb.toString();
		
	}
	
	/** BarteringAccountant implementation.
	 */
	public synchronized double evaluateSupplying(Task t,long xstart_time)
		throws GridException {

		if (t == null) {
			throw new GridException("evaluateSupplying(): illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("evaluateSupplying(): "+
									"illegal execution start time");
		}
		
		double supplied_favor =
			(double)(t.output().runtime() * (long)t.output().resourcePower());

		return supplied_favor;
		
	}

	/** BarteringAccountant implementation.
	 */
	public synchronized double evaluateConsumption(Task t,long xstart_time)
		throws GridException {
		
		if (t == null) {
			throw new GridException("evaluateConsumption(): illegal Task");
		}
		if (xstart_time < 0) {
			throw new GridException("evaluateConsumption(): "+
									"illegal execution start time");
		}

		double consumed_favor =
			(double)(t.output().runtime() * (long)t.output().resourcePower());
		 
		assert(consumed_favor >= 0.0);

		return consumed_favor;
		
	}
	
	private Peer managed_peer;

}

