/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// FavorsSupplyingNegotiation.java

package lbg.peer.strategy.negotiation.supplying;

import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.register.*;
import lbg.peer.register.negotiation.*;
import lbg.peer.service.*;
import lbg.peer.strategy.negotiation.*;
import lbg.peer.strategy.ranking.*;

/** Favors-ranked distribution of Consumption Grants.
 * @author Cyril Briquet
 */
public class FavorsSupplyingNegotiation
	implements SupplyingNegotiationPolicy {

	public FavorsSupplyingNegotiation(Peer managed_peer) throws GridException {
	
		if (managed_peer == null) {
			throw new GridException("FavorsSupplyingNegotiation(): "+
									"illegal Peer");
		}

		this.managed_peer = managed_peer;
		
	}

	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[FavorsSupplyingNegotiation]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** Implementation of Resource Supplying follows Andrade's
	 * initial Network of Favors paper:
     * available capacity is distributed first to Peers
	 * with highest Favors balance.
	 */
	public synchronized void evaluateSupplyingRequests()
		throws GridException {

		// Extract Peers which sent Supplying requests
		SupplyingRequestsContainer s_r =
			this.managed_peer.register().getNeighborhoodNegotiationProfile().
				getReceivedSupplyingRequests();
		if (s_r == null) return;
		
		// Compute available Resources
		int available_res = this.managed_peer.rms().countIdleResources();
		
		// Distribute available capacity to Peers with highest favors balance
		Ranking r = new FavorsSupplyingRanking(s_r.requestors());
		if (r.getRankedPeerCount() < 1) {
			throw new GridException("filterOutSupplyingTask(): "+
									"inconsistent FavorsSupplyingRanking");
		}
		
		int[] sorted_remaining_requestors_index = r.peersRanking();
		assert(sorted_remaining_requestors_index != null);
		int ranking_i = 0;
		
		while ((available_res > 0) && (s_r.pendingSupplyingRequests() > 0)) {

			// Select a Supplying Request
			int i = sorted_remaining_requestors_index[ranking_i];
			int count = s_r.pendingSupplyingRequests(i);
			if (count > available_res) count = available_res;
			
			// Use Supplying Requests (update Supplying Requests Container)
			s_r.grantSupplyingRequests(i,count);
			available_res -= count;
			ranking_i++;				
		}

		// Communicate Supplying decisions to Peers
		PeerProfile[] requestors = s_r.requestors();
		for (int i = 0 ; i < requestors.length ; i++) {
			int count = s_r.grantedSupplyingRequests(i);
			if (count > 0) {
				// Communicate Supplying decision
				ExternalPeerHandle peer_handle = requestors[i].getHandle();
				peer_handle.grantConsumption(this.managed_peer.id(),count);

				// Update stats
				this.managed_peer.register().getNeighborhoodNegotiationProfile().
					accountSentConsumptionGrants(peer_handle.id(),count);
			}
		}
		
	}

	// ****************************************************************

	private Peer managed_peer;

}

