/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// UnlimitedSupplyingNegotiation.java

package lbg.peer.strategy.negotiation.supplying;

import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.register.*;
import lbg.peer.register.negotiation.*;
import lbg.peer.service.*;
import lbg.peer.strategy.negotiation.*;

/** Unlimited distribution of Consumption Grants.
 * @author Cyril Briquet
 */
public class UnlimitedSupplyingNegotiation
	implements SupplyingNegotiationPolicy {

	public UnlimitedSupplyingNegotiation(Peer managed_peer)
		throws GridException {
	
		if (managed_peer == null) {
			throw new GridException("UnlimitedSupplyingNegotiation(): "+
									"illegal Peer");
		}

		this.managed_peer = managed_peer;
		
	}

	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[UnlimitedSupplyingNegotiation]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** Implementation of Resource Supplying follows Andrade's
	 * initial Network of Favors paper (p. 6, section 4.2):
	 * Supplying is granted to all requesting Peers.
	 */
	 public synchronized void evaluateSupplyingRequests()
		throws GridException {

		// Extract Peers which sent Supplying requests
		SupplyingRequestsContainer s_r =
			this.managed_peer.register().getNeighborhoodNegotiationProfile().
				getReceivedSupplyingRequests();
		if (s_r == null) return;
		
		// Randomly distribute available capacity between Peers
		for (int i = 0 ; i < s_r.size() ; i++) {
			int count = s_r.pendingSupplyingRequests(i);
			if (count > 0) {
				// Use Supplying Request (update Supplying Requests Container)
				s_r.grantSupplyingRequests(i,count);
			}
		}

		// Communicate Supplying decisions to Peers
		PeerProfile[] requestors = s_r.requestors();
		for (int i = 0 ; i < requestors.length ; i++) {
			int count = s_r.grantedSupplyingRequests(i);
			if (count > 0) {
				// Communicate Supplying decision
				ExternalPeerHandle peer_handle = requestors[i].getHandle();
				peer_handle.grantConsumption(this.managed_peer.id(),count);

				// Update stats
				this.managed_peer.register().getNeighborhoodNegotiationProfile().
					accountSentConsumptionGrants(peer_handle.id(),count);
			}
		}
		
	}

	private Peer managed_peer;

}

