/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ConservativeTimeStrideConsumptionRanking.java

package lbg.peer.strategy.ranking;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.common.math.*;
import lbg.peer.register.*;
import lbg.peer.register.bartering.*;
import lbg.peer.register.learning.consumption.*;
import lbg.peer.register.negotiation.*;
import lbg.task.*;

import java.io.Serializable;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

/** ConservativeTimeStrideConsumptionRanking ranks grantors
 * (i.e. potential suppliers who sent consumption grants)
 * by using for each grantor its worst normalized rank among
 * the MTBC and MCoS normalized ranks.
 * <p>Note: value normalization is performed with respect to the maximum value
 * among the set of target grantors.
 * @author Cyril Briquet
 */
public class ConservativeTimeStrideConsumptionRanking extends CommonRanking {

	public ConservativeTimeStrideConsumptionRanking(
			ConsumptionGrantsContainer c_g,
			JobConsumptionOutcomeTracker jcot,
			boolean do_blacklist)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("ConservativeTimeStride"+
									"ConsumptionRanking(): "+
									"illegal Consumption Grants Container");
		}
		if (jcot == null) {
			throw new GridException("ConservativeTimeStride"+
									"ConsumptionRanking(): "+
									"illegal Job Consumption outcome tracker");
		}
		
		final Ranking r_mcos =
			new MCoSConsumptionRanking(c_g,jcot,do_blacklist);
		final Ranking r_mtbc =
			new MTBCConsumptionRanking(c_g,jcot,do_blacklist);

		
		final int[] mcos_ranking = r_mcos.peersRanking();
		if (mcos_ranking == null) {
			this.ranked_peers = 0;
			this.unranked_peers = c_g.pendingConsumptionGrants();
			this.blacklisted_peers = r_mcos.getBlacklistedPeerCount();
			this.peers_ranking = null;
			return;
		}
		Map<Integer,Float> mcos_grantor_to_rank_map =
			new TreeMap<Integer,Float>();
		for (int i = 0 ; i < mcos_ranking.length ; ++i) {
			final int grantor_idx = mcos_ranking[i];
			final float mcos_normalized_rank =
				((float)i) / ((float)r_mcos.getRankedPeerCount());
			if (mcos_grantor_to_rank_map.
				put(grantor_idx,mcos_normalized_rank) != null) {
				throw new GridException("ConservativeTimeStride"+
										"ConsumptionRanking(): corrupted data");
			}
		}

		final int[] mtbc_ranking = r_mtbc.peersRanking();
		if (mtbc_ranking == null) {
			this.ranked_peers = 0;
			this.unranked_peers = c_g.pendingConsumptionGrants();
			this.blacklisted_peers = r_mtbc.getBlacklistedPeerCount();
			this.peers_ranking = null;
			return;
		}
		Map<Float,Set<Integer>> conservative_grantor_to_rank_map =
			new TreeMap<Float,Set<Integer>>();
		int tmp_ranked_peers = 0;
		int tmp_blacklisted_peers = 0;
		for (int i = 0 ; i < mtbc_ranking.length ; ++i) {
			final int grantor_idx = mtbc_ranking[i];
			
			PeerProfile grantor_profile = c_g.grantorProfile(grantor_idx);
			PeerBarteringProfile peer_bp =
				grantor_profile.getBarteringProfile();
			if (do_blacklist) {
				boolean should_blacklist_job_level =
					jcot.shouldBlackListPeer(grantor_profile.id());
				if (should_blacklist_job_level) {
					++tmp_blacklisted_peers;
					continue;
				}
				// no Peer-level blacklisting
			}
			
			final float mtbc_normalized_rank =
				((float)i) / ((float)r_mtbc.getRankedPeerCount());
			Float dummy = mcos_grantor_to_rank_map.get(grantor_idx);
			if (dummy == null) {
				dummy = zero;
			}
			final float mcos_normalized_rank = dummy;
			final float rank =
				Math.max(mtbc_normalized_rank,mcos_normalized_rank);
			Set<Integer> grantors_with_same_rank =
				conservative_grantor_to_rank_map.get(rank);
			if (grantors_with_same_rank == null) {
				grantors_with_same_rank = new TreeSet<Integer>();
				if (conservative_grantor_to_rank_map.
					put(rank,grantors_with_same_rank) != null) {
					throw new GridException("ConservativeTimeStride"+
											"ConsumptionRanking(): "+
											"corrupted data");
				}
			}
			if (grantors_with_same_rank.add(grantor_idx) == false) {
				throw new GridException("ConservativeTimeStride"+
										"ConsumptionRanking(): corrupted data");
			}
			++tmp_ranked_peers;
		}
		
		this.ranked_peers = tmp_ranked_peers;
		this.unranked_peers =
			c_g.pendingGrantorsIndex().size()-this.ranked_peers;
		this.blacklisted_peers = tmp_blacklisted_peers;
		
		if (tmp_ranked_peers == 0) {
			this.peers_ranking = null;
			return;
		}

		// Fill peers_ranking with Conservative-Time-Stride-ranked Grantors
		this.peers_ranking = new int[ranked_peers];
		int i = 0;
		for (Set<Integer> grantors_with_same_rank :
			conservative_grantor_to_rank_map.values()) {
			for (Integer grantor_idx : grantors_with_same_rank) {
				this.peers_ranking[i++] = grantor_idx;
			}
		}
		if (i != this.ranked_peers) {
			throw new GridException("ConservativeTimeStride"+
									"ConsumptionRanking(): corrupted data");
		}
		
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[CommonRanking]",
							   "[ConservativeTimeStrideConsumptionRanking]"));

		return sb.toString();
		
	}

	// ****************************************************************

	/** Ranking implementation.
	 * @return Peers ranking, or null if no Peers can be ranked
	 */
	public int[] peersRanking() {
		
		return this.peers_ranking;
		
	}

	/** Ranking implementation.
	 */
	public int getRankedPeerCount() {
		
		return this.ranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getUnrankedPeerCount() {
		
		return this.unranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getBlacklistedPeerCount() {
		
		return this.blacklisted_peers;
		
	}

	// ****************************************************************

	private final int[] peers_ranking;
	private final int ranked_peers;
	private final int unranked_peers;
	private final int blacklisted_peers;
	
	private static final Float zero = new Float(0);

}

