/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// FavorsConsumptionRanking.java

package lbg.peer.strategy.ranking;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.peer.register.*;
import lbg.peer.register.learning.consumption.*;
import lbg.peer.register.negotiation.*;
import java.io.Serializable;
import java.util.*;

/** FavorsConsumptionRanking ranks Grantors
 * (i.e. potential Supplier Peers who sent Consumption Grants)
 * by decreasing supplied Favors.
 * @author Cyril Briquet
 */
public class FavorsConsumptionRanking extends CommonRanking {

	/** Constructs a new FavorsConsumptionRanking.
	 * @param c_g Consumption Grants container
	 * @exception GridException if a parameter is illegal
	 */
	public FavorsConsumptionRanking(ConsumptionGrantsContainer c_g,
					  				JobConsumptionOutcomeTracker jcot,
									boolean do_blacklist)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("FavorsConsumptionRanking(): "+
									"illegal Consumption Grants Container");
		}
		if (jcot == null) {
			throw new GridException("FavorsConsumptionRanking(): "+
									"illegal Job Consumption outcome tracker");
		}
		
		// Obtain pending Grantors profiles set
		Collection<Integer> pending_grantors_index = c_g.pendingGrantorsIndex();

		TreeSet<FavorsKey> ts_peers_ranking =
			new TreeSet<FavorsKey>(default_cmp);
		
		// Rank Grantors by decreasing consumed Favors
		int tmp_ranked_peers = 0;
		int tmp_blacklisted_peers = 0;
		for (Integer grantor_idx : pending_grantors_index) {
			PeerProfile grantor_profile = c_g.grantorProfile(grantor_idx);
			double favors =
				grantor_profile.getBarteringProfile().getTotalConsumedFavors();
			
			if (do_blacklist) {
				boolean should_blacklist_job_level =
					jcot.shouldBlackListPeer(grantor_profile.id());
				if (should_blacklist_job_level) {
					++tmp_blacklisted_peers;
					continue;
				}
				// no Peer-level blacklisting
			}

			FavorsKey key = new FavorsKey(favors,grantor_idx);
			ts_peers_ranking.add(key);
			tmp_ranked_peers++;
		}

		this.ranked_peers = tmp_ranked_peers;
		this.unranked_peers =
			c_g.pendingGrantorsIndex().size()-this.ranked_peers;
		this.blacklisted_peers = tmp_blacklisted_peers;

		if (tmp_ranked_peers == 0) {
			this.peers_ranking = null;
			return;
		}
		
		this.peers_ranking = new int[this.ranked_peers];
		int i = 0;
		for (FavorsKey key : ts_peers_ranking) {
			this.peers_ranking[i] = key.i;
			++i;
		}
		if (i != this.ranked_peers) {
			throw new GridException("FavorsConsumptionRanking(): "+
									"corrupted data");
		}
		
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[CommonRanking]",
							   "[FavorsConsumptionRanking]"));

		return sb.toString();
		
	}

	// ****************************************************************

	/** Ranking implementation.
	 * @return Peers ranking, or null if no Peers can be ranked
	 */
	public int[] peersRanking() {
		
		return this.peers_ranking;
		
	}

	/** Ranking implementation.
	 */
	public int getRankedPeerCount() {
		
		return this.ranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getUnrankedPeerCount() {
		
		return this.unranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getBlacklistedPeerCount() {
		
		return this.blacklisted_peers;
		
	}
	
	// ****************************************************************

	private final int[] peers_ranking;
	private final int ranked_peers;
	private final int unranked_peers;
	private final int blacklisted_peers;

	private static final DecreasingFavorsComparator default_cmp =
		new DecreasingFavorsComparator();

	// ****************************************************************

	/** Helper class of FavorsConsumptionRanking.
	 */
	private static class FavorsKey {
		
		public FavorsKey(double favors,int i) //,PeerProfile peer_profile)
			throws GridException {
			
			/*
			if (peer_profile == null) {
				throw new GridException("FavorsKey(): "+
										"illegal Peer profile");
			}
			*/
			
			this.favors = favors;
			this.i = i;
			//this.peer = peer_profile;
			
		}
			
		/** Favors associated with target Peer (peers[i]).
		 */
		public final double favors;
		
		/** Index of target Peer (peers[i]).
		 */
		public final int i;
		
		/* Target Peer (peers[i]).
		 *
		public final PeerProfile peer;*/
		
	}
	
	/** Helper class of FavorsConsumptionRanking.
	 * Comparator of FavorsBalanceKeys (by decreasing order of Favors Balance).
	 */
	private static class DecreasingFavorsComparator
		implements Comparator<FavorsKey>, Serializable {
		
		public DecreasingFavorsComparator() {
			
		}
		
		public int compare(FavorsKey k1,FavorsKey k2) {
			
			if (k1.favors < k2.favors) {
				return +1;
			}
			else if (k1.favors > k2.favors) {
				return -1;
			}
			else /* if (k1.favors == k2.favors) */ {
				return (k1.i == k2.i) ? 0 : ((k1.i < k2.i) ? +1 : -1);
			}
			
		}
		
		static final long serialVersionUID = -8111277913627670039L;
		
	}

	// ****************************************************************

}

