/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// FavorsSupplyingRanking.java

package lbg.peer.strategy.ranking;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.peer.register.*;

import java.io.Serializable;
import java.util.Comparator;
import java.util.TreeSet;

/** FavorsSupplyingRanking ranks actual Consumers
 * (i.e. some of their Supplying Tasks have been accepted and queued)
 * by decreasing Favors balance.
 * @author Cyril Briquet
 */
public class FavorsSupplyingRanking extends CommonRanking {

	public FavorsSupplyingRanking(PeerProfile[] peers)
		throws GridException {
		
		if (peers == null) {
			throw new GridException("FavorsSupplyingRanking(): illegal Peers");
		}

		TreeSet<FavorsBalanceKey> ts_peers_ranking =
			new TreeSet<FavorsBalanceKey>(default_cmp);

		for (int i = 0 ; i < peers.length ; i++) {
			// Hypothesis: (k1.i != k2.i) always hold for data actually inserted
			double fb = peers[i].getBarteringProfile().getFavorsBalance();
			FavorsBalanceKey key = new FavorsBalanceKey(fb,i,peers[i]);
			ts_peers_ranking.add(key);
		}

		this.ranked_peers = peers.length;
		this.unranked_peers = 0;
		this.blacklisted_peers = 0;
		this.peers_ranking = new int[peers.length];
		this.ranked_peers_array = new PeerProfile[peers.length];
		int i = 0;
		for (FavorsBalanceKey key : ts_peers_ranking) {
			this.peers_ranking[i] = key.i;
			this.ranked_peers_array[i] = key.peer;
			++i;
		}

	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[CommonRanking]",
							   "[FavorsSupplyingRanking]"));

		return sb.toString();
		
	}

	// ****************************************************************

	/** Ranking implementation.
	 * @return Peers ranking, or null if no Peers can be ranked
	 */
	public int[] peersRanking() {
		
		return this.peers_ranking;
		
	}

    /**	
	 * @return Ranked Peers, or null if no Peers can be ranked
	 */
	public PeerProfile[] getRankedPeers() {
		
		return this.ranked_peers_array;
		
	}

	/** Ranking implementation.
	 */
	public int getRankedPeerCount() {
		
		return this.ranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getUnrankedPeerCount() {
		
		return this.unranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getBlacklistedPeerCount() {
		
		return this.blacklisted_peers;
		
	}

	// ****************************************************************

	private final int[] peers_ranking;
	private final PeerProfile[] ranked_peers_array;
	private final int ranked_peers;
	private final int unranked_peers;
	private final int blacklisted_peers;

	private static final DecreasingFBComparator default_cmp =
		new DecreasingFBComparator();

	// ****************************************************************

	/** Helper class of FavorsSupplyingRanking.
	 */
	private static class FavorsBalanceKey {
			
		public FavorsBalanceKey(double fb,int i,PeerProfile peer_profile)
			throws GridException {
				
			if (peer_profile == null) {
				throw new GridException("FavorsBalanceKey(): "+
										"illegal Peer profile");
			}
			
			this.fb = fb;
			this.i = i;
			this.peer = peer_profile;
			
		}
			
		/** Favors Balance of target Peer (peers[i]).
		 */
		public final double fb;
		
		/** Index of target Peer (peers[i]).
		 */
		public final int i;
		
		/** Target Peer (peers[i]).
		 */
		public final PeerProfile peer;
		
	}
	
	/** Helper class of FavorsSupplyingRanking.
	 * Comparator of FavorsBalanceKeys (by decreasing order of Favors Balance).
	 */
	private static class DecreasingFBComparator
		implements Comparator<FavorsBalanceKey>, Serializable {
		
		public DecreasingFBComparator() {
			
		}
		
		public int compare(FavorsBalanceKey k1,FavorsBalanceKey k2) {
			
			if (k1.fb < k2.fb) {
				return +1;
			}
			else if (k1.fb > k2.fb) {
				return -1;
			}
			else /* if (k1.fb == k2.fb) */ {
				return (k1.i == k2.i) ? 0 : ((k1.i < k2.i) ? +1 : -1);
			}
			
		}
		
		static final long serialVersionUID = -8111277913627670039L;
		
	}

	// ****************************************************************

}

