/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// GrantsConsumptionRanking.java

package lbg.peer.strategy.ranking;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.peer.register.*;
import lbg.peer.register.learning.consumption.*;
import lbg.peer.register.negotiation.*;
import java.io.Serializable;
import java.util.*;

/** GrantsConsumptionRanking ranks Grantors
 * (i.e. potential Supplier Peers who sent Consumption Grants)
 * by decreasing number of sent consumption grants.
 * @author Cyril Briquet
 */
public class GrantsConsumptionRanking extends CommonRanking {

	/** Constructs a new GrantsConsumptionRanking.
	 * @param c_g Consumption Grants container
	 * @exception GridException if a parameter is illegal
	 */
	public GrantsConsumptionRanking(ConsumptionGrantsContainer c_g,
					  				JobConsumptionOutcomeTracker jcot,
					  				boolean do_blacklist)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("GrantsConsumptionRanking(): "+
									"illegal Consumption Grants Container");
		}
		if (jcot == null) {
			throw new GridException("GrantsConsumptionRanking(): "+
									"illegal Job Consumption outcome tracker");
		}
		
		// Obtain pending Grantors profiles set
		Collection<Integer> pending_grantors_index = c_g.pendingGrantorsIndex();
		Set<GrantKey> pending_grantors_announced_availability =
			new TreeSet<GrantKey>(GrantsConsumptionRanking.default_cmp);
		
		// Rank Grantors by decreasing number of sent consumption grants
		int tmp_ranked_peers = 0;
		int tmp_blacklisted_peers = 0;
		for (Integer grantor_idx : pending_grantors_index) {
			int grants = c_g.initialConsumptionGrants(grantor_idx);
			if (grants < 1) {
				continue;
			}

			PeerProfile grantor_profile = c_g.grantorProfile(grantor_idx);

			if (do_blacklist) {
				boolean should_blacklist_job_level =
					jcot.shouldBlackListPeer(grantor_profile.id());
				if (should_blacklist_job_level) {
					++tmp_blacklisted_peers;
					continue;
				}
				// no Peer-level blacklisting
			}

			GrantKey grant_key = new GrantKey(grants,grantor_idx);
			if (pending_grantors_announced_availability.add(grant_key) ==
				false) {
				throw new GridException("GrantsConsumptionRanking(): "+
										"corrupted data");
			}

			++tmp_ranked_peers;
		}

		this.ranked_peers = tmp_ranked_peers;
		this.unranked_peers = pending_grantors_index.size()-this.ranked_peers;
		this.blacklisted_peers = tmp_blacklisted_peers;
		
		if (tmp_ranked_peers == 0) {
			this.peers_ranking = null;
			return;
		}
		
		// Fill peers_ranking with performance-ranked Grantors
		// that are estimated to be faster than the managed Peer
		this.peers_ranking = new int[ranked_peers];
		int i = 0;
		for (GrantKey grant_key : pending_grantors_announced_availability) {
			this.peers_ranking[i++] = grant_key.grantor_idx;
		}
		if (i != this.ranked_peers) {
			throw new GridException("GrantsConsumptionRanking(): "+
									"corrupted data");
		}
		
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[CommonRanking]",
							   "[GrantsConsumptionRanking]"));

		return sb.toString();
		
	}

	// ****************************************************************

	/** Ranking implementation.
	 * @return Peers ranking, or null if no Peers can be ranked
	 */
	public int[] peersRanking() {
		
		return this.peers_ranking;
		
	}

	/** Ranking implementation.
	 */
	public int getRankedPeerCount() {
		
		return this.ranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getUnrankedPeerCount() {
		
		return this.unranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getBlacklistedPeerCount() {
		
		return this.blacklisted_peers;
		
	}

	// ****************************************************************

	private final int[] peers_ranking;
	private final int ranked_peers;
	private final int unranked_peers;
	private final int blacklisted_peers;

	private static final GrantsComparator default_cmp = new GrantsComparator();

	// ****************************************************************
	
	/** Helper class of GrantsConsumptionRanking.
	 */
	private static class GrantKey {
		
		public GrantKey(int grants,int grantor_idx) {
			
			this.grants = grants;
			this.grantor_idx = grantor_idx;
			
		}
		
		public final int grants;
		public final int grantor_idx;
		
	}
	
	// ****************************************************************
	
	/** Helper class of GrantsConsumptionRanking.
	 * Comparator of GrantKeys (by decreasing grants).
	 */
	private static class GrantsComparator
		implements Comparator<GrantKey>, Serializable {
		
		public GrantsComparator() {
			
		}
		
		public int compare(GrantKey k1,GrantKey k2) {
			
			if (k1.grants < k2.grants) {
				return +1;
			}
			else if (k1.grants > k2.grants) {
				return -1;
			}
			else /* if (k1.grants == k2.grants) */ {
				if (k1.grantor_idx < k2.grantor_idx) {
					return -1;
				}
				else if (k1.grantor_idx > k2.grantor_idx) {
					return +1;
				}
				else /* if (k1.grantor_idx == k2.grantor_idx) */ {
					return 0;
				}
			}
			
		}
		
		static final long serialVersionUID = 8630173185887145247L;
		
	}

	// ****************************************************************

}

