/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// MC4Ranking.java

package lbg.peer.strategy.ranking;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.peer.register.*;
import lbg.peer.register.bartering.*;
import lbg.peer.register.learning.consumption.*;
import lbg.peer.register.negotiation.*;
import lbg.task.*;
import java.io.Serializable;
import java.util.*;

/** MC4Ranking ranks Grantors (i.e. potential Supplier Peers
 * who sent Consumption Grants) by decreasing MC4.
 * @author Cyril Briquet
 */
public class MC4Ranking extends CommonRanking {

	public MC4Ranking(ConsumptionGrantsContainer c_g,
					  JobConsumptionOutcomeTracker jcot,
					  boolean do_blacklist)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("MC4Ranking(): "+
									"illegal Consumption Grants Container");
		}
		if (jcot == null) {
			throw new GridException("MC4Ranking(): "+
									"illegal Job Consumption outcome tracker");
		}
		
		// Obtain pending Grantors profiles set
		Collection<Integer> pending_grantors_index = c_g.pendingGrantorsIndex();
		Set<MC4Key> pending_grantors_mc4_set =
			new TreeSet<MC4Key>(MC4Ranking.default_cmp);
		
		// Rank Grantors by decreasing MC4-ranked reliability
		int tmp_ranked_peers = 0;
		int tmp_blacklisted_peers = 0;
		for (Integer grantor_idx : pending_grantors_index) {
			PeerProfile grantor_profile = c_g.grantorProfile(grantor_idx);
			PeerBarteringProfile peer_bp =
				grantor_profile.getBarteringProfile();
			double mc4 = peer_bp.getConsumptionMC4();
			
			if (do_blacklist) {
				boolean should_blacklist_job_level =
					jcot.shouldBlackListPeer(grantor_profile.id());
				if (should_blacklist_job_level) {
					continue;
				}
				// no Peer-level blacklisting
			}
			
			// Sort by increasing MC4
			MC4Key mc4_key = new MC4Key(mc4,grantor_idx);
			if (pending_grantors_mc4_set.add(mc4_key) == false) {
				throw new GridException("MC4Ranking(): corrupted data");
			}
			tmp_ranked_peers++;
		}

		this.ranked_peers = tmp_ranked_peers;
		this.unranked_peers = pending_grantors_index.size()-this.ranked_peers;
		this.blacklisted_peers = tmp_blacklisted_peers;
		
		if (tmp_ranked_peers == 0) {
			this.peers_ranking = null;
			return;
		}

		// Fill peers_ranking with MC4-ranked Grantors
		this.peers_ranking = new int[ranked_peers];
		int i = 0;
		for (MC4Key mc4_key : pending_grantors_mc4_set) {
			this.peers_ranking[i++] = mc4_key.grantor_idx;
		}
		
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[CommonRanking]",
							   "[MC4Ranking]"));

		return sb.toString();
		
	}

	// ****************************************************************

	/** Ranking implementation.
	 * @return Peers ranking, or null if no Peers can be ranked
	 */
	public int[] peersRanking() {
		
		return this.peers_ranking;
		
	}

	/** Ranking implementation.
	 */
	public int getRankedPeerCount() {
		
		return this.ranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getUnrankedPeerCount() {
		
		return this.unranked_peers;
		
	}

	/** Ranking implementation.
	 */
	public int getBlacklistedPeerCount() {
		
		return this.blacklisted_peers;
		
	}
	
	// ****************************************************************

	private final int[] peers_ranking;
	private final int ranked_peers;
	private final int unranked_peers;
	private final int blacklisted_peers;

	private static final MC4Comparator default_cmp = new MC4Comparator();

	// ****************************************************************
	
	/** Helper class of MC4Ranking.
	 */
	private static class MC4Key {
		
		public MC4Key(double mc4,int grantor_idx) {
			
			this.mc4 = mc4;
			this.grantor_idx = grantor_idx;
			
		}
		
		public final double mc4;
		public final int    grantor_idx;
		
	}
	
	// ****************************************************************
	
	/** Helper class of MC4Ranking.
	 * Comparator of MC4Keys (by increasing MC4).
	 */
	private static class MC4Comparator
		implements Comparator<MC4Key>, Serializable {
		
		public MC4Comparator() {
			
		}
		
		public int compare(MC4Key k1,MC4Key k2) {
				
			if (k1.mc4 < k2.mc4) {
				return -1;
			}
			else if (k1.mc4 > k2.mc4) {
				return +1;
			}
			else /* if (k1.mc4 == k2.mc4) */ {
				if (k1.grantor_idx < k2.grantor_idx) {
					return -1;
				}
				else if (k1.grantor_idx > k2.grantor_idx) {
					return +1;
				}
				else /* if (k1.grantor_idx == k2.grantor_idx) */ {
					return 0;
				}
			}
			
		}
		
		static final long serialVersionUID = 8630173185887145247L;
		
	}

	// ****************************************************************

}

