/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// MCoSConsumptionRanking.java

package lbg.peer.strategy.ranking;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.common.math.*;
import lbg.peer.register.*;
import lbg.peer.register.bartering.*;
import lbg.peer.register.learning.consumption.*;
import lbg.peer.register.negotiation.*;
import lbg.task.*;
import java.io.Serializable;
import java.util.*;

/** MCoSConsumptionRanking ranks Grantors (i.e. potential Supplier Peers
 * who sent Consumption Grants)
 * first by decreasing mean Consumption completion stride,
 * then by decreasing (mean,current) normalized stride distance
 * between the current and the mean Consumption completion stride.
 * @author Cyril Briquet
 */
public class MCoSConsumptionRanking extends CommonRanking {

	public MCoSConsumptionRanking(ConsumptionGrantsContainer c_g,
								  JobConsumptionOutcomeTracker jcot,
								  boolean do_blacklist)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("MCoSConsumptionRanking(): "+
									"illegal Consumption Grants Container");
		}
		if (jcot == null) {
			throw new GridException("MCoSConsumptionRanking(): "+
									"illegal Job Consumption outcome tracker");
		}
		
		// Obtain pending Grantors profiles set
		Collection<Integer> pending_grantors_index = c_g.pendingGrantorsIndex();
		Set<StrideKey> pending_grantors_stride_set =
			new TreeSet<StrideKey>(MCoSConsumptionRanking.default_cmp);
		
		// Rank Grantors by decreasing stride-ranked reliability
		int tmp_ranked_peers = 0;
		int tmp_blacklisted_peers = 0;
		for (Integer grantor_idx : pending_grantors_index) {
			PeerProfile grantor_profile = c_g.grantorProfile(grantor_idx);
			PeerBarteringProfile peer_bp =
				grantor_profile.getBarteringProfile();
			double stride_distance = normalizedStrideDistance(peer_bp);
			double mean_stride =
				peer_bp.getMeanConsumptionCompletionStride();
			long current_stride =
				peer_bp.getRecentlyCompletedConsumptionTasksCount();
			
			if (do_blacklist) {
				boolean should_blacklist_job_level =
					jcot.shouldBlackListPeer(grantor_profile.id());
				if (should_blacklist_job_level) {
					++tmp_blacklisted_peers;
					continue;
				}
				// no Peer-level blacklisting
			}

			// Sort first by decreasing mean stride,
			// then by decreasing (mean,current) normalized stride distance.
			StrideKey stride_key =
				new StrideKey(stride_distance,mean_stride,grantor_idx);
			if (pending_grantors_stride_set.add(stride_key) == false) {
				throw new GridException("MCoSConsumptionRanking(): "+
										"corrupted data");
			}
			tmp_ranked_peers++;
		}

		this.ranked_peers = tmp_ranked_peers;
		this.unranked_peers = pending_grantors_index.size()-this.ranked_peers;
		this.blacklisted_peers = tmp_blacklisted_peers;
		
		if (tmp_ranked_peers == 0) {
			this.peers_ranking = null;
			return;
		}

		// Fill peers_ranking with stride-ranked Grantors
		this.peers_ranking = new int[ranked_peers];
		int i = 0;
		for (StrideKey stride_key : pending_grantors_stride_set) {
			this.peers_ranking[i++] = stride_key.grantor_idx;
		}
		
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[CommonRanking]",
							   "[MCoSConsumptionRanking]"));

		return sb.toString();
		
	}

	// ****************************************************************

	/** Ranking implementation.
	 * @return Peers ranking, or null if no Peers can be ranked
	 */
	public int[] peersRanking() {
		
		return this.peers_ranking;
		
	}

	/** Ranking implementation.
	 */
	public int getRankedPeerCount() {
		
		return this.ranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getUnrankedPeerCount() {
		
		return this.unranked_peers;
		
	}

	/** Ranking implementation.
	 */
	public int getBlacklistedPeerCount() {
		
		return this.blacklisted_peers;
		
	}

	// ****************************************************************

	/** Computes the normalized difference between the mean and the most recent
	 * Consumption Completion Stride of target Supplier.
	 * <p>A high normalized stride distance, i.e. close to 1.0,
	 * indicates a high probability of success of a Consumption Task
	 * submitted right now to the target Supplier.
	 * @param peer_bp target Supplier Bartering Profile
	 * @return normalized (i.e. included in [0.0..1.0]) stride distance
	 * @exception GridException if a parameter is illegal
	 */
	public static double
		normalizedStrideDistance(PeerBarteringProfile peer_bp)
		throws GridException {
		
		if (peer_bp == null) {
			throw new GridException("normalizedStrideDistance(): "+
									"illegal Peer Bartering Profile");
		}

		double mean_stride =
			peer_bp.getMeanConsumptionCompletionStride();
		long current_stride =
			peer_bp.getRecentlyCompletedConsumptionTasksCount();
		
		return CommonMath.normalizeValue(current_stride,mean_stride);
		
	}
	
	// ****************************************************************

	private final int[] peers_ranking;
	private final int ranked_peers;
	private final int unranked_peers;
	private final int blacklisted_peers;

	private static final StrideComparator default_cmp = new StrideComparator();

	// ****************************************************************
	
	/** Helper class of MCoSConsumptionRanking.
	 */
	private static class StrideKey {
		
		public StrideKey(double stride_distance,double mean_stride,
						 int grantor_idx) {
			
			this.stride_distance = stride_distance;
			this.mean_stride = mean_stride;
			this.grantor_idx = grantor_idx;
			
		}
		
		public final double stride_distance;
		public final double mean_stride;
		public final int grantor_idx;
		
	}
	
	// ****************************************************************
	
	/** Helper class of MCoSConsumptionRanking.
	 * Comparator of StrideKeys
	 * (by decreasing mean stride,
	 * then by decreasing (mean,current) normalized stride distance).
	 */
	private static class StrideComparator
		implements Comparator<StrideKey>, Serializable {
		
		public StrideComparator() {
			
		}
		
		public int compare(StrideKey k1,StrideKey k2) {
				
			if (k1.mean_stride < k2.mean_stride) {
				return +1;
			}
			else if (k1.mean_stride > k2.mean_stride) {
				return -1;
			}
			else /* if (k1.mean_stride == k2.mean_stride) */ {
				if (k1.stride_distance < k2.stride_distance) {
					return +1;
				}
				else if (k1.stride_distance > k2.stride_distance) {
					return -1;
				}
				else /* if (k1.stride_distance == k2.stride_distance) */ {
					if (k1.grantor_idx < k2.grantor_idx) {
						return -1;
					}
					else if (k1.grantor_idx > k2.grantor_idx) {
						return +1;
					}
					else /* if (k1.grantor_idx == k2.grantor_idx) */ {
						return 0;
					}
				}
			}
			
		}
		
		static final long serialVersionUID = -4660043055462230343L;
		
	}

	// ****************************************************************

}

