/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// MTTCConsumptionRanking.java

package lbg.peer.strategy.ranking;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.peer.register.*;
import lbg.peer.register.bartering.*;
import lbg.peer.register.learning.consumption.*;
import lbg.peer.register.negotiation.*;
import lbg.task.*;
import java.io.Serializable;
import java.util.*;

/** MTTCConsumptionRanking ranks Grantors (i.e. potential Supplier Peers
 * who sent Consumption Grants)
 * by increasing MTTC.
 * @author Cyril Briquet
 */
public class MTTCConsumptionRanking extends CommonRanking {

	public MTTCConsumptionRanking(ConsumptionGrantsContainer c_g,
								  JobConsumptionOutcomeTracker jcot,
								  boolean do_blacklist)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("MTTCConsumptionRanking(): "+
									"illegal Consumption Grants Container");
		}
		if (jcot == null) {
			throw new GridException("MTTCConsumptionRanking(): "+
									"illegal Job Consumption outcome tracker");
		}
		
		// Obtain pending Grantors profiles set
		Collection<Integer> pending_grantors_index = c_g.pendingGrantorsIndex();
		Set<MTTCKey> pending_grantors_mttc_set =
			new TreeSet<MTTCKey>(MTTCConsumptionRanking.default_inc_cmp);
		
		// Rank Grantors by increasing MTTC
		int tmp_ranked_peers = 0;
		int tmp_blacklisted_peers = 0;
		for (Integer grantor_idx : pending_grantors_index) {
			PeerProfile grantor_profile = c_g.grantorProfile(grantor_idx);
			PeerBarteringProfile peer_bp =
				grantor_profile.getBarteringProfile();
			double mttc = peer_bp.getConsumptionMTTC();
			
			if (do_blacklist) {
				boolean should_blacklist_job_level =
					jcot.shouldBlackListPeer(grantor_profile.id());
				if (should_blacklist_job_level) {
					++tmp_blacklisted_peers;
					continue;
				}
				// no Peer-level blacklisting
			}

			// Sort by increasing MTTC
			MTTCKey mttc_key = new MTTCKey(mttc,grantor_idx);
			if (pending_grantors_mttc_set.add(mttc_key) == false) {
				throw new GridException("MTTCConsumptionRanking(): "+
										"corrupted data");
			}
			tmp_ranked_peers++;
		}

		this.ranked_peers = tmp_ranked_peers;
		this.unranked_peers = pending_grantors_index.size()-this.ranked_peers;
		this.blacklisted_peers = tmp_blacklisted_peers;
		
		if (tmp_ranked_peers == 0) {
			this.peers_ranking = null;
			return;
		}

		// Fill peers_ranking with stride-ranked Grantors
		this.peers_ranking = new int[ranked_peers];
		int i = 0;
		for (MTTCKey mttc_key : pending_grantors_mttc_set) {
			this.peers_ranking[i++] = mttc_key.grantor_idx;
		}
		
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[CommonRanking]",
							   "[MTTCConsumptionRanking]"));

		return sb.toString();
		
	}

	// ****************************************************************

	/** Ranking implementation.
	 * @return Peers ranking, or null if no Peers can be ranked
	 */
	public int[] peersRanking() {
		
		return this.peers_ranking;
		
	}

	/** Ranking implementation.
	 */
	public int getRankedPeerCount() {
		
		return this.ranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getUnrankedPeerCount() {
		
		return this.unranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getBlacklistedPeerCount() {
		
		return this.blacklisted_peers;
		
	}
	
	// ****************************************************************

	private final int[] peers_ranking;
	private final int ranked_peers;
	private final int unranked_peers;
	private final int blacklisted_peers;

	private static final MTTCIncreasingComparator default_inc_cmp =
		new MTTCIncreasingComparator();
	private static final MTTCDecreasingComparator default_dec_cmp =
		new MTTCDecreasingComparator();

	// ****************************************************************
	
	/** Helper class of MTTCConsumptionRanking.
	 */
	private static class MTTCKey {
		
		public MTTCKey(double MTTC,int grantor_idx) {
			
			this.MTTC = MTTC;
			this.grantor_idx = grantor_idx;
			
		}
		
		public final double MTTC;
		public final int grantor_idx;
		
	}
	
	// ****************************************************************
	
	/** Helper class of MTTCConsumptionRanking.
	 * Comparator of MTTCKeys (by increasing MTTC).
	 */
	private static class MTTCIncreasingComparator
		implements Comparator<MTTCKey>, Serializable {
		
		public MTTCIncreasingComparator() {
			
		}
		
		public int compare(MTTCKey k1,MTTCKey k2) {
				
			if (k1.MTTC < k2.MTTC) {
				return -1;
			}
			else if (k1.MTTC > k2.MTTC) {
				return +1;
			}
			else /* if (k1.MTTC == k2.MTTC) */ {
				if (k1.grantor_idx < k2.grantor_idx) {
					return -1;
				}
				else if (k1.grantor_idx > k2.grantor_idx) {
					return +1;
				}
				else /* if (k1.grantor_idx == k2.grantor_idx) */ {
					return 0;
				}
			}
			
		}
		
		static final long serialVersionUID = -1797099406692034152L;
		
	}

	// ****************************************************************
	
	/** Helper class of MTTCConsumptionRanking.
	 * Comparator of MTTCKeys (by decreasing MTTC).
	 */
	private static class MTTCDecreasingComparator
		implements Comparator<MTTCKey>, Serializable {
		
		public MTTCDecreasingComparator() {
			
		}
		
		public int compare(MTTCKey k1,MTTCKey k2) {
				
			if (k1.MTTC < k2.MTTC) {
				return +1;
			}
			else if (k1.MTTC > k2.MTTC) {
				return -1;
			}
			else /* if (k1.MTTC == k2.MTTC) */ {
				if (k1.grantor_idx < k2.grantor_idx) {
					return -1;
				}
				else if (k1.grantor_idx > k2.grantor_idx) {
					return +1;
				}
				else /* if (k1.grantor_idx == k2.grantor_idx) */ {
					return 0;
				}
			}
			
		}
		
		static final long serialVersionUID = 1312875176103447021L;
		
	}

	// ****************************************************************
}

