/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PerformanceConsumptionRanking.java

package lbg.peer.strategy.ranking;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.peer.register.*;
import lbg.peer.register.learning.consumption.*;
import lbg.peer.register.negotiation.*;
import java.util.*;

/** PerformanceConsumptionRanking ranks Grantors
 * (i.e. potential Supplier Peers who sent Consumption Grants)
 * by decreasing expected performance, excluding those
 * estimated to be PERF_FACTOR times slower than the managed Peer.
 * @author Cyril Briquet
 */
public class PerformanceConsumptionRanking extends CommonRanking {

	/** Constructs a new PerformanceConsumptionRanking.
	 * @param c_g Consumption Grants container
	 * @exception GridException if a parameter is illegal
	 */
	public PerformanceConsumptionRanking(ConsumptionGrantsContainer c_g,
										 double estimated_mean_self_runtime,
										 JobConsumptionOutcomeTracker jcot,
										 boolean do_blacklist)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("PerformanceConsumptionRanking(): "+
									"illegal Consumption Grants Container");
		}
		if (estimated_mean_self_runtime < 0.0) {
			throw new GridException("PerformanceConsumptionRanking(): "+
									"illegal estimated mean self runtime");
		}
		if (jcot == null) {
			throw new GridException("PerformanceConsumptionRanking(): "+
									"illegal Job Consumption outcome tracker");
		}
		
		// Obtain pending Grantors profiles set
		Collection<Integer> pending_grantors_index = c_g.pendingGrantorsIndex();
		Map<Double,Set<Integer>> pending_grantors_performance =
			new TreeMap<Double,Set<Integer>>();
		
		double threshold = PERF_FACTOR * estimated_mean_self_runtime;
		
		// Rank Grantors by decreasing estimated performance
		int tmp_ranked_peers = 0;
		int tmp_blacklisted_peers = 0;
		for (Integer grantor_idx : pending_grantors_index) {
			PeerProfile grantor_profile = c_g.grantorProfile(grantor_idx);
			double mean_consumption_task_runtime =
				grantor_profile.getBarteringProfile().
					getMeanConsumptionTaskRuntime();
			
			// Don't rank slower Peers
			if (mean_consumption_task_runtime > threshold) {
				continue;
			}

			if (do_blacklist) {
				boolean should_blacklist_job_level =
					jcot.shouldBlackListPeer(grantor_profile.id());
				if (should_blacklist_job_level) {
					++tmp_blacklisted_peers;
					continue;
				}
				// no Peer-level blacklisting
			}

			Set<Integer> grantors_with_given_performance =
				pending_grantors_performance.get(mean_consumption_task_runtime);
			if (grantors_with_given_performance == null) {
				grantors_with_given_performance = new TreeSet<Integer>();
				pending_grantors_performance.
					put(mean_consumption_task_runtime,
						grantors_with_given_performance);
			}
			grantors_with_given_performance.add(grantor_idx);
			tmp_ranked_peers++;
		}

		this.ranked_peers = tmp_ranked_peers;
		this.unranked_peers = pending_grantors_index.size()-this.ranked_peers;
		this.blacklisted_peers = tmp_blacklisted_peers;
		
		if (tmp_ranked_peers == 0) {
			this.peers_ranking = null;
			return;
		}
		
		// Fill peers_ranking with performance-ranked Grantors
		// that are estimated to be faster than the managed Peer
		this.peers_ranking = new int[ranked_peers];
		int i = 0;
		for (Set<Integer> grantors_with_given_performance :
			 pending_grantors_performance.values()) {
			for (Integer grantor_idx : grantors_with_given_performance) {
				this.peers_ranking[i++] = grantor_idx;
			}
		}
		if (i != this.ranked_peers) {
			throw new GridException("PerformanceConsumptionRanking(): "+
									"corrupted data");
		}
		
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[CommonRanking]",
							   "[PerformanceConsumptionRanking]"));

		return sb.toString();
		
	}

	// ****************************************************************

	/** Ranking implementation.
	 * @return Peers ranking, or null if no Peers can be ranked
	 */
	public int[] peersRanking() {
		
		return this.peers_ranking;
		
	}

	/** Ranking implementation.
	 */
	public int getRankedPeerCount() {
		
		return this.ranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getUnrankedPeerCount() {
		
		return this.unranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getBlacklistedPeerCount() {
		
		return this.blacklisted_peers;
		
	}

	// ****************************************************************

	public static final double PERF_FACTOR = 2.0;

	private final int[] peers_ranking;
	private final int ranked_peers;
	private final int unranked_peers;
	private final int blacklisted_peers;

}

