/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Rankings.java

package lbg.peer.strategy.ranking;

import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.math.*;

import java.io.Serializable;
import java.util.*;

/**
 * @author Cyril Briquet
 */
public class Rankings {
	
	private Rankings() {
		
	}

	/** Merges the target rankings into one ranking.
	 * @param rankings target rankings
	 * @exception GridException if a parameter is illegal
	 */
	public static Ranking merge(Ranking... rankings)
		throws GridException {
	
		if ((rankings == null) || (rankings.length == 0)) {
			throw new GridException("merge(): illegal rankings");
		}
		
		// STEP 0 - identify Peers within the target rankings
		TreeSet<Integer> all_peers_idx = new TreeSet<Integer>();
		for (Ranking ranking : rankings) {
			int[] peers_idx = ranking.peersRanking();
			if (peers_idx != null) {
				for (int i = 0 ; i < peers_idx.length ; i++) {
					all_peers_idx.add(Tank.newInteger(peers_idx[i]));
				}
			}
		}
		int absolute_max_rank = all_peers_idx.size();

		// STEP 1 - build Peers metadata
		TreeMap<Integer,PeerRankingsMetaData> all_peers =
			new TreeMap<Integer,PeerRankingsMetaData>();
		for (Integer peer_idx : all_peers_idx) {
			PeerRankingsMetaData peer_data =
				new PeerRankingsMetaData(peer_idx,rankings,absolute_max_rank);
			if (all_peers.put(peer_idx,peer_data) != null) {
				throw new GridException("merge(): corrupted data");
			}
		}
		
		// STEP 2 - complete Peers metadata with exact rank for each ranking
		int r = 0;
		for (Ranking ranking : rankings) {
			int[] peers_idx = ranking.peersRanking();
			// Process target ranking
			if (peers_idx != null) {
				for (int i = 0 ; i < peers_idx.length ; i++) {
					PeerRankingsMetaData peer_md =
						all_peers.get(Tank.newInteger(peers_idx[i]));
					peer_md.setRank(r,i+1); // <-- [0] == rank 1, [1] == rank 2, ...
				}
			}
			r++;
		}

		// STEP 3 - compute, and store the merged normlaized rank of each Peer
		TreeMap<Float,TreeSet<PeerRankingsMetaData>> normalized_merged_ranking =
			new TreeMap<Float,TreeSet<PeerRankingsMetaData>>();
		for (PeerRankingsMetaData peer_md : all_peers.values()) {
			// Compute the normalized merged rank of target Peer
			float normalized_merged_rank = 0.0f;
			for (PeerRankingData peer_rd : peer_md.ranking_data) {
				float value = (peer_rd.isRanked()) ?
					peer_rd.getNormalizedRank() : 1.0f;
				normalized_merged_rank += value;
				/* with normalized weights --
				  normalized_merged_rank += (value/norm_weight[w_i++])); */
			}

			normalized_merged_rank =
				normalized_merged_rank / peer_md.ranking_data.length;
				/* with normalized weights -- no need for division */
				
			// Store the normalized merged rank of target Peer
			Float key = new Float(normalized_merged_rank);
			TreeSet<PeerRankingsMetaData> peers_with_same_mnr =
				normalized_merged_ranking.get(key);
			if (peers_with_same_mnr == null) {
				peers_with_same_mnr =
					new TreeSet<PeerRankingsMetaData>(Rankings.peers_md_cmp);
				normalized_merged_ranking.put(key,peers_with_same_mnr);
			}
			
			if (peers_with_same_mnr.add(peer_md) == false) {
				throw new GridException("merge(): corrupted data");
			}
		}
		
		//debugging code -- dumpData(all_peers,normalized_merged_ranking);
		
		// STEP 4 - create the merged Ranking
		int[] merged_ranking = new int[absolute_max_rank];
		int i = 0;
		for (TreeSet<PeerRankingsMetaData> peers_with_same_mnr :
				normalized_merged_ranking.values()) {
			for (PeerRankingsMetaData peer_md : peers_with_same_mnr) {
				merged_ranking[i++] = peer_md.idx;
			}
		}

		return new MergedRanking(merged_ranking);
		
	}
	
	// ****************************************************************

	public static boolean isRankingValid(Ranking ranking) {
		
		int[] peers_idx = ranking.peersRanking();
		
		if ((peers_idx == null) || (peers_idx.length < 1)) {
			//throw new GridException("isRankingValid(): no Peer is ranked");
			return false;
		}
		
		TreeSet<Integer> ts_peers_idx = new TreeSet<Integer>();
		for (int i = 0 ; i < peers_idx.length ; i++) {
			if (ts_peers_idx.add(Tank.newInteger(peers_idx[i]))
				== false) {
				//throw new GridException("isRankingValid(): "+
				//						  "same Peer ranked twice");
				return false;
			}
		}

		return true;
		
	}
	
	// ****************************************************************

	/** For testing purposes.
	 */
	public static void main(String[] args) {
		
		try {
			MergedRanking mr1 =
				new MergedRanking(new int[] { 5, 6, 7, 8, 4, 1, 2 } );
			if (Rankings.isRankingValid(mr1) == false) {
				throw new GridException("Test: illegal ranking "+mr1);
			}
			System.out.println(mr1+"\n");

			MergedRanking mr2 =
				new MergedRanking(new int[] { 6, 5, 7, 9, 4 } );
			if (Rankings.isRankingValid(mr2) == false) {
				throw new GridException("Test: illegal ranking "+mr2);
			}
			System.out.println(mr2+"\n");

			MergedRanking mr3 =
				new MergedRanking(new int[] { 8, 6 } );
			if (Rankings.isRankingValid(mr3) == false) {
				throw new GridException("Test: illegal ranking "+mr3);
			}
			System.out.println(mr3+"\n");

			Ranking mrm =  Rankings.merge(mr1,mr2,mr3);
			if (Rankings.isRankingValid(mrm) == false) {
				throw new GridException("Test: illegal ranking "+mrm);
			}
			System.out.println("\n"+mrm);
			
		}
		catch (GridException e) {
			e.printStackTrace();
		}
		
	}

	// ****************************************************************

	private static void
		dumpData(TreeMap<Integer,PeerRankingsMetaData> all_peers,
				 TreeMap<Float,TreeSet<PeerRankingsMetaData>>
				 	normalized_merged_ranking)
		throws GridException {

		if (all_peers != null) {
			for (PeerRankingsMetaData peer_md : all_peers.values()) {
				System.out.println(peer_md);
			}
		}
		
		if (normalized_merged_ranking != null) {
			System.out.print("\nMerged normalized ranking:\n");
			Iterator<Map.Entry<Float,TreeSet<PeerRankingsMetaData>>>
				normalized_merged_ranking_it =
					normalized_merged_ranking.entrySet().iterator();
			while (normalized_merged_ranking_it.hasNext()) {
				Map.Entry<Float,TreeSet<PeerRankingsMetaData>> nmr_val =
					normalized_merged_ranking_it.next();
				Float merged_normalized_ranking = nmr_val.getKey();
				TreeSet<PeerRankingsMetaData> peers_with_same_mnr =
					nmr_val.getValue();
				
				System.out.print(CommonMath.
									div_label_2dd(merged_normalized_ranking,1));
				System.out.print(" => { ");
				for (PeerRankingsMetaData peer_md : peers_with_same_mnr) {
					System.out.print("["+peer_md.idx+"] ");
				}
				System.out.print("}\n");
			}
		}
		
	}
	
	// ****************************************************************

	private static final PeerRankingsMetaDataComparator peers_md_cmp =
		new PeerRankingsMetaDataComparator();
	
	// ****************************************************************

	private static class PeerRankingsMetaData
		implements Comparable<PeerRankingsMetaData> {
	
		public PeerRankingsMetaData(int idx,Ranking[] rankings,
									int absolute_max_rank)
			throws GridException {
			
			if (idx < 0) {
				throw new GridException("PeerRankingsMetaData(): "+
										"illegal Peer index");
			}
			if ((rankings == null) || (rankings.length < 1)) {
				throw new GridException("PeerRankingsMetaData(): "+
										"illegal rankings");
			}
			if (absolute_max_rank < 1) {
				throw new GridException("PeerRankingsMetaData(): "+
										"illegal absolute max rank");
			}
			
			this.idx = idx;
			this.ranking_data = new PeerRankingData[rankings.length];
			int i = 0;
			for (Ranking ranking : rankings) {
				this.ranking_data[i++] =
					new PeerRankingData(ranking,absolute_max_rank);				
			}
			
			this.ranked_count = 0;
			this.hashcode = Integer.MIN_VALUE;
			
		}
		
		public synchronized String toString() {
			
			StringBuilder sb = new StringBuilder();
			
			sb.append("[PeerRankingsMetaData]");
			sb.append("\n  ->  ");
			sb.append("Peer [");
			sb.append(this.idx);
			sb.append("], ranked ");
			sb.append(this.ranked_count);
			sb.append(" times");
			for (PeerRankingData peer_rd : this.ranking_data) {
				sb.append("\n  ->  ");
				sb.append(peer_rd);
			}
			
			return sb.toString();
			
		}

		public int compareTo(PeerRankingsMetaData other) {

			if (this.idx < other.idx) {
				return -1;
			}
			else if (this.idx > other.idx) {
				return +1;
			}
			else /* if (this.idx == other.idx) */ {
				return 0;
			}
			
		}
		
		@Override
		public boolean equals(Object obj) {
			
			if (obj instanceof PeerRankingsMetaData == false) {
				return false;
			}
			
			return (this.compareTo((PeerRankingsMetaData)obj) == 0);
			
		}
		
		@Override
		public int hashCode() {
		
			if (this.hashcode == Integer.MIN_VALUE) {
				this.hashcode = CommonMath.generateHashCode(this.idx);
			}
			return this.hashcode;
		
		}

		// ****************************************************************

		public synchronized void setRank(int ranking_index,int rank)
			throws GridException {
		
			if ((ranking_index < 0) ||
				(ranking_index >= this.ranking_data.length)) {
				throw new GridException("setRank(): illegal ranking");
			}
			
			this.ranking_data[ranking_index].setRank(rank);
			this.ranked_count++;
			
		}
		
		public synchronized int getRankedCount() {
		
			return this.ranked_count;
			
		}

		// ****************************************************************

		public final int idx;
		public final PeerRankingData[] ranking_data;
		
		private int ranked_count;
		private int hashcode;
		
	}
	
	private static class PeerRankingData {

		public PeerRankingData(Ranking ranking,int absolute_max_rank)
			throws GridException {
			
			if (ranking == null) {
				throw new GridException("PeerRankingData(): illegal Ranking");
			}
			if (absolute_max_rank < 1) {
				throw new GridException("PeerRankingData(): "+
										"illegal absolute max rank");
			}
			if (isRankingValid(ranking) == false) {
				throw new GridException("PeerRankingData(): illegal ranking");
			}
			
			this.max_rank          = ranking.getPeerCount();
			this.absolute_max_rank = absolute_max_rank;
			this.rank              = -1;
			this.normalized_rank   = -1.0f;
			this.is_ranked         = false;
			
			if (this.max_rank > this.absolute_max_rank) {
				throw new GridException("PeerRankingData(): incompatible "+
										"max rank and absolute max rank");
			}
			
		}
		
		public synchronized String toString() {
			
			StringBuilder sb = new StringBuilder();
			
			sb.append("[PeerRankingData]");
			if (this.is_ranked) {
				sb.append(" ranked ");
				sb.append(this.rank);
				sb.append("/");
				sb.append(this.max_rank);
				sb.append(" (");
				sb.append(CommonMath.div_label_2dd(this.normalized_rank,1));
				sb.append(" normalized)");
			}
			else {
				sb.append(" unranked");
			}
			
			return sb.toString();
			
		}
		
		// ****************************************************************

		public synchronized boolean isRanked() {
		
			return this.is_ranked;
			
		}
		
		public synchronized int getMaxRank() {
			
			return this.max_rank;
			
		}

		/** Peer rank in target Ranking.
		 * @return Peer rank, in [1..getMaxRank()]
		 */
		public synchronized int getRank() throws GridException {
			
			if (this.is_ranked == false) {
				throw new GridException("getRank(): Peer not ranked "+
										"in target ranking");
			}
			
			return this.rank;
			
		}
		
		/** Normalizrd Peer rank in target Ranking.
		 * @return Peer rank, ]0.0..1.0]
		 */
		public synchronized float getNormalizedRank() throws GridException {
			
			if (this.is_ranked == false) {
				throw new GridException("getRank(): Peer not ranked "+
										"in target ranking");
			}

			return this.normalized_rank;
			
		}

		public synchronized void setRank(int rank) throws GridException {
		
			if (this.is_ranked) {
				throw new GridException("setRank(): already ranked");
			}
			if ((rank < 1) || (rank > max_rank)) {
				throw new GridException("setRank(): illegal rank");
			}
			
			this.rank = rank;
			this.normalized_rank = ((float)this.rank)/((float)this.max_rank);
			this.is_ranked = true;
			
		}
		
		// ****************************************************************
		
		private final int max_rank; // Max rank for target ranking
		private final int absolute_max_rank; // Max rank for all rankings
		private int rank;
		private float normalized_rank;
		private boolean is_ranked;
		
	}
	
	// ****************************************************************
	
	private static class PeerRankingsMetaDataComparator
		implements Comparator<PeerRankingsMetaData>, Serializable {
		
		protected PeerRankingsMetaDataComparator() {
		
		}
		
		// Sort by decreasing ranked count
		public int compare(PeerRankingsMetaData peer_md1,
						   PeerRankingsMetaData peer_md2) {
			
			if (peer_md1.getRankedCount() < peer_md2.getRankedCount()) {
				return +1;
			}
			else if (peer_md1.getRankedCount() > peer_md2.getRankedCount()) {
				return -1;
			}
			else /*if (peer_md1.getRankedCount()==peer_md2.getRankedCount())*/ {
				if (peer_md1.idx < peer_md2.idx) {
					return -1;
				}
				else if (peer_md1.idx > peer_md2.idx) {
					return +1;
				}
				else /* if (peer_md1.idx == peer_md2.idx) */ {
					return 0; // should not happen !
				}
			}
			
		}
	
		static final long serialVersionUID = 3332351255776416503L;
		
	}

	// ****************************************************************

}

