/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ReliableMLConsumptionRanking.java

package lbg.peer.strategy.ranking;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.peer.register.*;
import lbg.peer.register.learning.db.TaskSchedulingContext;
import lbg.peer.register.negotiation.*;
import lbg.task.*;
import static lbg.task.TaskOutcome.*;

import java.io.Serializable;
import java.util.*;

/** ReliableMLConsumptionRanking ranks Grantors (i.e. potential Supplier Peers
 * who sent Consumption Grants)
 * by decreasing predicted reliablity that is computed using ML techniques.
 * @author Cyril Briquet
 */
public class ReliableMLConsumptionRanking extends CommonRanking {

	public ReliableMLConsumptionRanking(ConsumptionGrantsContainer c_g,Task t)
		throws GridException {
		
		this(c_g,t,false);
		
	}

	public ReliableMLConsumptionRanking(ConsumptionGrantsContainer c_g,Task t,
									 	boolean keep_only_best_tsc)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("ReliableMLConsumptionRanking(): "+
									"illegal Consumption Grants Container");
		}
		if (t == null) {
			throw new GridException("ReliableMLConsumptionRanking(): "+
									"illegal Task");
		}
		
		TreeSet<ReliabilityKey> peers_ranking_set =
			new TreeSet<ReliabilityKey>(new DecreasingReliabilityComparator());

		for (Integer idx : c_g.pendingGrantorsIndex()) {
			// Hypothesis: (k1.i != k2.i) always hold for data actually inserted
			PeerProfile peer_profile = c_g.grantorProfile(idx);
			TaskSchedulingContext tsc =
				peer_profile.createConsumptionTaskSchedulingContext(t);
			double favors_balance =
				peer_profile.getBarteringProfile().getFavorsBalance();
				
			ReliabilityKey key = new ReliabilityKey(tsc,favors_balance,idx);
			peers_ranking_set.add(key);
		}
		
		int s = peers_ranking_set.size();
		if (s != c_g.pendingGrantorsIndex().size()) {
			throw new GridException("ReliableMLConsumptionRanking(): "+
									"corrupted data");
		}

		if (keep_only_best_tsc) {
			this.ranked_peers = 1;
			this.unranked_peers = s-1;
			this.blacklisted_peers = 0;

			this.peers_ranking = new int[1];
			this.peers_prediction = new TaskSchedulingContext[1];
			ReliabilityKey best_key = peers_ranking_set.first();
			this.peers_ranking[0] = best_key.i;
			this.peers_prediction[0] = best_key.tsc;
/* unused code --
			// Immediately release memory of unused TaskSchedulingContexts
			peers_ranking_set.remove(best_key);
			for (ReliabilityKey key : peers_ranking_set) {
				key.tsc.releaseAttributesArray();
			}
*/
		}
		else {
			this.ranked_peers = s;
			this.unranked_peers = 0;
			this.blacklisted_peers = 0;

			this.peers_ranking = new int[s];
			this.peers_prediction = new TaskSchedulingContext[s];
			int idx = 0;
			for (ReliabilityKey key : peers_ranking_set) {
				this.peers_ranking[idx] = key.i;
				this.peers_prediction[idx] = key.tsc;
				idx++;
			}
		}

	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ReliableMLConsumptionRanking]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	public TaskSchedulingContext[] peersPrediction() {
	
		return this.peers_prediction;
		
	}

	/** Ranking implementation.
	 * @return Peers ranking, or null if no Peers can be ranked
	 */
	public int[] peersRanking() {
	
		return this.peers_ranking;
		
	}

	/** Ranking implementation.
	 */
	public int getRankedPeerCount() {
		
		return this.ranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getUnrankedPeerCount() {
		
		return this.unranked_peers;
		
	}

	/** Ranking implementation.
	 */
	public int getBlacklistedPeerCount() {
		
		return this.blacklisted_peers;
		
	}

	// ****************************************************************

	private final TaskSchedulingContext[] peers_prediction;
	private final int[] peers_ranking;
	private final int ranked_peers;
	private final int unranked_peers;
	private final int blacklisted_peers;

	// ****************************************************************

	/** Helper class of ReliableMLConsumptionRanking.
	 */
	private static class ReliabilityKey {
			
		public ReliabilityKey(TaskSchedulingContext tsc,double fb,int i) {
			
			this.tsc = tsc;
			this.predicted_task_outcome = tsc.predictedTaskOutcome();
			this.fb = fb;
			this.i = i;
				
		}
			
		/** TaskSchedulingContext of target Peer (peers[i]).
		 */
		public TaskSchedulingContext tsc;
		
		/** Predicted Task outcome of target Peer (peers[i]).
		 */
		public TaskOutcome predicted_task_outcome;
	
		/** Favors Balance of target Peer (peers[i]).
		 */
		public double fb;
		
		/** Index of target Peer (peers[i]).
		 */
		public int i;
		
	}
	
	// ****************************************************************

	/** Helper class of ReliableMLConsumptionRanking.
	 * Comparator of ReliabilityKeys that sorts keys
	 * by decreasing order of Reliability,
	 * then by decreasing order of Favors Balance.
	 */
	private static class DecreasingReliabilityComparator
		implements Comparator<ReliabilityKey>, Serializable {
		
		public DecreasingReliabilityComparator() {
			
		}
		
		public int compare(ReliabilityKey k1,ReliabilityKey k2) {
			
			if (k1.predicted_task_outcome == k2.predicted_task_outcome) {
				if (k1.fb < k2.fb) {
					return +1;
				}
				else if (k1.fb > k2.fb) {
					return -1;
				}
				else /* if (k1.fb == k2.fb) */ {
					return (k1.i == k2.i) ? 0 : -1;
				}
			}
			else /*if (k1.predicted_task_outcome!=k2.predicted_task_outcome){*/{
				return (k1.predicted_task_outcome == TASK_SUCCESS) ? -1 : +1;
			}
	
		}
		
		static final long serialVersionUID = 7958425965755637398L;
		
	}

	// ****************************************************************

}

