/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// StorageAffinityConsumptionRanking.java

package lbg.peer.strategy.ranking;

import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.math.CommonMath.*;
import lbg.peer.register.*;
import lbg.peer.register.bartering.*;
import lbg.peer.register.learning.consumption.*;
import lbg.peer.register.negotiation.*;
import static lbg.peer.strategy.ranking.TCoRConsumptionRanking.*;
import lbg.task.*;
import java.util.*;

/** StorageAffinityConsumptionRanking ranks Grantors
 * (i.e. potential Supplier Peers who sent Consumption Grants)
 * by decreasing Storage Affinity for target Task.
 * @author Cyril Briquet
 */
public class StorageAffinityConsumptionRanking extends CommonRanking {

	/** Constructs a new StorageAffinityConsumptionRanking.
	 * <p>Note that Peers with no Storage Affinity are not ranked.
	 * @param c_g Consumption Grants container
	 * @param data_tracker data tracker
	 * @param t target Task
	 * @param jcot JobConsumptionOutcomeTracker
	 * @param do_blacklist true to activate blacklisting of
	 * suppliers estimated be unreliable at the Job-level
	 * @exception GridException if a parameter is illegal
	 */
	public StorageAffinityConsumptionRanking(ConsumptionGrantsContainer c_g,
											 ConsumptionDataTracker data_tracker,
											 Task t,
											 JobConsumptionOutcomeTracker jcot,
											 boolean do_blacklist,
											 float blacklist_proba)
		throws GridException {
		
		this(c_g,data_tracker,t,jcot,do_blacklist,blacklist_proba,false);
		
	}
	/** Constructs a new StorageAffinityConsumptionRanking.
	 * <p>Note that Peers with no Storage Affinity are not ranked.
	 * @param c_g Consumption Grants container
	 * @param data_tracker data tracker
	 * @param t target Task
	 * @param jcot JobConsumptionOutcomeTracker
	 * @param do_blacklist true to activate blacklisting of
	 * suppliers estimated be unreliable at the Job-level
	 * @exception GridException if a parameter is illegal
	 */
	public StorageAffinityConsumptionRanking(ConsumptionGrantsContainer c_g,
											 ConsumptionDataTracker data_tracker,
											 Task t,
											 JobConsumptionOutcomeTracker jcot,
											 boolean do_blacklist,
											 float blacklist_proba,
											 boolean apply_tcor)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("StorageAffinityConsumptionRanking(): "+
									"illegal Consumption Grants Container");
		}
		if (data_tracker == null) {
			throw new GridException("StorageAffinityConsumptionRanking(): "+
									"illegal Consumption data tracker");
		}
		if (t == null) {
			throw new GridException("StorageAffinityConsumptionRanking(): "+
									"illegal Task");
		}
		if (jcot == null) {
			throw new GridException("StorageAffinityConsumptionRanking(): "+
									"illegal Job Consumption outcome tracker");
		}
		
		// Rank past Suppliers Peers for Job of target Task,
		// by decreasing Storage Affinity
		Map<Long,Set<String>> sa_ranking = data_tracker.rankPeers(t);

		// Obtain pending Grantors profiles set
		Collection<String> pending_grantors_id_set = c_g.pendingGrantors();
		
		// Filters SA-ranked Peers, and keeps only those that are also Grantors
		ArrayList<String> dummy =
			new ArrayList<String>(Math.min(pending_grantors_id_set.size(),
										   sa_ranking.size()));
		int tmp_blacklisted_peers = 0;
		for (Set<String> s : sa_ranking.values()) {
			for (String sa_ranked_peer_id : s) {
				if (pending_grantors_id_set.contains(sa_ranked_peer_id)) {
					
					/* **** */
					if (do_blacklist) {
						boolean should_blacklist_job_level =
							jcot.shouldBlackListPeer(sa_ranked_peer_id);
						if (should_blacklist_job_level) {
							++tmp_blacklisted_peers;
							continue;
						}
					}
					if (apply_tcor) {
						int grantor_idx =
							c_g.pendingGrantorIndex(sa_ranked_peer_id);
						PeerProfile grantor_profile =
							c_g.grantorProfile(grantor_idx);
						PeerBarteringProfile peer_bp =
							grantor_profile.getBarteringProfile();
						double tcor =
							peer_bp.getConsumptionTasksCompletionRatio();
						boolean should_blacklist_peer_level =
							tcor < TCOR_THRESHOLD;
						if (should_blacklist_peer_level) {
							if (sampleUniformDistWithinUnity() <
								blacklist_proba) {
								++tmp_blacklisted_peers;
								continue;
							}
						}
					}
					/* **** */
					
					dummy.add(sa_ranked_peer_id);
				}
			}
		}
		int s = dummy.size();
		
		this.ranked_peers = s;
		this.unranked_peers =
			c_g.pendingGrantorsIndex().size()-this.ranked_peers;
		this.blacklisted_peers = tmp_blacklisted_peers;

		// Fill peers_ranking with SA-ranked Peers that are also Grantors
		if (s == 0) {
			this.peers_ranking = null;
			return;
		}

		int i = 0;
		this.peers_ranking = new int[s];
		for (String sa_ranked_peer_id : dummy) {
			this.peers_ranking[i++] =
				c_g.pendingGrantorIndex(sa_ranked_peer_id);
		}

	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[CommonRanking]",
							   "[StorageAffinityConsumptionRanking]"));

		return sb.toString();
		
	}

	// ****************************************************************

	/** Ranking implementation.
	 * @return Peers ranking, or null if no Peers can be ranked
	 */
	public int[] peersRanking() {
		
		return this.peers_ranking;
		
	}

	/** Ranking implementation.
	 */
	public int getRankedPeerCount() {
		
		return this.ranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getUnrankedPeerCount() {
		
		return this.unranked_peers;
		
	}

	/** Ranking implementation.
	 */
	public int getBlacklistedPeerCount() {
		
		return this.blacklisted_peers;
		
	}
	
	// ****************************************************************

	private final int[] peers_ranking;
	private final int ranked_peers;
	private final int unranked_peers;
	private final int blacklisted_peers;

}

