/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// TCaRConsumptionRanking.java

package lbg.peer.strategy.ranking;

import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.math.CommonMath.*;
import lbg.peer.register.*;
import lbg.peer.register.bartering.*;
import lbg.peer.register.learning.consumption.*;
import lbg.peer.register.negotiation.*;
import lbg.task.*;
import java.io.Serializable;
import java.util.*;

/** TCaRConsumptionRanking ranks Grantors (i.e. potential Supplier Peers
 * who sent Consumption Grants) by decreasing TCaR.
 * @author Cyril Briquet
 */
public class TCaRConsumptionRanking extends CommonRanking {

	public TCaRConsumptionRanking(ConsumptionGrantsContainer c_g,
								  JobConsumptionOutcomeTracker jcot,
								  boolean do_blacklist,float blacklist_proba)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("TCaRConsumptionRanking(): "+
									"illegal Consumption Grants Container");
		}
		if (jcot == null) {
			throw new GridException("TCoRConsumptionRanking(): "+
									"illegal Job Consumption outcome tracker");
		}
		
		// Obtain pending Grantors profiles set
		Collection<Integer> pending_grantors_index = c_g.pendingGrantorsIndex();
		Set<TCaRKey> pending_grantors_tcar_set =
			new TreeSet<TCaRKey>(TCaRConsumptionRanking.default_cmp);
		
		// Rank Grantors by decreasing TCaR-ranked reliability
		int tmp_ranked_peers = 0;
		int tmp_blacklisted_peers = 0;
		for (Integer grantor_idx : pending_grantors_index) {
			PeerProfile grantor_profile = c_g.grantorProfile(grantor_idx);
			PeerBarteringProfile peer_bp =
				grantor_profile.getBarteringProfile();
			double tcar = peer_bp.getConsumptionTasksCancellationRatio();

			if (do_blacklist) {
				boolean should_blacklist_job_level =
					jcot.shouldBlackListPeer(grantor_profile.id());
				if (should_blacklist_job_level) {
					++tmp_blacklisted_peers;
					continue;
				}
				boolean should_blacklist_peer_level = tcar > TCAR_THRESHOLD;
				if (should_blacklist_peer_level) {
					if (sampleUniformDistWithinUnity() < blacklist_proba) {
						++tmp_blacklisted_peers;
						continue;
					}
				}
			}
			
			// Sort by increasing TCaR
			TCaRKey tcar_key = new TCaRKey(tcar,grantor_idx);
			if (pending_grantors_tcar_set.add(tcar_key) == false) {
				throw new GridException("TCaRConsumptionRanking(): "+
										"corrupted data");
			}
			tmp_ranked_peers++;
		}

		this.ranked_peers = tmp_ranked_peers;
		this.unranked_peers = pending_grantors_index.size()-this.ranked_peers;	this.blacklisted_peers = tmp_blacklisted_peers;

		
		if (tmp_ranked_peers == 0) {
			this.peers_ranking = null;
			return;
		}

		// Fill peers_ranking with TCaR-ranked Grantors
		this.peers_ranking = new int[ranked_peers];
		int i = 0;
		for (TCaRKey tcar_key : pending_grantors_tcar_set) {
			this.peers_ranking[i++] = tcar_key.grantor_idx;
		}
		
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[CommonRanking]",
							   "[TCaRConsumptionRanking]"));

		return sb.toString();
		
	}

	// ****************************************************************

	/** Ranking implementation.
	 * @return Peers ranking, or null if no Peers can be ranked
	 */
	public int[] peersRanking() {
		
		return this.peers_ranking;
		
	}

	/** Ranking implementation.
	 */
	public int getRankedPeerCount() {
		
		return this.ranked_peers;
		
	}
	
	/** Ranking implementation.
	 */
	public int getUnrankedPeerCount() {
		
		return this.unranked_peers;
		
	}

	/** Ranking implementation.
	 */
	public int getBlacklistedPeerCount() {
		
		return this.blacklisted_peers;
		
	}

	public static final double TCAR_THRESHOLD = 0.80;

	// ****************************************************************

	private final int[] peers_ranking;
	private final int ranked_peers;
	private final int unranked_peers;
	private final int blacklisted_peers;

	private static final TCaRComparator default_cmp = new TCaRComparator();

	// ****************************************************************
	
	/** Helper class of TCaRConsumptionRanking.
	 */
	private static class TCaRKey {
		
		public TCaRKey(double tcar,int grantor_idx) {
			
			this.tcar = tcar;
			this.grantor_idx = grantor_idx;
			
		}
		
		public final double tcar;
		public final int grantor_idx;
		
	}
	
	// ****************************************************************
	
	/** Helper class of TCaRConsumptionRanking.
	 * Comparator of TCaRKeys (by increasing TCaR).
	 */
	private static class TCaRComparator
		implements Comparator<TCaRKey>, Serializable {
		
		public TCaRComparator() {
			
		}
		
		public int compare(TCaRKey k1,TCaRKey k2) {
				
			if (k1.tcar < k2.tcar) {
				return -1;
			}
			else if (k1.tcar > k2.tcar) {
				return +1;
			}
			else /* if (k1.tcar == k2.tcar) */ {
				if (k1.grantor_idx < k2.grantor_idx) {
					return -1;
				}
				else if (k1.grantor_idx > k2.grantor_idx) {
					return +1;
				}
				else /* if (k1.grantor_idx == k2.grantor_idx) */ {
					return 0;
				}
			}
			
		}
		
		static final long serialVersionUID = 8630173185887145247L;
		
	}

	// ****************************************************************

}

