/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// CommonConsumptionScheduling.java

package lbg.peer.strategy.scheduling.consumption;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.*;
import lbg.peer.core.qms.*;
import lbg.peer.core.qms.queue.*;
import lbg.peer.core.scheduler.*;
import lbg.peer.register.negotiation.*;
import lbg.peer.service.*;
import lbg.peer.strategy.scheduling.*;
import lbg.peer.strategy.ranking.*;
import lbg.task.*;

/**	Reliability-ranked use of Consumption Grants.
 * @author Cyril Briquet
 */
public abstract class CommonConsumptionScheduling
	implements ConsumptionSchedulingPolicy {

	public CommonConsumptionScheduling(Peer managed_peer,Scheduler scheduler)
		throws GridException {
	
		if (managed_peer == null) {
			throw new GridException("CommonConsumptionScheduling(): "+
									"illegal Peer");
		}
		if (scheduler == null) {
			throw new GridException("CommonConsumptionScheduling(): "+
									"illegal Scheduler");
		}
		
		this.managed_peer = managed_peer;
		this.scheduler = scheduler;
		
	}

	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[CommonConsumptionScheduling]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** ConsumptionSchedulingPolicy implementation.
	 * <p>Note: this implementation requires that getLocalQueueCursor()
	 * returns a non-null reference to an existing Local Queue Cursor.
	 * @exception GridException if an error occurs
	 * while using Consumption Grants to schedule Consumption Tasks
	 * @return number of successfully scheduled Consumption Tasks
	 */
	public synchronized int scheduleConsumptionTasks() throws GridException {
		
		// Extract Peers which sent Consumption grants
		ConsumptionGrantsContainer c_g =
			this.managed_peer.register().getNeighborhoodNegotiationProfile().
				getReceivedConsumptionGrants();

		if (c_g == null) return 0;
		
		LocalQueueCursor lqc = getLocalQueueCursor();
		if (lqc == null) {
			throw new GridException("scheduleConsumptionTasks(): "+
									"null Local Queue Cursor");
		}
		lqc.clear();

		//debugging code -- System.out.println("CommonConsumptionScheduling before loop");

		Measurement m_blacklisted_suppliers = MeasurementFactory.
			createSummarizedMeasurement("blacklisted suppliers");
		// Consume available capacity between Peers
		while (this.managed_peer.qms().hasWaitingLocalTasks() &&
			   (c_g.pendingConsumptionGrants() > 0)) {
			
			// Remove next Local Task from Local Tasks queue
			RequestContainer rc = this.scheduleNextConsumptionTask(lqc);
			
			// Exit from loop if all unscheduled Local Tasks've been enumerated
			if (rc == null) break;

			Task t = rc.getTask();
			Ranking ranking = rankSuppliers(c_g,rc);
			m_blacklisted_suppliers.
				addObservation(ranking.getBlacklistedPeerRatio());
			int[] peers_ranking = ranking.peersRanking();
			int peer_idx =
				(peers_ranking != null) ?
					peers_ranking[0] : c_g.randomPendingGrantorIndex();

			// Predict outcome -- WOULD BE DONE HERE
			
			assert(this.assertConsumptionGrantsAvailability(c_g,peer_idx));

			// Schedule extracted Task on best-ranked Peer
			ExternalPeerHandle supplier_peer = c_g.grantor(peer_idx);
			boolean is_accepted = this.managed_peer.xms().
				runLocalTaskOnSupplyingResource(t,supplier_peer);

			if (is_accepted) {
				// Use consumption grant (update Consuption Grants Container)
				c_g.useConsumptionGrants(peer_idx,1);

				// Store predicted outcome -- WOULD BE DONE HERE
				
				// Update stats
				this.managed_peer.register().getNeighborhoodNegotiationProfile().
					accountUsedConsumptionGrant(supplier_peer.id());
			}
			else {
				// Consumption Task was filtered out by Supplier:
				// fake to use (spill ...) Consumption Grants from
				// the Supplier Peer that didn't honor what has been negotiated
				c_g.spillAllConsumptionGrants(peer_idx);
				// Requeue the selected Task
				this.managed_peer.qms().requeueUnqueuedConsumptionTask(t);
			}
		}

		int scheduled_tasks = c_g.usedConsumptionGrants();
		
		if (scheduled_tasks > 0) {
			double blacklisted_supplier_ratio =
				m_blacklisted_suppliers.getMean();
			this.scheduler.
				accountSchedulerConsumptionDecision(scheduled_tasks,
													blacklisted_supplier_ratio);
		}

		return scheduled_tasks;
		
	}

	// ****************************************************************

	/** Rank Suppliers.
	 * @param c_g collected Consumption Grants container
	 * @param rc target Task container
	 * @return Suppliers Ranking
	 * @exception GridException if a problem occured while ranking Suppliers
	 */
	protected abstract Ranking
		rankSuppliers(ConsumptionGrantsContainer c_g,RequestContainer rc)
		throws GridException;

	// ****************************************************************

	/** Local Queue Cursor accessor.
	 * @return Local Queue Cursor, or null if none has been defined
	 */
	protected abstract LocalQueueCursor getLocalQueueCursor();
	
	/** Schedules first queued Local Task as a Consumption Task.
	 * @return first Consumption Task, may be null if all waiting Local Tasks
	 * are restricted to local execution
	 * @exception GridException if there is no Consumption Task to schedule
	 */
	protected RequestContainer scheduleFirstConsumptionTask()
		throws GridException {
		
		return this.managed_peer.qms().
			scheduleFirstConsumptionTask(this.managed_peer.basicConfig().
										 SCHEDULE_HEAD_JOB_ONLY);

	}
	
	/** Schedules next queued Local Task as a Consumption Task.
	 * The next Local Task is determined by following information provided
	 * by the Local Queue cursor.
	 * @param lqc Local Queue Cursor
	 * @return next Consumption Task, may be null if all waiting Local Tasks
	 * are restricted to local execution
	 * @exception GridException if a parameter is illegal
	 */
	protected RequestContainer scheduleNextConsumptionTask(LocalQueueCursor lqc)
		throws GridException {
		
		if (lqc == null) {
			throw new GridException("scheduleNextConsumptionTask(): "+
									"illegal Local Queue Cursor");
		}
		
		return this.managed_peer.qms().scheduleNextConsumptionTask(lqc);

	}

	// ****************************************************************

	protected static boolean
		assertConsumptionGrantsAvailability(ConsumptionGrantsContainer c_g,
											int peer_idx)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("checkConsumptionGrantsAvailability(): "+
									"illegal Consumption Grants Container");
		}
		
		// Check if (best-ranked) Peer still has a consumption grant
		if (c_g.pendingConsumptionGrants(peer_idx) <= 0) {
			return false;
		}
		
		return true;
		
	}
	
	// ****************************************************************

	private final Peer      managed_peer;
	private final Scheduler scheduler;

}

