/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DataConsumptionScheduling.java

package lbg.peer.strategy.scheduling.consumption;

import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.config.*;
import lbg.peer.core.qms.*;
import lbg.peer.core.qms.queue.*;
import lbg.peer.core.scheduler.*;
import lbg.peer.register.learning.consumption.*;
import lbg.peer.register.negotiation.*;
import lbg.peer.strategy.ranking.*;
import lbg.peer.strategy.scheduling.*;
import lbg.task.*;

/**	Storage Affinity-ranked use of Consumption Grants.
 * @author Cyril Briquet
 */
public class DataConsumptionScheduling extends CommonConsumptionScheduling
	implements ConsumptionSchedulingPolicy {

	public DataConsumptionScheduling(Peer managed_peer,Scheduler scheduler)
		throws GridException {
	
		this(managed_peer,scheduler,false);
		
	}
	
	DataConsumptionScheduling(Peer managed_peer,Scheduler scheduler,
							  boolean apply_tcor) throws GridException {
		
		super(managed_peer,scheduler);
		if (managed_peer == null) {
			throw new GridException("DataConsumptionScheduling(): "+
									"illegal Peer");
		}

		this.managed_peer = managed_peer;
		this.lqc = new LocalQueueCursor();
		this.apply_tcor = apply_tcor;
		
	}

	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[DataConsumptionScheduling]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** Rank Suppliers.
	 * @param c_g collected Consumption Grants container
	 * @param rc target Task container
	 * @return Suppliers Ranking
	 * @exception GridException if a problem occured while ranking Suppliers
	 */
	protected synchronized Ranking
		rankSuppliers(ConsumptionGrantsContainer c_g,RequestContainer rc)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("rankSuppliers(): illegal Consumption "+
									" Grants container");
		}
		if (rc == null) {
			throw new GridException("rankSuppliers(): illegal Task container");
		}
		
		JobConsumptionDataTracker job_data_tracker =
			rc.getJobConsumptionDataTracker();
		Task t = rc.getTask();
		JobConsumptionOutcomeTracker jcot =
			rc.getJobConsumptionOutcomeTracker();
		PeerStrategyConfiguration strategy_config =
			this.managed_peer.strategyConfig();
		boolean do_blacklist = strategy_config.DO_CONSUMPTION_BLACKLIST();
		float blacklist_proba =
			strategy_config.CONSUMPTION_BLACKLIST_PROBA();
		
		//debugging_msg_1(t,c_g);
		Ranking ranking =
			new StorageAffinityConsumptionRanking(c_g,job_data_tracker,t,
												  jcot,do_blacklist,
												  blacklist_proba,
												  this.apply_tcor);

		if (ranking.peersRanking() == null) {
			//debugging_msg_2();
			ranking =
				new StorageAffinityConsumptionRanking(c_g,this.managed_peer.
													  rms().dataTracker(),t,
													  jcot,do_blacklist,
													  blacklist_proba,
													  this.apply_tcor);
		}
		//debugging_msg_3(c_g,t,job_data_tracker,ranking);

		return ranking;
		
	}
	
	/** Local Queue Cursor accessor.
	 * @return Local Queue Cursor, or null if none has been defined
	 */
	protected synchronized LocalQueueCursor getLocalQueueCursor() {
		
		return this.lqc;
		
	}

	// ****************************************************************

	private void debugging_msg_1(Task t,ConsumptionGrantsContainer c_g)
		throws GridException {
	
		System.out.println("Consumer: will schedule Task "+t.id()+
						   " of Job "+t.jobID()+" with data "+
						   t.getInputMetaGridDataSet()[0].id);
		System.out.print("\tavailable suppliers =  ");
		java.util.Collection<String> zuppliers = c_g.pendingGrantors();
		System.out.println(zuppliers);

	}
	
	private void debugging_msg_2()
		throws GridException {
		
		System.out.println("Consumer: Storage-optimal suppliers ranking = "+
						   "no data in current Job tracker, "+
						   "now trying global tracker\n");
		
	}
	
	private void debugging_msg_3(ConsumptionGrantsContainer c_g,Task t,
							 	 JobConsumptionDataTracker job_data_tracker,
								 Ranking ranking)
		throws GridException {

		StringBuilder sb_msg = new StringBuilder();
		int[] dummy = ranking.peersRanking();
		if (dummy == null) {
			sb_msg.append("Consumer: Storage-optimal suppliers ranking = "+
						  "<empty> ");
		}
		else {
			sb_msg.append("Consumer: Storage-optimal suppliers ranking "+
						  "available = ");
			for (int i = 0 ; i < dummy.length ; i++) {
				sb_msg.append(c_g.grantor(dummy[i]).id());
				sb_msg.append(" ");
			}
		}
		sb_msg.append("\n\ttracked suppliers = { ");
		String did = t.getInputMetaGridDataSet()[0].id;
		java.util.Iterator<String> it = 
		  job_data_tracker.trackedSuppliers(did);
		if (it != null){
			while (it.hasNext()) {
				sb_msg.append(it.next());
				sb_msg.append(" ");
			}
		}
		it = this.managed_peer.rms().dataTracker().trackedSuppliers(did);
		if (it != null){
			sb_msg.append("/");
			while (it.hasNext()) {
				sb_msg.append(it.next());
				sb_msg.append(" ");
			}
		}
		sb_msg.append(" }\n");
		System.out.println(sb_msg.toString());

	}
	
	// ****************************************************************

	private final Peer managed_peer;
	private final LocalQueueCursor lqc;
	private final boolean apply_tcor;

}

