/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// RandomConsumptionScheduling.java

package lbg.peer.strategy.scheduling.consumption;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.*;
import lbg.peer.core.qms.*;
import lbg.peer.core.qms.queue.*;
import lbg.peer.core.scheduler.*;
import lbg.peer.register.negotiation.*;
import lbg.peer.service.*;
import lbg.peer.strategy.scheduling.*;
import lbg.peer.strategy.ranking.*;
import lbg.task.*;

/** Random use of Consumption Grants.
 * @author Cyril Briquet
 */
public class RandomConsumptionScheduling extends CommonConsumptionScheduling
	implements ConsumptionSchedulingPolicy {

	public RandomConsumptionScheduling(Peer managed_peer,Scheduler scheduler)
		throws GridException {
	
		super(managed_peer,scheduler);
		
		if (managed_peer == null) {
			throw new GridException("RandomConsumptionScheduling(): "+
									"illegal Peer");
		}

		this.managed_peer = managed_peer;
		
	}

	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[RandomConsumptionScheduling]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** ConsumptionSchedulingPolicy implementation.
	 * @return number of successfully scheduled Consumption Tasks
	 */
	public synchronized int scheduleConsumptionTasks()
		throws GridException {

		// Extract Peers which sent Consumption grants
		ConsumptionGrantsContainer c_g =
			this.managed_peer.register().getNeighborhoodNegotiationProfile().
				getReceivedConsumptionGrants();
		if (c_g == null) return 0;

		// Randomly consume available capacity between Peers
		int s = c_g.size();
		while (this.managed_peer.qms().hasWaitingLocalTasks() &&
			   (c_g.pendingConsumptionGrants() > 0)) {

			// First find a Consumption Grant ...
			int i = CommonMath.sampleUniformDist(0,s-1);
			while (c_g.pendingConsumptionGrants(i) <= 0) {
				i = (++i)%s;
			}
			
			// ... and, only then, remove a Task from queue and schedule it
			// (else it'd have to be requeued !)
			RequestContainer rc = super.scheduleFirstConsumptionTask();
			if (rc == null) {
				// All waiting Local Tasks restricted to Local execution
				break;
			}
			Task t = rc.getTask();
			
			// Schedule extracted Task on random Peer
			ExternalPeerHandle supplier_peer = c_g.grantor(i);
			boolean is_accepted = this.managed_peer.xms().
				runLocalTaskOnSupplyingResource(t,supplier_peer);
			
			if (is_accepted) {
				// Use Consumption Grant (update Consuption Grants Container)
				c_g.useConsumptionGrants(i,1);

				// Update stats
				this.managed_peer.register().getNeighborhoodNegotiationProfile().
					accountUsedConsumptionGrant(supplier_peer.id());
			}
			else {
				// Consumption Task was filtered out by Supplier:
				// fake to use (spill ...) Consumption Grants from
				// the Supplier Peer that didn't honor what has been negotiated
				c_g.spillAllConsumptionGrants(i);
				// Requeue the selected Task
				this.managed_peer.qms().requeueUnqueuedConsumptionTask(t);
			}

		}

		return c_g.usedConsumptionGrants();
		
	}

	// ****************************************************************

	/** Rank Suppliers.
	 * <p>Note: dummy implementation.
	 * @param c_g collected Consumption Grants container
	 * @param rc target Task container
	 * @return null
	 * @exception GridException if a problem occured while ranking Suppliers
	 */
	protected synchronized Ranking
		rankSuppliers(ConsumptionGrantsContainer c_g,RequestContainer rc)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("rankSuppliers(): illegal Consumption "+
									" Grants container");
		}
		if (rc == null) {
			throw new GridException("rankSuppliers(): illegal Task container");
		}
		
		return null;
		
	}

	/** Local Queue Cursor accessor.
	 * <p>Note: dummy implementation.
	 * @return null
	 */
	protected synchronized LocalQueueCursor getLocalQueueCursor() {
		
		return null;
		
	}

	// ****************************************************************

	private final Peer managed_peer;

}

