/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ReliableMLConsumptionScheduling.java

package lbg.peer.strategy.scheduling.consumption;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.peer.*;
import lbg.peer.core.qms.*;
import lbg.peer.core.qms.queue.*;
import lbg.peer.core.scheduler.*;
import lbg.peer.register.learning.db.TaskSchedulingContext;
import lbg.peer.register.negotiation.*;
import lbg.peer.service.*;
import lbg.peer.strategy.scheduling.*;
import lbg.peer.strategy.ranking.*;
import lbg.task.*;
import static lbg.task.TaskOutcome.*;

/**	Reliability-ranked use of Consumption Grants.
 * @author Cyril Briquet
 */
public class ReliableMLConsumptionScheduling
	extends CommonConsumptionScheduling
	implements ConsumptionSchedulingPolicy {

	public ReliableMLConsumptionScheduling(Peer managed_peer,
										   Scheduler scheduler)
		throws GridException {
	
		super(managed_peer,scheduler);

		if (managed_peer == null) {
			throw new GridException("ReliableMLConsumptionScheduling(): "+
									"illegal Peer");
		}

		this.managed_peer = managed_peer;
		this.lqc = new LocalQueueCursor();
		
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(CommonStringHandling.
					replaceTag(super.toString(),
							   "[CommonRanking]",
							   "[ReliableMLConsumptionRanking]"));

		return sb.toString();
		
	}

	// ****************************************************************

	/** ConsumptionSchedulingPolicy implementation.
	 * <p>Note: this implementation updates
	 * the target Tasks scheduling contexts.
	 * @return number of successfully scheduled Consumption Tasks
	 */
	public synchronized int scheduleConsumptionTasks() throws GridException {

		// Extract Peers which sent Consumption grants
		ConsumptionGrantsContainer c_g =
			this.managed_peer.register().getNeighborhoodNegotiationProfile().
				getReceivedConsumptionGrants();

		if (c_g == null) return 0;
		
		this.lqc.clear();

		// Randomly consume available capacity between Peers
		while (this.managed_peer.qms().hasWaitingLocalTasks() &&
			   (c_g.pendingConsumptionGrants() > 0)) {
			
			// Remove next Local Task from Local Tasks queue
			RequestContainer rc = super.scheduleNextConsumptionTask(this.lqc);

			// Exit from loop if all unscheduled Local Tasks've been enumerated
			if (rc == null) break;

			Task t = rc.getTask();
			Ranking ranking = rankSuppliers(c_g,rc);
			if (ranking.getRankedPeerCount() < 1) {
				throw new GridException("scheduleConsumptionTasks(): "+
										"inconsistent ReliableMLRanking");
			}
			int [] peers_idx = ranking.peersRanking();
			assert(peers_idx != null);
			int peer_idx = peers_idx[0];
			
			// Predict outcome
			TaskSchedulingContext tsc =
				((ReliableMLConsumptionRanking)ranking).peersPrediction()[0];
			if (tsc.predictedTaskOutcome() == TASK_FAILURE) {
				// Requeue the selected Task
				this.managed_peer.qms().requeueUnqueuedConsumptionTask(t);
				//debugging code -- System.out.println("Cannot schedule target "+"Consumption Task, predicted failure");
				continue;
			}
			
			assert(super.assertConsumptionGrantsAvailability(c_g,peer_idx));

			// Schedule extracted Task on best-ranked Peer
			ExternalPeerHandle supplier_peer = c_g.grantor(peer_idx);
			boolean is_accepted = this.managed_peer.xms().
				runLocalTaskOnSupplyingResource(t,supplier_peer);

			if (is_accepted) {
				// Use consumption grant (update Consuption Grants Container)
				c_g.useConsumptionGrants(peer_idx,1);

				// Store predicted outcome
				rc.setSchedulingContext(tsc);
				
				// Update stats
				this.managed_peer.register().getNeighborhoodNegotiationProfile().
					accountUsedConsumptionGrant(supplier_peer.id());
			}
			else {
				// Consumption Task was filtered out by Supplier:
				// fake to use (spill ...) Consumption Grants from
				// the Supplier Peer that didn't honor what has been negotiated
				c_g.spillAllConsumptionGrants(peer_idx);
				// Requeue the selected Task
				this.managed_peer.qms().requeueUnqueuedConsumptionTask(t);
			}
		}

		return c_g.usedConsumptionGrants();
		
	}

	// ****************************************************************

	/** Rank Suppliers.
	 * @param c_g collected Consumption Grants container
	 * @param rc target Task container
	 * @return Suppliers Ranking
	 * @exception GridException if a problem occured while ranking Suppliers
	 */
	protected synchronized Ranking
		rankSuppliers(ConsumptionGrantsContainer c_g,RequestContainer rc)
		throws GridException {
		
		if (c_g == null) {
			throw new GridException("rankSuppliers(): illegal Consumption "+
									" Grants container");
		}
		if (rc == null) {
			throw new GridException("rankSuppliers(): illegal Task container");
		}
		
		Task t = rc.getTask();
		Ranking ranking = new ReliableMLConsumptionRanking(c_g,t,true);

		return ranking;
		
	}
	
	/** Local Queue Cursor accessor.
	 * @return Local Queue Cursor, or null if none has been defined
	 */
	protected synchronized LocalQueueCursor getLocalQueueCursor() {
		
		return this.lqc;
		
	}

	// ****************************************************************

	private final Peer managed_peer;
	private final LocalQueueCursor lqc;

}

