/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// RelaxedFavorsSupplyingFiltering.java

package lbg.peer.strategy.scheduling.filtering;

import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.register.*;
import lbg.peer.strategy.ranking.*;
import lbg.peer.strategy.scheduling.*;
import lbg.task.Task;

/** RelaxedFavors Supplying filtering.
 * @author Cyril Briquet
 */
public class RelaxedFavorsSupplyingFiltering extends CommonSupplyingFiltering
	implements SupplyingFilteringPolicy {

	public RelaxedFavorsSupplyingFiltering(Peer managed_peer)
		throws GridException {
	
		super(managed_peer);
		
		if (managed_peer == null) {
			throw new GridException("RelaxedFavorsSupplyingFiltering(): "+
									"illegal Peer");
		}
		
		this.managed_peer = managed_peer;
		
	}

	/** SupplyingFilteringPolicy implementation.
	 */
	public synchronized boolean filterOutSupplyingTask(Task t)
		throws GridException {

		if (t == null) {
			throw new GridException("filterOutSupplyingTask(): "+
									"illegal Supplying Task");
		}

		if (isThresholdExceeded() == false) {
			return false; // can be queued, so do not filter out
		}
		else /* if (isThresholdExceeded() == true) */ {
			// Get NoF balance of consumer of target Supplying Task
			String consumer_id = t.consumerID();
			PeerProfile in_consumer =
				this.managed_peer.register().getProfile(consumer_id);
			double in_consumer_favors =
				in_consumer.getBarteringProfile().getFavorsBalance();

			// Rank waiting consumers
			PeerProfile[] waiting_consumers =
				this.managed_peer.qms().getWaitingConsumers();
			assert(waiting_consumers.length >= 1); // because threshold exceeded
			FavorsSupplyingRanking r = 
				new FavorsSupplyingRanking(waiting_consumers);
			if (r.getRankedPeerCount() < 1) {
				throw new GridException("filterOutSupplyingTask(): "+
										"inconsistent FavorsSupplyingRanking");
			}
			PeerProfile[] ranked_consumers = r.getRankedPeers();
			PeerProfile lo_consumer =
				ranked_consumers[r.getRankedPeerCount()-1];
			String lo_consumer_id = lo_consumer.id();
			int[] rankings = r.peersRanking();
			assert(rankings != null);
			double lo_consumer_favors =
				rankings[r.getRankedPeerCount()-1];
			
			// Compute if incoming Supplying Task should be filtered out
			if (in_consumer_favors > lo_consumer_favors) {
				this.managed_peer.qms().
					dequeueOldestWaitingSupplyingTaskOfConsumer(true,
																lo_consumer_id);
				return false; // can be queued, so do not filter out
				/* Note: the incoming Supplying Task and
				         the one with the lowest NoF balance
						 *cannot* have the same consumer Peer *in this case*,
						 because if they were, their NoF balance would be ==
				 */
			}
			else /* if (in_consumer_favors <= lo_consumer_favors) */ {
				return true; // filter out
			}
		}

	}
	
	// ****************************************************************

	private final Peer managed_peer;

}

