/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// StrictFavorsSupplyingFiltering.java

package lbg.peer.strategy.scheduling.filtering;

import lbg.common.exception.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.peer.*;
import lbg.peer.register.*;
import lbg.peer.strategy.ranking.*;
import lbg.peer.strategy.scheduling.*;
import lbg.task.Task;

/** StrictFavors Supplying filtering.
 * @author Cyril Briquet
 */
public class StrictFavorsSupplyingFiltering extends CommonSupplyingFiltering
	implements SupplyingFilteringPolicy {

	public StrictFavorsSupplyingFiltering(Peer managed_peer)
		throws GridException {
	
		super(managed_peer);
		
		if (managed_peer == null) {
			throw new GridException("StrictFavorsSupplyingFiltering(): "+
									"illegal Peer");
		}
		
		this.managed_peer = managed_peer;
		
	}

	/** SupplyingFilteringPolicy implementation.
	 */
	public synchronized boolean filterOutSupplyingTask(Task t) 
		throws GridException {

		if (t == null) {
			throw new GridException("filterOutSupplyingTask(): "+
									"illegal Supplying Task");
		}

		// Get NoF balance of consumer of target Supplying Task
		String consumer_id = t.consumerID();
		PeerProfile in_consumer =
			this.managed_peer.register().getProfile(consumer_id);
		double in_consumer_favors =
			in_consumer.getBarteringProfile().getFavorsBalance();

		// Rank all consumers
		PeerProfile[] all_consumers =
			this.managed_peer.qms().getAllConsumers(); /* waiting + running */
		if (all_consumers.length == 0) {
			return false;
		}
		assert(all_consumers.length >= 1);
		FavorsSupplyingRanking r = 
			new FavorsSupplyingRanking(all_consumers);
		PeerProfile[] ranked_consumers = r.getRankedPeers();
		if (r.getRankedPeerCount() < 1) {
			throw new GridException("filterOutSupplyingTask(): "+
									"inconsistent FavorsSupplyingRanking");
		}
		PeerProfile lo_consumer =
			ranked_consumers[r.getRankedPeerCount()-1];
		String lo_consumer_id = lo_consumer.id();
		int[] all_rankings = r.peersRanking();
		assert(all_rankings != null);
		double lo_consumer_favors =
			all_rankings[r.getRankedPeerCount()-1];

		// Rank running consumers
		boolean incoming_should_be_running = false;
			 // ... instead of lowest-ranked running Supplying Task
		String lo_running_consumer_id = null;
		//
		PeerProfile[] running_consumers =
			this.managed_peer.qms().getRunningConsumers();
		if (running_consumers.length == 0) {
			incoming_should_be_running = false;
		}
		else {
			assert(running_consumers.length >= 1);
			r = new FavorsSupplyingRanking(running_consumers);
			PeerProfile[] ranked_running_consumers = r.getRankedPeers();
			if (r.getRankedPeerCount() < 1) {
				throw new GridException("filterOutSupplyingTask(): "+
										"inconsistent FavorsSupplyingRanking");
			}
			/*PeerProfile lo_running_consumer =
				ranked_consumers[r.getRankedPeerCount()-1];*/
			lo_running_consumer_id = lo_consumer.id();
			int[] running_rankings = r.peersRanking();
			assert(running_rankings != null);
			double lo_running_consumer_favors =
				running_rankings[r.getRankedPeerCount()-1];
			if (in_consumer_favors > lo_running_consumer_favors) {
				incoming_should_be_running = true;
			}
			else /* if (in_consumer_favors > lo_running_consumer_favors) */ {
				incoming_should_be_running = false;
			}
		}

		// Compute if incoming Supplying Task should be filtered out
		if (isThresholdExceeded() == true) {
			if (in_consumer_favors <= lo_consumer_favors) {
				return true;
			}
			else /* if (in_consumer_favors > lo_consumer_favors) */ {
				this.managed_peer.qms().
					dequeueOldestWaitingSupplyingTaskOfConsumer(true,
																lo_consumer_id);
				if (incoming_should_be_running) {
					Task running_to_dequeue =
						identifyMostRecentRunningSupplyingTaskOfConsumer(
							lo_running_consumer_id);
					this.managed_peer.xms().
						preemptSupplyingTaskOnLocalResource(running_to_dequeue,
															DUMMY_LOG);
					this.managed_peer.schedulerController().
						activateSupplyingScheduling();
				}
				return false;
			}
		}
		else /* if (isThresholdExceeded() == false) */ {
			if (in_consumer_favors <= lo_consumer_favors) {
				return false;
			}
			else /* if (in_consumer_favors > lo_consumer_favors) */ {
				if (incoming_should_be_running) {
					Task running_to_dequeue =
						identifyMostRecentRunningSupplyingTaskOfConsumer(
							lo_running_consumer_id);
					this.managed_peer.xms().
						preemptSupplyingTaskOnLocalResource(running_to_dequeue,
															DUMMY_LOG);
					this.managed_peer.schedulerController().
						activateSupplyingScheduling();
				}
				return false;
			}
		}
		
	}


	// ****************************************************************

	private Task
		identifyMostRecentRunningSupplyingTaskOfConsumer(String consumer_id)
		throws GridException {
		
		if (consumer_id == null) {
			throw new GridException("identifyMostRecentRunningSupplyingTask"+
									"OfConsumer(): illegal consumer");
		}

		Task[] running_supplying_tasks =
			this.managed_peer.xms().listRunningSupplyingTasks();
		
		for (int i = 0 ; i < running_supplying_tasks.length ; ++i) {
			Task t = running_supplying_tasks[i];
			if (t.consumerID().equals(consumer_id)) {
				return t;
			}
		}
		
		throw new GridException("identifyMostRecentRunningSupplyingTask"+
								"OfConsumer(): cannot find target consumer");
		
	}
	
	// ****************************************************************

	private final Peer managed_peer;

}

