/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PreemptiveLocalScheduling.java

package lbg.peer.strategy.scheduling.local;

import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.core.qms.*;
import static lbg.peer.core.rms.ResourceState.*;
import lbg.peer.core.scheduler.*;
import lbg.peer.strategy.scheduling.*;
import lbg.resource.service.*;
import lbg.task.*;

/** Preemptive Local Scheduling.
 * @author Cyril Briquet
 */
public class PreemptiveLocalScheduling implements LocalSchedulingPolicy {

	public PreemptiveLocalScheduling(Peer managed_peer,Scheduler scheduler)
		throws GridException {
	
		if (managed_peer == null) {
			throw new GridException("PreemptiveLocalScheduling(): "+
									"illegal Peer");
		}
		if (scheduler == null) {
			throw new GridException("PreemptiveLocalScheduling(): "+
									"illegal Scheduler");
		}

		this.managed_peer = managed_peer;
		this.scheduler = scheduler;
	
	}

	// ****************************************************************

	/** LocalSchedulingPolicy implementation.
	 */
	public synchronized void scheduleLocalTasks() throws GridException {
		
		while (this.managed_peer.qms().hasWaitingLocalTasks()) {

			/* Algorithm:
				if (idle)                           -> select Task ;
													   locate Resource
				[] (no idle AND some BusySupplying) -> select Task ;
													   preempt ;
													   locate Resource
				[] (no idle AND no   BusySupplying) -> return
				fi;
				run();
				account();
			*/

			boolean has_idle_resource =
				this.managed_peer.rms().hasIdleResource();
			boolean has_busy_supplying_resource =
				this.managed_peer.rms().hasBusySupplyingResource();
			
			ResourceHandle rh = null;
			Task t = null;

			if (has_idle_resource == true) {
				t = selectLocalTask();
				rh = this.managed_peer.rms().locateResource(IDLE,t);
			}
			else /* if (has_idle_resource == false) */ {
				if (has_busy_supplying_resource == true) {
					t = selectLocalTask();
					this.scheduler.preemptLocalResources(t);
					rh = this.managed_peer.rms().locateResource(IDLE,t);
					if (rh == null) {
						// * with NoPreemption policy
						// * with any preemption policy:
						//   if data is too large to store on any Resource
						// * with adaptive preemption policy:
						//   if all running Supplying Tasks
						//   may have already been offered a 2nd chance
						this.managed_peer.qms().requeueUnqueuedLocalTask(t);
 						return;
					}
				}
				else /* if (has_busy_supplying_resource == false) */ {
					return;
				}
			}
			
			if (this.managed_peer.xms().runLocalTaskOnLocalResource(t,rh) ==
				false) {
				this.managed_peer.qms().requeueUnqueuedLocalTask(t);
				return;
			}

			this.scheduler.accountSchedulerLocalDecision();
		}
		
	}
	
	// ****************************************************************

	private Task selectLocalTask() throws GridException {
		
		RequestContainer rc = this.managed_peer.qms().
			scheduleFirstLocalTask(this.managed_peer.basicConfig().
								   SCHEDULE_HEAD_JOB_ONLY);
		if (rc == null) {
			throw new GridException("selectLocalTask(): "+
									"corrupted QueueManager data");
		}
		return rc.getTask();
		
	}

	// ****************************************************************

	private Peer managed_peer;
	private Scheduler scheduler;

}

