/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// AdaptivePreemption.java

package lbg.peer.strategy.scheduling.preemption;

import lbg.*;
import lbg.common.exception.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.peer.*;
import static lbg.peer.core.qms.RequestType.*;
import static lbg.peer.core.rms.ResourceState.*;
import lbg.peer.core.scheduler.*;
import lbg.peer.register.bartering.*;
import lbg.peer.service.*;
import lbg.peer.strategy.scheduling.*;
import lbg.resource.service.*;
import lbg.task.*;

/** Adaptive Preemption.
 * @author Cyril Briquet
 */
public class AdaptivePreemption implements SupplyingRunningPreemptionPolicy {

	public AdaptivePreemption(Peer managed_peer,Scheduler scheduler)
		throws GridException {
	
		if (managed_peer == null) {
			throw new GridException("AdaptivePreemption(): illegal Peer");
		}
		if (scheduler == null) {
			throw new GridException("AdaptivePreemption(): illegal Scheduler");
		}

		this.managed_peer = managed_peer;
		this.scheduler = scheduler;
		
	}

	// ****************************************************************

	/** SupplyingRunningPreemptionPolicy implementation.
	 * As a side effect, all waiting Supplying Tasks are cancelled.
	 */
	public synchronized void preemptLocalResources(Task prioritary_task)
		throws GridException {

		// First cancel waiting Supplying Tasks too far in the queue
		this.managed_peer.qms().preemptWaitingSupplyingTasks();
		
		// If there are not so many waiting Local Tasks, then preemption OK
		boolean preempt_only =
			!this.managed_peer.qms().shouldPreemptWaitingSupplyingTasks();

		 // for best results, PSufferage supposed to be activated
		 ResourceHandle rh = this.managed_peer.rms().
			locateResource(BUSY_SUPPLYING,prioritary_task);
		
		if (rh != null) {
			Task t = this.managed_peer.rms().getRunningTask(rh);
			
			boolean offer_second_chance = this.offerSecondChance(t);
			
			if (offer_second_chance == false) { // Act now
				if (this.managed_peer.xms().isSupplyingTaskGraced(t) == false) {
					if (preempt_only) { // Preempt
						// Actually preempt the Task
						this.managed_peer.xms().
							preemptSupplyingTaskOnLocalResource(t,DUMMY_LOG);
					}
					else { // Cancel
						// Actually cancel the Task
						this.managed_peer.xms().
							cancelSupplyingTaskOnLocalResource(t);
						
						// Communicate the cancellation to the Task consumer
						ExternalPeerHandle peer_handle =
							this.managed_peer.register().getHandle(t.consumerID());
						peer_handle.cancelLocalTaskOnSupplyingResource(
								this.managed_peer.id(),t);
					}
				}
			}
			else if (offer_second_chance) { // Offer a 2nd-chance delay
				boolean do_cancel = !preempt_only;
				this.managed_peer.xms().
					updateSupplyingTaskTimeOutAction(t,do_cancel,
													 this.getGracePeriod());
					// may or may not succeed, as all running Supplying Tasks
					// may have already been offered a 2nd chance
			}
			
			this.scheduler.accountSchedulerPreemptionDecision();
		}

	}
	
	// ****************************************************************

	public static long getGracePeriod() {
	
		return GRACE_PERIOD;
		
	}

	// ****************************************************************

	private boolean offerSecondChance(Task t) throws GridException {
		
		if (t == null) {
			throw new GridException("offerSecondChance(): illegal Task");
		}
		
		long start_time =
			this.managed_peer.xms().getStartTime(t,SUPPLYING_TASK);
		long now = LightweightBarteringGrid.getTime();
		long current_runtime = now-start_time;
		if (current_runtime < 0) {
			throw new GridException("offerSecondChance(): corrupted data");
		}
		
		long estimated_completion_time = this.estimateTaskCompletionTime(t);
		
		long estimated_remaining_time =
			estimated_completion_time - current_runtime;
		
		return
			(estimated_remaining_time >= 0) &&
			(estimated_remaining_time <= this.getGracePeriod());
		
	}
	
	private long estimateTaskCompletionTime(Task t) throws GridException {

		if (t == null) {
			throw new GridException("estimateTaskCompletionTime(): "+
									"illegal Task");
		}
		
		PeerBarteringProfile peer_bp = this.managed_peer.register().
			getProfile(t.consumerID()).getBarteringProfile();
		long completed_supplying_tasks =
			peer_bp.getTotalCompletedSupplyingTasks();

		long estimated_completion_time = 0;
		if (completed_supplying_tasks > 0) {
			estimated_completion_time =
				Math.round(peer_bp.estimateSupplyingTaskRuntime());
			if (estimated_completion_time < 0) {
				estimated_completion_time = 0;
			}
		}
		
		return estimated_completion_time;
		
	}
	
	// ****************************************************************

	private final Peer managed_peer;
	private final Scheduler scheduler;
	
	private static final long GRACE_PERIOD = 180; // 180;

	static {
		assert(GRACE_PERIOD > 0);
	}

}

