/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PeerWebStatusProvider.java

package lbg.peer.web;

import lbg.common.*;
import lbg.common.exception.*;
import static lbg.common.io.CommonStringHandling.getStackTrace;
import lbg.peer.*;
import lbg.task.*;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.net.Socket;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.logging.Logger;

/**
 * @author Cyril Briquet
 */
public class PeerWebStatusProvider implements Runnable {
	
	public PeerWebStatusProvider(Peer p,LinkedBlockingQueue<Socket> q)
		throws GridException {
		
		if (p == null) {
			throw new GridException("PeerWebStatusProvider(): illegal Peer");
		}
		if (q == null) {
			throw new GridException("PeerWebStatusProvider(): "+
									"illegal request queue");
		}
		
		this.p = p;
		this.thread_pool =
			new ThreadPool(this,6,true,true,"WebServer-service-daemon");
		this.q = q;
		
	}

	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PeerWebStatusProvider]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	// Runnable implementation
	public void run() {
		
		for ( ; ; ) { // request loop
			try {
				Socket s = this.q.take();
				doRequest(s);
			}
			catch (Throwable thr) {
				this.p.logger().
					info("PeerWebStatusProvider: "+getStackTrace(thr));
			}
		}
		
	}
	
	public void startService() throws GridException {
		
		this.thread_pool.startService();
		
	}

	// ****************************************************************

	private void doRequest(Socket s) throws GridException {
		
		if (s == null) {
			throw new GridException("doRequest(): illegal socket");
		}
		
		try {
			BufferedReader br = new BufferedReader(
					new InputStreamReader(s.getInputStream()));
			BufferedWriter bw = new BufferedWriter(
					new OutputStreamWriter(s.getOutputStream()));
			
			String status = this.status();
	
			bw.write(status);
			bw.flush(); // do *NOT* close bw
			br.close();
			s.close();
		}
		catch (IOException e) {
			throw new GridException("Could not open connection to "+
									s.getInetAddress().toString(),e);
		}
		
	}
	
	private String status() throws GridException {
		
		StringBuilder sb = new StringBuilder();

		int av_res = this.p.rms().countIdleResources();
		int wlt = this.p.qms().countWaitingLocalTasks();
		int wst = this.p.qms().countWaitingSupplyingTasks();
		int preq_len = this.p.qms().prequeue().size();
		int max_sched_req = ((lbg.peer.core.scheduler.LiveSchedulerController)this.p.schedulerController()).maxSize();
		int max_neg_req = ((lbg.peer.core.negotiator.LiveNegotiatorController)this.p.negotiatorController()).maxSize();
		int rlt = -1;
		int rct = -1;
		int rst = -1;
		Task[] running_supplying_tasks = null;
		synchronized (this.p.rms()) {
			rlt = this.p.xms().countRunningLocalTasks();
			rct = this.p.xms().countRunningConsumptionTasks();
			rst = this.p.xms().countRunningSupplyingTasks();
			running_supplying_tasks = this.p.xms().listRunningSupplyingTasks();
		}
		sb.append("\t").
			append(wlt).append(" WLT, ").
			append(wst).append(" WST, ").
			append(preq_len).append(" PQT, ").
			append("\n\t").
			append(max_sched_req).append(" MAX SCHED REQ, ").
			append(max_neg_req).append(" MAX NEGOTIATION REQ, ").
			append("\n\t").
			append(av_res).append(" available Resources, ").
			append("\n\t").
			append(rlt).append(" RLT, ").
			append(rct).append(" RCT, ").
			append(rst).append(" RST (suppliers = ");
		for (int i = 0 ; i < running_supplying_tasks.length ; ++i) {
			sb.append(running_supplying_tasks[i].consumerID()).append(" ");
		}
		sb.append("), ");
		sb.append("\n\t").
			append(this.p.qms().getCompletedLocalJobsCount()).
			append(" completed Lobal BoT").
			append("\n\t");
		
		double utilization = this.p.rms().getCachedUtilizationValue();
		long utilization_percent = Math.round(utilization*100);
		sb.append("utilization: ").append(utilization_percent).
			append("\n\t");
		
		return sb.toString();
		
	}
	
	// ****************************************************************

	private final Peer p;
	private final ThreadPool thread_pool;
	private final LinkedBlockingQueue<Socket> q;
	
}

