/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// BoT.java

package lbg.peer.workflow;

import lbg.common.*;
import lbg.common.exception.*;
import lbg.peer.core.qms.*;
import lbg.peer.core.qms.queue.*;
import lbg.protocol.data.*;
import lbg.task.*;
import java.util.*;

/** Bag of Tasks.
 * <p>Each Task is always in one of three states:
 * unscheduled, scheduled and completed.
 * <p>Tasks of Bag of Tasks are therefore separated into 3 sets,
 * each of which is backed by a heap.
 * <p>Scheduled and completed Tasks are stored according
 * to their nominal id. Unscheduled Tasks are stored by their scheduling id,
 * rather than by their nominal id,
 * so as to preserve Temporal Task Grouping,
 * i.e. scheduling decisions taken prior to the construction of the BoT.
 * @author Cyril Briquet
 */
/* Currently commented out:
   Statistics services to estimate remaining completion time and critical paths.
   Support for BoT graph management in order to link to other BoTs.*/
public class BoT {

	public BoT(int job_id, int bot_id, Task[] tasks, MetaGridData[] datas,
			   boolean is_supplying_task)
		throws GridException {
	
		if (job_id < 0) {
			throw new GridException("BoT(): illegal Job id");
		}

		if (bot_id < 0) {
			throw new GridException("BoT(): illegal id");
		}

		if (tasks == null) {
			throw new GridException("BoT(): illegal Tasks array");
		}
		
		if (datas == null) {
			throw new GridException("BoT(): illegal datas");
		}
		
		this.job_id = job_id;
		this.bot_id = bot_id;
		this.tasks_count = tasks.length;
		this.unique_datas = datas;
		this.nominal_to_scheduling_id = new TreeMap<Integer,Integer>();
		this.unscheduled_tasks_heap = new TreeMap<Integer,Task>();
		this.scheduled_tasks_heap = new TreeMap<Task,RequestContainer>();
		this.completed_tasks_heap = new TreeSet<Task>();

/*
		this.bot_heap = new TreeSet<BoT>();
		this.cumulated_nominal_runtime = 0;
		this.max_nominal_runtime_this_layer = 0;
		this.cp_runtime_next_layer = 0;
*/

		int sum_ids = 0;
		
		for (int i = 0 ; i < this.tasks_count ; i++) {
			if (tasks[i] == null) {
				throw new GridException("BoT(): illegal Task");
			}
			if (is_supplying_task == false) {
				int task_id = tasks[i].id();
				if ((task_id < 0) || (task_id >= tasks.length)) {
					throw new GridException("BoT(): corrupted Task id");
				}
				sum_ids += task_id;
			}
			// Store i-th Task
			Integer task_scheduling_id = Tank.newInteger(i); // PRESERVES TTG
			this.unscheduled_tasks_heap.put(task_scheduling_id,tasks[i]);
			// Store Task nominal-to-scheduling id mapping
			if (this.nominal_to_scheduling_id.
					put(tasks[i].taskIDAsKey(),task_scheduling_id) !=
				null) {
				throw new GridException("BoT(): corrupted "+
										"nominal-to-scheduling id map");
			}
/*
			this.cumulated_nominal_runtime += tasks[i].nominalRuntime();
			if (tasks[i].nominalRuntime() >
				this.max_nominal_runtime_this_layer) {
				this.max_nominal_runtime_this_layer =
					tasks[i].nominalRuntime();
			}
*/
		}

		if (is_supplying_task == false) {
			// If the Tasks id (excluding 0) don't add up
			// to the sum of the first k integers, there's a problem,
			// possibly corruption of Tasks by TemporalTasksGrouping
			int n = tasks.length-1;
			if (sum_ids != (n * (n+1) / 2)) {
				throw new GridException("BoT(): corrupted Task ids");
			}
		}
		
		assert(threeHeapsInvariant());

	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[BoT]");
		
		sb.append("\n  ->  [");
		sb.append(this.unscheduled_tasks_heap.size());
		sb.append(" unscheduled Tasks]");
		for (Task t : this.unscheduled_tasks_heap.values()) {
			sb.append("\n  ->  ");
			sb.append(t);			
		}
		sb.append("\n  ->  [");
		sb.append(this.scheduled_tasks_heap.size());
		sb.append(" scheduled Tasks]");
		for (Task t : this.scheduled_tasks_heap.keySet()) {
			sb.append("\n  ->  ");
			sb.append(t);			
		}
		sb.append("\n  ->  [");
		sb.append(this.completed_tasks_heap.size());
		sb.append(" completed Tasks]");
		for (Task t : this.completed_tasks_heap) {
			sb.append("\n  ->  ");
			sb.append(t);			
		}
		
		return sb.toString();
		
	}

	public int jobID() {
		
		return this.job_id;
		
	}

	public int id() {
		
		return this.bot_id;
		
	}

	// ****************************************************************

	public synchronized int tasksCount() {
	
		return this.tasks_count;
		
	}
	
	public synchronized Iterator<Task> unscheduledTasks() {
		
		return this.unscheduled_tasks_heap.values().iterator();
		
	}

	public synchronized boolean hasUnscheduledTasks() {
		
		return this.unscheduled_tasks_heap.size() > 0;
		
	}

	public synchronized int countUnscheduledTasks() {
		
		return this.unscheduled_tasks_heap.size();
		
	}

	public synchronized Iterator<Task> scheduledTasks() {
		
		return this.scheduled_tasks_heap.keySet().iterator();
		
	}

	public synchronized boolean hasScheduledTasks() {
		
		return this.scheduled_tasks_heap.size() > 0;
		
	}

	public synchronized int countScheduledTasks() {
		
		return this.scheduled_tasks_heap.size();
		
	}

	public synchronized Iterator<Task> completedTasks() {
		
		return this.completed_tasks_heap.iterator();
		
	}

	public synchronized boolean hasCompletedTasks() {
		
		return this.completed_tasks_heap.size() > 0;
		
	}

	public synchronized int countCompletedTasks() {

		return this.completed_tasks_heap.size();
		
	}

	public synchronized int countUncompletedTasks() {

		return
			this.unscheduled_tasks_heap.size()+this.scheduled_tasks_heap.size();
		
	}

	public synchronized boolean isCompleted() {
		
		assert(threeHeapsInvariant());

		return
			((this.unscheduled_tasks_heap.size() == 0) &&
			 (this.scheduled_tasks_heap.size() == 0));
		
	}

	// ****************************************************************

	/** Set of all data required by the BoT.
	 * @return set of all data required by the BoT
	 */
	public MetaGridData[] getUniqueMetaGridDatas() {
		
		return unique_datas.clone();
		
	}

	// ****************************************************************

	public synchronized RequestContainer selectFirstLocalTask()
		throws GridException {

		assert(threeHeapsInvariant());

		if (this.unscheduled_tasks_heap.isEmpty()) {
			throw new GridException("selectFirstLocalTask(): "+
									"all Tasks are already scheduled");
		}

		Integer task_scheduling_id = this.unscheduled_tasks_heap.firstKey();
		Task t = this.unscheduled_tasks_heap.get(task_scheduling_id);
		RequestContainer rc = new RequestContainer(t,task_scheduling_id);
		
		if (this.unscheduled_tasks_heap.remove(task_scheduling_id) == null) {
			throw new GridException("selectFirstLocalTask(): "+
									"corrupted unscheduled Tasks heap");
		}
		if (this.scheduled_tasks_heap.put(t,rc) != null) {
			throw new GridException("selectFirstLocalTask(): "+
									"corrupted scheduled Tasks heap");
		}
		
		assert(threeHeapsInvariant());

		return rc;
		
	}

	public synchronized RequestContainer selectFirstConsumptionTask()
		throws GridException {

		assert(threeHeapsInvariant());
		
		if (this.unscheduled_tasks_heap.isEmpty()) {
			throw new GridException("selectFirstConsumptionTask(): "+
									"all Tasks are already scheduled");
		}

		Integer task_scheduling_id = this.unscheduled_tasks_heap.firstKey();
		Task t = this.unscheduled_tasks_heap.get(task_scheduling_id);
		RequestContainer rc = new RequestContainer(t,task_scheduling_id);

		if (this.unscheduled_tasks_heap.remove(task_scheduling_id) == 
			null) {
			throw new GridException("selectFirstConsumptionTask(): "+
									"corrupted unscheduled Tasks heap");
		}
		if (this.scheduled_tasks_heap.put(t,rc) != null) {
			throw new GridException("selectFirstConsumptionTask(): "+
									"corrupted scheduled Tasks heap");
		}
		//debugging code -- System.out.println("[SELECT FIRST] Job "+t.jobID()+" Bot "+t.botID()+" Task "+t.id()+" TTG "+task_scheduling_id);
		
		assert(threeHeapsInvariant());
		
		return rc;
		
	}
	
	/** Selects next Consumption Task to schedule.
	 * @return next Consumption Task to schedule, or null if the Task
	 * referenced by the Local queue cursor was the less prioritary
	 * unscheduled Task
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized RequestContainer
		selectNextConsumptionTask(LocalQueueCursor lqc)
		throws GridException {
		
		if (lqc == null) {
			throw new GridException("selectNextConsumptionTask(): "+
									"illegal Local queue cursor");
		}

		assert(threeHeapsInvariant());
		
		if (this.unscheduled_tasks_heap.isEmpty()) {
			throw new GridException("selectNextConsumptionTask(): "+
									"all Tasks are already scheduled");
		}

		RequestContainer rc = null;
		Integer previous_task_id_as_key = lqc.taskSchedulingIDAsKey();
		for (Integer task_scheduling_id : this.unscheduled_tasks_heap.keySet()){
			// (skip previously enumerated Tasks)
			if (task_scheduling_id <= previous_task_id_as_key) {
				continue;
			}
			
			Task t = this.unscheduled_tasks_heap.get(task_scheduling_id);
			rc = new RequestContainer(t,task_scheduling_id);

			if (this.unscheduled_tasks_heap.remove(task_scheduling_id) == 
				null) {
				throw new GridException("selectNextConsumptionTask(): "+
										"corrupted unscheduled Tasks heap");
			}
			if (this.scheduled_tasks_heap.put(t,rc) != null) {
				throw new GridException("selectNextConsumptionTask(): "+
										"corrupted scheduled Tasks heap");
			}
			//debugging code -- System.out.println("[SELECT NEXT] Job "+t.jobID()+" Bot "+t.botID()+" Task "+t.id()+" TTG "+task_scheduling_id);
			break;
		}
		
		assert(threeHeapsInvariant());

		return rc;
		
	}

	/** Peeks at the Supplying Task that would selected for scheduling.
	 * <p>Recommendation: use this method only for waiting Supplying Tasks,
	 * as the Task is selected without update of the BoT state.
	 * @return Supplying Task that would selected for scheduling
	 */
	public synchronized Task peekFirstSupplyingTask() throws GridException {
		
		assert(threeHeapsInvariant());
		
		if (this.unscheduled_tasks_heap.isEmpty()) {
			throw new GridException("peekFirstSupplyingTask(): "+
									"all Tasks are already scheduled");
		}
		
		Integer task_scheduling_id = this.unscheduled_tasks_heap.firstKey();
		Task t = this.unscheduled_tasks_heap.get(task_scheduling_id);

		assert(threeHeapsInvariant());
		
		return t;

	}

	public synchronized RequestContainer selectFirstSupplyingTask()
		throws GridException {

		assert(threeHeapsInvariant());
		
		if (this.unscheduled_tasks_heap.isEmpty()) {
			throw new GridException("selectFirstSupplyingTask(): "+
									"all Tasks are already scheduled");
		}

		Integer task_scheduling_id = this.unscheduled_tasks_heap.firstKey();
		Task t = this.unscheduled_tasks_heap.get(task_scheduling_id);
		RequestContainer rc = new RequestContainer(t,task_scheduling_id);

		if (this.unscheduled_tasks_heap.remove(task_scheduling_id) == null) {
			throw new GridException("selectFirstSupplyingTask(): "+
									"corrupted unscheduled Tasks heap");
		}
		if (this.scheduled_tasks_heap.put(t,rc) != null) {
			throw new GridException("selectFirstSupplyingTask(): "+
									"corrupted scheduled Tasks heap");
		}
		
		assert(threeHeapsInvariant());

		return rc;
		
	}

	// ****************************************************************

	/**
	 * @param returning_t returning Task, i.e. copy of the original Task
	 * (that comes back from another Grid node)
	 * @param resync if true, resynchronizes the metadata of the original Task
	 * with the returning Task
	 */
	public synchronized RequestContainer flagTaskAsUnscheduled(Task returning_t,
															   boolean resync)
		throws GridException {
		
		assert(threeHeapsInvariant());

		RequestContainer rc =
			this.scheduled_tasks_heap.remove(returning_t);
		if (rc == null) {
			throw new GridException("flagTaskAsUnscheduled(): corrupted "+
									"scheduled Tasks heap");
		}

		Task current_t = rc.getTask();

		if (resync) {
			current_t.syncWith(returning_t);
		}

		Integer task_scheduling_id =
			this.nominal_to_scheduling_id.get(returning_t.taskIDAsKey());
		/* nasty bug -- if (this.unscheduled_tasks_heap.put(...,t)) { */
		if (this.unscheduled_tasks_heap.put(task_scheduling_id,current_t) !=
			null) {
			throw new GridException("flagTaskAsUnscheduled(): corrupted "+
									"unscheduled Tasks heap");
		}
		
		assert(threeHeapsInvariant());
		
		return rc;

	}

	// ****************************************************************

	public synchronized RequestContainer flagTaskAsCompleted(Task returning_t)
		throws GridException {
		
		assert(threeHeapsInvariant());

		RequestContainer rc =
			 this.scheduled_tasks_heap.remove(returning_t);
		if (rc == null) {
			throw new GridException("flagTaskAsCompleted(): corrupted "+
									"scheduled Tasks heap");
		}
		Task current_t = rc.getTask();
		current_t.syncWith(returning_t);
		/* nasty bug -- if (this.completed_tasks_heap.add(t) == false) { */
		if (this.completed_tasks_heap.add(current_t) == false) {
			throw new GridException("flagTaskAsCompleted(): corrupted "+
									"completed Tasks heap");
		}

		assert(threeHeapsInvariant());

		return rc;

	}
	
	// ****************************************************************
	
/*
	public synchronized long cumulatedNominalRuntime() {
		
		return this.cumulated_nominal_runtime;
		
	}

	public synchronized long criticalPathRuntime() {
		
		return this.max_nominal_runtime_this_layer+this.cp_runtime_next_layer;
		
	}
*/

	// ****************************************************************

	// TODO [WORKFLOW] BoT.*Layer() and BoT.*Runtime() methods
/*	
	public synchronized void addToNextLayer(BoT bot) throws GridException {
		
		if (bot == null) {
			throw new GridException("addToNextLayer(): illegal BoT");
		}
		
		if (this.bot_heap.add(bot) == false) {
			throw new GridException("addToNextLayer(): BoT already stored");
		}
		
		if (bot.criticalPathRuntime() > this.cp_runtime_next_layer) {
			this.cp_runtime_next_layer = bot.criticalPathRuntime();
		}
		
	}
	
	public synchronized void removeFromNextLayer(BoT bot) throws GridException {
		
		if (bot == null) {
			throw new GridException("addToNextLayer(): illegal BoT");
		}
		
		if (this.bot_heap.remove(bot) == false) {
			throw new GridException("addToNextLayer(): BoT not stored");
		}
		
		// TODO [WORKFLOW] potential performance bottleneck
		// => maybe keep BoT sorted by decreasing nominalRuntimes ?
		this.cp_runtime_next_layer = 0;
		Iterator<BoT> it = this.nextLayer();
		if (it != null) {
			while (it.hasNext()) {
				BoT dummy = it.next();
				if (dummy.criticalPathRuntime() > this.cp_runtime_next_layer) {
					this.cp_runtime_next_layer = dummy.criticalPathRuntime();
				}
			}
		}
		
	}
	
	public synchronized boolean isLastLayer() {
		
		return this.bot_heap.size() == 0;
		
	}
	
	// Returns next layer BoTs.
	public synchronized Iterator<BoT> nextLayer() {
		
		return isLastLayer() ? null : this.bot_heap.iterator();
		
	}
*/

	// ****************************************************************

	private synchronized boolean threeHeapsInvariant() {
		
		return this.tasks_count == this.unscheduled_tasks_heap.size()+
			this.scheduled_tasks_heap.size()+this.completed_tasks_heap.size();

	}

	// ****************************************************************

	private final int job_id;
	private final int bot_id;
	private final int tasks_count;
	private final Map<Integer,Integer> nominal_to_scheduling_id;
	private final TreeMap<Integer,Task> unscheduled_tasks_heap;
	private final Map<Task,RequestContainer> scheduled_tasks_heap;
	private final Set<Task> completed_tasks_heap;
	private final MetaGridData[] unique_datas;

/*	
	private Set<BoT> bot_heap;
	private long cumulated_nominal_runtime;
	private long max_nominal_runtime_this_layer;
	private long cp_runtime_next_layer;
*/

}

