/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// BoTFactory.java

package lbg.peer.workflow;

import lbg.*;
import lbg.common.exception.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.common.math.*;
import lbg.protocol.data.*;
import lbg.simulator.*;
import lbg.task.*;

import java.util.*;

/**
 * <p>Note: methods are not static because a BoTFactory may be instanciated
 * multiple times within the same VM, e.g. inside the LBG Simulator.
 * @author Cyril Briquet
 */
public class BoTFactory {

	/** Creates a new BoTFactory.
	 * @param enable_ttg true to enable TemporalTasksGrouping
	 * @exception GridException if a parameter is illegal
	 */
	public BoTFactory(boolean enable_ttg) throws GridException {
	
		this.current_id = 0;
		this.enable_ttg = enable_ttg;
		this.ttg = new TemporalTasksGrouping();
	
	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[BoTFactory]");

		return sb.toString();

	}
	
	// ****************************************************************

	/** Creates a Bag of Tasks from given JobDescriptor.
	 */
	public synchronized BoT createBoT(String consumer_peer_id,
									  int job_id, JobDescriptor jd,
									  boolean is_supplying_task)
			throws GridException {

		Task[] tasks = null;
		int bot_id = this.current_id; // It will be incremented later

		switch (LightweightBarteringGrid.getMode()) {
		case LIVE_MODE:
			if (jd == null) {
				throw new GridException("createBoT(): null JobDescriptor");
			}
			
			tasks = jd.toTasks(consumer_peer_id, job_id, bot_id);
			
			break;
		case SIM_MODE:

			if (jd == null) {
				tasks = generateSimulatedTasks(consumer_peer_id,job_id,bot_id);
			} else {
				throw new GridException("createBot(): passing non-null job"
					+ " descriptor in simulated mode");
			}

			break;
		default:
			throw new GridException("createBoT(): unsupported mode");
		}

		return createBoT(consumer_peer_id, job_id, tasks, is_supplying_task);
	}
	
	/** Creates a Bag of Tasks from given Tasks array.
	 * Mostly useful for either wrapping a Supplying Task
	 * (both Live and Simulated),
	 * or a subset of Tasks stored in a JobDescriptor
	 * if the BoT is part of a Workflow.
	 */
	public synchronized BoT createBoT(String consumer_peer_id,
									  int job_id, Task[] tasks,
									  boolean is_supplying_task)
			throws GridException {

		int bot_id = this.current_id++;

		// Gather unique MetaGridDatas. We need additional logic from TreeMap
		// (>< TreeSet)
		Map<MetaGridData,MetaGridData> mgd =
			new TreeMap<MetaGridData,MetaGridData>();
		for (Task t : tasks)
			for (MetaGridData m : t.getInputMetaGridDataSet())
				/* This will overwrite the old one but assuming we have X
				 * MetaGridDatas, Y of which are unique:
				 *  Adding only if required requires X checks and Y insertions
				 *  Adding anyway requires X insertions
				 * As the time for both insertions and checks is O(log N),
				 * We'd better not run any checks.
				 */
				mgd.put(m,m);

		// Temporal Task Grouping
		if (this.enable_ttg) {
			this.ttg.groupTasks(tasks);
		}

		return new BoT(job_id, bot_id, tasks,
			mgd.values().toArray(new MetaGridData[mgd.size()]),
			is_supplying_task);
	}

	// ****************************************************************
	
	private int current_id;
	private final boolean enable_ttg;
	private final TemporalTasksGrouping ttg;

	// ****************************************************************

	private static Task[] generateSimulatedTasks(String consumer_peer_id,
												 int job_id,int bot_id)
		throws GridException {
			
		Scenario sc = LightweightBarteringGrid.getScenario();
		int N = sc.getGridConfiguration().TASKS_PER_JOB(consumer_peer_id);
		int lo = sc.getGridConfiguration().TASK_LEN_LO(consumer_peer_id);
		int hi = sc.getGridConfiguration().TASK_LEN_HI(consumer_peer_id);
		
		//debugging code -- System.out.print("BoT ");
		boolean identical_data_for_all_bot = false;
		Task[] tasks = new Task[N];
		SimulatedGridDataFactory sim_input_grid_data_factory =
		 	new SimulatedGridDataFactory(!identical_data_for_all_bot,N);
		for (int task_id = 0 ; task_id < N ; task_id++) {
			int task_runtime = CommonMath.sampleUniformDist(lo,hi);
			tasks[task_id] =
				new Task(consumer_peer_id,job_id,bot_id,task_id,
						 DUMMY_BYTE_MATRIX,DUMMY_STRING_ARRAY,
						 DUMMY_STRING,DUMMY_OBJECT_ARRAY,
						 sim_input_grid_data_factory.
						 	createSimulatedGridDataArray(consumer_peer_id),
						 task_runtime);
		}
		//debugging code -- System.out.println();
		
		return tasks;

	}
	
}

