/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Job.java

package lbg.peer.workflow;

import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.register.learning.consumption.*;
import lbg.user.service.*;

/**
 * @author Cyril Briquet
 */
public class Job implements Comparable<Job> {

	/** Creates a new Local Job.
	 */
	public Job(int id,String owner_peer_id,
			   UserHandle uh,long submission_time,BoT tasks_root)
		throws GridException {
		
		this(id,owner_peer_id,uh,true,
			 submission_time,tasks_root);
		
	}

	/** Creates a new Supplying Job.
	 */
	public Job(int id,String owner_peer_id,long submission_time,BoT tasks_root)
		throws GridException {

		this(id,owner_peer_id,null,false,
			 submission_time,tasks_root);
		
	}
	
	/** Creates a new Job.
	 */
	private Job(int id,String owner_peer_id,
			   	UserHandle uh,boolean is_local_job,
				long submission_time,BoT tasks_root)
		throws GridException {
		
		if (id < 0) {
			throw new GridException("Job(): illegal id");
		}

		if (owner_peer_id == null) {
			throw new GridException("Job(): illegal Peer id");
		}
		
		if ((is_local_job) && (uh == null)) {
			throw new GridException("Job(): illegal User");
		}
		
		if (tasks_root == null) {
			throw new GridException("Job(): illegal BoT");
		}

		if (submission_time < 0) {
			throw new GridException("Job(): illegal submission time");
		}

		this.id = id;
		this.id_as_key = Tank.newInteger(id);
		this.owner_peer_id = owner_peer_id;
		this.uh = uh;
		this.tasks_root = tasks_root;
		this.submission_time = submission_time;
		
		this.job_consumption_outcome_tracker = (is_local_job) ?
			new JobConsumptionOutcomeTracker(this.owner_peer_id,this.id,
									  		 this.tasks_root.tasksCount()) :
			null;
		this.job_consumption_data_tracker = (is_local_job) ?
			new JobConsumptionDataTracker(this.owner_peer_id,this.id) :
			null;
		this.hashcode = Integer.MIN_VALUE;

	}

	public String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[Job");
		sb.append(this.id);
		sb.append("] submitter=");
		sb.append(this.owner_peer_id);
		sb.append(", submission time=");
		sb.append(this.submission_time);
		sb.append("\n  ->  ");
		sb.append(this.tasks_root);
		
		return sb.toString();
		
	}

	public int compareTo(Job other) {
	
		return this.id - other.id();
		
	}

	@Override
	public boolean equals(Object obj) {
		
		if (obj instanceof Job == false) {
			return false;
		}
		
		return (this.compareTo((Job)obj) == 0);
		
	}
	
	@Override
	public int hashCode() {
	
		if (this.hashcode == Integer.MIN_VALUE) {
			this.hashcode = CommonMath.generateHashCode(this.id);
		}
		return this.hashcode;
	
	}

	// ****************************************************************

	public int id() {
		
		return this.id;
		
	}
	
	public Integer idAsKey() {
		
		return this.id_as_key;
		
	}

	public String ownerID() {
	
		return this.owner_peer_id;
		
	}
	
	public UserHandle user() throws GridException {
		
		if (this.uh == null) {
			throw new GridException("user(): illegal access to Supplying Job");
		}
		return this.uh;
		
	}
	
	public BoT root() {
		
		return this.tasks_root;
		
	}

	public long submissionTime() {
		
		return this.submission_time;
		
	}
	
	public JobConsumptionOutcomeTracker consumptionOutcomeTracker()
		throws GridException {
		
		if (this.job_consumption_outcome_tracker == null) {
			throw new GridException("consumptionOutcomeTracker(): "+
									"illegal access to Supplying Job");
		}
		return this.job_consumption_outcome_tracker;
		
	}
	
	public JobConsumptionDataTracker consumptionDataTracker()
		throws GridException {
		
		if (this.job_consumption_data_tracker == null) {
			throw new GridException("consumptionDataTracker(): "+
									"illegal access to Supplying Job");
		}
		return this.job_consumption_data_tracker;
		
	}

	// ****************************************************************

	private final int id;
	private final Integer id_as_key;
	private final String owner_peer_id;
	private final UserHandle uh;
	private final BoT tasks_root;
	private final long submission_time;
	private final JobConsumptionOutcomeTracker job_consumption_outcome_tracker;
	private final JobConsumptionDataTracker job_consumption_data_tracker;
	private int hashcode;
	
}

