/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// JobDescriptor.java

package lbg.peer.workflow;

import java.io.Serializable;
import java.util.HashMap;
import java.util.TreeMap;
import java.util.Vector;
import java.util.Map.Entry;
import java.util.logging.Logger;

import lbg.common.exception.GridException;
import lbg.common.Tank;
import lbg.protocol.data.GridDataManager;
import lbg.protocol.data.GridData;
import lbg.task.Task;

/** Contains the description of all tasks to be in this Job Peer-side
 * @author Xavier Dalem
 */
public class JobDescriptor implements Serializable {

	/** Minimum size to enable BitTorrent on conversion heuristic, in bytes */
	private static long SIZE_THRESHOLD = 50*1024*1024; // 50Mi
	/** Number of occurences to enable BitTorrent. */
	private static int OCCUR_THRESHOLD = 2;
	
	//private Workflow w; // TODO [WORKFLOW]
	private Vector<TaskDescriptor> taskdescs = null;
	private TreeMap<String, byte[]> classinfo = null;
	private TreeMap<String, GridData[]> datas = null;
	public long submissionTime = -1;
	
	/* Creates an empty JobDescriptor.
	 * The descriptor can be later populated using the various add() methods.
	 *
	public JobDescriptor() throws GridException {
		this(null);
	}
	*/ // TODO [WORKFLOW]
	
	/* * Creates an empty JobDescriptor.
	 * The descriptor can be later populated using the various add() methods.
	 * @param w The Workflow
	 *
	public JobDescriptor(Workflow w) throws GridException {
		this.w = w;
	}
	*/ // TODO [WORKFLOW]
	
	/** Adds classes that will be required later.
	 * @param classnames Fully qualified (package.class) class names
	 * @param bytecodes Bytecodes for the corresponding classnames
	 */
	public void add(String[] classnames, byte[][] bytecodes)
			throws GridException {
		if (classnames == null || bytecodes == null
			|| classnames.length == 0 || classnames.length != bytecodes.length)
			throw new GridException("Cannot add inconsistent classes");
		
		if (classinfo == null)
			classinfo = new TreeMap<String, byte[]>();
		for (int i = 0; i < classnames.length; i++)
			classinfo.put(classnames[i], bytecodes[i]);
	}
	
	/** Adds datas that will be required later.
	 * This is a destructive process. If you add an option for an already added
	 * data, the old option will be overwritten.
	 * @param ids IDs of needed data's
	 * @param datas Options for the matching data's
	 */
	public void add(String[] ids, GridData[][] datas) throws GridException {
		if (datas == null)
			throw new GridException("Cannot add null datas");
		
		// Create an empty map even for empty data
		if (this.datas == null)
			this.datas = new TreeMap<String, GridData[]>();
		for (int i = 0; i < ids.length; i++)
			this.datas.put(ids[i], datas[i]);
		
		// As we could have two task descriptors using the same data's,
		// the value returned by put might legally be non-null
	}
	
	/** Adds a TaskDescriptor to to this JobDescriptor. Datas and classes
	 * this descriptor depends on should have been loaded previously.
	 * @param taskdesc The TaskDescriptor to add
	 */
	public void add(TaskDescriptor taskdesc) throws GridException {
		if (taskdesc == null)
			throw new GridException("Cannot add null descriptor");
		
		if (taskdescs == null)
			taskdescs = new Vector<TaskDescriptor>(1,1);
		
		String[] requirements = taskdesc.getClassNames();
		for (String s : requirements)
			if (classinfo.get(s) == null)
				throw new GridException(taskdesc.toString() + ": class " + s
					+ " wasn't loaded in the parent JobDescriptor");
		
		requirements = taskdesc.getDataIDs();
		for (String s : requirements)
			if (datas.get(s) == null)
				throw new GridException(taskdesc.toString() + ": data " + s
					+ " wasn't loaded in the parent JobDescriptor");
		
		taskdescs.add(taskdesc);
	}
	
	/** Convenience method to add everything needed for a TaskDescriptor at once
	 * @param taskdesc The TaskDescriptor to add
	 * @param classnames Fully qualified (package.class) class names
	 * @param bytecodes Bytecodes for the corresponding classnames
	 * @param ids IDs of needed data's
	 * @param datas Options for the mathing data's
	 */
	public void add(TaskDescriptor taskdesc, String[] classnames,
			byte[][] bytecodes, String[] ids, GridData[][] datas)
			throws GridException {
		add(classnames, bytecodes);
		add(ids, datas);
		add(taskdesc);
	}

	public void setSubmissionTime() throws GridException {
		if (this.submissionTime != -1) {
			throw new GridException("setSubmissionTime(): "+
									"submission time already set");
		}
		this.submissionTime = lbg.LightweightBarteringGrid.getTime();
	}

	public long getSubmissionTime() {
		return this.submissionTime;
	}

	/** Makes Tasks from the current description.<br>
	 * Calls to this method are always run completely (i.e. not cached) to
	 * avoid inconsistency when elements are added between 2 calls.
	 * @param consumer_peer_id ID of the Task owner
	 * @param job_id Job ID for the Tasks
	 * @param bot_id BoT ID for the Tasks
	 * @return all Tasks that could be built (null if none)
	 */
	public Task[] toTasks(String consumer_peer_id, int job_id, int bot_id)
		throws GridException {
		
		if (taskdescs == null || taskdescs.size() == 0)
			return null;
		
		int num_tasks = taskdescs.size();
		Task[] tasks = new Task[num_tasks];
		TaskDescriptor[] tds = new TaskDescriptor[num_tasks];
		tds = taskdescs.toArray(tds);
		
		for (int i = 0; i < num_tasks; i++) {
			// Get bytecodes for that task
			String[] names = tds[i].getClassNames();
			byte[][] bytecodes = new byte[names.length][];
			for (int j = 0; j < names.length; j++) {
				bytecodes[j] = classinfo.get(names[j]);
				if (bytecodes[j] == null)
					throw new GridException("toTasks(): cannot retreive "
						+ "bytecode for class " + names[j]);
			}
			
			// Get data for that task
			String[] dataIDs = tds[i].getDataIDs();
			GridData[][] tdata =
				new GridData[dataIDs.length][GridDataManager.NUM_SHARE_OPTIONS];
			for (int j = 0; j < dataIDs.length; j++) {
				tdata[j] = datas.get(dataIDs[j]);
				if (tdata[j] == null)
					throw new GridException("toTasks(): cannot retreive "
						+ "data of ID " + dataIDs[j]);
			}
			
			// TODO: set nominal runtime
			tasks[i] = new Task(consumer_peer_id, job_id, bot_id, i, bytecodes,
				names, tds[i].getMainClass(), tds[i].getParams(),
				tdata, 0x2A);
		}
		
		return tasks;

	}

	/** Convert all stored datas. The conversion depends on the configuration
	 * of the manager.
	 * @param manager The data manager
	 * @throws GridException if some data cannot be shared
	 */
	/* Let N be the number of unique data's and M be the total number of
	 * data's (SDR is then N/M), then the algorithm yields results in O(M+N)
	 * time and O(N) memory. Timeis irrelevant, as the host() operation will
	 * take a while
	 */
	public void convert(GridDataManager manager) throws GridException {

		final Logger DATALOGGER = Logger.getLogger("lbg.data");

		// On default policy, just apply it. Memory becomes O(1), time is O(N)
		if (manager.hasDefaultPolicy()) {
			// Conversion of data at JobDescriptor level is always local
			
			for (Entry<String, GridData[]> entry: datas.entrySet())
				// Overwrite old options
				datas.put(entry.getKey(), manager.host(
					manager.extract(entry.getValue()), true));
			
			return;
		}
		
		// No default policy, use an heuristic
		final HashMap<String, Integer> occurences =
			new HashMap<String, Integer>(datas.size());
		
		// Run through all occurences of data's and count them
		for (TaskDescriptor td: taskdescs) {
			for (String data_id: td.getDataIDs()) {
				Integer occur = occurences.get(data_id);
				
				if (occur == null) {
					final int seeds = manager.getSeedCount(data_id, true);
					DATALOGGER.fine("Recovered " + seeds + " seeds for data "
						+ data_id);
					
					occurences.put(data_id,
						Tank.newInteger(1+seeds));
				}
				else {
					occurences.put(data_id,
								   Tank.newInteger(occur.intValue() + 1));
				}
			}
		}
		
		// Data transfer protocol selection algorithm
		//
		// Resources are required to support all download options,
		// so that the Peer can choose which one to use
		for (Entry<String, Integer> entry: occurences.entrySet()) {
			final GridData data = manager.extract(datas.get(entry.getKey()));
			GridData[] converted = null;
			
			if (data.size < SIZE_THRESHOLD
				|| entry.getValue().intValue() < OCCUR_THRESHOLD) {
				// Use FTP, prepare BT for possible future downloads
				converted = manager.host(data, true, GridDataManager.HOST_FTP
					| GridDataManager.HOST_TORRENT_TRACK);
				DATALOGGER.finer("Selecting FTP for " + data.id);
			}
			
			else {
				// FTP is sub-optimal, use BitTorrent
				converted = manager.host(data, true,
					GridDataManager.HOST_TORRENT_TRACK);
				DATALOGGER.finer("Selecting BitTorrent for " + data.id);
			}
			
			// Insert result in map, overwriting the old one
			datas.put(entry.getKey(), converted);
		}
		// (very) simplified version of:
		// Baohua Wei, Gilles Fedak and Franck Cappello : "Scheduling
		// independent tasks sharing large data distributed with BitTorrent"
		
		DATALOGGER.fine("Done with automatic data protocol transfer selection");

	}

	/** Gets Job workflow.
	 */
	public Workflow workflow() {
		return null;
		//return w; // TODO [WORKFLOW]
	}
	
	/** String representation of the class descriptor, showing numbers of tasks
	 * and classes */
	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[JobDescriptor] describing ");
		sb.append((taskdescs != null) ? taskdescs.size() : 0); // :0)
		sb.append(" Tasks in ");
		sb.append((classinfo != null) ? classinfo.size() : 0); // :0)
		sb.append(" classes");
		
		return sb.toString();
		
	}

	static final long serialVersionUID = 8668681988721606818L;

}

