/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// JobFactory.java

package lbg.peer.workflow;

import lbg.*;
import lbg.common.exception.*;
import lbg.task.*;
import lbg.user.service.*;

/**
 * <p>Note: methods are not static because a JobFactory may be instanciated
 * multiple times within the same VM, e.g. inside the LBG Simulator.
 * @author Cyril Briquet
 */
public class JobFactory {

	/** Creates a new JobFactory.
	 * @param owner_peer_id owner Peer id
	 * @param enable_ttg true to enable TemporalTasksGrouping
	 * @exception GridException if a parameter is illegal
	 */
	public JobFactory(String owner_peer_id,boolean enable_ttg)
		throws GridException {
	
		if (owner_peer_id == null)  {
			throw new GridException("JobFactory(): "+
									"illegal Peer id");
		}

		this.owner_peer_id = owner_peer_id;
		this.current_id = 0;
		this.bf = new BoTFactory(enable_ttg);
		
	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[JobFactory] owner=");
		sb.append(this.owner_peer_id);
		sb.append(", current Job id=");
		sb.append(this.current_id);
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized Job createLocalJob(UserHandle uh,JobDescriptor jd)
		throws GridException {
		
		int job_id = this.current_id++;
		
		BoT bot = null;

		switch (LightweightBarteringGrid.getMode()) {
		case LIVE_MODE:
			bot = generateLiveBoT(job_id,jd);
			break;
		case SIM_MODE:
			bot = generateSimulatedBoT(job_id,jd);
			break;
		default:
			throw new GridException("createJob(): unsupported mode");
		}

		return new Job(job_id,this.owner_peer_id,uh,
					   LightweightBarteringGrid.getTime(),bot);
		
	}
	
	public synchronized Job createSupplyingJob(Task t)
		throws GridException {
		
		int job_id = this.current_id++;
		
		t.setSupplyingSpaceJobID(job_id); // QueueManager: ULTRA IMPORTANT
		
		BoT bot = generateSupplyingBoT(job_id,t);

		t.setSupplyingSpaceBoTID(bot.id()); // QueueManager: ULTRA IMPORTANT
		
		return new Job(job_id,t.consumerID(),
					   LightweightBarteringGrid.getTime(),bot);
		
	}

	// ****************************************************************

	private synchronized BoT generateSimulatedBoT(int job_id,JobDescriptor jd)
		throws GridException {
		
		if (jd == null) {
			throw new GridException("generateSimulatedBoT(): "+
									"illegal Job Descriptor");
		}
		
		if (jd.workflow() == null) {
			return this.bf.
				createBoT(owner_peer_id,job_id,(JobDescriptor)null,false);
		}

		throw new GridException("yet to be implemented");
		// TODO [WORKFLOW] -- decode Workflow, create simulated Tasks
		
	}
		
	// ****************************************************************

	private synchronized BoT generateLiveBoT(int job_id, JobDescriptor jd)
		throws GridException {
		if (jd == null) {
			throw new GridException("generateLiveBoT(): "+
									"illegal Job Descriptor");
		}
		// TODO [WORKFLOW] -- decode Workflow
		
		return this.bf.createBoT(owner_peer_id, job_id, jd, false);
	}
	
	private synchronized BoT generateSupplyingBoT(int job_id,Task t)
		throws GridException {
		
		if (t == null) {
			throw new GridException("generateBoT(): illegal Task");
		}

		Task[] tasks = { t }; 
		
		return this.bf.createBoT(t.consumerID(),job_id,tasks,true);
	}

	// ****************************************************************

	private final String owner_peer_id;
	private       int current_id;
	private final BoTFactory bf;

}

