/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// TaskDescriptor.java

package lbg.peer.workflow;

import lbg.common.exception.GridException;
import java.io.Serializable;

/**
 * Contains information relative to the creation of the Task : names of the
 * classes, name of the class containing the running routines, input parameters
 * and input data or ways to access it.
 * @see JobDescriptor
 * @author Xavier Dalem
 */
public class TaskDescriptor implements Serializable {

	private final String[] related_classes;
	private final String   main_class;
	private final Object[] params;
	private final String[] related_datas;
	
	/** Create a task description
	 * <p><strong>Warning:</strong> for efficiency reasons, we cannot afford to
	 * make copies of the possibly enormous parameter arrays. Make sure you
	 * don't modify it later.</p> 
	 * @param related_classes Names of all classes required for this task
	 * and whose bytecodes are kept in the parent JobDescriptor
	 * @param main_class The name of the class implementing GridApplication
	 * @param params Parameters for the task. Should at least be an empty array
	 * @param related_datas IDs of the input data's passed to the task
	 */
	public TaskDescriptor(String[] related_classes, String main_class,
		Object[] params, String[] related_datas) throws GridException {
		if (related_classes == null || related_classes.length == 0)
			throw new GridException("TaskDescriptor(): illegal related classes");
		if (main_class == null)
			throw new GridException("TaskDescriptor(): illegal main class");
		if (params == null)
			throw new GridException("TaskDescriptor(): illegal parameters");
		if (related_datas == null)
			throw new GridException("TaskDescriptor(): illegal datas");
		
		// Check for duplicates
		for (int i = 0; i < related_datas.length; i++)
			for (int j = 0; j < related_datas.length; j++) 
				if (i == j)
					continue;
				else if (related_datas[i].equals(related_datas[j]))
					throw new GridException("Duplicate data ID '"
						+ related_datas[i]
						+ "' in Task description. If you intend to use "
						+ "twice the same data, code your task to do so "
						+ "without uploading twice.");
		
		// Check main_class is part of related_classes. O(n) as related_classes
		// Isn't sorted and sorting would take O(n log n)
		boolean included = false;
		for (int i = 0; i < related_classes.length; i++)
			if (related_classes[i].equals(main_class)) {
				included = true;
				break;
			}
		if (!included)
			throw new GridException("Main class '" + main_class
				+ "' isn't included in the supplied sources");
		
		this.related_classes = related_classes;
		this.main_class = main_class;
		this.params = params;
		this.related_datas = related_datas;
	}

	/** Get all classes needed by the wannabe Task */
	public String[] getClassNames() {
		return related_classes.clone();
	}

	public String getMainClass() {
		return main_class;
	}

	public Object[] getParams() {
		return params.clone();
	}

	/** Return all data options for all data's this TaskDescriptor depends on.
	 */
	public String[] getDataIDs() {
		return related_datas.clone();
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[TaskDescriptor] for ");
		sb.append(main_class);
		sb.append(" (");
		sb.append(related_classes.length);
		sb.append(" related classes)");
		
		return sb.toString();
		
	}

	static final long serialVersionUID = 4846761527591266482L;

}

