/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// TemporalTasksGrouping.java

package lbg.peer.workflow;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.protocol.data.*;
import lbg.task.*;
import java.util.*;

/** Implementation of the TemporalTasksGrouping heuristic,
 * ordering Tasks of a Bag Of Tasks as sets of Tasks
 * sharing the same input data.
 * @author Cyril Briquet
 */
public class TemporalTasksGrouping {

	public TemporalTasksGrouping() throws GridException {

	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[TemporalTasksGrouping]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** Reorders a given set of Tasks with the TemporalTasksGrouping heuristic.
	 * <p>Precondition: only 1 data per Task
	 * <p>Note: groupTasksWithSingleData() updates the ordering
	 * inside target Tasks array.
	 * @deprecated
	 */
	@Deprecated
	public Task[] groupTasksWithSingleData(Task[] tasks)
		throws GridException {

		if ((tasks == null) || (tasks.length == 0)) {
			throw new GridException("groupTasksWithSingleData(): "+
									"illegal Tasks");
		}
		
		if (tasks.length == 1) {
			return tasks;
		}
		
		// STEP 1: count frequencies of data (or length of data-equal sequences)
		//         and associate Tasks of a data-equal sequence to its metadata
		TreeMap<String,DataEqualSequenceMetadata> freq =
			new TreeMap<String,DataEqualSequenceMetadata>();
		for (int t = 0 ; t < tasks.length ; t++) { // Iterate over Tasks
			MetaGridData[] data = tasks[t].getInputMetaGridDataSet();
			if (data.length > 1) {
				throw new GridException("groupTasksWithSingleData(): "+
										"unsupported Task data count");
			}

			boolean no_data = (data.length == 0);
			String data_id = no_data ?
				TemporalTasksGrouping.NO_DATA : data[0].id;
			DataEqualSequenceMetadata md = freq.get(data_id);
			if (md != null) {
				md.update(tasks[t]);
			}
			else {
				freq.put(data_id,
						 new DataEqualSequenceMetadata(no_data,tasks[t]));
			}
		}
		
		// STEP 2: sort data-equal sequences by decreasing length
		DataEqualSequenceMetadata[] sequences =
			new DataEqualSequenceMetadata[freq.size()];
		sequences = freq.values().toArray(sequences);
		Arrays.sort(sequences);
		
		// STEP 3: rearrange Tasks set
		//         by data-equal sequences of decreasing length
		int t = 0;
		for (int seq = 0 ; seq < sequences.length ; seq++) {
			DataEqualSequenceMetadata md = sequences[seq];
			Iterator<Task> it = md.tasks();
			while (it.hasNext()) {
				Task task = it.next();
				tasks[t++] = task;
			}
		}
		if (t != tasks.length) {
			throw new GridException("groupTasksWithSingleData(): "+
									"corrupted data");
		}
		
		assert(groupingInvariant(tasks));
		
		return tasks;

	}

	// ****************************************************************
	
	/** Reorders a given set of Tasks with the TemporalTasksGrouping heuristic.
	 * <p>Note: groupTasks() updates the ordering inside target Tasks array.
	 */
	public Task[] groupTasks(Task[] tasks) throws GridException {
	
		if ((tasks == null) || (tasks.length == 0)) {
			throw new GridException("groupTasks(): illegal Tasks");
		}
		
		if (tasks.length == 1) {
			return tasks;
		}
		
		// STEP 1: count frequencies of data (or length of data-equal sequences)
		//         and associate Tasks of a data-equal sequence to its metadata
		TreeMap<String,DataEqualSequenceMetadata> freq =
			new TreeMap<String,DataEqualSequenceMetadata>();
		for (int t = 0 ; t < tasks.length ; t++) { // Iterate over Tasks
			MetaGridData[] data = tasks[t].getInputMetaGridDataSet();

			boolean no_data = (data.length == 0);
			String inputdataset_id = no_data ?
				TemporalTasksGrouping.NO_DATA : tasks[t].getInputDataSetID();
			DataEqualSequenceMetadata md = freq.get(inputdataset_id);
			if (md != null) {
				md.update(tasks[t]);
			}
			else {
				freq.put(inputdataset_id,
						 new DataEqualSequenceMetadata(no_data,tasks[t]));
			}
		}
		
		// STEP 2: sort data-equal sequences by decreasing length
		DataEqualSequenceMetadata[] sequences =
			new DataEqualSequenceMetadata[freq.size()];
		sequences = freq.values().toArray(sequences);
		Arrays.sort(sequences);
		
		// STEP 3: rearrange Tasks array
		//         by data-equal sequences of decreasing length
		// Invariant: tasks[0..t-1] has been processed
		int t = 0;
		for (int seq = 0 ; seq < sequences.length ; seq++) {
			DataEqualSequenceMetadata md = sequences[seq];
			int count = 0;
			Iterator<Task> it = md.tasks();
			while (it.hasNext()) {
				Task task = it.next();
				// Sanity check: cmp Task to process with last processed Task
				if (count++ != 0) {
					if (areDataEqual(task.getInputMetaGridDataSet(),
									 tasks[t-1].getInputMetaGridDataSet()) ==
						false) {
						throw new GridException("groupTasks(): "+
												"not Data-Equal Tasks");
					}
				}

				tasks[t++] = task;
			}
		}
		if (t != tasks.length) {
			throw new GridException("groupTasks(): corrupted data");
		}
		
		assert(groupingInvariant(tasks));
		
		return tasks;
		
	}
	

	// ****************************************************************

	/** Data-Equality test between target Tasks data sets.
	 * @param ds1 1st target data set
	 * @param ds2 2nd target data set
	 * @exception GridException if a parameter is illegal
	 */
	public static boolean areDataEqual(MetaGridData[] ds1,MetaGridData[] ds2)
		throws GridException {
	
		if ((ds1 == null) || (ds2 == null)) {
			throw new GridException("areDataEqual(): illegal data sets");
		}
		
		if (ds1.length != ds2.length) {
			return false;
		}

		Set<String> datas1_set = new HashSet<String>(ds1.length);
		for (MetaGridData data : ds1) {
			datas1_set.add(data.id);
		}
		
		for (MetaGridData data : ds2) {
			if (datas1_set.contains(data.id) == false) {
				return false;
			}
		}
		
		return true;

	}

	/** Distance between target Tasks data sets,
	 * defined as the Storage Affinity of ds2 relative to ds1.
	 * <p>Note: distance(ds1,ds2) is usually not equal to distance(ds2,ds1)
	 */
	public static long distance(MetaGridData[] ds1,MetaGridData[] ds2)
		throws GridException {
		
		if ((ds1 == null) || (ds2 == null)) {
			throw new GridException("distance(): illegal data sets");
		}

		Set<MetaGridData> ds1_set = new HashSet<MetaGridData>(ds1.length);
		for (MetaGridData data : ds1) {
			ds1_set.add(data);
		}

		// Invariant:
		//	d contains sum of sizes of data processed in ds2 that are not in ds1
		long d = 0;
		for (MetaGridData data : ds2) {
			if (ds1_set.contains(data) == false) {
				d += data.size;
			}
		}
		
		return d;
		
	}

	// ****************************************************************

	private synchronized boolean groupingInvariant(Task[] tasks)
		throws GridException {

		if (tasks == null) {
			throw new GridException("groupingInvariant(): illegal Tasks");
		}
		
		int prv_seq_len = tasks.length+1; // anything larger than longest seq
		int cur_seq_len = 1;
		for (int k = 1 ; k < tasks.length ; k++) {
			if (areDataEqual(tasks[k-1].getInputMetaGridDataSet(),
							 tasks[k].getInputMetaGridDataSet()) == true) {
				cur_seq_len++;
			}
			else {
				// debugging -- System.out.print(cur_seq_len+" ");				
				if (cur_seq_len > prv_seq_len) {
					return false;
				}
				prv_seq_len = cur_seq_len;
				cur_seq_len = 1;
			}
		}
		
		return true;
		
	}

	// ****************************************************************

	private static final String NO_DATA = "NO_DATA_UNIQUE_TAG_42";

	// ****************************************************************

	private static class DataEqualSequenceMetadata
		implements Comparable<DataEqualSequenceMetadata> {
		
		public DataEqualSequenceMetadata(boolean no_data,
										 Task t) throws GridException {
			
			if (t == null) {
				throw new GridException("DataEqualSequenceMetadata(): "+
										"illegal task");
			}

			this.no_data = no_data;
			this.tasks_set = new TreeSet<Task>();
			this.tasks_set.add(t);
			
		}
		
		/**
		 * <p>Sort by decreasing size.
		 * <p>Note: size() is going to be 0 for Tasks without data.
		 */
		public int compareTo(DataEqualSequenceMetadata md) {
		
			int my_size    = tasks_set.size();
			int other_size = md.size();
			
			if (my_size < other_size) {
				return +1;
			}
			else if (my_size > other_size) {
				return -1;
			}
			else /* if (my_size == other_size) */ {
				return 0;
			}
			
		}
		
		@Override
		public boolean equals(Object obj) {
			
			if (obj instanceof DataEqualSequenceMetadata == false) {
				return false;
			}
			
			return (this.compareTo((DataEqualSequenceMetadata)obj) == 0);
			
		}

		@Override
		public int hashCode() {
		
			return CommonMath.generateHashCode(this.tasks_set.size());
		
		}

		public void update(Task t) throws GridException {
		
			if (t == null) {
				throw new GridException("update(): illegal task");
			}
			if (this.tasks_set.add(t) == false) {
				throw new GridException("update(): task"+
										"already in data-equal sequence");
			}
			
		}
		
		public int size() {
			
			return (this.no_data) ? 0 : this.tasks_set.size();
			
		}
		
		public Iterator<Task> tasks() {
		
			return this.tasks_set.iterator();
			

		}
		
		private final boolean no_data;
		private final TreeSet<Task> tasks_set;
		
	}

	// ****************************************************************

}

