/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// GridApplication.java

package lbg.protocol.api;

import lbg.common.exception.GridException;
import lbg.protocol.data.GridData;

/** This interface must be implemented by one class
 * in each Grid application submitted as a Task.<br><br>
 * This class must implement:<ul>
 * <li>A constructor <b>without any arguments</b></li>
 * <li>A method to receive its input parameters<li>
 * <li>A method to receive its input data files</li>
 * <li>A method to receive the id of the Supplier Peer that controls
 * the Resource running it<li>
 * <li>A method to be started</li>
 * <li>A method to export its results</li></ul>
 * These methods will always be called in that order. If your application does
 * not need one or both of the offered inputs, it should still provide the
 * methods, which will be empty. getResult() can be called more than once.
 * @author Xavier Dalem, Cyril Briquet
 */
public interface GridApplication {
	
	/** Sets the Grid application's input parameters.
	 * @param parameters An array of parameters as Objects. Can be empty, but
	 * never null.
	 * @throws GridException If the parameters are not what they should be.
	 */
	public void setParameters(Object[] parameters) throws GridException;
	
	/** Sets the Grid application's input data files.
	 * @param datas All input datas
	 * @throws GridException If the data is not what it should be.
	 */
	public void setInputData(GridData[] datas) throws GridException;
	
	/** Sets the id of the Supplier Peer that controls the Resource
	 * running the Grid application
	 * @param supplier_id Supplier Peer id
	 * @throws GridException If the Supplier Peer id is not what it should be.
	 */
	public void setSupplier(String supplier_id) throws GridException;

	/** Sets the playpen directory of the Resource
	 * where the Grid application can write and read temporary files.
	 * @param playpen_dir playpen directory
	 * @throws GridException If the playpen directory is not what it should be.
	 */
	public void setPlaypen(String playpen_dir) throws GridException;
	
	/** Gets computed results.
	 * <p>Note: if no results are produced, <em>new byte[0]</em>
	 * can be used as a legal value.
	 * @return computed results, or null if a transient error happened
	 */
	public byte[] getResult();
	
	/** Runs the Grid application.
	 * As this method cannot throw any exception, one should print his stack
	 * traces to stderr in case of error. They will be returned in the logs if
	 * the application completes (i.e. doesn't fail elsewhere). */
	public void compute();
	
}

