/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// AzureusListener.java

package lbg.protocol.data;

import org.gudy.azureus2.core3.download.*;
import org.gudy.azureus2.core3.util.DisplayFormatters;

/** An all-in singleton listener for Azureus. It will report status on
 * System.out or whatever PrintStream you happen to give him.<br>
 * This class is intended for debugging.
 * @author Xavier Dalem
 */
public class AzureusListener
	implements DownloadManagerListener
	         /*DownloadManagerTrackerListener*/ {
	private java.io.PrintStream out = System.out;
	private static AzureusListener instance = null;
	
	protected AzureusListener() {}
	
	public static synchronized AzureusListener getInstance() {
		if (instance == null)
			instance = new AzureusListener();
		return instance;
	}
	
	public synchronized void setPrintStream(java.io.PrintStream out) {
		this.out = out;
	}
	
	/** Reports stats on the download handled by the given manager */
	public void reportDownload(DownloadManager manager) {
		StringBuilder msg = new StringBuilder();
		int state = manager.getState();
		switch (state) {
			case DownloadManager.STATE_WAITING :
				msg.append("Waiting");
				break;
			case DownloadManager.STATE_ALLOCATING :
				msg.append("Allocating");
				break;
			case DownloadManager.STATE_CHECKING :
				msg.append("Checking");
				break;
			case DownloadManager.STATE_READY :
				msg.append("Ready");
				break;
			case DownloadManager.STATE_DOWNLOADING :
				msg.append("Downloading");
				break;
			case DownloadManager.STATE_FINISHING :
				msg.append("Finishing");
				break;
			case DownloadManager.STATE_SEEDING :
				msg.append("Seeding");
				break;
			case DownloadManager.STATE_STOPPED :
				msg.append("Stopped");
				break;
			case DownloadManager.STATE_ERROR :
				msg.append("Error : " + manager.getErrorDetails());
				break;
			default:
				msg.append("Unknown state " + state);
		}
		msg.append(" (" + manager.getTrackerStatus() + ") "
			+ manager.getNbSeeds() + " seeds, " + manager.getNbPeers()
			+ " peers");
		/*msg.append(", Scrape-Seeds: "+manager.getTrackerScrapeResponse().getSeeds());
		msg.append(", Scrape-Peers: "+manager.getTrackerScrapeResponse().getPeers());
		msg.append(", Scrape-health: "+manager.getHealthStatus());
		msg.append(", Scrape-status: "+manager.getTrackerScrapeResponse().getStatusString());*/
		
		DownloadManagerStats stats = manager.getStats();
		msg.append("   " + stats.getCompleted() / 10 + "."
			+ stats.getCompleted() % 10
			+ "% (" + DisplayFormatters.formatDownloaded(stats) + ")"
			+ " Down: " + DisplayFormatters.formatByteCountToKiBEtcPerSec(
				stats.getDataReceiveRate())
			+ " Up: " + DisplayFormatters.formatByteCountToKiBEtcPerSec(
				stats.getDataSendRate()));
		
		out.print(msg.toString() + "\n");
		out.flush();
	}
	
	/* DownloadManagerListener methods */
	public synchronized void stateChanged(DownloadManager manager, int state) {
		reportDownload(manager);
	}
	
	public synchronized void downloadComplete(DownloadManager manager) {
		reportDownload(manager);
	}
	
	public synchronized void completionChanged(DownloadManager manager,
		boolean bCompleted){
		reportDownload(manager);
	}
	
	public synchronized void positionChanged(DownloadManager manager,
		int oldPosition, int newPosition) {
		reportDownload(manager);
	}
	
	public synchronized void filePriorityChanged(DownloadManager manager,
		org.gudy.azureus2.core3.disk.DiskManagerFileInfo file) {
		reportDownload(manager);
	}
	
	/* DownloadManagerTrackerListener methods */
	/*public void scrapeResult(TRTrackerScraperResponse response) {
	}
	
	public void announceResult(TRTrackerAnnouncerResponse response) {
	}*/
}