/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ByteBoundedSlottedLRUCache.java

package lbg.protocol.data;

import java.util.concurrent.LinkedBlockingQueue;

/** This class implements a cache of a size bounded in number of entries.
 * 
 * @author Xavier Dalem
 *
 * @param <K> The type of the entries
 */
public class ByteBoundedSlottedLRUCache<K extends Comparable<? extends K>>
	extends SlottedLRUCache<K> {

	private final long byte_limit;
	private       long byte_count;
	
	public ByteBoundedSlottedLRUCache(int capacity,long byte_limit) {

		super(capacity);
		
		if (byte_limit < 1) {
			throw new IllegalArgumentException("ByteBoundedSlottedLRUCache(): "+
											   "illegal byte limit");
		}
		
		this.byte_limit = byte_limit;
		this.byte_count = 0;
		
	}

	@Override
	public Metadata<K> add(K entry, Metadata<K> meta) throws VetoException,
			IllegalStateException {
		
		assert(checkByteLimitInvariant());

		Metadata<K> retval = super.add(entry,meta);
		
		this.byte_count += meta.getWeight();

		assert(checkByteLimitInvariant());
		
		return retval;
		
	}

	@Override
	/** @inheritDoc
	 * Note that since the veto size cannot be known for sure, only vetoed
	 * entries currently in the cache are taken into account when checking for
	 * the size limit in bytes.
	 */
	public boolean canAdd(Metadata<K>[] entries, boolean mindVeto) {
		
		final boolean retval = super.canAdd(entries,mindVeto);
		
		long incoming_byte_count = 0;
		for (Metadata<K> entry : entries) {
			incoming_byte_count += entry.getWeight();
		}
		
		long veto_byte_count = 0;
		for (K key: veto) {
			final Metadata<K> meta;
			if ((meta = getMetadata(key)) != null)
				veto_byte_count += meta.getWeight();
		}
		
		return retval
			&& ((!mindVeto && incoming_byte_count <= this.byte_limit)
				|| (mindVeto && veto_byte_count + incoming_byte_count <= this.byte_limit));
	}

	@Override
	public boolean isFull() {
		return super.isFull() || (this.byte_count == this.byte_limit);
	}

	@Override
	public Metadata<K> remove(K entry) throws VetoException {

		assert(checkByteLimitInvariant());
		
		final Metadata<K> ret = super.remove(entry);

		this.byte_count -= ret.getWeight();

		assert(checkByteLimitInvariant());
		
		return ret;
	}
	
	private boolean checkByteLimitInvariant() {
	
		return (this.byte_count >= 0) && (this.byte_count <= this.byte_limit);
		
	}

}
