/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Cache.java

package lbg.protocol.data;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

/** An abstract cache.
 * <p>A Cache does not actually store entries. Rather, it stores identifiers for
 * these entries along with Metadata that could be useful for implementing
 * classes. In the following, we will consider that the identifiers are the
 * actual entries (and they could be, depending on what you store).<p>
 * <p>Additionnally, a Cache can have a veto on a set of entries, which means
 * that, no matter what, they will never me removed from the cache (and never
 * be candidate for removal). The veto, however, does not guarantee that the
 * entries are present in cache, just that if they were present when the veto
 * was set, or were added after it was set, they are still there.
 * Consequently, there is no bound on the veto size.</p>
 * @param <K> The type of the keys for the entries
 * 
 * @author Xavier Dalem
 */
public abstract class Cache<K extends Comparable<? extends K>> {
	/** A list of veto'ed entries */
	protected Set<K> veto;
	/** All entries know by this cache, including veto'ed ones */
	protected Map<K, Metadata<K>> entries;
	
	public Cache() {
		veto = new HashSet<K>();
		entries = new TreeMap<K, Metadata<K>>();
	}
	
	/** Sets a set of entries as non-removable.
	 * Veto'ed data do not have to be present in the cache. They could
	 * be added later.
	 * Furthermore, the veto could be larger than the cache. See the definition
	 * of the veto for more details.
	 * @param veto The new veto set (null to clear current veto)
	 */
	public final void setVeto(K[] veto) {
		this.veto.clear();
		
		if (veto != null) {
			for (K key: veto) {
				this.veto.add(key);
			}
		}
	}
	
	public final K[] getVeto(K[] a) {
		return veto.toArray(a);
	}
	
	/** Is this entry veto'ed ?
	 * 
	 * @param candidate The entry to check
	 * @return <tt>true</tt> if the entry is veto'ed, <tt>false</tt> in all
	 * other cases, including when the entry isn't even in cache.
	 */
	public final boolean hasVetoOn(K candidate) {
		return veto.contains(candidate);
	}
	
	/** Returns the number of elements in the veto.
	 * This method should be significantly faster than checking the size of
	 * the array returned by {@link #getVeto(Comparable[])}
	 */
	public final int vetoSize() {
		return veto.size();
	}
	
	/** Adds an entry to the cache.
	 * Setting meta data for an existing entry doesn't replace the existing
	 * metadata. Either {@link #getMetadata(Comparable)} and modify it or
	 * {@link #remove(Comparable)} then add it back.
	 * 
	 * @param entry The entry to add
	 * @param meta Metadata concerning this entry
	 * @return The metadata for the entry that was removed, if the cache
	 * implementation has a finite capacity and had to make room.
	 * @throws VetoException If the Veto prevents adding this entry in any way
	 */
	public abstract Metadata<K> add(K entry, Metadata<K> meta) throws VetoException;
	
	/** Checks if the cache can hold all entries.
	 * The check is based on the current veto, as this method returns true if
	 * all entries not in the veto can be stored in the remaining space.
	 * 
	 * @param entries The entries that would be added
	 * @param mindVeto <tt>true</tt> if the current veto should be taken into
	 * account. If <tt>false</tt>, it will be considered that the veto could
	 * be removed before actually inserting the data. In the latter case,
	 * adding without clearing the veto could fail even if the method returns
	 * <tt>true</tt>
	 * @return <tt>true</tt> if all entries can potentially be added.
	 */
	public abstract boolean canAdd(Metadata<K>[] entries, boolean mindVeto);
	
	/** Removes an entry from the cache.
	 * 
	 * @param entry The entry to remove
	 * @return <tt>true</tt> if the entry was in the cache
	 * @throws VetoException if the entry is veto'ed
	 */
	public Metadata<K> remove(K entry) throws VetoException {
		if (entry == null)
			throw new NullPointerException("null entry to remove");
		
		if (veto.contains(entry))
			throw new VetoException("Cannot remove veto'ed entry "+entry);
		
		return entries.remove(entry);
	}
	
	/** Selects an entry that could be removed, but doesn't remove it yet.
	 * 
	 * @return The entry that could be removed, or <tt>null</tt> if none can be
	 * removed but this is normal (e.g. because the cache has no limit
	 * or is empty)
	 * @throws VetoException If the veto prevents removal of any entry
	 */
	public abstract K selectForRemoval() throws VetoException;
	
	public abstract boolean isFull();
	
	public final boolean isEmpty() {
		return entries.size() == 0;
	}
	
	public final boolean has(K entry) {
		return entries.containsKey(entry);
	}
	
	public final Metadata<K> getMetadata(K entry) {
		return entries.get(entry);
	}
	
	/** Get the cache hit that inserting a set of entries would yield.
	 * <p>The cache hit is defined by <em>(number of candidate entries already
	 * in the cache) / (total number of candidate entries)</em>. It is thus
	 * comprised between 0 and 1 inclusive.</p>
	 * 
	 * @param entries The entries that would be inserted.
	 * @return The cache hit ratio
	 */
	public final float getCacheHit(K[] entries) {
		if (entries.length == 0)
			return 1;

		float already = 0;
		
		for (K entry: entries)
			if (this.entries.containsKey(entry))
				already++;
		
		return already/entries.length;
	}
	
	/** Returns the number of entries currently known by this cache.
	 * <p><b>This is not linked to any metric that could be used in a subclass
	 * to bound the cache.</b></p>
	 */
	public final int size() {
		return entries.size();
	}
}
