/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VetoException.java

package lbg.protocol.data;

import lbg.common.ThreadPool;
import lbg.common.exception.GridException;
import lbg.common.math.CommonMath;
import lbg.common.net.CommonNetHandling;
import static lbg.protocol.data.GridDataManager.DATALOGGER;
import static lbg.protocol.data.GridDataManager.MEK_CHECKSUM;
import static lbg.protocol.data.GridDataManager.MEK_LOCAL_FILE;
import static lbg.protocol.data.GridDataManager.MEK_HOSTED_TORRENT;
import static lbg.protocol.data.GridDataManager.MEK_ORIGINAL_GD;

import java.io.File;
import java.net.URL;

import org.gudy.azureus2.core3.config.COConfigurationManager;
import org.gudy.azureus2.core3.download.DownloadManager;
import org.gudy.azureus2.core3.download.DownloadManagerFactory;
import org.gudy.azureus2.core3.torrent.TOTorrent;
import org.gudy.azureus2.core3.torrent.TOTorrentCreator;
import org.gudy.azureus2.core3.torrent.TOTorrentException;
import org.gudy.azureus2.core3.torrent.TOTorrentFactory;
import org.gudy.azureus2.core3.tracker.host.TRHost;
import org.gudy.azureus2.core3.tracker.host.TRHostException;
import org.gudy.azureus2.core3.tracker.host.TRHostFactory;
import com.aelitis.azureus.core.AzureusCore;
import com.aelitis.azureus.core.AzureusCoreFactory;

/** An object with BT data tracking and sharing capabilities.
 * Data downloading is not directly supported by this object, but you may
 * require the service to be running in order to download BitTorrent data
 * 
 * @author Xavier Dalem
 */
public class EmbeddedBitTorrentServer {
	/* Metadata extra keys */
	private static final String MEK_AZ_TORRENT = "azureus.torrent";
	
	private static AzureusCore core = null;
	private static URL announceURL = null;
	private static String trackerHost = null;
	private static int trackerPort = 0;
	private static int dataPort = 0;
	
	/** Sets the announce URL according to other already set parameters.
	 * <p><b>Tracker host and port must have been set beforehand.</b></p>
	 * 
	 * @throws GridException
	 * @see #setTrackerHost(String)
	 * @see #setTrackerPort(int)
	 */
	private static void setAnnounceURL() throws GridException {
		if (core != null)
			return;
		
		if (trackerHost == null || trackerPort == 0)
			throw new GridException("Tracker host and port have to be set "
				+ "before calling setAnnounceURL()");
		
		URL announceURL = null;
		try {
			announceURL = new URL("http://"+trackerHost+':'+
				trackerPort+"/announce");
		}
		catch (java.net.MalformedURLException e) {
			throw new GridException("Not a valid host : http://"+
									trackerHost+':'+trackerPort+
									"/announce", e);
		}
		
		EmbeddedBitTorrentServer.announceURL = announceURL;
	}
	
	private static void setTrackerHost(String trackerHost) {
		if (core != null)
			return;
		
		assert (trackerHost != null);
		if (trackerHost.length() == 0)
			throw new IllegalArgumentException("Invalid tracker host");
		
		EmbeddedBitTorrentServer.trackerHost = trackerHost;
	}

	private static void setTrackerPort(int trackerPort) {
		if (core != null)
			return;
		
		if (!CommonNetHandling.isValidPort(trackerPort))
			throw new IllegalArgumentException("Invalid tracker port");
		
		EmbeddedBitTorrentServer.trackerPort = trackerPort;
	}
	
	private static void setDataPort(int dataPort) {
		if (core != null)
			return;
		
		if (!CommonNetHandling.isValidPort(dataPort))
			throw new IllegalArgumentException("Invalid data port");
		
		EmbeddedBitTorrentServer.dataPort = dataPort;
	}
	/** Readies the Azureus core by configuring, initializing and starting it.
	 * If the core is already started, nothing is done
	 * <p>This is a time consuming operation and has side effects, hence it
	 * should only done when needed.</p>
	 * <p><b>Warning</b>: This method has the side effect of modifying
	 * <tt>user.home</tt> system property, this in order to move azureus'
	 * config directory to cache directory (location is hardcoded as relative to
	 * home directory) and allow running multiple BitTorrent clients (think of
	 * NFS-Shared directories).</p>
	 * <p>Note that since 2007-04-28, this method should no longer be prone to
	 * race conditions due to the singleton nature of Azureus. (that's why
	 * it's static)</p>
	 */
	static private synchronized void prepareAzureus(String cacheDirectory) {
		// Might happen if two threads were waiting for monitor
		if (core != null)
			return;

		// Hacky: As config directory is hardcoded into Azurueus, let him
		// think our user directory is elsewhere
		System.setProperty("user.home", cacheDirectory);

		// Change Azureus ports
		COConfigurationManager.initialise();

		// Tricky: the announceURL doesn't actually set anything. If we
		// leaver things in state, we get decentralised torrents we can do
		// nothing with
		if (trackerHost != null) {
			COConfigurationManager.setParameter("Tracker IP", trackerHost);
			COConfigurationManager.setParameter("Tracker Port Enable",true);
			COConfigurationManager.setParameter("Tracker Port",
				trackerPort);
		}

		COConfigurationManager.setParameter("TCP.Listen.Port",
			dataPort);
		COConfigurationManager.setParameter("UDP.Listen.Port",
			dataPort);
		COConfigurationManager.setParameter("UDP.NonData.Listen.Port",
			dataPort);
		COConfigurationManager.setParameter("Tracker Public Enable",
			true);

		core = AzureusCoreFactory.create();
		core.start();

		DATALOGGER.info("Azureus service started" +
			((trackerHost == null)?"":
				(", tracker running on " + trackerHost + ":" + trackerPort)));
	}

	private String cacheDirectory;

	/** Creates a BitTorrent server.
	 * This server is backed by Azureus. For this reason, creating
	 * multiple servers is supported but won't actually create multiple
	 * services.
	 * 
	 * @param cacheDirectory
	 * @param trackerHost The BitTorrent tracker host (i.e. the local host).
	 * Set to null for no tracking possibilites.
	 * @param trackerPort The desired BitTorrent tracker port. Will be ignored
	 * if trackerHHost is null.
	 * @param btDataPort
	 * @throws GridException
	 */
	public EmbeddedBitTorrentServer(String cacheDirectory, String trackerHost,
			int trackerPort, int btDataPort) throws GridException {
		if (cacheDirectory == null)
			throw new NullPointerException("null cache directory");
		
		this.cacheDirectory = cacheDirectory;
		
		
		if (trackerHost != null) {
			EmbeddedBitTorrentServer.setTrackerHost(trackerHost);
			EmbeddedBitTorrentServer.setTrackerPort(trackerPort);
			EmbeddedBitTorrentServer.setAnnounceURL();
		}
		EmbeddedBitTorrentServer.setDataPort(btDataPort);
	}
	
	/** Starts the BitTorrent service.
	 * Has not effect if the service is already started.
	 */
	public void start() {
		prepareAzureus(cacheDirectory);
	}
	
	public void stop() {
		// TODO: Find a way to shutdown Azureus
	}
	
	/** Hosts the GridData and render it available for torrenting.
	 * The data should already be stored in a local file.
	 * You shouldn't try to track the same data twice.
	 */
	public synchronized TorrentGridData track(Metadata<String> meta)
			throws GridException {
		if (meta == null)
			throw new GridException("GridDataManager: cannot track null data");
		if (trackerHost == null)
			throw new GridException("GridDataManager was not initialised for"
				+ " tracking");
		
		start();
		
		final GridData previous = (GridData) meta.getExtra(MEK_HOSTED_TORRENT);
		if (previous != null) {
			if (! GridDataManager.equalHashes(previous.getChecksum(),
					(byte[]) meta.getExtra(MEK_CHECKSUM)))
				throw new GridException("Tracking inconsistent: tracker"
					+ " already knows data '" + meta.id
					+ "' but with a different hash");
		}
		
		// Get then location where the data was store()'d
		File file = (File) meta.getExtra(MEK_LOCAL_FILE);
		
		// Get the server
		// The singleton has been created _and_ initialised by the core
		TRHost host = TRHostFactory.getSingleton();
		TOTorrentCreator creator = null;
		try {
			creator =
				TOTorrentFactory.createFromFileOrDirWithComputedPieceLength( 
					file, announceURL);
		} catch (TOTorrentException e) {
			throw new GridException("Unable to announce the torrent for "
				+ meta.id, e);
		}
		
		// Create the torrent
		TOTorrent torrent;
		try {
			torrent = creator.create();
			// A publish torrent is a torrent Azureus tracks but doesn't have
			// data of. It could be perfectly normal for general use but we
			// don't want this
			assert(!(host.getHostTorrent(torrent) instanceof
				org.gudy.azureus2.core3.tracker.host.impl.TRHostTorrentPublishImpl));
		} catch (TOTorrentException e) {
			throw new GridException("Unable to create the torrent for "
				+ meta.id, e);
		}
		
		// Host torrent
		try {
			host.hostTorrent(torrent, true, false);
		} catch (TRHostException e) {
			throw new GridException("Unable to publish torrent for "
				+ meta.id, e);
		}
		
		// Avoid exception in org.gudy.azureus2.core3.tracker.server.impl.
		//		TRTrackerServerImpl.java:627 ("operation denied")
		host.getHostTorrent(torrent).start();
		
		// Dirty part, but seems unavoidable for Azureus
		File torrentFile = new File(file.getAbsolutePath() + "_"
			//+ new java.util.Random(new java.util.Date().getTime()).nextInt()
			+ CommonMath.sampleUniformDist()
			+ ".torrent");
		torrentFile.deleteOnExit();
		if (torrentFile.exists())
			throw new GridException("Could not save torrent file before"
				+ " sharing " + meta.id);
		try {
			torrent.serialiseToBEncodedFile(torrentFile);
		} catch (TOTorrentException e) {
			throw new GridException("Could not save torrent file before"
				+ " sharing " + meta.id, e);
		}
		
		// Tricky part : We created a tracker and "hosted" a file
		// Or did we ? In fact, hosting is just announcing so we have to
		// create a download manager that will seed the file (up to now, the
		// tracker is complete but we should provide the ressources too)
		DownloadManager dManager = DownloadManagerFactory.create(
			core.getGlobalManager(), null, torrentFile.getAbsolutePath(),
			file.getAbsolutePath(), DownloadManager.STATE_WAITING, false,
			false, null, null );
		dManager.initialize();
		// DEBUG: dManager.addListener( lbg.protocol.AzureusListener.getInstance());
		
		int state = 0;
		while ((state = dManager.getState()) != DownloadManager.STATE_READY) {
			if (state == DownloadManager.STATE_ERROR)
				throw new GridException("Torrenting error : "
					+ dManager.getErrorDetails());
			
			// Loop rather than register myself as a Listener: it would be
			// too much work for nothing (we just wait for the seeder to
			// hash the file)
			ThreadPool.nap(100);
		}
		dManager.startDownload();
		
		// Now the torrent will be shared as long as the peer runs
		
		
		TorrentGridData tgd = null;
		try {
			tgd = new TorrentGridData(meta.id, meta.getWeight(),
				torrent.serialiseToMap(), (byte[]) meta.getExtra(MEK_CHECKSUM),
				dManager);
		} catch (TOTorrentException e) {
			throw new GridException("Could not serialise torrent to create"
				+ " a TorrentGridData", e);
		}
		
		meta.setExtra(MEK_AZ_TORRENT, torrent);
		
		return tgd;
	}
	
	/** Removes all mention of an entry in BitTorrent structures.
	 * 
	 * @param meta The entry to remove
	 * @throws GridException If a step in removal failed.
	 */
	// Warning: As this relies heavily on Azureus, I cannot be sure
	// every data has been cleaned. Did my best, though.
	public synchronized void removeFromBitTorrent(Metadata<String> meta)
			throws GridException {
		if (meta == null)
			throw new GridException("Cannot remove null entry from BitTorrent");
		
		GridData gd = (GridData) meta.getExtra(MEK_ORIGINAL_GD);
		if (! (gd instanceof TorrentGridData))
			return;
		
		DownloadManager dm;
		// Remove downloader
		dm = ((TorrentGridData)gd).getDownloadManager();
		dm.stopIt(DownloadManager.STATE_STOPPED, true, false);
		try {
			core.getGlobalManager().removeDownloadManager(dm);
		} catch (org.gudy.azureus2.core3.global.
				GlobalManagerDownloadRemovalVetoException e) {
			throw new GridException("Failed to remove data downloader");
		}
		
		TOTorrent torrent = (TOTorrent) meta.getExtra(MEK_AZ_TORRENT);
		
		// Remove torrent from tracking
		if (torrent != null) {
			// Stop tracker download. Might be the same as the first download
			dm = core.getGlobalManager().getDownloadManager(torrent);
			dm.stopIt(DownloadManager.STATE_STOPPED, true, false);
			try {
				core.getGlobalManager().removeDownloadManager(dm);
			} catch (org.gudy.azureus2.core3.global.
					GlobalManagerDownloadRemovalVetoException e) {
				throw new GridException("Failed to remove data seeder");
			}
			
			TRHost host = TRHostFactory.getSingleton();
			host.getHostTorrent(torrent).stop();
			try {
				host.getHostTorrent(torrent).remove();
			} catch (org.gudy.azureus2.core3.tracker.host.
					TRHostTorrentRemovalVetoException e) {
				throw new GridException("Failed to remove data from tracker");
			}
		}
	}
	
	public String toString() {
		return new StringBuilder("BitTorrent Server (Azureus-based) [ announce:")
			.append(announceURL).append(" , tracker host:").append(trackerHost)
			.append(" , tracker port:").append(trackerPort)
			.append(" , data port:").append(dataPort)
			.append(" , tracking active:").append(core != null)
			.append(" ]").toString();
	}
}
