/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// EmbeddedFTPServer.java

package lbg.protocol.data;

import lbg.common.exception.GridException;
import static lbg.protocol.data.GridDataManager.DATALOGGER;

import java.io.File;
import java.util.Properties;
import java.util.logging.Level;

import org.apache.ftpserver.FtpConfigImpl;
import org.apache.ftpserver.FtpServer;
import org.apache.ftpserver.config.PropertiesConfiguration;
import org.apache.ftpserver.usermanager.BaseUser;

/** An FTP server with minimal functionnalities.
 * Its main purpose is to be easily used (i.e. not too much hassle).
 * @author Xavier Dalem
 */
public class EmbeddedFTPServer {
	/* Static fields and methods */
	/** Path to default Apache FTPServer properties file */
	private static final String DEFAULT_FTPSERVER_CONFIG =
		"/lbg/protocol/data/ftpserver_conf.properties";
	
	/** Default length for random Strings.
	 * @todo replace this with bounds.
	 */
	private static final int DEFAULT_RANDOM_LENGTH = 15;

	private static final java.util.Random rand = new java.util.Random(
		new java.util.Date().getTime());

	/** Get a String of random characters
	 * @param length The length of the random string
	 */
	private static String getRandomString(int length) {
		char[] alphabet = new char[] {
			'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
			'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
			'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
			'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
			'0', '1', '2', '3', '4', '5', '6', '7', '8', '9'
			/*' ', '&', '|', '@', '"', '#', '\'','(', ')', '[', ']', '{', '}',
		'!', '*', '%', '<', '>', '\\',',', '?', ';', '.', ':', '/',	'=',
		'+'*/
			/* Accentuated characters and currencies fail in ftp. Manual tests
			 * seem to work but as it already failed, keep safe*/
		};

		StringBuilder sb = new StringBuilder(length);

		for (int i = 0; i < length; i++)
			sb.append(alphabet[rand.nextInt(alphabet.length)]);

		return sb.toString();
	}
	
	/* Instance fields and methods */
	private final String hostname;
	private final String sharedDirectory;
	private FtpServer engine = null;
	private String user = null, password = null;
	/** Transfer rate for FTP: Unlimited */
	private int downRate = -1, upRate = -1;
	/** Requests port for ftp */
	private int port;
	private Properties config = null;
	
	public EmbeddedFTPServer(String hostname, String sharedDirectory) {
		if (hostname == null)
			throw new NullPointerException("null hostname");
		if (hostname.length() == 0)
			throw new IllegalArgumentException("Empty hostname");
		if (hostname == null)
			throw new NullPointerException("null shared directory");
		
		this.hostname = hostname;
		this.sharedDirectory = sharedDirectory;
	}
	
	/* package */ String getUser() {
		return user;
	}
	
	/* package */ String getPassword() {
		return password;
	}
	
	/* package */ int getPort() {
		return port;
	}

	/** Sets the FTP server configuration.<br>
	 * If a server is already running, it will need to be restarted for the
	 * changes to take effect. If no user information is given, it is
	 * either recovered from what was previously set or generated randomly.
	 * @param port FTP Requests port
	 * @see #start()
	 * @see #stop()
	 * @throws GridException if the parameters are invalid
	 */
	public void setConfig(int port) throws GridException {
		setConfig(port, (user != null)?user:
			getRandomString(DEFAULT_RANDOM_LENGTH),
			(password != null)?password:
				getRandomString(DEFAULT_RANDOM_LENGTH));
	}
	
	/** Sets the FTP server configuration.<br>
	 * If a server is already running, it will need to be restarted for the
	 * changes to take effect.
	 * @param port FTP Requests port
	 * @param user User name
	 * @param password Associated password
	 * @see #start()
	 * @see #stop()
	 * @throws GridException if the parameters are invalid
	 */
	public void setConfig(int port, String user,
			String password) throws GridException {
		if (user == null || user.equals("")
			|| password == null || password.equals("") || port < 1)
			throw new GridException("setFTPConfig: invalid parameters");
		
		this.port = port;
		this.user = user;
		this.password = password;
		
		//DEBUG: System.err.println("user: " + ftp_user + "\npass: " + ftp_password);

		java.io.InputStream res =
			GridDataManager.class.getResourceAsStream(DEFAULT_FTPSERVER_CONFIG);
		//http://www.javaworld.com/javaworld/javaqa/2003-08/01-qa-0808-property.html?page=2
		//http://mindprod.com/jgloss/getresourceasstream.html

		if (res == null)
			throw new GridException("Could not start FTP Server: default "
				+ "config not found");

		// Repeated instanciation for multiple levels of defaults.
		config = new Properties();
		try {
			config.load(res);
			// Set custom config. /me <3 implicit cast from int to String
			config.setProperty("config.socket-factory.address",hostname);
			config.setProperty("config.socket-factory.port", ""
				+ port);
			config.setProperty("config.user-manager.prop-file",
				sharedDirectory + File.separator + ".ftpserver_user.gen");
		} catch (java.io.IOException e) {
			throw new GridException("Could not start FTP Server: failed in "
				+ "loading config", e);
		}
	}
	
	/** Load an FTP Server and start it.<br>
	 * If a server is already running, do nothing. If you want to restart
	 * a server, stop it then call this method.<br>
	 * The config of the server is based on the following properties, used
	 * in this order (the laters thus replacing props from the formers)
	 * <ol>
	 * <li>Apache FtpServer built-in defaults</li>
	 * <li>LBG Custom defaults. Quick security patch (no default users) and
	 * default ports.</li>
	 * <li>The set that was previously given with a call to setFTPConfig()</li>
	 * </ol>
	 * @see #setConfig(int)
	 * @see #setConfig(int, String, String)
	 * @see #stop()
	 */
	public void start() throws GridException {
		if (engine != null)
			return;
		
		if (config == null)
			throw new GridException("FTP server was not configured before "
				+ "being started");
		
		try {
			engine = new RestrictedFtpServer(
				new FtpConfigImpl(
					new PropertiesConfiguration(config)));
			engine.start();
		} catch (Exception e) {
			throw new GridException("Could not start FTP Server: failed to "
				+ "instanciate and start Apache's server on " + hostname
				+ ":" + port , e);
		}
		
		// Set user and password.
		BaseUser bUser = new BaseUser();
		bUser.setName(user);
		bUser.setPassword(password);
		bUser.setEnabled(true);
		bUser.setHomeDirectory(sharedDirectory);
		bUser.setWritePermission(false);
		bUser.setMaxDownloadRate(downRate);
		bUser.setMaxUploadRate(upRate);
		try {
			((FtpConfigImpl) engine.getFtpConfig()).getUserManager()
				.save(bUser);
		} catch (org.apache.ftpserver.ftplet.FtpException e) {
			throw new GridException("Could not add random user", e);
		}
		DATALOGGER.info("Started FTP server at " + hostname + ":"
			+ port);
	}
	
	/** Stops the associated FTP Server if it's running.<br>
	 * All reference to it are lost, except the given properties if any.<br>
	 * Does nothing if no server is active.
	 * @see #setConfig(int)
	 * @see #setConfig(int, String, String)
	 */
	public void stop() {
		if (engine != null) {
			try {
				((FtpConfigImpl) engine.getFtpConfig()).getUserManager()
					.delete(user);
			} catch (org.apache.ftpserver.ftplet.FtpException e) {
				DATALOGGER.log(Level.WARNING,
					"Could not remove user before stopping", e);
			}
			// there's not much we can do about that

			engine.stop();
			DATALOGGER.info("Stopped FTP server");
		}
		engine = null;
	}
	
	/** Sets new user login and password for FTP Server.<br>
	 * The server needs to be restarted for this change to take effect. This
	 * is tedious in order to remember that all users currently relying
	 * on the running server are doomed.
	 * @throws GridException If the former user couldn't be deleted.
	 */
	public void forceNewUser() throws GridException {
		// Remove references to the old user first
		if (engine != null && user != null) {
			try {
				((FtpConfigImpl) engine.getFtpConfig()).getUserManager()
					.delete(user);
			} catch (org.apache.ftpserver.ftplet.FtpException e) {
				throw new GridException("Could not delete former user", e);
			}
		}
		user = getRandomString(DEFAULT_RANDOM_LENGTH);
		password = getRandomString(DEFAULT_RANDOM_LENGTH);
	}
	
	public String toString() {
		return new StringBuilder("FTP Server (Apache-based) [ host:")
			.append(hostname).append(" , port:").append(port)
			.append(" , root:").append(sharedDirectory).append(" ]")
			.toString();
	}
	
	public static void main(String[] argv) throws Exception {
		final int port;
		try {
			if (argv.length != 2)
				throw new Exception();
			port = Integer.parseInt(argv[1]);
		} catch (Exception e) {
			System.err.println("EmbeddedFTPServer HOST PORT");
			return;
		}
		
		System.err.println("LBG EmbeddedFTPServer tester: don't use this in"
				+ " production\n");
		EmbeddedFTPServer server = new EmbeddedFTPServer(argv[0],
			System.getProperty("user.dir"));
		server.setConfig(port);
		
		server.start();
		
		System.err.println(server);
		System.err.println("You can login with "+server.user+":"+server.password);
	}
}
