/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// FTPGridData.java
package lbg.protocol.data;

import java.io.File;
import java.net.InetAddress;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.enterprisedt.net.ftp.FTPClient;
import com.enterprisedt.net.ftp.FTPConnectMode;
import com.enterprisedt.net.ftp.FTPTransferType;

import lbg.common.exception.*;
import lbg.common.net.*;

/** Keep info to find a file that lies on an FTP
 * @author Xavier Dalem
 */
public class FTPGridData extends GridData {
	static final long serialVersionUID = 7037970282268736283L;
	private static final Logger LOGGER = Logger.getLogger("lbg.data.ftp");
	
	private String server_host, path, user, password, local_host;
	private int server_port, local_port;
	private byte[] checksum;
	private File disk_file = null;
	
	/** Makes a new FTPGridData.
	 * @param id The unique ID for the file. It should be of the form
	 * <code>requestingpeerID.userID.filename</code>
	 * @param path Absolute path for data on the FTP server
	 * @param host The host where data lies.
	 * @param port The port on the FTP server.
	 * @param user Login on FTP server.
	 * @param password Associated password to user.
	 * @param checksum Checksum of the given data
	 * @throws GridException if parameters are invalid.
	 */
	public FTPGridData(String id, long size, String host, int port,
			String path, String user, String password, byte[] checksum)
			throws GridException {
		super(id, size);
		
		if (host == null || host.equals(""))
			throw new GridException("FTPGridData: invalid host");
		if (!CommonNetHandling.isValidPort(port))
			throw new GridException("FTPGridData: invalid port:" + port);
		if (path == null || path.equals(""))
			throw new GridException("FTPGridData: invalid path");
		if (user == null || user.equals(""))
			throw new GridException("FTPGridData: invalid user");
		if (password == null) // Could be empty
			throw new GridException("FTPGridData: invalid password");
		if (checksum == null)
			throw new GridException("FTPGridData: invalid checksum");
		
		this.server_host = host;
		this.server_port = port;
		this.path = path;
		this.user = user;
		this.password = password;
		this.checksum = checksum;
	}
	
	void setLocalFileCopy(File file) {
		disk_file = file;
	}

	public void setLocalParameters(String hostname, int port)
		throws GridException {
			
		if ((hostname == null) || (hostname.length() == 0) ||
			(CommonNetHandling.isValidPort(port) == false)) {
			throw new GridException("FTPGridData: invalid active parameters");
		}
		
		this.local_host = hostname;
		this.local_port = port;
		
	}

	/** Save data in a file.
	 * This call is blocking.
	 * <p>FTP transfer is in passive mode, you don't need to worry
	 * about ports to open.</p>
	 * @param file The file where to put data
	 */
	@Override
	public void toFile(File file) throws GridException {
		if (file == null)
			throw new GridException("FTPGridData: null output file");
		if (file.exists())
			throw new GridException("FTPGridData: output file already exists");
		
		try {
			FTPClient client = new FTPClient();
			client.setRemoteHost(server_host);
			client.setRemotePort(server_port);
			client.setConnectMode(FTPConnectMode.ACTIVE);
			
			client.connect();
			client.login(user, password);
			
			//edtftpj needs a IPv4 or IPv6 string. Resolve hostname and
			// extract what we need.
			final String inet = InetAddress.getByName(local_host).toString();
			client.setActiveIPAddress(inet.substring(inet.indexOf('/')+1));
			client.setActivePortRange(local_port, local_port);
			
			//Always run in binary, Task will do with exactly what was submitted
			client.setDetectTransferMode(false);
			client.setType(FTPTransferType.BINARY);
			
			LOGGER.finer("Started downloading " + id
				+ " from ftp://"+server_host+":"+server_port+"/ "
				+ " with local info " + local_host + ":" + local_port);
			
			client.get(new java.io.FileOutputStream(file), path);
			
			LOGGER.finer("Finished downloading " + id
				+ " from FTP");
			
			client.quit();
		} catch (Exception e) {
			LOGGER.log(Level.WARNING, "Failed to get file "+id, e);
			throw new GridException("FTPGridData: could not get file at "
				// Don't print user and password on stderr
				+ "ftp://" +/* user + ":" + password + "@" +*/ server_host
				+ ":" + server_port + "/" + path, e);
		}
		
		// Everything went OK, we can say we got the file.
		disk_file = file;
	}
	
	@Override
	public byte[] getChecksum() throws GridException {
		return checksum;
	}
	
	/** @throws GridException If the file doesn't exist */
	@Override
	public java.io.InputStream getInputStream() throws GridException {
		if (disk_file == null)
			throw new GridException("FTPGridData: File " +
				id + " was not retrieved");
		
		java.io.InputStream is = null;
		try {
			is = new java.io.FileInputStream(disk_file);
		} catch (java.io.FileNotFoundException e) {
			throw new GridException("Could not find file " + id, e);
		}
		
		return is;
	}
	
	/* FTPGridData usage demo */
	/*public static void main(String[] argv) throws GridException {
		if (argv.length != 6) {
			System.err.println("\tSample FTPGridData usage:\n"
				+ "java lbg.protocol.FTPGridData ID HOST PORT PATH USER PASS");
			return;
		}
		
		FTPGridData gd = new FTPGridData(argv[0], argv[1],
			new Integer(argv[2]).intValue(), argv[3], argv[4], argv[5],
			new byte[0]);
		File f = new File("FTPGD_" + gd.id());
		gd.setComplementaryInfo(argv[1], 60409);
		gd.toFile(f);
		System.out.println("File saved to " + f.getAbsolutePath());
	}*/
}
