/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// GridData.java
package lbg.protocol.data;

import lbg.common.exception.GridException;
import static lbg.common.io.CommonFileHandling.FILE_SEP;
import java.io.*;

/** Parent class for all grid data. Offers some basic functionnality for
 * subclasses and forces others.
 * @author Xavier Dalem, Cyril Briquet
 */
public abstract class GridData implements Comparable<GridData>, Serializable {
	/** Data ID. Should be of the form 
	 * <code>requestingpeerID.userID.original_filename</code> */
	public final String id;
	/** Size (in bytes) of GridData. */
	public final long size;
	protected String client_hostname = null;

	/** The name of the hashing algorithm that is used by all subclasses. Refer
	 * to the API from java.security.MessageDigest for valid names.
	 */
	public static final String HASHING_ALGORITHM = "SHA-256";
	
	/** Make a new GridData
	 * @param id The unique ID for the file. It should be of the form
	 * <code>requestingpeerID.userID.filename</code>
	 * @throws GridException if the ID is invalid
	 */
	protected GridData(String id, long size) throws GridException {
		if (id == null || id.equals(""))
			throw new GridException("Invalid file ID");

		this.id = id;
		this.size = size;

	}

	public final int compareTo(GridData other) {

		return this.id.compareTo(other.id);

	}

	/** Save data in a file in system temporary directory.
	 * This call is blocking.
	 * @return a <code>java.io.File</code> to the created file.
	 */
	final File toFile()  throws GridException {
		File f = new File(System.getProperty("java.io.tmpdir")+FILE_SEP+id);
		toFile(f);
		return f;
	}
	
	/** Save data in a file.
	 * This call is blocking.
	 * @param file The file where to put data
	 */
	public abstract void toFile(File file) throws GridException;
	
	/** Tell the GridData that its contents are already stored in a local
	 * file. This is particularly useful in cases where the data is a copy of
	 * an already known one but needs to know where the former file was
	 * stored.
	 * Subclasses that need this should reimplement this method.
	 * @param file The local copy.
	 */
	void setLocalFileCopy(File file) {
	}
	
	/** Get checksum from data. It must be the checksum from the real
	 * data, not of data and combined details like it is the case for torrents.
	 * <br>The used algorithm is specified by HASHING_ALGORITHM.
	 * @see #HASHING_ALGORITHM
	 */
	public abstract byte[] getChecksum() throws GridException;
	
	/** Get an input stream to read the data */
	public abstract java.io.InputStream getInputStream() throws GridException;
}
