/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// GridDataManager.java

package lbg.protocol.data;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.common.math.*;
import lbg.common.net.*;

import java.io.File;
import java.io.InputStream;
import java.security.MessageDigest;
import java.util.logging.Logger;
import java.util.logging.Level;

/** Handles GridData. The data manager (hereafter GDM) can host and retrieve
 * GridData files, and cache them for later use.<br>
 * It is a merge of the previous separate Peer- and ResourceDataManager,
 * grouped for convenience and ease of use.<br>
 * A lot of work was done to make this a normal object instead of a
 * singleton (a choice that was initially lead by the use of Azureus
 * as BitTorrent engine). It is now much more convenient to use but could be
 * subject to some documented side effects.<br>
 * Related logger: lbg.data<br>
 * <b>Warnings:</b>
 * <ul>
 * <li>If you want a resource and a peer to share the same cache directory (but
 * running in different processes, thus having different managers), you
 * should set the resource cache size to infinite. Else, the resource will end
 * up deleting the peer's files with.</li>
 * <li>Empty files are supported but such files cannot be recovered between
 * sessions. This is in order to be more flexible with files failing to
 * download the first time. See {@link #store(GridData, boolean)}</li>
 * </ul>
 * @author Xavier Dalem
 * @see #setWorkingSet(GridData[]) Concept of Working Set
 * @see EmbeddedBitTorrentServer#prepareAzureus(String)
 * Warning about use of <tt>user.home</tt> in conjunction with this class
 */
public class GridDataManager {
	
	// Static fields and methods {{{
	/** Logger for data management */
	static final Logger DATALOGGER = Logger.getLogger("lbg.data");
	/** Denotes that the GDM cache has an infinite capacity. */
	public static final int INFINITE_CAPACITY = -1;
	
	/** Buffer size for read operations */
	private static final int READ_BUFFER_SIZE = 1024;
	
	/* Bunch of Metadata extra keys */
	/* package */ static final String MEK_HOSTED_TORRENT = "hosted.torrent";
	/* package */ static final String MEK_HOSTED_FTP = "hosted.ftp";
	/* package */ static final String MEK_ORIGINAL_GD = "original_data";
	/* package */ static final String MEK_CHECKSUM = "checksum";
	/* package */ static final String MEK_LOCAL_FILE = "stored_file";
	
	/** Download a TorrentGridData and act as a new seed.
	 * This behavior improves the swarm of a torrent but the node using this
	 * mode doesn't change anything to the GridData itself.
	 * Should be used as a mask for download preferences
	 * @see #host(GridData, boolean, int)
	 */
	public static final int HOST_TORRENT_SEED = 0x0001;
	/** Download a GridData and seed & track it as a TorrentGridData.
	 * Implies HOST_TORRENT_SEED.
	 * Should be used as a mask for download preferences
	 * @see #HOST_TORRENT_SEED
	 * @see #host(GridData, boolean, int)
	 */
	public static final int HOST_TORRENT_TRACK = 0x0003;
	/** Download a GridData and host it on an FTP Server.
	 * Should be used as a mask for download preferences
	 * @see #host(GridData, boolean, int)
	 */
	public static final int HOST_FTP = 0x0004;
	
	/** Index of the TorrentGridData in the GridData download options array. */
	public static final int SHARED_TORRENT = 0;
	/** Index of the FTPGridData in the GridData download options array. */
	public static final int SHARED_FTP = 1;
	/** Index of the InlineGridData in the GridData download options array. */
	public static final int SHARED_INLINE = 2;
	/** Number (counting 1) of elements in a valid GridData download options
	 * array. */
	public static final int NUM_SHARE_OPTIONS = SHARED_INLINE + 1;
	
	/** Small utility to compare two hashes. They should obviously have been
	 * made by the same algorithm for the result to be relevant
	 * @see GridData#HASHING_ALGORITHM
	 */
	/* package */ static boolean equalHashes(byte[] hash1, byte[] hash2) {
		if (hash1.length != hash2.length)
			return false;
		
		for (int i = 0; i < hash1.length; i++)
				if (hash1[i] != hash2[i])
					return false;
		
		return true;
	}
	
	/** Hashes a file using the default algorithm
	 * ({@value lbg.protocol.data.GridData#HASHING_ALGORITHM}).
	 * 
	 * @param f The file to hash
	 * @return the hash in a byte array
	 * @throws GridException If hashing cannot be completed
	 * @see GridData#HASHING_ALGORITHM
	 */
	public static byte[] hash(File f) throws GridException {
		return hash(f, GridData.HASHING_ALGORITHM);
	}
	
	/** Hashes a file using the supplied algorithm.
	 * 
	 * @param f The file to hash
	 * @param algorithm The desired hashing algorithm, as per
	 * {@link java.security.MessageDigest}
	 * @return the hash in a byte array
	 * @throws GridException If hashing cannot be completed
	 * @see #hash(File)
	 */
	public static byte[] hash(File f, String algorithm) throws GridException {
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance(algorithm);
		} catch (java.security.NoSuchAlgorithmException e) {
			throw new GridException("File '" + f.getName()
					+ "' already exists and hash cannot be checked", e);
		}
		
		// Hash
		try {
			InputStream is = new java.io.FileInputStream(f);
			byte[] buffer = new byte[READ_BUFFER_SIZE];
			int read = is.read(buffer, 0, READ_BUFFER_SIZE);
			while (read != -1) {
				md.update(buffer, 0, read);
				read = is.read(buffer, 0, READ_BUFFER_SIZE);
			}
			is.close();
		} catch (java.io.IOException e){
			throw new GridException("Could not hash " + f.getName());
		}
		
		return md.digest();
	}
	
	/** Returns a human-readable representation of a hash.
	 * @param hash The hash to print
	 * @return A String containing an hexadecimal representation of the hash
	 * (1 character per 4 bits)
	 */
	public static String hashToString(byte[] hash) {
		StringBuilder out = new StringBuilder();
		for (byte b: hash)
			out.append(String.format("%x%x", (b >> 4)&0xF, b & 0xF));
		return out.toString();
	}
	
	// Static fields and methods }}}
	
	
	// Instance fields and methods {{{
	
	private final Measurement measurement;
	/** Current cache directory */
	private final String cacheDirectory;
	private ThreadedDownloader wsDownloader = null;

	private final EmbeddedFTPServer ftpServer;
	private final EmbeddedBitTorrentServer btServer;
	private final String hostname;
	private final int ftpDataPort;
	private int sharePreferences;
	private final int[] downloadPreferences
		= new int[] {SHARED_INLINE, SHARED_FTP, SHARED_TORRENT};
	
	private final Cache<String> localCache;
	private final Cache<String> supplyingCache;
	private final Object cacheLock = new Object();
	
	/** Creates a GridDataManager.
	 * @param hostname Hostname of the server or client owning this manager.
	 * @param cacheDir Absolute path to cache directory, without
	 * trailing path separator (see java.io.File.separatorChar). The directory
	 * must be created and writable.
	 * @param btDataPort The port for TCP and UDP BitTorrent transfers. It
	 * won't be taken into account if an Azureus core was already started
	 * @param ftpDataPort Port for FTP data connections. Needed for
	 * FTPGridData downloads.
	 * @param trackerHost The BitTorrent tracker host (i.e. the local host).
	 * Set to null for no tracking capabilities.
	 * @param trackerPort The desired BitTorrent tracker port. Will be ignored
	 * if trackerHHost is null.
	 * @param capacity The number of supplying files this manager can handle.
	 * INFINITE_CAPACITY for no upper bound. See class description for details.
	 * @param sharePreferences Flags for setting how data given to host() will
	 * be shared. 0 if you don't plan to host anything.
	 * @param statsDirectory Where the cache hit stats should be saved. Set to
	 * <tt>null</tt> If you don't want to keep track of the stats.
	 * @see lbg.peer.config.PeerDataConfiguration#buildRcvPolicy(String)
	 * @throws GridException If the parameters are set incorrectly.
	 */
	public GridDataManager(String hostname, String cacheDir,
			int btDataPort, int ftpDataPort, String trackerHost,
			int trackerPort, int capacity, int sharePreferences,
			String statsDirectory)
			throws GridException {
		
		if (hostname == null || hostname.equals(""))
			throw new GridException("GridDataManager: invalid hostname");
		if (cacheDir == null)
			throw new GridException("GridDataManager: null cache directory");
		if (!CommonNetHandling.isValidPort(ftpDataPort))
			throw new GridException("GridDataManager: invalid FTP data port: "
				+ ftpDataPort);
		if ((capacity < 1) && (capacity != INFINITE_CAPACITY))
			throw new GridException("GridDataManager: Capacity for supplying"
				+ " cache must be strictly positive");
		
		this.hostname = hostname;
		this.cacheDirectory = cacheDir;
		
		this.ftpServer = new EmbeddedFTPServer(hostname, cacheDir);
		this.btServer = new EmbeddedBitTorrentServer(cacheDirectory,
			trackerHost, trackerPort, btDataPort);

		// Create cache directory
		CommonFileHandling.createDirectory(cacheDir);

		try {
			File dir = new File(cacheDir);
			if (! dir.canWrite())
				throw new java.io.IOException(cacheDir + " is not writable");
		} catch (java.io.IOException e) {
			throw new GridException("Cache directory " + cacheDir
				+ " is not valid", e);
		}

		this.ftpDataPort = ftpDataPort;
		
		this.sharePreferences = sharePreferences;
		
		localCache = new UnlimitedCache<String>();
		if (capacity != INFINITE_CAPACITY)
			supplyingCache = new SlottedLRUCache<String>(capacity);
		else
			supplyingCache = new UnlimitedCache<String>();

		// Measurement handling
		if (statsDirectory != null) {
			this.measurement =
				MeasurementFactory.
					createSummarizedMeasurement(this.hostname + ".cache_hit");
			Runtime.getRuntime().addShutdownHook(new Thread(
				new MeasurementSaver(this.measurement, statsDirectory)));
		}
		else {
			this.measurement = (Measurement)null;
		}
		
		// Shutdown cleaner
		Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
			public void run() {
				ftpServer.stop();
				btServer.stop();
			}
		}));
	}
	
	
	// Type-independent operations {{{
	
	/** @return A string representation of this manager. */
	public String toString() {
		StringBuilder sb = new StringBuilder();
		
		sb.append("[GridDataManager]");
		
		sb.append("\n  ->  Cache directory=").append(cacheDirectory);
		sb.append("\n  ->  local cache=").append(localCache);
		sb.append("\n  ->  supplying cache=").append(localCache);
		sb.append("\n  ->  BT.sharing=").append(btServer);
		sb.append("\n  ->  FTP.sharing=").append(ftpServer);
		sb.append("\n  ->  FTP.data port=").append(ftpDataPort);
		
		sb.append("\n  ->  Download preferences=");
		for (int pref: downloadPreferences) {
			switch (pref) {
				case SHARED_FTP:
					sb.append(" FTP");
					break;
				case SHARED_TORRENT:
					sb.append(" TORRENT");
					break;
				case SHARED_INLINE:
					sb.append(" INLINE");
					break;
				default:
					sb.append("*UNKNOWN*");
			}
		}
		
		sb.append("\n  ->  Sharing preferences=");
		if ((sharePreferences & HOST_FTP) != 0)
			sb.append(" FTP");
		if ((sharePreferences & HOST_TORRENT_TRACK) != 0)
			sb.append(" TORRENT");
		
		return sb.toString();
	}
	
	/** Returns the file that should be associated with a data ID. This offers
	 * no guarantee on whether the file exists or not.
	 * @param dataID The ID of the data to check.
	 * @return A File where data is or should be stored
	 */
	private File getDiskFile(String dataID) {
		// Note for future comitters: this has to stay deterministic
		return new File(cacheDirectory + File.separatorChar	+ dataID);
	}

	
	/** Detects whether the manager was configured with a default policy or not.
	 * 
	 * @return <tt>true</tt> if the manager has a default policy
	 */
	public boolean hasDefaultPolicy() {
		return this.sharePreferences != 0;
	}
	
	protected boolean isCached(String id) {
		synchronized(cacheLock) {
			return (localCache.has(id) || supplyingCache.has(id));
		}
	}
	
	protected Metadata<String> getMetadata(String id) {
		Metadata<String> meta = null;
		
		synchronized(cacheLock) {
			if ((meta = localCache.getMetadata(id)) == null)
				meta = supplyingCache.getMetadata(id);
		}
		
		return meta;
	}
	
	/** Checks if the passed GridData's would fit in the Working Set.
	 * This method does not change the state of the GridDataManager in any way.
	 * @param workingSet The candidate working set
	 * @return <tt>true</tt> if the working set could fit.
	 */
	public synchronized boolean checkWorkingSet(GridData[] workingSet) {
		@SuppressWarnings("unchecked")
		Metadata<String>[] metas = new Metadata[workingSet.length];
		
		for (int i = 0; i < workingSet.length; i++) {
			metas[i] = new Metadata<String>(workingSet[i].id);
			metas[i].setWeight(workingSet[i].size);
		}
		
		return supplyingCache.canAdd(metas, false);
	}
	
	/** Sets the working set.
	 * This call is not blocking.
	 * @param workingSet The new working set (an array of download options)
	 * @return <tt>false</tt> if working set is larger than cache or data
	 * couldn't be scheduled for download
	 */
	public synchronized boolean setWorkingSet(GridData[][] workingSet) {
		return setWorkingSet(extract(workingSet));
	}
	
	/** Sets the working set.
	 * <p>A Working Set is a set of GridData's that should stay in cache.
	 * When data has to be popped from cache, only those not in the WS will
	 * be candidate.</p>
	 * <p>This call is not blocking, data will be transferred in the
	 * background.</p>
	 * @param workingSet The new working set (an array of different data's).
	 * Set to null to remove working set management.
	 * @return <tt>false</tt> if working set is larger than cache or data
	 * couldn't be scheduled for download
	 */
	public synchronized boolean setWorkingSet(GridData[] workingSet) {
		if (! checkWorkingSet(workingSet))
			return false;
		
		// Clear
		if (workingSet == null) {
			supplyingCache.setVeto(null);
			return true;
		}
		
		// Set new WS
		String[] veto = new String[workingSet.length];
		for (int i = 0; i < veto.length; i++) {
			assert workingSet[i] != null : "null element in working set";
			veto[i] = workingSet[i].id;
		}
		supplyingCache.setVeto(veto);
		
		// Non-blockingly download all data
		if (wsDownloader == null)
			wsDownloader = new ThreadedDownloader(this);
		
		try {
			wsDownloader.setDownloadQueue(workingSet);
		} catch (GridException e) {
			DATALOGGER.log(Level.WARNING, "Failed to download Working Set", e);
			return false;
		}
		
		return true;
	}
	
	/** Blockingly wait for all required data's to be present in cache.
	 * Though blocking, the method will not hold the lock for the whole time.
	 * <p>The required data has to be in the current working set. Else, we can't
	 * be sure it will ever be downloaded.
	 * If the data was already known, the submitted one's local file copy
	 * is set.</p>
	 * <p><b>This method doesn't actually download the data.</b></p>
	 * @param requirements The data's to wait for
	 * @throws GridException if the active working set isn't included in the
	 * previously given working set.
	 * @see #setWorkingSet(GridData[])
	 */
	public synchronized void waitFor(GridData[] requirements)
			throws GridException {
		if (requirements == null)
			throw new GridException("waitFor: null requirements");
		if (requirements.length == 0)
			return;
		if (supplyingCache.vetoSize() <= 0)
			throw new GridException("Cannot wait for data in empty WS");
		
		int cache_hit_num = 0;
		// Check data before downloading. We run through data twice but
		// if something is wrong, we'd better know it before starting to
		// download anything
		for (GridData gd : requirements) {
			if (! supplyingCache.hasVetoOn(gd.id))
				throw new GridException("Cannot wait for data " + gd.id
					+ ": it's not in the working set");
			if (isCached(gd.id))
				cache_hit_num++;
		}

		// Task cache hit ratio (btwn 0.0 and 1.0)
		float cache_hit_ratio = ((float)cache_hit_num)/requirements.length;
		
		for (GridData gd : requirements) {
			boolean stored = isCached(gd.id);
			GridException e = wsDownloader.getException();
			if (e != null)
				throw e;
			while (! stored) {
				try { wait(); } catch (InterruptedException ie) {}
				e = wsDownloader.getException();
				if (e != null)
					throw e;
				stored = isCached(gd.id);
			}
			
			// Don't forget to set the copy. At worse, we set the same
			// thing twice.
			Cache<String> source = localCache;
			if (supplyingCache.has(gd.id))
				source = supplyingCache;
			gd.setLocalFileCopy((File) source.getMetadata(gd.id)
				.getExtra(MEK_LOCAL_FILE));
		}
		
		if (this.measurement != null)
			this.measurement.addObservation(cache_hit_ratio);
	}
	
	/* Reminder: extraction does not need to be synchronized: the parameters
	 * it relies on are set once, at instantiation-time */
	/** Extracts an array of data downloading options.
	 * The extraction is based on the preferences given at instantiation and
	 * available download types.<br>
	 * <b>Warning:</b> If preferences only list unavailable types, some or all
	 * entries in the returned array might be null. The only way to be sure to
	 * to always have a valid returned value is to ensure that the data
	 * preferences list all types.
	 * @param datas An array of arrays of options for downloading GridData's
	 */
	public GridData[] extract(GridData[][] datas) {
		GridData[] extracted = new GridData[datas.length];
		
		for (int i = 0; i < extracted.length; i++)
			extracted[i] = extract(datas[i]);
		
		return extracted;
	}
	
	/** Selects the preferred download option out of an array of possibilities.
	 * The selected option is
	 * @return The preferred option, or null if none matched.
	 */
	public GridData extract(GridData[] data) {
		GridData extracted = null;
		
		// Select best choice
		for (int j = 0; j < downloadPreferences.length; j++) {
			// First non-null matching option gets it
			if (data[downloadPreferences[j]] != null) {                       
				extracted = data[downloadPreferences[j]];
				break;
			}
		}
		
		return extracted;
	}
	
	/** Hosts given GridData.
	 * <p>Hosting some GridData is saving it locally and rendering it available
	 * for other peers / resources using different protocols depending on
	 * the flags.</p>
	 * @param data The GridData to host.
	 * @param isLocal Whether the data is local to the GDM owner (<tt>true</tt>)
	 * or from an extern consumer (supplying: <tt>false</tt>).
	 * The cache policy for local data is to leave data in cache as long as
	 * it's not explicitly deleted. Cache for supplying data is generally
	 * limited.
	 * @param prefs Flags for setting how data given to host() should be
	 * shared (overrides the default policy if any).
	 * @return An array of GridData's, each type being accessed by the
	 * SHARE_* static indices. Depending on the flags, some or all of the
	 * elements might be null.
	 * @throws GridException if hosting failed (could be at numerous places
	 * depending on the flags)
	 * @see #HOST_TORRENT_SEED
	 * @see #HOST_TORRENT_TRACK
	 * @see #HOST_FTP
	 * @see #SHARED_TORRENT
	 * @see #SHARED_FTP
	 * @see #extract(GridData[])
	 */
	public GridData[] host(GridData data, boolean isLocal,
			int prefs) throws GridException {
		if (data == null)
			throw new NullPointerException("Cannot host null data");
		if (prefs == 0)
			throw new GridException(
				"Cannot host without a proper sharing policy");
		
		GridData[] hosted = new GridData[3];
		for (int i = 0; i < hosted.length; hosted[i++]=null);
		
		/* HOST_TORRENT_SEED is actually not taken into account. All that's
		 * Needed to comply with definition is to download the data. */
		
		Metadata<String> meta = store(data, isLocal);
		
		if ((prefs & HOST_TORRENT_TRACK) != 0) {
			// Either it was just created or it wasn't tracked yet
			if (meta.getExtra(MEK_HOSTED_TORRENT) == null) {
				meta.setExtra(MEK_HOSTED_TORRENT, btServer.track(meta));
			}
			
			hosted[SHARED_TORRENT] = (GridData) meta.getExtra(MEK_HOSTED_TORRENT);
		}
		
		if ((prefs & HOST_FTP) != 0) {
			if (meta.getExtra(MEK_HOSTED_FTP) == null) {
				meta.setExtra(MEK_HOSTED_FTP, hostOnFTP(meta));
			}
			
			hosted[SHARED_FTP] = (GridData) meta.getExtra(MEK_HOSTED_FTP);
		}
		
		return hosted;
	}
	
	
	/** Hosts given GridData with default sharing policy.
	 * @throws GridException if the manager doesn't have a default policy
	 * @see #host(GridData, boolean, int)
	 * @see #extract(GridData[])
	 */
	public synchronized GridData[] host(GridData data, boolean isLocal)
			throws GridException {
		if (this.sharePreferences == 0)
			throw new GridException("Manager doesn't have a default policy");
		
		return host(data, isLocal, this.sharePreferences);
	}
	
	/** Store data in cache directory.<br>
	 * If the data isn't local, it is added to a queue of elements whose
	 * behavior was explained earlier.<br>
	 * This call is blocking.<br>
	 * Some data types may leave empty files when failing. These are deleted
	 * on next download if required (to avoid hash collision between the empty
	 * and the real file). For this reason, all "legally" empty files on disk
	 * but not in cache map will be downloaded again.<br>
	 * This method is not fully synchronized: as the call is blocking, it would
	 * slow down calls a lot. However, it should still be thread-safe.<br>
	 * <b>Note: </b> to avoid attacks, cached data is only removed when the new
	 * file has been successfully downloaded.You should be aware that the actual
	 * size of the cache directory could exceed the desired capacity (but only
	 * for one file and a reasonably short time). 
	 * @param data The data to store
	 * @param isLocal Whether the data is local or not. Ignored if data was
	 * already known (but not if it was recovered !).
	 * @return The CacheEntry of the now downloaded data.
	 * @throws GridException if the data wasn't successfully downloaded and
	 * stored.
	 */
	/* Note: each time a data is downloaded or encounters an exception, we
	 * notify the threads in waitFor()
	 */
	Metadata<String> store(GridData data, boolean isLocal)
			throws GridException {
		/* We have to gather all exceptions to wake threads potentially
		 * waiting to complete if we were to fail.
		 */
		try {
			if (data == null)
				throw new
					GridException("GridDataManager: cannot store null data");

			if (data instanceof TorrentGridData)
				btServer.start();
			
			Cache<String> destination =
				(isLocal) ? localCache : supplyingCache;

			// Previous info
			Metadata<String> meta = null;
			synchronized(cacheLock) {
			// If it was already known as supplying, it should be moved to local
			if (isLocal && supplyingCache.has(data.id)) {
				meta = supplyingCache.remove(data.id);
				assert(meta != null);
				final Metadata<String> safety = localCache.add(data.id, meta);
				assert (safety == null) : "Local cache should be unlimited" ;
			} else if (localCache.has(data.id)) {
				meta = localCache.getMetadata(data.id);
				// Data is already local, all operations should stay local
				destination = localCache;
			} else {
				// Will stay null if the data is unknown
				meta = supplyingCache.getMetadata(data.id);
				// destination is already OK
			}
			}
			
			byte[] checksum = data.getChecksum();
			
			if (meta != null) {
				// We already had information for this data, check consistency
				if (equalHashes(checksum, (byte[]) meta.getExtra(MEK_CHECKSUM))) {
					data.setLocalFileCopy((File) meta.getExtra(MEK_LOCAL_FILE));
					// Refresh cache (in case the cache needs it)
					final Metadata<String> safety = destination.add(data.id, meta);
					assert (safety == null): "A refresh shouldn't pop anything";
					
					// Awaken threads that were potentially waiting for this data
					// to be ready.
					synchronized(this) {
						notifyAll();
					}
					return meta;
				}
				
				throw new GridException("Data for '" + data.id + "' already "
					+ "recorded with a different hash");
					
			}
			

			// The data was not stored in cache. Maybe there is a cached
			// file from another session
			File diskFile = getDiskFile(data.id);

			// Sometimes, file transfers may create an empty file before
			// failing. Delete them. If we can't, we'll reach a hash collision
			// below
			if (diskFile.exists() && diskFile.length() == 0) {
				if (! diskFile.delete()) {
					// too bad, get ready for a collision
					DATALOGGER.warning("Failed to delete empty file " + data.id);
				}
			}

			if (diskFile.exists()) {
				// Hash the disk file
				byte[] diskChecksum = hash(diskFile);

				if (! equalHashes(checksum, diskChecksum)) {
					throw new GridException("File '" + data.id
						+ "' already exists with a different hash (existing"
						+ " file is " + diskFile.length() + "B long)"
						+ "\n\tdisk hash: " + hashToString(diskChecksum)
						+ "\n\texp. hash: " + hashToString(checksum));
				}

				// Data matches. Recover the disk entry, it can come from an
				// older session or another data manager operating on the same
				// directory
				data.setLocalFileCopy(diskFile);
			}

			else { // This is a totally new entry
				/* Quite a hack to check the class, but active FTP needs this
				 * (this is especially true in NAT environments) and we can't
				 * use passive FTP for scalability reasons. */
				if (data instanceof FTPGridData)
					((FTPGridData)data)
						.setLocalParameters(this.hostname,this.ftpDataPort);
				if (data instanceof TorrentGridData)
					((TorrentGridData)data).getDownloadManager();
				data.toFile(diskFile);
				if (data instanceof TorrentGridData)
					((TorrentGridData)data).getDownloadManager();
			}
			
			// Add this new or recovered data. Let the cache handle replacement
			// policy
			meta = new Metadata<String>(data.id);
			meta.setExtra(MEK_LOCAL_FILE, diskFile);
			meta.setWeight(data.size);
			meta.setExtra(MEK_CHECKSUM, checksum);
			
			if (! (data instanceof InlineGridData) // Don't use too much space
				&& (meta.getExtra(MEK_ORIGINAL_GD) == null)) // keep only the first match
				meta.setExtra(MEK_ORIGINAL_GD, data);
		
			
			try {
				// This is the only place we could pop an entry from a bounded
				// cache
				synchronized(cacheLock) {
					Metadata<String> removed = destination.add(data.id, meta);
					if (removed != null)
						deleteEntry(removed);
				}
			} catch (VetoException ex) {
				throw new GridException("Storing data " + data.id
					+ " failed: couldn't find another data to pop that is"
					+ " not in active working set.", ex);
			}
			
			// Awaken threads that were potentially waiting for this data
			// to be ready.
			synchronized(this) {
				notifyAll();
			}

			return meta;
		
		} catch (VetoException e) {
			// This means one of the operations consisting in moving
			// a data from supplying to local cache failed.
			synchronized(this) { notifyAll(); }
			throw new GridException(
				"Went against a veto while refreshing the cache", e);
		} catch (GridException e) {
			synchronized(this) { notifyAll(); }
			throw e;
		}
	}
	
	/** Ensures that all given data's are being seeded. If it's not the case,
	 * seed them.
	 * Given data's are assumed to be already downloaded.
	 * 
	 * @param datasOptions An array of arrays of data transfer options for
	 * data's. In other words, there are datasOptions.length data's to check.
	 */
	public synchronized void seed(GridData[][] datasOptions) throws GridException {
		btServer.start();
		
		for (GridData[] options: datasOptions) {
			if (options[SHARED_TORRENT] == null)
				// Not possible to seed without an initial tracked torrent 
				continue;
			
			GridData selected = extract(options);
			DATALOGGER.finest("Re-seeding " + selected.id
				+ ", first transferred as "
				+ selected.getClass().getSimpleName());
			
			File f = getDiskFile(options[SHARED_TORRENT].id);
			assert (f.exists()):"We can only seed existing files";
			
			final Metadata<String> meta =
			  getMetadata(options[SHARED_TORRENT].id);
			
			if (meta == null) {
				throw new GridException("Data " + options[SHARED_TORRENT].id +
										" is not known by the cache. "+
										"Only already downloaded data "+
										"can be seeded.");
			}

			if (meta.getExtra(MEK_ORIGINAL_GD) instanceof TorrentGridData) {
				//AzureusListener.getInstance().setPrintStream(System.err); // DEBUG
				//AzureusListener.getInstance().reportDownload(((TorrentGridData)meta.getExtra(MEK_ORIGINAL_GD)).getDownloadManager()); // DEBUG
				continue; // Already is being shared using BitTorrent
			}
			
			// We don't need to update any internal structure, we're just
			// starting a torrent download on an already complete and known file
			options[SHARED_TORRENT].toFile(f);
			
			// Set original grid data as this one, for future processing
			meta.setExtra(MEK_ORIGINAL_GD, options[SHARED_TORRENT]);
		}
		
		DATALOGGER.finer("Done sending data for seeding");
	}
	
	/** Get the number of seeds participating in a BitTorrent download
	 * for a particular file
	 * 
	 * @param dataID The identifier of the data to check for
	 * @param countPeers <tt>true</tt> if the BitTorrent Peers should also be
	 * counted. These could have no interesting pieces right now, but could
	 * become valuable sharers anyway.
	 * @return The number of seeds (and peers if asked)
	 */
	public int getSeedCount(String dataID, boolean countPeers) {
		Metadata<String> data;
		synchronized(cacheLock) {
			if ((data = getMetadata(dataID)) == null)
				return 0;
		}
		
		final GridData gd;
		TorrentGridData tgd;
		
		// The data could be tracked locally, in which case we have direct
		// access to the shared torrent
		if ((tgd = (TorrentGridData)data.getExtra(MEK_HOSTED_TORRENT)) != null)
			;
		// Or, it can be a re-seed, in which case we can get information
		// from the torrent that was used for downloading
		else if ((gd = (GridData)data.getExtra(MEK_ORIGINAL_GD)) instanceof TorrentGridData)
			tgd = (TorrentGridData)gd;
		else
			return 0;
		
		assert (tgd != null) : "TorrentGridData should be non-null";
		assert (tgd.getDownloadManager() != null)
			: "Download manager was used for download and shouldn't be null";
		
		/*
		We cannot rely on DownloadManager.getNb*() because, apparently, it
		only accounts Peers the current Peer has contacts with. Concretely,
		if the current Peer isn't downloading anything and nobody is
		downloading anything from him either, the current Peer know no seeds and
		no peers.
		The information we want can only be known by the tracker. As such,
		it must come from a periodic request of the Peer to the Tracker, the
		so-called 'Scrape'.
		See also:
			http://www.azureuswiki.com/index.php/Scrape
			http://www.google.com/search?q=bittorrent+scrape
		*/
		
		int seeds = tgd.getDownloadManager().getTrackerScrapeResponse().getSeeds();
		if (countPeers)
			seeds += tgd.getDownloadManager().getTrackerScrapeResponse().getPeers();
		
		return seeds;
	}
	
	/** Removes a selected entry from disk.
	 * The entry is also removed for all internal structures and not available
	 * for sharing anymore.
	 * @param toDelete The entry to delete
	 */
	private synchronized void deleteEntry(Metadata<String> toDelete) {
		try {
			btServer.removeFromBitTorrent(toDelete);
		} catch (GridException e) {
			DATALOGGER.log(Level.WARNING, "Torrent data " + toDelete.id
				+ " wasn't properly removed", e);
		}
		
		File toDeleteFile = (File) toDelete.getExtra(MEK_LOCAL_FILE);
		assert (toDeleteFile != null) : "File was stored but we lost 'where'";
		
		if (toDeleteFile.delete())
			DATALOGGER.fine("Data " + toDelete.id + " expired, it was"
				+ " removed from cache (was: "
				+ toDeleteFile.getAbsolutePath() + ")");
		else
			DATALOGGER.warning("Data " + toDelete.id+ " expired but could"
				+ " not be physically deleted from cache. It should be"
				+ " safe to delete it manually (left in: "
				+ toDeleteFile.getAbsolutePath() + ")");
	}
	
	/** Store data in cache and get the associated File.<br>
	 * Data saved this way is never considered as local. Peers wanting to share
	 * data should instead use {@link #host(GridData, boolean)}.<br>
	 * This call is blocking.
	 * @param data The data to store
	 * @return The file where the data was stored
	 * @throws GridException if the data wasn't successfully downloaded and
	 * stored.
	 */
	public synchronized File getFile(GridData data) throws GridException {
		return (File) store(data, false).getExtra(MEK_LOCAL_FILE);
	}
	
	/** Convenience method to automate extraction of data when downloading
	 * @param data An array of dowload options for a GridData (<b>not</b> an
	 * array of independent GridData's)
	 */
	public synchronized File getFile(GridData[] data) throws GridException {
		return (File) store(extract(data), false).getExtra(MEK_LOCAL_FILE);
	}
	// }}}
	
	// FTP operations {{{
	/** Restart the FTP server, forcing reload of the configuration. */
	public synchronized void restartFTPServer() throws GridException {
		ftpServer.stop();
		ftpServer.start();
	}
		
	
	/** Sets the FTP server configuration.<br>
	 * If a server is already running, it will need to be restarted for the
	 * changes to take effect. If no user information is given, it is
	 * either recovered from what was previously set or generated randomly.
	 * @param port FTP Requests port
	 * @see #restartFTPServer()
	 * @return the current GDM, for convenience
	 * @throws GridException if the parameters are invalid
	 */
	public GridDataManager setFTPConfig(int port) throws GridException {
		ftpServer.setConfig(port);
		return this;
	}
	
	/** Sets the FTP server configuration.<br>
	 * If a server is already running, it will need to be restarted for the
	 * changes to take effect.
	 * @param port FTP Requests port
	 * @param user User name
	 * @param password Associated password
	 * @see #restartFTPServer()
	 * @return the current GDM, for convenience
	 * @throws GridException if the parameters are invalid
	 */
	public GridDataManager setFTPConfig(int port, String user,	String password)
			throws GridException {
		ftpServer.setConfig(port, user, password);
		return this;
	}
	
	/** Hosts the given entry on FTP server
	 * @param ce The entry to host
	 * @return the converted FTPGridData
	 */
	private FTPGridData hostOnFTP(Metadata<String> meta) throws GridException {
		ftpServer.start();
		return new FTPGridData(meta.id, meta.getWeight(), hostname,
			ftpServer.getPort(), meta.id, ftpServer.getUser(),
			ftpServer.getPassword(), (byte[]) meta.getExtra(MEK_CHECKSUM));
	}
	// FTP operations }}}

	// Instance fields and methods }}}
}

