/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// InlineGridData.java
package lbg.protocol.data;

import lbg.common.exception.GridException;

import java.io.*;
import java.security.MessageDigest;

/** Grid data class where the data is actually contained by the class itself.
 * Nice for User to Peer communication, where the file will be sent once
 * @author Xavier Dalem
 */
public class InlineGridData extends GridData {
	static final long serialVersionUID = -1220620537770888733L;
	
	private byte[] data;
	private byte[] checksum;
	
	/** Create an InlineGridData with given id and data.
	 * @param id The file ID. For restrictions on this, see GridData
	 * @param data The data as an array of bytes
	 * @see GridData
	 */
	public InlineGridData(String id, byte[] data) throws GridException {
		super(id, data.length);
		
		this.data = data;
		checksum = null;
	}
	
	public void toFile(File file) throws GridException {
		if (file == null)
			throw new GridException("FTPGridData: null output file");
		
		if (file.exists())
			throw new GridException("File already exists");
		
		FileOutputStream fos = null;
		try {
			fos = new FileOutputStream(file);
		} catch (FileNotFoundException e) {
			throw new GridException("Could not create output stream for "
				+ file.getAbsolutePath(), e);
		}
		try {
			fos.write(data);
		} catch (IOException e) {
			throw new GridException("Could not write to "
				+ file.getAbsolutePath(), e);
		}
		try { fos.close(); } catch (IOException e) {}
	}
	
	public byte[] getChecksum() throws GridException {
		// Cache the checksum to avoid recomputing it numerous times
		if (checksum != null)
			return checksum;
		
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance(GridData.HASHING_ALGORITHM);
		} catch (java.security.NoSuchAlgorithmException e) {
			throw new GridException("Cannot hash data " + id, e);
		}
		
		checksum = md.digest(data);
		return checksum;
	}
	
	public InputStream getInputStream() {
		return new ByteArrayInputStream(data);
	}
}