/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// MetaGridData.java

package lbg.protocol.data;

import lbg.common.exception.*;
import lbg.common.math.*;
import java.io.Serializable;

/**
 * @author Cyril Briquet
 */
public class MetaGridData implements Comparable<MetaGridData>, Serializable {

	public MetaGridData(String id,long size,GridData[] grid_data)
		throws GridException {
	
		if ((id == null) || (id.length() < 1)) {
			throw new GridException("MetaGridData(): illegal data id");
		}
		if (size < 1) {
			throw new GridException("MetaGridData(): illegal data size");
		}
		if ((grid_data == null) ||
			(grid_data.length < 1)) {
			throw new GridException("MetaGridData(): illegal GridData array");
		}
		
		this.id = id;
		this.size = size;
		this.grid_data = grid_data;
		this.hashcode = Integer.MIN_VALUE;
	
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		sb.append("[MetaGridData] id=");
		sb.append(this.id);
		sb.append(" size=");
		sb.append(this.size);
		sb.append(" bytes");

		return sb.toString();

	}
	
	/**
	 * Note: this class has a natural ordering that is inconsistent
	 * with equals.
	 * <small>It is hypothesized that 2 MetaGridData with same id
	 * and different sizes will *never* be compared.
	 * As this method is called very frequently, the implementation
	 * of compareTo() skips the comparison of size,
	 * relying only on the comparison of id.</small>
	 */
	public int compareTo(MetaGridData other) {

		return this.id.compareTo(other.id);

	}

	@Override
	public boolean equals(Object obj) {
		
		if (obj instanceof MetaGridData == false) {
			return false;
		}
		
		MetaGridData other = (MetaGridData)obj;
		
		return (this.id.equals(other.id)) && (this.size == other.size);
		
	}

	@Override
	public int hashCode() {
	
		if (this.hashcode == Integer.MIN_VALUE) {
			this.hashcode = CommonMath.generateHashCode(this.id) ^ super.hashCode();
		}
		return this.hashcode;
	
	}


	// ****************************************************************

	public final String id;
	public final long size;
	public final GridData[] grid_data;
	private int hashcode;

	// ****************************************************************
	
	static final long serialVersionUID = 7652485949586916691L;

}

