/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Metadata.java

package lbg.protocol.data;

import java.util.Map;
import java.util.TreeMap;

/** Denotes metadata about a cache entry.
 * An entry is typically represented by an ID, and has an insertion time and a
 * size. While these metrics are often used, it wouldn't be possible to foresee
 * every other metric that could possibly be used, or every side data that could
 * be needed. For this reason, the {@link #setExtra(String, Object)} and
 * {@link #getExtra(String)} methods offer a contract similar to
 * {@link java.util.Map} to set what could be needed.
 * 
 * @author Xavier Dalem
 *
 * @param <K> The type of the entry key
 */
public class Metadata<K> {
	private long time;
	private long weight = 0;
	private final Map<String, Object> extras;
	public final K id;
	
	/** Creates a Metada, setting time to the current time.
	 * Weight is initalized to 0.
	 * 
	 * @see #setWeight(long)
	 * @see #setTime(long)
	 */
	public Metadata(K id) {
		setTime();
		this.id = id;
		this.extras = new TreeMap<String, Object>();
	}
	
	/** Set the time from epoch.
	 * 
	 * @see #setTime(long)
	 */
	public void setTime() {
		setTime(System.currentTimeMillis());
	}
	
	/** Sets the creation time of this Metadata.
	 * Time should be relative to a defined moment. It can be epoch (as returned
	 * by {@link System#currentTimeMillis()} but can very well be any other
	 * convention as long as it is consistent.
	 * 
	 * @param time The time in milliseconds
	 */
	public void setTime(long time) {
		this.time = time;
	}
	
	public long getTime() {
		return time;
	}
	
	/** Sets the weight of this metadata.
	 * If the metadata is associated with a file, the weight could be the size
	 * in bytes of that file.
	 * 
	 * @param weight The weight to set.
	 */
	public void setWeight(long weight) {
		this.weight = weight;
	}
	
	public long getWeight() {
		return weight;
	}
	
	public Object setExtra(String key, Object value) {
		return extras.put(key, value);
	}
	
	public Object getExtra(String key) {
		return extras.get(key);
	}
}
