/*
 * Copyright 2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Edits for Lightweight Bartering Grid by Xavier Dalem
 *
 */

// RestrictedFtpServer.java

package lbg.protocol.data;

import java.net.ServerSocket;
import java.net.Socket;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.ftpserver.interfaces.IConnectionManager;
import org.apache.ftpserver.interfaces.IFtpConfig;

/** A restricted FtpServer with less Debugging and without the ability to list
 * directories.
 * @author <a href="mailto:rana_b@yahoo.com">Rana Bhattacharyya</a>
 * @author Xavier Dalem
 */
class RestrictedFtpServer extends org.apache.ftpserver.FtpServer {

    private Thread runner;
    private ServerSocket serverSocket;
    private IFtpConfig ftpConfig;
    private boolean suspended;
    private static final Logger LOGGER = Logger.getLogger("lbg.data.ftp");
    

    public RestrictedFtpServer(IFtpConfig ftpConfig) {
        super(ftpConfig);
		this.ftpConfig = ftpConfig;
    }
	
	/**
     * Start the server. Open a new listener thread.
     */
    public void start() throws Exception {
        if (runner == null) {
			serverSocket = ftpConfig.getSocketFactory().createServerSocket();
			LOGGER.finest("Factory created command socket " + serverSocket);
			runner = new Thread(this);
			runner.setDaemon(true);
			runner.start();
        }
    }

    /**
     * Listen for client requests.
     */
    public void run() {
        IConnectionManager conManager = ftpConfig.getConnectionManager();
        while (runner != null) {
            try {
                
                // closed - return
                if(serverSocket == null) {
                	
                    return;
                }
                
                // accept new connection .. if suspended 
                // close immediately.
                Socket soc = serverSocket.accept();
                if(suspended) {
                    try {
                        soc.close();
                    }
                    catch(Exception ex) {
                    }
                    continue;
                }
                
                conManager.newConnection(
					new RestrictedRequestHandler(ftpConfig, soc));
            }
            catch (Exception ex) {
            	LOGGER.log(Level.SEVERE,
            		"Critical exception in FTP server, exiting serving thread",
            		ex);
                return;
            }
        }
    }
}

