/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SimulatedGridDataFactory.java

package lbg.protocol.data;

import lbg.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.common.math.*;
import lbg.simulator.*;

/** Factory of SimulatedGridData.
 * @author Cyril Briquet
 */
public class SimulatedGridDataFactory {
	
	/** Creates a new SimulatedGridDataFactory.
	 * @param random_salt true if filename salts should be selected randomly,
	 * false if they should be selected arbitrarily
	 * @param max_salt_value maximum salt value
	 * @exception GridException if a parameter is illegal
	 */
	public SimulatedGridDataFactory(boolean random_salt,int max_salt_value)
		throws GridException {
	
		if (max_salt_value < 0) {
			throw new GridException("SimulatedGridDataFactory(): "+
									"illegal maximum salt value");
		}
		
		this.random_salt = random_salt;
		this.max_files_in_bot = max_salt_value;
		this.next_salt = 1;
		
	}

	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[SimulatedGridDataFactory]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized SimulatedGridData
		createSimulatedGridData(String peer_id,String suffix)
		throws GridException {
		
		if (this.max_files_in_bot < 1) {
			throw new GridException("createSimulatedGridData(): "+
									"illegal access");
		}
		if (LightweightBarteringGrid.getMode() != SIM_MODE) {
			throw new GridException("SimulatedGridData(): wrong constructor");
		}

		if (peer_id == null) {
			throw new GridException("createSimulatedGridData(): "+
									"illegal Peer id");
		}
		
		StringBuilder sb_filename = createFileName(peer_id);
		if (suffix != null) {
			sb_filename.append(suffix);
		}
		String filename = sb_filename.toString();
		
		long size = CommonMath.sampleRandomFileSize(filename);
		
		return new SimulatedGridData(filename,size);
		
	}

	public synchronized SimulatedGridData
		createSimulatedGridData(String peer_id)
		throws GridException {

		return createSimulatedGridData(peer_id,null);
		
	}
	
	public synchronized GridData[][] 
		createSimulatedGridDataArray(String consumer_peer_id)
		throws GridException {
		
		if (consumer_peer_id == null) {
			throw new GridException("generateSimulatedGridData(): "+
									"illegal Consumer Peer id");
		}
		
		Scenario sc = LightweightBarteringGrid.getScenario();
		int size = sc.getGridConfiguration().DATA_PER_TASK(consumer_peer_id);

		if (size < 0) {
			throw new GridException("generateSimulatedGridData(): "+
									"illegal size");
		}
		
		GridData[][] dataset = new GridData[size][];
		
		// Loop over Task data files
		for (int i = 0 ; i < size ; i++) {
			dataset[i] = // with only 1 download option per file
				new GridData[] {
					createSimulatedGridData(consumer_peer_id,
											"_"+Integer.toString(i))
				};
		}
		
		return dataset;
		
	}
	
	// ****************************************************************

	private synchronized StringBuilder createFileName(String peer_id)
		throws GridException {
		
		StringBuilder sb_filename = new StringBuilder();
		sb_filename.append(peer_id);
		sb_filename.append(".simuser.simdata");
		sb_filename.append(CommonStringHandling.
							fillWithHeadingZerosLimit(this.next_salt,
													  this.max_files_in_bot));

		if (this.random_salt) {
			this.next_salt =
				CommonMath.sampleUniformDist(1,8);
		}
		else {
			this.next_salt = (++this.next_salt)%this.max_files_in_bot;
		}
		
		return sb_filename;
		
	}

	private synchronized long createFileSize(String filename)
		throws GridException {
		
		if (filename == null) {
			throw new GridException("createFileSize(): illegal file name");
		}
	
		return CommonMath.sampleRandomFileSize(filename);
		
	}

	// ****************************************************************

	private final boolean random_salt;
	private final int max_files_in_bot;
	private int next_salt;

	// ****************************************************************

}

