/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SlottedLRUCache.java

package lbg.protocol.data;

import java.util.concurrent.LinkedBlockingQueue;

/** This class implements a cache of a size bounded in number of entries.
 * 
 * @author Xavier Dalem
 *
 * @param <K> The type of the entries
 */
public class SlottedLRUCache<K extends Comparable<? extends K>> extends Cache<K> {
	private final LinkedBlockingQueue<K> usage;
	private final int capacity;
	
	public SlottedLRUCache(int capacity) {
		super();
		
		this.capacity = capacity;
		usage = new LinkedBlockingQueue<K>(capacity);
	}
	
	/** Adds an entry to the cache.
	 * <p>If the cache is full, an entry will be evicted if possible.
	 * If the entry was already present, it is updated as having been used
	 * recently (i.e. put on top of the LRU queue).</p>
	 * <p>Setting meta data for an existing entry doesn't replace the existing
	 * metadata. Either {@link #getMetadata(Comparable)} and modify it or
	 * {@link #remove(Comparable)} then add it back.</p>
	 * 
	 * @param entry The entry to add
	 * @param meta Metadata concerning this entry
	 * @return The entry that was popped, if any
	 * @throws VetoException if one entry should be popped but all are veto'ed
	 */
	@Override
	// TODO : maybe this should reset the time in the metadata
	public Metadata<K> add(K entry, Metadata<K> meta) throws VetoException,
			IllegalStateException {
		if (entry == null)
			throw new NullPointerException("null entry");
		if (meta == null)
			throw new NullPointerException("null metadata");
		
		Metadata<K> removedMeta = null;
		
		// Pop an entry if full and the new one wasn't know
		if (isFull() && !entries.containsKey(entry)) {
			K removedEntry = selectForRemoval();
			removedMeta = getMetadata(removedEntry);
			remove(removedEntry);
		}
		
		assert((! isFull()) || entries.containsKey(entry));
		
		// If we previously had that entry, we have to requeue it
		if (entries.containsKey(entry)) {
			final boolean popped = usage.remove(entry);
			assert(popped);
		} else // New one, insert it into entry list
			entries.put(entry, meta);
		
		final boolean inserted = usage.offer(entry);
		assert(inserted);
		
		return removedMeta;
	}

	@Override
	public boolean canAdd(Metadata<K>[] entries, boolean mindVeto) {
		if (mindVeto)
			// We can add them only if the general capacity, minus all data
			// that absolutely have to stay, is still enough to store all entries
			return ((capacity - veto.size()) >= entries.length);
		else
			return capacity >= entries.length;
	}

	@Override
	public boolean isFull() {
		return usage.size() == capacity;
	}

	@Override
	public Metadata<K> remove(K entry) throws VetoException {
		final Metadata<K> ret = super.remove(entry);
		final boolean changed = usage.remove(entry);
		
		assert ((ret == null) || ((ret != null ) && changed)) : "If entry was "
			+ "known, usage should have been changed";
		
		return ret;
	}

	@Override
	public K selectForRemoval() throws VetoException {
		if (isEmpty())
			return null;
		
		// We cannot check if veto.size() is equal to capacity : not all
		// veto'ed entries have to be in the entry list, so just because they
		// have the same size doesn't mean they have the same content
		
		K remove = null;
		for (K key: usage) {
			if (! veto.contains(key)) {
				remove = key;
				break;
			}
		}
		
		if (remove == null)
			throw new VetoException("All data is veto'ed, cannot select one"
				+ " for removal");
		
		return remove;
	}
	
	/** Returns a String representation of this cache */
	public String toString() {
		return "Slotted LRU cache (" + size() + "/" + capacity + " slots used)";
	}
	
	public String toStringDetailled() {
		StringBuilder builder = new StringBuilder(toString());
		
		builder.append("\n\tcached: ").append(usage.toString());
		builder.append("\n\tvetoed: ").append(veto.toString());
		
		return builder.toString();
	}

}
