/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ThreadedDownloader.java

package lbg.protocol.data;

import lbg.common.exception.GridException;

import java.util.concurrent.atomic.AtomicBoolean;

/** A ThreadedDownloader can handle a queue of GridData.
 * This object does <b>not</b> download the queued items in parallel. Its aim
 * is to allow the sequential download of a set of data's in a unique parallel
 * thread.
 * @author Xavier Dalem
 */
class ThreadedDownloader implements Runnable {
	/** Currently read data in queue */
	private int current = 0;
	/** Queue of scheduled-for-downloading data's */
	private GridData[] queue = null;
	private AtomicBoolean stopped = new AtomicBoolean(false);
	private boolean queue_has_changed = true;
	/** Holds the last exception while retrieving data. */
	private GridException excep = null;
	/** The GridDataManager this Downloader should retreive its files to */
	private final GridDataManager gdm;
	
	/**
	 * @param gdm The GridDataManager where the downloader should store the
	 * files.
	 */
	ThreadedDownloader(GridDataManager gdm) {
		this.gdm = gdm;
		Thread t = new Thread(this);
		t.setDaemon(true);
		t.start();
	}
	
	/** Sets the files that this downloader should try to get. Replaces the
	 * old queue (the current file is downloaded before the new queue is
	 * handled).
	 *
	 * @param items An array of GridData's (not options)
	 * @see GridDataManager#extract(GridData[][])
	 */
	synchronized void setDownloadQueue(GridData[] items)
			throws GridException {
		if (stopped.get())
			throw new GridException("Downloader asked to download items "
				+ "but was previously stopped");
		
		queue = items;
		queue_has_changed = true;
		notify();
	}
	
	/** Stops the thread when it's done with the current queue.
	 * A stopped downloader cannot handle a new queue.
	 */
	synchronized void stop() {
		stopped.set(true);
		queue_has_changed = true; // fast-stop
		notify();
	}
	
	/** @return <tt>null</tt> If everything went fine downloading the
	 * last WS.
	 */
	synchronized GridException getException() {
		return excep;
	}
	
	/* As long as we're not asked to stop :
	 *  - Wait for the queue to be nonempty, or a new one if we finished
	 *    the current one.
	 *  - For each item in the queue, try to download it. If it fails, nvm
	 * If the queue changes before we finish downloading it, just restart
	 * from the begining. If the data that has just be downloaded is
	 * useless, store() will take care of it.
	 *
	 * We could use an Executor to download all data at once. If so,
	 * we'd have to remove all thread from it when changing the queue. This
	 * parallel (and not sequential) behaviour would be interesting for
	 * download times but not for bandwidth: getFile() is blocking (and
	 * should stay that way) so all started files would have to be finished.
	 * Besides, we'd run into a problem: we could start downloading new
	 * files (high priority) at the same time as old files (low priority)
	 * and they would be competing for cache with the same priority, which
	 * would be plain wrong as an old file might then eject a new one. Yuk!
	 */
	public void run() {
		while (! stopped.get()) {
			
			synchronized (this) {
				while (queue == null)
					try { wait(); } catch (InterruptedException e) {}
				
				queue_has_changed = false;
			}
			
			this.excep = null;
			/* The synchronisation is messy but should be functionnal
			 * Since store() can take a while, we don't want to keep the lock
			 * for the whole loop. Hence, before every access, we check
			 * consistency.
			 */
			for (current = 0; current < queue.length; current++) {
				GridData gd = null;
				synchronized (this) {
					if (! queue_has_changed)
						gd = queue[current];
					else {
						current = 0;
						queue_has_changed = false;
						break;
					}
				}
				
				try {
					gdm.store(gd, false);
				} catch (GridException e) {
					this.excep = e;
					// Threads waiting on this were awaken by store()
				}
			}
			queue = null; // unset queue, useful to wait for events
		}
	}
}