/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// TorrentGridData.java
package lbg.protocol.data;

import lbg.common.exception.*;
import lbg.common.math.*;

/* Azureus Imports */
import com.aelitis.azureus.core.AzureusCore;
import com.aelitis.azureus.core.AzureusCoreFactory;
import org.gudy.azureus2.core3.disk.DiskManagerFileInfo; 
import org.gudy.azureus2.core3.download.DownloadManager;
import org.gudy.azureus2.core3.download.DownloadManagerFactory;
import org.gudy.azureus2.core3.download.DownloadManagerListener;
import org.gudy.azureus2.core3.torrent.TOTorrent;
import org.gudy.azureus2.core3.torrent.TOTorrentException;
import org.gudy.azureus2.core3.torrent.TOTorrentFactory;

import java.io.File;
import java.util.Map;

/** Grid data class where the data is shared using BitTorrent.
 * The file becomes available as soon as this object is instanciated.<br>
 * <b>This heavily relies on Azureus internals and could be broken with
 * numerous versions of Azureus.</b> (it actually broke on 2.5) I should find
 * another library (which doesn't seem to exist) or code my own (which I don't
 * have time for).<br>
 * @author Xavier Dalem
 */
public class TorrentGridData extends GridData
							 implements DownloadManagerListener {
	static final long serialVersionUID = -4028309881505044163L;
	
	private Map torrent;
	private File disk_file;
	private byte[] checksum;
	private transient DownloadManager dManager;
	
	/** Create a Torrent Grid Data from an id and a file
	 * @param id The file ID. For restrictions on this, see GridData
	 * @param az_serialized_map An azureus-serialized TOTorrent map. Such a map
	 * can be got from TOTorrent.serialiseToMap() and used with
	 * TOTorrentFactory.deserialiseFromMap(), after which you have to write the
	 * file as a .torrent on disk then reload it wih a TorrentDownloader
	 * because, somehow, the TorrentDownloader has no method to use an existing
	 * torrent and rather finds the data by himself...
	 * @param checksum The GridData.HASHING_ALGORITHM checksum for data
	 * @param manager The Download manager that was used to share the data
	 * when it was first tracked. This won't be used by {@link #toFile(File)}
	 * and only serves as a way to keep all useful information on the tracking
	 * side.
	 */
	protected TorrentGridData(String id, long size,
			java.util.Map az_serialized_map, byte[] checksum,
			DownloadManager manager) throws GridException {
		super(id, size);
		this.torrent = az_serialized_map;
		this.disk_file = null;
		this.checksum = checksum;
		this.dManager = manager;
	}
	
	void setLocalFileCopy(File file) {
		disk_file = file;
	}
	
	@Override
	public synchronized void toFile(File file) throws GridException {
		if (file == null)
			throw new GridException("TorrentGridData: null output file");
		// If the file exists, it's perfectly fine : this method can be used to
		// re-seed an existing file
		
		AzureusCore core = null;
		try {
			core = AzureusCoreFactory.getSingleton();
		}
		catch (com.aelitis.azureus.core.AzureusCoreException e) {
			throw new GridException("Core was not instanciated", e);
		}
		assert(core != null);
	
		// Dirty part, but seems unavoidable for Azureus
		File torrent_file = new File(file.getAbsolutePath() + "_"
			//+ new java.util.Random(new java.util.Date().getTime()).nextInt()
			+ CommonMath.sampleUniformDist()
			+ ".torrent");
		torrent_file.deleteOnExit();
		if (torrent_file.exists())
			throw new GridException("Could not save torrent file before"
				+ " downloading " + id);
		try {
			TOTorrent t = TOTorrentFactory.deserialiseFromMap(torrent);
			t.serialiseToBEncodedFile(torrent_file);
		} catch (TOTorrentException e) {
			throw new GridException("Could not save torrent file before"
				+ " downloading " + id, e);
		}
		
		// Azureus cannot set length on a non-existing file. So create an empty
		// file before him
		// Dunno if it's a hack but I didn't read anything about this
		if (! file.exists()) {
			try {
				java.io.FileOutputStream fos =
					new java.io.FileOutputStream(file);
				fos.close();
			} catch (java.io.IOException e) {
				throw new GridException("Could not create file " + id +
					" before downloading");
			}
		}
		assert(file.exists());
		
		dManager = DownloadManagerFactory.create(
			core.getGlobalManager(), null, torrent_file.getAbsolutePath(),
			file.getAbsolutePath(), DownloadManager.STATE_WAITING, false,
			false, null, null);
		dManager.initialize();
		dManager.addListener(this);
		//debugging code -- dManager.addListener(AzureusListener.getInstance());
		
		GridDataManager.DATALOGGER.finer("Started downloading " + id
			+ " from BitTorrent");
		
		while (dManager.getState() < DownloadManager.STATE_SEEDING) {
			try {
				wait();
			}
			catch (InterruptedException e) {}
			
			if (dManager.getState() == DownloadManager.STATE_READY) {
				dManager.startDownload();
			}
			/* debugging code --
			AzureusListener.getInstance().reportDownload(dManager);
			try { Thread.sleep(1000); } catch (InterruptedException e) {}*/
		}
		
		// We don't want to be awoken anymore by the possible seeding
		dManager.removeListener(this);
		//debugging code -- dManager.removeListener( AzureusListener.getInstance());
		
		switch (dManager.getState()) {
			case DownloadManager.STATE_SEEDING:
			case DownloadManager.STATE_STOPPED:
				break;
			case DownloadManager.STATE_ERROR:
				throw new GridException("Could not download " + id + ": "
					+ dManager.getErrorDetails());
			default:
				throw new GridException("Could not download " + id
					+ " for unknown reason");
		}
		
		disk_file = file;
		
		GridDataManager.DATALOGGER.finer("Finished downloading " + id
			+ " from BitTorrent");
	}
	
	/** Get the Download manager for this TorrentGridData.
	 * Quite a hack but Azureus doesn't offer a lot of alternatives
	 */
	DownloadManager getDownloadManager() {
		return dManager;
	}
	
	@Override
	public byte[] getChecksum() throws GridException {
		/* The following method cannot be used because it led to a bug where
		 * data cached in a previous session would never be recognised. The
		 * hash carried by a torrent doesn't depend only on the data, which
		 * would be the desired behaviour */
		/*Object o = torrent.get("info");
		if (o == null || (! (o instanceof java.util.Map)))
			throw new GridException("Hash for file " + id()
				+ " cannot be checked");
		
		o = ((java.util.Map)o).get("pieces");
		if (o == null || (!(o instanceof byte[])))
			throw new GridException("Hash for file " + id()
				+ " cannot be checked");*/
		
		return checksum;
	}
	
	/** @throws GridException If the file doesn't exist */
	@Override
	public java.io.InputStream getInputStream() throws GridException {
		if (disk_file == null)
			throw new GridException("File " + id + " was not retrieved");
		
		java.io.InputStream is = null;
		try {
			is = new java.io.FileInputStream(disk_file);
		} catch (java.io.FileNotFoundException e) {
			/* Note: this could happen if the data was stored in a file on a
			 * remote host. disk_file is not NULL, but there's no local
			 * copy of the file either. That's what setLocalFileCopy is for,
			 * btw.
			 * We can't erase the reference to the file at serialization time
			 * because passing a GridData between processes using pipes
			 * is legal.*/
			throw new GridException("Could not find file " + id, e);
		}
		
		return is;
	}
	
	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("TorrentGridData ");
		sb.append(torrent);
		
		if (disk_file != null) {
			sb.append(" at ");
			sb.append(disk_file.getAbsolutePath());
		}
		else {
			sb.append(", not yet stored");
		}
		
		return sb.toString();
		
	}
	
	/* DownloadManagerListener methods */
	public synchronized void stateChanged(DownloadManager manager,int state) {
		
		processEvent();
		
	}
	
	public synchronized void downloadComplete(DownloadManager manager) {
		
		processEvent();
		
	}
	
	public synchronized void completionChanged(DownloadManager manager,
											   boolean bCompleted){
	
		processEvent();
		
	}
	
	public synchronized void positionChanged(DownloadManager download,
											 int oldPosition,
											 int newPosition) {
		
		processEvent();
		
	}
	
	public synchronized void filePriorityChanged(DownloadManager download,
												 DiskManagerFileInfo file) {
	
		processEvent();

	}
	
	/* Rather than actively listening for changes, we set ourselves as a
	 * passive listener*/
	private void processEvent() {
		
		notifyAll();
		
	}

}

