/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// RemoteCall.java

package lbg.protocol.rpc;

import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.common.net.*;
import lbg.peer.service.LivePeerHandle;
import lbg.searchengine.service.LiveSearchEngineHandle;

import java.io.*;
import java.lang.reflect.Method;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.util.logging.Logger;

/** A class implementing simple and protected calls to distant host.
 * This class makes use of <code>java.lang.reflect</code>, and therefore allows
 * polymorphism
 * @author Xavier Dalem, Cyril Briquet
 */
public class RemoteCall implements Serializable {
	static final long serialVersionUID = -1480906902821136093L;
	
	/** Who do we make the call to ? */
	public enum CallType {
		PEER, USER, RESOURCE, SEARCH
	}
	
	private CallType destination;
	private String methodName;
	private Object[] params;
	private Class[] paramtypes;
	
	/**
	 * Creates a remote call.<br>
	 * <b>Warning:</b> you could experience problem with runtime types (e.g.
	 * with interfaces)
	 * @param destination The node we are talking to
	 * @param methodName The name of the method
	 * @param params The paramaters for called function
	 * @see #RemoteCall(CallType type, String methodName, Class[] paramtypes,
	 * Object[] params)
	 */
	public RemoteCall(CallType destination, String methodName,
			Object... params) throws GridException {
		if (destination == null || methodName == null || methodName.equals(""))
			throw new GridException("RemoteCall(): invalid parameters");
		this.destination = destination;
		this.methodName = methodName;
		this.params = params;
		
		// Get all param types;
		paramtypes = new Class[params.length];
		for (int i = 0; i < params.length; i++)
			paramtypes[i] = params[i].getClass();
	}
	
	/**
	 * Creates a remote call.
	 * This constructor, unlike RemoteCall(Calltype, String, Object...),
	 * is type-safe.
	 * @param destination The node we are talking to
	 * @param methodName The name of the method
	 * @param paramtypes The classes of the parameters
	 * @param params The paramaters for called function
	 * @see #RemoteCall(CallType type, String methodName, Object... params)
	 */
	public RemoteCall(CallType destination, String methodName,
		Class[] paramtypes, Object[] params) throws GridException {
		if (methodName == null || paramtypes == null || params == null)
			throw new GridException("RemoteCall(): invalid parameters");

		this.destination = destination;
		this.methodName = methodName;
		this.params = params;
		this.paramtypes = paramtypes;
	}
	
	/** Gets a handle from a Peer just by knowing his host and port.
	 * <br>This is simply a shorthand. You can get also handle without this.
	 * @param host Peer host
	 * @param port Peer port
	 * @throws GridException if getting the handle was not possible
	 */
	protected static LivePeerHandle
		getPeerHandle(String host, int port) throws GridException {
		RemoteCall rc = new RemoteCall(CallType.PEER, "createHandle",
			new Class[0], new Object[0]);
		
		Object ret = rc.send(host, port, true);
		if (! (ret instanceof LivePeerHandle))
			throw new GridException("Peer " + host
				+ " didn't provide a valid handle");
		
		return (LivePeerHandle) ret;
	}
	
	/** Gets a handle from a search engine just by knowing his host and port.
	 * <br>This is simply a shorthand. You can get also handle without this.
	 * @param host Search host
	 * @param port Search port
	 * @throws GridException if getting the handle was not possible
	 */
	protected static LiveSearchEngineHandle
		getSearchEngineHandle(String host, int port) throws GridException {
		RemoteCall rc = new RemoteCall(CallType.SEARCH, "createHandle",
			new Class[0], new Object[0]);
		
		Object ret = rc.send(host, port, true);
		if (! (ret instanceof LiveSearchEngineHandle))
			throw new GridException("Search engine " + host
				+ " didn't provide a valid handle");
		
		return (LiveSearchEngineHandle) ret;
	}
	
	/**
	 * Makes the <code>obj.methodName(params)</code> call.
	 * @param obj The object to make the call on (will be type-checked)
	 * @throws GridException if the call is not supported by that object.
	 * @throws java.lang.reflect.InvocationTargetException if the underlying
	 * method throws an exception.
	 * @return The result of the invoked method
	 */
	public Object doCall(Object obj) throws GridException,
		java.lang.reflect.InvocationTargetException {
		Class<?> c = null;
		try {
			switch (destination) {
				case PEER:
					c = Class.forName("lbg.peer.service.LivePeerService");
					break;
				case RESOURCE:
					c = Class.forName("lbg.resource.service.LiveResourceService");
					break;
				case USER:
					c = Class.forName("lbg.user.service.LiveUserService");
					break;
				case SEARCH:
					c = Class.forName("lbg.searchengine.service.LiveSearchEngineService");
					break;
			}
		} catch (ClassNotFoundException e) {
			throw new GridException("Critical error: could not find one of the"
				+ " base classes", e);
		}
		
		Object ret = null;
		try {
			Method m = c.getDeclaredMethod(methodName, paramtypes);
			if (m == null)
				throw new GridException("Could not find method " +
					destination.toString() + "." + methodName);
			
			ret=m.invoke(obj,params);
		} catch (java.lang.reflect.InvocationTargetException ite) {
			throw ite;
		} catch (Exception nsme) {
			throw new GridException("Invalid " + toString(), nsme);
		}
		
		return ret;
	}
	
	/**
	 * Opens a connection to host:port and send the call
	 * @param host The host to send it to
	 * @param port The port on where to send it
	 * @param do_return True if we wait for a return value
	 * @return The remote call return value, or null if do_return is false
	 * @throws GridException if the parameters are invalid
	 * @throws GridNetworkException if the remote host couldn't be reached or
	 * didn't format its messages as expected.
	 */
	public Object send(String host, int port, boolean do_return)
		throws GridException {
		
		return send(host,port,do_return,true);
		
	}
	
	/**
	 * Opens a connection to host:port and send the call
	 * @param host The host to send it to
	 * @param port The port on where to send it
	 * @param do_return true if we wait for a return value
	 * @param do_timeout false to deactivate the transmission timeout
	 * @return The remote call return value, or null if do_return is false
	 * @throws GridException if the parameters are invalid
	 * @throws GridNetworkException if the remote host couldn't be reached or
	 * didn't format its messages as expected.
	 */
	public Object send(String host, int port, boolean do_return,
					   boolean do_timeout) throws GridException {
	
		return send(host,port,do_return,do_timeout,false);

	}
	
	/**
	 * Opens a connection to host:port and send the call
	 * @param host The host to send it to
	 * @param port The port on where to send it
	 * @param do_return true if we wait for a return value
	 * @param do_timeout false to deactivate the transmission timeout
	 * @param do_failfast true to retry only once, and immediately,
	 * in case of comm. failure
	 * @return The remote call return value, or null if do_return is false
	 * @throws GridException if the parameters are invalid
	 * @throws GridNetworkException if the remote host couldn't be reached or
	 * didn't format its messages as expected.
	 */
	public Object send(String host, int port, boolean do_return,
					   boolean do_timeout,boolean do_failfast)
		throws GridException {
	
		if (host == null || host.equals("")
				|| !CommonNetHandling.isValidPort(port))
			throw new GridException("RemoteCall.send: invalid parameters");
		
		// Connect to peer
		Socket socket = null;
		ObjectOutputStream oos = null;
		ObjectInputStream ois = null;
		Object retval = null;

		boolean is_data_sent = false;
		final int timeout = 4000;
		int exception_counter = 0;
		final int exception_threshold = 3;
		while (is_data_sent == false) {
			try {
				// Establish connection
				try {
					socket = new Socket(host, port);
					if (do_timeout) {
						socket.setSoTimeout(timeout);
					}
					// Both streams must be obtained, even if no return value
					// is expected, else the data will get corrupted
					oos = new ObjectOutputStream(socket.getOutputStream());
					ois = new ObjectInputStream(socket.getInputStream());
				}
				catch (IOException e) {
					throw e;
				}
			
				// Send request
				try {
					oos.writeObject(this);
					oos.flush();
				} catch (IOException e) {
					if (e instanceof SocketTimeoutException) {
						throw e;
					}
					throw new GridNetworkException("Could not send request to "+
												   host+":"+ port,e);
				}
				
				// Get return value
				if (do_return) {
					try {
						retval = ois.readObject();
						if (retval == null) {
							throw new IOException("send(): illegal return value");
						}
					}
					catch (IOException e) {
						throw e;
					}
					catch (ClassNotFoundException e) {
						throw new GridException(e);
					}
				}
		
				// Close connection
				try { ois.close(); } catch (IOException e) { }
				try { oos.close(); } catch (IOException e) { }
				try { socket.close(); } catch (IOException e) { }
				
				is_data_sent = true;
			}
			catch (IOException e) {
				if (socket != null) {
					try { socket.close(); } catch (IOException cursed) {}
				}
				if (oos != null) {
					try { oos.close(); } catch (IOException cursed) {}
				}
				if (++exception_counter >= exception_threshold) {
					if (e instanceof SocketTimeoutException == false) {
						throw new GridNetworkException("RemoteCall.send(): "+
												   "could not connect to " +
												   host + ":" + port, e);
					}
				}

				if (do_failfast) {
					exception_counter = exception_threshold;
					continue;
				}
				
				int sleep_len = CommonMath.sampleUniformDist(2,12);

				// get Logger explicitly because this code executes in
				// the Peer using the handle, not in the Peer that created the handle
				Logger PEER_LOGGER = Logger.getLogger("lbg.peer");
				PEER_LOGGER.info("RemoteCall.send(): connection timed-out, "+
								 "backoff for "+sleep_len+
								 " seconds before retrying");
				
				ThreadPool.sleep(sleep_len);
				continue;
			}
		}
			
		return retval;
	}
	
	public CallType getType() {
		return destination;
	}
	
	/** Sums up the call as a string */
	public String toString() {
		StringBuilder sb = new StringBuilder("RemoteCall to ");
		switch (destination) {
			case PEER:
				sb.append("LivePeerService");
				break;
			case RESOURCE:
				sb.append("LiveResourceService");
				break;
			case USER:
				sb.append("LiveUserService");
				break;
			default:
		}
		sb.append("." + methodName + "(");
		for (int i = 0; i < paramtypes.length; i++) {
			sb.append(paramtypes[i].getName());
			if (i <  paramtypes.length - 1)
				sb.append(", ");
		}
		sb.append(")");
		return sb.toString();
	}
	
}

