/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// RemoteCallDaemon.java

package lbg.protocol.rpc;

import lbg.common.net.*;
import java.io.*;
import java.net.*;
import java.util.concurrent.Semaphore;
import java.util.logging.Level;
import java.util.logging.Logger;

import lbg.common.exception.*;

/**
 * A generic daemon able to handle calls from and on all entities : peers,
 * users and ressources. Replaces the now obsolete individual and
 * semi-individual daemons
 * @author Xavier Dalem
 */
public class RemoteCallDaemon implements Runnable {
	private Object callHandler;
	private int target_port;
	private RemoteCall.CallType target_type;
	private RemoteCallStatusListener listener = null;
	private static int DEFAULT_HANDLERS = 4;
	private int handlers;
	protected Logger logger;
	
	/** Create a daemon with the default number of handlers (does not start
	 * thread).
	 * @param target_type The type of calls we accept (all other will be
	 * discarded)
	 * @param target_port The port to listen on
	 * @param callHandler An object that is able to handle the call of relevant
	 * type (target_type)
	 */
	public RemoteCallDaemon(RemoteCall.CallType target_type, int target_port,
		Object callHandler, RemoteCallStatusListener l) throws GridException {
		this (target_type, target_port, callHandler, l, DEFAULT_HANDLERS);
	}
	
	/** Creates a daemon (does not start thread)
	 * @param target_type The type of calls we accept (all other will be
	 * discarded)
	 * @param target_port The port to listen on
	 * @param callHandler An object that is able to handle the call of relevant
	 * type (target_type)
	 * @param handlers The number of thread handlers
	 * @param listener A listener for daemon status changes (null for none)
	 */
	public RemoteCallDaemon(RemoteCall.CallType target_type, int target_port,
		Object callHandler, RemoteCallStatusListener listener, int handlers)
			throws GridException {
		if (callHandler == null || !CommonNetHandling.isValidPort(target_port)
			|| handlers < 1 || target_type == null)
			throw new GridException("RemoteCallDaemon: invalid arguments");
		this.callHandler = callHandler;
		this.target_port = target_port;
		this.target_type = target_type;
		this.handlers = handlers;
		this.listener = listener;
		
		// Report on correct logger depending on who called us.
		switch(target_type) {
			case PEER: logger = Logger.getLogger("lbg.peer");
				break;
			case RESOURCE: logger = Logger.getLogger("lbg.resource");
				break;
			case USER: logger = Logger.getLogger("lbg.user");
				break;
			case SEARCH: logger = lbg.searchengine.SearchEngineServer.SELOGGER;
				break;
		}
	}
	
	/** Launch the daemon. */
	public void run() {
		ServerSocket server = null;
		try {
			server = new ServerSocket(target_port);
		}
		catch (IOException e) {
			String this_hostname = "unknown";
			try {
				this_hostname = CommonNetHandling.getLocalHostName();
			}
			catch (Exception doomed) {}
			logger.log(Level.SEVERE,"Failed to create socket "+
									 this_hostname+":"+target_port, e);
			return;
		}
		
		// Create handlers pool
		String tmpname = Thread.currentThread().getName();
		final String myName = 
			tmpname.substring(0,tmpname.lastIndexOf("-daemon"))+
			"-service-daemon-";
		Semaphore available = new Semaphore(0);
		RemoteCallHandler[] pool = new RemoteCallHandler[handlers];
		for (int i = 0; i < handlers; i++)
			pool[i] =
			new RemoteCallHandler(available,callHandler,target_type,myName+i);
		// Wait for pool to be ready
		available.acquireUninterruptibly(handlers);
		// Ready ? restore state
		available.release(handlers);

		// Tell the listener we're ready
		if (listener != null)
			listener.isOnline();
		
		// wait for inbound connections
		while (true) {
			try {
				Socket client = null;

				// Get incoming call
				try {
					client = server.accept();
				}
				catch (IOException e) {
					throw new GridException(
							"Could not open connection to unknown host", e);
				}

				// Wait for a handler to be ready
				available.acquireUninterruptibly();
				// Find which one it was
				int candidate = 0;
				for (; candidate < handlers; candidate++)
					if (pool[candidate].idle())
						break;
				// We can't make a whole rotation without finding a good one
				assert(candidate != handlers);
				
				pool[candidate].handle(client);
			}
			catch (GridException e) {
				// We log everything from one place.
				// For better service, no error is fatal.
				logger.warning(e.getMessage() + ", stack trace follows");
					if (e.getCause() != null)
						logger.log(Level.WARNING, e.getMessage(), e.getCause());
					else
						logger.warning(e.getMessage());
			}
		}
		
		//try { server.close(); } catch (IOException e) {}
		// Useless as loop doesn't break
	}
	
	// ************************************************************************
	
	/** A handler for incoming calls. It is autostarted and waits for calls to
	 * be activated (this saves the cost of starting new Threads each time).
	 */
	private class RemoteCallHandler implements Runnable {
		/* Only one Thread will ever wait on this object in this implementation
		 * and it is the one started by this RemoteCallHandler.run().
		 * Some parts of the methods should be synchronized to be sure that we
		 * manage idling effectively
		 */
		private Socket socket;
		private boolean go;
		private final Semaphore avail;
		private final Object handler;
		private final RemoteCall.CallType target_type;
		
		/** Creates and starts a handler.
		 * 
		 * @param avail The semaphore on which the handler will announce it is
		 * idle and ready for provessing a new request
		 * @param handler The service handling the incoming requests
		 * @param target_type The type of calls this handler can answer
		 * @param name The name of this handler's Thread
		 */
		RemoteCallHandler(Semaphore avail, Object handler,
				RemoteCall.CallType target_type, String name) {
			socket = null;
			go = true;
			this.avail = avail;
			this.handler = handler;
			this.target_type = target_type;
			Thread t = new Thread(this,name);
			t.setDaemon(true);
			t.start();
		}
		
		boolean idle() {
			return (socket == null);
		}
		
		/** Ask the RemoteCallHandler to stop running after handling the
		 * current request. If no request was submitted, handler stops
		 * running immediately */
		void stop() {
			go = false;
			notify();
		}
		
		/** Handles a request. */
		void handle(Socket s) throws GridException {
			if (s == null)
				throw new GridException("Cannot handle null socket");
			if (!idle())
				throw new GridException("Handler is not idle");
			if (!go)
				throw new GridException("Handler was asked to stop running");
			
			synchronized(this) {
				socket = s;
				notify();
			}
		}
		
		public void run() {
			while (go) {
				avail.release(); // We are now idle and ready to handle a
					             // new request
				
				// Wait for a request to be submitted
				synchronized(this) {
					// We could have been asked to handle a new request
					// between avail.release() and socket check
					while (socket == null) {
						try  { wait(); } catch (InterruptedException e) {}
					}
				}
				// By who were we awaken ? If stop(), go was set to false
				if (!go) {
					avail.acquireUninterruptibly(); // We must acquire what we
					break;                          // released
				}
				
				ObjectInputStream ois = null;
				ObjectOutputStream oos = null;
				// Handle request
				try {
					// Get streams
					try {
						ois = new ObjectInputStream(socket.getInputStream());
						oos = new ObjectOutputStream(socket.getOutputStream());
					}
					catch (IOException e) {
						throw new GridException("Could not open connection to "+
												socket.getInetAddress().
												toString(),e);
					}

					// Get request
					Object obj = null;
					try {
						obj = ois.readObject();
					} catch (Exception e) {
						throw new GridException("Error while reading object "
							+ "from " + socket.getInetAddress().toString(),e);
					}

					if (! (obj instanceof RemoteCall)) {
						throw new GridException("Object received from "
							+ socket.getInetAddress().toString()
							+ " is not a RemoteCall");
					}

					RemoteCall rc = (RemoteCall) obj;
					obj = null;

					// We can only handle calls to peer
					if (rc.getType() != this.target_type) {
						throw new GridException("Call type " + rc.getType()
							+ " requested by "
							+ socket.getInetAddress().toString()
							+ " is not handled by this daemon");
					}

					// Handle the request
					Object ret = null;
					try {
						ret = rc.doCall(handler);
					} catch (GridException ge) {
						throw new GridException(
							"Could not make call requested by "
							+ socket.getInetAddress().toString() + ": "
							+ ge.getMessage(), ge.getCause());
					} catch (java.lang.reflect.InvocationTargetException ite) {
						throw new GridException(
							"Could not make call requested by "
							+ socket.getInetAddress().toString(),
							ite.getCause());
						/* The ITE itself is utterly uninformative */
					}

					// Return
					try {
						if (ret != null) {
							oos.writeObject(ret);
							oos.flush();
						}
					}
					catch (IOException e) {
						throw new GridException("Could not send return value to"
							+ " " + socket.getInetAddress().toString(), e);
					}

					// Close the connection
					try { oos.close(); } catch (IOException e) {}
					try { ois.close(); } catch (IOException e) {}
					try { socket.close(); } catch (IOException e) {}
				} catch (GridException e) {
					try { socket.close(); } catch (IOException doomed) {}
					if (ois != null) {
						try { ois.close(); } catch (IOException doomed) {}
					}
					if (oos != null) {
						try { ois.close(); } catch (IOException doomed) {}
					}
					// We log everything from one place.
					// For better service, no error is fatal.
					if (e.getCause() != null)
						logger.log(Level.WARNING, e.getMessage(), e.getCause());
					else
						logger.warning(e.getMessage());
				}
				
				oos = null;
				ois = null;
				// Don't forget this, it's what sets the idle status.
				socket = null;
			}
		}
	}
}

