/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LiveResource.java

package lbg.resource;

import java.util.logging.Level;
import java.util.logging.Logger;

import lbg.common.exception.*;
import lbg.peer.service.*;
import lbg.protocol.data.GridData;
import lbg.protocol.data.GridDataManager;
import lbg.protocol.rpc.*;
import lbg.resource.config.*;
import lbg.resource.execution.LiveTaskExecutionHandler;
import lbg.resource.service.*;
import lbg.task.*;

/** Implementation of a Live Resource.<br>
 * Related logger: lbg.resource
 * @author Xavier Dalem
 */
public class LiveResource extends Resource {
	
	private class ResourceRemover implements Runnable {
		private InternalPeerHandle peer;
		private ResourceHandle res;
		
		public ResourceRemover(InternalPeerHandle peer, ResourceHandle res) {
			this.peer = peer;
			this.res = res;
		}
		
		public void run() {
			try {
				flagAsRemoved();
				// Cancel running task
				peer.removeResource(res);
				RESLOGGER.fine("Removed resource " + res.id() + " from peer "
					+ peer.id());
			} catch (GridException e) {
				RESLOGGER.warning("Failed to properly remove resource from "
					+ "peer: " + e);
			}
		}
	}

	// ****************************************************************

	// We're going to use it more than once, better not ask it every time.
	static final Logger RESLOGGER = Logger.getLogger("lbg.resource");

	protected boolean is_removed = false;
	public synchronized boolean getRemovalState() { return is_removed; }
	protected synchronized void flagAsRemoved() { is_removed = true; }
	
	private GridDataManager gdm;
	private LiveTaskExecutionHandler lth;
	private ResourceDataConfiguration rdc;
	
	private Thread svc_thread;
	private boolean online;
	private Logger logger;

	// ****************************************************************
	
	/** Creates a Live resource
	 * Resource is not yet connected to peer.
	 * @param rbc The basic configuration for the resource
	 * @param rnc The network configuration for the resource
	 * @param rdc The data configuration for the resource
	 * @param owner_peer A handle to the peer that manages this resource
	 * @see #connect()
	 */
	public LiveResource(ResourceBasicConfiguration rbc,
						ResourceNetworkConfiguration rnc,
						ResourceDataConfiguration rdc,
						InternalPeerHandle owner_peer) throws GridException {
		
		super(rbc,rdc,rnc,owner_peer);
		
		if (rdc == null) {
			throw new GridException("LiveResource(): invalid data config");
		}
		
		this.logger = Logger.getLogger("lbg.resource"); // .id
		this.rdc = rdc;
		
		// Be nice to the peer when going away
		Runtime.getRuntime().addShutdownHook(new Thread(new ResourceRemover(
			owner_peer, super.service().createHandle())));
			
		this.gdm = new GridDataManager(super.netConfig().hostname(),
									   rdc.CACHE_DIR(),
									   super.netConfig().btDataPort(),
									   super.netConfig().ftpDataPort(),
									   null, 0,
									   rdc.SUPPLYING_CACHE_CAPACITY(),
									   0, rdc.STATS_DIR());
		
		this.svc_thread = null;
		this.online = false;

	}

	/** Connects the Resource to its owner Peer.
	 * <p>This cannot be done at creation-time because we maybe need another
	 * infrastructure (e.g. Sockets) before telling the Peer we're ready to
	 * process Tasks.</p>
	 */
	public void connect() throws GridException {
		super.connect();
		
		RESLOGGER.fine("Added resource " + id() + " to peer "+owner().id());
	}

	public synchronized String toString() {
		StringBuilder sb = new StringBuilder(super.toString());
		
		return sb.toString();
	}

	public ResourceDataConfiguration getDataConfig() {
		return this.rdc;
	}
	
	public GridDataManager gdm() {
		return this.gdm;
	}

	public synchronized boolean setWorkingSet(GridData[][] working_set)
		throws GridException {
		GridData[] ext_ws = gdm.extract(working_set);
		Task run = runningTask();
		
		if (run != null) {
			GridData[] t_data = null;
			
			// Get data from running task
			synchronized (run) {
				try {
					t_data = this.gdm.extract(
						run.getRawInputDataSet());
				} catch (GridException e1) {
					try {
						t_data = run.getInputDataSet();
					} catch (GridException e2) {
						RESLOGGER.log(Level.WARNING, "Failed to set working "
							+ "set: couldn't get data's from running task", e2);
					}
				}
			}
			
			for (GridData gd : t_data) {
				int i;
				for (i = 0; i < ext_ws.length
					&& (! ext_ws[i].id.equals(gd.id)); i++);
				
				if (i == ext_ws.length)
					RESLOGGER.warning("Submitted working set doesn't include " +
						gd.id + ", needed for running task");
			}
		}
		
		return gdm.setWorkingSet(ext_ws);
	}

	/** Runs a given Task by getting and loading all required information,
	 * then running the methods as specified by GridApplication.<br>
	 * The method returns immediately, starting the Task in another Thread.
	 * @see lbg.protocol.api.GridApplication
	 * @param t The task to run
	 */
	public synchronized void runTask(Task t) throws GridException {
		super.runTask(t);
		
		int mem_size = 800; // 512;
		
		RESLOGGER.info("Started Task for "+ t.consumerID() +
					   " at "+
					   new java.util.Date(taskStartTime()*1000).toString());
		this.lth = new LiveTaskExecutionHandler(this,owner(),t,mem_size);
		new Thread(this.lth).start();
	}

	/** Make final operations to register the running task as completed
	 * @param data The output data of the completed task
	 */
	public void completeTask(GridData data,byte[] logs) throws GridException {
		RESLOGGER.info("Finished Task for "+runningTask().consumerID()+
					   " at "+new java.util.Date()+
					   " (Task started at "+
					   new java.util.Date(taskStartTime()*1000).toString()+")");
		super.completeTask(data,logs);
	}

	/** Cancel the currently running task */
	public synchronized void cancelTask() throws GridException {
		RESLOGGER.info("Cancelled Task " + runningTask().id() + " for "+
					   runningTask().consumerID() +
					   ", that started at "+
					   new java.util.Date(taskStartTime()*1000).toString());
		
		super.cancelTask();
		
		lth.abort();
	}
	
	/** How much time should running this task take ?
	 * @todo Yet to be implemented
	 */
	public synchronized long expectedTaskRuntime() throws GridException {
		RESLOGGER.warning("expectedTaskRuntime is not yet implemented");
		return 51;
	}
	
	public synchronized void startService() throws GridException {

		if (this.online == true) {
			return;
		}
		else {
			this.online = true;
		}
		
		final LiveResource self = this;
		RemoteCallStatusListener rsl =
			new RemoteCallStatusListener() {
				public void isOnline() {
					try {
						self.connect();
					}
					catch (GridException e) {
						Logger.getLogger("lbg.resource").
							log(Level.WARNING,"Connecting resource failed", e);
					}
				}
			};
		RemoteCallDaemon rc_daemon =
			new RemoteCallDaemon(RemoteCall.CallType.RESOURCE,
								 super.netConfig().resourcePort(),
								 self.service(),rsl);

		this.svc_thread = new Thread(rc_daemon,"Peer-to-Resource-daemon");
		this.svc_thread.start();
		
	}
	
	public synchronized boolean isAlive() {
		
		return (this.svc_thread != null) ? this.svc_thread.isAlive() : false;
		
	}

	public Logger logger() {
		
		return this.logger;
		
	}

	// ****************************************************************

}

