/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Resource.java

package lbg.resource;

import lbg.*;
import lbg.common.exception.*;
import lbg.peer.service.*;
import lbg.protocol.data.*;
import lbg.resource.config.*;
import lbg.resource.service.*;
import lbg.task.*;

import java.util.logging.Logger;

/**
 * @author Cyril Briquet
 */
public abstract class Resource {

	Resource(ResourceBasicConfiguration res_basic_config,
			 ResourceDataConfiguration res_data_config,
			 ResourceNetworkConfiguration res_net_config,
			 InternalPeerHandle owner_peer)
		throws GridException {
		
		if (res_basic_config == null) {
			throw new GridException("Resource(): illegal basic configuration");
		}
		if (res_data_config == null) {
			throw new GridException("Resource(): illegal data configuration");
		}
		if (res_net_config == null) {
			throw new GridException("Resource(): "+
									"illegal network configuration");
		}
		if (owner_peer == null) {
			throw new GridException("Resource(): illegal owner Peer");
		}

		this.id = res_basic_config.RESOURCE_ID();
		this.power = res_basic_config.RESOURCE_POWER();
		this.cache_capacity = res_data_config.SUPPLYING_CACHE_CAPACITY();
		this.res_net_config = res_net_config;
		
		this.owner_peer = owner_peer;
		
		this.idle = true;
		this.running_task = null;
		this.start_time = 0;
		
		this.resource_service =
			ResourceServiceFactory.createResourceService(this);

	}
	
	public String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[Resource ").
			append(this.id).
			append("] @ ").
			append(this.res_net_config.hostname()).
			append(":").
			append(this.res_net_config.resourcePort());
		
		return sb.toString();

	}

	// ****************************************************************

	public String id() {
		
		return this.id;
		
	}
	
	public ResourceNetworkConfiguration netConfig() {
	
		return this.res_net_config;
		
	}
	
	public InternalPeerHandle owner() {
		
		return this.owner_peer;
		
	}

	public int power() {
		
		return this.power;
		
	}

	public synchronized ResourceService service() {
		
		return this.resource_service;
		
	}

	public abstract Logger logger();

	// ****************************************************************

	/** Connects the Resource to its owner Peer.
	 * <p>This cannot be done at creation-time because we maybe need another
	 * infrastructure (e.g. Sockets) before telling the Peer we're ready to
	 * process Tasks.</p>
	 */
	public synchronized void connect() throws GridException {

		this.owner_peer.
			addResource(this.resource_service.createHandle(),this.power());
		
	}

	public int cacheCapacity() {
		
		return this.cache_capacity;
		
	}

	public abstract boolean setWorkingSet(GridData[][] working_set)
		throws GridException;

	
	// ****************************************************************

	public synchronized boolean idle() {
	
		return this.idle;
		
	}

	public synchronized Task runningTask() {
		
		return this.running_task;
		
	}

	public synchronized void runTask(Task t) throws GridException {
		
		if (this.idle == false) {
			throw new GridException("runTask(): busy Resource");
		}
		if (t == null) {
			throw new GridException("runTask(): illegal Task");
		}
		
		this.idle = false;
		this.running_task = t;
		t.setStartTime();
		this.start_time = LightweightBarteringGrid.getTime();
	
	}
	
	/** Completes the execution of a Task by:
	 * <ul>
	 * <li>flagging the Task as complete (thus, creating a TaskOutput)
	 * <li>updating the Resource state
	 * </ul>
	 */
	public synchronized void completeTask(GridData data,byte[] logs)
		throws GridException {
		
		if (this.idle == true) {
			throw new GridException("completeTask(): idle Resource");
		}

		// Flag Task as completed
		this.running_task.flagAsCompleted(this.owner_peer.id(),
										  this.start_time,data,logs,this.power);
		
		// Update Resource state
		Task t = this.running_task;
		this.idle = true;
		this.running_task = null;
		
		// Upload results to Peer
		this.owner_peer.uploadLocallyCompletedTask(t);

	}

	public synchronized void cancelTask() throws GridException {

		// Sanity check
		if (this.idle == true) {
			Logger logger = this.logger();
			String msg = "cancelTask(): idle Resource";
			if (logger == null) {
				throw new GridException(msg);
			}
			else {
				logger.info(msg);
			}
		}
		
		// Update Resource state
		this.idle = true;
		this.running_task = null;
		
	}

	// ****************************************************************

	public synchronized long taskStartTime() {
		
		return this.start_time;
		
	}

	public synchronized long expectedTaskRuntime() throws GridException {
		
		Task t = this.running_task;
		if (t == null) {
			throw new GridException("expectedTaskRuntime(): no Task running");
		}
		
		return Math.max(1,(long)Math.ceil(((double)t.nominalRuntime())/
										  this.power));
		
	}

	// ****************************************************************

	private final String id;
	private final int power;
	private final int cache_capacity;
	private final ResourceNetworkConfiguration res_net_config;
	private final InternalPeerHandle owner_peer;
	private       boolean idle;
	private       Task running_task;
	private       long start_time;
	private final ResourceService resource_service;

}

