/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SimulatedResource.java

package lbg.resource;

import lbg.common.exception.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.common.math.*;
import lbg.peer.service.*;
import lbg.protocol.data.*;
import lbg.resource.config.*;
import lbg.resource.service.*;
import lbg.simulator.event.*;
import lbg.task.*;

import java.util.logging.Logger;
import org.apache.commons.math.distribution.ExponentialDistributionImpl;

/**
 * @author Cyril Briquet
 */
public class SimulatedResource extends Resource {

	public SimulatedResource(ResourceBasicConfiguration res_basic_config,
							 ResourceDataConfiguration res_data_config,
							 ResourceReliabilityConfiguration res_rel_config,
							 ResourceNetworkConfiguration res_net_config,
							 InternalPeerHandle owner_peer,
							 EventList el)
		throws GridException {

		super(res_basic_config,res_data_config,res_net_config,owner_peer);
		
		if (res_rel_config == null) {
			throw new GridException("SimulatedResource(): illegal "+
									"Resource reliability configuration");
		}
		if (el == null) {
			throw new GridException("SimulatedResource(): illegal Event List");
		}

		this.rfg = new ResourceFailureGenerator(res_rel_config.MTBF(),this);
		this.el = el;
		this.current_evt = null;
		
	}

	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder(super.toString());

		sb.append("\n  ->  ").
			append("power=").
			append(super.power()).
			append(", status=").
			append(super.idle() ? "idle" : "busy").
			append("\n  ->  ").
			append(super.service());
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized void runTask(Task t) throws GridException {

		super.runTask(t);
		
		Event evt = null;
		if (this.rfg.willFailTaskExecution()) {
			evt = this.el.factory(). // create FailedTaskEvent
				createFailedTaskEvent(t.consumerID(),owner().id(),t,this);
		}
		else {
			evt = this.el.factory(). // create CompletedTaskEvent
				createCompletedTaskEvent(t.consumerID(),owner().id(),t,this);
		}

		this.current_evt = (TaskEvent)evt;
		this.el.addEvent(evt);
		
	}

	public synchronized boolean setWorkingSet(GridData[][] working_set)
		throws GridException {
	
		return true;
		
	}

	public synchronized void cancelTask() throws GridException {
		
		/* There are 4 situations when a Task may be cancelled:
		   (1) Peer-initiated cancellation, Task failure
		   (2) Peer-initiated cancellation, Task-to-be-completed time-out
		   (3) Peer-initiated cancellation, Task-to-be-failed time-out
		   (4) Simulator-initiated cancellation, Task failure
		       (will call the Peer, so that it calls back cancelTask() in (1))
		*/
		
		if (this.current_evt == null) {
			// Actually cancel the Task, because the Peer has been asked to,
			// in order to simulate Task failure
			super.cancelTask();
		}
		else /* if (this.current_evt != null) */ {

			if (this.current_evt instanceof CompletedTaskEvent) {
				// Actually cancel the Task
				super.cancelTask();
				// Remove CompletedTaskEvent from EventList
				// so that Task does not complete as initially expected
				this.el.removeEvent(this.current_evt);
				// Clear current Event
				this.current_evt = null;
			}

			else if (this.current_evt instanceof FailedTaskEvent) {
				// Let first suppose that the cancellation is Peer-initiated
				boolean is_task_failure = false;
				try {
					// Remove FailedTaskEvent from EventList
					// so that Task does not fail as initially expected
					this.el.removeEvent(this.current_evt);
				}
				catch (GridException e) {
					// Of course
					is_task_failure = true;
				}
				if (is_task_failure == false) {
					// Indeed, cancellation was Peer-initiated:
					// cancel Task before it fails
					super.cancelTask();
					this.current_evt = null;
					return;
				}
				
				// Now it is certain that the cancellation was
				// initiated by the Simulator so as to simulate Task failure
				
				// Clear current Event
				this.current_evt = null;
				// Ask the Peer to cancel the Task
				Task t = super.runningTask();
				String consumer_id = t.consumerID();
				if (consumer_id.equals(super.owner().id())) {
					// Local Task
					super.owner().
						preemptLocalTaskOnLocalResource(t,DUMMY_LOG);
				}
				else {
					// Supplying Task
					super.owner().
						preemptSupplyingTaskOnLocalResource(consumer_id,t,DUMMY_LOG);
				}
			}

		}
		
	}

	/**
	 * @return null
	 */	 
	public Logger logger() {
		
		return null;
		
	}
	
	// ****************************************************************

	private static class ResourceFailureGenerator {
		
		public ResourceFailureGenerator(int mtbf,Resource res)
			throws GridException {
			
			if (mtbf < 0) {
				throw new GridException("ResourceFailureGenerator(): "+
										"illegal MTBF");
			}
			if (res == null) {
				throw new GridException("ResourceFailureGenerator(): "+
										"illegal Resource");
			}
			
			this.mtbf = mtbf;
			this.is_perfectly_reliable = (this.mtbf == 0);
			this.exp_dist = (this.is_perfectly_reliable) ?
				null : new ExponentialDistributionImpl(this.mtbf);
			this.res = res;
			
		}
		
		/**	Stochastically selects Task execution completion or failure.
		 * Task execution failures are rare events and constitute
		 * a Poisson process, which is modelled here
		 * by an exponential distribution to determine actual failure probability.
		 */
		private synchronized boolean willFailTaskExecution()
			throws GridException {
			
			if (this.is_perfectly_reliable) {
				return false;
			}
			
			long expected_task_runtime = this.res.expectedTaskRuntime();
			
			double failure_probability = 0.0;
			try {
				failure_probability =
					this.exp_dist.cumulativeProbability(expected_task_runtime);
			}
			catch (org.apache.commons.math.MathException e) {
				throw new GridException("willFailTaskExecution(): "+
										"Jakarta Commons Math internal error, "+
										"probably numerical instability",e);
			}
			if ((failure_probability < 0.0) || (failure_probability > 1.0)) {
				throw new GridException("willFailTaskExecution(): "+
										"illegal failure probabibility");
			}
			
			double x = CommonMath.sampleUniformDistWithinUnity();

			//System.out.println(x+" "+failure_probability);
			
			return (x <= failure_probability);
			
		}
		
		private final double mtbf;
		private final boolean is_perfectly_reliable;
		private final ExponentialDistributionImpl exp_dist;
		private final Resource res;
		
	}

/* for testing purposes - refactor into a test unit
	private static class DummyResource extends Resource {

		public DummyResource(int task_runtime) {
		
			this.task_runtime = task_runtime;
		
		}
		
		public Logger logger() {
			
			return null;
			
		}
		
		public long expectedTaskRuntime() {
			
			return this.task_runtime;
			
		}
		
		public boolean setWorkingSet(GridData[][] working_set)
			throws GridException {
		
			return true;
			
		}
		
		private final int task_runtime;
		
	}

	Resource() {
	
		this.id = null;
		this.power = 0;
		this.cache_capacity = 0;
		this.res_net_config = null;
		this.owner_peer = null;
		this.idle = true;
		this.running_task = null;
		this.start_time = 0;
		this.resource_service = null;
	
	}

	// Usage: java -cp jars/lightweight_bartering_grid.jar lbg.resource.SimulatedResource mtbf task_runtime
	public static void main(String[] args) throws GridException {
		
		int mtbf = Integer.parseInt(args[0]);
		int task_runtime = Integer.parseInt(args[1]);
		DummyResource res = new DummyResource(task_runtime);
		ResourceFailureGenerator rfg =
			new ResourceFailureGenerator(mtbf,res);

		int n = 1000;
		int total_failures = 0;
		for (int i = 0 ; i < n ; ++i) {
			if (rfg.willFailTaskExecution()) {
				++total_failures;
			}
		}
		
		System.out.println( Math.round(((float)(n-total_failures))/n*100)+"% Task execution success");
		
	}
	*/
	
	// ****************************************************************
	
	private final ResourceFailureGenerator rfg;
	private final EventList el;
	private       TaskEvent current_evt;

}

