/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ResourceDataConfiguration.java

package lbg.resource.config;

import lbg.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.protocol.data.*;
import java.util.Properties;

/** Resource data configuration.
 * @author Cyril Briquet
 */
public class ResourceDataConfiguration {

	/** Creates a new ResourceDataConfiguration.
	 */
	public ResourceDataConfiguration(Properties props) throws GridException {
		
		if (props == null) {
			throw new GridException("ResourceDataConfiguration(): "+
									"illegal Properties");
		}

		this.CACHE_DIR = loadProperty("CACHE_DIR",props);
		this.STATS_DIR = loadProperty("STATS_DIR",props);
		String candidate_capacity =
			loadProperty("SUPPLYING_CACHE_CAPACITY",props);
		this.SUPPLYING_CACHE_CAPACITY =
			((candidate_capacity == null) ||
			candidate_capacity.length() == 0) ? 
			GridDataManager.INFINITE_CAPACITY :
			Integer.parseInt(candidate_capacity);

		// ****************

		checkProperties();

	}

	/** Creates a new ResourceDataConfiguration (legal in SIM_MODE only).
	 */
	public ResourceDataConfiguration(int SUPPLYING_CACHE_CAPACITY)
		throws GridException {
		
		if (LightweightBarteringGrid.getMode() != SIM_MODE) {
			throw new GridException("ResourceDataConfiguration(): "+
									"wrong constructor");
		}
		
		if (SUPPLYING_CACHE_CAPACITY < 0) {
			throw new GridException("ResourceDataConfiguration(): "+
									"illegal Supplying cache capacity");
		}
		
		this.CACHE_DIR = null;
		this.STATS_DIR = null;
		this.SUPPLYING_CACHE_CAPACITY = SUPPLYING_CACHE_CAPACITY;

	}
	
	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ResourceDataConfiguration]");
		sb.append(format(true));
		
		return sb.toString();
		
	}
	
	/** Outputs the ResourceDataConfiguration,
	 * formatted either as a property file or as a log file.
	 * @param log_format true to format ResourceDataConfiguration as a log file
	 * @return formatted ResourceDataConfiguration
	 */
	public String format(boolean log_format) {
		
		StringBuilder sb = new StringBuilder();

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("CACHE_DIR = ");
		sb.append(this.CACHE_DIR);

		if (log_format) sb.append("  ->  ");
		sb.append("SUPPLYING_CACHE_CAPACITY = ");
		sb.append(this.SUPPLYING_CACHE_CAPACITY);

		return sb.toString();

	}
	
	// ****************************************************************

	public String CACHE_DIR() {
		
		return this.CACHE_DIR;
		
	}
	
	public String STATS_DIR() {
		
		return this.STATS_DIR;
		
	}

	public int SUPPLYING_CACHE_CAPACITY() {
		
		return this.SUPPLYING_CACHE_CAPACITY;
		
	}

	// ****************************************************************

	private void checkProperties() throws GridException {
		
		if ((this.CACHE_DIR == null) ||
			(this.CACHE_DIR.length() < 1)) {
			throw new GridException("ResourceDataConfiguration(): "+
									"illegal Resource cache directory");
		}
		// Don't check STATS_DIR, it is legal for it to be null or empty
		if (this.SUPPLYING_CACHE_CAPACITY < 1
			&& this.SUPPLYING_CACHE_CAPACITY != GridDataManager
			.INFINITE_CAPACITY) {
			throw new GridException("ResourceDataConfiguration(): "+
									"illegal Resource "+
									"Supplying cache capacity");
		}

	}

	// ****************************************************************

	private final String CACHE_DIR;
	private final String STATS_DIR;
	private int SUPPLYING_CACHE_CAPACITY;

}

