/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ResourceNetworkConfiguration.java

package lbg.resource.config;

import lbg.*;
import static lbg.VirtualizationLevel.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.common.net.*;

import java.util.Properties;

/** Resource network configuration
 * @author Cyril Briquet
 */
public class ResourceNetworkConfiguration {

	/** Creates a new ResourceNetworkConfiguration.
	 */
	public ResourceNetworkConfiguration(Properties props) throws GridException {
		
		if (props == null) {
			throw new GridException("ResourceNetworkConfiguration(): "+
									"illegal Properties");
		}

		this.RESOURCE_HOSTNAME = loadProperty("RESOURCE_HOSTNAME",props);
		this.RESOURCE_PORT = loadInt(loadProperty("RESOURCE_PORT",props));
		this.PEER_HOST = loadProperty("PEER_HOST",props);
		this.PEER_PORT = loadInt(loadProperty("PEER_PORT",props));
		this.BITTORRENT_DATA_PORT = loadInt(loadProperty("BITTORRENT_DATA_PORT",
														 props));
		this.FTP_DATA_PORT = loadInt(loadProperty("FTP_DATA_PORT",props));

		// ****************
		
		checkProperties();

	}
	
	/** Creates a new ResourceNetworkConfiguration.
	 */
	public ResourceNetworkConfiguration() throws GridException {

		if (LightweightBarteringGrid.getMode() != SIM_MODE) {
			throw new GridException("ResourceDataConfiguration(): "+
									"wrong constructor");
		}

		this.RESOURCE_HOSTNAME    = "worker.fakedomain.matrix";
		this.RESOURCE_PORT        = 31337;

		this.PEER_HOST            = "peer.fakedomain.matrix";
		this.PEER_PORT            = 31337;
		this.BITTORRENT_DATA_PORT = 31337;
		this.FTP_DATA_PORT        = 31337;

	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ResourceNetworkConfiguration]");
		sb.append(format(true));
		
		return sb.toString();
		
	}
	
	/** Outputs the ResourceNetworkConfiguration,
	 * formatted either as a property file or as a log file.
	 * @param log_format true to format ResourceNetworkConfiguration as a log file
	 * @return formatted ResourceNetworkConfiguration
	 */
	public String format(boolean log_format) {
		
		StringBuilder sb = new StringBuilder();

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("RESOURCE_HOSTNAME = ");
		sb.append(this.RESOURCE_HOSTNAME);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("RESOURCE_PORT = ");
		sb.append(this.RESOURCE_PORT);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_HOST = ");
		sb.append(this.PEER_HOST);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_PORT = ");
		sb.append(this.PEER_PORT);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("BITTORRENT_DATA_PORT = ");
		sb.append(this.BITTORRENT_DATA_PORT);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("FTP_DATA_PORT = ");
		sb.append(this.FTP_DATA_PORT);
		
		return sb.toString();

	}
	
	// ****************************************************************

	/**
	 * @return Resource host name, as reachable by owner Peer
	 */
	public String hostname() {
		
		return this.RESOURCE_HOSTNAME;
		
	}

	/**
	 * @return valid port for owner Peer-to-Resource communication
	 */
	public int resourcePort() {
		
		return this.RESOURCE_PORT;
		
	}

	/**
	 * @return owner Peer host name
	 */
	public String HOST() {
		
		return this.PEER_HOST;
		
	}
	
	/**
	 * @return valid port for Resource-to-owner Peer communication
	 */
	public int peerPort() {
		
		return this.PEER_PORT;
		
	}
	
	/**
	 * @return BitTorrent data port
	 */
	public int btDataPort() {
		
		return this.BITTORRENT_DATA_PORT;
		
	}
	
	public int ftpDataPort() {
		
		return this.FTP_DATA_PORT;
		
	}
	
	// ****************************************************************

	private void checkProperties() throws GridException {

		if (this.RESOURCE_HOSTNAME == null) {
			this.RESOURCE_HOSTNAME = CommonNetHandling.getLocalHostName();
		}
		if ((this.RESOURCE_HOSTNAME == null) ||
			(this.RESOURCE_HOSTNAME.length() < 1)) {
			throw new GridException("checkProperties(): "+
									"illegal Resource hostname");
		}
		if (CommonNetHandling.isValidPort(this.RESOURCE_PORT) == false) {
			throw new GridException("checkProperties(): "+
									"illegal Resource port");
		}
		if ((this.PEER_HOST == null) ||
			(this.PEER_HOST.length() < 1)) {
			throw new GridException("checkProperties(): "+
									"illegal owner Peer hostname");
		}
		if (CommonNetHandling.isValidPort(this.PEER_PORT) == false) {
			throw new GridException("checkProperties(): "+
									"illegal owner Peer port");
		}
		if (CommonNetHandling.isValidPort(this.BITTORRENT_DATA_PORT) ==
			false) {
			throw new GridException("checkProperties(): "+
									"illegal BitTorrent data port");
		}
		if (CommonNetHandling.isValidPort(this.FTP_DATA_PORT) == false) {
			throw new GridException("checkProperties(): "+
									"illegal FTP data port");
		}

	}

	// ****************************************************************

	private String RESOURCE_HOSTNAME;
	private int RESOURCE_PORT;
	private String PEER_HOST;
	private int PEER_PORT;
	private int BITTORRENT_DATA_PORT;
	private int FTP_DATA_PORT;

}

