/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ContainerPolicyManager.java

package lbg.resource.execution;

import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonFileHandling.FILE_SEP;

/** A ContainerPolicyManager specifies the security policy
 * of a LiveTaskExecutionServer.
 * <p>It grants permissions that define precisely the limits of sandboxing.
 * Notably: network communication and access to a playpen
 * (a playpen is a temporary disk storage space used by a GridApplication).
 * @author Cyril Briquet
 */
public class ContainerPolicyManager {

	public ContainerPolicyManager(String cache_dir) throws GridException {
		
		if ((cache_dir == null) || (cache_dir.length() < 1)) {
			throw new GridException("ContainerPolicyManager(): "+
									"illegal cache directory");
		}
		
		this.cache_dir = cache_dir+FILE_SEP;
		this.playpen_dir =
			this.cache_dir+ContainerPolicyManager.PLAYPEN_BASE_PREFIX+FILE_SEP;
		this.container_policy_path =
			this.cache_dir+ContainerPolicyManager.GRIDAPP_POLICY_FILE;
		
	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ContainerPolicyManager]\n");
		sb.append("\tcache directory=");
		sb.append(this.cache_dir);
		sb.append("\tplaypen directory=");
		sb.append(this.playpen_dir);
		
		return sb.toString();
		
	}

	// ****************************************************************

	public String getPolicy(String... target_dirs) throws GridException {
		
		if (target_dirs == null) {
			throw new GridException("getPolicy(): illegal directory list");
		}
		
		/*
		http://java.sun.com/j2se/1.5.0/docs/guide/security/PolicyFiles.html
		http://java.sun.com/j2se/1.5.0/docs/api/java/security/Permission.html
		http://java.sun.com/j2se/1.5.0/docs/api/java/io/FilePermission.html
		http://java.sun.com/j2se/1.5.0/docs/api/java/net/SocketPermission.html
		http://www.cloudgarden.com/JSAPI/docs/security.html
		http://java.sun.com/j2se/1.5.0/docs/guide/security/permissions.html
		http://www.javaworld.com/javaworld/jw-11-1997/jw-11-hood.html
		http://www.onjava.com/pub/a/onjava/2005/01/26/classloading.html
		http://java.sun.com/developer/technicalArticles/Networking/classloaders/		
		http://java.sun.com/docs/books/tutorial/deployment/jar/

		Jar sealing
		http://java.sun.com/developer/JDCTechTips/2001/tt0130.html
		*/
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("grant {\n").
			append(grantRutimePermission("setIO")).
			append(grantRutimePermission("createClassLoader")).
			append(grantRutimePermission("accessClassInPackage.*")).
			append(grantSocketPermission("listen, accept, connect, resolve")).
			append("\tpermission java.net.NetPermission \"specifyStreamHandler\";\n").
			append(grantFilePermission(this.cache_dir,
									   FilePermissionLevel.READ_WRITE,
									   false)).
			append(grantFilePermission(this.cache_dir,
									   FilePermissionLevel.READ_ONLY,
									   true)).
			append(grantFilePermission(this.playpen_dir,
									   FilePermissionLevel.READ_WRITE_DELETE,
									   true));
		for (String target_dir : target_dirs) {
			sb.append(grantFilePermission(target_dir,
										  FilePermissionLevel.READ_WRITE_DELETE,
										  true));
		}
		sb.append("};\n");
			
		return sb.toString();
	
	}
	
	public void writePolicy(String... target_dirs) throws GridException {
		
		CommonFileHandling.writeFile(this.container_policy_path,
									 this.getPolicy(target_dirs));
		
	}
	
	// ****************************************************************

	/* For testing purposes 
	   java  -cp jars/lightweight_bartering_grid.jar lbg.resource.execution.ContainerPolicyManager
	 */
	public static void main(String[] args) {
	
		try {
			ContainerPolicyManager p =
				new ContainerPolicyManager("/tmp/lbg_res_cache/");
			System.out.println(p.getPolicy("truc2","truc1"));
		}
		catch (GridException e) {
			e.printStackTrace();
		}
		
	}
	
	// ****************************************************************

	private String grantRutimePermission(String tag)
		throws GridException {
		
		if (tag == null) {
			throw new GridException("grantRutimePermission(): illegal tag");
		}
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("\tpermission java.lang.RuntimePermission \"").
			append(tag).
			append("\";\n");
		
		return sb.toString();

	}

	private String grantSocketPermission(String tag)
		throws GridException {
		
		if (tag == null) {
			throw new GridException("grantSocketPermission(): illegal tag");
		}
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("\tpermission java.net.SocketPermission ").
			append("\"*\", \"").
			append(tag).
			append("\";\n");
		
		return sb.toString();

	}

	/** Grant file permission for target file/directory.
	 * @param path target file or directory
	 * @param read_write true for read/write access, false for read only access
	 * @param all_files true to extend the permission to all files
	 * of target directory (it is assumed that the target path
	 * designates a directory, and no check is done to verify it)
	 */
	private String grantFilePermission(String path,
									   FilePermissionLevel perm_level,
									   boolean all_files)
		throws GridException {
		
		if (path == null) {
			throw new GridException("grantFilePermission(): "+
									"illegal path");
		}

		String perm_level_s = null;
		switch (perm_level) {
		case READ_ONLY:
			perm_level_s = "read";
			break;
		case READ_WRITE:
			perm_level_s = "read,write";
			break;
		case READ_WRITE_DELETE:
			perm_level_s = "read,write,delete";
			break;
		default:
			throw new GridException("grantFilePermission(): "+
									"illegal file permission");
		}
		
		StringBuilder sb = new StringBuilder();
		
		// Grant access to file or directory itself
		sb.append("\tpermission java.io.FilePermission \"").
			append(path).
			append("\",\""+perm_level_s+"\";\n");
		// Grant access to directory contents
		boolean recursive = true;
		if (all_files) {
			sb.append("\tpermission java.io.FilePermission \"").
				append(path).
				append(recursive ? "/-" : "/*").
				append("\",\""+perm_level_s+"\";\n");
		}

		return sb.toString();
	
	}
	
	// ****************************************************************

	public final String cache_dir;
	public final String playpen_dir;
	public final String container_policy_path;
	
	public static enum FilePermissionLevel {
		READ_ONLY,
		READ_WRITE,
		READ_WRITE_DELETE
	};

	// ****************************************************************

	public static String getPlaypenDir(String cache_dir) throws GridException {
	
		if (cache_dir == null) {
			throw new GridException("getPlaypenDir(): illegal cache direcory");
		}

		return cache_dir+ContainerPolicyManager.PLAYPEN_BASE_PREFIX+FILE_SEP;
		
	}
	
	public static final String GRIDAPP_POLICY_FILE = "__grid__app.policy";
	public static final String PLAYPEN_BASE_PREFIX = "__playpen";

}

