/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LiveTaskExecutionHandler.java

package lbg.resource.execution;

import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.getStackTrace;
import lbg.peer.service.InternalPeerHandle;
import lbg.protocol.data.*;
import lbg.resource.LiveResource;
import lbg.task.Task;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OptionalDataException;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;

/** A wrapper over GridApplications to allow better control and automatic
 * TaskOutput generation when an app ends.<br>
 * Related logger: lbg.resource
 * @author Xavier Dalem, Cyril Briquet
 */
public class LiveTaskExecutionHandler implements Runnable {

	/** Readies a handler to be started as a different Thread. The thread
	 * isn't started automatically (and the Task doesn't run until is starts).
	 * @param res owner Resource
	 * @param t Task to run
	 * @param iph handle to the owner Peer
	 * @param mem_size memory to allocate to the Task execution
	 * (expressed as a number of megabytes)
	 */
	public LiveTaskExecutionHandler(LiveResource res,InternalPeerHandle iph,
									Task t,int mem_size)
		throws GridException {
			
		if (res == null) {
			throw new GridException("LiveTaskExecutionHandler(): "+
									"illegal resource");
		}
		if (iph == null) {
			throw new GridException("LiveTaskExecutionHandler: illegal Peer");
		}
		if (t == null) {
			throw new GridException("LiveTaskExecutionHandler: illega Task");
		}
		if ((mem_size < 1) || (mem_size > 1024*1024)) {
			throw new GridException("LiveTaskExecutionHandler: "+
									"illegal memory size");
		}
		
		this.res      = res;
		this.iph      = iph;
		this.t        = t;
		this.mem_size = mem_size;
		this.runner   = null;

		// Create containment policy for the Grid application
		String cache_dir = this.res.getDataConfig().CACHE_DIR();
		this.cpm = new ContainerPolicyManager(cache_dir);
		
	}
	
	public synchronized String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LiveTaskExecutionHandler]\n");

		return sb.toString();
		
	}

	// ****************************************************************

	/** Start handling the task. */
	public synchronized void run() {
		
		ObjectOutputStream oos = null;
		ObjectInputStream ois = null;
		
		byte[] log = null; // Will hold the runtime logs on std(err|out)
		
		try {
			// Download data (asynchronously from main Resource thread)
			t.readyInputDataSet(res.gdm());

			// Write containment policy for the Grid application
			this.cpm.writePolicy();

			// Remove playpen dir
			CommonFileHandling.removeDirectory(this.cpm.playpen_dir,true);
			// Create playpen dir
			CommonFileHandling.createDirectory(this.cpm.playpen_dir);

			// Launch the Grid application container (LiveTaskRunner)
			// in a separate VM
			String[] cmd = new String[] {
				"java",
				"-ea",
				"-Djava.security.manager",
				"-Djava.security.policy="+this.cpm.container_policy_path,
				"-Xmx"+Integer.toString(this.mem_size)+"m",
				"-cp",
				System.getProperty("java.class.path"),
				"lbg.resource.execution.LiveTaskExecutionServer",
				this.res.owner().id(),
				this.cpm.playpen_dir // must be terminated by file.separator
			};
			this.runner = Runtime.getRuntime().exec(cmd);
			// Note: ProcessBuilder does not work here, so we exec directly

			// Upload target Task
			oos = new ObjectOutputStream(this.runner.getOutputStream());
			/* This used to fail with an EOFException. Refer to build
			   20060924 for a workaround if it was to fail again */
			oos.writeObject(t);
			oos.flush();

			// Read VM start error log, if any
			String vm_start_error_log = null;
			vm_start_error_log = CommonFileHandling.
				readFile(this.cpm.playpen_dir+
						 LiveTaskExecutionServer.LTC_ERROR_LOG_SUFFIX,true);
			// Design choice: if there *is* an error log file,
			// but it cannot be read, execution continues as if it did not exist
			if (vm_start_error_log != null) {
				throw new GridException(vm_start_error_log);
			}

			// Download results
			ois = new ObjectInputStream(new BufferedInputStream(
				this.runner.getInputStream()));
			
			// Execution is blocked here when a Task is running
			boolean success = false;
			try {
				success = ois.readBoolean();
			}
			catch (Exception e) {
				throw new GridException("run(): Task execution failure "+
										"(without awareness of runner VM)");
			}
			if (!success) {
				Object vm_runtime_error_log = ois.readObject();
				String s_vm_runtime_error_log =
					((vm_runtime_error_log instanceof String) &&
					 (vm_runtime_error_log != null)) ?
					("\n"+(String)vm_runtime_error_log) : "";
				throw new GridException("run(): Task execution failure "+
										"(with awareness of runner VM)"+
										s_vm_runtime_error_log);
			}

			Object data = ois.readObject();
			if (!(data instanceof GridData)) {
				throw new GridException("run(): illegal GridData");
			}

			log = new byte[ois.readInt()];
			ois.readFully(log);

			// Remove playpen dir
			CommonFileHandling.removeDirectory(this.cpm.playpen_dir,true);

			// Task execution success
			res.completeTask((GridData)data, log);
		}
		catch (Throwable failCause) {
			// This branch is reached upon a System.exit() call from
			// the Grid application (if allowed by containment policy...),
			// to the Resource middleware is not affected
			this.res.logger().log(Level.INFO, "Failed to run task", failCause);
			
			if (failCause instanceof OptionalDataException) {
				System.err.println(((OptionalDataException)failCause).eof);
			}
			// Wherever the task failed, it was before managing to complete it.
			
			// If resource is idle, the peer asked us to cancel. If not, we must
			// tell him we failed so that he can ask us.
			try {
				if (log == null) {
					log = new byte[0];
				}
				
				if ((res.idle() == false) && (res.getRemovalState() == false)) {
					ThreadPool.nap(200);
					if (t.consumerID().equals(iph.id())) { // Local cancel
						iph.preemptLocalTaskOnLocalResource(t,log);
					}
					else { // Supplying cancel
						iph.preemptSupplyingTaskOnLocalResource(t.consumerID(),
							t,log);
					}
				}
			}
			catch (Exception e) {
				this.res.logger().
					info("run(): could not contact Peer "+
						 "to report Task execution failure");
			}

			// Read VM start error log, if any
			String vm_start_error_log = null;
			try {
				String cache_dir = this.res.getDataConfig().CACHE_DIR();
				String playpen_dir =
					ContainerPolicyManager.getPlaypenDir(cache_dir);
				vm_start_error_log =
					CommonFileHandling.
						readFile(playpen_dir+
								 LiveTaskExecutionServer.LTC_ERROR_LOG_SUFFIX);
			}
			catch (GridException e) { }
			if (vm_start_error_log != null) {
				vm_start_error_log = "\nCaused by "+vm_start_error_log;
			}
			else {
				vm_start_error_log = "";
			}
			
			// Tell that we failed
			this.res.logger().
				info("run(): Task execution failure"+vm_start_error_log+
					getStackTrace(failCause));
		}
		finally {
			// Remove playpen dir, possibly for the second time
			// after Task completion/failure
			try {
				CommonFileHandling.removeDirectory(this.cpm.playpen_dir);
			}
			catch (GridException e) {
				this.res.logger().
					info("run(): could not clean playpen directory");
			}

			// Close communication channells with runner VM
			if (oos != null) {
				try { oos.close(); } catch (IOException doomed) {}
			}
			if (ois != null) {
				try { ois.close(); } catch (IOException doomed) {}
			}
			
			try {
				// Kill the runner VM process
				this.runner.destroy();
			}
			catch (Exception doomed) {} // goal is to catch RuntimeExceptions
			finally {
				this.runner = null;
			}

		}
		
	}
	
	/** Stops the managed task. Instantly kills the target thread and any
	 * thread ran by it. The resource should be reported as idle prior to
	 * calling this.
	 */
	public void abort() {

		// If resource is not idle, we could encounter some problems in the
		// running task.
		assert(this.res.idle());
		
		if (this.runner != null) {
			this.runner.destroy();
			this.runner = null;
		}
		
	}

	// ****************************************************************

	private final LiveResource res;
	private final InternalPeerHandle iph;
	private final Task t;
	private final int mem_size;
	private final ContainerPolicyManager cpm;
	private       Process runner;

}

