/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LiveTaskExecutionServer.java

package lbg.resource.execution;

import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.getStackTrace;
import lbg.protocol.data.GridData;
import lbg.task.Task;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintStream;
import java.util.Date;
import java.util.logging.Logger;

/** A LiveTaskExecutionServer provides a contained environment
 * for the execution of a Grid application.
 * <p>It manages I/O with the LiveTaskExecutionHandler that controls it.
 * It runs the target Grid application with a LiveTaskRunner.
 * <p>Related logger: lbg.resource
 * @author Cyril Briquet
 */
class LiveTaskExecutionServer {

	/** Constructs a new LiveTaskExecutionServer.
	 */
	LiveTaskExecutionServer(String supplier_id,String playpen_dir)
		throws GridException {

		if (supplier_id == null) {
			throw new GridException("LiveTaskExecutionServer(): illegal Supplier id");
		}
		if (playpen_dir == null) {
			throw new GridException("LiveTaskExecutionServer(): illegal playpen dir");
		}
		
		// As the standard streams are more or less protected
		// from Exceptions,
		// we must have a way of telling the Handler
		// whether Task execution succeeded or not
		boolean success = true;
		GridData output_data = null;
		Throwable error_log = (Throwable)null;
		
		try {
			/* for testing purposes only -- can safely be removed
			java.util.Vector v = new java.util.Vector();
			for (int a = 1,b = 2 ; a < b ; ) v.add(new int[1024*1024]);*/

			// Process constructor arguments
			if (supplier_id == null) {
				throw new GridException("LiveTaskExecutionServer(): "+
										"illegal Supplier id");
			}

			// Dowload target Task from the LiveTaskExecutionHandler
			output_data = null;
			Object task = ois.readObject();
			if (!(task instanceof Task)) {
				throw new GridException("LiveTaskExecutionServer(): "+
										"received data is not a Task");
			}
			
			// Actually run target Task
			LiveTaskRunner ltr =
				new LiveTaskRunner((Task)task,supplier_id,playpen_dir);
			output_data = ltr.go();
		}
		catch (Throwable fatal) {
			success = false;
			error_log = fatal;
			Logger.getLogger("lbg.resource").severe(fatal.getMessage());
		}
		
		try {
			// Upload Task execution-generated data to
			// the LiveTaskExecutionHandler controller
			try {
				if (LiveTaskExecutionServer.oos == null) {
					throw new GridException("LiveTaskExecutionServer(): "+
											"output stream unavailable");
				}
				LiveTaskExecutionServer.oos.writeBoolean(success);
				if (success) {
					LiveTaskExecutionServer.oos.writeObject(output_data);
					LiveTaskExecutionServer.oos.
						writeInt(LiveTaskExecutionServer.ebao.size());
					LiveTaskExecutionServer.oos.
						write(LiveTaskExecutionServer.ebao.toByteArray());
				}
				else /* if (!success) */ {
					LiveTaskExecutionServer.oos.writeObject(getStackTrace(error_log));
				}
				LiveTaskExecutionServer.oos.flush();
			}
			catch (IOException e) {
				Logger.getLogger("lbg.resource").
					severe("Cannot transmit results");
				throw new GridException("LiveTaskExecutionServer(): "+
										"cannot transmit results",e);
			}
		}
		catch (GridException e) {
			throw e;
		}
		catch (Throwable fatal) {
			Logger.getLogger("lbg.resource").severe(fatal.getMessage());
			throw new GridException("LiveTaskExecutionServer(): unexpected Exception"+
									" when transmitting results\n"+getStackTrace(fatal),fatal);
		}
		
	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LiveTaskExecutionServer]");

		return sb.toString();
		
	}
	
	// ****************************************************************

	public static void main(String[] args) throws GridException {

		try {
			if ((args.length != 2) ||
				(args[0] == null) || (args[0].length() < 1) ||
				(args[1] == null) || (args[1].length() < 1)) {
				throw new GridException("LiveTaskExecutionServer(): "+
										"illegal command line");
			}
			
			openCommChannels();
			/*LiveTaskExecutionServer ltc =*/ new LiveTaskExecutionServer(args[0],args[1]);
			closeCommChannels();
		}
		catch (Throwable thr) {
			Logger.getLogger("lbg.resource").
				severe("unexpected Exception in LiveTaskExecutionServer... "+
					   "May the Force be with You !\n"+thr);
			String error_log_fn = args[1]+LTC_ERROR_LOG_SUFFIX;
			String error_log_msg = getStackTrace(thr)+new Date();
			CommonFileHandling.writeFile(error_log_fn,error_log_msg);
		}
		finally {
			System.exit(0);
		}
		
	}
	
	/** Opens communication channels with the LiveTaskExecutionHandler.
	 */
	public static void openCommChannels() throws GridException {
	
		LiveTaskExecutionServer.ois  = null;
		try {
			LiveTaskExecutionServer.ois =
				new ObjectInputStream(new BufferedInputStream(System.in));
		}
		catch (IOException e) {
			Logger.getLogger("lbg.resource").
				severe("Cannot create input stream");
			throw new GridException("openCommChannels(): "+
									"cannot create input stream",e);
		}

		LiveTaskExecutionServer.oos = null;
		try {
			LiveTaskExecutionServer.oos = new ObjectOutputStream(System.out);
		}
		catch (IOException e) {
			Logger.getLogger("lbg.resource").
				severe("Cannot create output stream");
			throw new GridException("openCommChannels(): "+
									"cannot create output stream",e);
		}
		
		// Override input and output to prevent the launched task from
		// interfering with us. Meanwhile, we are able to keep log of its
		// output on stderr and stdout
		System.setIn(new NullInputStream());
		LiveTaskExecutionServer.ebao =
			new EnhancedByteArrayOutputStream(MAX_OUTPUT_SIZE,MAX_OUTPUT_SIZE);
		PrintStream ps = new PrintStream(LiveTaskExecutionServer.ebao);
		System.setOut(ps);
		System.setErr(ps);

	}
	
	/** Closes communication channels with the LiveTaskExecutionHandler.
	 */
	public static void closeCommChannels() throws GridException {
	
		// Close communication channels with the LiveTaskExecutionHandler
		try {
			if (LiveTaskExecutionServer.ois != null) {
 				ois.close();
			}
		}
		catch (IOException e) {
			Logger.getLogger("lbg.resource").
				severe("Cannot close input stream");
			throw new GridException("closeCommChannels(): "+
									"cannot close input stream",e);
		}
		
		try {
			if (LiveTaskExecutionServer.oos != null) {
				LiveTaskExecutionServer.oos.close();
			}
		 }
		 catch (IOException e) {
			Logger.getLogger("lbg.resource").
				severe("Cannot close output stream");
			throw new GridException("closeCommChannels(): "+
									"cannot close output stream",e);
		 }

	}
	
	// ****************************************************************

	/** Size limit of output data file
	 * (arbitrarily set to 256MiB).
	 */
	public static final int MAX_OUTPUT_SIZE = 64*1024*1024;

	public static final String LTC_ERROR_LOG_SUFFIX = "__ltc_error_log";
	
	// ****************************************************************
	
	private static ObjectInputStream ois;
	private static ObjectOutputStream oos;
	private static EnhancedByteArrayOutputStream ebao;

}

