/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LiveResourceHandle.java

package lbg.resource.service;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.protocol.data.GridData;
import lbg.protocol.rpc.RemoteCall;
import lbg.task.*;

/** A handle to a resource.
 * @author Xavier Dalem
 */
public class LiveResourceHandle implements ResourceHandle {
	
	static final long serialVersionUID = -837912632505263504L;
	
	private final String id;
	private final int cache_capacity;
	private final String res_host;
	private final int res_port;
	private       int hashcode;
	
	/** Create a handle to contact a LiveResource
	 * @param id The resource ID
	 * @param cache_capacity
	 * @param res_host A way to contact the resource
	 * @param res_port The port on which the resource listens
	 */
	public LiveResourceHandle(String id, int cache_capacity,
		String res_host,int res_port) throws GridException {
		if (id == null || res_host == null || res_port <= 0 || cache_capacity < 1) {
			throw new GridException("LiveResourceHandle(): illegal parameters");
		}

		this.id = id;
		this.cache_capacity = cache_capacity;
		this.res_host = res_host;
		this.res_port = res_port;
		this.hashcode = Integer.MIN_VALUE;
		
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LiveResourceHandle ");
		sb.append(this.id);
		sb.append("] on ");
		sb.append(this.res_host);
		sb.append(":");
		sb.append(this.res_port);
		
		return sb.toString();
		
	}

	public int compareTo(ResourceHandle other) {
		return this.id.compareToIgnoreCase(other.id());
	}
	
	public boolean equals(Object obj) {
		
		if (obj instanceof ResourceHandle == false) {
			return false;
		}
		
		return (this.compareTo((ResourceHandle)obj) == 0);
		
	}

	public String id() {
		return this.id;
	}

	public int cacheCapacity() {
		
		return this.cache_capacity;
		
	}
	
	@Override
	public int hashCode() {
	
		if (this.hashcode == Integer.MIN_VALUE) {
			this.hashcode = CommonMath.generateHashCode(this.id);
		}
		return this.hashcode;
	
	}
	
	/** Sets a resource's working set
	 * This call is non-blocking (only verification of the validity of the WS
	 * is done before returning: data is not yet downloaded).
	 * @return <tt>true</tt> if setting the Working Set succeeded.
	 * @see lbg.protocol.data.GridDataManager#setWorkingSet(GridData[])
	 * Concept of a Working Set
	 */
	public synchronized boolean setWorkingSet(GridData[][] working_set)
			throws GridException {
		Object ret = null;
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.RESOURCE,
				"setWorkingSet",
				new Class[] {Class.forName("[[Llbg.protocol.data.GridData;")},
				new Object[] {working_set} );
			ret = rc.send(res_host, res_port, true);
		} catch (ClassNotFoundException e2) {
			throw new GridException(
				"Critical: could not find one of the base classes", e2);
		}
		
		if (ret == null || !(ret instanceof Boolean))
			return true;
		
		return ((Boolean) ret).booleanValue();
	}

	/** Checks if a resource if idle.
	 * <br>If the Resource cannot be contacted, it is considered to be idle,
	 * so that removing it can be done without any further step
	 * @return the state of the Resource
	 */
	public synchronized boolean idle() {
		Object ret = null;
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.RESOURCE,
				"idle",	new Class[0], new Object[0]);
			ret = rc.send(res_host, res_port, true);
		} catch (GridException e) {
			e.printStackTrace();
			return false;
		}
		
		if (ret == null || !(ret instanceof Boolean))
			return true;
		
		return ((Boolean) ret).booleanValue();
	}

	/** Gets a resource's network state.<br>
	 * @return <tt>true</tt> if the resource could be reached and answered
	 * correctly
	 */
	public synchronized boolean isAlive() {
		Object ret = null;
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.RESOURCE,
				"isAlive",	new Class[0], new Object[0]);
			ret = rc.send(res_host, res_port, true);
		} catch (GridException e) {
			if (! (e instanceof GridNetworkException))
				e.printStackTrace();
			return false;
		}
		
		if (ret == null || !(ret instanceof Boolean)
				|| ((Boolean) ret).booleanValue() != true)
			return false;
		
		return true;
	}

	
	/** Get the running Task of a Resource.
	 * @return the Task, or null if the Resource cannot be contacted
	 * @deprecated
	 */
	@Deprecated
	public synchronized Task runningTask() {
		Object ret = null;
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.RESOURCE,
				"runningTask", new Class[0], new Object[0]);
			ret = rc.send(res_host, res_port, true);
		} catch (GridException e) {
			return null;
		}
		
		if (ret == null || !(ret instanceof Task))
			return null;
		
		return ((Task) ret);
	}
	
	/** Ask a Resource to run a Task
	 * @deprecated Specify new Working Set when running a task.
	 * @see #runTask(Task, GridData[][])
	 * @todo
	 */
	@Deprecated public synchronized void delete_me_runTask(Task t) throws GridException {
		if (t == null)
			throw new GridException("runTask: null task");
		
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.RESOURCE,
				"runTask", new Class[] {Class.forName("lbg.task.Task")},
				new Object[] {t} );
			rc.send(res_host, res_port, false);
		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}

	/** Sets a resource's working set then runs a task.
	 * This call is non-blocking (only verification of the validity of the WS
	 * is done before returning: data is not yet downloaded).
	 * @return <tt>true</tt> if setting the Working Set succeeded.
	 * @see lbg.protocol.data.GridDataManager#setWorkingSet(GridData[])
	 * Concept of a Working Set
	 */
	public boolean runTask(Task t,GridData[][] working_set)
			throws GridException {
		
		Object ret = null;
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.RESOURCE,
				"runTask",
				new Class[] {Class.forName("lbg.task.Task"),
					Class.forName("[[Llbg.protocol.data.GridData;")},
				new Object[] {t, working_set} );
			ret = rc.send(res_host, res_port, true);
		} catch (ClassNotFoundException e2) {
			throw new GridException(
				"Critical: could not find one of the base classes", e2);
		}
		
		if (ret == null || !(ret instanceof Boolean))
			return true;
		
		return ((Boolean) ret).booleanValue();
	}

	/** Ask a Resource to cancel running Task */
	public synchronized void cancelTask() throws GridException {
		RemoteCall rc = new RemoteCall(RemoteCall.CallType.RESOURCE,
			"cancelTask", new Class[0], new Object[0]);
		rc.send(res_host, res_port, false);
	}
	
	/** Ask a Resource to give expected Task runtime
	 * @throws GridException in case of communication error */
	public synchronized long expectedTaskRuntime() throws GridException {
		RemoteCall rc = new RemoteCall(RemoteCall.CallType.RESOURCE,
			"expectedTaskRuntime", new Class[0], new Object[0]);
		Object ret = rc.send(res_host, res_port, true);
		
		if (ret == null || !(ret instanceof Long))
			throw new GridException("Invalid return value");
		
		return ((Long) ret).longValue();
	}
	
	/** Ask a Resource when the running Task has started 
	 * @return the time, or -1 if the Resource cannot be contacted */
	public long taskStartTime() {
		Object ret = null;
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.RESOURCE,
				"taskStartTime", new Class[0], new Object[0]);
			ret = rc.send(res_host, res_port, true);
		} catch (GridException e) {
			return -1;
		}
		
		if (ret == null || !(ret instanceof Long))
			return -1;
		
		return ((Long) ret).longValue();
	}
}

