/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LiveResourceService.java

package lbg.resource.service;

import lbg.common.exception.GridException;
import lbg.protocol.data.GridData;
import lbg.resource.*;
import lbg.resource.config.ResourceNetworkConfiguration;
import lbg.task.Task;

/** Live implementation of the Resource service, handling requests to
 * a given Resource.
 * @author Xavier Dalem, Cyril Briquet
 */
public class LiveResourceService implements ResourceService {

	/** Create a service for a resource.
	 * @param owner_resource The resource concerned by this service
	 */
	protected LiveResourceService(LiveResource owner_resource)
		throws GridException {
		
		if (owner_resource == null) {
			throw new GridException("LiveResourceService(): "+
									"illegal owner Resource");
		}
		
		this.owner_resource = owner_resource;
		ResourceNetworkConfiguration rnc = this.owner_resource.netConfig();
		this.self_resource_handle =
			new LiveResourceHandle(this.owner_resource.id(),
								   this.owner_resource.cacheCapacity(),
								   rnc.hostname(),rnc.resourcePort());
		
	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LiveResourceService ");
		sb.append(this.owner_resource.id());
		sb.append("]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** Get a handle to the owner resource
	 */
	public ResourceHandle createHandle() throws GridException {
		
		return this.self_resource_handle;
		
	}

	public String id() {
		return this.owner_resource.id();
	}
	
	public int cacheCapacity() {
		return this.owner_resource.cacheCapacity();
	}

	public boolean setWorkingSet(GridData[][] working_set)
		throws GridException {
		return this.owner_resource.setWorkingSet(working_set);
	}

	public boolean runTask(Task t, GridData[][] working_set)
		throws GridException {
		
		if (! this.owner_resource.setWorkingSet(working_set))
			return false;
		
		this.owner_resource.runTask(t);
		return true;
	}

	/** Get the state of the resource
	 * @return <tt>true</tt> if the resource is idle
	 */
	public boolean idle() {
		return this.owner_resource.idle();
	}
	
	/** Gets the Resource status
	 * @return <tt>true</tt> (this method is only there for networking needs)
	 */
	public synchronized boolean isAlive() {
		return true;
	}

	/**
	 * Get the task currently run by the resource
	 * @deprecated
	 */
	@Deprecated
	public Task runningTask() {
		return this.owner_resource.runningTask();
	}

	/** Ask the resource to run a task
	 * @param t The task to run
	 * @deprecated
	 * @todo
	 */
	@Deprecated
	public void delete_me_runTask(Task t) throws GridException {
		this.owner_resource.runTask(t);
	}

	/** Cancel the currently running task */
	public void cancelTask() throws GridException {
		this.owner_resource.cancelTask();
	}

	/** How much time should running this task take ? */
	public long expectedTaskRuntime() throws GridException {
		return this.owner_resource.expectedTaskRuntime();		
	}

	/** Returns the Task start time in seconds */
	public long taskStartTime() {
		return this.owner_resource.taskStartTime();
	}

	// ****************************************************************

	private final Resource owner_resource;
	private final ResourceHandle self_resource_handle;

	// ****************************************************************

}

