/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SimulatedResourceHandle.java

package lbg.resource.service;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.protocol.data.*;
import lbg.task.*;

/**
 * @author Cyril Briquet
 */
public class SimulatedResourceHandle implements ResourceHandle {

	public SimulatedResourceHandle(String id,
								   SimulatedResourceService resource_service)
		throws GridException {
	
		if (id == null) {
			throw new GridException("SimulatedResourceHandle(): illegal id");
		
		}
		
		if (resource_service == null) {
			throw new GridException("SimulatedResourceHandle(): "+
									"illegal ResourceService");
		}

		this.id = id;
		this.resource_service = resource_service;
		this.hashcode = Integer.MIN_VALUE;
	
	}
	
	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[SimulatedResourceHandle ");
		sb.append(this.id);
		sb.append("]");
		
		return sb.toString();
		
	}

	public int compareTo(ResourceHandle other) {

		return this.id.compareToIgnoreCase(other.id());
		
	}

	@Override
	public boolean equals(Object obj) {
		
		if (obj instanceof ResourceHandle == false) {
			return false;
		}
		
		return (this.compareTo((ResourceHandle)obj) == 0);
		
	}
	
	@Override
	public int hashCode() {
	
		if (this.hashcode == Integer.MIN_VALUE) {
			this.hashcode = CommonMath.generateHashCode(this.id);
		}
		return this.hashcode;
	
	}

	// ****************************************************************

	public String id() {
		
		return this.id;
		
	}

	public int cacheCapacity() {
		
		return this.resource_service.cacheCapacity();
		
	}

	public synchronized boolean idle() {
		
		return this.resource_service.idle();
		
	}
	
	public synchronized boolean isAlive() {
	
		return this.resource_service.isAlive();
		
	}

	/**
	 * @deprecated
	 */
	@Deprecated
	public synchronized Task runningTask() {
		
		return this.resource_service.runningTask();
		
	}
	
	public boolean runTask(Task t,GridData[][] working_set)
		throws GridException {

		if (this.resource_service.idle() == false) {
			throw new GridException("runTask(): "+
									"corrupted resource state");
		}

		if (t == null) {
			throw new GridException("runTask(): illegal Task");
		}
		if (working_set == null) {
			throw new GridException("runTask(): illegal working set");
		}

		return this.resource_service.runTask(t,working_set);

	}
	
	/** Sets a resource's working set
	 * @see lbg.protocol.data.GridDataManager#setWorkingSet(GridData[])
	 * @todo check this
	 */
	public synchronized boolean setWorkingSet(GridData[][] working_set)
			throws GridException {

		if (working_set == null) {
			throw new GridException("setWorkingSet(): illegal working set");
		}
		
		return this.resource_service.setWorkingSet(working_set);
		
	}

	public synchronized void cancelTask() throws GridException {
		
		this.resource_service.cancelTask();

	}
	
	public synchronized long expectedTaskRuntime() throws GridException {
		
		return this.resource_service.expectedTaskRuntime();
		
	}

	public synchronized long taskStartTime() {
		
		return this.resource_service.taskStartTime();
		
	}

	// ****************************************************************

	static final long serialVersionUID = -4589862336181679218L;

	// ****************************************************************

	private final String id;
	private transient final SimulatedResourceService resource_service;
	private int hashcode;
	
}

