/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SearchEngineClient.java

package lbg.searchengine;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.peer.service.*;
import lbg.searchengine.service.SearchEngineHandle;
import lbg.task.Task;

import java.util.HashSet;
import java.util.TimerTask;
import java.util.TreeMap;

/** <p>A local database of peer handles, automatically updating from a
 * server.</p>
 * <p>Handle removal is not automatic and should be done when encountering
 * a problem with a handle.</p>
 * <p>The local database may hold handles that the server doesn't know of:
 * they might be locally added or currently not expired. The server may
 * hold handles that are not locally known: they will be downloaded at
 * next update.</p>
 * <p>The local database may hold the handle of the peer owning this
 * client, as he is not required to identify himself before contacting the
 * server.</p>
 * <p>This object is Thread-safe.</p>
 * 
 * @author Xavier Dalem, Cyril Briquet
 * @since 2007-03-21
 */
public abstract class SearchEngineClient extends TimerTask implements Iterable {

	/** Default handle count that should be returned by a search.
	 */
	public static final int DEFAULT_SEARCH_COUNT = 40;

	
	private SearchEngineHandle reference_engine;
	private ExternalPeerHandle owner_peer;
	private long last_response = -1;
	private final TreeMap<String, ExternalPeerHandle> local_database;
	/** Snapshot of the local database. Up to date if {@link #dirty} is true. */ 
	private ExternalPeerHandle[] snapshot = null;
	/** Actual number of handles in {@link #snapshot} (not necesserally equal
	 * to database size) */
	private int snapshot_length;
	/** HashSet used for random numbers generation. Kept for memory allocation
	 * and garbage collection concerns. */
	private final HashSet<Integer> rand_set;
	/** State of the snapshot */
	private boolean dirty = true;
	
	/** Creates a new Search Engine client.
	 * @param se_handle search engine server
	 * @param owner_peer manager Peer (can be null, then set afterwards,
	 * see {@link #setOwnerPeer(ExternalPeerHandle)})
	 * @param timeout time to wait before updating database (in seconds)
	 * @throws GridException if the parameters are invalid
	 */
	SearchEngineClient(SearchEngineHandle se_handle,
					   ExternalPeerHandle owner_peer,int timeout)
			throws GridException {
		
		if (se_handle == null) {
			throw new GridException("SearchEngineClient(): "+
									"illegal search engine server");
		}
		if (timeout <= 0) {
			throw new GridException("SearchEngineClient(): invalid timeout");
		}
		
		this.owner_peer = owner_peer;
		local_database = new TreeMap<String, ExternalPeerHandle>();
		
		renewReferenceServer(se_handle);
		
		rand_set = new HashSet<Integer>();
		
	}
	
	/** Sets this client's owner peer. The owner peer is the peer using the
	 * client and wanting to be registered to its search engine server.
	 * 
	 * <p>Registering an owner is not a requirement to access the server's
	 * information, but the owner will then not be listed in the server's
	 * database.</p>
	 * 
	 * <p>Registering an owner can be done only once.</p>
	 * 
	 * @param owner_peer The owner peer
	 * @throws GridException If this client already had an owner
	 */
	public void setOwnerPeer(ExternalPeerHandle owner_peer) throws GridException {
		if (this.owner_peer != null)
			throw new GridException(
				"Search engine client already has an owner peer");
		
		this.owner_peer = owner_peer;
		if (owner_peer != null)
			reference_engine.add(owner_peer);
	}
	
	/** Changes this client's reference server.
	 * 
	 * @param handle The new reference server
	 * @return The old reference server
	 * @throws GridException if the new server is invalid
	 */
	public SearchEngineHandle renewReferenceServer(SearchEngineHandle handle)
			throws GridException {
		if (handle == null)
			throw new GridException("Cannot add a null handle as reference");
		if (reference_engine != null
				&& handle.id().equals(reference_engine.id()))
			return handle;
		
		SearchEngineHandle old = reference_engine;
		reference_engine = handle;
		if (owner_peer != null)
			reference_engine.add(owner_peer);
		refresh();
		return old;
	}
	
	/** Refreshes the local database by recontacting the reference server.
	 */
	// Synchronized because we could run into race conditions with last_response
	public synchronized void refresh() throws GridException {
		SearchEngineResponse response = null;
		if (last_response == -1)
			response = reference_engine.getHandles();
		else
			response = reference_engine.getHandles(last_response);
		
		if (response == null)
			throw new GridException("null response from getHandles");
		
		last_response = response.timestamp;
		
		// Add everything, skipping errors
		if (response.handles.length > 0) {
			add(response.handles);
		}
	}
	
	/** Adds a handle to the local database. If the handle already exists, it
	 * is not overwritten.
	 * <p>You can expect log(n) times from this operation.</p>
	 * 
	 * @param handle The peer handle to add
	 */
	public void add(ExternalPeerHandle handle) {
		synchronized(local_database) {
			// Not overwriting handles can be useful to keep trusted information
			// If the handle becomes invalid (i.e. ceases to work), it should be
			// removed by the owner peer.
			if (! local_database.containsKey(handle.id())) {
				local_database.put(handle.id(), handle);
				dirty = true;
			}
		}
	}
	
	/** Adds a bunch of handles handle to the local database. If a handle
	 * already exists, it is not overwritten.
	 * <p>You can expect log(n) times from this operation.</p>
	 * 
	 * @param handles The peer handles to add
	 */
	public void add(ExternalPeerHandle[] handles) {
		synchronized(local_database) {
			
			for (ExternalPeerHandle handle: handles) {
				if (! local_database.containsKey(handle.id())) {
					local_database.put(handle.id(), handle);
					dirty = true;
				}
			}
		}
	}
	
	/** Removes a handle from the local database.
	 * 
	 * @param handle The handle to remove
	 * @throws GridException if the handle is invalid
	 */
	public void remove(ExternalPeerHandle handle) throws GridException {
		if (handle == null)
			throw new GridException("Cannot remove null handle");
		
		synchronized(local_database) {
			if (local_database.remove(handle.id()) != null)
				dirty = true;
		}
	}
	
	/** Searches for a handle in the local database.
	 * 
	 * <p>If the ID is currently known by the reference server, it is garanteed
	 * that this method won't return null, regardless of the current state of
	 * local database.</p>
	 * 
	 * @param id The ID of the handle to search for
	 * @return The found handle, or null if none matched
	 * @throws GridException if the ID is invalid
	 */
	public ExternalPeerHandle search(String id) throws GridException {
		if (id == null || id.length() == 0)
			throw new GridException("Cannot search for invalid ID");
		
		synchronized(local_database) {
			ExternalPeerHandle handle = local_database.get(id);
			if (handle == null) {
				// In case of cache miss, update
				refresh();
				handle = local_database.get(id);
			}
			return handle;
		}
	}
	
	/** Returns a complete snapshot of the local database.
	 * <p>Don't mess with the returned array if you want consistent behaviour
	 * as it's always the same one that is returned as long as the database
	 * doesn't change</p>
	 * <p>The array is not garanteed to be of the exact size of the database. If
	 * the array is larger, the first irrelevant element is set to
	 * <tt>null</tt>.</p>
	 * 
	 * @return An array of all handles locally known
	 */
	public ExternalPeerHandle[] searchAll() {

		synchronized(local_database) {
			// Database didn't change since last time
			if (!dirty)
				return snapshot;
			
			int size = local_database.size();

			// First time or first time with non-empty database
			if (snapshot == null
					|| (snapshot.length == 0 && size > 0))
				snapshot = new ExternalPeerHandle[size];
			
			// Too small array
			else if (snapshot.length < size) {
				// Use exponantial allocation
				int new_length = 2*snapshot.length;
				
				// Multiplying once might not be enough (e.g. if snapshot was
				// of 1 handle and we now have > 2)
				while (new_length < size)
					new_length *= 2;
				
				snapshot = new ExternalPeerHandle[new_length];
			}
			
			snapshot = local_database.values().toArray(snapshot);
			snapshot_length = size;

			dirty = false;
			return snapshot;
		}
	}
	
	/** Get k random handles from the cache.
	 * 
	 * @param k The number of random handles
	 * @return An array of k random handles.
	 * @throws GridException if <i>k > {@link #size()}</i>
	 */
	public ExternalPeerHandle[] search(int k) throws GridException {
		/* Rationale behind this separation of methods is that why we certainly
		 * cannot cache an already allocated array and modify it (other threads
		 * might be using it and we would then interact with them), the caller
		 * of these methods maybe just want multiple sets of K handles, where K
		 * is constant. The caller can then safely reuse his array as he has more
		 * information over how it's used (i.e. knows if it could cause conflicts).
		 */
		if (k < 1) {
			throw new GridException("search(): illegal handles count");
		}
		
		ExternalPeerHandle[] retval = null;

		synchronized(this.local_database) {
			ExternalPeerHandle[] dummy  =
				new ExternalPeerHandle[Math.min(k,this.local_database.size())];
			retval = search(dummy);
		}
		
		return retval;
	}
	
	/** Get {@value #DEFAULT_SEARCH_COUNT} random handles from the cache.
	 * 
	 * @return An array of {@value #DEFAULT_SEARCH_COUNT} random handles.
	 * @see #DEFAULT_SEARCH_COUNT
	 */
	public ExternalPeerHandle[] search() throws GridException {
		
		return search(SearchEngineClient.DEFAULT_SEARCH_COUNT);

	}	

	/** Get a number random handles from the cache.
	 * As this method fills it argument, it is probably more efficient from
	 * a memory POV that {@link #search(int)} (which allocates a new array
	 * every time).
	 * 
	 * @param handles an array of handles. It will be filled with the random
	 * handles
	 * @return The array of handles.
	 * @throws GridException if <i>handles.length > {@link #size()}</i>
	 */
	/* Another alorithm would be to generate all numbers between 0 and size(),
	 * and then to make a sequence of swaps as in
	 * http://www.maths.abdn.ac.uk/~igc/tch/mx4002/notes/node83.html
	 * The efficiency of both methods is dependant on the number of changes
	 * to local_database relative to the number of calls to search.
	 * Further optimisations would probably need to keep track of that ratio
	 * and change algorithm on the fly.
	 * An algorithm allowing the generation of an increasing sequence of K
	 * integers between N and M and requiring O(1) memory would be welcome
	 */
	private ExternalPeerHandle[] search(ExternalPeerHandle[] handles)
			throws GridException {
		synchronized(local_database) {
			
			int size = local_database.size();
			if (handles.length > size)
				throw new GridException("Cache only contains "+size+" handles");
			
			searchAll(); // refresh snapshot
			rand_set.clear(); // Prepare set for new operations
			
			for (int total = 0; total < handles.length; total++) {
				// Get a random unused handle (based on its index)
				int i = CommonMath.sampleUniformDist(0,snapshot_length-1);
				
				// Linear probing to meet the "unused" criteria
				while (rand_set.contains(lbg.common.Tank.newInteger(i)))
					i = (++i) % snapshot_length;
				
				// Register index as already taken
				rand_set.add(lbg.common.Tank.newInteger(i));
				
				// Assign random handle
				handles[total] = snapshot[i];
			}
			
			return handles;
		}
	}
	
	/** Returns the number of entries currently in this cache. */
	public int size() {
		synchronized(local_database) {
			return local_database.size();
		}
	}
	
	/** Returns a String representation of this client. */
	public String toString() {
		return new StringBuilder("[SearchEngineClient] connected to ")
			.append(reference_engine.toString())
			.append("\n  ->  ")
			.append(local_database.size())
			.append(" indexed peers")
			.toString();
	}
	
	/** Returns a detailed (i.e. including known peers) String representation
	 * of this client.
	 */
	public String toStringDetailed() {
		StringBuilder sb =
			new StringBuilder("[SearchEngineClient] Connected to ")
			.append(reference_engine.toString())
			.append("\n  ->  ");
		synchronized(local_database) {
				sb.append(local_database.size())
					.append(" indexed peers:");
				
				for (Object h: this)
					sb.append("\n  ->   ").append(((ExternalPeerHandle)h).id());
		}
		return sb.toString();
	}
	
	/** Returns an iterator over the handles known to this client. This is
	 * fast-fail and most likely to fail fast depending on timeout.<br>
	 * 
	 * {@link #search()} is a costy but more reliable alternative.
	 */
	public java.util.Iterator<ExternalPeerHandle> iterator() {
		return local_database.values().iterator();
	}
	
	public void run() {
		try {
			refresh();
		}
		catch (GridException e) {}
	}
	
	/** Stress test.
	 * <p>Usage: java lbg.searchengine.SearchEngineClient NUM_HANDLES SAMPLES RATIO. */
	/* Logging memory usage might be interesting */
	public static void main(String[] argv) throws Exception {
		// Fake search handle
		SearchEngineHandle handle = new SearchEngineHandle() {
			public void add(ExternalPeerHandle handle) throws GridException {}
			public SearchEngineResponse getHandles() throws GridException {
				return new SearchEngineResponse(new ExternalPeerHandle[0]);
			}
			public SearchEngineResponse getHandles(long since) throws GridException {
				return new SearchEngineResponse(new ExternalPeerHandle[0]);
			}
			public String id() {
				return "desu";
			}
		};
		
		SearchEngineClient client =
			new LiveSearchEngineClient(handle, null, 9001);
		
		int num_handles = new Integer(argv[0]).intValue();
		System.err.print("Generating " + num_handles + " handles... ");
		ExternalPeerHandle handles[] = new ExternalPeerHandle[num_handles];

		long time = System.currentTimeMillis();
		
		// Generate a set of dummy handles
		for (int i = 0; i < num_handles; i++) {
			handles[i] =
				new TestPeerHandle("peer"+CommonMath.sampleUniformDist());
		}
		time = System.currentTimeMillis() - time;
		System.err.print(time + "ms.\nAdding the handles to the client... ");
		
		// Everything is generated, add all at once
		time = System.currentTimeMillis();
		client.add(handles);
		time = System.currentTimeMillis() - time;
		
		// Search when creating new arrays each time
		final int samples = new Integer(argv[1]).intValue(); 
		final int K = (int) (new Double(argv[2]).doubleValue() * num_handles);
		System.err.print(time + "ms.\nRetrieving " + samples + " sets of " + K
			+ " random handles without array optimization... ");
		time = System.currentTimeMillis();
		for (int i = 0; i < samples; i++)
			client.search(K);
		time = System.currentTimeMillis() - time;
		
		// Search when keeping always the same array
		System.err.print(time + "ms.\nRetrieving " + samples + " sets of " + K
			+ " random handles with array optimization... ");
		handles = new ExternalPeerHandle[K];
		time = System.currentTimeMillis();
		for (int i = 0; i < samples; i++)
			client.search(handles);
		time = System.currentTimeMillis() - time;
		
		// Comparison of searchAll, search(db.size) and search(db.size/2)
		System.err.print(time + "ms.\n" + samples + " calls to searchAll() (approx)... ");
		time = System.currentTimeMillis();
		java.lang.reflect.Field f = SearchEngineClient.class.getDeclaredField("dirty");
		f.setAccessible(true);
		for (int i = 0; i < samples; i++) {
			f.setBoolean(client, true); // force realloc
			client.searchAll();
		}
		time = System.currentTimeMillis() - time;
		System.err.print(time + "ms.\n" + samples + " calls to search(" + num_handles + ")... ");
		time = System.currentTimeMillis();
		for (int i = 0; i < samples; i++)
			client.search(num_handles);
		time = System.currentTimeMillis() - time;
		System.err.print(time + "ms.\n" + samples + " calls to search(" + num_handles/2 + ")... ");
		time = System.currentTimeMillis();
		for (int i = 0; i < samples; i++)
			client.search(num_handles/2);
		time = System.currentTimeMillis() - time;
		System.err.println(time + "ms.");
	}
}

