/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SearchEngineServer.java

package lbg.searchengine;

import lbg.common.exception.GridException;
import lbg.common.net.*;
import lbg.peer.service.*;
import lbg.searchengine.service.*;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.logging.Logger;

/** Search Engine server.
 * @author Xavier Dalem
 * @since 2007-03-21
 * @todo Optimize getHandles() with a "dirty" flag
 */
public abstract class SearchEngineServer {
	
	public static final Logger SELOGGER = Logger.getLogger("lbg.searchengine");
	
	// ****************************************************************

	/* We need a list of entry times, a list of the associated handles and a
	 * HashSet of ids to speed up contains() operations. 
	 * However, as ArrayList's remove() operation is apparently reference-
	 * based, we need to keep the original object along with it's ID if we
	 * want to
	 * This structures guarantees the following efficiencies:
	 *   remove()	O(2n) from ArrayList (one call to indexof, one to remove)
	 *   add()		amortized O(n) from ArrayList
	 *   contains()	O(1) from HashMap
	 *   list()		O(n), maybe less for full dumps of the lists
	 * I cannot think of a custom structure that would yield significantly
	 * better results while keeping the entry times ease of use, so this will
	 * do for know.
	 */
	private ArrayList<Long> entry_times;
	        ArrayList<ExternalPeerHandle> handles_list;
	private HashMap<String, ExternalPeerHandle> handle_ids;
	
	/** Search engine ID */
	public final String id;
	/** Search engine hostname */
	public final String hostname;
	/** Search engine port */
	public final int port;
	/** Search engine cache refresh timeout */
	public final int timeout;
	/** Search Engine Service */
	private final SearchEngineService se_service;
	
	// ****************************************************************

	/** Creates a search engine server.
	 * @param id Search Engine id
	 * @param host Search Engine hostname
	 * @param port Search Engine port
	 * @param timeout Timeout before a handle is checked for activity
	 * (in seconds)
	 */
	SearchEngineServer(String id, String host, int port, int timeout)
		throws GridException {
		if (timeout <= 0)
			throw new GridException("Cannot set negative or null timeout");
		if (id == null || id.length() == 0)
			throw new GridException("Invalid ID");
		if (host == null || host.length() == 0)
			throw new GridException("Invalid host");
		if (!lbg.common.net.CommonNetHandling.isValidPort(port))
			throw new GridException("Invalid port");
		
		this.id = id;
		this.hostname = host;
		this.port = port;
		this.timeout = timeout;
		
		entry_times = new ArrayList<Long>();
		handles_list = new ArrayList<ExternalPeerHandle>();
		handle_ids = new HashMap<String, ExternalPeerHandle>();
		
		this.se_service =
			SearchEngineServiceFactory.createSearchEngineService(this);

	}
	
	public synchronized SearchEngineService service() {
		
		return this.se_service;
		
	}
	
	/** Adds a handle to this search engine. It is only added if the handle is
	 * alive as per {@link AbstractExternalPeerHandle#isAlive()} and a handle
	 * if the same ID isn't already known.
	 * 
	 * @param handle The handle to add
	 * @throws GridException if the handle is invalid
	 * @return <tt>true</tt> if a handle of the same ID wasn't already known
	 */
	public boolean add(ExternalPeerHandle handle)
			throws GridException {
		if (handle == null)
			throw new GridException("Cannot add null handle");
		
		if (! handle.isAlive()) {
			SELOGGER.info("Dropped request to add inactive handle "
				+ handle.toString());
			return true;
		}
		
		synchronized(entry_times) {
			// We only add elements if they weren't already registered
			if (! handle_ids.containsKey(handle.id())) {
				entry_times.add(lbg.LightweightBarteringGrid.getTime());
				handles_list.add(handle);
				// eeep... forget the following and win an infinte loop (I did)
				handle_ids.put(handle.id(), handle);
				return true;
			} else
				return false;
		}
	}
	
	/** Removes a given handle from the engine if it was known.
	 * This method is reference-safe (checks the handle ID and not the
	 * reference for removal).
	 * 
	 * @param handle The handle to remove
	 */
	protected final void remove(ExternalPeerHandle handle) {
		remove(handle.id());
	}
	
	/** Removes a given handle from the engine if it was known.
	 * 
	 * @param handleID The ID of the handle to remove
	 */
	protected void remove(String handleID) {
		synchronized(entry_times) {
			ExternalPeerHandle old = handle_ids.get(handleID);
			if (old != null) {
				handle_ids.remove(handleID);
				int i = handles_list.indexOf(old);
				handles_list.remove(i);
				entry_times.remove(i);
				SELOGGER.fine("Removed handle " + handleID);
			}
		}
	}
	
	public SearchEngineResponse getHandles() throws GridException {
		if (entry_times.size() == 0)
			return new SearchEngineResponse(new ExternalPeerHandle[0]);
		
		synchronized(entry_times) {
			return new SearchEngineResponse(handles_list.toArray(
				new ExternalPeerHandle[handles_list.size()]));
		}
	}
	
	public SearchEngineResponse getHandles(long since) throws GridException {
		if (entry_times.size() == 0)
			return new SearchEngineResponse(new ExternalPeerHandle[0]);
		
		synchronized(entry_times) {
			// >= can englobe more results than necessary but no less
			// Taking > makes race conditions in simulator as handles get
			// added in less than a millisecond
			if (entry_times.get(0) >= since) // should be faster (TODO: test)
				return new SearchEngineResponse(handles_list.toArray(
						new ExternalPeerHandle[handles_list.size()]));
			
			// Finding the M last elements from a list of N elements sorted in
			// increasing order and starting from the end will yield O(M) times
			// rather than O(N-M) times. If the server is crowded, M << N
			// and update requests are quite fast (above 'if' will save us
			// from infrequently updated clients)
			final int max_size = handles_list.size();
			int i = max_size-1;
			
			// we can stop at 1 because entry_times.get(0) >= since
			for (; i >= 1; i--) {
				if (entry_times.get(i) < since) {
					i++; // This element is useless
					break;
				}
			}
			
			// i could be equal to max_size
			ExternalPeerHandle[] new_ones =	new ExternalPeerHandle[max_size-i];
			
			// filling is O(M)
			for (int j = 0; j < new_ones.length; j++)
				new_ones[j] = handles_list.get(j+i);
			
			// Total: O(2M)
			
			return new SearchEngineResponse(new_ones);
		}
	}
	
	public String toString() {
		
		return new StringBuilder("[SearchEngineServer ").
			append(id).append("] @ ").
			append(hostname).append(':').append(port).
			toString();
		
	}
	
	// ****************************************************************

	/* Test to check handles. To have results, the calls to
	 * isAlive() should be commented
	 */
	/*public static void main(String[] argv) throws Exception {
		SearchEngineHandle handle = 
			new LiveSearchEngineHandle("nostromo", "nostromo", 4242);
		SearchEngineResponse res = null;
		System.out.println(handle.getHandles() + "\n---\n");
		handle.add(new lbg.peer.service.LivePeerHandle("hello1", "nostromo",
			42, 42, 42));
		System.out.println((res = handle.getHandles()) + "\n---\n");
		ThreadPool.sleep(2);
		handle.add(new lbg.peer.service.LivePeerHandle("hello2", "nostromo",
			42, 42, 42));
		System.out.println(handle.getHandles() + "\n---\n");
		System.out.println(handle.getHandles(res.timestamp) + "\n---\n");
		System.out.println(handle.getHandles() + "\n---\n");
	}*/

}

