/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// EnvironmentController.java

package lbg.simulator;

import lbg.common.exception.*;
import lbg.peer.*;
import lbg.peer.core.negotiator.*;
import lbg.peer.core.scheduler.*;

/**
 * @author Cyril Briquet
 */
public class EnvironmentController {

	public EnvironmentController(int peer_count) throws GridException {
		
		this.peer_count = peer_count;
		if (this.peer_count < 1) {
			throw new GridException("EnvironmentController(): "+
									"illegal Peers count");
		}
		
		this.peers = null;
		this.schedulers = new Scheduler[this.peer_count];
		this.negotiators = new Negotiator[this.peer_count];
		this.scf = new SchedulerControlFlags[this.peer_count];
		this.scf_has_changed = new boolean[this.peer_count];
		this.ncf = new NegotiatorControlFlags[this.peer_count];
		this.ncf_has_changed = new boolean[this.peer_count];
		for (int i = 0 ; i < this.peer_count ; i++) {
			this.scf[i] = new SchedulerControlFlags();
			this.scf_has_changed[i] = false;
			this.ncf[i] = new NegotiatorControlFlags();
			this.ncf_has_changed[i] = false;
		}
		this.can_schedule_and_negotiate = false;
		
	}

	public String toString() {
	
		StringBuilder sb = new StringBuilder();

		sb.append("[EnvironmentController]");
		
		return sb.toString();

	}
	
	// ****************************************************************

	public synchronized void setPeers(Peer[] peers) throws GridException {
		
		if (this.peers != null) {
			throw new GridException("setPeers(): Peers already set");
		}
		
		this.peers = peers.clone();
		
	}
	
	public synchronized void setNegotiatorAndScheduler(int peer_id,
													   Negotiator negotiator,
													   Scheduler scheduler)
		throws GridException {
		
		if ((peer_id < 0) || (peer_id >= this.peer_count)) {
			throw new GridException("setNegotiatorAndScheduler(): "+
									"illegal Peer id");
		}
		if (negotiator == null) {
			throw new GridException("setNegotiatorAndScheduler(): "+
									"illegal Negotiator");
		}
		if (scheduler == null) {
			throw new GridException("setNegotiatorAndScheduler(): "+
									"illegal Scheduler");
		}

		this.negotiators[peer_id] = negotiator;
		this.schedulers[peer_id] = scheduler;
		
	}
	
	// ****************************************************************

	public synchronized void
		notifyEnvironmentController(int peer_id,
									SchedulerController sched_ctrlr)
		throws GridException {
		
		if ((peer_id < 0) || (peer_id >= this.peer_count)) {
			throw new GridException("notifyEnvironmentController(): "+
									"illegal Peer id");
		}
		if (sched_ctrlr == null) {
			throw new GridException("notifyEnvironmentController(): "+
									"illegal Scheduler Controller");
		}
		
		sched_ctrlr.getAndResetControlFlags(this.scf[peer_id]);
		this.scf_has_changed[peer_id] = true;
		this.can_schedule_and_negotiate = true;
	
	}
	
	public synchronized void
		notifyEnvironmentController(int peer_id,
									NegotiatorController neg_ctrlr)
		throws GridException {
		
		if ((peer_id < 0) || (peer_id >= this.peer_count)) {
			throw new GridException("notifyEnvironmentController(): "+
									"illegal Peer id");
		}
		if (neg_ctrlr == null) {
			throw new GridException("notifyEnvironmentController(): "+
									"illegal Negotiator Controller");
		}

		neg_ctrlr.getAndResetControlFlags(this.ncf[peer_id]);
		this.ncf_has_changed[peer_id] = true;
		this.can_schedule_and_negotiate = true;
		
	}
	
	// ****************************************************************

	public synchronized boolean canScheduleAndNegotiate() {

		return this.can_schedule_and_negotiate;

	}

	public synchronized void scheduleAndNegotiate() throws GridException {

		for (int p = 0 ; p < this.peer_count ; p++) {
			if (this.ncf_has_changed[p]) {
				this.negotiators[p].negotiate(this.ncf[p]);
				this.ncf[p].clear();
				this.ncf_has_changed[p] = false;
			}
			if (this.scf_has_changed[p]) {
				this.schedulers[p].schedule(this.scf[p]);
				this.scf[p].clear();
				this.scf_has_changed[p] = false;
			}
		}
		
		this.can_schedule_and_negotiate = false;

	}
	
	// ****************************************************************

	public synchronized void processSelfEvents() throws GridException {
		
		for (Peer p : this.peers) {
			p.processSelfEvents();
		}
		
	}
	
	// ****************************************************************

	private       Peer[] peers;
	private final Scheduler[] schedulers;
	private final Negotiator[] negotiators;
	private final int peer_count;
	private final SchedulerControlFlags[] scf;
	private final boolean[] scf_has_changed;
	private final NegotiatorControlFlags[] ncf;
	private final boolean[] ncf_has_changed;
	private       boolean can_schedule_and_negotiate;
	
}

