/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Scenario.java

package lbg.simulator;

import lbg.common.exception.*;
import static lbg.common.io.CommonStringHandling.getStackTrace;
import lbg.peer.config.*;
import lbg.simulator.config.*;
import java.util.Properties;

/**
 * @author Cyril Briquet
 */
public class Scenario implements Cloneable {

	public Scenario(Properties scenario_props,String scenario_filename)
		throws GridException {
		
		this(scenario_props,scenario_filename,true,true,true);
		
	}

	public Scenario(Properties scenario_props,
					String scenario_filename,
					boolean load_grid_config,
					boolean load_grid_strategy_config,
					boolean load_randomizer_config)
		throws GridException {
		
		if (scenario_filename == null) {
			throw new GridException("Scenario(): illegal Scenario file name");
		}
		if (scenario_props == null) {
			throw new GridException("Scenario(): illegal Scenario Properties");
		}
		
		this.scenario_filename = scenario_filename;

		if (load_grid_config) {
			this.grid_config =
				new GridConfiguration(scenario_props);
		}
		if (load_grid_strategy_config) {
			this.grid_strategy_config =
				new GridStrategyConfiguration(scenario_props);
		}
		if (load_randomizer_config) {
			this.randomizer_config =
				new RandomizerConfiguration(scenario_props);
		}
		if (areConfigurationsConsistent() == false) {
			throw new GridException("Scenario(): inconsistent configurations");
		}
	
	}
	
	public synchronized String toString() {

		StringBuilder sb = new StringBuilder();
		
		sb.append("[Scenario ");
		sb.append(this.scenario_filename);
		sb.append("]");

		sb.append(format(true));
		
		return sb.toString();
		
	}
	
	/** Formats the Scenario as a property file.
	 * @return Scenario, formatted as a property file.
	 */
	public synchronized String format() {
		
		return format(false);
		
	}
		
	// ****************************************************************

	public synchronized GridConfiguration getGridConfiguration() {
		
		return this.grid_config;
		
	}

	public synchronized GridStrategyConfiguration
		getGridStrategyConfiguration() {
		
		return this.grid_strategy_config;
		
	}

	public synchronized RandomizerConfiguration getRandomizerConfiguration() {
		
		return this.randomizer_config;
		
	}

	// ****************************************************************

	public synchronized void loadGridConfiguration(Properties props)
		throws GridException {
	
		if (props == null) {
			throw new GridException("loadGridConfiguration(): "+
									"illegal Properties");
		}
		
		this.grid_config = new GridConfiguration(props);
		
		if (areConfigurationsConsistent() == false) {
			throw new GridException("Scenario(): inconsistent configurations");
		}

	}
	
	public synchronized void setGridConfiguration(GridConfiguration grid_config)
		throws GridException {
		
		if (grid_config == null) {
			throw new GridException("setGridConfiguration(): "+
									"illegal configuration");
		}
		
		this.grid_config = grid_config;
	
	}

	public synchronized void loadGridStrategyConfiguration(Properties props)
		throws GridException {

		if (props == null) {
			throw new GridException("loadGridStrategyConfiguration(): "+
									"illegal Properties");
		}
		
		this.grid_strategy_config =
			new GridStrategyConfiguration(props);
		
		if (areConfigurationsConsistent() == false) {
			throw new GridException("Scenario(): inconsistent configurations");
		}

	}
	
	public synchronized
		void setGridStrategyConfiguration(GridStrategyConfiguration
										  grid_strategy_config)
		throws GridException {
		
		if (grid_strategy_config == null) {
			throw new GridException("setGridStrategyConfiguration(): "+
									"illegal configuration");
		}
		
		this.grid_strategy_config = grid_strategy_config;
	
	}

	public synchronized void loadRandomizerConfiguration(Properties props)
		throws GridException {

		if (props == null) {
			throw new GridException("loadRandomizerConfiguration(): "+
									"illegal Properties");
		}
		
		this.randomizer_config = new RandomizerConfiguration(props);
		
	}
	
	public synchronized
		void setRandomizerConfiguration(RandomizerConfiguration
										randomizer_config)
		throws GridException {
		
		if (randomizer_config == null) {
			throw new GridException("setRandomizerConfiguration(): "+
									"illegal configuration");
		}
		
		this.randomizer_config = randomizer_config;
	
	}

	// ****************************************************************

	/** Outputs the Scenario,
	 * formatted either as a property file or as a log file.
	 * @param log_format true to format Scenario as a log file
	 * @return formatted Scenario
	 */
	private synchronized String format(boolean log_format) {
		
		StringBuilder sb = new StringBuilder();

		if (this.grid_config != null) {
			if (log_format) sb.append("\n  ->  ");
			sb.append(this.grid_config.format(log_format));
		}
		if (this.grid_strategy_config != null) {
			if (log_format) sb.append("\n  ->  ");
			sb.append(this.grid_strategy_config.format(log_format));
		}
		if (this.randomizer_config != null) {
			if (log_format) sb.append("\n  ->  ");
			sb.append(this.randomizer_config.format(log_format));
		}
		
		return sb.toString();

	}
	
	private synchronized boolean areConfigurationsConsistent() {
		
		if ((this.grid_config == null) ||
			(this.grid_strategy_config == null)) {
			return true;
		}
		
		return
			(this.grid_config.PEER_GROUPS()==this.grid_strategy_config.size());
		
	}
	

	// ****************************************************************

	@Override
	public synchronized Object clone() {

		Scenario sc = null;
		
		try {
			sc = (Scenario)super.clone();
			sc.setGridConfiguration((GridConfiguration)this.grid_config.clone());
			sc.setGridStrategyConfiguration((GridStrategyConfiguration)
											this.grid_strategy_config.clone());
			sc.setRandomizerConfiguration((RandomizerConfiguration)
										  this.randomizer_config.clone());
		}
		catch (GridException e) {
			throw new InternalError("clone(): cannot clone Scenario"+
									getStackTrace(e));
		}
		catch (CloneNotSupportedException e) {
			throw new InternalError("clone(): cannot clone Scenario");
		}
		
		return sc;
		
	}
	
	// ****************************************************************

	private final String                    scenario_filename;

	private       GridConfiguration         grid_config;
	private       GridStrategyConfiguration grid_strategy_config;
	private       RandomizerConfiguration   randomizer_config;

}

