/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Simulator.java

package lbg.simulator;

import lbg.*;
import lbg.common.exception.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.common.io.*;
import lbg.common.math.*;
import lbg.common.net.*;
import lbg.peer.workflow.*;
import lbg.protocol.data.*;
import lbg.resource.*;
import lbg.simulator.event.*;
import lbg.task.*;

import java.util.Date;
import java.util.Properties;

/** Lightweight Bartering Grid Simulator.
 * @author Cyril Briquet
 */
public class Simulator {

	public Simulator(Properties scenario_props) throws GridException {
	
		this(scenario_props,"<unknown filename>");
		
	}

	public Simulator(Properties scenario_props,String scenario_filename)
		throws GridException {

		LightweightBarteringGrid.setMode(VirtualizationLevel.SIM_MODE);

		System.out.print(LightweightBarteringGrid.welcome());

		this.env =
			new SimulatorEnvironment(this,scenario_props,scenario_filename);
		this.el = this.env.eventList();
		this.sim_output_grid_data_factory =
			new SimulatedGridDataFactory(true,1);
		this.base_extra_self_events = this.env.extraSelfEvents();
		this.remaining_extra_self_events = this.base_extra_self_events;
		
		this.submitted_jobs_count = 0;
		
		int[] JPP = this.env.scenario().getGridConfiguration().JOBS_PER_PEER();
		int[] GS = this.env.scenario().getGridConfiguration().PEER_GROUP_SIZE();
		float tj = 0;
		for (int g = 0 ; g < JPP.length ; g++) {
			tj += JPP[g] * GS[g];
		}
		this.one_percent_step = tj / 100.0f;
		this.current_percent = 0;
		
	}

	public synchronized String toString() {

		StringBuilder sb = new StringBuilder();

		sb.append("[Lightweight Bartering Grid Simulator]\n\n");
		sb.append(this.env);
		sb.append("\n\n");
		sb.append(this.el);

		return sb.toString();
		
	}

	// ****************************************************************

	/** Simulation loop.
	 */
	public synchronized void simulate() throws GridException {
		
		System.out.println("[Simulation starts on "+new Date()+
						   " @ "+CommonNetHandling.getLocalHostName()+"]\n");
		long sim_start_real_time = System.currentTimeMillis();
		System.out.print("[");
		for (int i = 0 ; i < 100 ; i++) System.out.print("-");
		System.out.print("]\n ");

		// ****************
		
		while (this.el.isEmpty() == false) {
		
			// Process Event
			EventContainer event_container = this.el.extractFirstEvent();
			handleEvent(event_container.event());
			
			// After the processed Event is the last Event to happen
			// at current time stamp, ask all Peers to reschedule
			if (event_container.isLastEventAtCurrentTimestamp()) {
				schedule();
			}

		}

		// ****************

		System.out.println("\n\n[Simulation successfully ends on "+new Date()+
						   " @ "+CommonNetHandling.getLocalHostName()+"]");

		long sim_end_real_time = System.currentTimeMillis();
		long sim_real_time = (sim_end_real_time-sim_start_real_time)/1000;
		System.out.println("  ~>  simulation.real_time = "+sim_real_time);

		System.out.println("  ->  "+this.el);

		//System.out.println(this.env.gridMeanJobCompletionTimeSummary());

		CommonFileHandling.dumpSeparator();
		CommonFileHandling.dump(this.env);
		System.out.println("\n\n  ->  simulation.real_time = "+sim_real_time);
		CommonFileHandling.dumpSeparator();

	}

	// ****************************************************************

	public static void main(String[] args) {
	
		try {
			checkArgs(args);
			
			Properties scenario_props =
				CommonFileHandling.loadProperties(args[0]);
			Simulator sim = new Simulator(scenario_props,args[0]);
			sim.simulate();
		}
		catch (GridException e) {
			LightweightBarteringGrid.handleGridException(e);
		}
		catch (Throwable thr) {
			LightweightBarteringGrid.
				handleGridException(new GridException(thr));
		}
		
	}

	// ****************************************************************

	/** Let every Peer Negotiate and schedule Tasks if activated,
	 * until no Peer is activated
	 * (i.e. a stable scheduling state has been reached).
	 *
	 * A Negotiation clearing system supports the scheduling process
	 * (i.e. when there are multiple calls of receiveSupplyingRequests()
	 * of a given Peer, at the same timestamp, NoF is be taken into account)
     *
	 */
	private synchronized void schedule() throws GridException {
		
		EnvironmentController ctrlr = this.env.controller();
		if (ctrlr.canScheduleAndNegotiate()) {
			ctrlr.scheduleAndNegotiate();
		}
		
	}

	private synchronized void handleEvent(Event evt) throws GridException {
		
		if (evt == null) {
			throw new GridException("handleEvent(): illegal Event");
		}
		
		if (evt instanceof SelfEvent) {
			processSelfEvent((SelfEvent)evt);
		}
		// ****************************************************************
		else if (evt instanceof SubmittedJobEvent) {
			processSubmittedJobEvent((SubmittedJobEvent)evt);
		}
		// ****************************************************************
		else if (evt instanceof TaskEvent) {
			if (evt instanceof CompletedTaskEvent) {
				processCompletedTaskEvent((CompletedTaskEvent)evt);
			}
			else if (evt instanceof FailedTaskEvent) {
				processFailedTaskEvent((FailedTaskEvent)evt);
			}
			else {
				throw new GridException("simulate(): "+
										"unmanaged type of TaskEvent");
			}
		}
		// ****************************************************************
		else {
			throw new GridException("simulate(): unknown Event type");
		}
		
	}
	
	// ****************************************************************

	private synchronized void processSubmittedJobEvent(SubmittedJobEvent evt)
		throws GridException {

		// ****************
		int dummy =
			Math.round(((float)(++this.submitted_jobs_count)) /
						this.one_percent_step);
		if (dummy > this.current_percent) {
			int delta = dummy - this.current_percent;
			this.current_percent = dummy;
			for (int i = 0 ; i < delta ; i++) System.out.print("*");
			System.gc(); /* <---- VERY IMPORTANT, RELEASE MEMORY EACH TIME % */
		}
		// ****************
		
		// Submit Job
		evt.user().submitJob(new JobDescriptor());

		// Rescheduling done in the main loop of Simulator, when required

	}

	private synchronized void processCompletedTaskEvent(CompletedTaskEvent evt)
		throws GridException {

		if (evt == null) {
			throw new GridException("processCompletedTaskEvent(): "+
									"illegal Event");
		}

		Resource res = evt.runner();
		if (res == null) {
			throw new GridException("processCompletedTaskEvent(): "+
									"illegal Resource");
		}
		
		Task t = evt.task();
		if (t == null) {
			throw new GridException("processCompletedTaskEvent(): "+
									"illegal Task");
		}
		
		// Complete Task (Resource viewpoint)
		SimulatedGridData sim_data =
			this.sim_output_grid_data_factory.
				createSimulatedGridData(t.consumerID(),"*task-result*");
		res.completeTask(sim_data,DUMMY_LOG);
		
		// Sanity check
		assert((t.output() != null) &&
			   (/**/(t.output().isLocalTask() &&
			    	 (t.consumerID().equals(evt.supplierID()) == true))/**/
				||
			    /**/(t.output().isSupplyingTask() &&
			     	 (t.consumerID().equals(evt.supplierID()) == false) &&
					 (t.consumerID().equals(evt.consumerID()) == true))/**/));
		
		// Rescheduling done in the main loop of Simulator, when required

	}

	private synchronized void
		processSelfEvent(SelfEvent evt) throws GridException {

		EnvironmentController ctrlr = this.env.controller();
		ctrlr.processSelfEvents();

		Event new_self_evt = this.el.factory().createSelfEvent();
		EventList el = this.env.eventList();
		if (el.isEmpty() == false) {
			el.addEvent(new_self_evt);
			this.remaining_extra_self_events = this.base_extra_self_events;
		}
		else { // No more Events to process, but time-outs may still occur...
			if (this.remaining_extra_self_events-- > 0) {
				el.addEvent(new_self_evt);
			}
		}

		// Rescheduling done in the main loop of Simulator, when required

	}

	private synchronized void processFailedTaskEvent(FailedTaskEvent evt)
		throws GridException {

		if (evt == null) {
			throw new GridException("processFailedTaskEvent(): "+
									"illegal Event");
		}
		
		Resource res = evt.runner();
		if (res == null) {
			throw new GridException("processFailedTaskEvent(): "+
									"illegal Resource");
		}
		
		Task t = evt.task();
		if (t == null) {
			throw new GridException("processFailedTaskEvent(): "+
									"illegal Task");
		}
		
		// Fail Task (Resource viewpoint)
		res.cancelTask();
		
		// Sanity check (.equals() not required)
		assert(t.isCompleted() == false);
		
		// Rescheduling done in the main loop of Simulator, when required

	}

	// ****************************************************************

	private static void checkArgs(String[] args) throws GridException {

		if ((args.length != 1) ||
			(args[0] == null) || (args[0].length() < 1)) {
			System.err.println("\nUsage: java -cp classpath "+
							   Simulator.class.getName()+
							   " scenario.txt\n");
			throw new GridException("checkArgs(): illegal arguments");
		}

	}

	// ****************************************************************

	private final SimulatorEnvironment env;
	private final EventList el;
	private final SimulatedGridDataFactory sim_output_grid_data_factory;
	private final long base_extra_self_events;
	private       long remaining_extra_self_events;

	private       int   submitted_jobs_count;
	private final float one_percent_step;
	private       int   current_percent;
	
}

