/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SimulatorEnvironment.java

package lbg.simulator;

import lbg.*;
import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.common.math.*;
import lbg.peer.*;
import lbg.peer.config.*;
import lbg.peer.core.*;
import lbg.peer.service.*;
import lbg.resource.*;
import lbg.resource.config.*;
import lbg.searchengine.*;
import lbg.searchengine.service.*;
import lbg.simulator.config.*;
import lbg.simulator.event.*;
import lbg.user.*;
import static java.lang.Math.max;
import java.util.Iterator;
import java.util.Properties;
import java.util.Vector;

/**
 * @author Cyril Briquet
 */
public class SimulatorEnvironment implements ISimulator {

	public SimulatorEnvironment(Simulator sim,Properties scenario_props,
							    String scenario_filename)
		throws GridException {

		if (sim == null) {
			throw new GridException("SimulatorEnvironment(): "+
									"illegal Simulator");
		}
		if (scenario_props == null) {
			throw new GridException("SimulatorEnvironment(): "+
									"illegal Scenario Properties");
		}
		if (scenario_filename == null) {
			throw new GridException("SimulatorEnvironment(): "+
									"illegal Scenario file name");
		}
		
		this.sc = new Scenario(scenario_props,scenario_filename);
		LightweightBarteringGrid.setSimulator(sim,this.sc); // <-- IMPORTANT !

		CommonMath.setSeed(this.sc.getRandomizerConfiguration().
							SIMULATION_SEED());
		
		this.el = new EventList(this.sc);
		this.search_engine_server =
			new SimulatedSearchEngineServer("SearchEngine","SearchEngine",
											64000,
											this.sc.getGridConfiguration().
											SEARCH_TIMEOUT); // dummy value
		// s.e. server timeout = SEARCH_TIMEOUT just to sync on s.e. clients

		if (isValidPeerCount() == false) {
			throw new GridException("generatePeers(): "+
									"corrupted Grid configuration");
		}

		assert(CommonFileHandling.dumpSeparator());
		assert(CommonFileHandling.dump(this));

		int peer_count = this.sc.getGridConfiguration().PEER_COUNT();
		this.peers_handles = new PeerHandle[peer_count];
		this.peers = new Peer[peer_count];
		this.env_ctrlr = new EnvironmentController(peer_count);
		this.extra_self_events = generatePeers();
		this.env_ctrlr.setPeers(this.peers);

		this.users = new UserAgent[peer_count];
		generateUserAgents();
		
		assert(CommonFileHandling.dump(dumpExtraSelfEvents()));
		assert(CommonFileHandling.dumpSeparator());

	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[SimulatorEnvironment]\n  ->  ");
		sb.append(this.sc);
		sb.append("\n\n  ->  ");
		sb.append(this.search_engine_server);
		sb.append("\n");

		if (this.peers_handles != null) {
			int p_c = this.sc.getGridConfiguration().PEER_COUNT();
			for (int i = 0 ; i < p_c ; i++) {
				sb.append("\n  ->  ");
				sb.append(this.peers_handles[i]);
			}
			if (this.peers == null) sb.append("\n");
		}

		if (this.peers != null) {
			try {
				double total_mean_utilization = 0.0;
				double[] group_mean_utilization =
					new double[this.sc.getGridConfiguration().PEER_GROUPS()];
				double total_mean_blacklisting_ratio = 0.0;
				double[] group_mean_blacklisting_ratio =
					new double[this.sc.getGridConfiguration().PEER_GROUPS()];
				double total_mean_job_completion_time = 0.0;
				double total_mean_job_completion_time_nm5 = 0.0;
				double[] group_mean_job_completion_times =
					new double[this.sc.getGridConfiguration().PEER_GROUPS()];
				double[] group_mean_job_completion_times_nm5 =
					new double[this.sc.getGridConfiguration().PEER_GROUPS()];
				double total_stddev_job_completion_time = 0.0;
				double[] group_stddev_job_completion_times =
					new double[this.sc.getGridConfiguration().PEER_GROUPS()];
				double total_favor_ratio = 0.0;
				double[] group_favor_ratios =
					new double[this.sc.getGridConfiguration().PEER_GROUPS()];
				int p = 0;
				Measurement mfr_m =
					MeasurementFactory.
						createSummarizedMeasurement("mean-favor-ratio");
				for (int g = 0 ;
					 g < this.sc.getGridConfiguration().PEER_GROUPS() ;
					 g++) {
					group_mean_utilization[g] = 0.0;
					group_mean_blacklisting_ratio[g] = 0.0;
					group_mean_job_completion_times[g] = 0.0;
					group_mean_job_completion_times_nm5[g] = 0.0;
					group_stddev_job_completion_times[g] = 0.0;
					group_favor_ratios[g] = 0.0;
					for (int i = 0 ;
						 i<this.sc.getGridConfiguration().PEER_GROUP_SIZE()[g] ;
						 i++,p++) {
						sb.append("\n\n\n  ->  ");
						sb.append(this.peers[p]);
						group_mean_utilization[g] +=
							CommonMath.filterToZeroIfNaN(
								this.peers[p].rms().utilization());
						group_mean_blacklisting_ratio[g] +=
							CommonMath.filterToZeroIfNaN(
								this.peers[p].scheduler().
									getBlacklistedSupplierRatio());
						group_mean_job_completion_times[g] +=
							CommonMath.filterToZeroIfNaN(
								this.peers[p].qms().
									getMeanJobCompletionTime());
						group_mean_job_completion_times_nm5[g] +=
							CommonMath.filterToZeroIfNaN(
								this.peers[p].qms().
									getMeanJobCompletionTimeOfNm5FirstKJobs());
						group_stddev_job_completion_times[g] +=
							CommonMath.filterToZeroIfNaN(
								this.peers[p].qms().
									getStdDevJobCompletionTime());
						double fr = this.peers[p].register().
							getNeighborhoodBarteringProfile().getFavorRatio();
						fr = CommonMath.filterToZeroIfNaN(fr);
						group_favor_ratios[g] += fr;
						mfr_m.addObservation(fr);
					}
					total_mean_job_completion_time +=
						group_mean_job_completion_times[g];
					total_mean_job_completion_time_nm5 +=
						group_mean_job_completion_times_nm5[g];
					total_mean_utilization += group_mean_utilization[g];
					total_mean_blacklisting_ratio +=
						group_mean_blacklisting_ratio[g];
					total_stddev_job_completion_time +=
						group_stddev_job_completion_times[g];
					total_favor_ratio +=  group_favor_ratios[g];
				}
				if (p != this.sc.getGridConfiguration().PEER_COUNT()) {
					sb.append(" (INTERNAL ERROR)");
				}
				sb.append("\n\n  ->  ");
				
				sb.append("[Grid]");
				sb.append("\n  ~>  ");
				sb.append("grid.job_completion_time.mean     = ");
				sb.append(CommonMath.div_label_2dd(total_mean_job_completion_time,
												   this.sc.
													getGridConfiguration().
													PEER_COUNT()));
				sb.append("\n  ~>  ");
				sb.append("grid.job_completion_time.stddev   = ");
				sb.append(CommonMath.div_label_2dd(total_stddev_job_completion_time,
												   this.sc.
													getGridConfiguration().
													PEER_COUNT()));
				sb.append("\n  ~>  ");
				sb.append("grid.job_completion_time.mean.nm5 = ");
				sb.append(CommonMath.div_label_2dd(total_mean_job_completion_time_nm5,
												   this.sc.
													getGridConfiguration().
													PEER_COUNT()));
				sb.append("\n  ~>  ");
				sb.append("grid.utilization.mean             = ");
				sb.append(CommonMath.div_label_2dd(total_mean_utilization,
												   this.sc.
													getGridConfiguration().
													PEER_COUNT()));
				sb.append("\n  ~>  ");
				sb.append("grid.blacklisting.suppliers.mean  = ");
				sb.append(CommonMath.div_label_2dd(total_mean_blacklisting_ratio,
												   this.sc.
													getGridConfiguration().
													PEER_COUNT()));
				sb.append("\n  ->  ");
				sb.append("Job completion time = { mean=");
				sb.append(CommonMath.div_label_2dd(total_mean_job_completion_time,
												   this.sc.
													getGridConfiguration().
													PEER_COUNT()));
				sb.append(", std dev=");
				sb.append(CommonMath.div_label_2dd(total_stddev_job_completion_time,
												   this.sc.
													getGridConfiguration().
													PEER_COUNT()));
				sb.append(", mean (on N-5 first Jobs)=");
				sb.append(CommonMath.div_label_2dd(total_mean_job_completion_time_nm5,
												   this.sc.
													getGridConfiguration().
													PEER_COUNT()));
				sb.append(" }");
				sb.append("\n  ~>  ");
				sb.append("grid.favors_ratio.mean = ");
				sb.append(CommonMath.div_label_2dd(total_favor_ratio,
												   this.sc.
													getGridConfiguration().
													PEER_COUNT()));
				
				sb.append("\n  ->  ");
				sb.append("{SIM} Favor Ratio (gained / donated) = { mean=");
				sb.append(CommonMath.div_label_2dd(total_favor_ratio,
												   this.sc.
													getGridConfiguration().
													PEER_COUNT()));
				sb.append(", std dev=");
				sb.append(CommonMath.div_label_2dd(mfr_m.getStandardDeviation(),
												   1.0));
				sb.append(" }");
				for (int g = 0 ;
					 g < this.sc.getGridConfiguration().PEER_GROUPS() ;
					 g++) {
					sb.append("\n  ->  ");
					sb.append("[Peer Group ");
					sb.append(g);
					sb.append("]");
					sb.append("\n  ->  ");
					sb.append("mean Job completion time = { mean=");
					sb.append(CommonMath.
								div_label_2dd(group_mean_job_completion_times[g],
											  this.sc.getGridConfiguration().
												PEER_GROUP_SIZE()[g]));
					sb.append(", std dev=");
					sb.append(CommonMath.
								div_label_2dd(group_stddev_job_completion_times[g],
											  this.sc.getGridConfiguration().
												PEER_GROUP_SIZE()[g]));
					sb.append(", mean (on N-5 first Jobs)=");
					sb.append(CommonMath.
								div_label_2dd(group_mean_job_completion_times_nm5[g],
											  this.sc.getGridConfiguration().
												PEER_GROUP_SIZE()[g]));
					sb.append(", expected-optimum-if-no-contention=");
					// (mean_TASK_LEN / mean_res_power) *
					// (TASKS_PER_JOB / mean_res_count)
					GridConfiguration grid_config =
						this.sc.getGridConfiguration();
					double mean_res_power =
						CommonMath.mean(grid_config.RES_POWER_LO()[g],
										grid_config.RES_POWER_HI()[g]);
					double mean_res_count = 
						grid_config.PEER_POWER()[g]/mean_res_power;
					double expected =
						(CommonMath.mean(grid_config.TASK_LEN_LO()[g],						
										grid_config.TASK_LEN_HI()[g]) /
						mean_res_power)
						*
						(((double)grid_config.TASKS_PER_JOB()[g]) /
						(mean_res_count));
					sb.append(CommonMath.div_label_2dd(expected,1.0));
					sb.append(" }");
					sb.append("\n  ->  ");
					sb.append("{SIM} Favor Ratio (gained / donated) = { mean=");
					sb.append(CommonMath.
								div_label_2dd(group_favor_ratios[g],
											  this.sc.getGridConfiguration().
												PEER_GROUP_SIZE()[g]));
					sb.append(" }");
				}
			}
			catch (GridException e) {
				sb.append("\n*An unexpected, Measurement-related Exception ");
				sb.append("occured while generating this log*\n");
				sb.append(CommonStringHandling.getStackTrace(e));
				sb.append("\n");
			}
		}

		if (this.users != null) {
			sb.append("\n");
			for (int i = 0 ; i < this.users.length ; i++) {
				sb.append("\n  ->  ");
				sb.append(this.users[i]);
			}
		}
		
		sb.append("\n\n  ->  ");
		sb.append(Tank.toStaticString());

		sb.append(dumpExtraSelfEvents());
		
		return sb.toString();

	}

	synchronized String gridMeanJobCompletionTimeSummary() {

		StringBuilder sb = new StringBuilder();
		
		if (this.peers != null) {
			try {
				double total_mean_job_completion_time = 0.0;
				double total_mean_job_completion_time_nm5 = 0.0;
				double[] group_mean_job_completion_times =
					new double[this.sc.getGridConfiguration().PEER_GROUPS()];
				double[] group_mean_job_completion_times_nm5 =
					new double[this.sc.getGridConfiguration().PEER_GROUPS()];
				int p = 0;
				for (int g = 0 ;
					 g < this.sc.getGridConfiguration().PEER_GROUPS() ;
					 g++) {
					group_mean_job_completion_times[g] = 0.0;
					group_mean_job_completion_times_nm5[g] = 0.0;
					for (int i = 0 ;
						 i<this.sc.getGridConfiguration().PEER_GROUP_SIZE()[g] ;
						 i++,p++) {
						group_mean_job_completion_times[g] +=
							CommonMath.filterToZeroIfNaN(
								this.peers[p].qms().
									getMeanJobCompletionTime());
						group_mean_job_completion_times_nm5[g] +=
							CommonMath.filterToZeroIfNaN(
								this.peers[p].qms().
									getMeanJobCompletionTimeOfNm5FirstKJobs());
					}
					total_mean_job_completion_time +=
						group_mean_job_completion_times[g];
					total_mean_job_completion_time_nm5 +=
						group_mean_job_completion_times_nm5[g];
				}
				if (p != this.sc.getGridConfiguration().PEER_COUNT()) {
					sb.append(" (INTERNAL ERROR)");
				}
				sb.append("\n  ->  ");
				sb.append("grid.job_completion_time.mean = ");
				sb.append(CommonMath.div_label_2dd(total_mean_job_completion_time,
												   this.sc.
													getGridConfiguration().
													PEER_COUNT()));
				sb.append("\n  ->  ");
				sb.append("grid.job_completion_time.mean.nm5 = ");
				sb.append(CommonMath.div_label_2dd(total_mean_job_completion_time_nm5,
												   this.sc.
													getGridConfiguration().
													PEER_COUNT()));
			}
			catch (GridException e) {}
		}
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized Scenario scenario() {
		
		return this.sc;
		
	}
	
	public long extraSelfEvents() {
	
		return this.extra_self_events;
		
	}
	
	public synchronized EventList eventList() {
		
		return this.el;
		
	}
	
	/* currently unused
	public synchronized PeerSearchEngine searchEngine() {
		
		return this.search_engine;
		
	}

	public synchronized Peer[] peers() {
		
		return this.peers;
		
	}

	public synchronized UserAgent[] users() {
		
		return this.users;
		
	}
	*/
	
	public synchronized EnvironmentController controller() {
	
		return this.env_ctrlr;
		
	}
	
	public synchronized void activateLocalScheduling() throws GridException {
	
		for (int p = 0 ; p < this.peers.length ; p++) {
			this.peers[p].schedulerController().activateLocalScheduling();
		}
		
	}

	// ****************************************************************

	/** ISimulator implementation.
	 */
	public synchronized void addJobSubmission(UserAgent user,String peer_id)
		throws GridException {
		
		if (user.submittedJobs() <
			this.sc.getGridConfiguration().JOBS_PER_PEER(peer_id)) {
			Event evt = this.el.factory().createJobEvent(user,peer_id);
			this.el.addEvent(evt);
		}
		/*else {
			System.out.println("All jobs submitted to Peer "+this.peer.id());
		}*/

	}

	// ****************************************************************

	/* currently unused
	public synchronized Peer resolvePeer(String id) throws GridException {
	
		return this.peers[this.sc.getGridConfiguration().resolvePeerID(id)];
		
	}
	*/
	
	// ****************************************************************

	/** Creates Resources for target Peer.
	 * <p>Power repartition between Resources depends on PEER_POWER, RES_COUNT,
	 * RES_POWER_LO, RES_POWER_HI..
	 * 
	 * <p>Power repartition algorithm:
	<pre>
	if (res_count == 0) {
		// random number of Resources, variable Resource power (in [lo..hi])
	}
	else if (res_count > 0) {
		if (peer_power > 0) {
			// fixed number of Resources, fixed Resource power (= peer_power / res_count)
		}
		else if (peer_power == 0) {
			// fixed number of Resources, variable Resource power (in [lo..hi])
		}
	}</pre>
	 */
	public static SimulatedResource[]
		createPeerResources(InternalPeerHandle owner_peer,
							int peer_power,int res_count,int lo_power,int hi_power,
							int cache_capacity,int mtbf,EventList event_list)
		throws GridException {
		
		if (owner_peer == null) {
			throw new GridException("createResources(): "+
									"illegal Peer");
		}

		if (peer_power <= 0) {
			throw new GridException("createResources(): "+
									"illegal Peer power specs");
		}
		if ((lo_power <= 0) || (hi_power < lo_power)) {
			throw new GridException("createResources(): "+
									"illegal Resource power specs ("+
									lo_power+","+hi_power+")");
		}

		if (cache_capacity < 1) {
			throw new GridException("createResources(): "+
									"illegal Resource cache capacity specs");
		}
		
		if (mtbf < 0) {
			throw new GridException("createResources(): "+
									"illegal Resource mean time "+
									"between failures");
		}

		// BEGIN attribution of Resource power
		Vector<Integer> res_power = new Vector<Integer>();
		
		if (res_count == 0) {

			if (hi_power > peer_power) {
				throw new GridException("createResources(): "+
										"illegal Resource power specs");
			}
			if (peer_power == 0) {
				throw new GridException("createResources(): "+
										"illegal Peer power spec");
			}

			int current_power = 0;
			while (current_power < peer_power) {
				int power =
					CommonMath.sampleUniformDist(lo_power,hi_power);
				if (current_power + power > peer_power) {
					power = peer_power-current_power;
				}
				res_power.add(Tank.newInteger(power));
				res_count++;
				current_power += power;
			}
			if (current_power < peer_power) {
				int power = peer_power-current_power;
				res_power.add(Tank.newInteger(power));
				res_count++;
				current_power += power;
			}		
		
			if (peer_power != current_power) {
				throw new GridException("createResources(): "+
										"corrupted Resource power sum");
			}
			
		}
		else if (res_count > 0) {
			
			if (peer_power > 0) {
			
				if (peer_power % res_count != 0) {
					throw new GridException("createResources(): "+
											"incompatible Resource count");
				}
			
				int power = peer_power / res_count;
				for (int i = 0 ; i < res_count ; i++) {
					res_power.add(Tank.newInteger(power));
				}
				
			}
			
			else if (peer_power == 0) { 
				
				for (int i = 0 ; i < res_count ; i++) {
					int power =
						CommonMath.sampleUniformDist(lo_power,hi_power);
					res_power.add(Tank.newInteger(power));
				}
				
			}
			
			else /* if (peer_power < 0) */ {
				
				throw new GridException("createResources(): "+
										"illegal Peer power spec");

			}
			
		}
		else /* if (res_count < 0) */ {
			
			throw new GridException("createResources(): "+
									"illegal Resource count");
			
		}
		// END attribution of Resource power

		// BEGIN Resource generation
		// SUM_i res_power[i] = peer_power
		SimulatedResource[] resource_set = new SimulatedResource[res_count];
		Iterator<Integer> it = res_power.iterator();
		for (int i = 0 ; i < res_count ; i++) {
			if (it.hasNext() == false) {
				throw new GridException("createResources(): "+
										"data inconsistency");
			}
			
			InternalPeerHandle internal_peer_handle = owner_peer;
			String res_i_id = generateResourceID(internal_peer_handle,i);
			int res_i_power = it.next();
			ResourceBasicConfiguration res_basic_config =
				new ResourceBasicConfiguration(res_i_id,res_i_power);
			ResourceDataConfiguration res_data_config =
				new ResourceDataConfiguration(cache_capacity);
			ResourceReliabilityConfiguration res_rel_config =
				new ResourceReliabilityConfiguration(mtbf);
			ResourceNetworkConfiguration res_net_config =
				new ResourceNetworkConfiguration();

			resource_set[i] =
				new SimulatedResource(res_basic_config,res_data_config,
									  res_rel_config,res_net_config,
									  internal_peer_handle,event_list);
		}
		if (it.hasNext() == true) {
			throw new GridException("createResources(): "+
									"data inconsistency");
		}
		// END Resource generation

		// Perform some sanity checks
		if ((res_count < 1) ||(resource_set.length < 1)) {
			throw new GridException("createResources(): "+
									"corrupted Resource count");
		}
		if (peer_power > 0) {
			int check_power_sum = 0;
			for (int i = 0 ; i < resource_set.length ; i++) {
				check_power_sum += resource_set[i].power();
			}
			if (check_power_sum != peer_power) {
				throw new GridException("createResources(): "+
										"corrupted Resource power count");
			}
		}
		/* debugging code -- 
		for (int i = 0 ; i < resource_set.length ; i++)
			System.out.print(resource_set[i].power()+" "); */
		
		return resource_set;

	}

	// ****************************************************************

	private synchronized boolean isValidPeerCount() {
		
		int peer_count = 0;
		for (int i = 0 ; i < this.sc.getGridConfiguration().PEER_GROUPS();i++) {
			peer_count += this.sc.getGridConfiguration().PEER_GROUP_SIZE()[i];
		}
		
		return peer_count == this.sc.getGridConfiguration().PEER_COUNT();
		
	}
	
	private synchronized long generatePeers() throws GridException {

		int peer_id = 0;

		ControllerFactory ctrlr_factory =
			new ControllerFactory(this.env_ctrlr,
								  this.sc.getGridConfiguration().
								  PEER_BASE_NAME());
		PeerServiceFactory peer_service_factory =
			new PeerServiceFactory(this.sc.getGridConfiguration().
									CLONE_SIMULATED_TRANSFERRED_OBJECTS());
		SearchEngineHandle se_handle =
			SearchEngineServiceFactory.
				createSearchEngineService(this.search_engine_server).
					createHandle();

		// Enumerate Peers sets
		long max_timer = 0;
		for (int i = 0 ; i < this.sc.getGridConfiguration().PEER_GROUPS();i++) {
			GridConfiguration grid_config = this.sc.getGridConfiguration();
			int N              = grid_config.PEER_GROUP_SIZE()[i];
			int peer_power     = grid_config.PEER_POWER()[i];
			int res_count      = grid_config.RES_COUNT()[i];
			int lo_power       = grid_config.RES_POWER_LO()[i];
			int hi_power       = grid_config.RES_POWER_HI()[i];
			int cache_capacity = grid_config.RES_CACHE_CAPACITY()[i];
			int mtbf           = grid_config.RES_MTBF()[i];

			GridStrategyConfiguration grid_strategy_config =
				this.sc.getGridStrategyConfiguration();
			PeerStrategyConfiguration peer_strategy_config =
				new PeerStrategyConfiguration(grid_strategy_config,i);
			PeerNetworkConfiguration peer_network_config =
				new PeerNetworkConfiguration();
			PeerDataConfiguration peer_data_config =
				new PeerDataConfiguration(grid_config,i);

			// Generate Peers for target Peers set
			for (int p = 0 ; p < N ; p++,peer_id++) {
				// Create Peer
				String id =
					generatePeerID(peer_id,
								   this.sc.getGridConfiguration().
								   	PEER_BASE_NAME());
				PeerBasicConfiguration peer_basic_config =
					new PeerBasicConfiguration(id,
											   this.sc.getGridConfiguration().
											   		SEARCH_TIMEOUT);
				Peer peer =
					new SimulatedPeer(peer_basic_config,peer_network_config,
							 		  peer_data_config,peer_strategy_config,
									  peer_service_factory,ctrlr_factory,
									  se_handle,
									  this.env_ctrlr,peer_id);
							 
				this.peers_handles[peer_id] = peer.service().createHandle();
				this.peers[peer_id] = peer;
				InternalPeerHandle internal_peer_handle =
					(InternalPeerHandle)this.peers_handles[peer_id];

				if (peer_power > 0) {
					// Create Resources
					SimulatedResource[] res =
						createPeerResources(internal_peer_handle,
											peer_power,res_count,
											lo_power,hi_power,
											cache_capacity,mtbf,this.el);
					//  Register Resources
					addResourcesToPeer(res,internal_peer_handle);
				}

				// Register Peer
				peer.searchEngine().
					setOwnerPeer((ExternalPeerHandle)this.peers_handles[peer_id]);

				// Compute Grid-wide maximum timer
				max_timer =
					max(max_timer,
						peer_strategy_config.REQUESTORS_TIME_THRESHOLD());
				max_timer =
					max(max_timer,
						peer_strategy_config.GRANTORS_TIME_THRESHOLD());
				max_timer =
					max(max_timer,
						peer_strategy_config.CONSUMPTION_TIMEOUT());
				max_timer =
					max(max_timer,
						this.sc.getGridConfiguration().SEARCH_TIMEOUT);
			}

		}
		
		// Force the Peers to refresh their ExternalPeerHandle cache
		int peer_count = this.sc.getGridConfiguration().PEER_COUNT();
		for (int p = 0 ; p < peer_count ; p++) {
			this.peers[p].searchEngine().refresh();
		}
		
		// Add one Negotiation Timeout Event to EventList
		Event evt = this.el.factory().createSelfEvent();
		this.el.addEvent(evt);
		
		if (peer_id != peer_count) {
			throw new GridException("generatePeers(): corrupted Peers count");
		}
		
		return max_timer+1;

	}
	
	private synchronized void generateUserAgents() throws GridException {
		
		int peer_count = this.sc.getGridConfiguration().PEER_COUNT();
		
		for (int i = 0 ; i < peer_count ; i++) {
			// Create User Agent
			UserPeerHandle user_peer_handle =
				(UserPeerHandle)this.peers_handles[i];
			String uid =
				generateUserID(user_peer_handle,
							   this.sc.getGridConfiguration().PEER_BASE_NAME());
			this.users[i] =
				new SimulatedUserAgent(uid,user_peer_handle,this);
			
			// Add one Job submission Event to EventList for target UserAgent
			String peer_id = this.peers_handles[i].id();
			UserAgent user = this.users[i];
			long time_shift =
				this.sc.getGridConfiguration().JOB_INIT_SHIFT(peer_id);
			if (user.submittedJobs() <
				this.sc.getGridConfiguration().JOBS_PER_PEER(peer_id)) {
				Event evt =
					this.el.factory().createJobEvent(user,peer_id,time_shift);
				this.el.addEvent(evt);
			}
			
		}
		
	}
	
	private synchronized String dumpExtraSelfEvents() {
		
		StringBuilder sb = new StringBuilder();

		sb.append("\n\n  ->  ");
		sb.append("extra self-events allowed after Event list ");
		sb.append("has become allowed=");
		sb.append(this.extra_self_events);
		sb.append(" (i.e. maximum of all time-out constants)");
		
		return sb.toString();
		
	}

	// ****************************************************************

	private static void addResourcesToPeer(SimulatedResource[] res,
										   InternalPeerHandle peer)
		throws GridException {

		if (peer == null) {
			throw new GridException("addReourcesToPeer(): illegal Peer");
		}
		
		for (int i = 0 ; i < res.length ; i++) {
			res[i].connect();
		}
		
	}

	private static String generatePeerID(int id,String PEER_BASE_NAME)
		throws GridException {
		
		StringBuilder sb = new StringBuilder(PEER_BASE_NAME);
		sb.append(CommonStringHandling.fillWithHeadingZeros(id,2));
		
		return sb.toString();
		
	}
	
	private static String generateResourceID(InternalPeerHandle owner_peer,
											 int id)
		throws GridException {
		
		if (owner_peer == null) {
			throw new GridException("generateResourceID(): illegal Peer");
		}
		
		StringBuilder sb = new StringBuilder(owner_peer.id());
		sb.append("[");
		sb.append(CommonStringHandling.fillWithHeadingZeros(id,2));
		sb.append("]");

		return sb.toString();
		
	}
	
	private static String generateUserID(UserPeerHandle used_peer,
										 String PEER_BASE_NAME)
		throws GridException {
		
		if (used_peer == null) {
			throw new GridException("generateResourceID(): illegal Peer");
		}

		int iid =
			Integer.parseInt(used_peer.id().
							 substring(PEER_BASE_NAME.length()));
		
		StringBuilder sb = new StringBuilder("grid_user_");
		sb.append(CommonStringHandling.fillWithHeadingZeros(iid,2));

		return sb.toString();
		
	}
	
	// ****************************************************************

	private final Scenario sc;
	private final EventList el;
	private final long extra_self_events;
	private final SearchEngineServer search_engine_server;
	private final EnvironmentController env_ctrlr;
	private final PeerHandle[] peers_handles;
	private final Peer[] peers;
	private final UserAgent[] users;

}

