/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// GridConfiguration.java

package lbg.simulator.config;

import lbg.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.*;

import java.util.Properties;

/**
 * @author Cyril Briquet
 */
public class GridConfiguration implements Cloneable {

	public GridConfiguration(Properties props) throws GridException {
		
		if (props == null) {
			throw new GridException("GridConfiguration(): "+
									"illegal Properties");
		}

		this.CLONE_SIMULATED_TRANSFERRED_OBJECTS =
			loadBoolean(
				loadProperty("CLONE_SIMULATED_TRANSFERRED_OBJECTS",props));
		this.PEER_BASE_NAME  =
			loadProperty("PEER_BASE_NAME",props);
		this.PEER_GROUP_SIZE   =
			loadIntArray(loadProperty("PEER_GROUP_SIZE",props));
		this.PEER_GROUPS =  this.PEER_GROUP_SIZE.length;

		this.PEER_POWER =
			loadIntArray(loadProperty("PEER_POWER",props));
		this.PEER_TTG_POLICY =
			loadBooleanArray(loadProperty("PEER_TTG_POLICY",props));
		this.PEER_STORAGE_AFFINITY =
			loadBooleanArray(loadProperty("PEER_STORAGE_AFFINITY",props));
		this.PEER_DATA_REPLICATION =
			loadBooleanArray(loadProperty("PEER_DATA_REPLICATION",props));
		this.PEER_IDLE_REPLICATION_RATIO =
			loadFloatArray(loadProperty("PEER_IDLE_REPLICATION_RATIO",props));
		this.RES_MTBF =
			loadIntArray(loadProperty("RES_MTBF",props));
		this.RES_COUNT =
			loadIntArray(loadProperty("RES_COUNT",props));
		this.RES_POWER_LO =
			loadIntArray(loadProperty("RES_POWER_LO",props));
		this.RES_POWER_HI =
			loadIntArray(loadProperty("RES_POWER_HI",props));
		this.RES_CACHE_CAPACITY =
			loadIntArray(loadProperty("RES_CACHE_CAPACITY",props));
		this.SEARCH_TIMEOUT =
			loadInt(loadProperty("SEARCH_TIMEOUT",props));
		
		if ((this.PEER_BASE_NAME == null) ||
			(this.PEER_BASE_NAME.length() < 2)) {
			throw new GridException("GridConfiguration(): "+
									"illegal Peer base name");
		}

		// no check on CLONE_SIMULATED_TRANSFERRED_OBJECTS
		
		if ((this.PEER_GROUP_SIZE == null) ||
			(this.PEER_GROUP_SIZE.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal number of Peer groups");
		}
		for (int i = 0 ; i < this.PEER_GROUP_SIZE.length ; i++) {
			if (this.PEER_GROUP_SIZE[i] < 1) {
				throw new GridException("GridConfiguration(): "+
										"illegal Peer group "+i+" size");
			}
		}
		
		if ((this.PEER_POWER == null) || (this.PEER_POWER.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal number of Peer groups power");
		}
		for (int i = 0 ; i < this.PEER_POWER.length ; i++) {
			if ((this.PEER_POWER[i] < 0) ||
				((this.PEER_POWER[i] == 0) && (this.RES_COUNT[i] < 0))) {
				throw new GridException("GridConfiguration(): "+
										"illegal Peer group "+i+" power");
			}
		}
		
		if ((this.PEER_TTG_POLICY == null) ||
			(this.PEER_TTG_POLICY.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal number of Peer groups TTG flags");
		}

		if ((this.PEER_STORAGE_AFFINITY == null) ||
			(this.PEER_STORAGE_AFFINITY.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal number of Peer groups "+
									"Storage Affinity flags");
		}

		if ((this.PEER_DATA_REPLICATION == null) ||
			(this.PEER_DATA_REPLICATION.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal number of Peer groups "+
									"data replication flags");
		}

		if ((this.PEER_IDLE_REPLICATION_RATIO == null) ||
			(this.PEER_IDLE_REPLICATION_RATIO.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal number of Peer groups power");
		}
		for (int i = 0 ; i < this.PEER_IDLE_REPLICATION_RATIO.length ; i++) {
			if ((this.PEER_IDLE_REPLICATION_RATIO[i] < 0.0) ||
				(this.PEER_IDLE_REPLICATION_RATIO[i] > 1.0)) {
				throw new GridException("GridConfiguration(): "+
										"illegal Peer group "+i+
										" idle replication ratio");
			}
		}
		
		if ((this.RES_MTBF == null) ||
			(this.RES_MTBF.length < 0)) {
			throw new GridException("GridConfiguration(): illegal Resource "+
									"mean times between failures");
		}
		for (int i = 0 ; i < this.RES_MTBF.length ; i++) {
			if (this.RES_MTBF[i] < 0) {
				throw new GridException("GridConfiguration(): illegal "+
										"Peer group "+i+" Resource mean time "+
										"between failures");
			}
		}
		
		if ((this.RES_COUNT == null) ||
			(this.RES_COUNT.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal Resource counts");
		}
		VirtualizationLevel mode = LightweightBarteringGrid.getMode();
		int lim = (mode == VirtualizationLevel.SIM_MODE) ? 0 : 1;
		for (int i = 0 ; i < this.RES_COUNT.length ; i++) {
			if (this.RES_COUNT[i] < lim) {
				throw new GridException("GridConfiguration(): "+
										"illegal Peer group "+i+
										" Resource count");
			}
		}

		if ((this.RES_POWER_LO == null) ||
			(this.RES_POWER_LO.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal number of Peer groups "+
									"lowest power");
		}
		for (int i = 0 ; i < this.RES_POWER_LO.length ; i++) {
			if ((this.RES_POWER_LO[i] < 0) && 
				((this.RES_POWER_LO[i] == 0) && (this.PEER_POWER[i] != 0))) {
				throw new GridException("GridConfiguration(): "+
										"illegal Peer group "+i+
										" lowest power");
			}
		}

		if ((this.RES_POWER_HI == null) ||
			(this.RES_POWER_HI.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal number of Peer groups "+
									"highest power");
		}
		for (int i = 0 ; i < this.RES_POWER_HI.length ; i++) {
			if ((this.RES_POWER_HI[i] < 0) ||
				((this.RES_POWER_HI[i] == 0) && (this.PEER_POWER[i] != 0)) ||
				(this.RES_POWER_HI[i] < this.RES_POWER_LO[i])) {
				throw new GridException("GridConfiguration(): "+
										"illegal Peer group "+i+
										" highest power");
			}
		}
		
		if ((this.RES_CACHE_CAPACITY == null) ||
			(this.RES_CACHE_CAPACITY.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal number of Peer groups "+
									"cache capacity");
		}
		for (int i = 0 ; i < this.RES_CACHE_CAPACITY.length ; i++) {
			if (this.RES_CACHE_CAPACITY[i] < 1) {
				throw new GridException("GridConfiguration(): "+
										"illegal Peer group "+i+
										" cache capacity");
			}
		}

		// ****************
		
		this.JOBS_PER_PEER =
				loadIntArray(loadProperty("JOBS_PER_PEER",props));
		this.JOB_INIT_SHIFT =
			loadIntArray(loadProperty("JOB_INIT_SHIFT",props));
		this.JOBS_INTER_TIME_LO =
			loadIntArray(loadProperty("JOB_INTER_TIME_LO",props));
		this.JOBS_INTER_TIME_HI =
			loadIntArray(loadProperty("JOB_INTER_TIME_HI",props));

		if ((this.JOBS_PER_PEER == null) ||
			(this.JOBS_PER_PEER.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal Jobs per Peer count");
		}

		if ((this.JOB_INIT_SHIFT == null) ||
			(this.JOB_INIT_SHIFT.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal initial Job shift");
		}

		if ((this.JOBS_INTER_TIME_LO == null) ||
			(this.JOBS_INTER_TIME_LO.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal Jobs interarrival time "+
									"low boundary count");
		}

		if ((this.JOBS_INTER_TIME_HI == null) ||
			(this.JOBS_INTER_TIME_HI.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal Jobs interarrival time "+
									"high boundary count");
		}

		for (int i = 0 ; i < this.JOBS_PER_PEER.length ; i++) {
			if (this.JOBS_PER_PEER[i] < 0) {
				throw new GridException("GridConfiguration(): "+
										"illegal Jobs per Peer ("+
										this.JOBS_PER_PEER[i]+")");
			}
		}
		
		for (int i = 0 ; i < this.JOB_INIT_SHIFT.length ; i++) {
			if (this.JOB_INIT_SHIFT[i] < 0) {
				throw new GridException("GridConfiguration(): "+
										"illegal initial Job shift ("+
										this.JOB_INIT_SHIFT[i]+")");
			}
		}

		if (this.JOBS_INTER_TIME_LO.length != this.JOBS_INTER_TIME_HI.length) {
			throw new GridException("GridConfiguration(): "+
									"Jobs interarrival times arrays mismatch");
		}
		for (int i = 0 ; i < this.JOBS_INTER_TIME_LO.length ; i++) {
			if ((this.JOBS_INTER_TIME_LO[i] < 0) ||
				((this.JOBS_INTER_TIME_LO[i]==0)&&(this.JOBS_PER_PEER[i]!=0))) {
				throw new GridException("GridConfiguration(): "+
										"illegal Jobs interarrival time "+
										"low boundary ("+
										this.JOBS_INTER_TIME_LO[i]+")");
			}
			if ((this.JOBS_INTER_TIME_HI[i] < 0) ||
				((this.JOBS_INTER_TIME_HI[i]==0)&&(this.JOBS_PER_PEER[i]!=0))) {
				throw new GridException("GridConfiguration(): "+
										"illegal Jobs interarrival time "+
										"high boundary ("+
										this.JOBS_INTER_TIME_HI[i]+")");
			}
			if (this.JOBS_INTER_TIME_LO[i] > this.JOBS_INTER_TIME_HI[i]) {
				throw new GridException("GridConfiguration(): illegal "+
										"Jobs interarrival times interval ["+
										this.JOBS_INTER_TIME_LO[i]+".."+
										this.JOBS_INTER_TIME_HI[i]+"]");
			}
		}

		// ****************

		this.TASKS_PER_JOB =
			loadIntArray(loadProperty("TASKS_PER_JOB",props));
		this.DATA_PER_TASK =
			loadIntArray(loadProperty("DATA_PER_TASK",props));
		this.TASK_LEN_LO =
			loadIntArray(loadProperty("TASK_LEN_LO",props));
		this.TASK_LEN_HI =
			loadIntArray(loadProperty("TASK_LEN_HI",props));

		if ((this.TASKS_PER_JOB == null) ||
			(this.TASKS_PER_JOB.length <= 0)) {
			throw new GridException("GridConfiguration(): "+
									"illegal Tasks per Job count");
		}
		
		if ((this.DATA_PER_TASK == null) ||
			(this.DATA_PER_TASK.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal data per Task count");
		}

		if ((this.TASK_LEN_LO == null) ||
			(this.TASK_LEN_LO.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal shortest Task length");
		}

		if ((this.TASK_LEN_HI == null) ||
			(this.TASK_LEN_HI.length < 1)) {
			throw new GridException("GridConfiguration(): "+
									"illegal longest Task length");
		}

		for (int i = 0 ; i < this.TASKS_PER_JOB.length ; i++) {
			if ((this.TASKS_PER_JOB[i] < 0) ||
				((this.TASKS_PER_JOB[i] == 0) && (this.JOBS_PER_PEER[i]!=0))) {
				throw new GridException("GridConfiguration(): "+
										"illegal Tasks per Job ("+
										this.TASKS_PER_JOB[i]+
										")");
			}
		}

		for (int i = 0 ; i < this.DATA_PER_TASK.length ; i++) {
			if (this.DATA_PER_TASK[i] < 0) {
				throw new GridException("GridConfiguration(): "+
										"illegal data per Tasks ("+
										this.DATA_PER_TASK[i]+
										")");
			}
		}
		
		if (this.TASK_LEN_LO.length != this.TASK_LEN_HI.length) {
			throw new GridException("GridConfiguration(): "+
									"Tasks length arrays mismatch");
		}
		for (int i = 0 ; i < this.TASK_LEN_LO.length ; i++) {
			if ((this.TASK_LEN_LO[i] < 0) ||
				((this.TASK_LEN_LO[i] == 0) && (this.JOBS_PER_PEER[i] != 0))) {
				throw new GridException("GridConfiguration(): "+
										"illegal shortest Task length ("+
										this.TASK_LEN_LO[i]+")");
			}
			if ((this.TASK_LEN_HI[i] < 0) ||
				((this.TASK_LEN_HI[i] == 0) && (this.JOBS_PER_PEER[i] != 0))) {
				throw new GridException("GridConfiguration(): "+
										"illegal longest Task length ("+
										this.TASK_LEN_HI[i]+")");
			}
			if (this.TASK_LEN_LO[i] > this.TASK_LEN_HI[i]) {
				throw new GridException("GridConfiguration(): illegal "+
										"Task lengths interval ["+
										this.TASK_LEN_LO[i]+".."+
										this.TASK_LEN_HI[i]+"]");
			}
		}

		// ****************

		if ((this.PEER_GROUPS != this.PEER_GROUP_SIZE.length) ||
			(this.PEER_GROUPS != this.PEER_POWER.length) ||
			(this.PEER_GROUPS != this.PEER_TTG_POLICY.length) ||
			(this.PEER_GROUPS != this.PEER_STORAGE_AFFINITY.length) ||
			(this.PEER_GROUPS != this.PEER_DATA_REPLICATION.length) ||
			(this.PEER_GROUPS != this.PEER_IDLE_REPLICATION_RATIO.length)) {
			throw new GridException("GridConfiguration(): "+
									"Peers sets mismatch (Peers config)");
		}
		
		if ((this.PEER_GROUPS != this.RES_MTBF.length) ||
			(this.PEER_GROUPS != this.RES_COUNT.length) ||
			(this.PEER_GROUPS != this.RES_POWER_LO.length) ||
			(this.PEER_GROUPS != this.RES_POWER_HI.length) ||
			(this.PEER_GROUPS != this.RES_CACHE_CAPACITY.length)) {
			throw new GridException("GridConfiguration(): "+
									"Peers sets mismatch (Resources config)");
		}
			
		if ((this.PEER_GROUPS != this.JOBS_PER_PEER.length) ||
			(this.PEER_GROUPS != this.JOB_INIT_SHIFT.length) ||
			(this.PEER_GROUPS != this.JOBS_INTER_TIME_LO.length) ||
			(this.PEER_GROUPS != this.JOBS_INTER_TIME_HI.length) ||
			(this.PEER_GROUPS != this.TASKS_PER_JOB.length) ||
			(this.PEER_GROUPS != this.DATA_PER_TASK.length) ||
			(this.PEER_GROUPS != this.TASK_LEN_LO.length) ||
			(this.PEER_GROUPS != this.TASK_LEN_HI.length)) {
			throw new GridException("GridConfiguration(): "+
									"Peers sets mismatch (workload)");
		}

		// ****************
		
		int dummy = 0;
		for (int g = 0 ; g < this.PEER_GROUPS ; g++) {
			dummy += this.PEER_GROUP_SIZE[g];
		}
		this.PEER_COUNT = dummy;
			
		this.peer_to_group_mapping = new int[this.PEER_COUNT];
		int y = 0;
		for (int g = 0 ; g < this.PEER_GROUPS ; g++) {
			int s = this.PEER_GROUP_SIZE[g];
			for (int p = 0 ; p < s ; p++) {
				this.peer_to_group_mapping[y++] = g;
			}
		}
		if (y != this.PEER_COUNT) {
			throw new GridException("GridConfiguration(): could not create "+
									"Peer to Group mapping");
		}
		
		if (this.SEARCH_TIMEOUT < 1) {
			throw new GridException("GridConfiguration(): "+
									"illegal search timeout");
		}

	}

	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[GridConfiguration]");
		sb.append(format(true));
		
		return sb.toString();
		
	}
	
	/** Outputs the GridConfiguration,
	 * formatted either as a property file or as a log file.
	 * @param log_format true to format GridConfiguration as a log file
	 * @return formatted GridConfiguration
	 */
	public String format(boolean log_format) {
		
		StringBuilder sb = new StringBuilder();

		if (log_format) {
			sb.append("\n");
			sb.append("  ->  ");
			sb.append(this.PEER_COUNT);
			sb.append(" Peer");
			if (this.PEER_COUNT > 1) sb.append("s");
			sb.append(", split into ");
			sb.append(this.PEER_GROUPS);
			sb.append(" group");
			if (this.PEER_GROUPS > 1) sb.append("s");
		}

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("CLONE_SIMULATED_TRANSFERRED_OBJECTS  = ");
		sb.append(this.CLONE_SIMULATED_TRANSFERRED_OBJECTS);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("SEARCH_TIMEOUT      = ");
		sb.append(this.SEARCH_TIMEOUT);
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_BASE_NAME      = ");
		sb.append(this.PEER_BASE_NAME);

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_GROUP_SIZE     = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.PEER_GROUP_SIZE[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_POWER          = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.PEER_POWER[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_TTG_POLICY             = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.PEER_TTG_POLICY[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_STORAGE_AFFINITY       = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.PEER_STORAGE_AFFINITY[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_DATA_REPLICATION       = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.PEER_DATA_REPLICATION[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_IDLE_REPLICATION_RATIO = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.PEER_IDLE_REPLICATION_RATIO[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("RES_MTBF            = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.RES_MTBF[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("RES_COUNT           = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.RES_COUNT[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("RES_POWER_LO        = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.RES_POWER_LO[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("RES_POWER_HI        = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.RES_POWER_HI[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("RES_CACHE_CAPACITY  = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.RES_CACHE_CAPACITY[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		
		sb.append("JOBS_PER_PEER       = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.JOBS_PER_PEER[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");

		sb.append("JOB_INIT_SHIFT      = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.JOB_INIT_SHIFT[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		
		sb.append("JOB_INTER_TIME_LO   = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.JOBS_INTER_TIME_LO[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		
		sb.append("JOB_INTER_TIME_HI   = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.JOBS_INTER_TIME_HI[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("TASKS_PER_JOB       = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.TASKS_PER_JOB[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("DATA_PER_TASK       = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.DATA_PER_TASK[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("TASK_LEN_LO         = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.TASK_LEN_LO[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("TASK_LEN_HI         = { ");
		for (int i = 0 ; i < this.PEER_GROUPS ; i++) {
			sb.append(this.TASK_LEN_HI[i]);
			if (i < this.PEER_GROUPS-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");
		
		return sb.toString();
		
	}

	// ****************************************************************

	public int resolvePeerID(String peer_id) throws GridException {
	
		if (peer_id == null) {
			throw new GridException("resolvePeerID(): illegal id");
		}
		
		int len = this.PEER_BASE_NAME.length();
		
		String s_dummy = peer_id.substring(len);
		
		if ((s_dummy == null) || (s_dummy.length() < 1)) {
			throw new GridException("resolvePeerID(): corrupted id");
		}
		
		int i_dummy = Integer.parseInt(s_dummy);
		
		if ((i_dummy < 0) || (i_dummy >= this.PEER_COUNT)) {
			throw new GridException("resolvePeerID(): out of range id");
		}
		
		return i_dummy;
		
	}

	// ****************************************************************

	public int size() {
		
		return this.PEER_GROUPS;
		
	}
	
	public boolean CLONE_SIMULATED_TRANSFERRED_OBJECTS() {
		
		return this.CLONE_SIMULATED_TRANSFERRED_OBJECTS;
		
	}

	public String PEER_BASE_NAME() {
		
		return this.PEER_BASE_NAME;
		
	}

	public int PEER_GROUPS() {
		
		return this.PEER_GROUPS;
		
	}
	
	public int[] PEER_GROUP_SIZE() {
		
		return this.PEER_GROUP_SIZE.clone();
		
	}

	public int PEER_COUNT() {
		
		return this.PEER_COUNT;
		
	}
	
	public int[] PEER_POWER() {
		
		return this.PEER_POWER.clone();
		
	}
	
	public boolean[] PEER_TTG_POLICY() {
		
		return this.PEER_TTG_POLICY.clone();
		
	}

	public boolean[] PEER_STORAGE_AFFINITY() {
		
		return this.PEER_STORAGE_AFFINITY.clone();
		
	}

	public boolean[] PEER_DATA_REPLICATION() {
		
		return this.PEER_DATA_REPLICATION.clone();
		
	}

	public float[] PEER_IDLE_REPLICATION_RATIO() {
		
		return this.PEER_IDLE_REPLICATION_RATIO.clone();
		
	}
	
	public int[] RES_MTBF() {
		
		return this.RES_MTBF.clone();
		
	}

	public int[] RES_COUNT() {
		
		return this.RES_COUNT.clone();
		
	}

	public int[] RES_POWER_LO() {
		
		return this.RES_POWER_LO.clone();
		
	}

	public int[] RES_POWER_HI() {
		
		return this.RES_POWER_HI.clone();
		
	}

	public int[] RES_CACHE_CAPACITY() {
		
		return this.RES_CACHE_CAPACITY.clone();
		
	}

	public int[] JOBS_PER_PEER() {
		
		return this.JOBS_PER_PEER.clone();
		
	}

	public  int JOBS_PER_PEER(String peer_id) throws GridException {
		
		int id = this.resolvePeerID(peer_id);
		
		return this.JOBS_PER_PEER[this.peer_to_group_mapping[id]];
		
	}

	public int[] JOB_INIT_SHIFT() {
		
		return this.JOB_INIT_SHIFT.clone();
		
	}

	public  int JOB_INIT_SHIFT(String peer_id) throws GridException {
		
		int id = this.resolvePeerID(peer_id);
		
		return this.JOB_INIT_SHIFT[this.peer_to_group_mapping[id]];
		
	}

	public int[] JOBS_INTER_TIME_LO() {

		return this.JOBS_INTER_TIME_LO.clone();

	}
	
	public  int JOBS_INTER_TIME_LO(String peer_id) throws GridException {
		
		int id = this.resolvePeerID(peer_id);
		
		return this.JOBS_INTER_TIME_LO[this.peer_to_group_mapping[id]];
		
	}

	public int[] JOBS_INTER_TIME_HI() {
		
		return this.JOBS_INTER_TIME_HI.clone();

	}
	
	public int JOBS_INTER_TIME_HI(String peer_id) throws GridException {
		
		int id = this.resolvePeerID(peer_id);
		
		return this.JOBS_INTER_TIME_HI[this.peer_to_group_mapping[id]];
		
	}

	public int[] TASKS_PER_JOB() {

		return this.TASKS_PER_JOB.clone();
	}

	public int[] DATA_PER_TASK() {

		return this.DATA_PER_TASK.clone();
	}

	public int TASKS_PER_JOB(String peer_id) throws GridException {
		
		int id = this.resolvePeerID(peer_id);
		
		return this.TASKS_PER_JOB[this.peer_to_group_mapping[id]];
		
	}

	public int DATA_PER_TASK(String peer_id) throws GridException {
		
		int id = this.resolvePeerID(peer_id);
		
		return this.DATA_PER_TASK[this.peer_to_group_mapping[id]];
		
	}

	public int[] TASK_LEN_LO()  {
		
		return this.TASK_LEN_LO.clone();
		
	}

	public int TASK_LEN_LO(String peer_id) throws GridException {
		
		int id = this.resolvePeerID(peer_id);
		
		return this.TASK_LEN_LO[this.peer_to_group_mapping[id]];
		
	}

	public int[] TASK_LEN_HI() {

		return this.TASK_LEN_HI.clone();
		
	}
	
	public int TASK_LEN_HI(String peer_id) throws GridException {
		
		int id = this.resolvePeerID(peer_id);
		
		return this.TASK_LEN_HI[this.peer_to_group_mapping[id]];
		
	}
	
	// ****************************************************************

	@Override
	public synchronized Object clone() {

		GridConfiguration grid_config = null;
		
		try {
			grid_config = (GridConfiguration)super.clone();
		}
		catch (CloneNotSupportedException e) {
			throw new InternalError("clone(): cannot clone "+
									"GridConfiguration");
		}
		
		return grid_config;
		
	}

	// ****************************************************************

	private final String PEER_BASE_NAME;
	private final boolean CLONE_SIMULATED_TRANSFERRED_OBJECTS;
	private final int PEER_GROUPS;
	private final int[] PEER_GROUP_SIZE;
	private final int PEER_COUNT;
	private final int[] PEER_POWER;
	private final boolean[] PEER_TTG_POLICY;
	private final boolean[] PEER_STORAGE_AFFINITY;
	private final boolean[] PEER_DATA_REPLICATION;
	private final float[] PEER_IDLE_REPLICATION_RATIO;

	private final int[] RES_MTBF;
	private final int[] RES_COUNT;
	private final int[] RES_POWER_LO;
	private final int[] RES_POWER_HI;
	private final int[] RES_CACHE_CAPACITY;
	
	private final int[] JOBS_PER_PEER;
	private final int[] JOB_INIT_SHIFT;
	private final int[] JOBS_INTER_TIME_LO;
	private final int[] JOBS_INTER_TIME_HI;
	
	private final int[] TASKS_PER_JOB;
	private final int[] DATA_PER_TASK;
	private final int[] TASK_LEN_LO;
	private final int[] TASK_LEN_HI;

	private final int[] peer_to_group_mapping;
	
	public  final int SEARCH_TIMEOUT;

}

