/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// GridStrategyConfiguration.java

package lbg.simulator.config;

import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.peer.strategy.accounting.*;
import lbg.peer.strategy.negotiation.*;
import lbg.peer.strategy.scheduling.*;

import java.util.Properties;

/**
 * @author Cyril Briquet
 */
public class GridStrategyConfiguration implements Cloneable {

	public GridStrategyConfiguration(Properties props) throws GridException {
		
		if (props == null) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Properties");
		}
		
		this.PEER_DEFAULT_ACCOUNTANT =
			loadStringArray(
				loadProperty(BarteringAccountantType.getPDPLabel(),
							 props));
		this.PEER_NEGOTIATION_SUPPLYING_POLICY =
			loadStringArray(
				loadProperty(SupplyingNegotiationType.getPDPLabel(),
							 props));
		this.PEER_NEGOTIATION_CONSUMPTION_POLICY =
			loadStringArray(
				loadProperty(ConsumptionNegotiationType.getPDPLabel(),
							 props));

		this.PEER_SCHEDULING_LOCAL_POLICY =
			loadStringArray(
				loadProperty(LocalSchedulingType.getPDPLabel(),
							 props));
		this.PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY =
			loadStringArray(
				loadProperty(SupplyingRunningPreemptionType.getPDPLabel(),
							 props));
		this.PEER_PREEMPTION_SUPPLYING_WAITING_POLICY =
			loadStringArray(
				loadProperty(SupplyingWaitingPreemptionType.getPDPLabel(),
							 props));
		this.PEER_FILTERING_SUPPLYING_POLICY =
			loadStringArray(
				loadProperty(SupplyingFilteringType.getPDPLabel(),
							 props));
		this.PEER_SCHEDULING_SUPPLYING_POLICY =
			loadStringArray(
				loadProperty(SupplyingSchedulingType.getPDPLabel(),
							 props));
		this.PEER_SCHEDULING_CONSUMPTION_POLICY =
			loadStringArray(
				loadProperty(ConsumptionSchedulingType.getPDPLabel(),props));

		this.REQUESTORS_COUNT_THRESHOLD =
			loadIntArray(loadProperty("REQUESTORS_COUNT_THRESHOLD",props));
		this.REQUESTORS_TIME_THRESHOLD =
			loadIntArray(loadProperty("REQUESTORS_TIME_THRESHOLD",props));
		this.GRANTORS_COUNT_THRESHOLD =
			loadIntArray(loadProperty("GRANTORS_COUNT_THRESHOLD",props));
		this.GRANTORS_TIME_THRESHOLD =
			loadIntArray(loadProperty("GRANTORS_TIME_THRESHOLD",props));
		this.REQUEST_SUPPLYING_TIME_THRESHOLD =
			loadIntArray(
				loadProperty("REQUEST_SUPPLYING_TIME_THRESHOLD",props));

		this.LOCAL_Q_LEN_PREEMPTION_THRESHOLD =
			loadIntArray(loadProperty("LOCAL_Q_LEN_PREEMPTION_THRESHOLD",props));
		
		this.PEER_PSUFFERAGE =
			loadBooleanArray(loadProperty("PEER_PSUFFERAGE",props));

		this.PEER_Q_FILTERING_THRESHOLD =
			loadIntArray(loadProperty("PEER_Q_FILTERING_THRESHOLD",props));
		this.CONSUMPTION_TIMEOUT =
			loadLongArray(loadProperty("CONSUMPTION_TIMEOUT",props));

		this.DO_CONSUMPTION_BLACKLIST =
			loadBooleanArray(loadProperty("DO_CONSUMPTION_BLACKLIST",props));
		this.CONSUMPTION_BLACKLIST_PROBA =
			loadFloatArray(loadProperty("CONSUMPTION_BLACKLIST_PROBA",props));

		// ****************

		if ((this.PEER_DEFAULT_ACCOUNTANT == null) ||
			(this.PEER_DEFAULT_ACCOUNTANT.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer default Accountant");
		}

		if ((this.PEER_NEGOTIATION_SUPPLYING_POLICY
			 == null) ||
			(this.PEER_NEGOTIATION_SUPPLYING_POLICY.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer target "+
									"Negotiation Supplying policy");
		}
		if ((this.PEER_NEGOTIATION_CONSUMPTION_POLICY
			 == null) ||
			(this.PEER_NEGOTIATION_CONSUMPTION_POLICY.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer target "+
									"Negotiation Consumption policy");
		}

		if ((this.PEER_SCHEDULING_LOCAL_POLICY
			 == null) ||
			(this.PEER_SCHEDULING_LOCAL_POLICY.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer target "+
									"Scheduling Local policy");
		}
		if ((this.PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY
			 == null) ||
			(this.PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer target "+
									"Preemption Supplying running policy");
		}
		if ((this.PEER_PREEMPTION_SUPPLYING_WAITING_POLICY
			 == null) ||
			(this.PEER_PREEMPTION_SUPPLYING_WAITING_POLICY.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer target "+
									"Preemption Supplying waiting policy");
		}
		if ((this.PEER_FILTERING_SUPPLYING_POLICY
			 == null) ||
			(this.PEER_FILTERING_SUPPLYING_POLICY.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer target "+
									"filtering Supplying policy");
		}
		if ((this.PEER_SCHEDULING_SUPPLYING_POLICY
			 == null) ||
			(this.PEER_SCHEDULING_SUPPLYING_POLICY.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer target "+
									"Scheduling Supplying policy");
		}
		if ((this.PEER_SCHEDULING_CONSUMPTION_POLICY ==
			 null) ||
			(this.PEER_SCHEDULING_CONSUMPTION_POLICY.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer target "+
									"Scheduling Consumption policy");
		}

		if ((this.REQUESTORS_COUNT_THRESHOLD == null) ||
			(this.REQUESTORS_COUNT_THRESHOLD.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer requestors count threshold");
		}
		if ((this.REQUESTORS_TIME_THRESHOLD == null) ||
			(this.REQUESTORS_TIME_THRESHOLD.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer requestors time threshold");
		}
		if ((this.GRANTORS_COUNT_THRESHOLD == null) ||
			(this.GRANTORS_COUNT_THRESHOLD.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer grantors count threshold");
		}
		if ((this.GRANTORS_TIME_THRESHOLD == null) ||
			(this.GRANTORS_TIME_THRESHOLD.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer grantors time threshold");
		}
		if ((this.REQUEST_SUPPLYING_TIME_THRESHOLD == null) ||
			(this.REQUEST_SUPPLYING_TIME_THRESHOLD.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer Request Supplying "+
									" time threshold");
		}

		if ((this.LOCAL_Q_LEN_PREEMPTION_THRESHOLD == null) ||
			(this.LOCAL_Q_LEN_PREEMPTION_THRESHOLD.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer local queue length "+
									"preemption threshold");
		}
		
		if (this.PEER_PSUFFERAGE == null) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal Peer preemption sufferage flag");
		}

		if ((this.PEER_Q_FILTERING_THRESHOLD == null) ||
			(this.PEER_Q_FILTERING_THRESHOLD.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal number of Peer groups power");
		}
		for (int i = 0 ; i < this.PEER_Q_FILTERING_THRESHOLD.length ; i++) {
			if (this.PEER_Q_FILTERING_THRESHOLD[i] <= 0) {
				throw new GridException("GridStrategyConfiguration(): "+
										"illegal Peer group "+i+
										" queue threshold");
			}
		}

		if ((this.CONSUMPTION_TIMEOUT == null) ||
			(this.CONSUMPTION_TIMEOUT.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal number of Peer groups "+
									"Consumption timeout");
		}
		for (int i = 0 ; i < this.CONSUMPTION_TIMEOUT.length ; i++) {
			if (this.CONSUMPTION_TIMEOUT[i] < 1) {
				throw new GridException("GridStrategyConfiguration(): "+
										"illegal Peer group "+i+
										" Consumption timeout");
			}
		}

		if ((this.DO_CONSUMPTION_BLACKLIST == null) ||
			(this.DO_CONSUMPTION_BLACKLIST.length < 1)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"illegal consumption blacklist flag");
		}

		if ((this.CONSUMPTION_BLACKLIST_PROBA == null) ||
			(this.CONSUMPTION_BLACKLIST_PROBA.length < 1)) {
			throw new GridException("GridStrategyConfiguration():  "+
									"illegal number of Peer groups "+
									"consumption blacklist probability");
		}
		for (int i = 0 ; i < this.CONSUMPTION_BLACKLIST_PROBA.length ; i++) {
			if ((this.CONSUMPTION_BLACKLIST_PROBA[i] < 0.0f) ||
				(this.CONSUMPTION_BLACKLIST_PROBA[i] > 1.0f)) {
				throw new GridException("GridStrategyConfiguration(): illegal "+
										"consumption blacklist probability");
			}
		}

		// ****************

		this.size = this.PEER_DEFAULT_ACCOUNTANT.length;

		if ((this.size !=
			 this.PEER_DEFAULT_ACCOUNTANT.length) ||
			(this.size !=
			 this.PEER_NEGOTIATION_SUPPLYING_POLICY.length) ||
			(this.size !=
			 this.PEER_NEGOTIATION_CONSUMPTION_POLICY.length)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"Peers sets mismatch "+
									"(negotiation policies)");
		}
			
		if ((this.size !=
			 this.PEER_SCHEDULING_LOCAL_POLICY.length) ||
			(this.size !=
			 this.PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY.length) ||
			(this.size !=
			 this.PEER_PREEMPTION_SUPPLYING_WAITING_POLICY.length) ||
			(this.size !=
			 this.PEER_SCHEDULING_SUPPLYING_POLICY.length) ||
			(this.size !=
			 this.PEER_FILTERING_SUPPLYING_POLICY.length) ||
			(this.size !=
			 this.PEER_SCHEDULING_CONSUMPTION_POLICY.length)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"Peers sets mismatch "+
									"(scheduling policies)");
		}

		if ((this.size !=
			 this.REQUESTORS_COUNT_THRESHOLD.length) ||
			(this.size !=
			 this.REQUESTORS_TIME_THRESHOLD.length) ||
			(this.size !=
			 this.GRANTORS_COUNT_THRESHOLD.length) ||
			(this.size !=
			 this.GRANTORS_TIME_THRESHOLD.length) ||
			(this.size !=
			this.REQUEST_SUPPLYING_TIME_THRESHOLD.length)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"Peers sets mismatch "+
									"(negotiation policies threshold)");
		}
		
		if ((this.size !=
			 this.LOCAL_Q_LEN_PREEMPTION_THRESHOLD.length) ||
			(this.size !=
			 this.PEER_PSUFFERAGE.length) ||
			(this.size !=
			 this.PEER_Q_FILTERING_THRESHOLD.length) ||
			(this.size !=
			 this.CONSUMPTION_TIMEOUT.length)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"Peers sets mismatch "+
									"(scheduling policies thresholds)");
		}
		
		if ((this.size !=
			 this.DO_CONSUMPTION_BLACKLIST.length) ||
			(this.size !=
			 this.CONSUMPTION_BLACKLIST_PROBA.length)) {
			throw new GridException("GridStrategyConfiguration(): "+
									"Peers sets mismatch "+
									"(blacklisting thresholds)");
		}
		
	}
	
	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[GridStrategyConfiguration]");
		sb.append(format(true));
		
		return sb.toString();
		
	}
	
	/** Outputs the GridStrategyConfiguration,
	 * formatted either as a property file
	 * or as a log file.
	 * @param log_format true to format GridStrategyConfiguration as a log file
	 * @return formatted GridStrategyConfiguration
	 */
	public String format(boolean log_format) {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_SCHEDULING_LOCAL_POLICY              = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.PEER_SCHEDULING_LOCAL_POLICY[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY  = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_PREEMPTION_SUPPLYING_WAITING_POLICY  = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.PEER_PREEMPTION_SUPPLYING_WAITING_POLICY[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_SCHEDULING_SUPPLYING_POLICY          = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.PEER_SCHEDULING_SUPPLYING_POLICY[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_FILTERING_SUPPLYING_POLICY           = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.PEER_FILTERING_SUPPLYING_POLICY[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_SCHEDULING_CONSUMPTION_POLICY        = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.PEER_SCHEDULING_CONSUMPTION_POLICY[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_NEGOTIATION_SUPPLYING_POLICY         = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.PEER_NEGOTIATION_SUPPLYING_POLICY[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_NEGOTIATION_CONSUMPTION_POLICY       = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.PEER_NEGOTIATION_CONSUMPTION_POLICY[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_DEFAULT_ACCOUNTANT                   = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.PEER_DEFAULT_ACCOUNTANT[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("REQUESTORS_COUNT_THRESHOLD                = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.REQUESTORS_COUNT_THRESHOLD[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("REQUESTORS_TIME_THRESHOLD                 = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.REQUESTORS_TIME_THRESHOLD[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("GRANTORS_COUNT_THRESHOLD                  = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.GRANTORS_COUNT_THRESHOLD[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("GRANTORS_TIME_THRESHOLD                   = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.GRANTORS_TIME_THRESHOLD[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("REQUEST_SUPPLYING_TIME_THRESHOLD          = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.REQUEST_SUPPLYING_TIME_THRESHOLD[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("LOCAL_Q_LEN_PREEMPTION_THRESHOLD          = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.LOCAL_Q_LEN_PREEMPTION_THRESHOLD[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_PSUFFERAGE                           = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.PEER_PSUFFERAGE[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("PEER_Q_FILTERING_THRESHOLD                = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.PEER_Q_FILTERING_THRESHOLD[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("CONSUMPTION_TIMEOUT                       = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.CONSUMPTION_TIMEOUT[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");
		
		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("DO_CONSUMPTION_BLACKLIST                  = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.DO_CONSUMPTION_BLACKLIST[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		sb.append("\n");
		if (log_format) sb.append("  ->  ");
		sb.append("CONSUMPTION_BLACKLIST_PROBA               = { ");
		for (int i = 0 ; i < this.size ; i++) {
			sb.append(this.CONSUMPTION_BLACKLIST_PROBA[i]);
			if (i < this.size-1) sb.append(",");
			sb.append(" ");
		}
		sb.append("}");

		return sb.toString();

	}

	// ****************************************************************

	public int size() {
		
		return this.size;
		
	}
	
	public String[] PEER_DEFAULT_ACCOUNTANT() {
	
		return this.PEER_DEFAULT_ACCOUNTANT.clone();
		
	}

	public String[] PEER_NEGOTIATION_SUPPLYING_POLICY() {
	
		return this.PEER_NEGOTIATION_SUPPLYING_POLICY.clone();
		
	}

	public String[] PEER_NEGOTIATION_CONSUMPTION_POLICY() {
	
		return this.PEER_NEGOTIATION_CONSUMPTION_POLICY.clone();
		
	}

	public String[] PEER_SCHEDULING_LOCAL_POLICY() {
	
		return this.PEER_SCHEDULING_LOCAL_POLICY.clone();
		
	}
	
	public String[] PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY() {
	
		return this.PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY.clone();
		
	}

	public String[] PEER_PREEMPTION_SUPPLYING_WAITING_POLICY() {
	
		return this.PEER_PREEMPTION_SUPPLYING_WAITING_POLICY.clone();
		
	}

	public String[] PEER_FILTERING_SUPPLYING_POLICY() {
	
		return this.PEER_FILTERING_SUPPLYING_POLICY.clone();
		
	}

	public String[] PEER_SCHEDULING_SUPPLYING_POLICY() {
	
		return this.PEER_SCHEDULING_SUPPLYING_POLICY.clone();
		
	}

	public String[] PEER_SCHEDULING_CONSUMPTION_POLICY() {
	
		return this.PEER_SCHEDULING_CONSUMPTION_POLICY.clone();
		
	}
	
	public int[] REQUESTORS_COUNT_THRESHOLD() {
		
		return this.REQUESTORS_COUNT_THRESHOLD.clone();
		
	}
	
	public int[] REQUESTORS_TIME_THRESHOLD() {
		
		return this.REQUESTORS_TIME_THRESHOLD.clone();
		
	}
	
	public int[] GRANTORS_COUNT_THRESHOLD() {
		
		return this.GRANTORS_COUNT_THRESHOLD.clone();
		
	}
	
	public int[] GRANTORS_TIME_THRESHOLD() {
		
		return this.GRANTORS_TIME_THRESHOLD.clone();
		
	}
	
	public int[] REQUEST_SUPPLYING_TIME_THRESHOLD() {
		
		return this.REQUEST_SUPPLYING_TIME_THRESHOLD.clone();
		
	}
	
	public int[] LOCAL_Q_LEN_PREEMPTION_THRESHOLD() {
		
		return this.LOCAL_Q_LEN_PREEMPTION_THRESHOLD.clone();
		
	}
	
	public boolean[] PEER_PSUFFERAGE() {
		
		return this.PEER_PSUFFERAGE.clone();
		
	}
	
	public int[] PEER_Q_FILTERING_THRESHOLD() {
		
		return this.PEER_Q_FILTERING_THRESHOLD.clone();
		
	}
	
	public long[] CONSUMPTION_TIMEOUT() {
		
		return this.CONSUMPTION_TIMEOUT.clone();
		
	}

	public boolean[] DO_CONSUMPTION_BLACKLIST() {
		
		return this.DO_CONSUMPTION_BLACKLIST.clone();
		
	}

	public float[] CONSUMPTION_BLACKLIST_PROBA() {
		
		return this.CONSUMPTION_BLACKLIST_PROBA.clone();
		
	}
	
	// ****************************************************************

	@Override
	public synchronized Object clone() {

		GridStrategyConfiguration grid_strategy_config = null;
		
		try {
			grid_strategy_config = (GridStrategyConfiguration)super.clone();
		}
		catch (CloneNotSupportedException e) {
			throw new InternalError("clone(): cannot clone "+
									"GridStrategyConfiguration");
		}
		
		return grid_strategy_config;
		
	}

	// ****************************************************************

	private final String[]   PEER_DEFAULT_ACCOUNTANT;
	private final String[]   PEER_NEGOTIATION_SUPPLYING_POLICY;
	private final String[]   PEER_NEGOTIATION_CONSUMPTION_POLICY;
	private final String[]   PEER_SCHEDULING_LOCAL_POLICY;
	private final String[]   PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY;
	private final String[]   PEER_PREEMPTION_SUPPLYING_WAITING_POLICY;
	private final String[]   PEER_FILTERING_SUPPLYING_POLICY;
	private final String[]   PEER_SCHEDULING_SUPPLYING_POLICY;
	private final String[]   PEER_SCHEDULING_CONSUMPTION_POLICY;
	private final int[]      REQUESTORS_COUNT_THRESHOLD;
	private final int[]      REQUESTORS_TIME_THRESHOLD;
	private final int[]      GRANTORS_COUNT_THRESHOLD;
	private final int[]      GRANTORS_TIME_THRESHOLD;
	private final int[]      REQUEST_SUPPLYING_TIME_THRESHOLD;
	private final int[]      LOCAL_Q_LEN_PREEMPTION_THRESHOLD;
	private final boolean[]  PEER_PSUFFERAGE;
	private final int[]      PEER_Q_FILTERING_THRESHOLD;
	private final long[]     CONSUMPTION_TIMEOUT;
	private final boolean[]  DO_CONSUMPTION_BLACKLIST;
	private final float[]    CONSUMPTION_BLACKLIST_PROBA;
	private final int size;

}

