/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// EventFactory.java

package lbg.simulator.event;

import lbg.*;
import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.resource.*;
import lbg.simulator.*;
import lbg.user.*;
import lbg.task.*;

/** Event factory.
 * @author Cyril Briquet
 */
public class EventFactory {

	/** Creates a new EventFactory.
	 * @param sc target Scenario
	 * @exception GridException if a parameter is illegal
	 */
	public EventFactory(Scenario sc) throws GridException {
	
		if (sc == null) {
			throw new GridException("EventFactory(): illegal Scenario");
		}
		
		this.sc = sc;
		this.current_id = 0;
	
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		sb.append("[EventFactory]");
		sb.append("current Event id=");
		sb.append(this.current_id);

		return sb.toString();
		
	}

	// ****************************************************************

	/** Creates a SubmittedJobEvent so that target UserAgent submits,
	 * at a legal random time from now, a Job to target Peer.
	 * @param user target UserAgent
	 * @param owner_id target Peer id
	 * @return created SubmittedJobEvent
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized Event createJobEvent(UserAgent user,String owner_id)
		throws GridException {

		long now = LightweightBarteringGrid.getTime();
		
		return createJobEvent(user,owner_id,now);
		
	}

	/** Creates a SubmittedJobEvent so that target UserAgent submits,
	 * at a legal random time from target base time, a Job to target Peer.
	 * @param user target UserAgent
	 * @param owner_id target Peer id
	 * @param base_time target base time
	 * @return created SubmittedJobEvent
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized Event createJobEvent(UserAgent user,String owner_id,
											 long base_time)
		throws GridException {
		
		long inter_arrival_time =
			CommonMath.sampleUniformDist(this.sc.getGridConfiguration().
											JOBS_INTER_TIME_LO(owner_id),
										 this.sc.getGridConfiguration().
										 	JOBS_INTER_TIME_HI(owner_id));
		
		long timestamp = base_time + inter_arrival_time;
		
		return new SubmittedJobEvent(this.current_id++,timestamp,user,owner_id);	
	
	}
	
	/** Creates a CompletedTaskEvent so that target Task,
	 * submitted by target Consumer Peer, is completed
	 * by target Resource of target Supplier Peer,
	 * after the expected Task runtime from now.
	 * @param consumer_id target Consumer Peer id
	 * @param supplier_id target Supplier Peer id
	 * @param t target Task
	 * @param res target Resource
	 * @return created CompletedTaskEvent
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized Event createCompletedTaskEvent(String consumer_id,
													   String supplier_id,
													   Task t,
													   Resource res)
		throws GridException {

			if (res == null) {
				throw new GridException("createCompletedTaskEvent(): "+
										"illegal Resource");
			}
			
			long now = LightweightBarteringGrid.getTime();
			
			long timestamp = now + res.expectedTaskRuntime();
			
			return
				new CompletedTaskEvent(this.current_id++,timestamp,
									   consumer_id,supplier_id,t,res);
			
	}

	/** Creates a FailedTaskEvent so that target Task,
	 * submitted by target Consumer Peer, is cancelled
	 * by target Resource of target Supplier Peer,
	 * between now and the expected Task runtime from now.
	 * @param consumer_id target Consumer Peer id
	 * @param supplier_id target Supplier Peer id
	 * @param t target Task
	 * @param res target Resource
	 * @return created FailedTaskEvent
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized Event createFailedTaskEvent(String consumer_id,
													String supplier_id,
													Task t,
													Resource res)
		throws GridException {

			if (res == null) {
				throw new GridException("createFailedTaskEvent(): "+
										"illegal Resource");
			}
			
			long now = LightweightBarteringGrid.getTime();
			
			long timestamp =
				now + CommonMath.sampleUniformDist(1,res.expectedTaskRuntime());
			
			return
				new FailedTaskEvent(this.current_id++,timestamp,
									consumer_id,supplier_id,t,res);
			
	}

	/** Creates a SelfEvent.
	 * @return created SelfEvent
	 * @exception GridException if a parameter is illegal
	 */
	public synchronized Event createSelfEvent() throws GridException {
	
		long now = LightweightBarteringGrid.getTime();

		long inter_arrival_time = 1;
			
		long timestamp = now + inter_arrival_time;
		
		return new SelfEvent(this.current_id++,timestamp);	
	
	}

	// ****************************************************************

	private final Scenario sc;
	private long current_id;
	
}

