/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// EventList.java

package lbg.simulator.event;

import lbg.*;
import lbg.common.exception.*;
import lbg.simulator.*;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * current_time = timestamp of Events last extracted from the EventList
 * @author Cyril Briquet
 */
public class EventList {

	public EventList(Scenario sc) throws GridException {
	
		if (sc == null) {
			throw new GridException("EventList(): illegal Scenario");
		}
		this.ef = new EventFactory(sc);

		this.tm_event_list    = new TreeMap<Long,TreeSet<Event>>();
		this.tm_reverse_table = new TreeMap<Long,Long>();
		this.list_size = 0;
		this.list_size_max = 0;
		this.event_count = 0;

	}

	public synchronized String toString() {

		StringBuilder sb = new StringBuilder();
		sb.append("[EventList] current time=");
		try {
			sb.append(LightweightBarteringGrid.getTime());
		}
		catch (GridException e) {
			sb.append("?");
		}
		sb.append(", size=");
		sb.append(this.list_size);
		sb.append(", max size=");
		sb.append(this.list_size_max);
		sb.append(", Events count=");
		sb.append(this.event_count);
		//sb.append("\n");
		
		Iterator<TreeSet<Event>> it = this.tm_event_list.values().iterator();
		while (it.hasNext()) {
			TreeSet<Event> ts = it.next(); 
			Iterator<Event> it2 = ts.iterator();
			while (it2.hasNext()) {
				Event evt = it2.next();
				sb.append("\n");
				sb.append("  ->  ");
				sb.append(evt);
			}
		}
		
		
		return sb.toString();
		
	}
	
	// ****************************************************************

	public EventFactory factory() {
		
		return this.ef;
		
	}
	
	/** Returns true if EventList is empty, i.e. contains no Event.
	 * @return true if Event List is empty
	 */
	public synchronized boolean isEmpty() {
		
		return this.list_size <= 0;
		
	}

	public synchronized void addEvent(Event evt) throws GridException {

		assert(eventListInvariant());

		if (evt == null) {
			throw new GridException("addEvent(): null Event");
		}
		
		if (evt.timestamp() <= LightweightBarteringGrid.getTime()) {
			throw new GridException("addEvent(): timestamp in the past\n"+evt);
		}

		// Locate entry insertion point inside main table
		Long key = Long.valueOf(evt.timestamp());
		TreeSet<Event> ts = this.tm_event_list.get(key);
		if (ts == null) {
			ts = new TreeSet<Event>();
		}

		// Add entry to main table
		if (ts.add(evt) == false) {
			throw new GridException("addEvent(): "+
									"Event id conflict in EventList");
		}
		this.tm_event_list.put(key,ts); // timestamp conflict is normal
		
		// Add entry to reverse table
		if (this.tm_reverse_table.put(evt.idAsKey(),key) != null) {
			throw new GridException("addEvent(): corrupted reverse table");
		}
		
		// Update EventList size stats
		++this.event_count;
		++this.list_size;
		if (this.list_size > this.list_size_max) {
			this.list_size_max = this.list_size;
		}
			
		assert(eventListInvariant());

	}
	
	/**
	 * side effect: time is updated to the timestamp of the extracted Event
	 */
	public synchronized EventContainer extractFirstEvent()
		throws GridException {

		assert(eventListInvariant());

		if (this.tm_event_list == null) {
			throw new GridException("firstEvent(): corrupted EventList");
		}
		
		if (this.tm_event_list.size() == 0) {
			return null;
		}
		
		// Locate main table entries with lowest timestamp
		Long key = this.tm_event_list.firstKey();
		long ikey = key.longValue();

		TreeSet<Event> ts = this.tm_event_list.get(key);
		if (ts == null) {
			throw new GridException("firstEvent(): corrupted EventList");
		}

		// Remove main table entry if no more Events with the same timestamp
		Event the_event = ts.first();
		boolean last_event_at_current_timestamp = false;
		ts.remove(the_event);
		if (ts.size() == 0) {
			this.tm_event_list.remove(key);
			last_event_at_current_timestamp = true;
		}

		// Remove reverse table entry
		this.tm_reverse_table.remove(the_event.idAsKey());

		// Update current simulation time
		if (LightweightBarteringGrid.getTime() > ikey) {
			throw new GridException("firstEvents(): time skew");
		}
		LightweightBarteringGrid.setTime(ikey);

		// Update EventList size stats
		--this.list_size;
		if (this.list_size < 0) {
			throw new GridException("firstEvents(): negative EventList size");
		}
		
		assert(eventListInvariant());

		return new EventContainer(the_event,last_event_at_current_timestamp);
		
	}
	
	public synchronized void removeEvent(Event evt) throws GridException {

		assert(eventListInvariant());

		// Lookup entry in main table
		Long key = this.tm_reverse_table.get(evt.idAsKey());
		if (key == null) {
			throw new GridException("removeEvent(): damaged reverse table, "+
									"cannot find "+evt);
		}
		
		TreeSet<Event> ts = this.tm_event_list.get(key);
		if (ts == null) {
			throw new GridException("removeEvent(): damaged main table");
		}
		
		// Remove entry from main table
		if (ts.remove(evt) == false) {
			throw new GridException("removeEvent(): corrupted main table");
		}

		// If target Event was alone at t=timestamp,
		// remove container TreeSet from main table
		if (ts.size() == 0) {
			this.tm_event_list.remove(key);
		}

		// Remove entry from reverse table
		if (this.tm_reverse_table.remove(evt.idAsKey()) == null) {
			throw new GridException("removeEvent(): corrupted reverse table");
		}

		// Update EventList size stats
		--this.list_size;
		if (this.list_size < 0) {
			throw new GridException("firstEvents(): negative EventList size");
		}
		
		assert(eventListInvariant());

	}

	public synchronized boolean containsEvent(Event evt) throws GridException {
	
		if (this.list_size <= 0) {
			return false;
		}
		
		Long key = this.tm_reverse_table.get(evt.idAsKey());
		if (key == null) {
			throw new GridException("containtsEvent(): damaged reverse table");
		}
		
		TreeSet<Event> ts = this.tm_event_list.get(key);
		if (ts == null) {
			throw new GridException("containtsEvent(): corrupted main table");
		}
		
		return ts.contains(evt);

	}

	// ****************************************************************

	private synchronized boolean eventListInvariant() {
		
		return (this.list_size == this.tm_reverse_table.size());
		
	}

	// ****************************************************************

	private final EventFactory ef;

	// tm_event_list:
	//   KEY = timestamp
	//   VALUE = TreeSet of Events happening at t=timestamp
	private final TreeMap<Long,TreeSet<Event>> tm_event_list;
	
	// tm_reverse_table
	//  KEY = Event id
	//  VALUE = tm_event_list key of target Event (i.e. a timestamp)
	private final Map<Long,Long> tm_reverse_table;
	
	private       int list_size;
	private       int list_size_max;
	
	private       long event_count;

}

