/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// GlobalPropertiesStore.java

package lbg.simulator.tools;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.common.math.*;

import java.util.ArrayList;
import java.util.Map;
import java.util.TreeMap;
import java.util.Properties;

/**
 * @author Cyril Briquet
 */
public class GlobalPropertiesStore {

	/** Constructs a new GlobalPropertiesStore.
	 */
	public GlobalPropertiesStore() throws GridException {

		this.global_props_keys = new TreeMap<String,Measurement>();
		
	}
	
	public String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		synchronized(this.global_props_keys) {
			sb.append("[GlobalPropertiesStore]");
		}

		
		return sb.toString();
		
	}

	// ****************************************************************

	public void storeProperty(String log_name,String key,String value)
		throws GridException {

		if (log_name == null) {
			throw new GridException("analyzeLogProperty(): illegal "+
									"Log name");
		}
		if (key == null) {
			throw new GridException("storeProperty(): illegal key");
		}
		if (value == null) {
			throw new GridException("storeProperty(): illegal value");
		}

		// Filter value
		if (value.equals("NaN")) {
			return;
		}

		// Convert value
		double val = CommonStringHandling.loadDouble(value);

		// Store value
		synchronized(this.global_props_keys) {
			Measurement mea = this.global_props_keys.get(key);
			if (mea == null) {
				mea = MeasurementFactory.createSummarizedMeasurement(key);
				this.global_props_keys.put(key,mea);
			}
			mea.addObservation(val,log_name);
		}
		
	}
	
	/** Returns stored Measurement.
	 * @param key target key
	 * @return stored Measurement, guaranteed to be non-null
	 * @exception GridException if a parameter is illegal,
	 * or a Measurement cannot be found for target key
	 */
	public Measurement getMeasurement(String key)
		throws GridException {
	
		if (key == null) {
			throw new GridException("getMeasurement(): illegal key");
		}
		
		Measurement mea = null;
		
		synchronized(this.global_props_keys) {
			mea = this.global_props_keys.get(key);
		}

		if (mea == null) {
			throw new GridException("getMeasurement(): Measurement not found");
		}
		
		return mea;
		
	}
	
	public int size() {
	
		int size = -1;
		
		synchronized(this.global_props_keys) {
			size = this.global_props_keys.size();
		}
		
		return size;
		
	}

	public ArrayList<String> keys() {
	
		ArrayList<String> al = new ArrayList<String>(size());
		
		synchronized(this.global_props_keys) {
			for (String key : this.global_props_keys.keySet()) {
				al.add(key);
			}
		}
		
		return al;
		
	}

	// ****************************************************************

	private final Map<String,Measurement> global_props_keys;
	
}

