/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LogAnalyzer.java

package lbg.simulator.tools;

import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonFileHandling.*;
import static lbg.peer.config.PeerStrategyConfiguration.getPDPList;

import java.util.concurrent.atomic.AtomicInteger;
import java.util.Map;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.TreeMap;

/** Analyzes a given LBG simulator log file.
 * @author Cyril Briquet
 */
public abstract class LogAnalyzer {

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LogAnalyzer]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** Analyzes target log.
	 * @param log_name log filenmame
	 * @param global_props_store global properties store where to store
	 * the extracted properties
	 * @param sb_encountered_exceptions encountered Exceptions store,
	 * where Exceptions found in log files will be reported
	 * @param encountered_exceptions counter of encountered Exceptions
	 * @return CSV-formatted summary of values of extracted properties
	 */
	public static String analyzeLog(String log_name,
					   		 		GlobalPropertiesStore global_props_store,
									StringBuffer sb_encountered_exceptions,
									AtomicInteger encountered_exceptions)
		throws GridException {

		if (isLogFilename(log_name) == false) {
			throw new GridException("LogAnalyzer(): illegal "+
									"log filename "+log_name);
		}
		if (global_props_store == null) {
			throw new GridException("LogAnalyzer(): illegal "+
									"global properties keys");
		}
		if (sb_encountered_exceptions == null) {
			throw new GridException("LogAnalyzer(): illegal "+
									"encountered Exceptions store");
		}
		if (encountered_exceptions == null) {
			throw new GridException("LogAnalyzer(): illegal "+
									"encountered Exceptions counter");
		}

		// Load log contents
		String log_contents = readFile(log_name);

		// Check simulation success
		String e_msg = reportSimulationException(log_contents);
		
		if (e_msg == null) {
			// Success, can process log
			String props_values_csv_summary =
				analyzeLogContents(log_name,log_contents,global_props_store);
			return ", OK"+props_values_csv_summary;
		}
		else {
			encountered_exceptions.incrementAndGet();
			sb_encountered_exceptions.append(log_name).append(":\n").
				append(e_msg).append("\n\n");
			return ", FAILED";
		}

	}

	/** Analyzes scenario of target log.
	 * @param log_name log filenmame
	 * @return CSV summary of scenario
	 */
	public static String analyzeScenario(String log_name)
		throws GridException {
		
		if (isLogFilename(log_name) == false) {
			throw new GridException("LogAnalyzer(): illegal "+
									"log filename "+log_name);
		}

		// Load scenario
		String scenario_name = log_name.substring(LOG_PREFIX.length());
		Properties scenario_props = loadProperties(scenario_name);

		// Build CSV summary of scenario
		StringBuilder sb_csv = new StringBuilder();
		sb_csv.append(log_name);
		String[] pdp_list = getPDPList();
		for (String target_pdp : pdp_list) {
			processScenarioProperty(scenario_props,target_pdp,sb_csv);
		}
		
		return sb_csv.toString();
	
	}
	
	public static boolean isStrategyProperty(String prop)
		throws GridException {
		
		if (prop == null) {
			throw new GridException("isStrategyColumn(): illegal property");
		}
		
		String[] pdp_list = getPDPList();
		for (String target_pdp : pdp_list) {
			if (prop.equals(target_pdp)) {
				return true;
			}
		}
		
		return false;

	}

	// ****************************************************************

	/** Analyze log contents.
	 * @return CSV-formatted summary of values of extracted properties
	 */
	private static String
		analyzeLogContents(String log_name,String log_contents,
						   GlobalPropertiesStore global_props_store)
		throws GridException {
		
		if (log_name == null) {
			throw new GridException("analyzeLogContents(): "+
									"illegal log filaneme ");
		}
		if (log_contents == null) {
			throw new GridException("analyzeLogContents(): "+
									"illegal log contents for "+log_name);
		}
		if (global_props_store == null) {
			throw new GridException("analyzeLogContents(): "+
									"illegal global properties store");
		}

		// Flagged lines (i.e. beginning by "~>") are analyzed
		StringTokenizer st = new StringTokenizer(log_contents,"~"); // tilde
		if (st.countTokens() < 1) {
			throw new GridException("analyzeLogContents(): corrupted data");
		}
		
		// Skip what's before the 1st tilde
		st.nextToken();
		
		// Loop over the flagged lines
		Map<String,String> local_props_store = new TreeMap<String,String>();
		while (st.hasMoreTokens()) {
			analyzeLogProperty(log_name,st.nextToken(),
							   global_props_store,local_props_store);
		}
		
		// Generate CSV-formatted summary of values of extracted properties
		StringBuilder sb = new StringBuilder();
		for (Map.Entry<String,String> prop : local_props_store.entrySet()) {
			sb.append(", ").append(prop.getValue());
		}
		
		return sb.toString();
		
	}

	// The algorithm to analyze a given flag line
	private static void
		analyzeLogProperty(String log_name,String log_property,
						   GlobalPropertiesStore global_props_store,
						   Map<String,String> local_props_store)
		throws GridException {

		if (log_name == null) {
			throw new GridException("analyzeLogProperty(): illegal "+
									"Log name");
		}
		if (log_property == null) {
			throw new GridException("analyzeLogProperty(): illegal "+
									"Log Property");
		}
		if (global_props_store == null) {
			throw new GridException("analyzeLogProperty(): "+
									"illegal global properties store");
		}
		
		// Extract data before end of line
		/* Precondition: data ==">  a.b.c = property_value \n junk junk junk" */
		StringTokenizer st = new StringTokenizer(log_property,"\n\r");
		if (st.countTokens() < 1) {
			throw new GridException("analyzeLogProperty(): illegal "+
									"data |"+log_property+"|");
		}
		log_property = st.nextToken();

		// Remove leading ">"
		/* Precondition: data == ">  a.b.c = property_value " */
		if ((log_property.length() < 4) || (log_property.charAt(0) != '>')) {
			throw new GridException("analyzeLogProperty(): illegal "+
									"data |"+log_property+"|");
		}
		log_property = log_property.substring(1);

		// Extract key and value
		/* Precondition: data == "  a.b.c = property_value " */
		st = new StringTokenizer(log_property,"=");
		if (st.countTokens() != 2) {
			throw new GridException("analyzeLogProperty(): illegal "+
									"data |"+log_property+"|");
		}
		String key   = (st.nextToken()).trim();
		String value = (st.nextToken()).trim();
		if ((key == null) || (key.length() < 1)) {
			throw new GridException("analyzeLogProperty(): illegal data");
		}
		if ((value == null) || (value.length() < 1)) {
			throw new GridException("analyzeLogProperty(): illegal data");
		}
		
		// Store Property (log_name, key, value) in global Properties keys store
		/* Precondition: key == "a.b.c", value == "property_value" */
		// debugging code -- System.out.println(key+" = "+value);
		global_props_store.storeProperty(log_name,key,value);
		
		// Store Property (key, value) in local Properties keys store
		if (local_props_store.put(key,value) != null) {
			throw new GridException("analyzeLogProperty(): key "+key+
									"present twice in "+log_name);
		}
		
	}
	
	/** Reports GridException in simulation log file, if any.
	 * @param log_contents log file contents
	 * @return null if no GridException was found,
	 * or a String with the cause of a GridException if one is found
	 */
	private static String reportSimulationException(String log_contents)
		throws GridException {
		
		if (log_contents == null) {
			throw new GridException("reportSimulationException(): "+
									"illegal log contents");
		}
		
		String EXCEPTION_MSG = /*"lbg.common.Grid*/ "Exception: ";
		int e_idx = log_contents.indexOf(EXCEPTION_MSG);
		if (e_idx > -1) { // GridException
			int endline_idx =
				log_contents.indexOf("\n\n",e_idx+EXCEPTION_MSG.length());
			while ((e_idx >=0) && (log_contents.charAt(--e_idx) != '\n')) ;
			if (log_contents.charAt(e_idx) == '\n') ++e_idx;
			return log_contents.substring(e_idx,endline_idx);
		}
		else { // Success
			return null;
		}

	}
	
	private static void
		processScenarioProperty(Properties scenario_props,String target_prop,
							    StringBuilder sb) throws GridException {
		
		if (scenario_props == null) {
			throw new GridException("processScenarioProperty(): "+
									"illegal scenario properties");
		}
		if (target_prop == null) {
			throw new GridException("processScenarioProperty(): "+
									"illegal target property");
		}
		if (sb == null) {
			throw new GridException("processScenarioProperty(): "+
									"illegal StringBuilder");
		}
		
		sb.append(", ");

		String[] values = CommonStringHandling.
			loadStringArray(scenario_props.getProperty(target_prop));
		sb.append(values[0]);
	
	}

	// ****************************************************************

}

