/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LogAggregator.java

package lbg.simulator.tools.aggregator;

import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonFileHandling.*;
import lbg.simulator.tools.*;

import java.util.concurrent.atomic.AtomicInteger;
import java.util.ArrayList;
import java.util.StringTokenizer;

/** Aggregates execution logs of the Lightweight Bartering Grid Simulator,
 * so as to produce statistically significant results
 * based on multiple executions.
 * @author Cyril Briquet
 */
public class LogAggregator implements Runnable {

	/** Constructs a new LogAggregator.
	 * @param list_of_logs filename of target file containing a space-separated
	 * (i.e. StringTokenizer parsable, easily shell-generated) list
	 * of filenames of logs to aggregate.
	 * @param thread_pool_size thread pool size
	 */
	public LogAggregator(String list_of_logs,int thread_pool_size)
		throws GridException {
		
		if (list_of_logs == null) {
			throw new GridException("LogAggregator(): illegal "+
									"list of logs filenames");
		}
		if ((thread_pool_size < 1) || (thread_pool_size > 1024)) {
			throw new GridException("LogAggregator(): illegal "+
									"thread pool size");
		}
		
		System.out.println(WELCOME_LA);
		
		this.thread_pool =
			new ThreadPool(this,thread_pool_size,false,false,"LogAggregator");
		this.thread_pool_size = thread_pool_size;
		
		this.global_run = true;
		this.next_log_to_complete = 0;
		this.completed_logs = 0;
		this.encountered_errors = new AtomicInteger(0);
		this.sb_encountered_errors = new StringBuffer();
		this.encountered_exceptions = new AtomicInteger(0);
		this.sb_encountered_exceptions = new StringBuffer();
		ArrayList<String> al_log_names = loadLogNames(list_of_logs);
		this.logs_count = al_log_names.size();
		this.logs_names = al_log_names.toArray(new String[] {});
		this.logs_csv_summary = new String[this.logs_count];
		for (int i = 0 ; i < this.logs_count ; i++) {
			this.logs_csv_summary[i] = null;
		}
		this.global_props_store = new GlobalPropertiesStore();
		this.sync_on_me = new Object();
		
		// Extract base scenario filename
		this.scenario_base_name =
			extractBaseScenarioFilename(this.logs_names[0]);
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LogAggregator]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** Runnable implementation.
	 */
	public void run() {

		try {
			boolean thread_run = true;
			while (thread_run && global_run) {
		
				int l = -1;
				synchronized(this.sync_on_me) {
					// All logs processed, or being processed by other threads
					if (this.next_log_to_complete >= this.logs_count) {
						thread_run = false;
					}
					// Let's select the next log to process
					else {
						l = ++this.next_log_to_complete;
					}
				}
	
				if (thread_run) {
					try {
						analyzeLog(l);
					}
					catch (GridException e) {
						this.encountered_errors.incrementAndGet();
						String stack_trace =
							CommonStringHandling.getStackTrace(e);
							this.sb_encountered_errors.
								append(this.logs_names[l-1]).
								append(":\n").append(stack_trace).append("\n");

						/*global_run = false; -- uncomment to halt processing
						  thread_run = false; -- upon log analysis failure    */
					}
					finally {
						// Update completed logs status
						int c_l = 0;
						synchronized(this.sync_on_me) {
							c_l = ++this.completed_logs;
						}
						
						// Display progress status
						if (((c_l % 100 == 0) && (c_l > 0)) ||
							(c_l == this.logs_count)) {
							System.out.print("\rAnalyzed "+c_l+" logs");
						}
						
					}
				}

			}
		}
		catch (Throwable thr) {
			this.encountered_errors.incrementAndGet();
			String stack_trace = CommonStringHandling.getStackTrace(thr);
			this.sb_encountered_errors.append("[log file unknown]").
				append(":\n").append(stack_trace).append("\n");
			
			this.global_run = false;
		}
		
		synchronized(this) {
			this.notify();
		}

	}

	// ****************************************************************

	public synchronized void analyzeLogs() throws GridException {

		long t_begin = System.currentTimeMillis();
		
		// Launch multiple processing threads
		this.thread_pool.startService();
		
		long t_end = System.currentTimeMillis();
		System.out.println("Launched "+this.thread_pool_size+" threads in "+
						   (t_end-t_begin)+" ms");

		t_begin = t_end;
		
		// Wait for all threads to return
		waitForThreadsToReturn();
		
		t_end = System.currentTimeMillis();
		System.out.println("\rAnalyzed "+this.logs_count+" logs in "+
						   (t_end-t_begin)+" ms");
		
	}
	
	public synchronized void aggregateLogs() throws GridException {

		long t_begin = System.currentTimeMillis();

		String timestamp = CommonStringHandling.generateTimeStamp();
		
		LogAggregatorReportManager la_rep_mgr =
			new LogAggregatorReportManager(timestamp,this.scenario_base_name,
										   this.logs_count,
										   this.logs_csv_summary,
										   this.global_props_store,
										   this.sb_encountered_errors.
										   	toString(),
										   this.sb_encountered_exceptions.
										   	toString());
		
		la_rep_mgr.saveLogAnalysisSummary();
		la_rep_mgr.saveLogSpreadsheetReport();
		
		long t_end = System.currentTimeMillis();
		System.out.println("Aggregated "+this.global_props_store.size()+
						   " properties from "+this.logs_count+" log files in "+
						   (t_end-t_begin)+" ms");
		int e_ex = this.encountered_exceptions.get();
		if (e_ex > 0) {
			System.out.println("**** There w"+(e_ex > 1 ? "ere " : "as ")+
							   this.encountered_exceptions+
							   " Exception"+(e_ex > 1 ? "s" : "")+
							   " in the logs ****");
		}
		int e_er = this.encountered_errors.get();
		if (e_er > 0) {
			System.out.println("**** There w"+(e_er > 1 ? "ere " : "as ")+
							   this.encountered_errors+
							   " error"+(e_er > 1 ? "s" : "")+
							   " during the analysis of the logs ****");
		}

		t_begin = System.currentTimeMillis();

		String spreadsheet_fn = la_rep_mgr.getSpreadsheetFilename();
		StrategyPerformanceAnalyzer spa =
			new StrategyPerformanceAnalyzer(timestamp,spreadsheet_fn);
		
		spa.analyzeStrategyPerformance();
		spa.saveStrategyPerformanceAnalysisReport();
		
		t_end = System.currentTimeMillis();
		System.out.println("Analyzed strategy performance from "+
						   this.logs_count+" log files in "+
						   (t_end-t_begin)+" ms");
		

	}
	
	// ****************************************************************

	public static void main(String[] args) {

		try {
			long t_begin = System.currentTimeMillis();
			
			checkArgs(args);
			int thread_pool_size =
				(args.length > 1) ? Integer.parseInt(args[1]) :
				DEFAULT_THREAD_POOL_SIZE;
			
			LogAggregator l_ag =
				new LogAggregator(args[0],thread_pool_size);
			l_ag.analyzeLogs();
			l_ag.aggregateLogs();
			
			long t_end = System.currentTimeMillis();
			
			System.out.println("Analysis completed in "+(t_end-t_begin)+" ms");
		}
		catch (GridException e) {
			e.printStackTrace();
		}
		catch (Throwable thr) {
			thr.printStackTrace();
		}
		finally {
			System.exit(0);
		}
		
	}

	// ****************************************************************

	public final int thread_pool_size;
	public static final int DEFAULT_THREAD_POOL_SIZE = 10;

	/* How to determine THREAD_POOL_SIZE ?
	   
	   Total runtime in function of logs count, threads count
	   
	   Intel Core Duo (2 CPU)
		  t  |   1  |   5  |  10  |  20  |  40  |
		-----+------+------+------+------+------+
		1000 | 162' | 166' |  95' |  94' |  88' |
		1700 | 275' | 201' | 159' | 153' | 149' |
		
		Intel Xeon (4 x 2 CPU = 8 CPU)
		  t  |   1  |   5  |  10  |  20  |  40  |
		-----+------+------+------+------+------+
		2410 |  84' |  18' |  12' |  13' |  13' |
		
		=> 10 threads == good overall value (will not flood single core CPU)
	*/

	// ****************************************************************

	private synchronized void waitForThreadsToReturn() throws GridException {

		synchronized (this) {
			while (global_run && (this.completed_logs < this.logs_count)) {
				try {
					this.wait();
				}
				catch (Exception e) {
					throw new GridException(e);
				}
			}
		}

	}

	private void analyzeLog(int l) throws GridException {

		if ((l < 0) || (l > this.logs_count)) {
			throw new GridException("analyzeLog(): illegal log "+l);
		}
		if (this.logs_csv_summary[l-1] != null) {
			throw new GridException("analyzeLog(): log "+l+
									"already processed");
		}

		String log_name = this.logs_names[l-1];
		
		// Scenario analysis (initial)
		this.logs_csv_summary[l-1] = LogAnalyzer.analyzeScenario(log_name);
		// Why do initial scenario analysis here ? So that the log's entry
		// appears in the CSV spreadsheet even if the log cannot be analyzed
		// (i.e. an error occurs and a GridException is thrown by analyzeLog())
		
		// Log analysis
		String props_values_csv_summary = null;
		try {
			props_values_csv_summary =
				LogAnalyzer.analyzeLog(log_name,this.global_props_store,
									   this.sb_encountered_exceptions,
									   this.encountered_exceptions);
		}
		catch (GridException e) {
			this.logs_csv_summary[l-1] = this.logs_csv_summary[l-1]+", DNF";
			throw(e);
		}
		
		// Scenario analysis (final)
		this.logs_csv_summary[l-1] =
			this.logs_csv_summary[l-1]+props_values_csv_summary;
		
	}
	
	// ****************************************************************

	private static void checkArgs(String[] args) throws GridException {
	
		boolean ok = true;

		if (ok) {
			if (args == null) {
				ok = false;
			}
		}

		if (ok) {
			if ((args.length <= 0) || (args.length >= 3)) {
				ok = false;
			}
		}

		if (ok && (args.length >= 2)) {
			ok = ok && (args[1] != null);
			if (ok) {
				try {
					Integer.parseInt(args[1]);
				}
				catch (NumberFormatException e) {
					ok = false;
				}
			}
		}
		
		if (ok && (args.length >= 1)) {
			ok = ok && ((args[0] != null) &&
				        (args[0].length() >= 1+TXT_EXT.length()));
		}
		
		if (ok == false) {
			System.out.println("Usage: "+
							   "java -cp classpath "+
							   LogAggregator.class.getName()+
							   " list_of_logs.txt [THREAD POOL SIZE]\n");
			throw new GridException("checkArgs(): illegal arguments");
		}
		
	}
	
	// ****************************************************************

	private static ArrayList<String> loadLogNames(String list_of_logs)
		throws GridException {
		
		if (list_of_logs == null) {
			throw new GridException("loadLogNames(): "+
									"illegal list of logs filename");
		}
		
		String s = CommonFileHandling.readFile(list_of_logs);
		StringTokenizer st = new StringTokenizer(s);
		ArrayList<String> al_log_names = new ArrayList<String>();
		while (st.hasMoreTokens()) {
			String log_name = st.nextToken();
			if (isLogFilename(log_name) == false) {
				throw new GridException("loadLogNames(): illegal "+
										"log filename "+log_name);
			}
			al_log_names.add(log_name);
		}
		
		if (al_log_names.size() < 1) {
			throw new GridException("loadLogNames(): no log to aggregate");
		}
		
		return al_log_names;
		
	}
	
	private static String extractBaseScenarioFilename(String one_log_name)
		throws GridException {
		
		if (one_log_name == null) {
			throw new GridException("extractBaseScenarioFilename(): "+
									"illegal log name");
		}
		
		StringTokenizer st = new StringTokenizer(one_log_name,".");
		String dummy = st.nextToken(); // remove ".txt"
		final String sep = "-";
		st = new StringTokenizer(dummy,sep);
		st.nextToken(); // skip "log-"
		StringBuilder sb = new StringBuilder();
		int acc = 0;
		while (st.hasMoreTokens()) {
			String tok = st.nextToken();
			if (st.hasMoreTokens()) {
				if (++acc > 1) {
					sb.append(sep);
				}
				sb.append(tok);
			}
		}
		sb.append(LOG_EXT);
		
		return sb.toString();
		
	}
	

	// ****************************************************************

	private final ThreadPool thread_pool;
	private final int logs_count;
	private final String[] logs_names;
	private final String[] logs_csv_summary;
	private final String scenario_base_name;
	private final GlobalPropertiesStore global_props_store;
	private final Object sync_on_me;

	private       boolean global_run;
	private       int next_log_to_complete;
	private       int completed_logs;

	private       AtomicInteger encountered_errors;
	private final StringBuffer sb_encountered_errors;
	private       AtomicInteger encountered_exceptions;
	private final StringBuffer sb_encountered_exceptions;

	// ****************************************************************

	public static final String WELCOME_LA =
		"==== LBG Simulation Log Aggregator ====\n";

	// ****************************************************************

}

