/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package lbg.simulator.tools.aggregator;

import lbg.common.exception.*;
import java.util.StringTokenizer;
import java.util.TreeMap;

/** Log Properties trie.
 * <p>A Log Properties trie stores all Properties of a log..
 * It is a modified trie whose nodes and leafs are Strings.
 * Properties keys are stored in the nodes and values are stored in the leafs.
 * <p>For example, if a given log has 4 properties "a.b.c = val1",
 * "a.b.d = val2", "e = val3" and "e.f = val4",
 * the trie may be represented like this:
<pre>root -> "a" -> "b" -> "c" -> "val1"
            -> "d" -> "val2"
     -> "e" -> "f" -> "val3"
     -> "e" -> "val4"</pre>
 * @author Cyril Briquet
 */
public class LogPropertiesTrie {

	public LogPropertiesTrie() throws GridException {
		
		this.root = new LogPropertiesTrieNode();

	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LogPropertiesTrie]");
		try {
			sb.append("\n");
			sb.append(this.enumerate());
		}
		catch (GridException e) {
			sb.append("  -->  warning: could not enumerate !");
		}
		
		return sb.toString();
		
	}

	// ****************************************************************
	
	public synchronized void add(String subkeys,String value)
		throws GridException {
		
		if (subkeys == null) {
			throw new GridException("add(): illegal subkeys");
		}
		if (value == null) {
			throw new GridException("add(): illegal value");
		}
		
		StringTokenizer st_subkeys = new StringTokenizer(subkeys,".");
		this.root.add(st_subkeys,value);

	}
	
	public synchronized String enumerate() throws GridException {

		StringBuilder log_sb = new StringBuilder(); 
		
		this.root.enumerate(log_sb,new StringBuilder());
		
		return log_sb.toString();

	}

	public synchronized String get(String subkeys)
		throws GridException {

		if (subkeys == null) {
			throw new GridException("get(): illegal subkeys");
		}
		
		StringTokenizer st_subkeys = new StringTokenizer(subkeys,".");
		return this.root.get(st_subkeys);
		
	}

	// ****************************************************************

	/*java -cp jars/lightweight_bartering_grid.jar
	       lbg.simulator.tools.aggregator.LogPropertiesTrie*/

	// For testing purposes
	public static void main(String[] args) {
	
		try {
			LogPropertiesTrie lpt = new LogPropertiesTrie();
			
			lpt.add("a.b.c","val1");
			lpt.add("a.b.d","val2");
			lpt.add("e.f","val3");
			lpt.add("e","val4");
			System.out.println(lpt);
			System.out.println(lpt.get("e"));
			System.out.println(lpt.get("a.b.d"));
			System.out.println(lpt.get("a.b.c"));
			System.out.println(lpt.get("e.f"));
		}
		catch (GridException e) {
			e.printStackTrace();
		}
		finally {
			System.exit(0);
		}
		
	}

	
	private LogPropertiesTrieNode root;

		// ****************************************************************

	/** Node of a Log Properties trie.
	 */
	class LogPropertiesTrieNode {
	
		/** Constructs a new terminal Log Properties trie node.
		public LogPropertiesTrieNode(String tag) throws GridException {
			
			if (tag == null) {
				throw new GridException("LogPropertiesTrieNode(): illegal tag");
			}
			
			this.tag = tag;
			this.next = null;
	
		}*/
		
		/** Constructs a new non-terminal Log Properties trie node.
		 */
		public LogPropertiesTrieNode() throws GridException {
			
			this.value = null;
			this.next = new TreeMap<String,LogPropertiesTrieNode>();
	
		}
	
		public synchronized String toString() {
		
			StringBuilder sb = new StringBuilder();
			
			sb.append("[LogPropertiesTrieNode]");
			
			return sb.toString();
			
		}
	
		// ****************************************************************
	
		public synchronized void setValue(String value) throws GridException {
	
			if (this.value != null) {
				throw new GridException("setValue(): value already set");
			}
			if ((value == null) || (value.length() < 1)) {
				throw new GridException("setValue(): illegal value");
			}
			
			this.value = value;
	
			
		}
	
		public synchronized void add(StringTokenizer subkeys,String value)
			throws GridException {
			
			if (subkeys == null) {
				throw new GridException("add(): illegal subkeys");
			}
			if (value == null) {
				throw new GridException("add(): illegal value");
			}
			
			int count = subkeys.countTokens();
			
			if (count == 0) {
				this.setValue(value);
			}
			else /* if (count > 0) */ {
				addNode(subkeys,value);
			}
			
		}
	
		public synchronized void enumerate(StringBuilder log_sb,
										   StringBuilder recursion_sb)
			throws GridException {
	
			if (log_sb == null) {
				throw new GridException("enumerate(): "+
										"illegal log StringBuilder");
			}
			if (recursion_sb == null) {
				throw new GridException("enumerate(): "+
										"illegal recursion StringBuilder");
			}
			
			if (this.isLeaf()) {
				log_sb.append("  ->  ");
				log_sb.append(recursion_sb.toString());
				log_sb.append(" = ");
				log_sb.append(this.value);
				log_sb.append("\n");
			}
	
			for (String subkey : this.next.keySet()) {
				StringBuilder new_sb = new StringBuilder(recursion_sb);
				if (new_sb.length() > 0) new_sb.append(".");
				new_sb.append(subkey);
	
				LogPropertiesTrieNode lptn = this.next.get(subkey);
				lptn.enumerate(log_sb,new_sb);
			}
	
		}
	
		public synchronized String get(StringTokenizer subkeys)
			throws GridException {
			
			if (subkeys == null) {
				throw new GridException("get(): illegal subkeys");
			}
			
			if (subkeys.countTokens() == 0) {
				return (this.isLeaf() == true) ? this.value : null;
			}
			else /* if (subkeys.countTokens() > 0) */ {
				return getNode(subkeys);
			}
			
		}
		
		
		// ****************************************************************
	
		private synchronized boolean isLeaf() {
			
			return (this.value != null);
			
		}
	
		private synchronized void addNode(StringTokenizer subkeys,String value)
			throws GridException {
			
			if (subkeys == null) {
				throw new GridException("addNode(): illegal subkeys");
			}
			if (value == null) {
				throw new GridException("addNode(): illegal value");
			}
	
			String key = subkeys.nextToken();
			
			if (key == null) {
				throw new GridException("addNode(): illegal subkey");
			}
			
			// Get LogPropertiesTrieNode for key
			LogPropertiesTrieNode lptn = this.next.get(key);
			if (lptn == null) {
				lptn = new LogPropertiesTrieNode();
				this.next.put(key,lptn);
			}
			
			// Add rest of subkeys, value to the extracted LogPropertiesTrieNode
			lptn.add(subkeys,value);
	
		}
		
		private synchronized String getNode(StringTokenizer subkeys)
			throws GridException {
			
			if (subkeys == null) {
				throw new GridException("getNode(): illegal subkeys");
			}
	
			String key = subkeys.nextToken();
			
			if (key == null) {
				throw new GridException("getNode(): illegal subkey");
			}
			
			// Get LogPropertiesTrieNode for key
			LogPropertiesTrieNode lptn = this.next.get(key);

			// Not found !
			if (lptn == null) {
				return null;
			}
			
			// Found, continue traversal
			return lptn.get(subkeys);
			
		}
	
		// ****************************************************************
	
		private String value;
		private TreeMap<String,LogPropertiesTrieNode> next;
		
	}
	
	// ****************************************************************

}

