/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PolicyPerformanceAnalyzer.java

package lbg.simulator.tools.aggregator;

import lbg.common.exception.*;
import lbg.common.math.*;
import static lbg.common.math.CommonMath.*;
import static lbg.peer.config.PeerStrategyConfiguration.getPDPLabels;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * @author Cyril Briquet
 */
public class PolicyPerformanceAnalyzer {
	
	/** Constructs a new PolicyPerformanceAnalyzer.
	 */
	public PolicyPerformanceAnalyzer(String pdp_label,String[] ranked_policies,
									 Float[] ranked_mbrt)
		throws GridException {
		
		this(pdp_label,ranked_policies,ranked_mbrt,1.0f);
		
	}
	
	/** Constructs a new PolicyPerformanceAnalyzer.
	 */
	public PolicyPerformanceAnalyzer(String pdp_label,String[] ranked_policies,
									 Float[] ranked_mbrt,
									 float best_row_analysis_ratio)
		throws GridException {

		if (pdp_label == null) {
			throw new GridException("PolicyPerformanceAnalyzer() "+
									"illegal PDP label");
		}
		if (ranked_policies == null) {
			throw new GridException("PolicyPerformanceAnalyzer() "+
									"illegal ranked policies");
		}
		if (ranked_mbrt == null) {
			throw new GridException("PolicyPerformanceAnalyzer() "+
									"illegal ranked MBRTs");
		}
		if (ranked_policies.length != ranked_mbrt.length) {
			throw new GridException("PolicyPerformanceAnalyzer() "+
									"ranked policies and MBRTs count mismatch");
		}
		if ((best_row_analysis_ratio <= 0.0f) ||
			(best_row_analysis_ratio > 1.0f)) {
			throw new GridException("PolicyPerformanceAnalyzer() "+
									"illegal best row analysis ratio");
		}

		this.pdp_label = pdp_label;
		this.pdp_labels = new TreeSet<String>();
		String[] tmp_pdp_labels = getPDPLabels(pdp_label);
		for (String l : tmp_pdp_labels) {
			if (this.pdp_labels.add(l) == false) {
				throw new GridException("PolicyPerformanceAnalyzer(): "+
										"duplicate policy labels for PDP "+
										this.pdp_label);
			}
		}
		this.ranked_policies = ranked_policies;
		this.ranked_mbrt = ranked_mbrt;
		this.best_row_analysis_ratio = best_row_analysis_ratio;
		this.analysis_count =
			Math.max(Math.round(this.best_row_analysis_ratio*
					 			((float)ranked_policies.length)),1);
		assert((this.analysis_count > 0) &&
			   (this.analysis_count <= this.ranked_policies.length));
		
		this.policy_rank_analysis_map = new TreeMap<String,Measurement>();
		this.policy_mbrt_analysis_map = new TreeMap<String,Measurement>();
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PolicyPerformanceAnalyzer ").
			append(this.pdp_label).
			append(" over top ").
			append(div_label(Math.round(best_row_analysis_ratio*100.0f))).
			append("% best strategies").append("]");
		
		for (String l : this.pdp_labels) {
			sb.append("\n  ->  ").append(l).append("  ").append("{ ");
			
			Measurement m_rank = this.policy_rank_analysis_map.get(l);
			if (m_rank == null) {
				sb.append(" **CORRUPTED DATA**");
			}
			else {
				sb.append("mean rank = ").
					append(div_label(m_rank.getMean())).
					append(", stddev rank = ").
					append(div_label(m_rank.getStandardDeviation()));
			}
			
			Measurement m_mbrt = this.policy_mbrt_analysis_map.get(l);
			if (m_mbrt == null) {
				sb.append(" **CORRUPTED DATA**");
			}
			else {
				sb.append(", mean MBRT = ").
					append(div_label(m_mbrt.getMean())).
					append(", stddev MBRT = ").
					append(div_label(m_mbrt.getStandardDeviation()));
			}

			sb.append(" }");
		}
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized String outputAsCSV() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(this.pdp_label).
			append(" (top ").
			append(div_label(Math.round(best_row_analysis_ratio*100.0f))).
			append("% best)").append(", ").
			append("strategy.rank.mean").append(", ").
			append("strategy.rank.min").append(", ").
			append("strategy.rank.max").append(", ").
			append("strategy.rank.stddev").append(", ").
			append("grid.job_completion_time.mean").append(", ").
			append("grid.job_completion_time.min").append(", ").
			append("grid.job_completion_time.max").append(", ").
			append("grid.job_completion_time.stddev").append("\n");
		
		for (String l : this.pdp_labels) {
			sb.append(l).append(", ");
			
			Measurement m_rank = this.policy_rank_analysis_map.get(l);
			if (m_rank == null) {
				sb.append("NaN, NaN, NaN, NaN");
			}
			else {
				sb.append(div_label(m_rank.getMean())).append(", ").
					append(div_label(m_rank.getMin())).append(", ").
					append(div_label(m_rank.getMax())).append(", ").
					append(div_label(m_rank.getStandardDeviation()));
			}
			sb.append(", ");
			
			Measurement m_mbrt = this.policy_mbrt_analysis_map.get(l);
			if (m_mbrt == null) {
				sb.append("NaN, NaN, NaN, NaN");
			}
			else {
				sb.append(div_label(m_mbrt.getMean())).append(", ").
					append(div_label(m_mbrt.getMin())).append(", ").
					append(div_label(m_mbrt.getMax())).append(", ").
					append(div_label(m_mbrt.getStandardDeviation()));
			}
			sb.append("\n");
		}
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized Iterator<String> getSortedPDPLabels() {
	
		return this.pdp_labels.iterator();
		
	}

	public synchronized float getBestRowAnalysisRatio() {
		
		return this.best_row_analysis_ratio;
		
	}

	/** Returns rank analysis measurement for target policy.
	 * @param l target policy
	 * @return rank analysis measurement for target policy
	 * (can be null)
	 */
	public synchronized Measurement
		getPolicyRankAnalysisMeasurement(String l)
		throws GridException {

		if (l == null) {
			throw new GridException("getPolicyRankAnalysisMeasurement(): "+
									"illegal policy label");
		}
		
		Measurement m = this.policy_rank_analysis_map.get(l);

		return m;
		
	}

	/** Returns MBRT analysis measurement for target policy.
	 * @param l target policy
	 * @return MBRT analysis measurement for target policy
	 * (can be null)
	 */
	public synchronized Measurement
		getPolicyMBRTAnalysisMeasurement(String l)
		throws GridException {
		
		if (l == null) {
			throw new GridException("getPolicyMBRTAnalysisMeasurement(): "+
									"illegal policy label");
		}

		Measurement m = this.policy_mbrt_analysis_map.get(l);
		
		return m;
		
	}

	// ****************************************************************
	
	public synchronized void analyze() throws GridException {
		
		int rank = 1;
		for (String l : this.ranked_policies) {
			addObservation(this.policy_rank_analysis_map,
						   l,rank);
			addObservation(this.policy_mbrt_analysis_map,
						   l,this.ranked_mbrt[rank-1]);
			if (rank >= this.analysis_count) {
				break;
			}
			else {
				++rank;
			}
		}
		
		if (policyAnalysisMapsInvariant() == false) {
			throw new GridException("analyze(): corrupted analysis");
		}
		
	}
	
	// ****************************************************************
	
	private synchronized void addObservation(Map<String,Measurement> map,
								String l,float val)
		throws GridException {
		
		Measurement m = map.get(l);
		if (isLegalPolicyLabel(l) == false) {
			throw new GridException("addObservation(): illegal policy label "+l+
									"for PDP "+this.pdp_label);
		}
		if (m == null) {
			m = MeasurementFactory.createSummarizedMeasurement(l);
			if (map.put(l,m) != null) {
				throw new GridException("addObservation(): corrupted data");
			}
		}
		m.addObservation(val);
		
	}
	
	private synchronized boolean isLegalPolicyLabel(String l) {
		
		if (l == null) {
			return false;
		}
		if (this.pdp_labels.contains(l) == false) {
			return false;
		}
		
		return true;
		
	}
	
	// ****************************************************************

	private synchronized boolean policyAnalysisMapsInvariant()
		throws GridException {
		
		if ((this.policy_rank_analysis_map.size() != this.pdp_labels.size()) ||
			(this.policy_mbrt_analysis_map.size() != this.pdp_labels.size())) {
			if (this.policy_rank_analysis_map.size() !=
				this.policy_mbrt_analysis_map.size()) {
				return false;
			}
			if (this.best_row_analysis_ratio == 1.0f) {
				return false;
			}
			if (this.policy_rank_analysis_map.size() == 0) {
				return false;
			}
		}
		
		int total_rank_observations = 0;
		int total_mbrt_observations = 0;
		for (String l : this.pdp_labels) {
			final Measurement m_rank = this.policy_rank_analysis_map.get(l);
			final Measurement m_mbrt = this.policy_mbrt_analysis_map.get(l);

			if ((m_rank == null) && (m_mbrt == null)) {
				if (best_row_analysis_ratio == 1.0f) {
					return false;
				}
				else {
					continue;
				}
			}
			else if (((m_rank == null) && (m_mbrt != null)) ||
			    	 ((m_rank != null) && (m_mbrt == null))) {
				return false;
			}
			else /* if ((m_rank != null) && (m_mbrt != null)) */ {
				total_rank_observations += m_rank.getN();
				total_mbrt_observations += m_mbrt.getN();
			}
		}
		
		if ((total_rank_observations != this.analysis_count) ||
			(total_mbrt_observations != this.analysis_count)) {
			return false;
		}
		
		return true;
		
	}
	
	// ****************************************************************

	private final String      pdp_label;
	private final Set<String> pdp_labels;
	private final String[]    ranked_policies;
	private final Float[]     ranked_mbrt;
	private final float       best_row_analysis_ratio;
	private final int         analysis_count;
	private final Map<String,Measurement> policy_rank_analysis_map;
	private final Map<String,Measurement> policy_mbrt_analysis_map;
	
}

