/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PolicyPerformanceGroupAnalyzer.java

package lbg.simulator.tools.aggregator;

import lbg.common.exception.*;
import lbg.common.math.*;
import static lbg.common.math.CommonMath.*;
import static lbg.common.io.CommonStringHandling.DUMMY_STRING;

import static java.lang.Math.round;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * @author Cyril Briquet
 */
public class PolicyPerformanceGroupAnalyzer {
	
	/** Constructs a new PolicyPerformanceGroupAnalyzer.
	 */
	public PolicyPerformanceGroupAnalyzer(String pdp_label,
										  String[] ranked_policies,
										  Float[] ranked_mbrt)
		throws GridException {
		
		this(pdp_label,ranked_policies,ranked_mbrt,1.0f);
		
	}
	
	/** Constructs a new PolicyPerformanceGroupAnalyzer.
	 */
	public PolicyPerformanceGroupAnalyzer(String pdp_label,
										  String[] ranked_policies,
										  Float[] ranked_mbrt,
										  float... best_row_analysis_ratios)
		throws GridException {

		if (pdp_label == null) {
			throw new GridException("PolicyPerformanceGroupAnalyzer(): "+
									"illegal PDP label");
		}
		if (ranked_policies == null) {
			throw new GridException("PolicyPerformanceGroupAnalyzer(): "+
									"illegal ranked policies");
		}
		if (ranked_mbrt == null) {
			throw new GridException("PolicyPerformanceGroupAnalyzer(): "+
									"illegal ranked MBRTs");
		}
		if ((best_row_analysis_ratios == null) ||
			(best_row_analysis_ratios.length < 1)) {
			throw new GridException("PolicyPerformanceGroupAnalyzer(): "+
									"illegal best row analysis ratios");
		}
		
		this.pdp_label = pdp_label;
		this.ranked_policies = ranked_policies;
		this.ranked_mbrt = ranked_mbrt;
		this.best_row_analysis_ratios = best_row_analysis_ratios;
		this.ppa_db = new ArrayList<PolicyPerformanceAnalyzer>();
		
	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PolicyPerformanceGroupAnalyzer");
		
		return sb.toString();
		
	}

	public synchronized String outputAsCSV() throws GridException {
		
		StringBuilder sb = new StringBuilder();
		
		for (PolicyPerformanceAnalyzer ppa : this.ppa_db) {
			sb.append(ppa.outputAsCSV()).append("\n");
		}
		
		return sb.toString();

	}

	public synchronized String outputAsCSVSummary() throws GridException {
		
		StringBuilder sb = new StringBuilder();
		
		if (this.ppa_db.size() == 0) {
			return DUMMY_STRING;
		}
		
		// MBRT graph
		sb.append("MBRT");
		Iterator<String> sorted_pdp_labels =
			this.ppa_db.get(0).getSortedPDPLabels();
		while (sorted_pdp_labels.hasNext()) {
			String l = sorted_pdp_labels.next();
			sb.append(", ").append(l);
		}
		sb.append("\n");
		
		for (PolicyPerformanceAnalyzer ppa : this.ppa_db) {
			sb.append("'").
				append(div_label(round(ppa.getBestRowAnalysisRatio()*100.0f))).
				append("%");
			sorted_pdp_labels = ppa.getSortedPDPLabels();
			while (sorted_pdp_labels.hasNext()) {
				String l = sorted_pdp_labels.next();
				Measurement m_mbrt = ppa.getPolicyMBRTAnalysisMeasurement(l);
				sb.append(", ").
					append((m_mbrt != null) ? m_mbrt.getMean() : DUMMY_STRING);
			}
			sb.append("\n");
		}
		sb.append("\n");

		// rank graph
		sb.append("rank");
		sorted_pdp_labels = this.ppa_db.get(0).getSortedPDPLabels();
		while (sorted_pdp_labels.hasNext()) {
			String l = sorted_pdp_labels.next();
			sb.append(", ").append(l);
		}
		sb.append("\n");
		
		for (PolicyPerformanceAnalyzer ppa : this.ppa_db) {
			sb.append("'").
				append(div_label(round(ppa.getBestRowAnalysisRatio()*100.0f))).
				append("%");
			sorted_pdp_labels = ppa.getSortedPDPLabels();
			while (sorted_pdp_labels.hasNext()) {
				String l = sorted_pdp_labels.next();
				Measurement m_rank = ppa.getPolicyRankAnalysisMeasurement(l);
				sb.append(", ").
					append((m_rank != null) ? m_rank.getMean() : DUMMY_STRING);
			}
			sb.append("\n");
		}
		sb.append("\n");
		
		return sb.toString();

	}

	// ****************************************************************

	public synchronized void analyze() throws GridException {
		
		for (float best_row_analysis_ratio : this.best_row_analysis_ratios) {
			PolicyPerformanceAnalyzer ppa =
				new PolicyPerformanceAnalyzer(this.pdp_label,
											  this.ranked_policies,
											  this.ranked_mbrt,
											  best_row_analysis_ratio);
			ppa.analyze();
			this.ppa_db.add(ppa);
		}
		
	}

	// ****************************************************************

	private final String pdp_label;
	private final String[] ranked_policies;
	private final Float[] ranked_mbrt;
	private final float[] best_row_analysis_ratios;
	private final ArrayList<PolicyPerformanceAnalyzer> ppa_db;
	
}

