/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// StrategyPerformanceAnalyzer.java

package lbg.simulator.tools.aggregator;

import lbg.common.exception.*;
import lbg.common.math.*;
import lbg.common.io.*;
import static lbg.common.io.CommonFileHandling.CSV_EXT;
import lbg.peer.strategy.negotiation.*;
import lbg.peer.strategy.scheduling.*;
import lbg.simulator.tools.spreadsheet.*;

import java.util.ArrayList;
import java.util.Iterator;

/** A StrategyPerformanceAnalyzer analyzes the performance of
 * the bartering policies stored in a target spreadsheet
 * generated by the LogAggregator.
 * @author Cyril Briquet
 */
public class StrategyPerformanceAnalyzer {

	/** Constructs a new StrategyPerformanceAnalyzer.
	 */
	public StrategyPerformanceAnalyzer(String timestamp,String spreadsheet_fn)
		throws GridException {

		if (timestamp == null) {
			throw new GridException("StrategyPerformanceAnalyzer() "+
									"illegal timestamp");
		}
		if (spreadsheet_fn == null) {
			throw new GridException("StrategyPerformanceAnalyzer() "+
									"illegal spreadsheet filename");
		}

		this.timestamp = timestamp;
		this.spreadsheet_fn = spreadsheet_fn;
		this.ppga_db = new ArrayList<PolicyPerformanceGroupAnalyzer>();

	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[StrategyPerformanceAnalyzer]");
		
		return sb.toString();
		
	}

	// ****************************************************************
	
	public synchronized void analyzeStrategyPerformance()
		throws GridException {
		
		LogSpreadsheet spreadsheet = new LogSpreadsheet(this.spreadsheet_fn);

		Float[] ranked_mbrt = LogSpreadsheetSorter.
			getSortedDataColumn(spreadsheet,"grid.job_completion_time.mean");
		
		analyzePolicyPerformance(spreadsheet,ranked_mbrt,
								 LocalSchedulingType.getPDPLabel());
		analyzePolicyPerformance(spreadsheet,ranked_mbrt,
								 ConsumptionSchedulingType.getPDPLabel());
		analyzePolicyPerformance(spreadsheet,ranked_mbrt,
								 SupplyingSchedulingType.getPDPLabel());
		
		analyzePolicyPerformance(spreadsheet,ranked_mbrt,
								 SupplyingFilteringType.getPDPLabel());
		analyzePolicyPerformance(spreadsheet,ranked_mbrt,
								 SupplyingRunningPreemptionType.getPDPLabel());
		analyzePolicyPerformance(spreadsheet,ranked_mbrt,
								 SupplyingWaitingPreemptionType.getPDPLabel());
		
		analyzePolicyPerformance(spreadsheet,ranked_mbrt,
								 ConsumptionNegotiationType.getPDPLabel());
		analyzePolicyPerformance(spreadsheet,ranked_mbrt,
								 SupplyingNegotiationType.getPDPLabel());
		
	}
	
	public synchronized void saveStrategyPerformanceAnalysisReport()
		throws GridException {
		
		StringBuilder sb = new StringBuilder();
		
		Iterator<PolicyPerformanceGroupAnalyzer> it = this.ppga_db.iterator();
		while (it.hasNext()) {
			PolicyPerformanceGroupAnalyzer ppga = it.next();
			sb.append(ppga.outputAsCSV()).append("\n");
		}
		
		sb.append("'================\n\n");
		
		it = this.ppga_db.iterator();
		while (it.hasNext()) {
			PolicyPerformanceGroupAnalyzer ppga = it.next();
			sb.append(ppga.outputAsCSVSummary()).append("\n");
		}
		
		String report_filename =
			"log_aggregator_strat_perf_rank_"+timestamp+CSV_EXT;
		
		CommonFileHandling.writeFile(report_filename,sb.toString());
		
		System.out.println("Wrote strategy performance analysis report to "+
						   report_filename);
		
	}

	// ****************************************************************

	public static void main(String[] args) {

		try {
			checkArgs(args);

			String timestamp = CommonStringHandling.generateTimeStamp();
			String spreadsheet_fn = args[0];

			StrategyPerformanceAnalyzer spa =
				new StrategyPerformanceAnalyzer(timestamp,spreadsheet_fn);
			
			spa.analyzeStrategyPerformance();
			spa.saveStrategyPerformanceAnalysisReport();
		}
		catch (GridException e) {
			e.printStackTrace();
		}
		catch (Throwable thr) {
			thr.printStackTrace();
		}
		finally {
			System.exit(0);
		}
		
	}
	
	private static void checkArgs(String[] args) throws GridException {
		
		boolean ok = true;

		if (ok) {
			if (args.length != 1) {
				ok = false;
			}
		}
		
		if (ok) {
			if ((args[0] == null) || (args[0].length() < 1+CSV_EXT.length())) {
				ok = false;
			}
		}
		
		if (ok == false) {
			System.out.println("Usage: "+
							   "java -cp classpath "+
							   StrategyPerformanceAnalyzer.class.getName()+
							   " log_aggregator_report_*.csv\n");
			throw new GridException("checkArgs(): illegal arguments");
		}
		
	}
	
	// ****************************************************************

	private synchronized void
		analyzePolicyPerformance(LogSpreadsheet spreadsheet,Float[] ranked_mbrt,
								 String pdp_label)
		throws GridException {
		
		if (spreadsheet == null) {
			throw new GridException("analyzePolicyPerformance(): "+
									"illegal log spreadsheet");
		}
		if (ranked_mbrt == null) {
			throw new GridException("analyzePolicyPerformance(): "+
									"illegal ranked MBRTs");
		}
		if (pdp_label == null) {
			throw new GridException("analyzePolicyPerformance(): "+
									"illegal PDP label");
		}

		String[] ranked_policies = LogSpreadsheetSorter.
			getSortedStrategyColumn(spreadsheet,
									"grid.job_completion_time.mean",
									pdp_label);

		PolicyPerformanceGroupAnalyzer ppga =
				new PolicyPerformanceGroupAnalyzer(pdp_label,ranked_policies,
												   ranked_mbrt,
												   best_row_analysis_ratios);
		ppga.analyze();
		this.ppga_db.add(ppga);
		
	}

	private static final float[] best_row_analysis_ratios =
		{ 1.0f, 0.8f, 0.2f, 0.05f, 0.01f };

	// ****************************************************************

	private final String timestamp;
	private final String spreadsheet_fn;
	private final ArrayList<PolicyPerformanceGroupAnalyzer> ppga_db;
	
}

