/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PolicyEnumerator.java

package lbg.simulator.tools.generator;

import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonFileHandling.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.peer.config.*;
import lbg.peer.strategy.accounting.*;
import lbg.peer.strategy.negotiation.*;
import lbg.peer.strategy.scheduling.*;
import lbg.simulator.*;
import lbg.simulator.config.*;
import lbg.simulator.tools.jdf.*;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;

/** Given an incomplete Scenario file (i.e. not describing Peers strategy),
 * a PolicyEnumerator enumerates all combinations
 * of Accountant, Scheduler and Negotiator,
 * and generates a scenario for each combination.
 *
 * <p>The generated scenarios are written to disk,
 * as a well as a Job Description File to run them.
 *
 * @author Cyril Briquet
 */ 
public class PolicyEnumerator extends ScenarioSetGenerator {

	public PolicyEnumerator(String[] args) throws GridException {

		super(args);
		
		String scenario_complete_filename = args[0];
		String jar_complete_filename = args[1];
		
		if ((scenario_complete_filename == null) ||
			(scenario_complete_filename.length() < 1+TXT_EXT.length())) {
			throw new GridException("PolicyEnumerator(): "+
									"illegal Scenario filename");
		}
		if ((jar_complete_filename == null) ||
			(jar_complete_filename.length() < 1+JAR_EXT.length())) {
			throw new GridException("PolicyEnumerator(): "+
									"illegal JAR filename");
		}
		
		this.scenario_complete_filename = scenario_complete_filename;
		this.scenario_filename =
			CommonFileHandling.
			extractFileNameFromPath(this.scenario_complete_filename);
		this.scenario_base_name =
			CommonFileHandling.
				extractBaseFileNameFromFileName(this.scenario_filename,TXT_EXT);

		this.jar_complete_filename = jar_complete_filename;
		/*this.jar_filename =
			CommonFileHandling.
				extractFileNameFromPath(this.jar_complete_filename);*/
		/*this.jar_base_name =
			CommonFileHandling.
				extractBaseFileNameFromFileName(this.jar_filename,JAR_EXT);*/

		this.main_class = (args.length > 5) ? args[5] : null;

		Properties scenario_props =
			CommonFileHandling.loadProperties(this.scenario_complete_filename);
		this.base_sc =
			new Scenario(scenario_props,this.scenario_complete_filename,
						 true,true,true);

		this.expected_sc_count = PeerStrategyConfiguration.getPolicyCount();
		
		this.gen_scenarii_filenames = new TreeSet<String>();

		Middleware m = Middleware.valueOf(args[2]);
		this.jdfwf = new JDFWriterFactory(m,false);
		this.bucket_count =
			(args.length > 3) ? loadInt(args[3]) : 1;
		if (this.bucket_count < 1) {
			throw new GridException("PolicyEnumerator(): illegal # jdf splits "+
									"("+this.bucket_count+")");
		}
		
		this.add_freerider = (args.length > 4) ? loadBoolean(args[4]) : false;

	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[PolicyEnumerator]");
		
		return sb.toString();
		
	}
	
	// ****************************************************************

	public int getExpectedScenariiCount() {
	
		return this.expected_sc_count;
		
	}
	
	public int getGeneratedScenariiCount() {
		
		return this.generated_sc_count;
		
	}
	

	public JDFWriterFactory getJDFWriterFactory() {
		
		return this.jdfwf;
		
	}
	
	public Scenario getBaseScenario() {
	
		return this.base_sc;
		
	}
	
	public String getScenarioCompleteFileName() {
		
		return this.scenario_complete_filename;
		
	}

	public String getScenarioFileName() {
		
		return this.scenario_filename;
		
	}

	public String getScenarioBaseName() {
		
		return this.scenario_base_name;
		
	}
	
	public String getJarCompleteFileName() {
		
		return this.jar_complete_filename;
		
	}

	public int getBucketCount() {
		
		return this.bucket_count;
		
	}
	
	public String getMainClass() {
	
		return this.main_class;
		
	}
	
	public Set<String> getGeneratedScenariiFileNames() {
		
		return this.gen_scenarii_filenames;
		
	}

	// ****************************************************************

	public void checkArgs(String[] args) throws GridException {
	
		boolean ok = true;

		if (ok) {
			if (args == null) {
				ok = false;
			}
		}
		
		if (ok) {
			if (((args.length < 3) || (args.length > 6)) ||
				(args[0] == null) ||
				(args[0].length() < 1+TXT_EXT.length()) ||
				(args[1] == null) ||
				(args[1].length() < 1+JAR_EXT.length()) ||
				((args.length == 4) && (args[3] == null)) ||
				((args.length == 5) &&
				 ((args[4] == null) || (args[3] == null))) ||
				((args.length == 6) &&
				 ((args[5] == null)||(args[4] == null)||(args[3] == null)))) {
				ok = false;
			}
		}
		
		if (ok) {
			try {
				Middleware.valueOf(args[2]);
			}
			catch (Exception e) {
				ok = false;
			}
		}

		if (ok == false) {
			System.out.println("Usage: "+
							   "java -cp classpath "+
							   PolicyEnumerator.class.getName()+
							   " scenario.txt lbg.jar "+
							   "LBGRID|OURGRID [# jdf splits] "+
							   "[add freerider boolean flag] [Main-Class]\n");
			throw new GridException("checkArgs(): illegal arguments");
		}
		
	}

	// ****************************************************************

	public synchronized void generateScenarios() throws GridException {

		BarteringAccountantType[] bat_values =
			BarteringAccountantType.values();

		ConsumptionNegotiationType[] cnt_values =
			ConsumptionNegotiationType.values();
		SupplyingNegotiationType[] snt_values =
			SupplyingNegotiationType.values();
		
		SupplyingRunningPreemptionType[] rpt_values =
			SupplyingRunningPreemptionType.values();
		SupplyingWaitingPreemptionType[] wpt_values =
			SupplyingWaitingPreemptionType.values();
		SupplyingFilteringType[] sft_values =
			SupplyingFilteringType.values();
		
		LocalSchedulingType[] lst_values =
			LocalSchedulingType.values();
		ConsumptionSchedulingType[] cst_values =
			ConsumptionSchedulingType.values();
		SupplyingSchedulingType[] sst_values =
			SupplyingSchedulingType.values();

		this.generated_sc_count = 0;
		for (BarteringAccountantType ba : bat_values) {
		for (ConsumptionNegotiationType cn : cnt_values) {
		for (SupplyingNegotiationType sn : snt_values) {
			for (LocalSchedulingType ls : lst_values) {
			for (SupplyingRunningPreemptionType rp : rpt_values) {
			for (SupplyingWaitingPreemptionType wp : wpt_values) {
			for (SupplyingFilteringType sf : sft_values) {
			for (ConsumptionSchedulingType cs : cst_values) {
			for (SupplyingSchedulingType ss : sst_values) {
				PolicyEnumeratorPayload sc_payload =
					new PolicyEnumeratorPayload(ba,cn,sn,ls,rp,wp,sf,cs,ss);
				if (canAddStrategyPolicies(sc_payload)) {
					generateScenario(++this.generated_sc_count,sc_payload);
			}
			}
			}
			}
			}
			}
			}
		}
		}
		}
		
		// ****************

		System.out.println("Generated "+this.generated_sc_count+
						   " scenarii (out of "+this.expected_sc_count+
						   " possible scenarii)");
		
	}
	
	// ****************************************************************

	protected void updateScenario(Scenario sc,ScenarioPayload sc_payload)
		throws GridException {

		if (sc == null) {
			throw new GridException("updateScenario(): illegal Scenario");
		}
		if (sc_payload == null) {
			throw new GridException("updateScenario(): "+
									"illegal Scenario payload");
		}
		
		Properties props = new Properties();
		
		GridStrategyConfiguration grid_strategy_config =
			sc.getGridStrategyConfiguration();
		addStrategyParameters(props,grid_strategy_config);
		
		GridConfiguration grid_config = sc.getGridConfiguration();
		addStrategyPolicies(props,grid_config,
							(PolicyEnumeratorPayload)sc_payload,
							this.add_freerider);
		
		sc.loadGridStrategyConfiguration(props);
		
	}

	// ****************************************************************

	public static void main(String[] args) {

		try {
			ScenarioSetGenerator.run(new PolicyEnumerator(args));
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		finally {
			System.exit(0);
		}
		
	}

	// ****************************************************************

	private static boolean
		canAddStrategyPolicies(PolicyEnumeratorPayload sc_payload)
		throws GridException {
		
		if (sc_payload == null) {
			throw new GridException("canAddStrategyPolicies(): "+
									"illegal Scenario payload");
		}

		BarteringAccountantType ba =
			sc_payload.getBarteringAccountantType();
		ConsumptionNegotiationType cn =
			sc_payload.getConsumptionNegotiationType();
		SupplyingNegotiationType sn =
			sc_payload.getSupplyingNegotiationType();
		LocalSchedulingType ls =
			sc_payload.getLocalSchedulingType();
		SupplyingRunningPreemptionType rp =
			sc_payload.getSupplyingRunningPreemptionType();
		SupplyingWaitingPreemptionType wp =
			sc_payload.getSupplyingWaitingPreemptionType();
		SupplyingFilteringType sf =
			sc_payload.getSupplyingFilteringType();
		ConsumptionSchedulingType cs =
			sc_payload.getConsumptionSchedulingType();
		SupplyingSchedulingType ss =
			sc_payload.getSupplyingSchedulingType();
		
		if ((ba == null) || (cn == null) || (sn == null) || (ls == null) ||
			(rp == null) || (wp == null) || (sf == null) || (cs == null) ||
			(ss == null)) {
			throw new GridException("canAddStrategyPolicies(): "+
									"illegal parameter");
		}
		
		// Global strategy is either Bartering-based or purely local
		if (sn.isBarteringStrategy() || cn.isBarteringStrategy() ||
			ss.isBarteringStrategy() || cs.isBarteringStrategy() ||
			ba.isBarteringStrategy() || sf.isBarteringStrategy()) {
			if (!sn.isBarteringStrategy() || !cn.isBarteringStrategy() ||
				!ss.isBarteringStrategy() || !cs.isBarteringStrategy() ||
				!ba.isBarteringStrategy() || !sf.isBarteringStrategy()) {
				return false;
			}
		}
		
		// If Local scheduler does not preempt,
		// do not enumerate Preemption policies
		if ((ls.isPreemptionStrategy() == false) &&
			((rp.isPreemptionStrategy() == true) ||
			 (wp.isPreemptionStrategy() == true))) {
			return false;
		}

		// If Local scheduler does preempt,
		// do not enumerate policies that totally do not preempt
		if ((ls.isPreemptionStrategy() == true) &&
			((rp.isPreemptionStrategy() == false) &&
			 (wp.isPreemptionStrategy() == false))) {
			return false;
		}

		/*
		if (((ba == BarteringAccountantType.NoEvalAccountant) ||
			 (ba == BarteringAccountantType.LBGRelativePowerAccountant)) ==
			false) {
			return false;
		}
		*/
		
		return true;

	}
	
	private static void
		addStrategyPolicies(Properties props,GridConfiguration grid_config,
							PolicyEnumeratorPayload sc_payload,
							boolean add_freerider)
		throws GridException {
		
		if (props == null) {
			throw new GridException("addStrategyPolicies(): "+
									"illegal Properties");
		}
		if (grid_config == null) {
			throw new GridException("addStrategyPolicies(): "+
									"illegal Grid configuration");
		}
		if (sc_payload == null) {
			throw new GridException("addStrategyPolicies(): "+
									"illegal Scenario payload");
		}

		BarteringAccountantType ba =
			sc_payload.getBarteringAccountantType();
		ConsumptionNegotiationType cn =
			sc_payload.getConsumptionNegotiationType();
		SupplyingNegotiationType sn =
			sc_payload.getSupplyingNegotiationType();
		LocalSchedulingType ls =
			sc_payload.getLocalSchedulingType();
		SupplyingRunningPreemptionType rp =
			sc_payload.getSupplyingRunningPreemptionType();
		SupplyingWaitingPreemptionType wp =
			sc_payload.getSupplyingWaitingPreemptionType();
		SupplyingFilteringType sf =
			sc_payload.getSupplyingFilteringType();
		ConsumptionSchedulingType cs =
			sc_payload.getConsumptionSchedulingType();
		SupplyingSchedulingType ss =
			sc_payload.getSupplyingSchedulingType();
		
		if ((ba == null) || (cn == null) || (sn == null) || (ls == null) ||
			(rp == null) || (wp == null) || (sf == null) || (cs == null) ||
			(ss == null)) {
			throw new GridException("addStrategyPolicies(): "+
									"illegal parameter");
		}
		
		int group_count = grid_config.PEER_GROUPS();

		setGroupProperty(props,BarteringAccountantType.getPDPLabel(),
						 ba.name(),group_count);
		setGroupProperty(props,SupplyingNegotiationType.getPDPLabel(),
						 sn.name(),group_count);
		setGroupProperty(props,ConsumptionNegotiationType.getPDPLabel(),
						 cn.name(),group_count);
		setGroupProperty(props,LocalSchedulingType.getPDPLabel(),
						 ls.name(),group_count);
		setGroupProperty(props,SupplyingRunningPreemptionType.getPDPLabel(),
						 rp.name(),group_count);
		setGroupProperty(props,SupplyingWaitingPreemptionType.getPDPLabel(),
						 wp.name(),group_count);
		setGroupProperty(props,SupplyingFilteringType.getPDPLabel(),
						 sf.name(),group_count);
		setGroupProperty(props,SupplyingSchedulingType.getPDPLabel(),
						 ss.name(),group_count);
		setGroupProperty(props,ConsumptionSchedulingType.getPDPLabel(),
						 cs.name(),group_count);
		
		if (add_freerider) {
			/* Freerider policies:
				PEER_PREEMPTION_SUPPLYING_RUNNING_POLICY = { FullCancellation }
				PEER_PREEMPTION_SUPPLYING_WAITING_POLICY =
				  { FullWaitingPreemption }
				PEER_FILTERING_SUPPLYING_POLICY   = { NoSupplyingFiltering }
				PEER_SCHEDULING_SUPPLYING_POLICY  = { NoSupplyingScheduling }
				PEER_NEGOTIATION_SUPPLYING_POLICY = {  NoSupplyingNegotiation }
			*/
			setGroupProperty(props,SupplyingNegotiationType.getPDPLabel(),
							 sn.name(),"NoSupplyingNegotiation",group_count);
			setGroupProperty(props,SupplyingRunningPreemptionType.getPDPLabel(),
							 rp.name(),"FullCancellation",group_count);
			setGroupProperty(props,SupplyingWaitingPreemptionType.getPDPLabel(),
							 wp.name(),"FullWaitingPreemption",group_count);
			setGroupProperty(props,SupplyingFilteringType.getPDPLabel(),
							 sf.name(),"NoSupplyingFiltering",group_count);
			setGroupProperty(props,SupplyingSchedulingType.getPDPLabel(),
							 ss.name(),"NoSupplyingScheduling",group_count);
		}

	}

	private static void
		addStrategyParameters(Properties props,
							  GridStrategyConfiguration grid_strategy_config)
		throws GridException {
		
		if (props == null) {
			throw new GridException("addStrategyParameters(): "+
									"illegal Properties");
		}
		if (grid_strategy_config == null) {
			throw new GridException("addStrategyParameters(): "+
									"illegal Grid strategy configuration");
		}
		
		setGroupProperty(props,"REQUESTORS_COUNT_THRESHOLD",
						 grid_strategy_config.REQUESTORS_COUNT_THRESHOLD());
		setGroupProperty(props,"REQUESTORS_TIME_THRESHOLD",
						 grid_strategy_config.REQUESTORS_TIME_THRESHOLD());
		
		setGroupProperty(props,"GRANTORS_COUNT_THRESHOLD",
						 grid_strategy_config.GRANTORS_COUNT_THRESHOLD());
		setGroupProperty(props,"GRANTORS_TIME_THRESHOLD",
						 grid_strategy_config.GRANTORS_TIME_THRESHOLD());
		
		setGroupProperty(props,"REQUEST_SUPPLYING_TIME_THRESHOLD",
						 grid_strategy_config.
						 	REQUEST_SUPPLYING_TIME_THRESHOLD());
		
		setGroupProperty(props,"LOCAL_Q_LEN_PREEMPTION_THRESHOLD",
						 grid_strategy_config.
						 	LOCAL_Q_LEN_PREEMPTION_THRESHOLD());
		setGroupProperty(props,"PEER_Q_FILTERING_THRESHOLD",
						 grid_strategy_config.PEER_Q_FILTERING_THRESHOLD());
		
		setGroupProperty(props,"CONSUMPTION_TIMEOUT",
						 grid_strategy_config.CONSUMPTION_TIMEOUT());
		
		setGroupProperty(props,"PEER_PSUFFERAGE",
						 grid_strategy_config.PEER_PSUFFERAGE());

		setGroupProperty(props,"DO_CONSUMPTION_BLACKLIST",
						 grid_strategy_config.DO_CONSUMPTION_BLACKLIST());
		setGroupProperty(props,"CONSUMPTION_BLACKLIST_PROBA",
						 grid_strategy_config.CONSUMPTION_BLACKLIST_PROBA());

	}

	// ****************************************************************

	private final String  scenario_complete_filename;
	private final String  scenario_filename;
	private final String  scenario_base_name;
	private final String  jar_complete_filename;
	private final String  main_class;
	private final boolean add_freerider;
	
	private final Scenario base_sc;
	private final int expected_sc_count;
	private       int generated_sc_count;
	private final Set<String> gen_scenarii_filenames;

	private final JDFWriterFactory jdfwf;
	private	final int bucket_count;

}

