/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ScenarioRandomizer.java

package lbg.simulator.tools.generator;

import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.common.math.*;
import static lbg.common.io.CommonFileHandling.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.simulator.*;
import lbg.simulator.tools.jdf.*;

import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;

/** Given an arbitrary Scenario file,
 * a ScenarioRandomizer generates ranomized versions of the Scenario.
 * The randomization is implemented by a random SIMULATION_SEED field.
 *
 * <p>The generated scenarios are written to disk,
 * as a well as a Job Description File to run them.
 *
 * @author Cyril Briquet
 */
public class ScenarioRandomizer extends ScenarioSetGenerator {

	public ScenarioRandomizer(String[] args) throws GridException {

		super(args);

		String scenario_complete_filename = args[0];
		String jar_complete_filename = args[1];
		int ncopy = Integer.parseInt(args[2]);

		if ((scenario_complete_filename == null) ||
			(scenario_complete_filename.length() < 1+TXT_EXT.length())) {
			throw new GridException("ScenarioRandomizer(): "+
									"illegal Scenario filename");
		}
		if ((jar_complete_filename == null) ||
			(jar_complete_filename.length() < 1+JAR_EXT.length())) {
			throw new GridException("ScenarioRandomizer(): "+
									"illegal JAR filename");
		}
		if (ncopy < 1) {
			throw new GridException("ScenarioRandomizer(): "+
									"illegal number of copies");
		}
		
		this.scenario_complete_filename = scenario_complete_filename;
		this.scenario_filename =
			CommonFileHandling.
			extractFileNameFromPath(this.scenario_complete_filename);
		this.scenario_base_name =
			CommonFileHandling.
				extractBaseFileNameFromFileName(this.scenario_filename,TXT_EXT);

		this.jar_complete_filename = jar_complete_filename;
		/*this.jar_filename =
			CommonFileHandling.
				extractFileNameFromPath(this.jar_complete_filename);*/
		/*this.jar_base_name =
			CommonFileHandling.
				extractBaseFileNameFromFileName(this.jar_filename,JAR_EXT);*/

		this.main_class = (args.length == 5) ? args[4] : null;
		this.ncopy = ncopy;
		
		Properties scenario_props =
			CommonFileHandling.loadProperties(this.scenario_complete_filename);
		this.base_sc =
			new Scenario(scenario_props,this.scenario_complete_filename);
		CommonMath.setSeed(this.base_sc.getRandomizerConfiguration().
							SIMULATION_SEED());
		
		this.gen_scenarii_filenames = new TreeSet<String>();
		
		Middleware m = Middleware.valueOf(args[3]);
		this.jdfwf = new JDFWriterFactory(m,false);
		this.bucket_count = 1;

	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[ScenarioRandomizer]");
		
		return sb.toString();
		
	}
	
	// ****************************************************************

	public int getExpectedScenariiCount() {
	
		return this.ncopy;
		
	}

	public int getGeneratedScenariiCount() {
	
		return getExpectedScenariiCount();
		
	}
	
	public JDFWriterFactory getJDFWriterFactory() {
		
		return this.jdfwf;
		
	}

	public Scenario getBaseScenario() {
	
		return this.base_sc;
		
	}

	public String getScenarioCompleteFileName() {
		
		return this.scenario_complete_filename;
		
	}

	public String getScenarioFileName() {
		
		return this.scenario_filename;
		
	}

	public String getScenarioBaseName() {
		
		return this.scenario_base_name;
		
	}

	public String getJarCompleteFileName() {
		
		return this.jar_complete_filename;
		
	}

	public int getBucketCount() {
		
		return this.bucket_count;
		
	}
	
	public String getMainClass() {
	
		return this.main_class;
		
	}

	public Set<String> getGeneratedScenariiFileNames() {
		
		return this.gen_scenarii_filenames;
		
	}

	// ****************************************************************

	public void checkArgs(String[] args) throws GridException {
	
		boolean ok = true;

		if (ok) {
			if (args == null) {
				ok = false;
			}
		}

		if (ok) {
			if (((args.length < 4) || (args.length > 5)) ||
				(args[0] == null) ||
				(args[0].length() < 1+TXT_EXT.length()) ||
				(args[1] == null) ||
				(args[1].length() < 1+JAR_EXT.length()) ||
				(args[2] == null) ||
				(args[2].length() < 1) ||
				((args.length == 5) && (args[4] == null))) {
				ok = false;
			}
		}
		
		if (ok) {
			try {
				Integer.parseInt(args[2]);
			}
			catch (NumberFormatException e) {
				throw new GridException("main(): illegal number of copies");
			}
		}
		
		if (ok) {
			try {
				Middleware.valueOf(args[3]);
			}
			catch (Exception e) {
				ok = false;
			}
		}
		
		if (ok == false) {
			System.out.println("Usage: "+
							   "java -cp classpath "+
							   ScenarioRandomizer.class.getName()+
							   " scenario.txt lbg.jar "+
							   "ncopy LBGRID|OURGRID [Main-Class]\n");
			throw new GridException("checkArgs(): illegal arguments");
		}
		
	}

	// ****************************************************************

	public synchronized void generateScenarios() throws GridException {

		for (int idx = 0 ; idx < this.ncopy ; ++idx) {

			ScenarioPayload sc_payload =
				new ScenarioRandomizerPayload(
					Integer.toString(CommonMath.sampleUniformDist()));

			generateScenario(idx,sc_payload);
		}
	
		System.out.println("Generated "+this.ncopy+" scenarii");
	
	}

	protected void updateScenario(Scenario sc,ScenarioPayload sc_payload)
		throws GridException {
		
		if (sc == null) {
			throw new GridException("updateScenario(): illegal Scenario");
		}
		if (sc_payload == null) {
			throw new GridException("updateScenario(): "+
									"illegal Scenario payload");
		}
		
		Properties props = new Properties();
		String sim_seed = 
			((ScenarioRandomizerPayload)sc_payload).getSimulationSeed();
		setProperty(props,"SIMULATION_SEED",sim_seed);
		sc.loadRandomizerConfiguration(props);
		
	}
	
	// ****************************************************************

	public static void main(String[] args) {

		try {
			ScenarioSetGenerator.run(new ScenarioRandomizer(args));
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		finally {
			System.exit(0);
		}
		
	}
	
	// ****************************************************************

	private final String scenario_complete_filename;
	private final String scenario_filename;
	private final String scenario_base_name;
	private final String jar_complete_filename;
	private final String main_class;
	private final int ncopy;
	
	private final Scenario base_sc;
	private final Set<String> gen_scenarii_filenames;
	private final JDFWriterFactory jdfwf;
	private	final int bucket_count;

}

