/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ScenarioSetGenerator.java

package lbg.simulator.tools.generator;

import lbg.*;
import lbg.common.exception.*;
import static lbg.common.io.CommonFileHandling.*;
import static lbg.common.io.CommonStringHandling.*;
import lbg.simulator.*;
import lbg.simulator.tools.jdf.*;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

/**
 * @author Cyril Briquet
 */
public abstract class ScenarioSetGenerator {

	protected ScenarioSetGenerator(String[] args) throws GridException {
		
		LightweightBarteringGrid.setMode(VirtualizationLevel.SIM_MODE);
		checkArgs(args);

	}
	
	// ****************************************************************

	public abstract int getExpectedScenariiCount();
	
	public abstract int getGeneratedScenariiCount();

	public abstract JDFWriterFactory getJDFWriterFactory();

	public abstract Scenario getBaseScenario();
	
	public abstract String getScenarioCompleteFileName();

	public abstract String getScenarioFileName();

	public abstract String getScenarioBaseName();

	public abstract String getJarCompleteFileName();

	public abstract int getBucketCount();

	public abstract String getMainClass();
	
	public abstract Set<String> getGeneratedScenariiFileNames();
	
	// ****************************************************************

	public abstract void checkArgs(String[] args) throws GridException;

	/** Enumerates scenarios. For each scenario, calls generateScenario()
	 * and passes the scenario index and payload.
	 */
	public abstract void generateScenarios() throws GridException;
	
	public synchronized void generateJDF() throws GridException {
		
		final ArrayList<Set<String>> sc_buckets =
			generateScenariiBuckets(getGeneratedScenariiFileNames().iterator(),
									getGeneratedScenariiCount(),
									getBucketCount());
		
		for (int b = 0 ; b < getBucketCount() ; ++b) {
			generateBucketJDF(b,sc_buckets);
		}
		
	}

	// ****************************************************************

	protected void generateScenario(int idx,ScenarioPayload sc_payload)
		throws GridException {
		
		String scenario_filename = generateScenarioFileName(idx);
		Scenario sc = (Scenario)getBaseScenario().clone();

		updateScenario(sc,sc_payload);
		storeScenario(sc,scenario_filename);

		Set<String> gen_scenarii_filenames = getGeneratedScenariiFileNames();
		synchronized(gen_scenarii_filenames) {
			if (gen_scenarii_filenames.add(scenario_filename) == false) {
				throw new GridException("generateScenario(): "+
										"corrupted scenarii filenames DB, "+
										"scenario "+scenario_filename+
										" already generated");
			}
		}
		
	}

	/** Updates target scenario with target payload.
	 * For maximum efficiency, should not be synchronized.
     */
	protected abstract void updateScenario(Scenario sc,
										   ScenarioPayload sc_payload)
		throws GridException;

	// ****************************************************************

	/**
	 * @param idx Scenario index, in [0..expected scenarii-1]
	 */
	protected String generateScenarioFileName(int idx) throws GridException {

		if ((idx < 0) || (idx >= getExpectedScenariiCount())) {
			throw new GridException("generateScenarioFileName(): "+
									"illegal Scenario index "+idx+
									" ("+getExpectedScenariiCount()+
									" expected)");
		}
		
		StringBuilder sb_gen_scenario_filename = new StringBuilder();
		sb_gen_scenario_filename.append(getScenarioBaseName());
		sb_gen_scenario_filename.append("-");
		sb_gen_scenario_filename.
			append(fillWithHeadingZerosLimit(idx,getExpectedScenariiCount()));
		sb_gen_scenario_filename.append(TXT_EXT);

		return sb_gen_scenario_filename.toString();

	}

	// ****************************************************************

	protected void
		generateBucketJDF(int b,ArrayList<Set<String>> sc_buckets)
		throws GridException {
		
		if ((b < 0) || (b > getBucketCount())) {
			throw new GridException("generateBucketJDF(): "+
									"illegal bucket index");
		}
		if ((sc_buckets == null) || (sc_buckets.size() < getBucketCount())) {
			throw new GridException("generateBucketJDF(): "+
									"illegal scenarii filenames bucket");
		}

		String b_suffix = fillWithHeadingZerosLimit(b+1,getBucketCount());
		
		StringBuilder sb_job_label = new StringBuilder();
		sb_job_label.
			append(extractBaseFileNameFromPath(getScenarioFileName(),TXT_EXT));
		if (getBucketCount() > 1) {
			sb_job_label.append(b_suffix);
		}
		sb_job_label.append(TXT_EXT);
		
		StringBuilder sb_scenario_complete_filename = new StringBuilder();
		sb_scenario_complete_filename.
			append(extractBaseFileNameFromPath(getScenarioCompleteFileName(),
											   TXT_EXT));
		if (getBucketCount() > 1) {
			sb_scenario_complete_filename.append(b_suffix);
		}
		sb_scenario_complete_filename.append(TXT_EXT);

		Iterator<String> it_sc_bucket = sc_buckets.get(b).iterator();

		JDFWriter jdfw =
			getJDFWriterFactory().
				createJDFWriter(sb_job_label.toString(),
								sb_scenario_complete_filename.toString(),
								getJarCompleteFileName(),
								it_sc_bucket,getMainClass());
		jdfw.writeJDF();

		
	}

	// ****************************************************************

	public static void run(ScenarioSetGenerator sse) throws GridException {

		if (sse == null) {
			throw new GridException("run(): illegal ScenarioSetGenerator");
		}
		sse.generateScenarios();
		sse.generateJDF();
		
	}
	
	// ****************************************************************

	protected static void storeScenario(Scenario sc,String scenario_filename)
		throws GridException {

		if (sc == null) {
			throw new GridException("generateScenario(): "+
									"illegal Scenario");
		}
		if (scenario_filename == null) {
			throw new GridException("generateScenario(): "+
									"illegal Scenario filename");
		}
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("# ").append(scenario_filename).append("\n").
			append(doNotEditNotice()).
			append(sc.format()).append("\n");

		// ****************

		//System.out.println(""+scenario_filename+"\n"+sb);
		writeFile(scenario_filename,sb.toString());

	}

	// ****************************************************************
	
	protected static synchronized ArrayList<Set<String>>
		generateScenariiBuckets(Iterator<String> it_sc_all,
								int generated_sc_count,
								int bucket_count)
		throws GridException {
		
		if ((it_sc_all == null) || (it_sc_all.hasNext() == false)) {
			throw new GridException("generateScenariiBuckets(): "+
									"illegal scenarii filenames DB");
		}
		if (bucket_count < 1) {
			throw new GridException("generateScenariiBuckets(): "+
									"illegal bucket count");
		}
		if (generated_sc_count < 1) {
			throw new GridException("generateScenariiBuckets(): "+
									"illegal scenarii count");
		}
		if (bucket_count > generated_sc_count) {
			bucket_count = generated_sc_count;
		}

		final int sc_per_bucket =
			(int)Math.ceil(((float)generated_sc_count) / bucket_count);
		assert(sc_per_bucket > 1);
		
		final ArrayList<Set<String>> sc_buckets =
			new ArrayList<Set<String>>(bucket_count);
		int added_sc = 0;
		for (int b = 0 ; b < bucket_count ; ++b) {
			Set<String> ts = new TreeSet<String>();
			sc_buckets.add(ts);
			for (int s = 0 ; s < sc_per_bucket ; ++s) {
				if (it_sc_all.hasNext()) {
					if (ts.add(it_sc_all.next()) == false) {
						throw new GridException("generateScenariiBuckets(): "+
												"corrupted "+
												"scenarii filenames DB");
					}
					++added_sc;
				}
				else {
					if (b != bucket_count-1) {
						throw new GridException("generateScenariiBuckets(): "+
												"corrupted "+
												"scenarii filenames DB");
					}
					else {
						break;
					}
				}
			}
		}
		
		if (added_sc != generated_sc_count) {
			throw new GridException("generateScenariiBuckets(): "+
									"corrupted "+
									"scenarii filenames DB (added "+
									added_sc+" to .jdf, while "+
									generated_sc_count+" were generated)");
		}
		
		if (it_sc_all.hasNext()) {
			throw new GridException("generateScenariiBuckets(): "+
									"corrupted "+
									"scenarii filenames DB");
		}
		
		return sc_buckets;
		
	}

	// ****************************************************************

}

