/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LBGridJDFWriter.java

package lbg.simulator.tools.jdf;

import lbg.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonFileHandling.*;
import static lbg.common.io.CommonStringHandling.*;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * @author Cyril Briquet
 */
public class LBGridJDFWriter
	implements JDFWriter {
	
	protected LBGridJDFWriter(String job_label,
							  String scenario_complete_filename,
							  String jar_complete_filename,
							  Iterator<String> scenarii,
							  String main_class,
							  boolean upload_jar_dep)
		throws GridException {
		
		if ((job_label == null) ||
			(job_label.length() < 1)) {
			throw new GridException("LBGridJDFWriter(): "+
									"illegal Job label");
		}

		if ((scenario_complete_filename == null) ||
			(scenario_complete_filename.length() < 1+TXT_EXT.length())) {
			throw new GridException("LBGridJDFWriter(): "+
									"illegal Scenario filename");
		}
		if ((jar_complete_filename == null) ||
			(jar_complete_filename.length() < 1+JAR_EXT.length())) {
			throw new GridException("LBGridJDFWriter(): "+
									"illegal JAR filename "+
									jar_complete_filename);
		}
		if ((scenarii == null) || (scenarii.hasNext() == false)) {
			throw new GridException("LBGridJDFWriter(): "+
									"illegal Scenarii");
		}

		this.upload_jar_dep = upload_jar_dep;
		this.jar_dependencies = LightweightBarteringGrid.getJarDependencies();

		this.scenario_complete_filename = scenario_complete_filename;
		this.scenario_filename =
			CommonFileHandling.
			extractFileNameFromPath(this.scenario_complete_filename);
		this.scenario_base_name =
			CommonFileHandling.
				extractBaseFileNameFromFileName(this.scenario_filename,TXT_EXT);

		this.jar_complete_filename = jar_complete_filename;
		this.jar_filename =
			CommonFileHandling.
				extractFileNameFromPath(this.jar_complete_filename);
		/*this.jar_base_name =
			CommonFileHandling.
				extractBaseFileNameFromFileName(this.jar_filename,JAR_EXT);*/

		this.main_class = (main_class != null) ?
			main_class :
			CommonFileHandling.extractMainClass(this.jar_complete_filename);
		
		this.job_label = job_label;
		this.scenarii = scenarii;
		
		/* legacy code -- this.default_task_d = null; */
		this.tasks_d = new ArrayList<TaskDescription>();
		
		/* legacy code
		// Add to default input several .jar + 1 optional common scenario
		TaskDescription task_d = new TaskDescription();
		task_d.setInputs(new String[] { this.scenario_filename,
								 		this.jar_filename });
		addDefaultTaskDescription(task_d);
		*/
		
		this.buildJDF();
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LBGridJDFWriter]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized String getJDFContent() throws GridException {

		return this.jdf_content;
		
	}
			
	public synchronized String getJDFFileName() throws GridException {

		return this.jdf_filename;
		
	}
	
	public synchronized void writeJDF() throws GridException {
		
		CommonFileHandling.writeFile(this.jdf_filename,this.jdf_content);

		System.out.println("Wrote Job Description File to "+this.jdf_filename);
		
	}
	
	// ****************************************************************

	/* legacy code
	private synchronized void addDefaultTaskDescription(TaskDescription task_d)
		throws GridException {
		
		if (task_d == null) {
			throw new GridException("addDefaultTaskDescription(): "+
									"illegal Task Description");
		}
		if (this.default_task_d != null) {
			throw new GridException("addDefaultTaskDescription(): "+
									"default Task Description already set");
		}
		
		this.default_task_d = task_d;
		
	}
	*/

	private synchronized void buildJDF() throws GridException {

		/* legacy code
		TaskDescription default_task_d = new TaskDescription();
		String[] s_default_inputs =
			new String[] { this.jar_filename };
		default_task_d.setInputs(s_default_inputs);
		jdfw.addDefaultTaskDescription(default_task_d);
		*/

		while (this.scenarii.hasNext()) {
			String gen_sc_fn = this.scenarii.next();
			TaskDescription task_d = new LBGridTaskDescription();

			String[] s_local_inputs = new String[] {
				// jar_fn, -- target .jar processed separately from data in LBG
				gen_sc_fn
			};
			String[] s_inputs = (this.upload_jar_dep) ?
				mergeArrays(s_local_inputs,this.jar_dependencies) :
				s_local_inputs;

			task_d.setInputs(s_inputs);

			//task_d.setParameters(new String[] { gen_sc_fn });

			task_d.setJar(this.jar_filename);

			task_d.setMainClass(this.main_class);

			this.tasks_d.add(task_d);
		}
		
		/* **************** */

		StringBuilder sb_jdf_filename = new StringBuilder();
		sb_jdf_filename.append(this.scenario_base_name);
		sb_jdf_filename.append(".jdf");

		StringBuilder sb_jdf_content = new StringBuilder();
		sb_jdf_content.append("# Lightweight Bartering Grid ").
			append("Job Description File").
			append("\n").
			append(CommonStringHandling.doNotEditNotice()).
			append("\n").
			append("# Job: ").
			append(this.job_label).
			append("\n\n").
			append("task_list = ");

		int i = 1;
		int s = tasks_d.size();
		for (TaskDescription task_d : tasks_d) {
			sb_jdf_content.append("lbg_task").append(i);
			if (i++ < s) sb_jdf_content.append(", ");
		}
		sb_jdf_content.append("\n\n");

		i = 1;
		for (TaskDescription task_d : tasks_d) {
			sb_jdf_content.append(task_d.format("lbg_task"+i++)).append("\n");
		}
		
		this.jdf_filename = sb_jdf_filename.toString();
		this.jdf_content = sb_jdf_content.toString();
		
		System.out.println("Created Job Description File "+this.jdf_filename);
		
	}

	// ****************************************************************

	private final boolean  upload_jar_dep;
	private final String[] jar_dependencies;

	private final String job_label;
	private final String scenario_complete_filename;
	private final String scenario_filename;
	private final String scenario_base_name;
	private final String jar_complete_filename;
	private final String jar_filename;
	private final String main_class;
	
	private final Iterator<String> scenarii;
	private final ArrayList<TaskDescription> tasks_d;

	private       String jdf_filename;
	private       String jdf_content;

}

