/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LogChartFactory.java

package lbg.simulator.tools.panorama;

import lbg.common.*;
import static lbg.common.ThreadPool.*;
import lbg.common.exception.*;
import static lbg.common.io.CommonChartHandling.*;
import lbg.simulator.tools.spreadsheet.*;

import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.LinkedBlockingQueue;

/** Factory to generate charts from given logs.
 * @author Cyril Briquet
 */
public class LogChartFactory implements Runnable {

	public LogChartFactory(LogSpreadsheet spreadsheet,boolean do_generate_pdf)
		throws GridException {
		
		if (spreadsheet == null) {
			throw new GridException("LogChartFactory(): illegal spreadsheet");
		}

		this.spreadsheet = spreadsheet;
		this.q = new LinkedBlockingQueue<LogChartRequest>();
		this.thread_pool =
			new ThreadPool(this,DEFAULT_THREAD_POOL_SIZE,true,false,
						   "LogChartFactory");
		this.do_generate_pdf = do_generate_pdf;
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LogChartFactory]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	public void requestLogChart(LogChartRequest log_chart_req)
		throws GridException {
		
		if (log_chart_req == null) {
			throw new GridException("requestLogChart(): "+
									"illegal log chart request");
		}
		
		// Launch multiple processing threads
		this.thread_pool.startService();
		
		try {
			// Enqueue request to process
			this.q.put(log_chart_req);
		}
		catch (InterruptedException e) {
			throw new GridException(e);
		}
		
	}
	
	/** Runnable implementation.
	 */
	public void run() {
		
		try {
			for ( ; ; ) {
				LogChartRequest log_chart_req = this.q.take();
				if (log_chart_req == null) {
					throw new GridException("run(): corrupted requests queue");
				}
				processLogChartRequest(log_chart_req);
				System.out.println("Processed chart request... "+
								   this.q.size()+" still queued");
			}
		}
		catch (Throwable thr) {
			thr.printStackTrace();
		}
		
	}
	
	// ****************************************************************

	private void processLogChartRequest(LogChartRequest log_chart_req)
		throws GridException {
		
		if (log_chart_req == null) {
			throw new GridException("processLogChartRequest(): "+
									"illegal log chart request");
		}
		
		System.out.println("Processing log chart request...");

		String[] col_titles = log_chart_req.getColumnTitles();
		// Sort spreadsheet, get sorted column
		Float[] sorted_column =
			LogSpreadsheetSorter.
				getSortedDataColumn(this.spreadsheet,col_titles);
		
		// Dummy data conversion
		double[] sorted_column_d =
			new double[sorted_column.length];
		for (int d = 0 ; d < sorted_column.length ; ++d) {
			sorted_column_d[d] = sorted_column[d];
		}
		
		String x_label = col_titles[col_titles.length > 1 ?
									col_titles.length-2 : 0];
		String y_label = col_titles[col_titles.length-1];
		
		StringBuilder sb_pdf_fn = new StringBuilder();
		sb_pdf_fn.append(this.spreadsheet.getFilename());
		sb_pdf_fn.append("__");
		sb_pdf_fn.append(y_label);
		sb_pdf_fn.append("_fct_of_");
		sb_pdf_fn.append(x_label);
		String pdf_fn = sb_pdf_fn.toString().replace('.','_');
		pdf_fn += ".pdf";
		
		System.out.println("Generating log chart image...");

		// Generate chart image, also save as .pdf
		BufferedImage chart_img =
			(this.do_generate_pdf ?
			generateChartImage(saveAsPDF(generateChart(null,x_label,y_label,
													   sorted_column_d),
										 pdf_fn)) :
			generateChartImage(generateChart(null,x_label,y_label,
											 sorted_column_d)));

		log_chart_req.setChartImage(chart_img);

		System.out.println("Completed generation of log chart image");
		
		// Notify listener that the chart image is available
		log_chart_req.listener.chartDrawn(log_chart_req);
		
		System.out.println("Completed processing of log chart request");

	}

	// ****************************************************************

	private final LogSpreadsheet spreadsheet;
	private final LinkedBlockingQueue<LogChartRequest> q;
	private final ThreadPool thread_pool;
	private final boolean do_generate_pdf;

}

