/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LogPanorama.java

package lbg.simulator.tools.panorama;

import lbg.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.peer.config.PeerStrategyConfiguration.getPDPCount;
import static lbg.peer.config.PeerStrategyConfiguration.getPDPList;
import lbg.simulator.tools.spreadsheet.*;

import java.awt.image.BufferedImage;
import java.util.concurrent.LinkedBlockingQueue;

/** Analyzes a given LBG simulator log file.
 * @author Cyril Briquet
 */
public class LogPanorama implements LogChartDrawingListener {

	public LogPanorama(String[] args) throws GridException {
		
		checkArgs(args);

		this.spreadsheet_fn = args[0];
		boolean do_generate_pdf =
			(args.length > 1) ? Boolean.parseBoolean(args[1]) : false;

		this.q = new LinkedBlockingQueue<LogChartRequest>();
		this.panorama_frame = new LogPanoramaFrame(this.spreadsheet_fn,this.q);
		
		this.spreadsheet = new LogSpreadsheet(spreadsheet_fn);
		
		this.chart_factory =
			new LogChartFactory(this.spreadsheet,do_generate_pdf);
		this.log_chart_req_factory = new LogChartRequestFactory(this);

	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LogPanorama]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** LogChartDrawingListener implementation.
	 */
	public void chartDrawn(LogChartRequest log_chart_req) throws GridException {
		
		if (log_chart_req == null) {
			throw new GridException("chartDrawn(): illegal log chart request");
		}
		
		try {
			// Enqueue chart to display
			this.q.put(log_chart_req);
		}
		catch (InterruptedException e) {
			throw new GridException(e);
		}
		
	}
	
	// ****************************************************************

	public void display() throws GridException {

		this.panorama_frame.setup((getPDPCount()*2) + 2);
		
		String mjc_col_title = "grid.job_completion_time.mean";
		String srt_col_title = "simulation.real_time";

		String[] pdp_list = getPDPList();
		for (int i = 0 ; i < pdp_list.length ; ++i) {
			this.chart_factory.requestLogChart(this.log_chart_req_factory.
				createLogChartRequest(pdp_list[i],mjc_col_title));
			
			this.chart_factory.requestLogChart(this.log_chart_req_factory.
				createLogChartRequest(pdp_list[i],srt_col_title));
		}
		
		this.chart_factory.requestLogChart(this.log_chart_req_factory.
			createLogChartRequest(mjc_col_title));
		this.chart_factory.requestLogChart(this.log_chart_req_factory.
			createLogChartRequest(srt_col_title));
		
	}
	
	// ****************************************************************

	public static void main(String[] args) {
		
		try {
			LightweightBarteringGrid.setMode(VirtualizationLevel.SIM_MODE);
			
			LogPanorama panorama = new LogPanorama(args);
			panorama.display();
		}
		catch (GridException e) {
			e.printStackTrace();
			System.exit(0);
		}
		
	}

	// ****************************************************************

	private static void checkArgs(String[] args) throws GridException {
	
		boolean ok = true;

		if (ok) {
			if (args == null) {
				ok = false;
			}
		}

		if (ok)  {
			if ((args.length < 1) || (args.length > 2)) {
				ok = false;
			}
		}
		
		if (ok) {
			if (CommonFileHandling.isCSVFileName(args[0]) == false) {
				ok = false;
			}
		}
		
		if (ok) {
			if ((args.length > 1) && (args[1] == null)) {
				ok = false;
			}
		}

		if (ok == false) {
			System.out.println("Usage: "+
							   "java -cp classpath "+
							   LogPanorama.class.getName()+
							   " log_aggregator_report.csv "+
							   "[do_generata_pdf boolean value]\n");
			throw new GridException("checkArgs(): illegal arguments");
		}
		
	}
	
	// ****************************************************************

	private final String spreadsheet_fn;

	private final LinkedBlockingQueue<LogChartRequest> q;
	private final LogPanoramaFrame panorama_frame;

	private final LogSpreadsheet spreadsheet;

	private final LogChartFactory chart_factory;
	private final LogChartRequestFactory log_chart_req_factory;


}

