/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LogPanoramaFrame.java

package lbg.simulator.tools.panorama;

import lbg.common.*;
import lbg.common.exception.*;
import static lbg.common.net.CommonNetHandling.*;

import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.image.BufferedImage;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import java.net.URL;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * @author Cyril Briquet
 */
public class LogPanoramaFrame implements Runnable {

	public LogPanoramaFrame(String spreadsheet_fn,
							LinkedBlockingQueue<LogChartRequest> q)
		throws GridException {
		
		if (spreadsheet_fn == null) {
			throw new GridException("LogPanoramaFrame(): "+
									"illegal spreadsheet filename");
		}
		if (q == null) {
			throw new GridException("LogPanoramaFrame(): illegal queue");
		}
		
		this.spreadsheet_fn   = spreadsheet_fn;
		this.q                = q;
		this.thread_pool = new ThreadPool(this,1,false,true,"LogPanoramaFrame");

		this.expected_charts  = 0;
		this.chart_img_db     = new TreeMap<Integer,JLabel>();
		this.boot_chart_label = default_boot_url == null ?
								new JLabel("    Loading...    ") :
								new JLabel(new ImageIcon(default_boot_url));

		
		this.main_frame_title = "LogPanorama ["+this.spreadsheet_fn+"]";
		this.main_frame = new JFrame("LogPanorama [loading...]");
		this.main_frame.setSize(DEFAULT_INITIAL_WIDTH,DEFAULT_INITIAL_HEIGHT);
		this.main_frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		this.pan = new JPanel(new FlowLayout());
		this.main_frame.setContentPane(this.pan);

		displayChartImageDB();

		this.main_frame.setVisible(true);
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LogPanoramaFrame]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	/** Runnable implementation.
	 */
	public void run() {
		
		try {
			for ( ; ; ) {
				processChart(this.q.take());
			}
		}
		catch (Throwable thr) {
			thr.printStackTrace();
		}
		
	}
	
	public synchronized void setup(int expected_charts) throws GridException {
		
		if (expected_charts < 1) {
			throw new GridException("setup(): illegal expected charts count");
		}

		// Launch processing threads
		this.thread_pool.startService();

		// Clear chart image database
		this.expected_charts = expected_charts;
		this.chart_img_db.clear();

	}

	// ****************************************************************

	public static final int DEFAULT_INITIAL_WIDTH  = 64;
	public static final int DEFAULT_INITIAL_HEIGHT = 64;
	
	// ****************************************************************

	private synchronized void processChart(LogChartRequest log_chart_req)
		throws GridException {
		
		if (log_chart_req == null) {
			throw new GridException("displayChart(): "+
									"illegal log chart request");
		}

		int chart_id = log_chart_req.chart_id;
		JLabel chart_img_label =
			new JLabel(new ImageIcon(log_chart_req.getChartImage()));
		
		if (this.chart_img_db.put(chart_id,chart_img_label) != null) {
			throw new GridException("displayChart(): duplicate chart image "+
									"for chart id "+chart_id);
		}

		displayChartImageDB();
		
	}
	
	private synchronized void displayChartImageDB() {

		// Clear panel
		this.pan.removeAll();

		// Add chart images to panel
		if ((this.expected_charts > 0) &&
			(this.chart_img_db.size() == this.expected_charts)) {
		
			int side_size = (int)Math.ceil(Math.sqrt(this.expected_charts));
			this.pan.setLayout(new GridLayout(side_size,side_size));

			for (int i = 0 ; i < this.expected_charts ; i++) {
				JLabel chart_img_label = this.chart_img_db.get(i);
				this.pan.add((chart_img_label == null) ?
							 this.boot_chart_label : chart_img_label);
				//this.pan.add(new JLabel(new ImageIcon(chart_img)));
			}
			
			this.main_frame.setTitle(this.main_frame_title);
			
		}
		else {
			
			this.pan.setLayout(new FlowLayout());
			
			this.pan.add(this.boot_chart_label);
			
			this.main_frame.setTitle("LogPanorama [loading... "+
									 this.chart_img_db.size()+" charts]");
			
		}
	
		// Display panel
		this.main_frame.pack();
		this.main_frame.setLocationRelativeTo(null);
		
	}
	
	// ****************************************************************

	private final String spreadsheet_fn;
	private final LinkedBlockingQueue<LogChartRequest> q;
	private final ThreadPool thread_pool;

	private final JLabel boot_chart_label;
	private       int expected_charts;
	private final Map<Integer,JLabel> chart_img_db;

	private final String main_frame_title;
	private final JFrame main_frame;
	private final JPanel pan;

	private static final URL default_boot_url = loadRandomURLSafely(
		new String[] {
		"http://farm1.static.flickr.com/134/330987206_1ae5c9a63c.jpg",
		"http://farm2.static.flickr.com/1134/975758085_7a5d741921.jpg",
		"http://farm3.static.flickr.com/2095/1577960091_353deeb51f.jpg",
		"http://farm1.static.flickr.com/72/172146299_842d4d2d7a.jpg"
	});

}

