/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LogSpreadsheet.java

package lbg.simulator.tools.spreadsheet;

import lbg.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonStringHandling.DUMMY_STRING;
import lbg.simulator.tools.*;

import java.io.BufferedReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.TreeMap;

/** Analyzes a given LBG simulator log file.
 * @author Cyril Briquet
 */
public class LogSpreadsheet {

	public LogSpreadsheet(String spreadsheet_fn) throws GridException {

		if ((spreadsheet_fn == null) ||
			(CommonFileHandling.isCSVFileName(spreadsheet_fn) == false)) {
			throw new GridException("LogSpreadsheet(): "+
									"illegal CSV spreadsheet filename");
		}
		
		this.spreadsheet_fn = spreadsheet_fn;
		this.col_idx_to_title_map = new ArrayList<String>(42);
		this.col_title_to_idx_map = new TreeMap<String,Integer>();
		this.is_strategy_column   = new ArrayList<Boolean>(42);
		this.data_columns     = new TreeMap<String,Float[]>();
		this.strategy_columns = new TreeMap<String,String[]>();

		loadSpreadsheet();
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LogSpreadsheet]");
		
		return sb.toString();
		
	}

	// ****************************************************************

	public synchronized String getFilename() {
		
		return this.spreadsheet_fn;
	
	}
	
	public synchronized int getColumnsCount() {
		
		return this.cols_count;
		
	}

	public synchronized int getLinesCount() {
		
		return this.lines_count;
		
	}

	public synchronized String[] getColumnsTitles() {
	
		String[] col_titles = new String[this.col_idx_to_title_map.size()];
		
		int col_idx = 0;
		for (String col_title : this.col_idx_to_title_map) {
			col_titles[col_idx++] = col_title;
		}
		
		return col_titles;

	}
	
	public synchronized boolean isStrategyColumn(String col_title)
		throws GridException {
		
		if (col_title == null) {
			throw new GridException("isStrategyColumn(): illegal column title");
		}
		
		Integer col_idx = col_title_to_idx_map.get(col_title);
		if (col_idx == null) {
			throw new GridException("isStrategyColumn(): illegal column");
		}
		
		if ((col_idx < 0) || (col_idx >= this.is_strategy_column.size())) {
			throw new GridException("isStrategyColumn(): corrupted data");
		}
		return this.is_strategy_column.get(col_idx);
		
	}

	public synchronized String[] getStrategyColumn(String col_title)
		throws GridException {

		if (col_title == null) {
			throw new GridException("getStrategyColumn(): "+
									"illegal column title");
		}

		return this.strategy_columns.get(col_title);

	}

	public synchronized Float[] getDataColumn(String col_title)
		throws GridException {

		if (col_title == null) {
			throw new GridException("getDataColumn(): illegal column title");
		}

		return this.data_columns.get(col_title);

	}

	// ****************************************************************

	public synchronized String dump() throws GridException {

		// Prepare first line
		StringBuilder sb_titles = new StringBuilder();
		int col_idx = 0;
		for (String col_title : this.col_idx_to_title_map) {
			if (col_idx++ > 0) {
				sb_titles.append(", ");
			}
			sb_titles.append(col_title);
		}
		sb_titles.append("\n");

		// Allocate memory for following lines
		StringBuilder[] sb_lines = new StringBuilder[this.lines_count];
		for (int l = 0 ; l < this.lines_count ; ++l) {
			sb_lines[l] = new StringBuilder();
		}

		// Prepare following lines
		col_idx = 0;
		for (String col_title : this.col_idx_to_title_map) {
			int col_idx_check = this.col_title_to_idx_map.get(col_title);
			if (col_idx != col_idx_check) {
				throw new GridException("dump(): illegal data");
			}
			if (isStrategyColumn(col_title)) {
				String[] column = this.strategy_columns.get(col_title);
				if (col_idx > 0) {
					for (int l = 0 ; l < this.lines_count ; ++l) {
						sb_lines[l].append(", ").append(column[l]);
					}
				}
				else {
					for (int l = 0 ; l < this.lines_count ; ++l) {
						sb_lines[l].append(column[l]);
					}
				}
			}
			else {
				Float[] column = this.data_columns.get(col_title);
				if (col_idx > 0) {
					for (int l = 0 ; l < this.lines_count ; ++l) {
						String s = CommonStringHandling.
							formatIntegerDoubleAsInteger(column[l]);
						sb_lines[l].append(", ").append(s);
					}
				}
				else {
					for (int l = 0 ; l < this.lines_count ; ++l) {
						String s = CommonStringHandling.
							formatIntegerDoubleAsInteger(column[l]);
						sb_lines[l].append(s);
					}
				}
			}
			++col_idx;
		}
		
		StringBuilder sb = new StringBuilder();
		sb.append(sb_titles);
		for (int l = 0 ; l < this.lines_count ; ++l) {
			sb.append(sb_lines[l]).append("\n");
		}
		
		return sb.toString();

	}

	public static void main(String[] args) {
	
		try {
			LogSpreadsheet spreadsheet = new LogSpreadsheet(args[0]);
	
			CommonFileHandling.writeFile("___dump.txt",spreadsheet.dump());
		}
		catch (GridException e) {
			e.printStackTrace();
		}
		finally {
			System.exit(0);
		}
		
	}

	
	// ****************************************************************

	private synchronized void loadSpreadsheet() throws GridException {

		// Analyze spreadsheet columns titles
		BufferedReader br = CommonFileHandling.openBR(spreadsheet_fn);
		String line = CommonFileHandling.readLine(br,spreadsheet_fn);
		if (line == null) {
			throw new GridException("loadSpreadsheet(): empty file");
		}
		String[] dummy_s = new String[0];
		Float[] dummy_f  = new Float[0];
		StringTokenizer st = new StringTokenizer(line,", ");
		int current_col = 0;
		while (st.hasMoreTokens()) {
			String col_title = st.nextToken();
			if (this.col_idx_to_title_map.add(col_title) == false) {
				throw new GridException("loadSpreadsheet(): "+
										"corrupted data (1)");
			}
			if (this.col_title_to_idx_map.put(col_title,current_col) != null) {
				throw new GridException("loadSpreadsheet(): "+
										"corrupted data (2)");
			}
			boolean is_strategy_column =
				LogAnalyzer.isStrategyProperty(col_title);
			this.is_strategy_column.add(is_strategy_column);
			if (is_strategy_column) {
				if (this.strategy_columns.put(col_title,dummy_s) != null) {
					throw new GridException("loadSpreadsheet(): "+
											"corrupted data (3)");
				}
			}
			else {
				if (this.data_columns.put(col_title,dummy_f) != null) {
					throw new GridException("loadSpreadsheet(): "+
											"corrupted data (4)");
				}
			}
			++current_col;
		}

		this.cols_count = current_col;
		if (this.cols_count != this.col_idx_to_title_map.size()) {
			throw new GridException("loadSpreadSheetLine(): "+
									"corrupted data (5)");
		}
		if (this.cols_count != this.is_strategy_column.size()) {
			throw new GridException("loadSpreadSheetLine(): "+
									"corrupted data (6)");
		}

		// Extract lines from spreadsheet file
		ArrayList<String> lines = new ArrayList<String>(4096);
		while ((line = CommonFileHandling.readLine(br,spreadsheet_fn))!=null) {
			lines.add(line);
		}
		CommonFileHandling.closeBR(br,spreadsheet_fn);
		this.lines_count = lines.size();

		// Initialize columns
		for (String strategy_col_title : this.strategy_columns.keySet()) {
			this.strategy_columns.
				put(strategy_col_title,new String[lines_count]);
		}
		for (String data_col_title : this.data_columns.keySet()) {
			this.data_columns.
				put(data_col_title,new Float[lines_count]);
		}
		
		// Load spreadsheet lines into data matix
		for (int l = 0 ; l < lines_count ; ++l) {
			loadSpreadSheetLine(l,lines.get(l));
		}
		
	}

	private synchronized void loadSpreadSheetLine(int l,String line)
		throws GridException {
		
		if ((l < 0) || (l >= lines_count)) {
			throw new GridException("loadSpreadsheetLine(): illegal line id");
		}
		if (line == null) {
			throw new GridException("loadSpreadsheetLine(): illegal line");
		}

		StringTokenizer st = new StringTokenizer(line,", ");
		int col_idx = 0;
		while (st.hasMoreTokens()) {
			String val = st.nextToken();
			insertValue(l,col_idx,val);
			++col_idx;
		}
		
		// If the file line is incomplete, i.e. because the experiment failed,
		// continue inserting dummy values into the spreadsheet line so that
		// the spreadsheet can be sorted
		final int col_count = this.col_idx_to_title_map.size();
		while (col_idx < col_count) {
			insertValue(l,col_idx,DUMMY_STRING);
			++col_idx;
		}

	}
	
	private synchronized void insertValue(int l,int col_idx,String val)
		throws GridException {
		
		if ((l < 0) || (l >= lines_count)) {
			throw new GridException("insertValue(): illegal line id");
		}
		if ((col_idx < 0) || (col_idx >= this.col_idx_to_title_map.size())) {
			throw new GridException("insertValue(): illegal column index");
		}
		if (val == null) {
			throw new GridException("insertValue(): illegal value");
		}
		
		String col_title = this.col_idx_to_title_map.get(col_idx);
		
		boolean is_strategy_column = this.is_strategy_column.get(col_idx);
		
		if (is_strategy_column) {
			String[] column = this.strategy_columns.get(col_title);
			if (column == null) {
				throw new GridException("insertValue(): "+
										"corrupted strategy column");
			}
			column[l] = val;
		}
		else {
			Float[] column = this.data_columns.get(col_title);
			if (column == null) {
				throw new GridException("insertValue(): "+
										"corrupted strategy column");
			}
			float f = 0.0f;
			try {
				f = Float.parseFloat(val);
			}
			catch (NumberFormatException e) { f = -42.0f; }
			column[l] = f;
		}

	}
	
	// ****************************************************************

	private final String spreadsheet_fn;
	
	private final ArrayList<String>   col_idx_to_title_map;
	private final Map<String,Integer> col_title_to_idx_map;
	private final ArrayList<Boolean>  is_strategy_column;

	private       int cols_count;
	private       int lines_count;

	private final Map<String,Float[]>  data_columns;
	private final Map<String,String[]> strategy_columns;

	// ****************************************************************

}

