/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LogSpreadsheetSorter.java

package lbg.simulator.tools.spreadsheet;

import lbg.common.exception.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

/** LogSpreadsheet sorter.
 * @author Cyril Briquet
 */
public abstract class LogSpreadsheetSorter {

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LogSpreadsheetSorter]");
		
		return sb.toString();
		
	}
	
	// ****************************************************************

	/** Get last column of target columns sorted
	 * in function of the other target columns
	 * (first target column is first sort criterion and so on).
	 */
	public static Float[] getSortedDataColumn(LogSpreadsheet spreadsheet,
										  	  String... col_titles)
		throws GridException {
		
		if (spreadsheet == null) {
			throw new GridException("getSortedDataColumn(): "+
									"illegal spreadsheet");
		}
		if ((col_titles == null) || (col_titles.length < 1)) {
			throw new GridException("getSortedDataColumn(): "+
									"no column titles given");
		}
		
		String target_col_title = col_titles[col_titles.length-1];
		if (spreadsheet.isStrategyColumn(target_col_title) == true) {
			throw new GridException("getSortedDataColumn(): "+
									"column type not supported");
		}
		
		ArrayList<Integer> sorted_lines = LogSpreadsheetSorter.
			sortColumns(spreadsheet,col_titles);
		
		Float[] column = spreadsheet.getDataColumn(target_col_title);
		if (column == null) {
			throw new GridException("getSortedDataColumn(): "+
									"corrupted spreadsheet, cannot find "+
									"column "+target_col_title);
		}
		if (column.length != sorted_lines.size()) {
			throw new GridException("getSortedDataColumn(): "+
									"corrupted spreadsheet column");
		}

		Float[] sorted_column = new Float[column.length];
		
		int ol = 0;
		for (Integer il : sorted_lines) {
			sorted_column[ol++] = column[il];
		}
		
		return sorted_column;
		
	}

	// ****************************************************************

	/** Get last column of target columns sorted
	 * in function of the other target columns
	 * (first target column is first sort criterion and so on).
	 */
	public static String[] getSortedStrategyColumn(LogSpreadsheet spreadsheet,
												   String... col_titles)
		throws GridException {
		
		if (spreadsheet == null) {
			throw new GridException("getSortedStrategyColumn(): "+
									"illegal spreadsheet");
		}
		if ((col_titles == null) || (col_titles.length < 1)) {
			throw new GridException("getSortedStrategyColumn(): "+
									"no column titles given");
		}
		
		String target_col_title = col_titles[col_titles.length-1];
		if (spreadsheet.isStrategyColumn(target_col_title) == false) {
			throw new GridException("getSortedStrategyColumn(): "+
									"column type not supported");
		}
		
		ArrayList<Integer> sorted_lines = LogSpreadsheetSorter.
			sortColumns(spreadsheet,col_titles);
		
		String[] column = spreadsheet.getStrategyColumn(target_col_title);
		if (column == null) {
			throw new GridException("getSortedStrategyColumn(): "+
									"corrupted spreadsheet, cannot find "+
									"column "+target_col_title);
		}
		if (column.length != sorted_lines.size()) {
			throw new GridException("getSortedStrategyColumn(): "+
									"corrupted spreadsheet column");
		}
		
		String[] sorted_column = new String[column.length];
		
		int ol = 0;
		for (Integer il : sorted_lines) {
			sorted_column[ol++] = column[il];
		}
		
		return sorted_column;
		
	}

	// ****************************************************************

	public static ArrayList<Integer> sortColumns(LogSpreadsheet spreadsheet,
									   			 String... col_titles)
		throws GridException {
		
		//System.out.println("Sorting spreadsheet...");
		
		// Interface method both easy to use and
		// enabling to streamline the recursive sort process
		
		if (spreadsheet == null) {
			throw new GridException("sortColumns(): illegal spreadsheet");
		}
		if ((col_titles == null) || (col_titles.length < 1)) {
			throw new GridException("sortColumns(): no column titles given");
		}
		
		Set<Integer> input_lines = new TreeSet<Integer>();
		int lines_count = spreadsheet.getLinesCount();
		for (int l = 0 ; l < lines_count ; l++) {
			input_lines.add(l);
		}
		
		ArrayList<Integer> sorted_lines = LogSpreadsheetSorter.
			sortColumns(spreadsheet,input_lines,col_titles,0);

		//System.out.println("Completed spreadsheet sort");
		
		return sorted_lines;
		
	}

	// ****************************************************************

	private static ArrayList<Integer>
		sortColumns(LogSpreadsheet spreadsheet,Set<Integer> input_lines,
					String[] col_titles,int target_col)
		throws GridException {
		
		if (spreadsheet == null) {
			throw new GridException("sortColumns(): illegal spreadsheet");
		}
		if (input_lines == null) {
			throw new GridException("sortColumns(): no lines to sort");
		}
		if (col_titles == null) {
			throw new GridException("sortColumns(): no column titles given");
		}
		if ((target_col < 0) || (target_col > col_titles.length)) {
			throw new GridException("sortColumns(): illegal current column");
		}
		
		ArrayList<Integer> sorted_lines = new ArrayList<Integer>();

		if (target_col == col_titles.length) { // base case
			sorted_lines.addAll(input_lines);
		}
		else { // general case
			// Map unique elements of target column to
			// the set of line indexes where they happen to be
			String col_title = col_titles[target_col];
			Collection<Set<Integer>> groups_of_lines_with_equal_val = 
				group(input_lines,
					  (spreadsheet.isStrategyColumn(col_title) ?
						  spreadsheet.getStrategyColumn(col_title) :
						  spreadsheet.getDataColumn(col_title)));
			/* BEGIN old code
			if (spreadsheet.isStrategyColumn(col_title)) {
				String[] column = spreadsheet.getStrategyColumn(col_title);
				map = cool(input_lines,column);
			}
			else {
				Float[] column = spreadsheet.getDataColumn(col_title);
				map = cool(input_lines,column);
			}
			   END old code */
			
			// Recursively process (depth-first)
			// each set of line indexes with common value in target column
			for (Set<Integer> lines_with_equal_val :
					groups_of_lines_with_equal_val) {
				if (lines_with_equal_val.size() <= 1) {
					// Merge implicitly sorted line
					sorted_lines.addAll(lines_with_equal_val);
				}
				else {
					// Process input lines
					ArrayList<Integer> sorted_lines_subset =
						sortColumns(spreadsheet,lines_with_equal_val,
									col_titles,target_col+1);
					// Merge sorted lines
					sorted_lines.addAll(sorted_lines_subset);
				}
			}
		}
		
		assert(areLinesSetsConsistent(input_lines,sorted_lines));
		
		return sorted_lines;
		
	}

	/** Groups sets of line indexes with common value in target column.
	 * <p>In other words, maps unique elements of target column to
     * the set of line indexes where they happen to be.
	 * @param input_lines set of input lines
	 * @param column target column
	 * @return groups of sets of line indexes with common value in target column
	 * @exception GridException if a parameter is illegal
	 */
	private static <T>
		Collection<Set<Integer>> group(Set<Integer> input_lines,T[] column)
		throws GridException {
		
		if (input_lines == null) {
			throw new GridException("group(): no lines to sort");
		}
		if (column == null) {
			throw new GridException("group(): illegal column");
		}
		
		Map<T,Set<Integer>> map = new TreeMap<T,Set<Integer>>();
		
		for (Integer l : input_lines) {
			if ((l < 0) || (l > column.length)) {
				throw new GridException("group(): corrupted lines set");
			}
			T criterion = column[l];
			if (criterion == null) {
				throw new GridException("group(): illegal column element "+
										"at line "+l);
			}
			Set<Integer> lines_with_equal_val = map.get(criterion);
			if (lines_with_equal_val == null) {
				lines_with_equal_val = new TreeSet<Integer>();
				map.put(criterion,lines_with_equal_val);
			}
			if (lines_with_equal_val.add(l) == false) {
				throw new GridException("group(): corrupted data");
			}
		}

		assert(areLinesSetsConsistent(input_lines,map.values()));
		
		return map.values();
		
	}
	
	// ****************************************************************

	private static boolean
		areLinesSetsConsistent(Set<Integer> input_lines,
						       ArrayList<Integer> sorted_lines)
		throws GridException {
	
		if (input_lines == null) {
			throw new GridException("areLinesSetsConsistent(): "+
									"illegal input lines");
		}
		if (sorted_lines == null) {
			throw new GridException("areLinesSetsConsistent(): "+
									"illegal sorted lines");
		}
		
		if (input_lines.size() != sorted_lines.size()) {
			return false;
		}
		
		for (Integer l : sorted_lines) {
			if (input_lines.contains(l) == false) {
				return false;
			}
		}
	
		return true;
		
	}

	private static boolean
		areLinesSetsConsistent(Set<Integer> input_lines,
						       Collection<Set<Integer>> sorted_lines)
		throws GridException {
	
		if (input_lines == null) {
			throw new GridException("areLinesSetsConsistent(): "+
									"illegal input lines");
		}
		if (sorted_lines == null) {
			throw new GridException("areLinesSetsConsistent(): "+
									"illegal sorted lines");
		}

		int sorted_lines_size = 0;

		for (Set<Integer> sorted_lines_subset : sorted_lines) {
			for (Integer l : sorted_lines_subset) {
				if (input_lines.contains(l) == false) {
					return false;
				}
				++sorted_lines_size;
			}
		}

		if (input_lines.size() != sorted_lines_size) {
			return false;
		}

		return true;
		
	}
	
	// ****************************************************************

}

