/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Task.java

package lbg.task;

import lbg.*;
import lbg.common.*;
import lbg.common.exception.*;
import static lbg.common.io.CommonStringHandling.DUMMY_STRING_ARRAY;
import static lbg.common.io.CommonStringHandling.DUMMY_MGD_ARRAY;
import static lbg.common.math.CommonMath.generateHashCode;
import lbg.protocol.data.*;
import java.io.Serializable;
import java.util.Arrays;

/**
 * @author Cyril Briquet, Xavier Dalem
 */
public class Task implements Cloneable, Comparable<Task>, Serializable {

	/** Constructs a new Task.
	 */
	public Task(String consumer_peer_id,int job_id,int bot_id,int task_id,
				byte[][] jars,String[] classnames,String main_class,
				Object[] parameters,GridData[][] datas,long nominal_runtime)
		throws GridException {
		
		if (consumer_peer_id == null) {
			throw new GridException("Task(): illegal consumer Peer id");
		}

		if (job_id < 0) {
			throw new GridException("Task(): illegal Job id");
		}

		if (bot_id < 0) {
			throw new GridException("Task(): illegal BoT id");
		}

		if (task_id < 0) {
			throw new GridException("Task(): illegal Task id");
		}

		if (jars == null) {
			throw new GridException("Task(): illegal jarfile");
		}

		if (classnames == null) {
			throw new GridException("Task(): illegal class names");
		}

		if (main_class == null) {
			throw new GridException("Task(): illegal main class name");
		}

		if (parameters == null) {
			throw new GridException("Task(): illegal input parameters");
		}

		if (datas == null) {
			throw new GridException("Task(): illegal data");
		}

		if (nominal_runtime < 0) {
			throw new GridException("Task(): illegal nominal runtime");
		}

		this.consumer_peer_id = consumer_peer_id;

		this.job_id = job_id;
		this.job_id_as_key = Tank.newInteger(this.job_id);
		this.tmp_job_id = -1;
		this.bot_id = bot_id;
		this.tmp_bot_id = -1;
		this.task_id = task_id;
		this.task_id_as_key = Tank.newInteger(this.task_id);

		this.jars = jars;
		this.main_class = main_class;
		this.classnames = classnames;
		this.parameters = parameters;
		this.datas      = datas;
		this.nominal_runtime = nominal_runtime;

		this.start_time = -1;
		this.is_completed = false;;
		this.task_output = null;

		this.preemption_count = 0;
		this.cancellation_count = 0;
		this.runtime_before_last_cancellation = 0;
		this.supplier_before_last_cancellation = null;
		
		/* BEGIN data-related preprocessing */
		// Build dependencies array
		this.readied_datas = null;
		this.input_data_set_size = 0;
		if (this.datas.length == 0) {
			this.dependencies = DUMMY_STRING_ARRAY;
			this.input_meta_grid_data_set = DUMMY_MGD_ARRAY;
			this.input_data_set_id = null;
			return;
		}

		//debugging code -- System.out.print("| ");
		this.dependencies = new String[this.datas.length];
		this.input_meta_grid_data_set = new MetaGridData[this.datas.length];
		// Loop over inout data files
		for (int i = 0; i < this.datas.length; i++) {
			this.dependencies[i] = null;
			this.input_meta_grid_data_set[i] = null;
			// Loop over download options
			for (int j = 0 ; j < this.datas[i].length ; j++) {
				GridData gd = this.datas[i][j];
				if (gd != null) {
					//debugging code -- System.out.print(gd.id+" ");
					this.dependencies[i] = gd.id;
					this.input_meta_grid_data_set[i] =
						new MetaGridData(gd.id,gd.size,this.datas[i]);
					this.input_data_set_size += gd.size; // <- MUST break loop !
					break;
				}
			}
			assert((this.dependencies[i] != null) &&
				   (this.input_meta_grid_data_set[i] != null));
		}
		//debugging code -- System.out.print(" |");
		
		// Build an input dataset id (and check its validity)
		Arrays.sort(this.dependencies);
		StringBuilder sb = new StringBuilder(this.dependencies[0]);
		for (int i = 1 ; i < this.dependencies.length; i++) {
			int cmpval = this.dependencies[i-1].compareTo(this.dependencies[i]);
			if (cmpval == 0) {
				throw new GridException("Task(): data redundancy detected ("+
										this.dependencies[i]+")");
			}
			else if (cmpval > 0) {
				throw new GridException("Task(): corrupted data");
			}
			sb.append(this.dependencies[i]);
			sb.append(" ");
		}
		this.input_data_set_id = sb.toString();
		/* END data-related preprocessing */

		this.hashcode = 0;
		
	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		sb.append("[Task ");
		sb.append(this.task_id);
		sb.append("] consumer=");
		sb.append(this.consumer_peer_id);
		sb.append(", ");
		sb.append("Job id=");
		sb.append(this.job_id);
		sb.append(", ");
		sb.append("BoT id=");
		sb.append(this.bot_id);
		sb.append(", ");
		sb.append("Supplying space Job id=");
		sb.append(this.tmp_job_id);
		sb.append(", ");
		sb.append("Supplying space BoT id=");
		sb.append(this.tmp_bot_id);
		sb.append(", ");

		sb.append(this.jars.length);
		sb.append(" bytecode");
		if (this.jars.length > 1) sb.append("s");
		sb.append(", ");
		sb.append(this.parameters.length);
		sb.append(" input parameter");
		if (this.parameters.length > 1) sb.append("s");
		sb.append(", ");

		if (this.dependencies != null) {
			sb.append(" input dataset = { ");
			int i = 0;
			for (String data_id : this.dependencies) {
				sb.append(data_id);
				if (++i < this.dependencies.length) sb.append(", ");
			}
			sb.append(" }, ");
		}
		else {
			sb.append("no input dataset, ");
		}

		sb.append("nominal runtime=");
		sb.append(this.nominal_runtime);
		if (this.is_completed) {
			sb.append(", ");
			sb.append("completed\n");
			sb.append(this.task_output);
		}

		return sb.toString();

	}

	/**
	 * <p>Note: this class has a natural ordering that is inconsistent
	 * with equals (see comment here below).
	 * <p><small>It is hypothesized that 2 Tasks from different BoTs will *never* be
	 * compared. As this method is called very frequently, the implementation
	 * of compareTo() skips the comparison of BoT id,
	 * relying only on the comparison of Task id.</small>
	 */
	public int compareTo(Task other) {
	
		return this.task_id - other.id();
		
	}

	@Override
	public boolean equals(Object obj) {
		
		if (obj instanceof Task == false) {
			return false;
		}
		
		Task other = (Task)obj;
		
		return (this.job_id == other.jobID()) &&
			(this.bot_id == other.botID()) && (this.task_id == other.id());
		
	}

	@Override
	public synchronized int hashCode() {
	
		if (this.hashcode == Integer.MIN_VALUE) {
			if (this.tmp_job_id == -1) {
				this.hashcode =
					generateHashCode(this.job_id,
									 generateHashCode(
										 generateHashCode(this.bot_id),
										 generateHashCode(this.task_id)));
			}
			else {
				this.hashcode =
					generateHashCode(this.tmp_job_id,
									 generateHashCode(
										 generateHashCode(this.tmp_bot_id),
										 generateHashCode(this.task_id)));
			}
		}
		return this.hashcode;
	
	}
	
	@Override
	public synchronized Object clone() {
	
		Task t = null;
		
		try {
			t = (Task)super.clone();

			try {
				t.syncWith(this);
			}
			catch (GridException e) {
				throw new CloneNotSupportedException(e.getMessage());
			}
		}
		catch (CloneNotSupportedException e) {
			throw new InternalError("clone(): cannot clone Task");
		}
		
		return t;
		
	}

	// ****************************************************************

	public String consumerID() {
		
		return this.consumer_peer_id;
		
	}

	public int jobID() {
		
		return this.job_id;
		
	}
	
	public Integer jobIDAsKey() {
		
		return this.job_id_as_key;
		
	}

	public int botID() {
		
		return this.bot_id;
		
	}

	public int id() {
		
		return this.task_id;
		
	}

	public Integer taskIDAsKey() {
		
		return this.task_id_as_key;
		
	}

	public byte[][] jars() {
		
		return this.jars;
		
	}
	
	public String[] names() {
		
		return this.classnames;
		
	}
	
	public Object[] inputParameters() {
		
		return this.parameters;
		
	}
	
	public String mainClass() {
	
		return this.main_class;
		
	}
	
	public long nominalRuntime() {
		
		return this.nominal_runtime;
		
	}
	
	// ****************************************************************

	public synchronized void setSupplyingSpaceJobID(int tmp_job_id)
		throws GridException {
	
		if (tmp_job_id < 0) {
			throw new GridException("Task(): illegal Job id");
		}

		this.tmp_job_id = tmp_job_id;
		
	}

	public synchronized int getSupplyingSpaceJobID() {
	
		return this.tmp_job_id;
		
	}

	public synchronized void setSupplyingSpaceBoTID(int tmp_bot_id)
		throws GridException {
	
		if (tmp_bot_id < 0) {
			throw new GridException("Task(): illegal Job id");
		}

		this.tmp_bot_id = tmp_bot_id;
		
	}

	public synchronized int getSupplyingSpaceBoTID() {
	
		return this.tmp_bot_id;
		
	}

	// ********************DATA-CENTRIC OPERATIONS*********************

	public GridData[][] getRawInputDataSet() throws GridException {

		if (this.readied_datas != null) {
			throw new GridException("getRawInputDataSet(): "+
									"input data set for Task "+toString()+
									"was readied for execution, "+
									"use getReadiedDatas() instead");
		}

		return this.datas;

	}

	public MetaGridData[] getInputMetaGridDataSet() throws GridException {

		return this.input_meta_grid_data_set;

	}

	public String getInputDataSetID() {
		
		return this.input_data_set_id;
		
	}
	
	public long getInputDataSetSize() {
	
		return this.input_data_set_size;
		
	}

	// ****************************************************************

	/** Readies input data set for execution.
	 * This :<ul>
	 * <li>downloads all data needed for execution and makes them available
	 * in an unidemnsionnal array, making it easier to get the GridData that
	 * were downloaded (without the proper ones, it's not possible to
	 * get the InputStreams)</li>
	 * <li>seeds the data, even if it was first downloaded by FTP</li>
	 * <li>erases other now redundant download options</li></ul>
	 * @throws GridException If all data's cannot be downloaded
	 * @since 2006-12-08
	 */
	public GridData[] readyInputDataSet(GridDataManager gdm)
		throws GridException {

		this.readied_datas = gdm.extract(this.datas);
		gdm.waitFor(this.readied_datas);
		gdm.seed(this.datas);
		this.datas = null;
		
		return this.readied_datas;

	}

	/** Returns readied data's.
	 * @see #readyInputDataSet(lbg.protocol.data.GridDataManager)
	 * @since 2006-12-08
	 */
	public GridData[] getInputDataSet() throws GridException {

		if (this.readied_datas == null) {
			throw new GridException("getInputDataSet(): data not yet readied");
		}

		return this.readied_datas;

	}

	// ****************************************************************

	public synchronized void setStartTime() throws GridException {
		
		this.start_time = LightweightBarteringGrid.getTime();
		
	}
	
	/**
	 * @deprecated
	 */
	@Deprecated
	synchronized long getStartTime() throws GridException {
		
		if (this.start_time < 0) {
			throw new GridException("getStartTime(): corrupted start time");
		}
		
		return this.start_time;
		
	}

	// ****************************************************************

	public synchronized void
		flagAsCompleted(String supplier_peer_id,long start_time,
						GridData data,byte[] logs,int power)
		throws GridException {
		
		if (this.is_completed == true) {
			throw new GridException("flagAsCompleted(): "+
									"Task marked as completed");
		}

		this.is_completed = true;
		this.task_output =
			new TaskOutput(this,supplier_peer_id,start_time,data,logs,power);
		
	}

	public synchronized boolean isCompleted() {
		
		return this.is_completed;
		
	}
	
	public synchronized TaskOutput output() throws GridException {
		
		if (this.is_completed == false) {
			throw new GridException("output(): Task not completed");
		}
		if (this.task_output == null) {
			throw new GridException("output(): unavailable Task output");
		}
		
		return this.task_output;
		
	}
	
	// ****************************************************************

	public synchronized void flagAsCancelled(String supplier_peer_id)
		throws GridException {
		
		if (this.is_completed == true) {
			throw new GridException("flagAsCancelled(): "+
									"Task marked as completed");
		}
		if (supplier_peer_id == null) {
			throw new GridException("flagAsCancelled(): "+
									"illegal supplier Peer id");
		}

		++this.cancellation_count;
		this.runtime_before_last_cancellation =
			LightweightBarteringGrid.getTime()-this.start_time;
		this.supplier_before_last_cancellation = supplier_peer_id;
		
	}
	
	public synchronized int getCancellationCount() {
		
		return this.cancellation_count;
		
	}
	
	public synchronized void flagAsPreempted() {
	
		++this.preemption_count;
		
	}
	
	public synchronized int getPreemptionCount() {
		
		return this.preemption_count;
		
	}

	// ****************************************************************

	/**
	 * @deprecated
	 */
	@Deprecated
	synchronized boolean hasBeenStarted() {
	
		return this.start_time >= 0;
		
	}
	
	/**
	 * @deprecated
	 */
	@Deprecated
	synchronized long getRuntimeBeforeLastCancellation()
		throws GridException {
		
		if (this.cancellation_count <= 0) {
			throw new GridException("runtimeBeforeLastCancellation(): "+
									"Task has not been cancelled");
		}
		
		return this.runtime_before_last_cancellation;
		
	}

	/**
	 * @deprecated
	 */
	@Deprecated
	synchronized String getSupplierBeforeLastCancellation()
		throws GridException {
		
		if (this.cancellation_count <= 0) {
			throw new GridException("supplierBeforeLastCancellation(): "+
									"Task has not been cancelled");
		}
		
		return this.supplier_before_last_cancellation;
		
	}

	// ****************************************************************

	public synchronized void syncWith(Task returning_self)
		throws GridException {
		
		if (returning_self == null) {
			throw new GridException("syncWith(): illegal Task");
		}
		
		this.start_time =
			returning_self.hasBeenStarted() ? returning_self.getStartTime():-1;
		this.is_completed = returning_self.isCompleted();
		this.task_output = (this.is_completed) ?
			returning_self.output() : null;
	
		this.preemption_count = 0;
		this.cancellation_count = returning_self.getCancellationCount();
		if (this.cancellation_count > 0) {
			this.runtime_before_last_cancellation =
				returning_self.getRuntimeBeforeLastCancellation();
			this.supplier_before_last_cancellation =
				returning_self.getSupplierBeforeLastCancellation();
		}
		
	}
	
	// ****************************************************************

	/** Strips the Task from now-unuseful payload,
	 * such as metada, output, ...
	 * <p>This is a transitional method: will become unneeded when .jar files
	 * and output data files are integrated into the data transfer architecture.
	 */
	public synchronized void deleteMeatada() throws GridException {
	
		if ((this.classnames == null) ||
			(this.jars == null) || (this.parameters == null) ||
			(this.dependencies == null) ||
			(this.input_meta_grid_data_set == null)) {
			throw new GridException("diet(): metadata already deleted");
		}
		
		this.datas = null;
		this.classnames = null;
		this.dependencies = null;
		this.input_meta_grid_data_set = null;
		this.jars = null;
		this.parameters = null;
		this.task_output.clear();
		
	}

	// ****************************************************************

	static final long serialVersionUID = 1574164816422994645L;
	
	// ****************************************************************

	private final String consumer_peer_id;

	private final int job_id;     // Job id in consumption space
	private final Integer job_id_as_key;
	private       int tmp_job_id; // Job id in supplying space

	private final int bot_id;     // BoT id in consumption space
	private       int tmp_bot_id; // BoT id in supplying space

	private final int task_id;
	private final Integer task_id_as_key;

	private       byte[][] jars;
	private final String main_class;
	private       String[] classnames;
	private       Object[] parameters;
	private       GridData[][] datas;
	private       long nominal_runtime;

	private       GridData[] readied_datas;
	private       String[] dependencies;
	private       MetaGridData[] input_meta_grid_data_set;
	private final String input_data_set_id;
	private       long input_data_set_size;

	private       long start_time;
	private       boolean is_completed;
	private       TaskOutput task_output;

	private       int    preemption_count;
	private       int    cancellation_count;
	private       long   runtime_before_last_cancellation;
	private       String supplier_before_last_cancellation;
	
	private       int hashcode;

	// ****************************************************************

}

