/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// TaskOutput.java

package lbg.task;

import lbg.*;
import lbg.common.exception.*;
import lbg.protocol.data.*;
import java.io.Serializable;

/** A TaskOutput is an immutable Object that contains
 * both data and metadata resulting from the successful execution
 * of the target Task by a Resource of the target Supplier Peer.
 * @author Cyril Briquet, XavierDalem
 */
public class TaskOutput implements Serializable {

	/** Constructs a new TaskOutput.
	 * @param t target Task
	 * @param supplier_peer_id Supplier Peer id
	 * @param start_time Task start time
	 * @param data Task execution results
	 * @param power power of the Resource that executed the Task
	 */
	public TaskOutput(Task t,String supplier_peer_id,long start_time,
					  GridData data,byte[] logs,int power)
		throws GridException {
		
		if (t == null) {
			throw new GridException("TaskOutput(): illegal Task");
		}
		if (supplier_peer_id == null) {
			throw new GridException("TaskOutput(): illegal supplier Peer id");
		}
		if (start_time > LightweightBarteringGrid.getTime()) {
			throw new GridException("TaskOutput(): illegal Task start time");
		}
		if (data == null) {
			throw new GridException("TaskOutput(): invalid data");
		}
		if (logs == null) {
			throw new GridException("TaskOutput(): illegal logs array");
		}
		if (power < 1) {
			throw new GridException("TaskOutput(): illegal Power");
		}

		this.t = t;
		this.supplier_peer_id = supplier_peer_id;
		this.is_local = this.t.consumerID().equals(this.supplier_peer_id);
		this.start_time = start_time;
		this.end_time = LightweightBarteringGrid.getTime();
		this.runtime = this.end_time-this.start_time;
		this.data = data;
		this.logs = logs;
		this.resource_power = power;

	}

	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		sb.append("[TaskOutput]");
		if (this.is_local) {
			sb.append(" local"); 
		}
		else {
			sb.append(" supplier=");
			sb.append(this.supplier_peer_id);
		}
		sb.append(", power=");
		sb.append(this.resource_power);
		sb.append(", start time=");
		sb.append(this.start_time);
		sb.append(", end time=");
		sb.append(this.end_time);
		sb.append(", runtime=");
		sb.append(this.runtime);

		return sb.toString();

	}
	
	// ****************************************************************

	/** Peer that executed the Task.
	 * @return Supplier Peer id
	 */
	public String supplierID() {
		
		return this.supplier_peer_id;
		
	}
	
	/** Indicates whether the Task has been executed by its owner Peer,
	 * or by a Supplier Peer.
	 * @return true if the Task has been executed by its owner Peer
	 */
	public boolean isLocalTask() {
		
		return this.is_local;
		
	}

	/** Indicates whether the Task has been executed by its owner Peer,
	 * or by a Supplier Peer.
	 * @return true if the Task has been executed by a Supplier Peer
	 */
	public boolean isSupplyingTask() {
		
		return !this.is_local;
		
	}
	
	/** Time at which Task execution was effectively started.
	 * @return Task execution start time
	 */
	public long taskStartTime() {
		
		return this.start_time;
		
	}

	/** Time at which Task execution was effectively completed.
	 * @return Task execution end time
	 */
	public long taskEndTime() {
		
		return this.end_time;
		
	}

	/** Effective Task runtime (i.e. excluding queueing),
	 * as set by the Supplier.
	 * <p>This information should be used in SIM mode only,
	 * as it may be unreliable in LIVE mode.
	 * @return effective Task runtime
	 */
	 // Used only in: NeighborhoodBarteringProfile, 3 OG* BarteringAccountant
	public long runtime() {
		
		return this.runtime;
		
	}
	
	/** Power of the Resource that run the Task.
	 * <p>This information should be used in SIM mode only,
	 * as it is meaningless (i.e. arbitrary value, really) in LIVE mode.
	 * @return power of the Resource that run the Task.
	 */
	public int resourcePower() {
		
		return this.resource_power;
		
	}
	
	/** Output data.
	 */
	public GridData getOutputData() throws GridException {
		
		if (this.data == null) {
			throw new GridException("getOutputData(): output already cleared");
		}
		
		return this.data;
		
	}
	
	/** Log.
	 */
	public byte[] getLog() throws GridException {
		
		if (this.logs == null) {
			throw new GridException("getLog(): output already cleared");
		}
		
		return this.logs;
		
	}

	// ****************************************************************

	void clear() throws GridException {
	
		if ((this.data == null) || (this.logs == null)) {
			throw new GridException("clear(): output already cleared");
		}
		
		this.data = null;
		this.logs = null;
		
	}

	// ****************************************************************

	static final long serialVersionUID = -5860126438916101416L;

	// ****************************************************************

	private final Task t;
	private final String supplier_peer_id;
	private final boolean is_local;
	private final long start_time;
	private final long end_time;
	private final long runtime;
	private       GridData data;
	private       byte[] logs;
	private final int resource_power;

}

