/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// EmbeddedLiveUserAgent.java

package lbg.user;

import lbg.*;
import lbg.common.*;
import lbg.common.exception.*;
import lbg.common.io.*;
import lbg.common.net.*;
import lbg.middleware.util.*;
import lbg.peer.service.*;
import lbg.peer.workflow.*;
import lbg.protocol.rpc.*;

import java.util.Properties;

/** An EmbeddedLiveUserAgent is a LiveUserAgent intended to be run
 * within another software.
 * <p>The User service is immediately created in a separate thread.
 * @author Cyril Briquet
 */
public class EmbeddedLiveUserAgent extends LiveUserAgent {

	/** Creates an EmbeddedLiveUserAgent.
	 * @param id User id
	 * @param user_hostname User hostname
	 * @param user_port User port
	 * @param peer_handle used Peer
	 * @param output_dir path to the directory where log files will be saved
	 * @throws GridException if a parameter is illegal,
	 * or if the used Peer cannot be contacted
	 */
	public EmbeddedLiveUserAgent(String id,String user_hostname,int user_port,
								 UserPeerHandle peer_handle,String output_dir)
		throws GridException {

		super(id,user_hostname,user_port,peer_handle,output_dir);
		
	}
	
	public synchronized String toString() {
	
		return CommonStringHandling.
			replaceTag(super.toString(),
					   "LiveUserAgent","EmbeddedLiveUserAgent");
		
	}
	
	// ****************************************************************

	/** Submits a Job.
	 * @param jdf_props descriptor of the Job to submit, stored as Properties
	 * @return submitted job id
	 * @throws GridException if a Job with the same id has already been sent
	 */
	public synchronized int submitJob(Properties jdf_props)
		throws GridException {

		String current_dir = System.getProperty("user.dir");
		PropertiesJobLoader pjl = new PropertiesJobLoader();
		JobDescriptor jd = pjl.loadJobDescription(this,jdf_props,current_dir);
		
		return super.submitJob(jd);
		
	}
	
	// ****************************************************************
	
	/** Creates a new EmbeddedLiveUserAgent (convenience method).
	 * <p>The User service is immediately created in a separate thread.
	 * @param user_hostname User hostname
	 * @param user_port User port
	 * @param peer_hostname Peer hostname
	 * @param peer_port User port
	 * @param user_logs_dir path to the directory where log files will be saved
	 * @return EmbeddedLiveUserAgent, or null if one cannot be created
	 */
	public static EmbeddedLiveUserAgent
		startUser(String user_hostname,int user_port,
				  String peer_hostname,int peer_port,String user_logs_dir) {
		
		EmbeddedLiveUserAgent user = null;

		try {
			// Set LBG mode (exactly once !)
			LightweightBarteringGrid.setMode(lbg.VirtualizationLevel.LIVE_MODE);

			// Contact the LBG Peer and obtain a Peer handle
			UserPeerHandle peer_handle =
				HandleDownloader.downloadPeerHandle(peer_hostname,peer_port);

			// Create LBG User (will be immediately up and running)
			user = new EmbeddedLiveUserAgent("EmbeddedUserAgent",
										  	 user_hostname,user_port,
											 peer_handle,user_logs_dir);
			user.startService(true);
		}
		catch (GridException e) {
			e.printStackTrace();
			user = null;
		}

		return user;

	}
	
	/** Makes target EmbeddedLiveUserAgent submit a Job from target JDF
	 * (convenience method).
	 * @param user target User
	 * @param jdf_props target JDF
	 * @return Job id if Job was successfully submitted, -1 otherwise
	 */
	public static int 
		submitJobToTheGrid(EmbeddedLiveUserAgent user,Properties jdf_props) {
	
		int submitted_job_id = -1;
		try {
			submitted_job_id = user.submitJob(jdf_props);
		
		}
		catch (GridException e) {
			for (int i = 0 ; i < 64 ; i++) {
				System.out.print("*");
			}
			System.out.println();
			System.out.println("A GridException *has been caught*.");
			System.out.println("This message is intended "+
							   "for informational purposes only.");
			e.printStackTrace();
			for (int i = 0 ; i < 64 ; i++) {
				System.out.print("*");
			}

			return -1;
		}
		
		return submitted_job_id;
	
	}
	
	/** Gracefully shutdown target EmbeddedLiveUserAgent (convenience method).
	 * <p>Basically, waits for all Jobs to complete cleanly,
	 * i.e. that all results are successfully uploaded.
	 * @param user target User
	 */
	public static void shutdownUser(EmbeddedLiveUserAgent user) {
		
		while (user.hasRunningJobs()) {
			// Sleep 1 second
			try {
				ThreadPool.sleep(1);
			}
			catch (GridException e) {}
		}
		
	}
	
	/** Describes, as a Properties and with the target values,
	 * a Job to be sumbitted by an EmbeddedLiveUserAgent (convenience method).
	 * @param task_count Task count
	 * @param embedded_task_jar path to application jar of embedded Task
	 * @param embedded_task_main main class of application jar
	 * @param ctrlr_hostname hostname of server to be contacted by embedded Task
	 * @param ctrlr_port port of server to be contacted by embedded Task
     * @return embdded BoT description, or null if one or more parameters
	 * describing the Job is not legal
	 */
	public static Properties
		describeEmbeddedJob(int task_count,
							String embedded_task_jar,String embedded_task_main,
							int embedded_task_port,
							String ctrlr_hostname,int ctrlr_port) {
		
		if ((task_count < 1) ||
			(CommonFileHandling.isJarFileName(embedded_task_jar) == false) ||
			(embedded_task_main == null) ||
			(embedded_task_main.length() < 1) ||
			(CommonNetHandling.isValidPort(embedded_task_port) == false) ||
			(ctrlr_hostname == null) ||
			(ctrlr_hostname.length() < 1) ||
			(CommonNetHandling.isValidPort(ctrlr_port) == false)) {
			return null;
		}

		/* Expected Properties:
		   task_list = embedded_task0, embedded_task1, ...
		   ...
		   embedded_taski.jar = eval(embedded_task_jar)
		   embedded_taski.main_class = eval(embedded_task_main)
		   embedded_taski.parameters = eval(ctrlr_hostname), eval(ctrlr_port)
		   ...
		*/

		Properties props = new Properties();
		StringBuilder task_list = new StringBuilder();
		
		String base_name = "embedded_task";
		String parameters = ctrlr_hostname+", "+ctrlr_port+", "+
							embedded_task_port;
		
		for (int i = 0 ; i < task_count ; i++) {
			// Set properties of Task i
			String task_name = base_name+i;
			
			props.setProperty(task_name+".jar",embedded_task_jar);
			props.setProperty(task_name+".main_class",embedded_task_main);
			props.setProperty(task_name+".parameters",parameters);
			
			// Add Task i to Task list
			task_list.append(task_name);
			if (i < task_count-1) task_list.append(", ");
		}
		
		// Set Task list property
		props.setProperty("task_list",task_list.toString());
		
		return props;
		
	}

	// ****************************************************************
	
	// For testing purposes
	public static void main(String[] args) throws Exception {
		demo();
	}

	// ****************************************************************
	
	private static void demo() {
		
		String peer_hostname     = "remotehost";
		int    peer_port         = 4242;
		String user_hostname     = "localhost";
		int    user_port         = 2424;
		String user_logs_dir     = "/tmp";
		
		String ctrlr_hostname    = user_hostname;
		int    ctrlr_port        = 2600;
		
		String embedded_task_jar  = "/tmp/leet.jar";
		String embedded_task_main = "com.leet.CodeOfTheNode";
		int    embedded_task_port = 1234;

		// Create User
		EmbeddedLiveUserAgent user =
			EmbeddedLiveUserAgent.startUser(user_hostname,user_port,
					  					    peer_hostname,peer_port,
										    user_logs_dir);
		
		/* BEGIN SUBMISSION LOOP */
		// Create JDF
		int task_count = 24;
		Properties jdf_props =
			EmbeddedLiveUserAgent.
				describeEmbeddedJob(task_count,
									embedded_task_jar,embedded_task_main,
									embedded_task_port,
									ctrlr_hostname,ctrlr_port);
		try {
			jdf_props.store(System.out,"this is a test"); // simple test
		}
		catch (java.io.IOException e) { }

		// Submit Job
		int submitted_job_id =
			EmbeddedLiveUserAgent.submitJobToTheGrid(user,jdf_props);
		
		// Check if Job was correctly submitted
		if (submitted_job_id >= 0) {
			System.out.println("Submitted Job #"+submitted_job_id);
		}
		else /* if (job_id < 0) */{
			System.out.println("Could not submit a new Job");
		}
		/* END SUBMISSION LOOP */
		
		// Gracefully shutdown User
		System.out.println("Wait for all submitted Jobs to complete cleanly");
		EmbeddedLiveUserAgent.shutdownUser(user);
		
	}
	
	// ****************************************************************

}

