/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LiveUserAgent.java

package lbg.user;

import lbg.*;
import static lbg.common.Tank.newInteger;
import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonFileHandling.*;
import lbg.common.math.*;
import lbg.common.net.*;
import lbg.peer.service.*;
import lbg.peer.workflow.*;
import lbg.protocol.rpc.*;
import lbg.middleware.execution.*;
import lbg.task.*;
import lbg.user.service.*;

import java.io.File;
import java.util.Map;
import java.util.TreeMap;
import java.util.Vector;

/** UserAgent, Live Implementation.<br>
 * Related logger: lbg.user
 * @author Xavier Dalem, Cyril Briquet
 */
public class LiveUserAgent extends UserAgent {
	
	// Keep it instead of always re-asking it
	private static java.util.logging.Logger UA_LOGGER =
		java.util.logging.Logger.getLogger("lbg.user");
	
	private final Measurement measurement;
	private final String host;
	private final int port;
	private final String output_dir;
	private final UserService user_service;
	private final Map<Integer,JobDescriptor> submitted_jobs;
	private final boolean save_results;
	private final GridClassLoader job_checker;
	
	private boolean i_am_processing_uploaded_results;
	private long completed_jobs;
	private long completed_tasks;
	/** Holds references to all the listeners */
	private final Vector<LiveUserAgentListener> listeners;

	private Thread svc_thread;
	private boolean online;

	// ****************************************************************

	/** Creates a Live UserAgent.
	 * <p>Measurements are activated by default.
	 * <p>Output data files are saved by default.
	 * @param id User ID
	 * @param host User host (to contact him from the outside)
	 * @param port User port (local)
	 * @param peer Parent peer
	 * @param output_dir The path of a directory where files will be
	 * output. Filenames will be of the form
	 * <code><i>[user_id]</i>.<i>[peer_id]</i>.job<i>[jobID]</i>.bot<i>[botID]</i>.task<i>[taskID]</i>.dat</code>
	 * @param stats_dir The path of a directory where job runtime stats will be
	 * saved
	 * @throws GridException if the parameters are erroneous
	 */
	public LiveUserAgent(String id, String host, int port, UserPeerHandle peer,
						 String output_dir, String stats_dir)
		throws GridException {

		this(id,host,port,peer,output_dir,stats_dir,true,true);
	}
	
	/** Creates a Live UserAgent.
	 * <p>Measurements are deactivated by default.
	 * <p>Output data files are not saved by default.
	 * @param id User ID
	 * @param host User host (to contact him from the outside)
	 * @param port User port (local)
	 * @param peer Parent peer
	 * @param output_dir The path of a directory where files will be
	 * output. Filenames will be of the form
	 * <code><i>[user_id]</i>.<i>[peer_id]</i>.job<i>[jobID]</i>.bot<i>[botID]</i>.task<i>[taskID]</i>.dat</code>
	 * @throws GridException if the parameters are erroneous
	 */
	public LiveUserAgent(String id, String host, int port, UserPeerHandle peer,
						 String output_dir)
		throws GridException {

		this(id,host,port,peer,output_dir,null,false,false);
		
	}

	/** Creates a Live UserAgent.
	 * @param id User ID
	 * @param host User host (to contact him from the outside)
	 * @param port User port (local)
	 * @param peer Parent peer
	 * @param output_dir The path of a directory where files will be
	 * output. Filenames will be of the form
	 * <code><i>[user_id]</i>.<i>[peer_id]</i>.job<i>[jobID]</i>.bot<i>[botID]</i>.task<i>[taskID]</i>.dat</code>
	 * @param stats_dir The path of a directory where job runtime stats will be
	 * saved
	 * @param activate_measurement true if Measurements should be activated
	 * @param save_results true to save results to disk
	 * @throws GridException if the parameters are erroneous
	 */
	public LiveUserAgent(String id, String host, int port, UserPeerHandle peer,
						 String output_dir, String stats_dir,
						 boolean activate_measurement,boolean save_results)
		throws GridException {
		
		super(id,peer);
		
		this.i_am_processing_uploaded_results = false;
		
		if (host == null || (CommonNetHandling.isValidPort(port) == false) ||
			(output_dir == null) ||
			((stats_dir == null) && (activate_measurement == true))) {
			throw new GridException("LiveUserAgent(): invalid parameters");
		}
		
		this.host = host;
		this.port = port;
		
		if (! (new File(output_dir).isDirectory())) {
			throw new GridException("LiveUserAgent(): given output directory "
				+ "is not a directory or doesn't exist");
		}
		this.output_dir = output_dir;
		this.save_results = save_results;

		this.user_service = UserServiceFactory.createUserService(this);
		
		this.submitted_jobs	= new TreeMap<Integer,JobDescriptor>();
		this.completed_jobs	=  0;
		this.completed_tasks = 0;
		
		this.listeners = new Vector<LiveUserAgentListener>();
		
		// Shutdown cleaner
		if (activate_measurement) {
			this.measurement =
				MeasurementFactory.
					createDescriptiveMeasurement(id + ".job_runtime");
			Runtime.getRuntime().addShutdownHook(new Thread(
				new MeasurementSaver(measurement,stats_dir)));
		}
		else {
			this.measurement = null;
		}

		this.job_checker = new GridClassLoader();

		this.svc_thread = null;
		this.online = false;
		
	}

	/** String representation of the User.
	 */
	public synchronized String toString() {
		
		String super_string = CommonStringHandling.
			replaceTag(super.toString(),"UserAgent","LiveUserAgent");
		StringBuilder sb = new StringBuilder(super_string);

		sb.append(", ");
		sb.append(this.submitted_jobs.size());
		sb.append(" Jobs submitted, ");
		sb.append(this.completed_jobs);
		sb.append(" completed Jobs received, ");
		sb.append(this.completed_tasks);
		sb.append(" completed Tasks received");
		
		return sb.toString();
	}
	
	// ****************************************************************

	public synchronized UserService service() {
		
		return this.user_service;
		
	}
	
	/** Get user host */
	public String host() {
		return host;
	}
	
	/** Get user port */
	public int port() {
		return port;
	}
	
	/** Register this object as a listener.
	 * @param listen The object that wants to be aware of our actions.
	 */
	public synchronized void registerListener(LiveUserAgentListener listen) {
		this.listeners.add(listen);
	}
	
	/** Unregister this object as a listener.
	 * @param listen The object that doesn't want to be aware of our actions
	 * anymore.
	 */
	public synchronized void unRegisterListener(LiveUserAgentListener listen) {
		this.listeners.remove(listen);
	}
	
	/** Notifies a User that a Task has been completed.
	 * @param t completed Task
	 * @throws GridException if User failed to save output data file
	 */
	public synchronized void notifyCompletedTask(Task t) throws GridException {

		if (t == null) {
			throw new GridException("notifyCompletedTask: null task");
		}
		
		this.i_am_processing_uploaded_results = true;
		
		// Keep track of completed Tasks
		this.completed_tasks++;
		
		updateStatistics(t, false);

		this.i_am_processing_uploaded_results = false;
	}

	/** Notifies a User that the last Task of a Job has been completed.
	 * @param t completed Task
	 * @throws GridException if User failed to save output data file
	 */
	public synchronized void notifyCompletedJob(Task t) throws GridException {
		
		if (t == null) {
			throw new GridException("notifyCompletedJob: null task");
		}
		
		this.i_am_processing_uploaded_results = true;

		// Keep track of completed Tasks and Jobs
		this.completed_tasks++;
		this.completed_jobs++;
		
		updateStatistics(t, true);

		this.i_am_processing_uploaded_results = false;
	}

	/** Saves result, outputs info, computes runtime */
	private void updateStatistics(Task t, boolean is_job) throws GridException {

		JobDescriptor job = submitted_jobs.get(newInteger(t.jobID()));
		
		StringBuilder sb = new StringBuilder();
		
		if (job == null) {
			sb.append("Got a Task from a former session. "+
					  "No statistics available.");
		}
		else {
			long runtime = LightweightBarteringGrid.getTime() -
						   job.getSubmissionTime();
		
			if (is_job) {
				sb.append("Job ");
				if (this.measurement != null) {
					this.measurement.addObservation(runtime);
				}
				for (LiveUserAgentListener l : listeners) {
					l.jobCompletion(job, t.jobID());
				}
			}
			else {
				sb.append("Task ").append(t.id()).append(" from job ");
				for (LiveUserAgentListener l : listeners) {
					l.taskCompletion(t);
				}
			}
			sb.append(t.jobID()).append(" completed in ").append(runtime)
				.append(" seconds.");
		}
		
		File[] res = saveResult(t);
		if (res.length != 2) {
			throw new GridException("updateStatistics(): "+
									"illegal resulsts array");
		}
		sb.append(" Data ").append((this.save_results) ? "":"*not* ").
			append("saved to ").append(res[0].getAbsolutePath());
		if (res[1] != null) {
			sb.append(", logs to ").append(res[1].getAbsolutePath());
		}
		else {
			sb.append(", there were no logs");
		}
		
		UA_LOGGER.info(sb.toString());
	}

	/** Saves task result and logs to files in output_dir.
	 * This method has a wget-like behaviour: if the desired file already
	 * exists, the new file will have the following format:
	 * <tt>base_filename.i</tt>, where <i>i</i> is a number starting 0
	 * @param t The task to save
	 * @return An array of files where info was saved : { data, logs }. The
	 * second element is <tt>null</tt> if there were no logs.
	 */
	private synchronized File[] saveResult(Task t) throws GridException {

		if (t == null) {
			throw new GridException("saveResult(): null task");
		}
		
		// Build output data filename
		File data = getFirstMatchingFile(
			new File(output_dir+FILE_SEP+id()+'.'+t.output().getOutputData().id+
					 OUTPUT_DATA_EXT));
		
		// Actually save the output data file
		if (this.save_results) {
			t.output().getOutputData().toFile(data);
		}
		
		// Actually save the log file
		File logs = null;
		if (t.output().getLog().length != 0) {
			logs = getFirstMatchingFile(
				new File(output_dir+ FILE_SEP+
						 LOG_PREFIX+id()+'.'+t.output().getOutputData().id+
						 LOG_EXT));
		
			CommonFileHandling.writeFile(logs,new String(t.output().getLog()));
		}
		
		return new File[] { data, logs };
	}

	/** Submits a Job.
	 * @param jd descriptor of the Job to submit
	 * @return submitted job id
	 * @throws GridException if a Job with the same id has already been sent
	 */
	public synchronized int submitJob(JobDescriptor jd) throws GridException {

		if (jd == null) {
			throw new GridException("submitJob: null job");
		}

		job_checker.load(jd);

		UA_LOGGER.finer("Task loading checks completed successfully");

		jd.setSubmissionTime();
		int job_id = usedPeer().submitJob(this.user_service.createHandle(),jd);

		// Keep track of the job
		if (this.submitted_jobs.put(newInteger(job_id),jd) != null) {
			throw new GridException("submitJob(): Job id conflict");
		}
		
		return job_id;
	}

	public synchronized long submittedJobs() {
		
		return this.submitted_jobs.size();
		
	}

	public synchronized long completedJobs() {
		
		return this.completed_jobs;
		
	}
	
	public synchronized boolean hasRunningJobs() {
		
		if (this.i_am_processing_uploaded_results == true) {
			// as updateStatistics() can take some time,
			// a boolean guard must be added to avoid race conditions...
			return true;
		}
		
		return (this.submitted_jobs.size()-this.completed_jobs > 0);
		
	}
	
	public synchronized boolean isAlive() {
		
		return (this.svc_thread != null) ? this.svc_thread.isAlive() : false;
		
	}

	public synchronized void startService() throws GridException {
		
		startService(false);
	
	}

	public synchronized void startService(boolean set_daemon)
		throws GridException {

		if (this.online == true) {
			return;
		}
		else {
			this.online = true;
		}
		
		RemoteCallDaemon rc_daemon =
			new RemoteCallDaemon(RemoteCall.CallType.USER,
								 this.port,this.service(),null);

		this.svc_thread =
			new Thread(rc_daemon,"Peer-to-UserAgent-daemon");
		if (set_daemon) {
			this.svc_thread.setDaemon(true);
		}
		this.svc_thread.start();
		
	}

	// ****************************************************************

}

