/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// SimulatedUserAgent.java

package lbg.user;

import lbg.common.*;
import lbg.common.exception.*;
import lbg.peer.service.*;
import lbg.peer.workflow.*;
import lbg.simulator.*;
import lbg.task.*;
import lbg.user.service.*;
import java.util.TreeMap;

/**
 * @author Cyril Briquet
 */
public class SimulatedUserAgent extends UserAgent {

	public SimulatedUserAgent(String id,UserPeerHandle peer,ISimulator sim)
		throws GridException {
	
		super(id,peer);

		if (sim == null) {
			throw new GridException("SimulatedUserAgent(): illegal Simulator");
		}
		
		this.sim = sim;
		
		this.user_service = UserServiceFactory.createUserService(this);
		
		this.submitted_jobs	= new TreeMap<Integer,JobDescriptor>();
		this.completed_jobs	=  0; //new TreeMap<Integer,JobDescriptor>();
		this.completed_tasks = 0;
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder(super.toString());
		
		sb.append(", submitted ");
		sb.append(this.submitted_jobs.size());
		sb.append(" Jobs, received ");
		sb.append(this.completed_jobs);
		sb.append(" completed Jobs ");
		if (this.submitted_jobs.size() != this.completed_jobs) {
			sb.append("(*");
			sb.append(this.submitted_jobs.size()-this.completed_jobs);
			sb.append(" MISSING*) ");
		}
		sb.append("+ ");
		sb.append(this.completed_tasks);
		sb.append(" completed Tasks");
		
		/* debugging code --
		Iterator<JobDescriptor> it = this.submitted_jobs.values().iterator();
		while (it.hasNext()) {
			sb.append("\n  ->  ");
			sb.append(it.next());
		}
		*/
		
		return sb.toString();
		
	}
	
	// ****************************************************************

	public synchronized UserService service() {
		
		return this.user_service;
		
	}

	// ****************************************************************

	public synchronized int submitJob(JobDescriptor jd) throws GridException {
		
		UserPeerHandle peer = super.usedPeer();
		
		// Submit Job
		int job_id = peer.submitJob(user_service.createHandle(),jd);
		
		// Keep track of submitted Job
		if (this.submitted_jobs.put(Tank.newInteger(job_id),jd) !=
			null) {
			throw new GridException("submitJob(): Job id conflict");
		}

		// Add one Job submission Event to the EventList, if required
		this.sim.addJobSubmission(this,peer.id());
		
		return job_id;
		
	}
	
	public synchronized long submittedJobs() {
		
		return this.submitted_jobs.size();
		
	}

	public synchronized long completedJobs() {
		
		return this.completed_jobs;
		
	}

	public synchronized boolean hasRunningJobs() {
		
		return (this.submitted_jobs.size()-this.completed_jobs > 0);
		
	}

	// ****************************************************************

	public synchronized void notifyCompletedTask(Task t) throws GridException {
		
		// Keep track of completed Tasks
		this.completed_tasks++;

	}
	
	public synchronized void notifyCompletedJob(Task t) throws GridException {

		// Keep track of completed Tasks
		this.completed_tasks++;
		
		// Keep track of completed Jobs
		Integer job_id_as_key = t.jobIDAsKey();
		/*JobDescriptor dummy_job =*/ this.submitted_jobs.get(job_id_as_key);
		if (this.submitted_jobs.get(job_id_as_key) == null) {
			throw new GridException("notifyCompletedJob(): unknown Job");
		}
		
		/*
		if (this.completed_jobs.put(job_id_as_key,dummy_job) !=
			null) {
			throw new GridException("notifyCompletedJob(): Job id conflict");
		}*/
		this.completed_jobs++;

	}

	// ****************************************************************

	private ISimulator sim;
	private UserService user_service;
	
	private TreeMap<Integer,JobDescriptor> submitted_jobs;
	//private TreeMap<Integer,JobDescriptor> completed_jobs;
	private long completed_jobs;
	private long completed_tasks;
	
}

