/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// UAJobSubmission.java

package lbg.user.server;

import lbg.common.exception.GridException;
import lbg.user.EmbeddedLiveUserAgent;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.ProtocolException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Properties;

/** A LiveUserAgentServer is a server translating job submission requests
 * received on a socket into calls made to an EmbeddedLiveUserAgent.
 *
 * <p>The requests follow a simple text-based protocol: A request header
 * is first sent. This header holds on one line. Each request can be followed by
 * zero or more parameters. There is one parameter per line.
 * See <code>RequestFactory</code> for a list of supported request headers.
 * 
 * <p>A typical session established with the LiveUserAgentServer (LUAS) is
 * as follows:
 * <ul>
 * <li>1) the LUAS accepts a single connection on its local port
 * (which is then closed).</li>
 * <li>2) the LUAS handles the request send on the accepted socket.</li>
 * <li>3a) The request is a job submission: the LUAS submits the job to
 * the EmbeddedLiveUserAgent and notifies the external software component of
 * the result of the submission.</li>
 * <li>3b) The request is a shutdown request:
 * the LUAS shuts the EmbeddedLiveUserAgent down.
 * When it is done, it notifies the connected external software component
 * and terminates itself.</li>
 * </ul>
 * 
 * @author Gerard Dethier
 */
public class LiveUserAgentServer {
	
	private final EmbeddedLiveUserAgent user;
	private       ServerSocket servSock;
	
	/**
	 * The LiveUserAgentServer contructor. It instantiates an EmbeddedLiveUserAgent which connects
	 * to a given Peer and creates a server socket to wait requests from an external software component.
	 * 
	 * @param serverPort The post on which the LUAS waits a connection from an external software component.
	 * @param userHost The local host.
	 * @param userPort The port to be used by the EmbeddedLiveUserAgent.
	 * @param userDir A directory in which the EmbeddedLiveUserAgent can write data.
	 * @param peerHost The host on which a Peer is running.
	 * @param peerPort The port on which the Peer is listening.
	 * 
	 * @throws Exception If the EmbeddedLiveUserAgent
	 * or the server socket could not be created.
	 */
	private LiveUserAgentServer(int serverPort, String userHost, int userPort,
								String userDir, String peerHost, int peerPort)
		throws Exception {
		
		this.user = EmbeddedLiveUserAgent.
			startUser(userHost, userPort, peerHost, peerPort, userDir);
		if (this.user == null) {
			throw new Exception("Could not create EmbeddedLiveUserAgent.");
		}
		servSock = new ServerSocket(serverPort);
	}

	/**
	 * The request handling loop. This method returns when a shutdown request is received and the
	 * EmbeddedLiveUserAgent shutdown is completed.
	 * 
	 * @throws IOException If a transmission problem occured between the LUAS and the external
	 * software component.
	 */
	private void requestHandlingLoop() throws IOException {
		Socket s = servSock.accept();
		servSock.close();
		servSock = null;
		System.out.println("New Connection...");

		
		// Creation of reader and writer for text protocol
		BufferedReader reader = new BufferedReader(new InputStreamReader(s.getInputStream()));
		PrintStream writer = new PrintStream(new BufferedOutputStream(s.getOutputStream()));
		
		// Request handling loop
		while(true) {
			String header = reader.readLine();

			System.out.println("Handling new request...");
			UAServerRequest req =
				UAServerRequestFactory.getRequestObject(header);
			if(req == null) {
				throw new ProtocolException("Unknwon header: "+header);
			}

			req.readRequest(reader);
			
			try {
				if(handleRequest(req, writer)) {
					break;
				}
			}
			catch (GridException e) { /* The show must go on... */ }
		}
	}

	/**
	 * Request handling method.
	 * 
	 * @param req The request.
	 * @param writer The writer to send informations to external software component.
	 * 
	 * @return True if the request handling must stop (received a shutdown request); false otherwise.
	 * 
	 * @throws IOException If a transmission problem occured between the LUAS and the external
	 * @throws GridException for unsupported request types
	 * software component.
	 */
	private boolean handleRequest(UAServerRequest req,PrintStream writer)
		throws GridException, IOException {
		
		if(req instanceof UAJobSubmission) {
			UAJobSubmission js = (UAJobSubmission) req;
			String jdfFile = js.getPropertiesFile();
			System.out.println("Job described in properties file "+jdfFile+" will be submitted.");
			FileInputStream fis = new FileInputStream(jdfFile);
			Properties prop = new Properties();
			prop.load(fis);
			fis.close();
			System.out.println("Properties file read, submitting job.");
			try {
				int jobId = this.user.submitJob(prop);
				System.out.println("Received jobId="+jobId);
				if(jobId == -1) {
					writer.println("NOK");
					writer.println("Could not submit job.");
					writer.flush();
				} else {
					writer.println("OK");
					writer.println(jobId);
					writer.flush();
				}
			} catch(GridException e) {
				writer.println("NOK");
				writer.println(e.getMessage());
				writer.flush();
			}
			System.out.println("Client notified about job submission.");
			return false;
		} else if(req instanceof UAShutdown) {
			System.out.println("Shutting down user.");
			EmbeddedLiveUserAgent.shutdownUser(this.user);
			writer.println("OK");
			writer.flush();
			System.out.println("User closed.");
			return true;
		} else {
			throw new GridException("Unknown request type "+req.getClass());
		}
	}

	/**
	 * The LiveUserAgentServer has the following usage:
	 * 
	 * LiveUserAgentServer serverPort userHostName userPort userDir peerHostName peerPort
	 * 
	 * serverPort is the local TCP port on which the server will listen.
	 * userHostName is the host name of the machine running this server.
	 * userPort is the TCP port on which the EmbeddedLiveUserAgent will listen.
	 * userDir is the output directory for the EmbeddedLiveUserAgent.
	 * peerHostName is the host name of the machine running a Peer.
	 * peerPort is the TCP port on which the Peer is listening.
	 * 
	 * @param args The parameters of the application.
	 */
	public static void main(String[] args) {
		
		if(args.length != 6) {
			System.out.println("Usage: LiveUserAgentServer port userHostName userPort userDir peerHostName peerPort");
		}
		
		int serverPort = Integer.parseInt(args[0]);
		String userHost = args[1];
		int userPort = Integer.parseInt(args[2]);
		String userDir = args[3];
		String peerHost = args[4];
		int peerPort = Integer.parseInt(args[5]);
		
		// Creation of user directory if it does not exist.
		File dir = new File(userDir);
		if(!dir.exists() && !dir.mkdirs()) {

			System.err.println("Could not create user log directory: "+userDir);
			System.exit(-1);

		} else {

			LiveUserAgentServer server;
			try {
				server = new LiveUserAgentServer(serverPort, userHost, userPort, userDir, peerHost, peerPort);
				server.requestHandlingLoop();
			} catch (Exception e) {
				e.printStackTrace();
			}

		}
	}

}

