/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//LiveUserHandle.java

package lbg.user.service;

import lbg.common.exception.GridException;
import lbg.protocol.rpc.*;
import lbg.task.*;

/** Live Implementation of the User Handle
 * @author Xavier Dalem
 */
public class LiveUserHandle implements UserHandle, java.io.Serializable {

	static final long serialVersionUID = -3566919217324621883L;
	
	private final String id;
	private final String host;
	private final int port;

	/** Create a handle
	 * @param id User id
	 * @param user_host User host
	 * @param user_port User port
	 * @throws GridException if the parameters are erroneous */
	public LiveUserHandle(String id, String user_host, int user_port)
		throws GridException {
		if (id == null || user_host == null || user_port < 1) {
			throw new GridException("LiveUserHandle(): illegal parameters");
		}
		this.id = id;
		host = user_host;
		port = user_port;
	}
	
	/** String representation of the User Handle */
	public String toString() {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LiveUserHandle ");
		sb.append(id);
		sb.append(" - ");
		sb.append(host);
		sb.append(":");
		sb.append(port);
		sb.append("]");
		
		return sb.toString();
		
	}
	
	/** Get user id */
	public String id() {
		return id;
	}

	/** Notify a user of a completed task
	 * @param t The completed task
	 * @throws GridException if call cannot be sent */
	public void notifyCompletedTask(Task t) throws GridException {
		if (t == null)
			throw new GridException("notifyCompletedTask: null task");
		
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.USER,
				"notifyCompletedTask", new Class[]
				{Class.forName("lbg.task.Task")}, new Object[] {t} );
			rc.send(host, port, false);
		}
		catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}

	/** Notify a user of a completed job
	 * @param t Last task from the completed job
	 * @throws GridException if call cannot be sent */
	public void notifyCompletedJob(Task t) throws GridException {
		if (t == null)
			throw new GridException("notifyCompletedJob: null task");
		
		try {
			RemoteCall rc = new RemoteCall(RemoteCall.CallType.USER,
				"notifyCompletedJob", new Class[]
				{Class.forName("lbg.task.Task")}, new Object[] {t} );
			rc.send(host, port, false);
		} catch (ClassNotFoundException e) {
			throw new GridException(
				"Critical: could not find one of the base classes", e);
		}
	}
}

