/*
 *  Copyright (c) 2005-2008, Cyril Briquet, parts Xavier Dalem
 *
 *  This file is part of Lightweight Bartering Grid.
 *  Lightweight Bartering Grid is free software;
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Lightweight Bartering Grid is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Lightweight Bartering Grid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LogAggregatorReportManager.java

package lbg.simulator.tools.aggregator;

import lbg.common.exception.*;
import lbg.common.io.*;
import static lbg.common.io.CommonFileHandling.CSV_EXT;
import lbg.common.math.*;
import static lbg.common.net.CommonNetHandling.*;
import static lbg.peer.config.PeerStrategyConfiguration.getPDPList;
import lbg.simulator.tools.*;

/**
 * @author Cyril Briquet
 */
public class LogAggregatorReportManager {

	/** Constructs a new LogAggregatorReportManager.
	 */
	public LogAggregatorReportManager(String timestamp,
									  String scenario_base_name,
									  int logs_count,String[] logs_csv_summary,
									  GlobalPropertiesStore global_props_store,
									  String encountered_errors,
									  String encountered_exceptions)
		throws GridException {
		
		if (timestamp == null) {
			throw new GridException("LogAggregatorReportManager() "+
									"illegal timestamp");
		}
		if (scenario_base_name == null) {
			throw new GridException("LogAggregatorReportManager() "+
									"illegal scenario base name");
		}
		if (logs_csv_summary == null) {
			throw new GridException("LogAggregatorReportManager() "+
									"illegal log csv summaries");

		}
		if (logs_count != logs_csv_summary.length) {
			throw new GridException("LogAggregatorReportManager() "+
									"illegal log count");
		}
		if (global_props_store == null) {
			throw new GridException("LogAggregatorReportManager() "+
									"illegal global properties store");
		}
		if (encountered_errors == null) {
			throw new GridException("LogAggregatorReportManager() "+
									"illegal encountered errors");
		}
		if (encountered_exceptions == null) {
			throw new GridException("LogAggregatorReportManager() "+
									"illegal encountered exceptions");
		}
		
		this.timestamp = timestamp;
		this.spreadsheet_fn = "log_aggregator_report_"+this.timestamp+CSV_EXT;
		this.scenario_base_name = scenario_base_name;
		this.logs_count = logs_count;
		this.logs_csv_summary = logs_csv_summary.clone();
		this.global_props_store = global_props_store;
		this.encountered_errors = encountered_errors;
		this.encountered_exceptions = encountered_exceptions;
		
	}
	
	public synchronized String toString() {
	
		StringBuilder sb = new StringBuilder();
		
		sb.append("[LogAggregatorReportManager]");
		
		return sb.toString();
		
	}

	// ****************************************************************
	
	public synchronized void saveLogAnalysisSummary() throws GridException {
		
		StringBuilder sb = new StringBuilder();
		
		sb.append(LogAggregator.WELCOME_LA).append("\n\n").
			append("Aggregated ").append(this.logs_count).append(" logs\n").
			append("\tfor scenario ").append(this.scenario_base_name).
			append("\n").
			append("\ton ").append(timestamp).append(" @ ").
			append(getLocalHostName()).append("\n\n\n");

		// Loop over all global Properties
		for (String key : this.global_props_store.keys()) {

			// Exploit data aggregated for current Property
			Measurement mea = this.global_props_store.getMeasurement(key);

			sb.append(key);
			sb.append("\n\tmean=");
			sb.append(CommonMath.div_label_2dd(mea.getSum(),
											   mea.getObservations()));
			sb.append("\n\tmin=");
			sb.append(CommonMath.div_label_2dd(mea.getMin(),1));
			sb.append(" (");
			sb.append(mea.getMinTag());
			sb.append(")");
			sb.append("\n\tmax=");
			sb.append(CommonMath.div_label_2dd(mea.getMax(),1));
			sb.append(" (");
			sb.append(mea.getMaxTag());
			sb.append(")");
			sb.append("\n");

		}
		
		sb.append("\n").append("\n").
			append("Errors while analyzing logs:\n\n").
			append(this.encountered_errors).append("\n").
			append("Exceptions found in logs:\n\n").
			append(this.encountered_exceptions);
		
		String summary_filename =
			"log_aggregator_summary_"+timestamp+".txt";
		
		CommonFileHandling.writeFile(summary_filename,sb.toString());
		
		System.out.println("Wrote analysis summary to "+summary_filename);
		
	}

	public synchronized void saveLogSpreadsheetReport() throws GridException {
		
		StringBuilder sb = new StringBuilder();
		
		// Generate 1st line (column titles)
		sb.append("SCENARIO");
		String[] pdp_list = getPDPList();
		for (String target_pdp : pdp_list) {
			sb.append(", ").append(target_pdp);
		}
		sb.append(", STATUS");
		for (String key : global_props_store.keys()) {
			sb.append(", ").append(key);
		}		
		sb.append("\n");
		
		// Generate data lines (1 per log file)
		for (String log_csv : this.logs_csv_summary) {
			sb.append(log_csv).append("\n");
		}
		
		CommonFileHandling.writeFile(this.spreadsheet_fn,sb.toString());

		System.out.println("Wrote spreadsheet report to "+this.spreadsheet_fn);

	}
	
	public String getSpreadsheetFilename() {
	
		return this.spreadsheet_fn;
		
	}

	// ****************************************************************

	private final String timestamp;
	private final String spreadsheet_fn;
	private final String scenario_base_name;
	private final int logs_count;
	private final String[] logs_csv_summary;
	private final GlobalPropertiesStore global_props_store;
	private final String encountered_errors;
	private final String encountered_exceptions;

}

